// -*- c -*-
// (setq c-basic-offset 4)
// ("clsFilename.h")

#include "clsFilename.h"

#include <fstream>
#include <sstream>
#include <iomanip>

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#ifdef WIN32
#define WCHARCONV(x) (wxString)wxConvLibc.cMB2WC(x)
#else
#define WCHARCONV(x) (wxString)wxConvUTF8.cMB2WC(x)
#endif

#ifdef WIN32
#define DELIM L"\\"
#define DELIMWX wxT("\\")
#else
#define DELIM L"/"
#define DELIMWX wxT("/")
#endif

SplitPath_cls::
SplitPath_cls(std::wstring pathStr){
    msFullPath = pathStr;
    std::wstring::size_type tail = msFullPath.size();

    std::wstring::size_type pos_period = msFullPath.rfind(L".");
    if (pos_period != std::wstring::npos){
	msExt = msFullPath.c_str() + pos_period;
	msFullPathMinusExt = msFullPath;
	msFullPathMinusExt.resize(pos_period);
    }

    std::wstring::size_type pos_delim = msFullPath.rfind(DELIM);
    if (pos_delim != std::wstring::npos){
	msFilename = msFullPath.substr(pos_delim + 1,
				       tail - pos_delim - msExt.size() - 1);
	msDirPlusDrive = msFullPath.substr(0, pos_delim);
    }
}

std::wstring SplitPath_cls::
GetPathStr(int flag){
    switch(flag){
    case _FN_EXT:
	return msExt;
    case _FN_FNAME_EXT:
	if (msExt != L"")
	    return msFilename + msExt;
	else
	    return msFilename;
    case _FN_FNAME:
	return msFilename;
    case _FN_DRIVE_DIR:
	return msDirPlusDrive;
    case _FN_FULLPATH:
	return msDirPlusDrive + DELIM + msFilename + msExt;
    }
    return msFullPath;
}

std::wstring SplitPath_cls::
GetPathDirLast(){
    std::wstring::size_type tail      = msDirPlusDrive.size();
    std::wstring::size_type pos_delim = msDirPlusDrive.rfind(DELIM);
    if (pos_delim != std::wstring::npos)
	return msDirPlusDrive.substr(pos_delim + 1, tail - pos_delim - 1) + DELIM;
    return L"";
}

// --------------

SplitPathWX_cls::
SplitPathWX_cls(wxString pathStr){
    mwsFullPath = pathStr;
    size_t tail = mwsFullPath.Len();
    
    int pos_period = mwsFullPath.Find(wxT('.'), TRUE); //TRUE=fromEnd
    if (pos_period == wxNOT_FOUND){
	mwsExt = wxT("");
	mwsFullPathMinusExt = mwsFullPath;
    }
    else{
	mwsExt = mwsFullPath.c_str() + pos_period;
	mwsFullPathMinusExt = mwsFullPath;
	mwsFullPathMinusExt.Truncate(pos_period);
    }
    
    int pos_delim = mwsFullPath.Find(*DELIMWX, TRUE); //TRUE=fromEnd
    if (pos_delim != wxNOT_FOUND){
	mwsFilename = mwsFullPath.SubString(pos_delim + 1,
					    tail - mwsExt.Len() - 1);
	mwsDirPlusDrive = mwsFullPath.SubString(0, pos_delim - 1);
    }
    else
	mwsFilename = mwsFullPath.SubString(0, tail - mwsExt.Len() - 1);
}

wxString SplitPathWX_cls::
GetPathWXStr(int flag){
    switch(flag){
    case _FN_EXT:
	return mwsExt;
    case _FN_FNAME_EXT:
	if (mwsExt != wxT(""))
	    return mwsFilename + mwsExt;
	else
	    return mwsFilename;
    case _FN_FNAME:
	return mwsFilename;
    case _FN_DRIVE_DIR:
	return mwsDirPlusDrive;
    case _FN_DRIVE_DIR_FNAME:
	return mwsFullPathMinusExt;
    case _FN_FULLPATH:
	return mwsDirPlusDrive + DELIMWX + mwsFilename + mwsExt;
    }
    return mwsFullPath;
}

wxString SplitPathWX_cls::
GetPathDirLast(){
    size_t tail   = mwsDirPlusDrive.Len();
    int pos_delim = mwsDirPlusDrive.Find(*DELIMWX, TRUE); //TRUE=fromEnd
    if (pos_delim != wxNOT_FOUND)
	return mwsDirPlusDrive.SubString(pos_delim, tail) + DELIMWX;
    return wxT("");
}

wxString SplitPathWX_cls::
GetImageTypeStr(){
    if (mwsExt == wxT(".jpeg") || mwsExt == wxT(".jpg") ||
	mwsExt == wxT(".JPEG") || mwsExt == wxT(".JPG"))
	return wxT("jpeg");
    if (mwsExt == wxT(".png") || mwsExt == wxT(".PNG"))
	return wxT("png");
    if (mwsExt == wxT(".gif") || mwsExt == wxT(".GIF"))
	return wxT("gif");
    if (mwsExt == wxT(".tga") || mwsExt == wxT(".TGA"))
	return wxT("tga");
    if (mwsExt == wxT(".tiff") || mwsExt == wxT(".tif") ||
	mwsExt == wxT(".TIFF") || mwsExt == wxT(".TIF"))
	return wxT("tiff");
    if (mwsExt == wxT(".hdr") || mwsExt == wxT(".HDR"))
	return wxT("hdr");
    if (mwsExt == wxT(".exr") || mwsExt == wxT(".EXR"))
	return wxT("exr");
    return wxT("unknown_format");
}

// --------------

Filename_cls::
Filename_cls(){
    mSPath = L"";
}

Filename_cls::
Filename_cls(const wchar_t *path){
    mSPath = path;
}

Filename_cls::
Filename_cls(std::wstring pathStr){
    mSPath = pathStr;
}

std::wstring Filename_cls::
GetPathStr(){
    return mSPath;
}

const wchar_t* Filename_cls::
c_str(){
    return mSPath.c_str();
}

void Filename_cls::
Set(std::wstring strPath){
    mSPath = strPath;
}

void Filename_cls::
DoubleSlash(std::wstring &str){
    std::wstring::size_type posBackSlash = 0;
    while ((posBackSlash = str.find(L"\\", posBackSlash)) != std::wstring::npos){
	str.replace(posBackSlash, 1, L"\\\\");
	posBackSlash+=2;
    }
}

void Filename_cls::
DoubleSlash(wxString &str){
    str.Replace(wxT("\\"), wxT("\\\\"), TRUE);
}

//-----

FilenameWX_cls::
FilenameWX_cls(){
    mwsPath = wxT("");
}

FilenameWX_cls::
FilenameWX_cls(const wxChar *path){
    mwsPath = path;
}

FilenameWX_cls::
FilenameWX_cls(wxString pathStr){
    mwsPath = pathStr;
}

wxString FilenameWX_cls::
GetPathWXStr(int flag){
    SplitPathWX_cls CPath(mwsPath);

    return CPath.GetPathWXStr(flag);
}

const wxChar* FilenameWX_cls::
c_str(){
    return mwsPath;
}

void FilenameWX_cls::
Set(wxString strPath){
    mwsPath = strPath;
}

void FilenameWX_cls::
DoubleSlash(wxString &str){
    str.Replace(wxT("\\"), wxT("\\\\"), TRUE);
}

//----

#include "../MYAPI/MsgBox.h"

// ----------------------

FileSequenceWX_cls::
FileSequenceWX_cls() : FilenameWX_cls() {
    mDigitLength = -1;
    mChannel = 0;
    mZSliceInit = 0;
}

FileSequenceWX_cls::
FileSequenceWX_cls(const wxChar *path) : FilenameWX_cls(path){
    Set(mwsPath);
}

FileSequenceWX_cls::
FileSequenceWX_cls(wxString pathStr) : FilenameWX_cls(pathStr){
    Set(mwsPath);
}

void FileSequenceWX_cls::
Set(const wxString pathStr){
    FilenameWX_cls::Set(pathStr);
    int formatPos = SequenceFormatPos();
    if (formatPos == wxNOT_FOUND) InitFormat();
    else SplitFormat(formatPos);
}

int FileSequenceWX_cls::
SequenceFormatPos(){
    for (int index = 1; index < 6; index++){
	wxString buf;
	buf.Printf(wxT("%s0%dd"), wxT("%"), index);
	int pos = mwsPath.Find(buf.c_str());
	if (pos != wxNOT_FOUND) {
	    mDigitLength = index;
	    mDigit = 0;
	    return pos;
	}
    }
    return wxNOT_FOUND;
}

void FileSequenceWX_cls::
InitFormat(){
    // %0[1-5]d甲
    for (int i = 1; i < 6; i++){
	wxString buf;
	buf.Printf(wxT("%s0%dd"), wxT("%"), i);
	if (mwsPath.Find(buf.c_str()) != wxNOT_FOUND) {
	    mDigitLength = i; //?
	    mwsFormat = mwsPath;
	    return;
	}
    }

    SplitPathWX_cls CPath(mwsPath);
    wxString strPathMinusExt = CPath.mwsFullPathMinusExt;

    int posTail = (int)strPathMinusExt.Len() - 1;
    if (posTail > 999) return;

    while (strPathMinusExt[posTail] >= '0' && 
	   strPathMinusExt[posTail] <= '9'){
	posTail--;
	if (posTail > 999) break;
    }

    posTail++;

    if (posTail == strPathMinusExt.Len()){
	mDigit = 0; mDigitLength = -1;
	mwsFileHead = strPathMinusExt;
	mwsFormat = mwsPath;
	return;
    }
    
    wxString strFH = CPath.mwsFullPathMinusExt.SubString(0, posTail - 1);
    wxString strDigit = CPath.mwsFullPathMinusExt.c_str() + posTail;
    mDigit = atoi(strDigit.ToAscii());
    mDigitLength = (int)(CPath.mwsFullPathMinusExt.Len() - posTail);
    wxString strFormatMain;
    strFormatMain.Printf(wxT("%s0%dd"), wxT("%"), mDigitLength);

    mwsFileHead = strFH;
    mwsFormat = strFH + strFormatMain + CPath.mwsExt;
    mwsExt = CPath.mwsExt;
}

void FileSequenceWX_cls::
SplitFormat(int pos){
    SplitPathWX_cls CPath(mwsPath);
    mwsFileHead = mwsPath.SubString(0, pos);
    mwsFormat = mwsPath;
    mwsExt = CPath.mwsExt;
}

void FileSequenceWX_cls::
CalcMaxOffset(){
    if (mDigitLength == -1) {
	mSeqLength = 1;
	return;
    }

    if (1==0) // ԕὠfBNgω鎞݂̂ɁBĂCHH
	if( mwsExt == wxT(".bmp") || mwsExt == wxT(".BMP") ||
	    mwsExt == wxT(".png") || mwsExt == wxT(".PNG") ||
	    mwsExt == wxT(".tif") || mwsExt == wxT(".TIF") ||
	    mwsExt == wxT(".tiff") || mwsExt == wxT(".TIFF")) {
	    // 摜̘AԂ́A{PXebv̂
	    SplitPathWX_cls CSplit0(GetModifiedWXStr(0));
	    SplitPathWX_cls CSplit1(GetModifiedWXStr(1));
	    if (CSplit0.GetPathWXStr(_FN_DRIVE_DIR) == 
		CSplit1.GetPathWXStr(_FN_DRIVE_DIR)){
		mSeqLength = 1;
		return;
	    }
	}
    
    mSeqLength    = 0; 
    int increment = 1000;
    int offset    = 0;

#ifdef WIN32
    std::wfstream ifstream(GetModifiedWXStr(offset).c_str(), std::ios::in);
#else
    wxString wxFileIn = wxConvFile.cWC2WX(GetModifiedWXStr(offset).c_str());
    std::ifstream ifstream(wxFileIn.ToAscii(), std::ios::in);
#endif

    wxString wxMes = wxT("Can not open file");
    
    if (ifstream.fail()) {// ȃt@C
	MsgBoxf(NULL, WCHARCONV("t@Cǂ߂܂"),
		"FileSequence_cls::CalcMaxOffset\n%ls\nFile open fails", 
		GetModifiedWXStr(offset).c_str() );
	return ; } 
    ifstream.close();

    bool exitFlag = FALSE;

    while (TRUE){
	do {
	    if (increment == 1){ // increment1J艺ȂI
		exitFlag = TRUE; break;
	    }
	    increment /= 10;
	
	    ifstream.clear();
#ifdef WIN32
	    ifstream.open(GetModifiedWXStr(offset + increment).c_str(),
			  std::ios::in);
#else
	    wxFileIn = wxConvFile.cWC2WX(GetModifiedWXStr(offset + increment).c_str());
	    ifstream.open(wxFileIn.ToAscii(), std::ios::in);
#endif
	    if (!ifstream.fail()) ifstream.close();
	    /*
	    else
		MsgBoxf(NULL, "FileSequence_cls::CalcMaxOffset",
			"open %ls fails %d", 
			GetModifiedWXStr(offset + increment).c_str(),
			mSeqLength);
	    */
	    
	} while (ifstream.fail()); // t@CincrementJ艺
	
	mSeqLength += increment;
	offset  += increment;
	if (exitFlag == TRUE) break;
	increment *= 10;
    }

    //not yet
}

wxString FileSequenceWX_cls::
GetFormat(){
    return mwsFormat;
}

int FileSequenceWX_cls::
CountNumericalExpression(wxString strFormat, std::vector < int > &VOrder){
    int count = 0;
    wxString strTmp = strFormat;
    int loc = 0;
    int locMin = 999;
    int order = 0;
    wxString strTagFormat = wxT("%%0%dd");
    
    while (TRUE){
	for (int i = 1; i < 9; i++){
	    wxString strTag;
	    strTag.Printf(strTagFormat,i);
	    if ((loc = strTmp.Find(strTag)) != -1){
		if (loc < locMin) locMin = loc;
		order = i;
	    }
	}
	
	if (locMin != 999){
	    count++;
	    strTmp = strTmp.Mid(locMin+4);
	    locMin = 999;
	    VOrder.push_back(order);
	}
	else break;
    }

    return count;
}

wxString FileSequenceWX_cls::
GetModifiedWXStr(int offset, int stereo, bool bModifyAll){
    wxString buf;
    wxString strFormat = GetFormat();
    if (stereo != 0){
	strFormat.Replace(wxT("%"), wxT("%s%"));
	wxString strStereo;
	if (stereo == 1) strStereo = wxT("L");
	if (stereo == 2) strStereo = wxT("R");
	buf.Printf(strFormat, strStereo.c_str(), mDigit + offset);
    }
    else{
	std::vector <int> VOrder;
	if (CountNumericalExpression(strFormat, VOrder) <= 1)
	    buf.Printf(strFormat, mDigit + offset);
	else{
	    if (bModifyAll == TRUE)
		buf.Printf(strFormat, mDigit + offset, mZSliceInit);
	    else{
		// ŏ%0ndu
		wxString strTagT = wxT("%0");
		strTagT << VOrder[0] << wxT("d");
		strFormat.Replace(strTagT, wxT("%%"), FALSE);
		wxString strTagZ = wxT("%0");
		strTagZ << VOrder[1] << wxT("d");
		strFormat.Replace(strTagZ, wxT("%s"), FALSE);
		strFormat.Replace(wxT("%%"), strTagT, FALSE);
		buf.Printf(strFormat, mDigit + offset, strTagZ.c_str());
	    }
	}
    }
    return buf;
}

wxString FileSequenceWX_cls::
GetInfoStr(){
    wxString strInfo;
    strInfo << mwsFormat;
    if (mDigitLength != -1)
	strInfo << wxT(" (") << mDigit << wxT("-") << mDigit + mSeqLength << wxT(")");
    return strInfo;
}

wxString FileSequenceWX_cls::
GetChannelStr(){
    wxString strChannel;
    if (mChannel == 0) strChannel = wxT("Red");
    if (mChannel == 1) strChannel = wxT("Green");
    if (mChannel == 2) strChannel = wxT("Blue");
    if (mChannel == 3) strChannel = wxT("Alpha (not implemented)");
    return strChannel;
}

void FileSequenceWX_cls::
AppendDefaultSequenceTag(){
    Set(mwsFileHead+wxT("%04d")+mwsExt);
}
