// (setq c-basic-offset 4)

#include "clsReadBodyImgSeq.h"
#include "../clsDF3.h"
#include "../clsDF3Sequence.h"
#include "../GL/clsColoring.h"
extern Coloring_cls GCColoring;

#include "../../COMMON/clsFilename.h"
#include "../../MAINCANVAS/clsPreference.h"
#include "../../MYAPI/MsgBox.h"

#include <wx/wx.h>
#include <wx/bitmap.h>
#include <wx/stopwatch.h>
#include <fstream>
#include <time.h>

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2

// ----------------

ThreadReaderOneImage::
ThreadReaderOneImage() : 
    wxThread(wxTHREAD_DETACHED){
}

void ThreadReaderOneImage::
SetParameters(bool *pbFinishedFlag, ReadBodyImgSeq_cls *pCParent,
	      wxWindow *pCMsgParent, int initz, int finz, int offset,
	      int bitmapType, FileSequenceWX_cls CFileImgSeq){
    mpbFinishedFlagLink = pbFinishedFlag;
    mpCParent = pCParent;
    mpCMsgParent = pCMsgParent;
    mInitZ = initz;
    mFinZ = finz;
    mOffset = offset;
    mBitmapType = bitmapType;
    mCFileImgSeq = CFileImgSeq;
    
    if ( Create() != wxTHREAD_NO_ERROR )
	MsgBoxExit(mpCMsgParent, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadReaderOneImage::
Entry(){
    int sizeSxT = mpCParent->mpC->GetCurrSizeSxT(Z);

    for (int z = mInitZ; z < mFinZ; z++){
	wxImage CImage;
	if (CImage.LoadFile(mCFileImgSeq.GetModifiedWXStr(z + mOffset), wxBITMAP_TYPE_PNG) == FALSE){
	    MsgBox(mpCMsgParent, _("read error thread"),
		   mpCParent->mpC->GetFilePathWXStr() + _("can not open"));
	    
	    return 0;
	}

	for (unsigned int addr = 0; addr < sizeSxT; addr++)
	    mpCParent->mpC->mpMainXYZ[Z][sizeSxT * z + addr]
		= CImage.GetData()[addr * 3 + mpCParent->mChannel];
    }
	
    *mpbFinishedFlagLink = TRUE;
    
    return 0;
}

// ----------------

void ReadBodyImgSeq_cls::
InitializeMember(DF3_cls *pCDF3){
    mpC = pCDF3;
}

void ReadBodyImgSeq_cls::
InitializeStaticMembers(Preference_cls *pCPreference){
    mspCPreferenceLink = pCPreference;
}

int ReadBodyImgSeq_cls::
GetBitmapType(wxString filename){
    SplitPathWX_cls CPath(filename);
    if (CPath.GetPathWXStr(_FN_EXT) == wxT(".bmp") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".BMP"))
	return wxBITMAP_TYPE_BMP;

    if (CPath.GetPathWXStr(_FN_EXT) == wxT(".png") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".PNG"))
	return wxBITMAP_TYPE_PNG;
    
    if (CPath.GetPathWXStr(_FN_EXT) == wxT(".tif") || CPath.GetPathWXStr(_FN_EXT) == wxT(".tiff") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".TIF") || CPath.GetPathWXStr(_FN_EXT) == wxT(".TIFF"))
	return wxBITMAP_TYPE_TIF;
    return -1;
}

bool ReadBodyImgSeq_cls::
ReadImgWithThreads(wxWindow *pC, int offset, int threadNum){
    FileSequenceWX_cls CFileImgSeq(mpC->GetFilePathFormat());
    int bitmapType = GetBitmapType(CFileImgSeq.GetModifiedWXStr(0)); //0ő\
    if (bitmapType == -1) return FALSE;
    
    bool *pbAFinished;
    ThreadReaderOneImage **ppCThread;
    pbAFinished = new bool [threadNum];
    ppCThread = new ThreadReaderOneImage* [threadNum];
    int zSizeEach = mpC->mCurrSize[Z]/threadNum;
    for (int i = 0; i < threadNum; i++){
	pbAFinished[i] = FALSE;
	int zIni = zSizeEach * i;
	int zFin = zSizeEach * (i+1);
	if (i == threadNum - 1) zFin = mpC->mCurrSize[Z];
	ppCThread[i] = new ThreadReaderOneImage;
	ppCThread[i]->SetParameters
	    (&pbAFinished[i], this, pC, zIni, zFin, offset, bitmapType, 
	     CFileImgSeq);
	ppCThread[i]->Run();
    }
    
    bool bFinishedAll = FALSE;
    while (bFinishedAll == FALSE){
	wxMilliSleep(3);
	bFinishedAll = TRUE;
	for (int core = 0; core < threadNum; core++) 
	    bFinishedAll &= pbAFinished[core];
    }

    delete [] pbAFinished;
    delete [] ppCThread;

    return TRUE;
}

bool ReadBodyImgSeq_cls::
ReadImg1Layer(int z, int offset){
    FileSequenceWX_cls CFileImgSeq(mpC->GetFilePathFormat());
    int bitmapType = GetBitmapType(CFileImgSeq.GetModifiedWXStr(z + offset));
    if (bitmapType == -1) return FALSE;
    wxImage CImage;
    wxString strFilePath = CFileImgSeq.GetModifiedWXStr(z + offset);
    if (CImage.LoadFile(strFilePath, bitmapType) == FALSE){
	MsgBox(NULL, strFilePath + _("can not open"),
	       _("read error thread"));
	return FALSE;
    }
    
    int sizeSxT = mpC->GetCurrSizeSxT(Z);
    unsigned char *pTarget = mpC->mpMainXYZ[Z];
    unsigned char *pSRC = CImage.GetData();
    for (unsigned int addr = 0; addr < sizeSxT; addr++)
	pTarget[sizeSxT*z + addr]
	    = pSRC[addr * 3 + mChannel];
    return TRUE;
}

bool ReadBodyImgSeq_cls::
ReadImg1LayerDownsized(int z, int offset){
    int trueSizeS    = mpC->GetTrueSizeS(Z);
    int currSizeSxT  = mpC->GetCurrSizeSxT(Z);
    int currSizeS    = mpC->GetCurrSizeS(Z);
    int downsizeX = DF3Sequence_cls::msCDownSizeSrc.mpBody[X];
    int downsizeY = DF3Sequence_cls::msCDownSizeSrc.mpBody[Y];
    int downsizeZ = DF3Sequence_cls::msCDownSizeSrc.mpBody[Z];
    FileSequenceWX_cls CFileImgSeq(mpC->GetFilePathFormat());
    int zForFile = z * downsizeZ;
    int bitmapType = GetBitmapType(CFileImgSeq.GetModifiedWXStr(zForFile + offset));
    if (bitmapType == -1) return FALSE;
    wxImage CImage;
    wxString strFilePath = CFileImgSeq.GetModifiedWXStr(zForFile + offset);
    if (CImage.LoadFile(strFilePath, bitmapType) == FALSE){
	MsgBox(NULL, strFilePath + _("can not open"),
	       _("read error thread"));
	return FALSE;
    }

    unsigned char *pTarget = mpC->mpMainXYZ[Z];
    unsigned char *pOffsetPointZ = pTarget + currSizeSxT*z;
    unsigned char *pSRC = CImage.GetData();
    for (unsigned int y = 0; y < mpC->mCurrSize[Y]; y++)
	for (unsigned int x = 0; x < mpC->mCurrSize[X]; x++)
	    *(pOffsetPointZ + currSizeS*y + x) = 
		pSRC[(trueSizeS*y*downsizeY + x * downsizeX)*3 + mChannel];
    return TRUE;
}

bool ReadBodyImgSeq_cls::
MainProcess(bool type, wxWindow* pC){

    static unsigned int coreNumber = wxThread::GetCPUCount();
    if (mspCPreferenceLink->mCThreadNumImgRead.mBody > coreNumber)
	mspCPreferenceLink->mCThreadNumImgRead.mBody = coreNumber;
    
    Initialize(mpC->GetCurrSize(Z));

    mpC->mOldPointXYZ = 10;
    assert(mpC->IsAllocated() == TRUE);

    wxStopWatch CSW;
   
    int offset = mpC->mpCParent->mDefaultZSliceInit;

    if (DF3Sequence_cls::msCDownSizeSrc.mpBody[X] == 1 &&
	DF3Sequence_cls::msCDownSizeSrc.mpBody[Y] == 1 && 
	DF3Sequence_cls::msCDownSizeSrc.mpBody[Z] == 1)
	for (unsigned int z = mNext; z < mpC->mCurrSize[Z]; z++){
	    
	    if (mspCPreferenceLink->mCbMultiThreadImgRead.mbBody == TRUE){
		int threadNum = mspCPreferenceLink->mCThreadNumImgRead.mBody;
		if (threadNum > 3) threadNum = 3;
		if (threadNum != 1){
		    if (ReadImgWithThreads(pC, offset, threadNum) == FALSE)
			return FALSE;
		    mNext = mpC->mCurrSize[Z];
		    break;
		}
	    }

	    if (ReadImg1Layer(z, offset) == FALSE) return FALSE;
	    mNext = z + 1;
	    if (IsResumeConditionSatisfied(type) == TRUE) break;
	}
    else{
	// ---- READ with DownSize ----
	for (unsigned int z = mNext; z < mpC->mCurrSize[Z]; z++){
	    if (ReadImg1LayerDownsized(z, offset) == FALSE) return FALSE;
	    mNext = z + 1;
	    if (IsResumeConditionSatisfied(type) == TRUE) break;
	}
    }
    
    if (IsFinishConditionSatisfied() == FALSE) return TRUE;
    
    Finalize();
    mpC->mbDirty = TRUE;

    //MsgBoxf(pC, "!!", "%d ms", CSW.Time());
    
    return TRUE;
}

bool ReadBodyImgSeq_cls::
ReadHeaderInfo(wxWindow *pCWindow){
    
    //FileSequenceWX_cls CFileImgSeq(mpC->GetFilePathWXStr());
    FileSequenceWX_cls CFileImgSeq(mpC->GetFilePathFormat());
    
    int index = 0;
    int offset = mpC->mpCParent->mDefaultZSliceInit;
    while (TRUE){
	wxString strFilePath = CFileImgSeq.GetModifiedWXStr(index + offset);
	FILE *fp;
#ifdef WIN32
	fp = fopen(wxFNCONV(strFilePath.ToAscii()), "rb");
#else
	fp = fopen(strFilePath.ToAscii(), "rb");
#endif
	if (fp == NULL) break;
	index++;
	fclose(fp);
    }
    
    mpC->mTrueSize[Z] = index;

    int bitmapType = GetBitmapType(CFileImgSeq.GetModifiedWXStr(0));
    wxImage CImage;
    //int zInit = mpC->mpCParent->mDefaultZSliceInit;
    if (CImage.LoadFile(/*CFileImgSeq.GetModifiedWXStr(zInit)*/mpC->GetFilePathWXStr(), bitmapType) == FALSE){
	MsgBox(pCWindow, /*CFileImgSeq.GetModifiedWXStr(zInit)*/mpC->GetFilePathWXStr() + _("can not open"),
	       _("read error Header Info Image"));
	return FALSE; }
    
    mpC->mTrueSize[X] = CImage.GetWidth();
    mpC->mTrueSize[Y] = CImage.GetHeight();

    return TRUE;
}

Preference_cls *ReadBodyImgSeq_cls::mspCPreferenceLink;
//unsigned int ReadBodyImgSeq_cls::msReadChannel = 0;
