// "clsColoring.h"
// (setq c-basic-offset 4)

#include "clsColoring.h"
//#include "../UNDO/clsUndoManager.h"

#include <math.h>

#include <sstream>
#include <cmath>

#include "../../POVRAY/clsPovRay.h"

#include "../../DF3/GL/clsDF3GL.h"
#include "../../DF3/GL/clsPreviewProperty.h"
#include "../../WIDGETCTRL/clsGraphPanel.h"
#include "../../WIDGETCTRL/clsIconSelector.h"
#include "../../WIDGETCTRL/clsSliderInt.h"
#include "../../WIDGETCTRL/EX/clsButtonEX.h"
#include "../../WIDGETCTRL/EX/clsColourDialogEX.h"

#include "../../../wxWidgetEnum.h"
#include "../../../wxWidgetEnumMenu.h"

//#include <wx/colordlg.h>

extern wxColourDialogEX* GpCClrDialog;

#define ENDL std::endl
#define ENDLWX wxT("\n")
#define LOOP_XYZ for(int xyz = 0; xyz < 3; xyz++)
#define R 0
#define G 1
#define B 2
#define LOOP_RGB for(unsigned int rgb = 0; rgb < 3; rgb++)

#define _CLR_RED_YELLOW 1
#define _CLR_COLD_COLOR 2
#define _CLR_GREEN_YELLOW_RED 3
#define _CLR_BLACK_WHITE 4
#define _CLR_MODELATE_GRAY 5
#define _CLR_SPECTRUM 6
#define _CLR_SPECTRUM_RBG 7
#define _CLR_SPECTRUM_RBGR 8
#define _CLR_INCANDESCENT 9
#define _CLR_COLD_STEEL 10

#ifdef WIN32
#define _MEM_PANEL_OFFSET 14
#else
#define _MEM_PANEL_OFFSET 10
#endif

extern UndoManager_cls GCUndoList;

BEGIN_EVENT_TABLE(CustomColorGUI_cls, wxDialog)
EVT_SIZE(CustomColorGUI_cls::OnSize)
EVT_BUTTON(btnCtrlCustomClrAdd, CustomColorGUI_cls::OnButtonAdd)
EVT_BUTTON(btnCtrlCustomClrRemove, CustomColorGUI_cls::OnButtonRemove)
EVT_BUTTON(btnCtrlCustomClrClose,  MultiParametersGUI_cls::OnButtonClose)
EVT_BUTTON(btnCtrlCustomClrAddKey, CustomColorGUI_cls::OnButtonAddKey)
EVT_BUTTON(btnCtrlCustomClrRemKey, CustomColorGUI_cls::OnButtonRemKey)
EVT_BUTTON(btnCtrlCustomClrColor,  CustomColorGUI_cls::OnButtonColor)
EVT_BUTTON(btnCtrlCustomClrInvert, CustomColorGUI_cls::OnButtonInvert)
EVT_BUTTON(btnCtrlCustomClrChannel,CustomColorGUI_cls::OnButtonChannelShift)
EVT_KEY_DOWN(CustomColorGUI_cls::OnKeyDown)
END_EVENT_TABLE()

CustomColorGUI_cls::
CustomColorGUI_cls(CustomColor_cls* pCParent, wxWindow *pC,
		   wxPoint pos, wxSize size) :
mpCGraphColor(NULL),
    mpCPanelColorBar(NULL), mpCPanelColorBar2(NULL),
    MultiParametersGUI_cls(pCParent, pC, _("Custom Color"),
			   pos, size, pCParent->mpBitmapLink, wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER)
{
    mTemplate = 0;
    mspCThis = this;
    mpCParentColor = pCParent;

    SetCloseButtonID(btnCtrlCustomClrClose);
    SetScreenRectPos(&pCParent->mCScreenRect);

    // $B%5%$%6!<$r:n$k(B
    wxBoxSizer       *pBoxSizerSrc   = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer       *pBoxSizerLFT   = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerLBTM  = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer       *pBoxSizerRGT   = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerRMID  = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer       *pBoxSizerRBTM  = new wxBoxSizer(wxHORIZONTAL);
    
    // GRAPH
    mpCGraphColor = new GraphPanel_cls(this, wxID_ANY, wxDefaultPosition, wxSize(-1, -1));

    mpCGraphColor->mbDisplayCurrent = TRUE;
    mpCGraphColor->mbDisplayKeyPoint = TRUE;
    mpCGraphColor->mBorderSizeOffset[0] = 1;
    mpCGraphColor->SetFunction(&GetCustomColorR);
    mpCGraphColor->SetColorLastItem(wxColour(255, 0, 0));
    mpCGraphColor->SetFunction(&GetCustomColorG);
    mpCGraphColor->SetColorLastItem(wxColour(0, 255, 0));
    mpCGraphColor->SetFunction(&GetCustomColorB);
    mpCGraphColor->SetColorLastItem(wxColour(0, 0, 255));
    mpCGraphColor->mOffset[0] = _MEM_PANEL_OFFSET;

    mpCPanelColorBar2 = new PanelColorBar_cls(this, wxID_ANY, wxDefaultPosition, wxSize(-1, 18));
    mpCPanelColorBar2->mbDisplayCurrent = TRUE;
    mpCPanelColorBar2->mbDisplayBorder = FALSE;
    mpCPanelColorBar2->mOffset[1] = 1;
    mpCPanelColorBar2->mBorderSizeOffset[0] = 1;
    mpCPanelColorBar2->SetFunction(&GetCustomColorRxD);
    mpCPanelColorBar2->SetFunction(&GetCustomColorGxD);
    mpCPanelColorBar2->SetFunction(&GetCustomColorBxD);
    mpCPanelColorBar2->mOffset[0] = _MEM_PANEL_OFFSET;
    mpCPanelColorBar2->mstrLabel = wxT("color * density");

    mpCPanelColorBar = new PanelColorBar_cls(this, wxID_ANY, wxDefaultPosition, wxSize(-1, 18));
    mpCPanelColorBar->mbDisplayCurrent = TRUE;
    mpCPanelColorBar->mbDisplayBorder = FALSE;
    mpCPanelColorBar->mOffset[1] = 1;
    mpCPanelColorBar->mBorderSizeOffset[0] = 1;
    mpCPanelColorBar->SetFunction(&GetCustomColorR);
    mpCPanelColorBar->SetFunction(&GetCustomColorG);
    mpCPanelColorBar->SetFunction(&GetCustomColorB);
    mpCPanelColorBar->mOffset[0] = _MEM_PANEL_OFFSET;

    mpCSlider = new SliderInt_cls(this, wxID_ANY, mpCParentColor->mIndexInt, 0, 255);
    Connect(mpCSlider->GetId(), wxEVT_SCROLL_CHANGED,
	    wxCommandEventHandler(CustomColorGUI_cls::OnSlider));
    Connect(mpCSlider->GetId(), wxEVT_SCROLL_PAGEUP,
	    wxCommandEventHandler(CustomColorGUI_cls::OnSliderPageUp));
    Connect(mpCSlider->GetId(), wxEVT_SCROLL_PAGEDOWN,
	    wxCommandEventHandler(CustomColorGUI_cls::OnSliderPageDown));
        
    pBoxSizerLFT->Add(mpCPanelColorBar2, 0, wxEXPAND);
    pBoxSizerLFT->Add(mpCPanelColorBar, 0, wxEXPAND);
    pBoxSizerLFT->Add(mpCGraphColor, 1, wxEXPAND);
    pBoxSizerLFT->Add(mpCSlider, 0, wxEXPAND);

    mpCButtonAddKey = 
	new wxButtonEX (this, btnCtrlCustomClrAddKey, _("Add Key"));
    mpCButtonRemKey = 
	new wxButtonEX (this, btnCtrlCustomClrRemKey, _("Remove Key"));
    mpCButtonColor = 
	new wxButtonEX (this, btnCtrlCustomClrColor, wxT(""), wxDefaultPosition, wxSize(80, -1));

    pBoxSizerLBTM->Add(mpCButtonAddKey, 0, wxEXPAND);
    pBoxSizerLBTM->Add(mpCButtonRemKey, 0, wxEXPAND); 
    pBoxSizerLBTM->Add(-1, -1, 1, wxEXPAND);
    pBoxSizerLBTM->Add(mpCButtonColor, 0, wxEXPAND); 

    wxButtonEX *pCButtonChannel
	= new wxButtonEX (this, btnCtrlCustomClrChannel, _("Clr Shift"));
    pBoxSizerRBTM->Add( pCButtonChannel, 0, wxEXPAND);
    
    wxButtonEX *pCButtonInvert
	= new wxButtonEX (this, btnCtrlCustomClrInvert, _("Invert"));
    pBoxSizerRBTM->Add( pCButtonInvert, 0, wxEXPAND);

    pBoxSizerRBTM->Add( -1, -1, 1, wxEXPAND);

    /*
    wxButtonEX *pCButtonClose 
	= new wxButtonEX (this, btnCtrlCustomClrClose, wxT("Close"));
    pBoxSizerRBTM->Add( pCButtonClose,  0, wxEXPAND);
    */

    SetMNumMain(pBoxSizerRGT, 
		btnCtrlCustomClrAdd, btnCtrlCustomClrRemove);
    
    SetMNumParam(pBoxSizerRGT);
    pBoxSizerRGT->Add( pBoxSizerRMID,  0, wxEXPAND);
    pBoxSizerRGT->Add( -1, -1, 1);
    pBoxSizerRGT->Add( pBoxSizerRBTM,  0, /*wxALIGN_RIGHT*/wxEXPAND);
        
    pBoxSizerLFT->Add( pBoxSizerLBTM,  0, wxEXPAND);

    pBoxSizerSrc->Add( pBoxSizerLFT, -1, wxEXPAND);
    pBoxSizerSrc->Add( pBoxSizerRGT, 0, wxEXPAND);

    SetSizerAndFit(pBoxSizerSrc);

    UpdateValues();
};

CustomColorGUI_cls::
~CustomColorGUI_cls(){
    MakeModal(FALSE);
}

void CustomColorGUI_cls::
SetMNumParam(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr =
	InitializeMNumPanel(&mpCMNumParam, this);
    pCurr->mbHideButton = FALSE;

    wxString listTemplate[11] = 
	{ wxT("-- select --"), _("Red Yellow"), _("Cold Color"),
	  _("Green Red"), _("Black White"), _("Modelate Gray"), 
	  _("Optical"), _("SpectrumRBG"), _("SpectrumRBGR"), 
	  _("Incandescent"), _("Cold Steel") };
    
    pCurr->mTextWidth = 60;
    pCurr->mComboWidth = 108;
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mTemplate,
			  wxT("Template"), TRUE, 11, listTemplate);
    
    pCurr->mTextWidth = -1;
    int boxWidth = 2;
    mIndexIndex = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Index"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &mpCParentColor->mIndexInt);
    pCurr->SetRange(0, 255);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParentColor->mIndexInt);
    pCurr->SetRange(0, 255);

    mIndexR = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("R"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &mpCParentColor->mCacheColor[0]);
    pCurr->SetRange(0, 255);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParentColor->mCacheColor[0]);
    pCurr->SetRange(0, 255);

    mIndexG = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("G"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &mpCParentColor->mCacheColor[1]);
    pCurr->SetRange(0, 255);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParentColor->mCacheColor[1]);
    pCurr->SetRange(0, 255);

    mIndexB = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("B"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &mpCParentColor->mCacheColor[2]);
    pCurr->SetRange(0, 255);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParentColor->mCacheColor[2]);
    pCurr->SetRange(0, 255);

    pCurr->CombinePointer(_GP_BOOL, &mpCParentColor->mbCacheMultiplyLevel, wxT("*density"));
    
    //pCurr->SetFuncCheck(&FuncHookSyncFromPanel, NULL);

#ifdef WIN32
    pCurr->mMinimumWidth = 180;
#else
    pCurr->mMinimumWidth = 220;
#endif

    pCurr->FinishSetting(pBoxSizer, TRUE);

    /*
      Connect(pCurr->mVCPointers[1].mID, wxEVT_SCROLL_CHANGED,
      wxCommandEventHandler(CustomColorGUI_cls::OnSliderParam));
      Connect(pCurr->mVCPointers[4].mID, wxEVT_SCROLL_CHANGED,
      wxCommandEventHandler(CustomColorGUI_cls::OnSliderParam));
      Connect(pCurr->mVCPointers[7].mID, wxEVT_SCROLL_CHANGED,
      wxCommandEventHandler(CustomColorGUI_cls::OnSliderParam));
      Connect(pCurr->mVCPointers[10].mID, wxEVT_SCROLL_CHANGED,
      wxCommandEventHandler(CustomColorGUI_cls::OnSliderParam));
    */

    pCurr->SetFuncHook(&FuncHookSyncFromPanel, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&FuncHookSyncFromPanel, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookCombo, _MN_HOOK_COMBOBOX_CHANGE);
}

void CustomColorGUI_cls::
RemakeIconList(){
    mVIcon.clear();
    for (int i = 0; i< mpCParent->mCItemNum.mBody; i++){
	unsigned char *pData = (unsigned char*)malloc(16*16*3);
	for (int x = 0; x < 16; x++){
	    double value = (double)x/16;
	    pData[x*3+0] = mpCParentColor->GetCustomColor(value, R, TRUE, i) * 255;
	    pData[x*3+1] = mpCParentColor->GetCustomColor(value, G, TRUE, i) * 255;
	    pData[x*3+2] = mpCParentColor->GetCustomColor(value, B, TRUE, i) * 255;
	}
	for (int y = 1; y < 16; y++)
	    memcpy(pData+y*16*3, pData, 16*3);
	wxImage image(16, 16, pData);
	wxBitmap bitmap(image);
	wxIcon icon; icon.CopyFromBitmap(bitmap);
	mVIcon.push_back(icon);
    }
    ((IconSelector_cls*)mpCMNumMain->mVCPointers[mIndexIcon].mpPointerCtrl)
	->RemakeImageList(&mVIcon);
}

void CustomColorGUI_cls::
EnableColorCtrl(bool bEnable){
    mpCMNumParam->EnableItems(mIndexIndex, 3, bEnable);
    mpCMNumParam->EnableItems(mIndexR, 3, bEnable);
    mpCMNumParam->EnableItems(mIndexG, 3, bEnable);
    mpCMNumParam->EnableItems(mIndexB, 3, bEnable);
}

void CustomColorGUI_cls::
UpdateValues(bool bRefresh){
    mpCParentColor->AdjustAutoControl();
    mpCGraphColor->ClearKeyPoint();
    if (mpCParentColor->GetPActiveItem() != NULL)
	for (unsigned int index = 0; index < mpCParentColor->GetActiveItemKeyNum(); index++){
	    double x = mpCParentColor->GetPActiveItem()->mVCKeys[index].mDoubleIndex;
	    double *py = mpCParentColor->GetPActiveItem()->GetPKeyFrameValue(index);
	    mpCGraphColor->SetKeyPoint(0, x, py[0]);
	    mpCGraphColor->SetKeyPoint(1, x, py[1]);
	    mpCGraphColor->SetKeyPoint(2, x, py[2]);
	}
      
    if (mpCParentColor->IsKeyPoint() == TRUE){
	mpCButtonRemKey->Enable(TRUE);
	//mpCMNumParam->Enable(TRUE);
	EnableColorCtrl(TRUE);
	mpCParentColor->mCacheIndexInt = mpCParentColor->mIndexInt;
    }
    else{
	mpCButtonRemKey->Disable();
	EnableColorCtrl(FALSE);
	//mpCMNumParam->Disable();
    }

    UpdateComboBox(); 
    mpCMNumMain->UpdateValues();
    mpCParentColor->UpdateCacheColor();
    mpCMNumParam->UpdateValues();
    
    mpCGraphColor->Refresh(TRUE);

    if (mpCParentColor->GetPActiveItem() == NULL){
	mpCButtonAddKey->Disable();
	mpCButtonColor->Disable();
	mpCButtonColor->SetBackgroundColour(wxT("GRAY"));
    }
    else{
	mpCButtonAddKey->Enable(TRUE);
	mpCButtonColor->Enable(TRUE);
	mpCButtonColor->SetBackgroundColour
	    (wxColour(mpCParentColor->mCacheColorTrim[0],
		      mpCParentColor->mCacheColorTrim[1],
		      mpCParentColor->mCacheColorTrim[2]));
    }

    wxCommandEvent com1(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshColorBar);
    GetParent()->AddPendingEvent(com1);

    if (bRefresh == TRUE){
	wxCommandEvent com2(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	GetParent()->AddPendingEvent(com2);
    }
}

void CustomColorGUI_cls::
UpdateCurrentIndex(bool bRefresh){
    mpCParentColor->mIndexInt = mpCSlider->GetValue();
    mpCGraphColor->mCurrentDoubleIndex = mpCParentColor->GetDoubleIndex();
    mpCPanelColorBar->mCurrentDoubleIndex = mpCParentColor->GetDoubleIndex();
    mpCPanelColorBar2->mCurrentDoubleIndex = mpCParentColor->GetDoubleIndex();
    UpdateValues(bRefresh);
}

void CustomColorGUI_cls::
PageUp(){
    if (mpCParentColor->GetPActiveItem() == NULL) return;
    mpCParentColor->mIndexInt = 
	(int)(mpCParentColor->GetPActiveItem()->GetPrevKeyPosDoubleIndex
	      (mpCParentColor->GetDoubleIndex(), 1.0/255) * 255 + 0.500);
    mpCSlider->SetValue(mpCParentColor->mIndexInt);
    UpdateCurrentIndex();
}

void CustomColorGUI_cls::
PageDown(){
    if (mpCParentColor->GetPActiveItem() == NULL) return;
    mpCParentColor->mIndexInt = 
	(int)(mpCParentColor->GetPActiveItem()->GetNextKeyPosDoubleIndex
	      (mpCParentColor->GetDoubleIndex(), 1.0/255) * 255 + 0.500);
    mpCSlider->SetValue(mpCParentColor->mIndexInt);
    UpdateCurrentIndex();
}

void CustomColorGUI_cls::
OnSize(wxSizeEvent& event){
    mpCGraphColor->Refresh(FALSE);
    mpCPanelColorBar->Refresh(FALSE);
    mpCPanelColorBar2->Refresh(FALSE);
    event.Skip();
}

void CustomColorGUI_cls::
OnButtonAddKey(wxCommandEvent &event){
    double pKey[4] = {
	GetCustomColorR(mpCParentColor->GetDoubleIndex()),
	GetCustomColorG(mpCParentColor->GetDoubleIndex()),
	GetCustomColorB(mpCParentColor->GetDoubleIndex()),
	1.0};

    KeyFrame_cls CKey(4), CKeySlave(0);
    CKey.Set(mpCParentColor->GetDoubleIndex(), _PARAM_HERMITE_AUTO, pKey);
    mpCParentColor->GetPActiveItem()->SetKey(CKey, CKeySlave, 1.0/255);
    
    UpdateValues(TRUE);
}

void CustomColorGUI_cls::
OnButtonRemKey(wxCommandEvent &event){
    if (mpCParentColor->GetPActiveItem()->DeleteKey
	((double)mpCParentColor->mIndexInt/255, 1.0/255) == TRUE)
	UpdateValues(TRUE);
}

void CustomColorGUI_cls::
OnButtonAdd(wxCommandEvent &event){
    GCUndoList.OpenBlock();
    MultiParametersGUI_cls::OnButtonAdd(event);
    double pKey[4] = {1, 1, 0, 1};
    
    KeyFrame_cls CKey(4), CKeySlave(0);
    CKey.Set(1, _PARAM_HERMITE_AUTO, pKey);
    mpCParentColor->GetPActiveItem()->SetKey(CKey, CKeySlave, 1.0/255);

    GCUndoList.CloseBlock();
    UpdateValues(TRUE); //$B#22sIA2h$K$J$k$1$I$7$g$&$,$J$$(B?

    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshCombo);
    GetParent()->AddPendingEvent(com);
}

void CustomColorGUI_cls::
OnButtonRemove(wxCommandEvent &event){
    MultiParametersGUI_cls::OnButtonRemove(event);
    UpdateValues(TRUE); //$B#22sIA2h$K$J$k$1$I$7$g$&$,$J$$(B?

    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshCombo);
    GetParent()->AddPendingEvent(com);
}

void CustomColorGUI_cls::
OnButtonColor(wxCommandEvent &event){
    if (mpCParentColor->IsKeyPoint() == FALSE) {
	double pKey[4] = {
	    GetCustomColorR(mpCParentColor->GetDoubleIndex()),
	    GetCustomColorG(mpCParentColor->GetDoubleIndex()),
	    GetCustomColorB(mpCParentColor->GetDoubleIndex()),
	    1.0};
	
	KeyFrame_cls CKey(4), CKeySlave(0);
	CKey.Set(mpCParentColor->GetDoubleIndex(), _PARAM_HERMITE_AUTO, pKey);
	mpCParentColor->GetPActiveItem()->SetKey(CKey, CKeySlave, 1.0/255);
	
    }
    
    mpCParentColor->mCacheIndexInt = mpCParentColor->mIndexInt;
    
    GpCClrDialog->SetColourInt(mpCParentColor->mCacheColorTrim);
			       /*mpCParentColor->mCacheColorTrim[0],
			       mpCParentColor->mCacheColorTrim[1],
			       mpCParentColor->mCacheColorTrim[2]);*/
    if (GpCClrDialog->ShowModal() == wxID_OK){
	mpCParentColor->mCacheColor[0] = GpCClrDialog->GetR();
	mpCParentColor->mCacheColor[1] = GpCClrDialog->GetG();
	mpCParentColor->mCacheColor[2] = GpCClrDialog->GetB();
	mpCParentColor->SyncFromCache();
    }

    UpdateValues();
}

void CustomColorGUI_cls::
OnButtonInvert(wxCommandEvent &event){
    Parameters_cls CNewParam = *mpCParentColor->GetPActiveItem();
    for (int i = 0; i < CNewParam.GetKeyNumber(); i++){
	CNewParam.mVCKeys[i].mDoubleIndex = 1.0 - CNewParam.mVCKeys[i].mDoubleIndex;
    }
    CNewParam.SortKey();
    mpCParentColor->GetPActiveItem()->
	CopyParameterFromP(&CNewParam);
    
    mpCParentColor->UpdateCacheColor();
    mpCParentColor->UpdateColorBar();
    mpCParentColor->UpdateItems(TRUE);
    UpdateValues();

    if (DF3GL_cls::mspCProperty->mCTextureType.mBody == _PREVIEW_RGBA){
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemReTexture);
	GetParent()->AddPendingEvent(com);
    }
}

void CustomColorGUI_cls::
OnButtonChannelShift(wxCommandEvent &event){
    Parameters_cls CNewParam = *mpCParentColor->GetPActiveItem();
    for (int i = 0; i < CNewParam.GetKeyNumber(); i++){
	int dimension = 3; // color$B$O(B3$B$^$G(B
	double value00 = CNewParam.mVCKeys[i].mpValue[0];
	double value01 = CNewParam.mVCKeys[i].mpDValue1[0];
	double value02 = CNewParam.mVCKeys[i].mpDValue2[0];
	for (int j = 0; j < dimension - 1; j++){
	    CNewParam.mVCKeys[i].mpValue[j] = 
		CNewParam.mVCKeys[i].mpValue[j+1];
	    CNewParam.mVCKeys[i].mpDValue1[j] = 
		CNewParam.mVCKeys[i].mpDValue1[j+1];
	    CNewParam.mVCKeys[i].mpDValue2[j] = 
		CNewParam.mVCKeys[i].mpDValue2[j+1];
	}
	CNewParam.mVCKeys[i].mpValue[dimension-1] = value00;
	CNewParam.mVCKeys[i].mpDValue1[dimension-1] = value01;
	CNewParam.mVCKeys[i].mpDValue2[dimension-1] = value02;
    }

    mpCParentColor->GetPActiveItem()->
	CopyParameterFromP(&CNewParam);

    mpCParentColor->UpdateCacheColor();
    mpCParentColor->UpdateColorBar();
    mpCParentColor->UpdateItems(TRUE);
    UpdateValues();

    if (DF3GL_cls::mspCProperty->mCTextureType.mBody == _PREVIEW_RGBA){
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemReTexture);
	GetParent()->AddPendingEvent(com);
    }
}

void CustomColorGUI_cls::
OnKeyDown(wxKeyEvent& event){
    wxPostEvent(GetParent()->GetEventHandler(), event);
    event.Skip();
}

#include "../../MYAPI/MsgBox.h"

/*
  void CustomColorGUI_cls::
  OnSliderParam(wxCommandEvent &event){
  if (event.GetEventType() == wxEVT_SCROLL_THUMBRELEASE)
  mpCParentColor->SyncFromCache();
  }
*/

void CustomColorGUI_cls::
OnSlider(wxCommandEvent &event){
    // $B%3%T!<>e$G<B9T$5$l$k$N$G!"(BmspCThis != this
    mspCThis->UpdateCurrentIndex();
}

void CustomColorGUI_cls::
OnSliderPageUp(wxCommandEvent &event){
    // $B%3%T!<>e$G<B9T$5$l$k$N$G!"(BmspCThis != this
    mspCThis->PageUp();
}

void CustomColorGUI_cls::
OnSliderPageDown(wxCommandEvent &event){
    // $B%3%T!<>e$G<B9T$5$l$k$N$G!"(BmspCThis != this
    mspCThis->PageDown();;
}

void CustomColorGUI_cls::
UpdateActiveOnTL(){
    UpdateValues(TRUE);
}

void CustomColorGUI_cls::
UpdateComboBoxPropagate(){
}

double CustomColorGUI_cls::
GetCustomColorR(double value, void *pClue){
    return mspCThis->mpCParentColor->GetCustomColor(value, 0);
}

double CustomColorGUI_cls::
GetCustomColorG(double value, void *pClue){
    return mspCThis->mpCParentColor->GetCustomColor(value, 1);
}

double CustomColorGUI_cls::
GetCustomColorB(double value, void *pClue){
    return mspCThis->mpCParentColor->GetCustomColor(value, 2);
}

double CustomColorGUI_cls::
GetCustomColorRxD(double value, void *pClue){
    return mspCThis->mpCParentColor->GetCustomColor(value, 0) * value;
}

double CustomColorGUI_cls::
GetCustomColorGxD(double value, void *pClue){
    return mspCThis->mpCParentColor->GetCustomColor(value, 1) * value;
}

double CustomColorGUI_cls::
GetCustomColorBxD(double value, void *pClue){
    return mspCThis->mpCParentColor->GetCustomColor(value, 2) * value;
}

CustomColorGUI_cls  *CustomColorGUI_cls::mspCThis;

void CustomColorGUI_cls::
FuncHookSyncFromPanel(void *WXUNUSED(pArg), void *WXUNUSED(pEvent)){
    mspCThis->mpCParentColor->SyncFromCache();
}

void CustomColorGUI_cls::
FuncHookCombo(void *pArg, void *pEvent){
    if (pArg != mspCThis->mpCMNumParam) return;
    mspCThis->mpCParentColor->SetColorTemplate(mspCThis->mTemplate);
    mspCThis->mpCMNumParam->SetFocusIgnoringChildren();
    //MsgBoxf(mspCThis, "!!", "%d", mspCThis->mTemplate);
}

// ---

CustomColor_cls::
CustomColor_cls(wxWindow *pC, Coloring_cls *pCColoring, wxBitmap *pBitmap) :
    MultiParameters_cls(L"CUSTOMCOLOR_NUM=", L"Color", 4, FALSE)
{
    mpCColoringLink = pCColoring;
    mpBitmapLink = pBitmap;
    mIndexInt = 0;
    mRefreshId = 0;
    double pInit[4] = {1, 0, 0, 1};
    SetInitParam(pInit);

    mCacheIndexInt = 0;
    mCacheColor[0] = mCacheColor[1] = mCacheColor[2] = 127;

    for (int i = 0; i < _CLR_CUSTOM_NUM_MAX; i++)
	mbMultiplyLevel[i] = TRUE;

    InitGUI(pC); 
};

CustomColor_cls::
~CustomColor_cls(){
}

double CustomColor_cls::
GetDoubleIndex(){
    return (double)mIndexInt/255;
}

void CustomColor_cls::
SetToDefault(){
}

void CustomColor_cls::
SetColorTemplate(int index){
    if (index == 0) return;
    // CHECK_ME $B$3$lLa$5$J$/$FBg>fIW!)(B
    GetPActiveItem()->SetFrameNumber(256, 1);
    GCUndoList.OpenBlock();
    switch(index){
    case _CLR_RED_YELLOW: {
	double key0[4] = {1.00, 0.00, 0.00, 1};
	double key1[4] = {1.00, 1.00, 0.00, 1};
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(1.0, key1);
	break; }
    case _CLR_COLD_COLOR: {
	double key0[4] = {0.00, 0.00, 1.00, 1};
	double key1[4] = {0.00, 1.00, 1.00, 1};
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(1.0, key1);
	break; }
    case _CLR_GREEN_YELLOW_RED: {
	double key0[4] = {0.00, 1.00, 0.00, 1};
	double key1[4] = {1.00, 1.00, 0.00, 1};
	double key2[4] = {1.00, 0.00, 0.00, 1};
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(0.5, key1);
	GetPActiveItem()->InsertKey(1.0, key2);
	break; }
    case _CLR_BLACK_WHITE:{
	double key0[4] = {0, 0, 0, 1};
	double key1[4] = {1, 1, 1, 1};
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(1, key1);
	break; }
    case _CLR_MODELATE_GRAY: {
	double key0[4] = {0.50, 0.50, 0.50, 1};
	double key1[4] = {0.75, 0.75, 0.75, 1};
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(1, key1);
	break; }
    case _CLR_SPECTRUM: {
	double key0[4] = {0.03, 0.00, 0.07, 1}; //380
	double key1[4] = {0.32, 0.00, 0.77, 1}; //430
	double key2[4] = {0.00, 0.31, 0.59, 1}; //480
	double key3[4] = {0.00, 0.78, 0.00, 1}; //530
	double key4[4] = {0.82, 0.56, 0.00, 1}; //580
	double key5[4] = {0.84, 0.00, 0.00, 1}; //630
	double key6[4] = {0.26, 0.00, 0.00, 1}; //680
	//double key7[4] = {0.05, 0.00, 0.00, 1}; //730
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(0.167, key1);
	GetPActiveItem()->InsertKey(0.333, key2);
	GetPActiveItem()->InsertKey(0.500, key3);
	GetPActiveItem()->InsertKey(0.666, key4);
	GetPActiveItem()->InsertKey(0.833, key5);
	GetPActiveItem()->InsertKey(1.000, key6);
	break; }
    case _CLR_SPECTRUM_RBG: {
	double key0[4] = {1.00, 0.00, 0.00, 1}; 
	double key1[4] = {0.50, 0.00, 0.50, 1}; 
	double key2[4] = {0.00, 0.00, 1.00, 1}; 
	double key3[4] = {0.00, 0.50, 0.50, 1}; 
	double key4[4] = {0.00, 1.00, 0.00, 1}; 
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(0.250, key1);
	GetPActiveItem()->InsertKey(0.500, key2);
	GetPActiveItem()->InsertKey(0.750, key3);
	GetPActiveItem()->InsertKey(1.000, key4);
	break; }
    case _CLR_SPECTRUM_RBGR: {
	double key0[4] = {1.00, 0.00, 0.00, 1}; 
	double key1[4] = {0.50, 0.00, 0.50, 1}; 
	double key2[4] = {0.00, 0.00, 1.00, 1}; 
	double key3[4] = {0.00, 0.50, 0.50, 1}; 
	double key4[4] = {0.00, 1.00, 0.00, 1}; 
	double key5[4] = {0.50, 0.50, 0.00, 1}; 
	double key6[4] = {1.00, 0.00, 0.00, 1}; 
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(0.167, key1);
	GetPActiveItem()->InsertKey(0.333, key2);
	GetPActiveItem()->InsertKey(0.500, key3);
	GetPActiveItem()->InsertKey(0.666, key4);
	GetPActiveItem()->InsertKey(0.833, key5);
	GetPActiveItem()->InsertKey(1.000, key6);
	break; }
    case _CLR_INCANDESCENT: {
	double key0[4] = {0.00, 0.00, 0.00, 1};
	double key1[4] = {0.73, 0.00, 0.00, 1};
	double key2[4] = {1.00, 0.55, 0.20, 1};
	double key3[4] = {0.97, 0.94, 0.07, 1};
	double key4[5] = {0.97, 0.97, 0.83, 1};
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(0.5940, key1);
	GetPActiveItem()->InsertKey(0.8097, key2);
	GetPActiveItem()->InsertKey(0.8998, key3);
	GetPActiveItem()->InsertKey(1.0000, key4);
	break; }
    case _CLR_COLD_STEEL: {
	double key0[4] = {1.00, 1.00, 1.00, 1};
	double key1[4] = {0.05, 0.02, 0.14, 1};
	double key2[4] = {0.36, 0.73, 0.76, 1};
	GetPActiveItem()->InitSet(key0);
	GetPActiveItem()->InsertKey(0.6477, key1);
	GetPActiveItem()->InsertKey(1.0000, key2);
	break; }
    }
    mpCGUI->mTemplate = 0;
    GCUndoList.CloseBlock();
    UpdateCacheColor();
    UpdateColorBar();
    UpdateItems(TRUE);
    mpCGUI->UpdateValues();

    if (DF3GL_cls::mspCProperty->mCTextureType.mBody == _PREVIEW_RGBA){
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemReTexture);
	mpCGUI->GetParent()->AddPendingEvent(com);
    }
}

void CustomColor_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new CustomColorGUI_cls(this, pC,
				    mCScreenRect.GetPosition(),
				    mCScreenRect.GetSize());
    SetPDialog(mpCGUI);
}

void CustomColor_cls::
UpdateItems(bool bForce){
    if (mpCGUI != NULL) mpCGUI->mpCMNumParam->UpdateValuesAll(bForce);
    mRefreshId++; 
}

bool CustomColor_cls::
IsRefreshed(int id){
    if (mRefreshId == id) return FALSE;
    return TRUE;
}

bool CustomColor_cls::
IsShortClrMapPossible(int index){
    if (mVCParam[index].mVCKeys.size() <= 2)
	return TRUE;
    return FALSE;
}

bool CustomColor_cls::
IsMultiplyLevel(int index){
    return mbMultiplyLevel[index];
}

void CustomColor_cls::
AddItemPropagate(){
    if (mpCGUI != NULL)	mpCGUI->UpdateValues();
}

void CustomColor_cls::
RemoveItemPropagate(){
    if (mpCGUI != NULL) mpCGUI->UpdateValues();
}

void CustomColor_cls::
ReadStrLinePropagate(){
}

bool CustomColor_cls::
IsKeyPoint(){
    if (GetPActiveItem() == NULL) return FALSE;
    return
	GetPActiveItem()->GetKeyFrameIndexFromDouble
	(GetDoubleIndex(), 1.0/255) != -1;
}

double CustomColor_cls::
GetCustomColor(double value, unsigned int index, bool bTrim, int customIndex){
    if (mCItemNum.mBody == 0) return 0.5;
    if (customIndex >= mCItemNum.mBody) customIndex = mActiveItem;

    double color;
    if (customIndex == -1)
	color = mVCParam[mActiveItem].GetPInterpolatedValue(value, 1.0/255)[index];
    else
	color = mVCParam[customIndex].GetPInterpolatedValue(value, 1.0/255)[index];
    if (bTrim == TRUE){
	if (color < 0) color = 0;
	if (color > 1) color = 1;
    }
    return color;
}

std::vector < float > CustomColor_cls::
GetV4CustomColor(double value, bool bTrim, int customIndex, float factor){
    std::vector < float > VRet(4); VRet[3] = 1;
    LOOP_RGB VRet[rgb] = 
	GetCustomColor(value, rgb, bTrim, customIndex) * factor;
    return VRet;
}

void CustomColor_cls::
UpdateCacheColor(){
    LOOP_RGB {
	mCacheColor[rgb] = (GetCustomColor(GetDoubleIndex(), rgb, FALSE)*255)+0.1;
	mCacheColorTrim[rgb] = GetCustomColor(GetDoubleIndex(), rgb, TRUE)*255;
    }
    mbCacheMultiplyLevel = mbMultiplyLevel[mActiveItem];
}

wxString Coloring_cls::
GetClrMapItem(int clrindex, int df3index, float clrMap, 
	      bool bShade, float shadeFactor,
	      bool bMultiplyLevel, float transpalency, wxString strRGBA,
	      bool bHalfMapping, bool bLevelAdjust, bool bShadeDiminish,
	      float monoClr[3]){
    float clrMapLeveled;
    float clr[4], clrinv[4];
    wxString strClrItem;

    // FIX_ME!! $B%l%Y%kJd@5$,MW$k$+$I$&$+$NH=Dj$O$I$&$9$k!)(B
    if (bLevelAdjust == TRUE)
	clrMapLeveled = 
	    PreviewProperty_cls::GetLevelMapping(clrMap, df3index);
    else
	clrMapLeveled = clrMap;
    
    if (clrMap == 1) clrMapLeveled = 0; 
    
    if (clrindex == 0 || monoClr != NULL)
	LOOP_RGB clr[rgb] = monoClr[rgb];
    else
	GetColor(clr, clrMapLeveled, clrindex);

    if (bMultiplyLevel == TRUE)
	LOOP_RGB clr[rgb] *= clrMapLeveled; 
    
    if (bShade == TRUE){
	float norm = sqrt(clr[0]*clr[0] + clr[1]*clr[1] + clr[2]*clr[2]);
	if (norm != 0){
	    LOOP_RGB clrinv[rgb] = clr[rgb] / norm; 
	    LOOP_RGB clrinv[rgb] = 1.0 - clrinv[rgb];
	    float norm2 = sqrt(clrinv[0]*clrinv[0] + clrinv[1]*clrinv[1] + clrinv[2]*clrinv[2]);
	    LOOP_RGB clrinv[rgb] /= norm2;
	    LOOP_RGB clrinv[rgb] *= norm ;
	    
	    LOOP_RGB clr[rgb] = clr[rgb] + shadeFactor * (clrinv[rgb] - clr[rgb]);
	}
	if (bShadeDiminish == TRUE) // $B5[<}$rH>J,$^$G(B
	    LOOP_RGB clr[rgb] *= 1.0f - clrMapLeveled * 0.5;
    }
    
    if (bHalfMapping == TRUE) clrMap = clrMap / 2 + 0.5;

    float clrMax = 0;
    LOOP_RGB if (clr[rgb] > clrMax) clrMax = clr[rgb];
    int adjust = 0;
    if (clrMax < 0.1) adjust = 1;
    if (clrMax < 0.01) adjust = 2;
    if (clrMax == 0) adjust = 0;

    if (transpalency == 0)
	strClrItem << L"\t[" << wxString::Format(wxT("%.3f"), clrMap) <<
	    PovOut3f(strRGBA, clr, 3 + adjust) << L"]" << ENDLWX;
    else{
	clr[3] = transpalency;
	strClrItem << L"\t[" << wxString::Format(wxT("%.3f"), clrMap) <<
	    PovOut4f(strRGBA, clr, 4 + adjust) << L"]" << ENDLWX;
    }

    return strClrItem;
}

wxString CustomColor_cls::
GetShortClrMap(int index, int df3index, bool bShade, float shadeFactor, 
	       bool bMultiplyLevel, float transpalency, wxString strRGBA,
	       bool bHalfMapping, bool bLevelAdjust, bool bShadeDiminish){
    wxString strMap;
    float clrMap;
    for (int i = 0; i < mVCParam[index].mVCKeys.size(); i++){
	clrMap = mVCParam[index].mVCKeys[i].mDoubleIndex;
	if (clrMap > 0.998) clrMap = 0.998;
	strMap << mpCColoringLink->
	    GetClrMapItem(index + _CLR_CUSTOM, df3index, clrMap, bShade, shadeFactor, 
			  bMultiplyLevel, transpalency, strRGBA,
			  bHalfMapping, bLevelAdjust, bShadeDiminish);
    }
    return strMap;
}

void CustomColor_cls::
SyncFromCache(){
    mbMultiplyLevel[mActiveItem] = mbCacheMultiplyLevel;

    double pKey[4] = {
	(double)mCacheColor[0]/255,
	(double)mCacheColor[1]/255,
	(double)mCacheColor[2]/255,
	1.0};
   
    if (mIndexInt == mCacheIndexInt){
	double doubleIndex = GetDoubleIndex();
	int index = GetPActiveItem()->
	    GetKeyFrameIndexFromDouble(doubleIndex, 1.0/255);
	double *pKeyFrameValue = GetPActiveItem()->GetPKeyFrameValue(index);
	if (pKeyFrameValue == NULL) return;
	if (pKeyFrameValue[0] == pKey[0] &&
	    pKeyFrameValue[1] == pKey[1] &&
	    pKeyFrameValue[2] == pKey[2]) return;
	
	KeyFrame_cls CKey(4), CKeySlave(0);
	CKey.Set(GetDoubleIndex(), _PARAM_HERMITE_AUTO, pKey);
	GetPActiveItem()->SetKey(CKey, CKeySlave, 1.0/255);
	
	mpCGUI->UpdateValues(TRUE);
    }
    else{
	int keyIndex = 
	    GetPActiveItem()->
	    GetKeyFrameIndexFromDouble((double)mCacheIndexInt/255, 1.0/255);
	//MsgBoxf(NULL, "key", "%d", keyIndex);
	if (keyIndex == -1) return;
	GetPActiveItem()->MoveKey
	    (keyIndex, (double)(mIndexInt - mCacheIndexInt)/255, 255);
	mpCGUI->mpCSlider->SetValueInt(mIndexInt);
	mpCGUI->UpdateCurrentIndex(TRUE);
    }

    UpdateItems();

    if (DF3GL_cls::mspCProperty->mCTextureType.mBody == _PREVIEW_RGBA){
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemReTexture);
	mpCGUI->GetParent()->AddPendingEvent(com);
    }
}

void CustomColor_cls::
UpdateColorBar(){
    if (mpCGUI != NULL){
	mpCGUI->RemakeIconList();
	mpCGUI->mpCPanelColorBar->Refresh(FALSE);
	mpCGUI->mpCPanelColorBar2->Refresh(FALSE);
    }
}

// ----------------------------------------

Coloring_cls::
Coloring_cls(){
    std::vector < float > VColor0_0; PushBack4(VColor0_0, 1, 1, 1, 1);
    
    std::vector < float > VColor1_0; PushBack4(VColor1_0, 1, 0, 0, 1);
    std::vector < float > VColor1_1; PushBack4(VColor1_1, 1, 1, 0, 0.5);
    std::vector < float > VColor1_2; PushBack4(VColor1_2, 1, 1, 1, 0.25);
    std::vector < float > VColor2_0; PushBack4(VColor2_0, 1, 0, 0, 1);
    std::vector < float > VColor2_1; PushBack4(VColor2_1, 1, 0, 1, 0.5);
    std::vector < float > VColor2_2; PushBack4(VColor2_2, 1, 1, 1, 0.25);
    std::vector < float > VColor3_0; PushBack4(VColor3_0, 1, 0, 0, 1);
    std::vector < float > VColor3_1; PushBack4(VColor3_1, 0, 1, 1, 0.5);
    std::vector < float > VColor3_2; PushBack4(VColor3_2, 1, 1, 1, 0.25);

    std::vector < float > VColor4_0; PushBack4(VColor4_0, 0, 1, 0, 1);
    std::vector < float > VColor4_1; PushBack4(VColor4_1, 0, 1, 1, 0.5);
    std::vector < float > VColor4_2; PushBack4(VColor4_2, 1, 1, 1, 0.25);
    std::vector < float > VColor5_0; PushBack4(VColor5_0, 0, 1, 0, 1);
    std::vector < float > VColor5_1; PushBack4(VColor5_1, 1, 1, 0, 0.5);
    std::vector < float > VColor5_2; PushBack4(VColor5_2, 1, 1, 1, 0.25);
    std::vector < float > VColor6_0; PushBack4(VColor6_0, 0, 1, 0, 1);
    std::vector < float > VColor6_1; PushBack4(VColor6_1, 1, 0, 1, 0.5);
    std::vector < float > VColor6_2; PushBack4(VColor6_2, 1, 1, 1, 0.25);

    std::vector < float > VColor7_0; PushBack4(VColor7_0, 0, 0, 1, 1);
    std::vector < float > VColor7_1; PushBack4(VColor7_1, 1, 0, 1, 0.5);
    std::vector < float > VColor7_2; PushBack4(VColor7_2, 1, 1, 1, 0.25);
    std::vector < float > VColor8_0; PushBack4(VColor8_0, 0, 0, 1, 1);
    std::vector < float > VColor8_1; PushBack4(VColor8_1, 0, 1, 1, 0.5);
    std::vector < float > VColor8_2; PushBack4(VColor8_2, 1, 1, 1, 0.25);
    std::vector < float > VColor9_0; PushBack4(VColor9_0, 0, 0, 1, 1);
    std::vector < float > VColor9_1; PushBack4(VColor9_1, 1, 1, 0, 0.5);
    std::vector < float > VColor9_2; PushBack4(VColor9_2, 1, 1, 1, 0.25);

    //std::vector < float > VColor10_0; PushBack4(VColor10_0, 1, 0.5, 0.25, 1);

    std::vector < std::vector < float > > ColorTemplate0;
    ColorTemplate0.push_back(VColor0_0);

    std::vector < std::vector < float > > ColorTemplate1;
    ColorTemplate1.push_back(VColor1_0);
    ColorTemplate1.push_back(VColor1_1);
    ColorTemplate1.push_back(VColor1_2);

    std::vector < std::vector < float > > ColorTemplate2;
    ColorTemplate2.push_back(VColor2_0);
    ColorTemplate2.push_back(VColor2_1);
    ColorTemplate2.push_back(VColor2_2);

    std::vector < std::vector < float > > ColorTemplate3;
    ColorTemplate3.push_back(VColor3_0);
    ColorTemplate3.push_back(VColor3_1);
    ColorTemplate3.push_back(VColor3_2);

    std::vector < std::vector < float > > ColorTemplate4;
    ColorTemplate4.push_back(VColor4_0);
    ColorTemplate4.push_back(VColor4_1);
    ColorTemplate4.push_back(VColor4_2);

    std::vector < std::vector < float > > ColorTemplate5;
    ColorTemplate5.push_back(VColor5_0);
    ColorTemplate5.push_back(VColor5_1);
    ColorTemplate5.push_back(VColor5_2);

    std::vector < std::vector < float > > ColorTemplate6;
    ColorTemplate6.push_back(VColor6_0);
    ColorTemplate6.push_back(VColor6_1);
    ColorTemplate6.push_back(VColor6_2);

    std::vector < std::vector < float > > ColorTemplate7;
    ColorTemplate7.push_back(VColor7_0);
    ColorTemplate7.push_back(VColor7_1);
    ColorTemplate7.push_back(VColor7_2);

    std::vector < std::vector < float > > ColorTemplate8;
    ColorTemplate8.push_back(VColor8_0);
    ColorTemplate8.push_back(VColor8_1);
    ColorTemplate8.push_back(VColor8_2);

    std::vector < std::vector < float > > ColorTemplate9;
    ColorTemplate9.push_back(VColor9_0);
    ColorTemplate9.push_back(VColor9_1);
    ColorTemplate9.push_back(VColor9_2);

    /*
      std::vector < std::vector < float > > ColorTemplate10;
      ColorTemplate10.push_back(VColor10_0);
    */

    mVVVAColors.push_back(ColorTemplate0);
    mVVVAColors.push_back(ColorTemplate1);
    mVVVAColors.push_back(ColorTemplate2);
    mVVVAColors.push_back(ColorTemplate3);
    mVVVAColors.push_back(ColorTemplate4);
    mVVVAColors.push_back(ColorTemplate5);
    mVVVAColors.push_back(ColorTemplate6);
    mVVVAColors.push_back(ColorTemplate7);
    mVVVAColors.push_back(ColorTemplate8);
    mVVVAColors.push_back(ColorTemplate9);
    
    mpThis = this;
}

Coloring_cls::
~Coloring_cls(){
    delete mpCCustomColor;
}

void Coloring_cls::
PushBack4(std::vector < float > &Vec, float r, float g, float b, float a){
    Vec.push_back(r);
    Vec.push_back(g);
    Vec.push_back(b);
    Vec.push_back(a);
}

unsigned int Coloring_cls::
GetPVVSize(int colorType){
    //if (colorType == -1) colorType = mCIndex.mBody;
    if (colorType >= _CLR_CUSTOM)
	return 3;
    return (unsigned int)GetPVV(colorType)->size();
}

void Coloring_cls::
CopyRGB(float pRGB[3], int cindex, int colorType){
    //if (colorType == -1) colorType = mCIndex.mBody;
    if (colorType >= _CLR_CUSTOM){
	LOOP_RGB pRGB[rgb] = 0;
	for (int i = 0; i <=  cindex; i++){
	    LOOP_RGB pRGB[rgb] /= 2;
	    LOOP_RGB pRGB[rgb] += mpCCustomColor->GetCustomColor((double)(cindex+1)/3, rgb, TRUE, colorType - _CLR_CUSTOM);
	}
	return;
    }

    if (colorType == 0){ // $BC1?'MQ$N$H$j$"$($:$N=hM}(B
	pRGB[0] = mVVVAColors[colorType][0][0];
	pRGB[1] = mVVVAColors[colorType][0][1];
	pRGB[2] = mVVVAColors[colorType][0][2];
    }
    else{
	pRGB[0] = mVVVAColors[colorType][cindex][0];
	pRGB[1] = mVVVAColors[colorType][cindex][1];
	pRGB[2] = mVVVAColors[colorType][cindex][2];
    }
}

float Coloring_cls::
GetAlphaFactor(int cindex, int colorType){
 
    if (colorType == -1 || colorType == 0) return 1;
    if (colorType >= _CLR_CUSTOM)
	return (double)(3-cindex)/6;
    else
	return (*GetPVV(colorType))[cindex][_CLR_ALPHA];
}

std::vector < std::vector < float > >* Coloring_cls::
GetPVV(int colorType){
    if (colorType >= _CLR_CUSTOM)
	return &mVVVAColors[0];
    return &mVVVAColors[colorType];
}

void Coloring_cls::
GetColor(float pRGB[3], float value, int clrindex, bool bTrim){
    pRGB[0] = pRGB[1] = pRGB[2] = 0;
    //if (value == 0) return;
    
    if (clrindex >= _CLR_CUSTOM){
	LOOP_RGB pRGB[rgb] = mpCCustomColor->GetCustomColor(value, rgb, bTrim, clrindex - _CLR_CUSTOM);
	return;
    }

    if (clrindex == 0){
	pRGB[0] = pRGB[1] = pRGB[2] = 1;
	return;
    }
    
    for (unsigned int cindex = 0; cindex < GetPVV(clrindex)->size(); cindex++){
	float color[3];
	CopyRGB(color, cindex, clrindex);

	float coef = 0;
	switch(cindex){
	case 0: coef = 1 - value * value; break;
	case 1: coef = 4 * (0.25f - (value - 0.5f) * (value - 0.5f)) ; break;
	case 2: coef = value * value   ; break;
	}

	for (int clrIndex = 0; clrIndex < 3; clrIndex++)
	    pRGB[clrIndex] += color[clrIndex] * coef; 
    }
    
    float norm = sqrt(pRGB[0]*pRGB[0] + pRGB[1]*pRGB[1] + pRGB[2]*pRGB[2]);
    
    if (norm == 0){ // $B%0%l%$%9%1!<%k$N;~H/@8(B
	pRGB[0] = pRGB[1] = pRGB[2] = 1;
	norm = sqrt(3.0); 
    }

    pRGB[0] = pRGB[0] / norm;
    pRGB[1] = pRGB[1] / norm;
    pRGB[2] = pRGB[2] / norm;
    
}

wxString Coloring_cls::
GetPovStrColorMapMacro(int df3index, int clrindex, bool bShade, float shadeFactor, float monoClr[3], bool bMultiplyLevel, unsigned int shadePattern){
    wxString strMap;
    strMap << L"// clrindex=" << df3index << clrindex << ENDLWX;
    strMap << L"#macro CLRMAP" << df3index << clrindex;
    if (bShade == TRUE)
	strMap << L"SHADE";
    if (bMultiplyLevel == FALSE)
	strMap << L"CLR";
    strMap << L"()" << ENDLWX << 
	GetPovStrColorMap(df3index, clrindex, bShade, shadeFactor, monoClr,
			  bMultiplyLevel, shadePattern);
    strMap << L"#end" << ENDLWX;
    return strMap;
}

wxString Coloring_cls::
GetPovStrPigment(int clrindex, float value, float alpha, int transmitType){
    float clr[4];
    GetColor(clr, value, clrindex);
    wxString strPigment;
    switch(transmitType){
    case 0:
	strPigment << PovOut3(L"color rgb", clr[R], clr[G], clr[B]);
	break;
    case _TRANSMIT_TRANSMIT:
	strPigment << PovOut4(L"color rgbt", clr[R], clr[G], clr[B], 1.0f-alpha);
	break;
    case _TRANSMIT_FILTER:
	strPigment << PovOut4(L"color rgbf", clr[R], clr[G], clr[B], 1.0f-alpha);
	break;
    }
    return strPigment;
}

wxString Coloring_cls::
GetPovStrColorMap(int df3index, int clrindex, bool bShade, float shadeFactor, 
		  float monoClr[3], bool bMultiplyLevel, unsigned int shadePattern,
		  float transpalency, 
		  int transmitType, bool bHalfMapping, bool bLevelAdjust,
		  bool bShortEnable){
    
    wxString strRGBA;
    if (transpalency == 0) transmitType = 0;
    switch (transmitType){
    case 0:
	strRGBA = wxT(" rgb"); transpalency = 0; break;
    case _TRANSMIT_TRANSMIT:
	strRGBA = wxT(" rgbt");	break;
    case _TRANSMIT_FILTER:
	strRGBA = wxT(" rgbf"); break;
    }
    wxString strMap;

    strMap << L"color_map {" << ENDLWX;

    if (bMultiplyLevel == FALSE && // $BG;EY$r$+$1$k$H$-$O>JN,IT2D(B
	bShortEnable == TRUE &&
	clrindex >= _CLR_CUSTOM &&
	mpCCustomColor->IsShortClrMapPossible(clrindex - _CLR_CUSTOM) == TRUE){
	
	strMap << mpCCustomColor->GetShortClrMap
	    (clrindex - _CLR_CUSTOM, df3index, bShade, shadeFactor, 
	     bMultiplyLevel,
	     transpalency, strRGBA, bHalfMapping, bLevelAdjust, 
	     shadePattern);
	strMap << wxT("}") ;
	return strMap;
    }
    
    //float clr0[4];
    //float clrMapLeveled;
    bool bMultiplyLevel2 = TRUE;
    if (clrindex >= _CLR_CUSTOM) 
	mpCCustomColor->IsMultiplyLevel(clrindex - _CLR_CUSTOM);
    
    for (float clrMap = 0; clrMap <= 1.001; clrMap += 0.02f){
	if (clrMap >= 0.999) clrMap = 0.999;
	strMap << GetClrMapItem
	    (clrindex, df3index, clrMap, bShade, shadeFactor, 
	     bMultiplyLevel * bMultiplyLevel2, transpalency, strRGBA,
	     bHalfMapping, bLevelAdjust, shadePattern, monoClr);
    }
    
    if (bMultiplyLevel == FALSE)
	strMap << GetClrMapItem(clrindex, df3index, 1.00, bShade, shadeFactor, 
				bMultiplyLevel, transpalency, strRGBA,
				bHalfMapping, bLevelAdjust, shadePattern, monoClr);
    
    strMap << L"}" << ENDLWX;

    return strMap;
}

wxString Coloring_cls::
GetPovStrClr(int df3index, int clrindex, float level, 
	       float transpalency, int transmitType){
    wxString strRGBA;
    if (transpalency == 0) transmitType = 0;
    switch (transmitType){
    case 0:
	strRGBA = wxT(" rgb"); transpalency = 0; break;
    case _TRANSMIT_TRANSMIT:
	strRGBA = wxT(" rgbt");	break;
    case _TRANSMIT_FILTER:
	strRGBA = wxT(" rgbf"); break;
    }

    wxString strClr;
    strClr << L"color ";
    float clr[4];

    GetColor(clr, level, clrindex);
    
    if (transpalency == 0)
	strClr << PovOut3f(strRGBA, clr, 3);
    else{
	clr[3] = transpalency;
	strClr << PovOut4f(strRGBA, clr, 4);
    }

    return strClr;
}

Coloring_cls* Coloring_cls::mpThis;
Coloring_cls  GCColoring;
