// -*- c -*- 
// "clsDF3GL.h"
// (setq c-basic-offset 4)

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
extern PFNGLTEXIMAGE3DPROC glTexImage3D;
extern PFNGLCOMPRESSEDTEXIMAGE3DPROC glTexCompressedImage3D;
extern PFNGLBLENDCOLORPROC glBlendColor;
extern PFNGLBLENDEQUATIONEXTPROC glBlendEquationEXT;
extern PFNGLCOLORTABLEEXTPROC glColorTableEXT;
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "clsColoring.h"
extern Coloring_cls GCColoring;
#include "clsDF3GL.h"
#include "clsDraw3DTexBoxA.h"
#include "clsDraw3DTexBoxB.h"
#include "clsPreviewProperty.h"
#include "../clsDF3SeqCtrl.h"
#include "../clsDF3Sequence.h"

#include "../../COMMON/MyVector.h"
#include "../../FILTER/clsFilterCtrl.h"
#include "../../MYGL/clsPositionAndDirection.h"
#include "../../MAINCANVAS/clsPreference.h"

#include "../../../clsDrawingBox.h"

#include <time.h>
#include <algorithm>

extern bool GbGLLockForMultithread;
extern bool GbGLForceWaitThreadEnd;

#define X 0
#define Y 1
#define Z 2

#define S 0
#define T 1

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#ifdef WIN32
#define _FLAG_SET_CURRENT
#endif
#ifdef __WXGTK__
#define _FLAG_SET_CURRENT
#endif

#define ENDL std::endl

#include <assert.h>

DF3GLInfo_cls::
DF3GLInfo_cls(){
    LOOP_XYZ mClipBoxMin[xyz] = 0;
    LOOP_XYZ mClipBoxMax[xyz] = 0;
}

DF3GLInfo_cls::
DF3GLInfo_cls(PositionAndDirection_cls CCam,
	      double clipBoxMin[3], double clipBoxMax[3]){
    mCCam = CCam;
    //LOOP_XYZ mbMirror[xyz] = bMirror[xyz];
    LOOP_XYZ mClipBoxMin[xyz] = clipBoxMin[xyz];
    LOOP_XYZ mClipBoxMax[xyz] = clipBoxMax[xyz];
}

bool operator==(const DF3GLInfo_cls& CL, const DF3GLInfo_cls& CR){
    bool retval = TRUE;
    if (!(CL.mCCam == CR.mCCam)) retval = FALSE;
    //LOOP_XYZ if (CL.mbMirror[xyz] != CR.mbMirror[xyz]) retval = FALSE;
    LOOP_XYZ if (CL.mClipBoxMin[xyz] != CR.mClipBoxMin[xyz]) retval = FALSE;
    LOOP_XYZ if (CL.mClipBoxMax[xyz] != CR.mClipBoxMax[xyz]) retval = FALSE;
    return retval;
}

DF3GL_cls::
DF3GL_cls(DF3Sequence_cls* pCParent) : 
    DF3_cls(pCParent)
{
    //mDownSizeLayer = 1;
    mOldPointTex = 0;
    mbFakeDeleted = FALSE;
    m3DTexCompressed = 0;
    m3DTexUsedByte = 0;
}

DF3GL_cls::
~DF3GL_cls(){
    if (IsTexObjectGenFinished() == TRUE)
	DeleteTexObjects(_DF3_FORCE_FINISH);
}

void DF3GL_cls::
InitializeStaticMembers(wxWindow *pCWindow, DF3SeqCtrl_cls *pCDF3SeqCtrl,
			wxBitmap *pBitmapProp, Preference_cls *pCPreference){
    mspCProperty = 
	new PreviewProperty_cls(pCWindow, pCDF3SeqCtrl, pBitmapProp);
    mspCDF3SeqCtrlLink = pCDF3SeqCtrl;
    mspCDraw3DTexBoxA = 
	new Draw3DTexBoxA_cls(pCDF3SeqCtrl, mspCProperty);
    mspCDraw3DTexBoxB = 
	new Draw3DTexBoxB_cls(pCDF3SeqCtrl, mspCProperty);
    mspCPreferenceLink = pCPreference;
}

bool DF3GL_cls::
IsTexObjectGenFinished(){
    if (mCTexObjectGen[X].IsFinished() == TRUE &&
	mCTexObjectGen[Y].IsFinished() == TRUE &&
	mCTexObjectGen[Z].IsFinished() == TRUE) return TRUE;
    return FALSE; 
}

bool DF3GL_cls::
IsTexObjectGenInProcess(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectGen[xyz].IsInProgress();
    return r; }
bool DF3GL_cls::
IsTexObjectGenFinishConditionSatisfied(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectGen[xyz].IsFinishConditionSatisfied();
    return r; }


bool DF3GL_cls::
IsTexObjectBindFinished(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectBind[xyz].IsFinished();
    return r; }
bool DF3GL_cls::
IsTexObjectBindInProgress(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectBind[xyz].IsInProgress();
    return r; }
bool DF3GL_cls::
IsTexObjectBindFinishConditionSatisfied(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectBind[xyz].IsFinishConditionSatisfied();
    return r; }


bool DF3GL_cls::
IsTexObjectDeleteFinished(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectDelete[xyz].IsFinished();
    return r; }
bool DF3GL_cls::
IsTexObjectDeleteInProgress(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectDelete[xyz].IsInProgress();
    return r; }
bool DF3GL_cls::
IsTexObjectDeleteFinishConditionSatisfied(){
    bool r = TRUE; LOOP_XYZ r &= mCTexObjectDelete[xyz].IsFinishConditionSatisfied();
    return r; }

// ----------------

bool DF3GL_cls::
GenerateTexObjects(bool type, bool bFakable){
    if (mspCProperty->mbUse3DTexture == FALSE){
	assert(IsAllocatedXY() == TRUE);
	assert(IsXYConstructFinished() == TRUE);
    }

    // TexDelete$BESCf$@$C$?$i$^$:$O=*$o$i$;$k(B
    if (IsTexObjectDeleteInProgress() == TRUE) 
	DeleteTexObjects(_DF3_FORCE_FINISH); 
    
    assert(IsTexObjectGenFinished() == FALSE);
    
    mOldPointTex = 0;
    
    LOOP_XYZ
	mCTexObjectGen[xyz].Initialize(GetCurrSize(xyz));
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glEnable(GL_TEXTURE_2D);
    
    LOOP_XYZ{
	if (mCTexObjectGen[xyz].mNext == 0)
	    mV2Texture[xyz].resize(GetCurrSize(xyz));;

	for (unsigned int index = mCTexObjectGen[xyz].mNext;index < mV2Texture[xyz].size(); index ++){
	    
	    glGenTextures(1, &mV2Texture[xyz][index]);
			  
	    mCTexObjectGen[xyz].mNext = index + 1;
	    if (mCTexObjectGen[xyz].IsResumeConditionSatisfied(type) == TRUE)
		goto DUAL_LOOP_EXITA;
	}
    }
    
 DUAL_LOOP_EXITA:

    glPopAttrib();
    
    if (IsTexObjectGenFinishConditionSatisfied() == FALSE) return FALSE;
    
    LOOP_XYZ
	mCTexObjectDelete[xyz].Clear();

    LOOP_XYZ
	mCTexObjectGen[xyz].Finalize();

    if (mspCProperty->mbUse3DTexture == TRUE){
	if (type == _DF3_FORCE_FINISH && 
	    bFakable == TRUE && mbFakeDeleted == TRUE)
	    mbFakeDeleted = FALSE;
	else
	    glGenTextures(1, &mTexture3D);
    }

    mOldPointTex = 1;
    
    return TRUE;
}

bool DF3GL_cls::
BindTexObjects(bool type){
    if (mspCProperty->mbUse3DTexture == FALSE){
	assert(IsAllocatedXY() == TRUE);
	assert(IsXYConstructFinished() == TRUE);
    }
    assert(IsTexObjectGenFinished() == TRUE);

    LOOP_XYZ
	mCTexObjectBind[xyz].Initialize((unsigned int)mV2Texture[xyz].size());
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glEnable(GL_TEXTURE_2D);

    bool bCompress = (mspCPreferenceLink->mCTexCompress.mBody == 1);

    LOOP_XYZ{
	for (unsigned int index = mCTexObjectBind[xyz].mNext
		 ;index < mV2Texture[xyz].size(); index++){
	    
	    if (mspCProperty->mbUse3DTexture == FALSE){
		glBindTexture(GL_TEXTURE_2D, mV2Texture[xyz][index]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA,
			     GetCurrSizeS(xyz), GetCurrSizeT(xyz), 0,
			     GL_ALPHA, GL_UNSIGNED_BYTE,
			     &mpMainXYZ[xyz][GetCurrSizeS(xyz)*GetCurrSizeT(xyz)*index]);
	    }
	    
	    mCTexObjectBind[xyz].mNext = index + 1;
	    if (mCTexObjectBind[xyz].IsResumeConditionSatisfied(type) == TRUE)
		goto DUAL_LOOP_EXITB;
	}

    }
    
 DUAL_LOOP_EXITB:

    if (IsTexObjectBindFinishConditionSatisfied() == FALSE){
	glPopAttrib();
	return FALSE;
    }

    //$B$R$H$^$:(B2DTexture$B$,=*$o$C$?$i:G8e$K(B3D$B%F%/%9%A%c$rMQ0U(B
    if (mspCProperty->mbUse3DTexture == TRUE){
	glEnable(GL_TEXTURE_3D);
	int alignment = 8;
	if (mCurrSize[X] % 8 == 0) alignment = 8;
	else if (mCurrSize[X] % 4 == 0) alignment = 4;
	else if (mCurrSize[X] % 2 == 0) alignment = 2;
	else alignment = 1;
	
#ifdef WIN32
	glTexImage3D = (PFNGLTEXIMAGE3DPROC) wglGetProcAddress("glTexImage3D");
	glTexCompressedImage3D = (PFNGLCOMPRESSEDTEXIMAGE3DPROC) wglGetProcAddress("glCompressedTexImage3D");
#endif
	glBindTexture(GL_TEXTURE_3D, mTexture3D);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);

	glPixelStorei(GL_UNPACK_ALIGNMENT, alignment);

	// $B:F=hM}$N@'Hs$O8~$3$&$GH=Dj(B
	ProcessXYZData();
	
	unsigned int totalSize = GetCurrentTotalSize();
	int textureFactor = 1;

	switch (mspCProperty->mCTextureType.mBody){
	case _PREVIEW_ALPHA: {
	    unsigned char *pTgt = mpProcessedXYZ[Z];
	    if (mspCProperty->IsFilterProcessRequired(mpCParent->mSelfIndex) == FALSE)
		pTgt = mpMainXYZ[Z];

	    int format =  bCompress ? GL_COMPRESSED_ALPHA : GL_ALPHA;
	    glTexImage3D(GL_TEXTURE_3D, 0, format, 
			 mCurrSize[X], mCurrSize[Y], mCurrSize[Z], 0,
			 GL_ALPHA, GL_UNSIGNED_BYTE, pTgt);
	    break; }
	case _PREVIEW_LUMINANCE_ALPHA: {
	    textureFactor = 2;
	    unsigned char *pLuminanceAlpha = new unsigned char [totalSize*2];
	    double factor = mspCProperty->mCLuminanceFactor.mBody;
	    
	    if (mspCProperty->IsFilterProcessRequired(mpCParent->mSelfIndex) == FALSE)
		for (unsigned int i = 0; i < totalSize; i++){
		    pLuminanceAlpha[i*2] = (255 - mpMainXYZ[Z][i]) * factor;
		    pLuminanceAlpha[i*2+1] = mpMainXYZ[Z][i];
		}
	    else
		for (unsigned int i = 0; i < totalSize; i++){
		    pLuminanceAlpha[i*2] = (255 - mpProcessedXYZ[Z][i]) * factor;
		    pLuminanceAlpha[i*2+1] = mpProcessedXYZ[Z][i];
		}
	    
	    int format =  bCompress ? GL_COMPRESSED_LUMINANCE_ALPHA : GL_LUMINANCE_ALPHA;
	    glTexImage3D(GL_TEXTURE_3D, 0, format, 
			 mCurrSize[X], mCurrSize[Y], mCurrSize[Z], 0,
			 GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, pLuminanceAlpha);
	    
	    delete pLuminanceAlpha;
	    break; }	
	case _PREVIEW_RGBA: {
	    textureFactor = 4;
	    unsigned char *pRGBA = new unsigned char [totalSize*4];
	    double factor = mspCProperty->mCLuminanceFactor.mBody;
	    int format =  bCompress ? GL_COMPRESSED_RGBA_S3TC_DXT5_EXT : GL_RGBA;
	    
	    int colorType = mpCParent->mpCFilterCtrl->GetVolumeColorType();
	    if (colorType == -1) colorType = 0; // Volume$B$,L5$$$H(B-1$B$G$^$:$$!#(B
	    unsigned char colorTable[256][4];
	    for (int i = 0; i < 256; i++){
		float fcolor[3];
		GCColoring.GetColor(fcolor, (float)i/255, colorType, TRUE);
		// ADD_HOC!! $B2?$GH?E>$7$J$$$H$$$1$J$$!)IA2h$N@_DjJQ!)(B
		LOOP_RGB colorTable[i][rgb] = 
		    255 - (unsigned char)(fcolor[rgb] * 255);
		colorTable[i][3] = i;
	    }
	    
	    unsigned char *pProcessTarget;
	    if (mspCProperty->
		IsFilterProcessRequired(mpCParent->mSelfIndex) == TRUE)
		pProcessTarget = mpProcessedXYZ[Z];
	    else
		pProcessTarget = mpMainXYZ[Z];

	    // -1::$B?'$K<+J,<+?H$rMxMQ(B $B$=$l0J30(B::$B30It%G!<%?(B
	    int dataType = mpCParent->mpCFilterCtrl->GetVolumeDataType() - 1;
	    unsigned char *pProcessTarget2;
	    if (dataType == -1){
		// $B<+J,<+?H$K$OC1$J$k%^%C%T%s%0$G9T$1$k(B
		float ftableR[256], ftableG[256], ftableB[256], ftableA[256];
		for (int i = 0; i < 256; i++){
		    ftableR[i] = (float)colorTable[i][0]/255;
		    ftableG[i] = (float)colorTable[i][1]/255;
		    ftableB[i] = (float)colorTable[i][2]/255;
		    ftableA[i] = (float)colorTable[i][3]/255;
		}
		
		glPixelTransferi( GL_MAP_COLOR, TRUE );
		glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, ftableR );
		glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, ftableG );
		glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, ftableB );
		glPixelMapfv(GL_PIXEL_MAP_I_TO_A, 256, ftableA );
		
		glTexImage3D(GL_TEXTURE_3D, 0, format, 
			     mCurrSize[X], mCurrSize[Y], mCurrSize[Z], 0,
			     GL_COLOR_INDEX, GL_UNSIGNED_BYTE, pProcessTarget);
		
		glPixelTransferi( GL_MAP_COLOR, FALSE );
		
		delete pRGBA;
		break;
	    }
	    else{ 
		if (mspCProperty->IsFilterProcessRequired(dataType) == TRUE)
		    pProcessTarget2 = mspCProperty->
			mpCDF3SeqCtrlLink->GetPDF3Seq(dataType)->GetPDF3()->mpProcessedXYZ[Z];
		else
		    pProcessTarget2 = mspCProperty->
			mpCDF3SeqCtrlLink->GetPDF3Seq(dataType)->GetPDF3()->mpMainXYZ[Z];
	    }
	    
	    printf("pProcessTarget2=%p\n%p\n%p\ntotalSize=%d\n", pProcessTarget2,
		   mspCProperty->
		   mpCDF3SeqCtrlLink->GetPDF3Seq(dataType)->GetPDF3()->mpProcessedXYZ[Z],
		   mspCProperty->
		   mpCDF3SeqCtrlLink->GetPDF3Seq(dataType)->GetPDF3()->mpMainXYZ[Z], totalSize);
	    if (pProcessTarget2 == NULL) { //$BMxMQ$7$?$$%G!<%?$,$^$@=hM}A0(B
		delete pRGBA;
		glPopAttrib();
		// $B$=$b$=$b%U%!%$%k@_DjA0$J$i!"%k!<%W2sHr$N0Y$KI=<($r(B
		if (mspCProperty->mpCDF3SeqCtrlLink->GetPDF3Seq(dataType)->mVCFileSequence.size() == 0) return TRUE;
		return FALSE;
	    }
	    
	    int threadNum = mspCPreferenceLink->GetThreadNumFilter();
	    
	    threadNum = 1; // $BB.$/$J$k!)(B
	    if (threadNum == 1){
		if (dataType == -1)
		    for (unsigned int i = 0, i4 = 0; i < totalSize; i++, i4+=4)
			memcpy(pRGBA+i4, colorTable[pProcessTarget[i]], 4);
		else
		    for (unsigned int i = 0, i4 = 0; i < totalSize; i++, i4+=4){
			memcpy(pRGBA+i4, colorTable[pProcessTarget2[i]], 3);
			pRGBA[i4+3] = pProcessTarget[i];
		    }
	    }
	    else {
		bool bFinished[2] = {FALSE, FALSE};
		ThreadMappingByCPU *pThread[2];
		for (int th = 0; th < 2; th++){
		    pThread[th] = new ThreadMappingByCPU;
		    pThread[th]->SetParameters
			(this, pRGBA, pProcessTarget, pProcessTarget2, 
			 th * totalSize / 2, (th + 1) * totalSize / 2,
			 colorTable, &bFinished[th]);
		}
		for (int th = 0; th < 2; th++) pThread[th]->Run();

		// JOINABLE$B$O(BWait$B$G=hM}$,Ht$s$G$7$^$$(BGL$B$GLq2p(B
		while (bFinished[0] == FALSE) wxMilliSleep(1);
		while (bFinished[1] == FALSE) wxMilliSleep(1);
	    }

	    glTexImage3D(GL_TEXTURE_3D, 0, format, 
			 mCurrSize[X], mCurrSize[Y], mCurrSize[Z], 0,
			 GL_RGBA, GL_UNSIGNED_BYTE, pRGBA);

	    delete pRGBA;
	    break; }
	}
	
	glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_COMPRESSED, &m3DTexCompressed);
	if (m3DTexCompressed == 1)
	    glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_COMPRESSED_IMAGE_SIZE, &m3DTexUsedByte);
	else
	    m3DTexUsedByte = mCurrSize[X] * mCurrSize[Y] * mCurrSize[Z] * textureFactor;
    }

    glPopAttrib();

    LOOP_XYZ 
	mCTexObjectBind[xyz].Finalize();

    //GbGLLockForMultithread = FALSE;   
    return TRUE;
}

bool DF3GL_cls::
DeleteTexObjects(bool type, bool bFakable){
    // TexGen$BESCf$@$C$?$i$^$:$O=*$o$i$;$k(B
    if (IsTexObjectGenInProcess() == TRUE){
	GenerateTexObjects(_DF3_FORCE_FINISH);
	type = _RESUMABLE_FORCE_FINISH;
    }
    
    // Generate$B$5$l$F$J$$!J:G=i!K$O$=$N$^$^5"$k(B
    if (IsTexObjectGenFinished() == FALSE){
	mOldPointTex = 0;
	Clean();
	return TRUE; }

    LOOP_XYZ
	mCTexObjectDelete[xyz].Initialize((unsigned int)mV2Texture[xyz].size());
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glEnable(GL_TEXTURE_2D);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    LOOP_XYZ
	for (unsigned int index = mCTexObjectDelete[xyz].mNext
		 ; index < mV2Texture[xyz].size(); index++){
	    
	    // 2D$B%F%/%9%A%c3+J|(B
	    if (mspCProperty->mbUse3DTexture == FALSE)
		glDeleteTextures(1, &mV2Texture[xyz][index]);
	    
	    mCTexObjectDelete[xyz].mNext = index + 1;
	    if (mCTexObjectDelete[xyz].IsResumeConditionSatisfied(type) == TRUE)
		goto DUAL_LOOP_EXITC;
	}
    
 DUAL_LOOP_EXITC:

    //mOldPointTex ++; $B%G%j!<%HESCf$N$b$N$O$5$i$K3+J|$rM%@h(B
    if (IsTexObjectDeleteFinishConditionSatisfied() == FALSE) {
	glPopAttrib();
	return FALSE;
    }

    // 3D$B%F%/%9%A%c3+J|(B
    if (mspCProperty->mbUse3DTexture == TRUE){
	if (type == _DF3_FORCE_FINISH && bFakable == TRUE)
	    mbFakeDeleted = TRUE;
	else
	    glDeleteTextures(1, &mTexture3D);
    }
    
    LOOP_XYZ{
	mCTexObjectGen[xyz] .Clear();
	mCTexObjectBind[xyz].Clear(); 
    }
    
    Clean();

    glPopAttrib();
    LOOP_XYZ
	mCTexObjectDelete[xyz].Finalize();

    return TRUE;
}

int  DF3GL_cls::
GetTextureBindPercent(){
    if (IsTexObjectDeleteInProgress() == TRUE) 
	return 100 - 
	    33 * mCTexObjectDelete[X].GetPercent() / 100 -
	    33 * mCTexObjectDelete[Y].GetPercent() / 100 -
	    33 * mCTexObjectDelete[Z].GetPercent() / 100;
    if (IsTexObjectBindFinished()) return 100;
    return 
	18 * mCTexObjectGen[X].GetPercent() / 100 +  
	18 * mCTexObjectGen[Y].GetPercent() / 100 +
	18 * mCTexObjectGen[Z].GetPercent() / 100 +
	15 * mCTexObjectBind[X].GetPercent() / 100 +
	15 * mCTexObjectBind[Y].GetPercent() / 100 +
	15 * mCTexObjectBind[Z].GetPercent() / 100;
}

void DF3GL_cls::
DrawDataInitSetting(){
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glPushMatrix();
    
    glDisable(GL_LIGHTING);

    glDisable(GL_CULL_FACE);

    glEnable(GL_TEXTURE_2D);
    //glEnable(GL_TEXTURE_3D); 
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //$B%"%k%U%!(B
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE); // $BC1=c2C;;(B
    
    glDepthMask(GL_FALSE);
    
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
    glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_S , GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_T , GL_CLAMP);
}

void DF3GL_cls::
DrawDataEndSetting(){
    glPopMatrix();
    glPopAttrib();
}

void DF3GL_cls::
SetBoxScale(double boxScaleX, double boxScaleY, double boxScaleZ){
    msBoxScaleX = boxScaleX;
    msBoxScaleY = boxScaleY;
    msBoxScaleZ = boxScaleZ;
}

void DF3GL_cls::
SetBlendFunction(){
    switch (mspCProperty->mCRenderType.mBody){
    case 0:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //$B%"%k%U%!9g@.(B
	break;
    case 1:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE); // $B2C;;(B
	break;
    }
}

void DF3GL_cls::
GetCurrentColour(float clr[3], float clrinv[3]){
    LOOP_RGB clrinv[rgb] = 0;
    LOOP_RGB clr[rgb] = mpCParent->mColorRGB[rgb];

    float norm = sqrt(clr[0]*clr[0] + clr[1]*clr[1] + clr[2]*clr[2]);
    if (norm != 0){
	LOOP_RGB clrinv[rgb] = 1.0 - clr[rgb]/norm;
	float norminv = sqrt(clrinv[0]*clrinv[0] + clrinv[1]*clrinv[1] + clrinv[2]*clrinv[2]);
	LOOP_RGB clrinv[rgb] /= norminv;
	LOOP_RGB clrinv[rgb] *= norm;
    }
}

void DF3GL_cls::
Bind3DTexture(){
    glBindTexture(GL_TEXTURE_3D, mTexture3D);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, 
		    mspCProperty->GetGLTexFilter());
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, 
		    mspCProperty->GetGLTexFilter());
}

void DF3GL_cls::
Draw3DNormal(PositionAndDirection_cls CCam, double alpha, bool bMirror[3],
	     double clipBoxMin[3], double clipBoxMax[3],
	     double *pShadeOffset, bool bDrawVolume){

    mspCDraw3DTexBoxA->CalcAllSlice(CCam, clipBoxMin, clipBoxMax);
    
    float clr[3], clrinv[3];
    GetCurrentColour(clr, clrinv);
    
    Bind3DTexture();
    double fogAlpha = 4*1.7 / mspCProperty->GetNQuality();
    mspCDraw3DTexBoxA->DrawAllSlice(alpha, clr, clrinv, fogAlpha, bDrawVolume,
				    pShadeOffset);
}

// !!!!
void DF3GL_cls::
Draw3DNormalWithOtherGL(PositionAndDirection_cls CCam, double *Aalpha, bool bMirror[3],
			double clipBoxMin[3], double clipBoxMax[3],
			double *pShadeOffset){

    mspCDraw3DTexBoxA->CalcAllSlice(CCam, clipBoxMin, clipBoxMax);

    float clr[_DF3_MAX_SLOT][3];
    float clrinv[_DF3_MAX_SLOT][3];
    bool AbDrawVolume[_DF3_MAX_SLOT];
    int Atexture3D[_DF3_MAX_SLOT];

    for (unsigned int i = 0; i < msVpDF3GL.size(); i++){
	msVpDF3GL[i]->GetCurrentColour(clr[i], clrinv[i]);
	AbDrawVolume[i] = msVpDF3GL[i]->mpCParent->mpCFilterCtrl->
	    GetVolumeValidity();
	if (AbDrawVolume[i] == TRUE)
	    Atexture3D[i] = msVpDF3GL[i]->mTexture3D;
	else
	    Atexture3D[i] = -1;
    }
    
    double fogAlpha = 4*1.7 / mspCProperty->GetNQuality();
    mspCDraw3DTexBoxA->DrawAllSliceWithOtherGL
	(Aalpha, (int)msVpDF3GL.size(), mspCProperty->GetGLTexFilter(),
	 clr, clrinv, Atexture3D, fogAlpha, AbDrawVolume,
	 pShadeOffset);
}

void DF3GL_cls::
DrawSlice3D(unsigned int axis, bool order,
	    double clipBoxMin[3], double clipBoxMax[3], float alpha,
	    double *pShadeOffset){
    
    float clr[3], clrinv[3];
    GetCurrentColour(clr, clrinv);

    mspCDraw3DTexBoxB->DrawSlice3D(axis, GetCurrSize(axis), order, clr, clrinv, 
				   clipBoxMin, clipBoxMax, alpha, pShadeOffset);
}

void DF3GL_cls::
DrawSlice3DWithOtherGL(unsigned int axis, bool order,
		       double clipBoxMin[3], double clipBoxMax[3], 
		       double *Aalpha,
		       double *pShadeOffset){

    float (*clr)[3] = new float [msVpDF3GL.size()][3];
    float (*clrinv)[3] = new float [msVpDF3GL.size()][3];
    int *Atexture3D = new int [msVpDF3GL.size()];
    for (unsigned int i = 0; i < msVpDF3GL.size(); i++){
	msVpDF3GL[i]->GetCurrentColour(clr[i], clrinv[i]);
	Atexture3D[i] = msVpDF3GL[i]->mTexture3D;
    }

    mspCDraw3DTexBoxB->DrawSlice3DWithOtherGL
	(axis, (int)msVpDF3GL.size(), GetCurrSize(axis), order, 
	 clr, clrinv, Atexture3D,
	 clipBoxMin, clipBoxMax, Aalpha, pShadeOffset);
    
    delete [] clr; 
    delete [] clrinv;
    delete [] Atexture3D;
}

void DF3GL_cls::
DrawSliceST(double coord, 
	    double sMin, double sMax, double tMin, double tMax){
    glBegin(GL_QUADS);
    glTexCoord2f(sMin, tMin); glVertex3d(sMin, tMin, coord);
    glTexCoord2f(sMax, tMin); glVertex3d(sMax, tMin, coord);
    glTexCoord2f(sMax, tMax); glVertex3d(sMax, tMax, coord);
    glTexCoord2f(sMin, tMax); glVertex3d(sMin, tMax, coord);
    glEnd();
}

void DF3GL_cls::
DrawSlice2D(unsigned int axis, int index,
	    double sMin, double sMax, double tMin, double tMax,
	    float clr[3], float clrinv[3], float alpha){

    glBindTexture(GL_TEXTURE_2D, mV2Texture[axis][index]);
    double coord = (double)(index + 0.5) / GetCurrSize(axis);

    if (mspCProperty->mCShadingType.mBody == 1){
	
	glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	float clrshade[3];
	LOOP_RGB clrshade[rgb] = clr[rgb] + mspCProperty->mCAbsorptionComplement.mBody * (clrinv[rgb] - clr[rgb]);
    
	if (mspCProperty->mCAbsorption.mBody != 0){
	    glColor4f(clrshade[0], clrshade[1], clrshade[2], 
		      alpha * mspCProperty->mCAbsorption.mBody);
	    DrawSliceST(coord, sMin, sMax, tMin, tMax);
	}
    }
    
    SetBlendFunction();
    glColor4f(clr[0], clr[1], clr[2], alpha);
    if (glBlendEquationEXT != NULL) glBlendEquationEXT(GL_FUNC_ADD_EXT);
    DrawSliceST(coord, sMin, sMax, tMin, tMax);
}

void DF3GL_cls::
DrawSlice(unsigned int axis, bool order,
	  double clipBoxMin[3], double clipBoxMax[3], float alpha){
    
    float clr[3], clrinv[3];
    GetCurrentColour(clr, clrinv);

    glTranslated(-0.5, -0.5, -0.5);
    
    double sMin = 0; //1. / (GetCurrSizeS(axis) + 1);
    double tMin = 0; //1. / (GetCurrSizeS(axis) + 1);
    double sMax = 1; //(double)GetCurrSizeS(axis) / (GetCurrSizeS(axis) + 1);
    double tMax = 1; //(double)GetCurrSizeT(axis) / (GetCurrSizeT(axis) + 1);

    int indexMin = clipBoxMin[axis] * GetCurrSize(axis);
    int indexMax = clipBoxMax[axis] * GetCurrSize(axis);

    switch (axis){
    case X: 
	sMin = clipBoxMin[Y]; tMin = clipBoxMin[Z];
	sMax = clipBoxMax[Y]; tMax = clipBoxMax[Z];
	break;
    case Y: 
	sMin = clipBoxMin[X]; tMin = clipBoxMin[Z];
	sMax = clipBoxMax[X]; tMax = clipBoxMax[Z];
	break;
    case Z:
	sMin = clipBoxMin[X]; tMin = clipBoxMin[Y];
	sMax = clipBoxMax[X]; tMax = clipBoxMax[Y];
	break;
    }
    
    glEnable(GL_TEXTURE_2D);

    if (order == TRUE)
	for (int index = indexMin; index < indexMax;
	     index+= DF3Sequence_cls::msCDownSizeLayer.mpBody[axis])
	    DrawSlice2D(axis, index, sMin, sMax, tMin, tMax,
			clr, clrinv, alpha);
    else
	for (int index = indexMax - DF3Sequence_cls::msCDownSizeLayer.mpBody[axis];
	     index >= indexMin;
	     index-= DF3Sequence_cls::msCDownSizeLayer.mpBody[axis])
	    DrawSlice2D(axis, index, sMin, sMax, tMin, tMax,
			clr, clrinv, alpha);
}

void DF3GL_cls::
DrawData(PositionAndDirection_cls CCam, double *Aalpha, 
	 unsigned int axis, bool bMirror[3],
	 double clipBoxMin[3], double clipBoxMax[3],
	 double *pShadeOffset){

    bool order = TRUE ^ bMirror[axis];

    mOldPointTex = 0;
    mOldPointXYZ = 0;

    // $B%U%)%0$@$1$N>l9g$b!*(B
    //if (mbValid == FALSE) return;

    /*
    // $B%U%!%$%k$rJD$8$?$H$-$K2a5n$N%F%/%9%A%c$,$"$k$N$GBP=h(B
    if (GetFilePathWXStr() == wxT(""))
	Aalpha[mpCParent->mSelfIndex] = 0; 
    */

    double scalarProduct;
    double axisDirection[3][3] = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}};
    
    // $B$3$N%+%a%i$O:BI8JQ49:Q$_$N%+%a%i$N0LCV!*(B 
    double CamTgt[3];
    CamTgt[X] = CCam.mTgt[X] / msBoxScaleX;
    CamTgt[Y] = CCam.mTgt[Y] / msBoxScaleY;
    CamTgt[Z] = CCam.mTgt[Z] / msBoxScaleZ;
    MyVector::Normalize(CamTgt, 1.0);

    scalarProduct = MyVector::GetScalarProduct(CamTgt, axisDirection[axis]);
    if (scalarProduct > 0) order = !order;
    if (scalarProduct < 0) scalarProduct *= -1;
    
    scalarProduct = scalarProduct * 1.2 - 0.2;
    if (scalarProduct < 0) scalarProduct = 0;
    
    if (mspCProperty->mbUse3DTexture == TRUE &&
	mspCProperty->mCPreviewType3D.mBody == 1){
	if (axis == X){
	    glPushAttrib(GL_ALL_ATTRIB_BITS);
	    glPushMatrix();
	    glEnable(GL_TEXTURE_3D);
	    for (unsigned int i = 0; i < msVpDF3GL.size(); i++)
		Aalpha[i] *= 1.7 / mspCProperty->GetNQuality();
	    /*
	    Draw3DNormal(CCam, alpha * 1.7, bMirror, clipBoxMin, clipBoxMax,
			 pShadeOffset, bDrawVolume);
	    */
	    Draw3DNormalWithOtherGL(CCam, Aalpha, bMirror, clipBoxMin, clipBoxMax,
				    pShadeOffset);
	    glDisable(GL_TEXTURE_3D);
	    glPopMatrix();
	    glPopAttrib();
	}
	return;
    }

    double boxSize;
    switch(axis){
    case X: boxSize = msBoxScaleX; break;
    case Y: boxSize = msBoxScaleY; break;
    case Z: boxSize = msBoxScaleZ; break; }
    if (boxSize < 0) boxSize *= -1;

    double alphaSum = 0;
    for (unsigned int i = 0; i < msVpDF3GL.size(); i++){
	Aalpha[i] *= scalarProduct * boxSize / GetCurrSize(axis);
	if (Aalpha[i] > 1) Aalpha[i] = 1;
	alphaSum+=Aalpha[i];
    }
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glPushMatrix(); 
    if (alphaSum > 0 && mspCDF3SeqCtrlLink->mAbShow[axis] == TRUE){
	if (mspCProperty->mbUse3DTexture == TRUE){
	    if (mspCProperty->mCPreviewType3D.mBody == 0){
		glEnable(GL_TEXTURE_3D);
		Bind3DTexture();
		DrawSlice3DWithOtherGL(axis, order, clipBoxMin, clipBoxMax, Aalpha, pShadeOffset);
		glDisable(GL_TEXTURE_3D);
	    }
	}
	else{
	    switch(axis){
	    case Z:
		break;
	    case Y:
		glRotatef(90, 1, 0, 0);
		glScalef(1, 1, -1);
		break;
	    case X:
		glRotatef(90, 1, 0, 0);
		glRotatef(90, 0, 1, 0);
		break;
	    } 
	    // FIX ME !!
	    DrawSlice(axis, order, clipBoxMin, clipBoxMax, Aalpha[0]);
	}
    }
    glPopMatrix();
    glPopAttrib();
}

unsigned int DF3GL_cls::
GetAllocatedTextureKB(){
    if (IsTexObjectGenFinished() == FALSE &&
	IsTexObjectGenInProcess() == FALSE) return 0;

    int memory1ch;
   
    if (mspCProperty->mbUse3DTexture == FALSE) {
	memory1ch = GetCurrentTotalSize() * 3 / 1024 ;
	return memory1ch; }
    
    return m3DTexUsedByte / 1024;
}

double DF3GL_cls::msBoxScaleX = 0;
double DF3GL_cls::msBoxScaleY = 0;
double DF3GL_cls::msBoxScaleZ = 0;

DF3SeqCtrl_cls      *DF3GL_cls::mspCDF3SeqCtrlLink = NULL;
Draw3DTexBoxA_cls   *DF3GL_cls::mspCDraw3DTexBoxA = NULL;
Draw3DTexBoxB_cls   *DF3GL_cls::mspCDraw3DTexBoxB = NULL;
Preference_cls      *DF3GL_cls::mspCPreferenceLink = NULL;

std::vector < std::vector < Position_cls > > DF3GL_cls::mVVCIntersect;

// --- $BAj8_1#JCMQ$N(B Adhoc $B$J5!G=72(B
std::vector < DF3GL_cls* > DF3GL_cls::msVpDF3GL;
std::vector < double > DF3GL_cls::msVAlpha;
// --- $BAj8_1#JCMQ$N(B Adhoc $B$J5!G=72(B $B%3%3$^$G(B

// ------

#include "../../MYAPI/MsgBox.h"

ThreadMappingByCPU::
ThreadMappingByCPU() : wxThread(wxTHREAD_DETACHED) {}

void ThreadMappingByCPU::
SetParameters(DF3GL_cls *pCParent,
	      unsigned char *pTo, unsigned char *pFrom, unsigned char *pFrom2,
	      int ini, int fin,
	      unsigned char (*pColorTable)[4], bool *pbFinished){
    mpCParent = pCParent;
    mpTo = pTo; mpFrom = pFrom; mpFrom2 = pFrom2;
    mIni = ini; mFin = fin;
    mpColorTable = pColorTable;
    mpbFinished = pbFinished;

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadMappingByCPU::
Entry(){
    unsigned char *pToIndex = mpTo + mIni * 4;
    unsigned char *pFromIndex = mpFrom + mIni;
    unsigned char *pFromIndex2 = mpFrom2 + mIni;
    unsigned char *pToIndexEnd = mpTo + mFin * 4;
    unsigned char (*pColorTable)[4] = mpColorTable;

    // $B%k!<%WFb$,$3$NDxEY$G$b!"$b$&%k!<%WH=Dj$N:o8:$O$[$H$s$I8z$+$J$$!D(B
    if (mpFrom == mpFrom2){
	// $BL)EY$H?'$rF1$8%G!<%?$+$i(B
	if ((int)(pToIndexEnd - pToIndex) % 16 == TRUE)
	    for (; pToIndex < pToIndexEnd; pToIndex+=16, pFromIndex+=4){
		memcpy(pToIndex   , pColorTable[*pFromIndex   ], 4);
		memcpy(pToIndex+4 , pColorTable[*(pFromIndex+1)], 4);
		memcpy(pToIndex+8 , pColorTable[*(pFromIndex+2)], 4);
		memcpy(pToIndex+12, pColorTable[*(pFromIndex+3)], 4);
	    }
	else
	    for (; pToIndex < pToIndexEnd; pToIndex+=4, pFromIndex++)
		memcpy(pToIndex, pColorTable[*pFromIndex], 4);
    }
    else{
	if ((int)(pToIndexEnd - pToIndex) % 16 == TRUE)
	    for (; pToIndex < pToIndexEnd; pToIndex+=16, pFromIndex+=4, pFromIndex2+=4){
		memcpy(pToIndex   , pColorTable[*pFromIndex2   ], 3);
		*(pToIndex+3) = *(pFromIndex);
		memcpy(pToIndex+4 , pColorTable[*(pFromIndex2+1)], 3);
		*(pToIndex+7) = *(pFromIndex+1);
		memcpy(pToIndex+8 , pColorTable[*(pFromIndex2+2)], 3);
		*(pToIndex+11) = *(pFromIndex+2);
		memcpy(pToIndex+12, pColorTable[*(pFromIndex2+3)], 3);
		*(pToIndex+15) = *(pFromIndex+3);
	    }
	else
	    for (; pToIndex < pToIndexEnd; pToIndex+=4, pFromIndex++, pFromIndex2++){
		memcpy(pToIndex, pColorTable[*pFromIndex2], 3);
		*(pToIndex + 3) = *(pFromIndex);
	    }
    }
    
    *mpbFinished = TRUE;
	
    return 0;
}

// ------

PositionEX_cls::
PositionEX_cls(double posX, double posY, double posZ, unsigned int index) :
    Position_cls(posX, posY, posZ)
{
    mIndex = index;
    mAngleTan = 0;
}

PositionEX_cls::
PositionEX_cls(const PositionEX_cls &CPos){
    LOOP_XYZ mPos[xyz] = CPos.mPos[xyz];
    LOOP_XYZ mVel[xyz] = CPos.mVel[xyz];
    mIndex = CPos.mIndex;
    mAngleTan = CPos.mAngleTan;
}

bool operator<(const PositionEX_cls& CLeft, 
	       const PositionEX_cls& CRight)
{
    return CLeft.mAngleTan < CRight.mAngleTan;
}
