// "clsDF3GL.cpp"

#ifndef _DF3GL_H_
#define _DF3GL_H_

#include "../clsDF3.h"
#include "../../MYGL/clsPositionAndDirection.h"
#include "../../CONFITEM/clsConfItem.h"

#if WIN32
#include <sys/types.h>
#include <process.h>
#include <windows.h>
#endif

#include <GL/glu.h>

#include <vector>
#include <string>
#include <sstream>

class DF3GL_cls; 
class DF3SeqCtrl_cls;
class FilterCtrl_cls;
class Direction_cls;
class Draw3DTexBoxA_cls;
class Draw3DTexBoxB_cls;
class PreviewProperty_cls;
class wxGLContext;
class wxGLCanvas;

class DF3GLInfo_cls{
 public:
    DF3GLInfo_cls();
    DF3GLInfo_cls(PositionAndDirection_cls CCam,
		  double clipBoxMin[3], double clipBoxMax[3]);

    PositionAndDirection_cls mCCam;
    double mClipBoxMin[3], mClipBoxMax[3];
};

bool operator==(const DF3GLInfo_cls& CL, const DF3GLInfo_cls& CR);

// $B=i4|CM$H$7$F(BDF3_cls$B%3%s%9%H%i%/%?Fb$N(Bthis$B$O4m81$J$N$G!"(B
// DF3Sequence_cls$B$N%3%s%9%H%i%/%?$GBeF~$7$F$$$k(B
class TexObjectGen_cls : public ResumableProgress_cls{
  public:
    DF3GL_cls *mpC;
};

class TexObjectBind_cls : public ResumableProgress_cls{
 public:
    DF3GL_cls *mpC;
};

class TexObjectDelete_cls : public ResumableProgress_cls{
 public:
    DF3GL_cls *mpC;
};

class DF3GL_cls : public DF3_cls {
 public:
    DF3GL_cls(DF3Sequence_cls* pCParent);
    ~DF3GL_cls();

    static DF3SeqCtrl_cls      *mspCDF3SeqCtrlLink;
    static Draw3DTexBoxA_cls   *mspCDraw3DTexBoxA;
    static Draw3DTexBoxB_cls   *mspCDraw3DTexBoxB;
    static Preference_cls      *mspCPreferenceLink;
    static void InitializeStaticMembers
	(wxWindow *pCWindow, DF3SeqCtrl_cls *pCDF3SeqCtrl,
	 wxBitmap *pBitmapProp, Preference_cls *pCPreference);

    // --- $BAj8_1#JCMQ$N(B Adhoc $B$J5!G=72(B
    static std::vector < DF3GL_cls* > msVpDF3GL;
    static std::vector < double > msVAlpha;
    void   Draw3DNormalWithOtherGL
	(PositionAndDirection_cls CCam, double *Aalpha, bool bMirror[3],
	 double clipBoxMin[3], double clipBoxMax[3],
	 double *pShadeOffset);
    // --- $BAj8_1#JCMQ$N(B Adhoc $B$J5!G=72(B $B%3%3$^$G(B
    
    // delete$B$7$F(Bgenerate$B$9$k$H%Q%U%)!<%^%s%9Nt2=$,Aa$$(B
    // $B$9$0I=<($9$k>l9g$O$G$-$k$@$1M^@)$9$k(B
    bool   mbFakeDeleted;
    bool   GenerateTexObjects(bool type, bool bFakable = FALSE);
    bool   BindTexObjects(bool type);
    bool   DeleteTexObjects(bool type, bool bFakeable = FALSE);

    int    GetTextureBindPercent();
    void   GetCurrentColour(float clr[3], float clrinv[3]);
    void   Bind3DTexture();

    static void   DrawDataInitSetting();
    static void   DrawDataEndSetting();
    static double msBoxScaleX, msBoxScaleY, msBoxScaleZ;
    static std::vector < std::vector < Position_cls > > mVVCIntersect; //$B6&DL(B
    
    void   SetBoxScale(double boxScaleX, double boxScaleY, double boxScaleZ);
    void   SetBlendFunction();

    void   Draw3DNormal
	(PositionAndDirection_cls CCam, double alpha, bool bMirror[3],
	 double clipBoxMin[3], double clipBoxMax[3],
	 double *pShadeOffset, bool bDrawVolume);

    
    void   DrawSlice3D(unsigned int axis, bool order,
		       double clipBoxMin[3], double clipBoxMax[3],
		       float alpha, double *pShadeOffset);
    void   DrawSlice3DWithOtherGL
	(unsigned int axis, bool order,
	 double clipBoxMin[3], double clipBoxMax[3],
	 double *Aalpha, double *pShadeOffset);
    void   DrawSliceST(double coord, 
		       double sMin, double sMax, double tMin, double tMax);
    void   DrawSlice2D(unsigned int axis, int index,
		       double sMin, double sMax, double tMin, double tMax,
		       float clr[3], float clrinv[3], float alpha);
    
    void   DrawSlice(unsigned int axis, bool order,
		     double clipBoxMin[3], double clipBoxMax[3], float alpha);
    void   DrawData(PositionAndDirection_cls CCam, double *Alpha,
		    unsigned int axis, bool bMirror[3],
		    double clipBoxMin[3], double clipBoxMax[3],
		    double *pShadeOffset);

    unsigned int GetAllocatedTextureKB();

    TexObjectGen_cls mCTexObjectGen[3];
    bool   IsTexObjectGenFinished();
    bool   IsTexObjectGenInProcess();
    bool   IsTexObjectGenFinishConditionSatisfied();
    
    TexObjectBind_cls mCTexObjectBind[3];
    bool   IsTexObjectBindFinished();
    bool   IsTexObjectBindInProgress();
    bool   IsTexObjectBindFinishConditionSatisfied();

    TexObjectDelete_cls mCTexObjectDelete[3];
    bool   IsTexObjectDeleteFinished();
    bool   IsTexObjectDeleteInProgress();
    bool   IsTexObjectDeleteFinishConditionSatisfied();

 public:
    std::vector < GLuint > mV2Texture[3];
    GLuint mTexture3D;
    int m3DTexCompressed;
    int m3DTexUsedByte;

    unsigned int   mOldPointTex;
    unsigned char *mpRGBA; //$B%^%k%A%9%l%C%IBP:v(B

};

// ------

class ThreadMappingByCPU : public wxThread {
 public:
    ThreadMappingByCPU();
    void SetParameters(DF3GL_cls *pCParent,
		       unsigned char *pTo, unsigned char *pFrom, 
		       unsigned char *pFrom2,
		       int ini, int fin,
		       unsigned char (*pColorTable)[4],
		       bool *pBFinished);
    
    virtual ExitCode Entry();
    
    DF3GL_cls *mpCParent;
    unsigned char *mpTo, *mpFrom, *mpFrom2;
    int mIni, mFin;
    unsigned char (*mpColorTable)[4];
    bool *mpbFinished;
};

// ------

class PositionEX_cls : public Position_cls{
 public:
    PositionEX_cls(double posX, double posY, double posZ, unsigned int index);
    PositionEX_cls(const PositionEX_cls &CPos);
    
    unsigned int mIndex;
    double mAngleTan;
};

bool operator<(const PositionEX_cls& CLeft, 
	       const PositionEX_cls& CRight);

#endif
