// "clsDraw3DTexBoxB.h"

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
extern PFNGLTEXIMAGE3DPROC glTexImage3D;
extern PFNGLBLENDEQUATIONEXTPROC glBlendEquationEXT;
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "clsDraw3DTexBoxB.h"
#include "clsPreviewProperty.h"
#include "../clsDF3Sequence.h"
#include "../clsDF3SeqCtrl.h"

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)

#define S 0
#define T 1

Draw3DTexBoxB_cls::
Draw3DTexBoxB_cls(DF3SeqCtrl_cls *pCDF3SeqCtrl, 
		  PreviewProperty_cls *pCProperty){
    mpCDF3SeqCtrlLink = pCDF3SeqCtrl;
    mpCPropertyLink = pCProperty;

    mspFuncDrawSlice3D[0] = &DrawSlice3DX;
    mspFuncDrawSlice3D[1] = &DrawSlice3DY;
    mspFuncDrawSlice3D[2] = &DrawSlice3DZ;
}

void Draw3DTexBoxB_cls::
DrawSlice3DX(double coord, double sMin, double sMax, double tMin, double tMax,
	     double *pShadeOffset){
    double coordTex = coord - pShadeOffset[X];
    double sMinTex = sMin - pShadeOffset[Y], sMaxTex = sMax - pShadeOffset[Y];
    double tMinTex = tMin - pShadeOffset[Z], tMaxTex = tMax - pShadeOffset[Z];

    glBegin(GL_QUADS);
    glTexCoord3f(coordTex, sMinTex, tMinTex); glVertex3d(coord, sMin, tMin);
    glTexCoord3f(coordTex, sMaxTex, tMinTex); glVertex3d(coord, sMax, tMin);
    glTexCoord3f(coordTex, sMaxTex, tMaxTex); glVertex3d(coord, sMax, tMax);
    glTexCoord3f(coordTex, sMinTex, tMaxTex); glVertex3d(coord, sMin, tMax);
    glEnd();
}

void Draw3DTexBoxB_cls::
DrawSlice3DY(double coord, double sMin, double sMax, double tMin, double tMax,
	     double *pShadeOffset){
    double coordTex = coord - pShadeOffset[Y];
    double sMinTex = sMin - pShadeOffset[Z], sMaxTex = sMax - pShadeOffset[Z];
    double tMinTex = tMin - pShadeOffset[X], tMaxTex = tMax - pShadeOffset[X];

    glBegin(GL_QUADS);
    glTexCoord3f(tMinTex, coordTex, sMin); glVertex3d(tMin, coord, sMin);
    glTexCoord3f(tMinTex, coordTex, sMax); glVertex3d(tMin, coord, sMax);
    glTexCoord3f(tMaxTex, coordTex, sMax); glVertex3d(tMax, coord, sMax);
    glTexCoord3f(tMaxTex, coordTex, sMin); glVertex3d(tMax, coord, sMin);
    glEnd();
}

void Draw3DTexBoxB_cls::
DrawSlice3DZ(double coord, double sMin, double sMax, double tMin, double tMax,
	     double *pShadeOffset){
    double coordTex = coord - pShadeOffset[Z];
    double sMinTex = sMin - pShadeOffset[X], sMaxTex = sMax - pShadeOffset[X];
    double tMinTex = tMin - pShadeOffset[Y], tMaxTex = tMax - pShadeOffset[Y];

    glBegin(GL_QUADS);
    glTexCoord3f(sMinTex, tMinTex, coordTex); glVertex3d(sMin, tMin, coord);
    glTexCoord3f(sMaxTex, tMinTex, coordTex); glVertex3d(sMax, tMin, coord);
    glTexCoord3f(sMaxTex, tMaxTex, coordTex); glVertex3d(sMax, tMax, coord);
    glTexCoord3f(sMinTex, tMaxTex, coordTex); glVertex3d(sMin, tMax, coord);
    glEnd();
}

void Draw3DTexBoxB_cls::
DrawSlice3DMain(unsigned int axis, double sizeAxis, int index,
		double sMin, double sMax, double tMin, double tMax,
		float clr[3], float clrinv[3],
		float alpha, double *pShadeOffset){
    
    double offset[3] = {0, 0 ,0}, offset0[3] = {0, 0, 0};
    if (pShadeOffset != NULL)
	LOOP_XYZ offset[xyz] = pShadeOffset[xyz];

    double coord = (double)(index + 0.5) / sizeAxis;
    if (DF3GL_cls::mspCProperty->mCShadingType.mBody == 1){
	glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	float clrshade[3];
	LOOP_RGB clrshade[rgb] = clr[rgb] + mpCPropertyLink->mCAbsorptionComplement.mBody * (clrinv[rgb] - clr[rgb]);
	if (mpCPropertyLink->mCAbsorption.mBody != 0){
	    glColor4f(clr[0], clr[1], clr[2], 
		      alpha * mpCPropertyLink->mCAbsorption.mBody);
	    (*mspFuncDrawSlice3D[axis])(coord, sMin, sMax, tMin, tMax, offset);
	}
    }
    SetBlendFunction();
    glColor4f(clr[0], clr[1], clr[2], alpha);
    glBlendEquationEXT(GL_FUNC_ADD_EXT);
    (*mspFuncDrawSlice3D[axis])(coord, sMin, sMax,  tMin, tMax, offset0);
}

void Draw3DTexBoxB_cls::
DrawSlice3D(unsigned int axis, double sizeAxis, bool order,
	    float clr[3], float clrinv[3],
	    double clipBoxMin[3], double clipBoxMax[3], float alpha,
	    double *pShadeOffset){
    
    glTranslated(-0.5, -0.5, -0.5);
    double sMin = 0;//1. / (GetCurrSizeS(axis) + 1);
    double tMin = 0;//1. / (GetCurrSizeS(axis) + 1);
    double sMax = 1;//(double)GetCurrSizeS(axis) / (GetCurrSizeS(axis) + 1);
    double tMax = 1;//(double)GetCurrSizeT(axis) / (GetCurrSizeT(axis) + 1);

    int indexMin = clipBoxMin[axis] * sizeAxis;
    int indexMax = clipBoxMax[axis] * sizeAxis;

    //if (clipBoxMin[axis] > clipBoxMax[axis]) order = !order;
    
    int downSizeLayer = 
	DF3Sequence_cls::msCDownSizeLayer.mpBody[axis];
	//mpCDF3SeqCtrlLink->GetPDF3Seq(_DF3_SLOT_0_FIX_ME)->mDownSizeLayer[axis];
    alpha *= downSizeLayer;

    if (order == TRUE){
	switch (axis){
	case X: {
	    sMin = clipBoxMin[Y]; tMin = clipBoxMin[Z];
	    sMax = clipBoxMax[Y]; tMax = clipBoxMax[Z];
	    for (int index = indexMin; index < indexMax; index+= downSizeLayer)
		DrawSlice3DMain(axis, sizeAxis, index, 
				sMin, sMax, tMin, tMax,	clr, clrinv, alpha, pShadeOffset);
	    break; }
	case Y: {
	    sMin = clipBoxMin[Z]; tMin = clipBoxMin[X];
	    sMax = clipBoxMax[Z]; tMax = clipBoxMax[X];
	    for (int index = indexMin; index < indexMax; index+= downSizeLayer)
		DrawSlice3DMain(axis, sizeAxis, index, 
				sMin, sMax, tMin, tMax, clr, clrinv, alpha, pShadeOffset);
	    break; }
	case Z: {
	    sMin = clipBoxMin[X]; tMin = clipBoxMin[Y];
	    sMax = clipBoxMax[X]; tMax = clipBoxMax[Y];
	    for (int index = indexMin; index < indexMax; index+= downSizeLayer)
		DrawSlice3DMain(axis, sizeAxis, index, 
				sMin, sMax, tMin, tMax, clr, clrinv, alpha, pShadeOffset);
	    break; }
	}
    }
    else{ //΂`
	switch (axis){
	case X: {
	    sMin = clipBoxMin[Y]; tMin = clipBoxMin[Z];
	    sMax = clipBoxMax[Y]; tMax = clipBoxMax[Z];
	    for (int index = indexMax - downSizeLayer; 
		 index >= indexMin; index-= downSizeLayer)
		DrawSlice3DMain(axis, sizeAxis, index, 
				sMin, sMax, tMin, tMax, clr, clrinv, alpha, pShadeOffset);
	    break; }
	case Y: {
	    sMin = clipBoxMin[Z]; tMin = clipBoxMin[X];
	    sMax = clipBoxMax[Z]; tMax = clipBoxMax[X];
	    for (int index = indexMax - downSizeLayer; 
		 index >= indexMin; index-= downSizeLayer)
		DrawSlice3DMain(axis, sizeAxis, index, 
				sMin, sMax, tMin, tMax, clr, clrinv, alpha, pShadeOffset);
	    break; }
	case Z: {
	    sMin = clipBoxMin[X]; tMin = clipBoxMin[Y];
	    sMax = clipBoxMax[X]; tMax = clipBoxMax[Y];
	    for (int index = indexMax - downSizeLayer; 
		 index >= indexMin; index-= downSizeLayer)
		DrawSlice3DMain(axis, sizeAxis, index, 
				sMin, sMax, tMin, tMax, clr, clrinv, alpha, pShadeOffset);
	    break; }
	}
    }
}

void Draw3DTexBoxB_cls::
DrawSlice3DWithOtherGL(unsigned int axis, int indexNum, 
		       double sizeAxis, bool order,
		       float (*clr)[3], float (*clrinv)[3], int *Atexture3D,
		       double clipBoxMin[3], double clipBoxMax[3], 
		       double *aAlpha, double *pShadeOffset){
    
    glTranslated(-0.5, -0.5, -0.5);
    double sMin = 0;//1. / (GetCurrSizeS(axis) + 1);
    double tMin = 0;//1. / (GetCurrSizeS(axis) + 1);
    double sMax = 1;//(double)GetCurrSizeS(axis) / (GetCurrSizeS(axis) + 1);
    double tMax = 1;//(double)GetCurrSizeT(axis) / (GetCurrSizeT(axis) + 1);

    int indexMin = clipBoxMin[axis] * sizeAxis;
    int indexMax = clipBoxMax[axis] * sizeAxis;

    int downSizeLayer = DF3Sequence_cls::msCDownSizeLayer.mpBody[axis];
    //mpCDF3SeqCtrlLink->GetPDF3Seq(_DF3_SLOT_0_FIX_ME)->mDownSizeLayer[axis];

    
    for (unsigned int df3 = 0; df3 < indexNum; df3++)
	aAlpha[df3] *= downSizeLayer;
    
    if (order == TRUE){
	switch (axis){
	case X: {
	    sMin = clipBoxMin[Y]; tMin = clipBoxMin[Z];
	    sMax = clipBoxMax[Y]; tMax = clipBoxMax[Z];
	    for (int index = indexMin; index < indexMax; 
		 index+= downSizeLayer)
		for (unsigned int df3 = 0; df3 < indexNum; df3++){
		    glBindTexture(GL_TEXTURE_3D, Atexture3D[df3]);
		    DrawSlice3DMain(axis, sizeAxis, index, 
				    sMin, sMax, tMin, tMax,	
				    clr[df3], clrinv[df3], aAlpha[df3],
				    pShadeOffset);
		}
	    break; }
	case Y: {
	    sMin = clipBoxMin[Z]; tMin = clipBoxMin[X];
	    sMax = clipBoxMax[Z]; tMax = clipBoxMax[X];
	    for (int index = indexMin; index < indexMax;
		 index+= downSizeLayer)
		for (unsigned int df3 = 0; df3 < indexNum; df3++){
		    glBindTexture(GL_TEXTURE_3D, Atexture3D[df3]);
		    DrawSlice3DMain(axis, sizeAxis, index, 
				    sMin, sMax, tMin, tMax, 
				    clr[df3], clrinv[df3], aAlpha[df3],
				    pShadeOffset);
		}
	    break; }
	case Z: {
	    sMin = clipBoxMin[X]; tMin = clipBoxMin[Y];
	    sMax = clipBoxMax[X]; tMax = clipBoxMax[Y];
	    for (int index = indexMin; index < indexMax;
		 index+= downSizeLayer)
		for (unsigned int df3 = 0; df3 < indexNum; df3++){
		    glBindTexture(GL_TEXTURE_3D, Atexture3D[df3]);
		    DrawSlice3DMain(axis, sizeAxis, index, 
				    sMin, sMax, tMin, tMax, 
				    clr[df3], clrinv[df3], aAlpha[df3],
				    pShadeOffset);
		}
	    break; }
	}
    }
    else{ //΂`
	switch (axis){
	case X: {
	    sMin = clipBoxMin[Y]; tMin = clipBoxMin[Z];
	    sMax = clipBoxMax[Y]; tMax = clipBoxMax[Z];
	    for (int index = indexMax - downSizeLayer; 
		 index >= indexMin; index-= downSizeLayer)
		for (unsigned int df3 = 0; df3 < indexNum; df3++){
		    glBindTexture(GL_TEXTURE_3D, Atexture3D[df3]);
		    DrawSlice3DMain(axis, sizeAxis, index, 
				    sMin, sMax, tMin, tMax, 
				    clr[df3], clrinv[df3], aAlpha[df3],
				    pShadeOffset);
		}
	    break; }
	case Y: {
	    sMin = clipBoxMin[Z]; tMin = clipBoxMin[X];
	    sMax = clipBoxMax[Z]; tMax = clipBoxMax[X];
	    for (int index = indexMax - downSizeLayer; 
		 index >= indexMin; index-= downSizeLayer)
		for (unsigned int df3 = 0; df3 < indexNum; df3++){
		    glBindTexture(GL_TEXTURE_3D, Atexture3D[df3]);
		    DrawSlice3DMain(axis, sizeAxis, index, 
				    sMin, sMax, tMin, tMax, 
				    clr[df3], clrinv[df3], aAlpha[df3],
				    pShadeOffset);
		}
	    break; }
	case Z: {
	    sMin = clipBoxMin[X]; tMin = clipBoxMin[Y];
	    sMax = clipBoxMax[X]; tMax = clipBoxMax[Y];
	    for (int index = indexMax - downSizeLayer; 
		 index >= indexMin; index-= downSizeLayer)
		for (unsigned int df3 = 0; df3 < indexNum; df3++){
		    glBindTexture(GL_TEXTURE_3D, Atexture3D[df3]);
		    DrawSlice3DMain(axis, sizeAxis, index, 
				    sMin, sMax, tMin, tMax, 
				    clr[df3], clrinv[df3], aAlpha[df3],
				    pShadeOffset);
		}
	    break; }
	}
    }
}

void (*Draw3DTexBoxB_cls::mspFuncDrawSlice3D[3])(double coord, double sMin, double sMax, double tMin, double tMax, double *pShadeOffset) = {NULL, NULL, NULL};

