// (setq c-basic-offset 4)
// "clsPreviewProperty."

#include "clsPreviewProperty.h"

#include "../../../wxWidgetEnum.h"
#include "../../../wxWidgetEnumMenu.h"

#include "../../DF3/clsDF3SeqCtrl.h"
#include "../../DF3/clsDF3Sequence.h"
#include "../../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../../FILTER/clsFilterCtrl.h"
#include "../../GLCANVAS/clsMainGLCanvas.h"
#include "../../MYGL/clsFBO.h"
#include "../../MYAPI/MsgBox.h"
#include "../../WIDGETCTRL/clsGraphPanel.h"

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

// ÓIɃCxgт
BEGIN_EVENT_TABLE(PreviewPropertyGUI_cls, wxDialog)
     EVT_SIZE(PreviewPropertyGUI_cls::OnSize)
     EVT_BUTTON(wxID_CANCEL   , PreviewPropertyGUI_cls::OnButtonClose)
     EVT_BUTTON(btnPreviewShade, PreviewPropertyGUI_cls::OnButtonShowItem)
END_EVENT_TABLE()

PreviewPropertyGUI_cls::
PreviewPropertyGUI_cls(PreviewProperty_cls *pCParent, wxWindow *pC,
		       wxPoint pos, wxSize size) :
wxDialogEXKey(pC, wxID_ANY, _("Preview Property"),
	      pos, size, pCParent->mpBitmapLink){
    mpCParent = pCParent;
    SetScreenRectPos(&pCParent->mCScreenRect);

    // TCU[
    wxBoxSizer       *pBoxSizerSrc    = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerParam  = new wxBoxSizer(wxVERTICAL);
    
    SetMNumRenderParam(pBoxSizerParam);
    SetMNumShadeParam (pBoxSizerParam);
    
    pBoxSizerSrc->Add(pBoxSizerParam, 0, wxEXPAND);
    
    mpCGraphModulator = new GraphPanel_cls(this, wxID_ANY, wxDefaultPosition, wxSize(-1, 60));
    mpCGraphModulator->SetFunction(&PreviewProperty_cls::GetLevelMapping, 256);
    mpCGraphModulator->SetFunction(&LevelMap_cls::GetLevelMap, 256);
    pBoxSizerSrc->Add(mpCGraphModulator, 0, wxEXPAND);

    pBoxSizerSrc->Add(pBoxSizerSrc->GetMinSize().GetWidth(), -1, 0, 0);
    SetSizerAndFit(pBoxSizerSrc);

    int minimumWidth = GetClientSize().GetWidth() - 8;
    minimumWidth = mpCMNumRenderParam->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumShadeParam->GetPaddingWidth(minimumWidth);
    mpCMNumRenderParam->SetPaddingWidth(minimumWidth);
    mpCMNumShadeParam->SetPaddingWidth(minimumWidth);

    mspCThis = this;
}

PreviewPropertyGUI_cls::
~PreviewPropertyGUI_cls(){
    MakeModal(FALSE);
}

void PreviewPropertyGUI_cls::
SetMNumRenderParam(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumRenderParam, this);
    int boxWidth = 2;
        
    pCurr->mTextWidth = -1;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Draw Texture Method"));
    pCurr->CombinePointer(_GP_BOOL, &mpCParent->mbUse3DTexture, _("Use 3D Texture"), FALSE);
    wxString listHiding[3] = 
	{_("Mutual Hide"), _("No Hide"), _("Inverse Hide") };
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mpCParent->mCHiding.mBody,
			  wxT(""), TRUE, 3, listHiding, 0);

#ifdef WIN32
    pCurr->mComboWidth = 108;
#else
    pCurr->mComboWidth = 130;
#endif

    wxString listTexture[3] = 
	{_("ALPHA (Less Memory)"), _("LUMINANCE_ALPHA"),
	 _("RGBA (High Quality)") };
    mIndexComboFormat = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCTextureType.mBody,
	 _("Texture Type"), TRUE, 3, listTexture, 1);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Draw Texture Method"));
    wxString strListMode[2] = {_("0: Alpha Blend"), _("1: Add") };
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, 
			  &mpCParent->mCRenderType.mBody,
			  wxT(""), TRUE, 2, strListMode, 1);
    wxString listFilter[2] = {_("NEAREST"), _("LINEAR")};
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mpCParent->mCTexFilter.mBody,
			  wxT(""), TRUE, 2, listFilter, 1);
    
    wxString listPreviewType[2] = {_("Lattice like"), _("Normal to Camera")};
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Preview Type"));
    mIndexComboSliceType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCPreviewType3D.mBody,
	 _("Preview Type"), TRUE, 2, listPreviewType);

    wxString listQuality[13] = {_("32 normal layers"), _("48 noamal layers"), _("64 normal layers"), _("96 normal layers"), _("128 normal layers"), _("192 normal layers"), _("256 normal layers"), _("384 normal layers"), _("512 normal layers"), _("768 normal layers"), _("1024 normal layers"), _("1536 normal layers"), _("2048 normal layers")};
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mpCParent->mCPreviewNormalQuality.mBody,
			  _("Quality"), TRUE, 13, listQuality);
    
#ifdef WIN32
    pCurr->mMinimumWidth = 340;
#else
    pCurr->mMinimumWidth = 400;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_COMBOBOX);
    pCurr->SetFuncHook(&FuncHookFormat, _MN_HOOK_COMBOBOX_CHANGE);
}

void PreviewPropertyGUI_cls::
SetMNumShadeParam(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumShadeParam, this);
    int boxWidth = 2;
    /*
    pCurr->AssignMinimizeButtonBoth(mpCBTNShade, btnPreviewShade,
				    _("[+]Shade Settings"));
    */
    pCurr->mTextWidth = -1;
    pCurr->CombinePointerButton(btnPanelHide, _("[+]Shade Settings"));

#ifdef WIN32
    pCurr->mComboWidth = 108;
#else
    pCurr->mComboWidth = 130;
#endif

    pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Shading"));

    wxString pListShade[2] = 
	{_("no Emboss"), _("Emboss Shading") };
    
    mIndexShading = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCShadingType.mBody,
	 wxT(""), TRUE, 2, pListShade);

    wxString pListPattern[2] = 
	{_("constant absorption"), _("weak at dense") };
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCShadingPattern.mBody,
	 wxT(""), TRUE, 2, pListPattern);
    
    mIndexLuminance = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Luminance Factor"));

    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mpCParent->mCLuminanceFactor.mBody);
    pCurr->SetRange(1, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCLuminanceFactor.mBody);

    mIndexAbsorption = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Absorption Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mpCParent->mCAbsorption.mBody);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCAbsorption.mBody);

    mIndexComplement = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Complement"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mpCParent->mCAbsorptionComplement.mBody);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCAbsorptionComplement.mBody);

    mIndexOffset = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Shade Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mpCParent->mCShadeOffset.mBody);
    pCurr->SetRange(1, 300, _MN_LASTITEM, 10000);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCShadeOffset.mBody);

    /*
    mIndexShadePattern = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Shading Pattern"));
    */
    wxString pListCoord[2] = 
	{_("local offset"), _("global offset") };
    mIndexShadePattern = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCShadingCoord.mBody,
	 wxT("Shading Pattern"), TRUE, 2, pListCoord);

        
#ifdef WIN32
    pCurr->mMinimumWidth = 340;
#else
    pCurr->mMinimumWidth = 400;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);

    //pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_COMBOBOX);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_ASSIGN);

    pCurr->SetFuncHook
	(FuncHookHideButton, _MN_HOOK_BUTTON, pCurr->mVCPointers[0].mID);
}

void PreviewPropertyGUI_cls::
FuncHookFormat(void *pC, void *pEvent){
    MultiNumericalInputsPanel_cls* pCurr = (MultiNumericalInputsPanel_cls*)pC;
    if (((wxCommandEvent*)pEvent)->GetId() == 
	pCurr->mVCPointers[mspCThis->mIndexComboFormat].GetPCtrlWindow()->GetId()){
	mspCThis->mpCParent->mpCDF3SeqCtrlLink->ReleaseAllDF3Seq(_DF3SEQ_TEXTURE);
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	mspCThis->mpCParent->mpCParentWindow->AddPendingEvent(com);
    }
}

void PreviewPropertyGUI_cls::
FuncHookRefresh(void *pC, void *pEvent){
    mspCThis->HideAndShowAll();
    wxCommandEvent comRefresh(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    mspCThis->mpCParent->mpCParentWindow->AddPendingEvent(comRefresh);
}

void PreviewPropertyGUI_cls::
FuncHookHideButton(void *pPanel, void *pArg){
    MultiNumericalInputsPanel_cls* pCurr = ((MultiNumericalInputsPanel_cls*)pPanel);
    pCurr->TogglePanelOpenStatus(1);
    mspCThis->HideAndShowAll();
}

void PreviewPropertyGUI_cls::
HideAndShowAll(){
    MultiNumericalInputsPanel_cls* pCurr = (MultiNumericalInputsPanel_cls*)mpCMNumRenderParam;

    if (mpCParent->mbUse3DTexture == FALSE){
	mspCThis->mpCParent->mCPreviewType3D.mBody = 0;
	pCurr->mVCPointers[mspCThis->mIndexComboSliceType].Disable();
    }

    pCurr->mVCPointers[mspCThis->mIndexComboSliceType+1].Enable
	(mspCThis->mpCParent->mCPreviewType3D.mBody != 0);

    pCurr = (MultiNumericalInputsPanel_cls*)mpCMNumShadeParam;

    if (pCurr->mbPanelOpen == FALSE){
	ResizeDialog();
	mpCGraphModulator->Refresh(TRUE);
	return;
    }

    pCurr->mVCPointers[mIndexShading].Enable
	(mpCParent->mbBlendEquationEX == TRUE);
    
    bool bShading = (mspCThis->mpCParent->mCShadingType.mBody == 1);

    pCurr->ShowItems(mIndexLuminance, 3, 
		     bShading && (mpCParent->mCTextureType.mBody == _PREVIEW_LUMINANCE_ALPHA));
    pCurr->ShowItems(mIndexAbsorption, 3, bShading);
    pCurr->ShowItems(mIndexComplement, 3, bShading);
    pCurr->ShowItems(mIndexOffset, 3, bShading);
    pCurr->ShowItems(mIndexShadePattern, 1, bShading);
    
    /*
    pCurr->mVCPointers[mIndexLuminance+1].Enable
	(mspCThis->mpCParent->mCTextureType.mBody == _PREVIEW_LUMINANCE_ALPHA);
    pCurr->mVCPointers[mIndexLuminance+2].Enable
	(mspCThis->mpCParent->mCTextureType.mBody == _PREVIEW_LUMINANCE_ALPHA);
    */

    ResizeDialog();
    mpCGraphModulator->Refresh(TRUE);
}

void PreviewPropertyGUI_cls::
OnSize(wxSizeEvent &event){
    event.Skip();
    SetSizerAndFit(GetSizer());
}

void PreviewPropertyGUI_cls::
OnButtonShowItem(wxCommandEvent& event){
    mpCMNumShadeParam->ShowPanel(TRUE);
    GetSizer()->Layout();
    Fit();
}

void PreviewPropertyGUI_cls::
UpdateGraph(){
    mpCGraphModulator->Refresh();
}

void PreviewPropertyGUI_cls::
OnButtonClose(wxCommandEvent &event){
    mpCParent->UpdateItems();
    CloseEXKey();
}

PreviewPropertyGUI_cls* PreviewPropertyGUI_cls::mspCThis;

// --------------------------------

PreviewProperty_cls::
PreviewProperty_cls(wxWindow* pCParentWindew, DF3SeqCtrl_cls* pCDF3SeqCtrl,
		    wxBitmap* pBitmap)
{
    mspCThis = this;
    //mspCFilterCtrlLink = pCFilterCtrl;
    mpCParentWindow = pCParentWindew;
    mpCDF3SeqCtrlLink = pCDF3SeqCtrl;
    mpCGUI = NULL;
    mpBitmapLink = pBitmap;

    SetToDefault();
    InitGUI(pCParentWindew);
    
    mbUse3DTexture = FALSE;
}

PreviewProperty_cls::
~PreviewProperty_cls(){
    if (mpCGUI != NULL) delete mpCGUI;
}

void PreviewProperty_cls::
SetToDefault(){
    mCConfListPreview.Clear();
    mCConfListPreview.Add(&mCPreviewType3D, L"PreviewType3D=", L"1");
    mCConfListPreview.Add(&mCPreviewNormalQuality, L"PreviewNormalQuality=", L"4");
    mCConfListPreview.Add(&mCTextureType,   wxT("PreviewTextureType="), L"0");
    mCConfListPreview.Add(&mCRenderType,    wxT("PreviewType="), L"0");
    mCConfListPreview.Add(&mCHiding,        wxT("PreviewHiding="), L"0");
    mCConfListPreview.Add(&mCAbsorption,    wxT("PreviewAbsorption="), L"0");
    mCConfListPreview.Add(&mCAbsorptionComplement, wxT("PreviewAbsorptionComplement="), L"0");
    mCConfListPreview.Add(&mCLuminanceFactor, wxT("PreviewLuminanceFactor="), L"1");
    mCConfListPreview.Add(&mCShadingType,    L"PreviewShading=", L"0");
    mCConfListPreview.Add(&mCShadingCoord,   L"PreviewShadingCoord=", L"0");
    mCConfListPreview.Add(&mCShadingPattern, L"PreviewShadingPattern=", L"0");
    mCConfListPreview.Add(&mCShadeOffset, wxT("PreviewShadeOffset="), L"0.001");
    mCConfListPreview.Add(&mCTexFilter  , wxT("PreviewTexFilter="), L"1");
}

void PreviewProperty_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new PreviewPropertyGUI_cls(this, pC,
					mCScreenRect.GetPosition(),
					mCScreenRect.GetSize());
    SetPDialog(mpCGUI);
}

void PreviewProperty_cls::
UpdateItems(bool bForce){
    if (mpCGUI != NULL){
	mpCGUI->mpCMNumRenderParam->UpdateValuesAll(bForce);
	mpCGUI->mpCMNumShadeParam->UpdateValuesAll(bForce);
    }
}

void PreviewProperty_cls::
Set3DTextureCondition(int majorVersionGL, int minorVersionGL){
    if (majorVersionGL >= 2) mbUse3DTexture = TRUE;
    if (majorVersionGL == 1 && minorVersionGL >= 2) mbUse3DTexture = TRUE;
}



bool PreviewProperty_cls::
IsFilterProcessRequired(unsigned int df3index){
    if (/*DF3_cls::msbUseProcessedXYZ == TRUE*/ TRUE && 
	mpCDF3SeqCtrlLink->GetPFilterCtrl(df3index)->CountDataFilterNumber() != 0) return TRUE;
    return FALSE;
}

double PreviewProperty_cls::
GetLevelMapping(double value, void *pClue){
    return mspCThis->mpCDF3SeqCtrlLink->GetPFilterCtrlActive()->GetTotalLevelMapping(value); //!!
}

double PreviewProperty_cls::
GetLevelMapping(double value, int df3index){
    return mspCThis->mpCDF3SeqCtrlLink->GetPFilterCtrl(df3index)->GetTotalLevelMapping(value); //!!
}

int PreviewProperty_cls::
GetNQuality(){
    int qindex = mCPreviewNormalQuality.mBody;
    static int quality[13] = {32, 48, 64, 96, 128, 192, 256, 384, 512, 768, 1024, 1536, 2048};
    return quality[qindex];
}

GLuint PreviewProperty_cls::
GetGLTexFilter(){
    if (mCTexFilter.mBody == 1)
	return GL_LINEAR;
    return GL_NEAREST;
}

void PreviewProperty_cls::
ReadStrLine(std::wstring strLine){
    mCConfListPreview.ReadStrLineAll(strLine);
}

wxString PreviewProperty_cls::
SerializeConf(){
    return mCConfListPreview.SerializeAll();
}

PreviewProperty_cls* PreviewProperty_cls::mspCThis;
//FilterCtrl_cls* PreviewProperty_cls::mspCFilterCtrlLink;
