// -*- c -*- 
// (setq c-basic-offset 4)
// "clsDF3.h"

#include "clsDF3.h"
#include "clsDF3SeqCtrl.h"
#include "clsDF3Sequence.h"
#include "clsDF3SeqGUI.h"
#include "GL/clsPreviewProperty.h"

#include "../FILTER/clsFilterCtrl.h"
#include "../MYAPI/MsgBox.h"
#include "FILEIO/clsExportDF3.h"
#include "FILEIO/clsExportImgSeq.h"

#include <fstream>
#include <time.h>

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#define X 0
#define Y 1
#define Z 2

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

// ----------------------

DF3_cls::
DF3_cls(DF3Sequence_cls* pCParent) 
{
    mSelfIndex = -1; // $BL$Dj5A(B
    mFileType = _DF3_FILETYPE_DF3;
    
    mbAllocated = FALSE;
    mbAllocatedXYAxis = FALSE;
    mbDirty = FALSE;
    mbValid = TRUE;
    mpMainXYZ[Z] = NULL;
    mpProcessedXYZ[Z] = NULL;
    LOOP_XYZ
	mCurrSize[xyz] = mTrueSize[xyz] = 0;
    //mDownsizing = 1;
    mwsFilePath = L"";
    mwsFilePathFormat = L"";
    mwsFilePathLoaded = L"";
    mOldPointXYZ = 0;
    mpCParent = pCParent;
    //mpDownSizeLayer = DF3Sequence_cls::msCDownSizeLayer.mpBody;
    //mpDownSizeSrc   = DF3Sequence_cls::msCDownSizeSrc.mpBody;
}

DF3_cls::
~DF3_cls(){
    if (mbAllocated == TRUE) Release();
}

void DF3_cls::
SetSelfIndex(unsigned int index){
    mSelfIndex = index;
}

unsigned int DF3_cls::
GetSelfIndex(){
    return mSelfIndex;
}

void DF3_cls::
Allocate(){
    assert(mbAllocated == FALSE);

    LOOP_XYZ
	mCurrSize[xyz] = mTrueSize[xyz] / 
	DF3Sequence_cls::msCDownSizeSrc.mpBody[xyz];
    if (GetCurrentTotalSize() > (long long)2*1024*1024*1024){
	mbValid = FALSE;
	MsgBox(wxGetTopLevelParent(mpCParent->mspCGUILink), 
	       wxString::Format(_T("%s\n\nsize: %dx%dx%d = %lld byte\n"),
				_("Too big data size (2GB limit is set)"),
				mCurrSize[X], mCurrSize[Y], mCurrSize[Z], 
				GetCurrentTotalSize()),
	       _("Allocation Error"));
	mCReadBodyDF3.Clear();
	mCReadBodyImgSeq.Clear();

	return;
    }
    
    int totalSize = GetCurrentTotalSize();
    //if (totalSize == 0 || mwsFilePath == wxT("")) return;

    mpMainXYZ[Z] = new unsigned char [ totalSize ];
    if (/*msbUseProcessedXYZ == TRUE*/TRUE){
	mpProcessedXYZ[Z] = new unsigned char [ totalSize ];
	mbAllocatedProcessed = TRUE;
	mbProcessFinished = FALSE;
    }
    else{
	mpProcessedXYZ[Z] = mpMainXYZ[Z];
	mbAllocatedProcessed = FALSE;
	mbProcessFinished = FALSE;
    }

    mAllocatedSize = totalSize;
    mbAllocated = TRUE;
    mCReadBodyDF3.Clear();
    mCReadBodyImgSeq.Clear();
}

void DF3_cls::
AllocateXY(){
    assert(mbAllocatedXYAxis == FALSE);

    assert(mCConstructXY[X].IsCleared() == TRUE);
    assert(mCConstructXY[Y].IsCleared() == TRUE);

    mpMainXYZ[X] = new unsigned char [ GetCurrentTotalSize() ];
    mpMainXYZ[Y] = new unsigned char [ GetCurrentTotalSize() ];
    mbAllocatedXYAxis = TRUE;
    mCConstructXY[X].Clear();
    mCConstructXY[Y].Clear();
}

void DF3_cls::
FakeRelease(){
    if (mwsFilePathLoaded != L""){
	if (mbValid == TRUE) assert (mbAllocated == TRUE);
	if (mbAllocatedXYAxis == TRUE) ReleaseXY();

	// $B%H!<%?%k%5%$%:$,0c$&>l9g$N$_<B:]$K(Bdelete$B$7$F:n$jD>$7(B
	//printf("fake release %d %d %d\n", mSelfIndex, mAllocatedSize, GetCurrentTotalSize());
	if (GetCurrentTotalSize() != mAllocatedSize){
	    if (mbAllocated == TRUE) delete mpMainXYZ[Z];
	    if (mbAllocatedProcessed == TRUE) delete mpProcessedXYZ[Z];
	}
    }

    if (GetCurrentTotalSize() != mAllocatedSize){
	mpMainXYZ[Z] = NULL;
	mpProcessedXYZ[Z] = NULL;
	mbAllocated = FALSE;
	mbAllocatedProcessed = FALSE;
    }

    mCReadBodyDF3.Clear();
    mCReadBodyImgSeq.Clear();
    mCLevelMap.Clear();
    mwsFilePathLoaded = L"";
}

void DF3_cls::
Release(){
    if (mwsFilePathLoaded != L""){
	if (mbValid == TRUE) assert(mbAllocated == TRUE);
	if (mbAllocatedXYAxis == TRUE) ReleaseXY();
	if (mbAllocated == TRUE) delete mpMainXYZ[Z];
	if (mbAllocatedProcessed == TRUE) delete mpProcessedXYZ[Z];
    }
    mpMainXYZ[Z] = NULL;
    mpProcessedXYZ[Z] = NULL;
    mbAllocated = FALSE;
    mbAllocatedProcessed = FALSE;

    mCReadBodyDF3.Clear();
    mCReadBodyImgSeq.Clear();
    mCLevelMap.Clear();
    mwsFilePathLoaded = L"";

    //printf("release %d %d %ld\n", mSelfIndex, mAllocatedSize, GetCurrentTotalSize());
}

void DF3_cls::
ReleaseXY(){
    assert(mbAllocatedXYAxis == TRUE);

    delete mpMainXYZ[X];
    delete mpMainXYZ[Y];
    mpMainXYZ[X] = NULL;
    mpMainXYZ[Y] = NULL;
    mbAllocatedXYAxis = FALSE;
    mCConstructXY[X].Clear();
    mCConstructXY[Y].Clear();
}

bool DF3_cls::
IsAllocated(){
    return mbAllocated;
}

bool DF3_cls::
IsAllocatedXY(){
    return mbAllocatedXYAxis;
}

bool DF3_cls::
IsXYConstructFinished(){
    return 
	mCConstructXY[X].IsFinished() &
	mCConstructXY[Y].IsFinished();
}

bool DF3_cls::
IsXYFinishConditionSatisfied(){
    return 
	mCConstructXY[X].IsFinishConditionSatisfied() &
	mCConstructXY[Y].IsFinishConditionSatisfied();
}

int  DF3_cls::
GetConstructPercent(){
    return (mCConstructXY[X].GetPercent() +
	    mCConstructXY[Y].GetPercent()) / 2;
}

unsigned int  DF3_cls::
GetAllocatedXYKB(){
    if (IsAllocatedXY() == FALSE) return 0;
    return GetCurrentTotalSize() * 2 / 1024 ; // $B=hM}8e$N%G!<%?4^$`(B
}

unsigned int  DF3_cls::
GetAllocatedXYZKB(){
    if (IsAllocated() == FALSE) return 0;
    
    unsigned int size = GetCurrentTotalSize() / 1024;
    if (IsAllocatedXY() == TRUE) size += GetCurrentTotalSize() * 2 / 1024 ;
    if (/*msbUseProcessedXYZ == */TRUE) size += GetCurrentTotalSize() / 1024;
    return size;
    
}

bool DF3_cls::
IsReadFinished(){
    switch(mFileType){
    case _DF3_FILETYPE_DF3:
	return mCReadBodyDF3.IsFinished();
    case _DF3_FILETYPE_IMAGE:
	return mCReadBodyImgSeq.IsFinished();
    }
    return FALSE;
}

bool DF3_cls::
IsFileChanged(){
    if (mwsFilePath != mwsFilePathLoaded) return TRUE;
    return FALSE;
}

void DF3_cls::
SetFileChangedTag(){
    if (mwsFilePathLoaded != L"") mwsFilePathLoaded = L".";
}

bool DF3_cls::
IsDirty(){
    return mbDirty;
}

void DF3_cls::
Clean(){
    mbDirty = FALSE;
}

void DF3_cls::
SetFilePath(wxString strFilePath, wxString strFilePathFormat, int channel){
    SplitPathWX_cls CPath(strFilePath);
    if( CPath.GetPathWXStr(_FN_EXT) == wxT(".bmp") || 
	CPath.GetPathWXStr(_FN_EXT) == wxT(".BMP") || 
	CPath.GetPathWXStr(_FN_EXT) == wxT(".png") || 
	CPath.GetPathWXStr(_FN_EXT) == wxT(".PNG") || 
	CPath.GetPathWXStr(_FN_EXT) == wxT(".tif") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".TIF") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".tiff") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".TIFF"))
	mFileType = _DF3_FILETYPE_IMAGE;
    else
	mFileType = _DF3_FILETYPE_DF3;
    mwsFilePath = strFilePath;
    mwsFilePathFormat = strFilePathFormat;
    //mChannel = channel;
    mCReadBodyImgSeq.mChannel = channel;
}

wxString DF3_cls::
GetFilePathWXStr(){
    return mwsFilePath;
}

wxString DF3_cls::
GetFilePathFormat(){
    return mwsFilePathFormat;
}

bool DF3_cls::
ReadHeader(wxWindow *pCWindow){
    mbValid = FALSE; //FALSE$BCf$K%@%$%d%m%0$,=P$J$$$H%k!<%W(B
    switch(mFileType){
    case _DF3_FILETYPE_DF3:
	mbValid = mCReadBodyDF3.ReadHeaderInfo(pCWindow);
	break;
    case _DF3_FILETYPE_IMAGE:
	mbValid = mCReadBodyImgSeq.ReadHeaderInfo(pCWindow);
	break;
    }

    if (mbValid == TRUE) mwsFilePathLoaded = mwsFilePath;
    return mbValid;
}

bool DF3_cls::
ReadBodyIfNotFinished(bool type, wxWindow *pCWindow){
    mbValid = FALSE; //FALSE$BCf$K%@%$%d%m%0$,=P$J$$$H%k!<%W(B
    switch(mFileType){
    case _DF3_FILETYPE_DF3:
	if (mCReadBodyDF3.IsFinished() == FALSE){
	    mbValid = mCReadBodyDF3.MainProcess(type, pCWindow);
	    if (mbValid == FALSE) return FALSE;
	    return TRUE; }
	else 
	    mbValid = TRUE;
	break;
    case _DF3_FILETYPE_IMAGE:
	if (mCReadBodyImgSeq.IsFinished() == FALSE){
	    mbValid = mCReadBodyImgSeq.MainProcess(type, pCWindow);
	    if (mbValid == FALSE) return FALSE;
	    return TRUE; }
	else
	    mbValid = TRUE;
	break;
    }
    return FALSE;
}

bool DF3_cls::
ExportDF3(wxString strFilename, bool bMirroring[3]){

    bool bInMemory = mCReadBodyImgSeq.IsFinished();

    if (IsAllocated() == FALSE){
	if (ReadHeader(wxGetTopLevelParent(mpCParent->mspCGUILink)) == FALSE)
	    return FALSE;
	Allocate();
    }

    if (mFileType == _DF3_FILETYPE_IMAGE){
	if (mCReadBodyImgSeq.IsFinished() == FALSE)
	    mCReadBodyImgSeq.MainProcess(_DF3_FORCE_FINISH, wxGetTopLevelParent(mpCParent->mspCGUILink));
    }
    else{
	if (mCReadBodyDF3.IsFinished() == FALSE)
	    mCReadBodyDF3.MainProcess(_DF3_FORCE_FINISH, wxGetTopLevelParent(mpCParent->mspCGUILink));
    }
    
    ExportDF3_cls CExport(this);
    bool retVal = CExport.OutputDF3(strFilename, bMirroring);

    // $B85$+$i%a%b%jFb$K$J$1$l$P3+J|(B
    if (bInMemory == FALSE) Release();

    return retVal;
}

bool DF3_cls::
ExportIMGSeq(FileSequenceWX_cls CFileSequence, wxStatusBar *pCStatusBar){
    ExportIMGSeq_cls CExport(this);
    bool retVal = CExport.OutputIMGSeq(CFileSequence, pCStatusBar);
    
    return retVal;
}

// ----------------


bool DF3_cls::
ProcessXYZData(){
    if (/*msbUseProcessedXYZ == */FALSE || 
	(mbProcessFinished == TRUE &&
	 mpCParent->mspCParentCtrlLink->
	 IsReprocessRequired(mpCParent->mSelfIndex) == FALSE))
	return FALSE; 
    
    if (IsAllocated() == FALSE) return TRUE;

    bool bTimeRemap = mpCParent->IsRemap();
    float remapFactor = mpCParent->GetRemapInterpolateFactor();
    
    mbProcessFinished = TRUE; // MsgBox$B$,=P$k$H:FEY$-$F$7$^$&$N$G@h$K(B
    mpCParent->mpCFilterCtrl->ProcessXYZData(mpProcessedXYZ[Z], mpMainXYZ[Z], mCurrSize, bTimeRemap, remapFactor, mpCNextDF3->mpMainXYZ[Z]);
    return TRUE;
}

bool DF3_cls::
ConstructXYAxisData(bool type){
    mOldPointXYZ = 5;

    assert(IsAllocatedXY() == TRUE);

    mbConstructType = type;

    ThreadConstructX_cls ThreadX(this, X), ThreadY(this, Y);
    
    ThreadX.Run() ; ThreadY.Run();
    ThreadX.Wait(); ThreadY.Wait();

    if (mbConstructType == _DF3_FORCE_FINISH)
	while (IsXYFinishConditionSatisfied() == FALSE) wxMilliSleep(20);
    
    if (IsXYFinishConditionSatisfied() == FALSE) return FALSE;
    
    mCConstructXY[X].Finalize();
    mCConstructXY[Y].Finalize();
    
    return TRUE;
}

PreviewProperty_cls *DF3_cls::mspCProperty = NULL;

// ------

ThreadConstructX_cls::
ThreadConstructX_cls(DF3_cls *pC, int axis) : wxThread(wxTHREAD_JOINABLE) 
{
    mpC = pC;
    mAxis = axis;
    
    if ( Create() != wxTHREAD_NO_ERROR ) 
	MsgBoxExit(NULL, "Error!", "Can't create thread!");
}

wxThread::ExitCode ThreadConstructX_cls::
Entry(){
    mpC->mCConstructXY[mAxis].Initialize(mpC->GetCurrSize(mAxis));
    
    unsigned int *pNext = &(mpC->mCConstructXY[mAxis].mNext);

    if (mAxis == X){
	// X$B<4$N>l9g(B
	for (unsigned int u = *pNext; u < mpC->GetCurrSize(mAxis); u++){
	    unsigned int uindex = u * mpC->GetCurrSizeSxT(mAxis);
	    for (unsigned int t = 0; t < mpC->GetCurrSizeT(mAxis); t++){
		unsigned int tindex = t * mpC->GetCurrSizeS(mAxis);
		for (unsigned int s = 0; s < mpC->GetCurrSizeS(mAxis); s++)
		    mpC->mpMainXYZ[mAxis][uindex + tindex + s] =
			mpC->mpMainXYZ[Z][mpC->GetIndex(u, s, t)];
	    }
	    *pNext = u + 1;
	    if (mpC->mCConstructXY[mAxis].IsResumeConditionSatisfied
		(mpC->mbConstructType) == TRUE) break;
	}
    }
    else {
	// Y$B<4$N>l9g(B
	for (unsigned int u = *pNext; u < mpC->GetCurrSize(mAxis); u++){
	    unsigned int uindex = u * mpC->GetCurrSizeSxT(mAxis);
	    for (unsigned int t = 0; t < mpC->GetCurrSizeT(mAxis); t++){
		unsigned int tindex = t * mpC->GetCurrSizeS(mAxis);
		for (unsigned int s = 0; s < mpC->GetCurrSizeS(mAxis); s++)
		    mpC->mpMainXYZ[mAxis][uindex + tindex + s] =
			mpC->mpMainXYZ[Z][mpC->GetIndex(s, u, t)];
	    }
	    *pNext = u + 1;
	    if (mpC->mCConstructXY[mAxis].IsResumeConditionSatisfied
		(mpC->mbConstructType) == TRUE) break;
	}
    }
    
    return 0;
}

// -----

/*
ThreadReader_cls::
ThreadReader_cls(DF3_cls *pC) : wxThread(wxTHREAD_DETACHED)
{
    mpC = pC;
    if ( Create() != wxTHREAD_NO_ERROR ) 
	MsgBoxExit(NULL, "Error!", "Can't create thread!");
}

void* ThreadReader_cls::
Entry(){
    if (mpC->IsAllocated() == TRUE){
	mpC->mpCParent->mbSubThreadReader = FALSE;
	return NULL;
    }
    mpC->ReadHeader();
    mpC->Allocate();
    mpC->ReadBody(_DF3_ENABLE_RESUME);
    mpC->mpCParent->mbSubThreadReader = FALSE;
    mpC->mpCParent->Refresh(TRUE);
    return NULL;
}
*/

void ThreadReader_cls::
OnExit(){
    /*
    mpC->mpCParent->mbSubThreadReader = FALSE;
    mpC->Unlock();
    */
}
