// -*- c++ -*- 
// "clsDF3Sequence.cpp"
// (setq c-basic-offset 4)

#ifndef _DF3_SEQUENCE_H_
#define _DF3_SEQUENCE_H_

class DF3Sequence_cls; 
class DF3_cls; 

#include "GL/clsDF3GL.h"
#include "../POVRAY/clsCheckTreeFile.h"
#include "../COMMON/clsFilename.h"
#include "../CONFITEM/clsConfItem.h"
#include "../PARAMETER/clsParameters.h"
//#include "../../clsCameraPath.h"

class DF3SeqGUI_cls; 
class CameraPath_cls;

#include <wx/wx.h>
#include <wx/glcanvas.h>

#include <vector>

#define _DF3_SEQ_MAX 2400

#ifdef WIN32
#define _MEM_PANEL_OFFSET 14
#else
#define _MEM_PANEL_OFFSET 10
#endif

#define _TIME_CTRL_CURRENT -1

#define _DF3SEQ_APPEND_NOADJUST 0
#define _DF3SEQ_APPEND_ADJUST 1

#define _DF3SEQ_TEXTURE 1
#define _DF3SEQ_MEMORY  2
#define _DF3SEQ_ALL 255

#define _DF3SEQ_SKIPTEXPROCESS 1

#define _IDLEPROCESS_DOSOMEPROCESS 1

class DF3SeqCtrl_cls;
class DF3SeqGUI_cls; 
class FilterCtrl_cls;
class PovRay_cls;
class TimeCtrl_cls;

class DF3Sequence_cls {
public:
    DF3Sequence_cls(int selfIndex, DF3SeqCtrl_cls *pCParentCtrl, TimeCtrl_cls *pCTime);
    ~DF3Sequence_cls();
    
    int mSelfIndex;
    int mPreIndex;
    static DF3SeqCtrl_cls *mspCParentCtrlLink;
    static DF3SeqGUI_cls *mspCGUILink;
    static TimeCtrl_cls *mspCTimeLink;

    FilterCtrl_cls *mpCFilterCtrl;

    void SetIndividualFileNames();
    void ClearFileSequence();
    void AppendFileSequence(wxString strFilepath, bool bSelectChannel = FALSE,
			    bool bImageTimeSequence = FALSE);
    
    // ---

    //bool CheckUpdate();
    void BeginBusyCursor();
    void EndBusyCursor();
    int  IdleProcessesMain(unsigned int index, bool type, 
			   bool bSkipTextureProcess = FALSE);
    bool IdleProcessesSubTreadReader(unsigned int index);
    
    // ---

    void UpdatePreFetchList();
    bool IdleProcessesFetch();

    bool *mAPreFetchList;
    bool mbPreFetchListAllocated;
    int  mPreFetchListAllocatedNumber;
    double *mpAlpha;
    bool mbAlphaAlocatted;
    int  mAlphaAllocatedNumber;
    
    bool IsPlayTexConditionSatisfied(unsigned int step);

    int  GetQuickIndex();
    bool DrawData(PositionAndDirection_cls CCam, 
		  double boxScaleX, double boxScaleY, double boxScaleZ,
		  unsigned int axis, bool bMirror[3],
		  double clipBoxMin[3], double clipBoxMax[3],
		  double *pShadeOffset);
    
    void EvaluateOldPoint();
    void ReleaseAll(int flagBit = _DF3SEQ_ALL, bool bFakePermitted = TRUE);
   
    void ReleaseStep(unsigned int index, int flagBit = _DF3SEQ_ALL,
		     bool bFakePermitted = TRUE);
    bool IdleProcessesRelease(bool type);
    void SetFileChangedTagAll();

    // -----

    void SetMaxIndex(unsigned int maxFrame);
    
    unsigned int mMaxIndex; // $B:G8e$N%9%F%C%W$r4^$`(B
    unsigned int mDigit;    // mMaxIndex$B$NM-8z7e?t(B
    
    //ConfItemList_cls mCConfListScene;
    //ConfItemINT_cls  mCAppendType;
   
    bool IsRemap();
    float GetRemapInterpolateFactor();
    unsigned int GetRemappedIndex(int frame = _TIME_CTRL_CURRENT);
    unsigned int GetDataSize(int axis);

    // -----
public:
    DF3GL_cls* GetPDF3();
    std::vector < DF3GL_cls > mVDF3GL;  

    /*
    CheckTreeFile_cls* GetpCheckTreeFile();
    std::vector < CheckTreeFile_cls > mVCCheckTreeFile;
    */

public:
    bool         mbSubThreadReader;
    ThreadReader_cls* pSubThreadReader;

    static ConfItemUINTARRAY_cls msCDownSizeLayer; // $BI=<($N4V0z$-(B
    static ConfItemUINTARRAY_cls msCDownSizeSrc;   // $B%G!<%?$H$7$F$N4V0z$-(B
    
    float mFactorAlpha;
    float mColorRGB[3];
    int   GetColorLayerNum();
    void  SetRGBAtoGL(unsigned int cindex);

    float GetLastLayerPovScale(unsigned int acxis, unsigned int interpolateType);
    
    std::vector < FileSequenceWX_cls > mVCFileSequence; 
    ConfItemUINT_cls mCSequenceNumber;
    int              mDefaultDigit; // 0$B0J30$N$H$-$O<jF0$G7h$a$i$l$F$$$k(B
    int              mDefaultZSliceInit; //$B#2$D$a$N(B%d$B$G@_Dj$5$l$?$H$-$N=i4|CM(B
    int              mDefaultChannel;
    
    wxString     GetCurrentFWXStr();

    // ----

    void ReadStrLineSceneConf(std::wistream &istream, std::wstring strLine);
    void ReadStrLineSequence(std::wistream &istream, std::wstring strLine);
    wxString      SerializeFile();
    
    std::wstring  SerializeFileName();
    std::wstring  SerializeLayerSize(unsigned int interpolateType);
    
    // -- $B>pJsIA2hMQ(B--$B:G=*E*$K$O$$$i$J$/$J$k(B
    int          mOldIndexTex;
    int          mOldIndexXYZ;
    int          mOldPointMaxTex;
    int          mOldPointMaxXYZ;
};

class MaxElement_cls {
 public:
    int          mRangeMin, mRangeMax;
    unsigned int mMax;
    int          mIndex;

    MaxElement_cls(int rangeMin, int rangeMax);
    
    unsigned int GetMax();
    int          GetIndex();

    void         SetIndexAndElement(int index, unsigned int value);
};

#endif
