// -*- c -*- 
// "clsLevelMap.h"

#include "clsLevelMap.h"
#include "clsDF3.h"

#include <stdlib.h>
#include <math.h>

#define X 0
#define Y 1
#define Z 2

LevelMap_cls::
LevelMap_cls() : 
    mpCParent(NULL),
    mpMap(NULL), mbAllocated(FALSE), mbCalculated(FALSE)
{}

LevelMap_cls::
~LevelMap_cls(){
    if (mbAllocated == TRUE) delete [] mpMap;
}

void LevelMap_cls::
SetParent(DF3_cls *pCParent){
    mpCParent = pCParent;
}

void LevelMap_cls::
SetCurrent(LevelMap_cls* pThisCurrent){
    mpThisCurrent = pThisCurrent;
}

void LevelMap_cls::
Allocate(){
    mpMap = new double [256];
    mbAllocated = TRUE;
}

void LevelMap_cls::
Clear(){
    if (mbAllocated == TRUE) delete [] mpMap;
    mbAllocated = FALSE;
    mbCalculated = FALSE;
}

void LevelMap_cls::
Calculate(){
    if (mbAllocated == FALSE) Allocate();
    
    if (mpCParent->IsReadFinished() == FALSE) return;
    
    int totalSize = mpCParent->GetCurrentTotalSize();
    if (totalSize == 0) return;
    for (int i = 0; i < 256; i++) mpMap[i] = 0;
    for (int index = 0; index < totalSize; index++)
	mpMap[mpCParent->mpMainXYZ[Z][index]] += 1;
    double max = 0;
    for (int i = 0; i < 256; i++) 
	if (max < mpMap[i]) max = mpMap[i];
    if (max != 0) 
	for (int i = 0; i < 256; i++) 
	    mpMap[i] = log10(9999.0 * mpMap[i] / max + 1) / 4;
    mbCalculated = TRUE;
    return;
}

double LevelMap_cls::
GetLevelMap(double value, void *pClue){
    if (mpThisCurrent == NULL) return 0.2;
    if (mpThisCurrent->mbCalculated == FALSE) mpThisCurrent->Calculate();
    if (mpThisCurrent->mbCalculated == FALSE) return 0.5;
    int index = value * 255; 
    if (index < 0) index = 0;
    if (index > 255) index = 255;
    return mpThisCurrent->mpMap[index];
}

LevelMap_cls* LevelMap_cls::mpThisCurrent = NULL;
