// -*- c -*-
// "clsHotSpot.h"

#include "clsHotSpot.h"

#include "../MYAPI/MsgBox.h"
#include "../COMMON/MyVector.h"
#include "../DF3/clsDF3Sequence.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../PARAMETER/clsParametersPanel.h"
#include "../WIDGETCTRL/clsAtoiComboCtrl.h"

#include "../../clsFrameMain.h"
#include "../../clsMain.h"
#include "../../clsTimeCtrl.h"

#include "../../wxWidgetEnum.h"

#define LOOP_RGBA for (int rgba = 0; rgba < 4; rgba++)

BEGIN_EVENT_TABLE(HotSpotGUI_cls, wxDialog)
     EVT_BUTTON(wxID_CANCEL,           MultiParametersGUI_cls::OnButtonClose)
     EVT_BUTTON(buttonCtrlSpotAdd    , MultiParametersGUI_cls::OnButtonAdd)
     EVT_BUTTON(buttonCtrlSpotRemove , MultiParametersGUI_cls::OnButtonRemove)
END_EVENT_TABLE()

HotSpotGUI_cls::
HotSpotGUI_cls(HotSpots_cls* pCParent, wxWindow *pC,
	       wxPoint pos, wxSize size) :
     MultiPointInputGUI_cls(pC, pCParent, _("Hot Spot Configure"), pos, size, 
			   L"HOTSPOT_NUM=", L"HSpot", 5)
{
    mspCParent = pCParent;
    mspCThis = this;
    mVCDisplayList.push_back(_GLOBJ_WIREBOX);
    mVIcon.resize(1);
#ifdef WIN32
    wxIcon icon(wxT("ICONS/HSpot.ico"), wxBITMAP_TYPE_ICO);
#else
    wxIcon icon(wxT("ICONS/HSpot.png"), wxBITMAP_TYPE_PNG);
#endif
    SetIcon(icon);
    mVIcon.resize(1);
    mVIcon[0] = icon;

    double pInit[5] = {0, 0, 0, 0.25, 2};
    SetInitParam(pInit);
    
    //SetCloseButtonID(buttonCtrlSpotClose);
    SetScreenRectPos(&pCParent->mCScreenRect);

    // $B%5%$%6!<$r:n$k(B
    wxBoxSizer       *pBoxSizerSrc   = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerBTM   = new wxBoxSizer(wxHORIZONTAL);

    SetMNumMain(pBoxSizerSrc, buttonCtrlSpotAdd, buttonCtrlSpotRemove);
    SetMNumParam(pBoxSizerSrc);
    
    pBoxSizerSrc->Add( pBoxSizerBTM,  0, wxALIGN_RIGHT);

    SetSizerAndFit(pBoxSizerSrc);
    FuncHookItemNumberChanged(NULL, (MultiParameters_cls*)this);
}

HotSpotGUI_cls::
~HotSpotGUI_cls(){
}

void HotSpotGUI_cls::
SetMNumParam(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumParam, this);
#ifdef WIN32
    pCurr->mTextWidth         = 80;
#else
    pCurr->mTextWidth         = 100;
#endif
    pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));

#ifdef WIN32
    pCurr->mMinimumWidth = 180;
#else
    pCurr->mMinimumWidth = 220;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(&FuncHookApply, _MN_HOOK_ASSIGN);
}

void HotSpotGUI_cls::
ReconnectMNumParam(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumParam;

    if (mCItemNum.mBody == 0) {
	pCurr->mVCPointers[0].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[1].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[2].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[3].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[4].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->UpdateValues();
	return;
    }
    
    double *ptr = mVCParam[mActiveItem].GetPInterpolatedValue(mspCParent->mpCParentFrame->mpCParent->mpCTime->GetDoubleIndex());
    pCurr->mVCPointers[0].
	CombinePointer(_GP_DOUBLE, &ptr[0], wxT("x"));
    pCurr->mVCPointers[1].
	CombinePointer(_GP_DOUBLE, &ptr[1], wxT("y"));
    pCurr->mVCPointers[2].
	CombinePointer(_GP_DOUBLE, &ptr[2], wxT("z"));
    pCurr->mVCPointers[3].
	CombinePointer(_GP_DOUBLE, &ptr[3], wxT("scale"));
    pCurr->mVCPointers[4].
	CombinePointer(_GP_DOUBLE, &ptr[4], wxT("strength"));
    pCurr->UpdateValues();
}

void HotSpotGUI_cls::
UpdateActiveOnTL(){
    mspCParent->UpdateActiveOnTL();
}

// ----

void HotSpotGUI_cls::
ReadStrLinePropagate(){
    FuncHookItemNumberChanged(NULL, (MultiParameters_cls*)this);
    UpdateValues();
    //UpdateActiveOnTL();
    UpdateComboBox();
}

void HotSpotGUI_cls::
FuncHookApply(void *pArg, void *pEvent){
    mspCThis->OnApplyBase();
}

HotSpots_cls* HotSpotGUI_cls::mspCParent;
HotSpotGUI_cls* HotSpotGUI_cls::mspCThis;

// -----------------------------------

// std::vector$B$NA}8:$,$"$C$F$b!";X$7<($9%]%$%s%?$,JQ2=$7$J$$$3$HA0Ds!'(B
// ($BI,$:$7$bJ]>Z$5$l$FL5$$$+$b$7$l$J$$$N$G$=$N$&$A3NG'$;$M$P!D(B)

// -----------------------------------

HotSpots_cls::
HotSpots_cls()
{
    SetTag(wxT("[HSpot]"));
}

HotSpots_cls::
~HotSpots_cls(){
    //$B<+F0$G%G%9%H%i%/$H$7$F$$$k$N$GL5$/$FBg>fIW(B
    //if (mpCGUI != NULL) delete mpCGUI;
}

void HotSpots_cls::
InitGUI(wxWindow* pC){
    mpCGUI = new HotSpotGUI_cls(this, pC,
				mCScreenRect.GetPosition(),
				mCScreenRect.GetSize()
				);
    SetPGUI(mpCGUI);
    mpCGUI->UpdateComboBox();
}

double HotSpots_cls::
CalcModifier(double posFocus[3], double posBox[3], double scaleBox){
    double mod = 1;
    for (unsigned int i = 0; i < mpCGUI->mCItemNum.mBody; i++){
	double distance = 
	    MyVector::GetNormRelative(mpCGUI->mVCParam[i].mpInterpolatedValue, posBox);
	if (distance < scaleBox * 0.7 + mpCGUI->mVCParam[i].mpInterpolatedValue[3])
	    mod *= mpCGUI->mVCParam[i].mpInterpolatedValue[4];
    }
    return mod;
}

