// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsCenterOfMass.h"

#include "clsCenterOfMass.h"

#include "../../DF3/clsDF3SeqGUI.h"
#include "../../FILTER/clsFilterCtrl.h"
#include "../../FILTER/FILTERUNDO/FilterUndo.h"
#include "../../MAINCANVAS/clsPreference.h"
#include "../../MAINCANVAS/clsPreviewPanel.h"
#include "../../MYAPI/MsgBox.h"
#include "../../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../../../wxWidgetEnum.h"
#include "../../../wxWidgetEnumMenu.h"

#include <GL/glu.h>

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

MACRO_FilterConstructor(FilterCenterOfMass,,)

void FilterCenterOfMassGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    FilterCenterOfMassData_cls *pCData = mpCParent->GetPCData();
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);
    
    pCurr->mTextWidth = 65;
    int boxWidth = 2;

    pCurr->mItemWidth = -1;
#ifdef WIN32
    pCurr->mComboWidth = 108;
#else
    pCurr->mComboWidth = 98;
#endif

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    wxString listType[2] = { _("[0-1] Coord."), _("Display Coord.") };
    mIndexType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &pCData->mCType.mBody,
	 wxT(""), TRUE, 2, listType);
    int indexCustomA = pCurr->CombinePointerButton
	(buttonCtrlPanelItemsCustomA, _("Update"));
        
    mIndexMessage = pCurr->CombinePointer
	(_GP_STATIC_TEXT, &mpCParent->mstrMessage);

    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    pCurr->UpdateValues();

    mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;

    pCurr->SetFuncHook(&FuncHookCombo, _MN_HOOK_COMBOBOX_CHANGE);
    //pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookChanged, _MN_HOOK_CHANGE);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);
}

MACRO_SetCustomFunctions(FilterCenterOfMassGUI_cls,UpdateResult(),,)
MACRO_HideAndShowEmpty(FilterCenterOfMassGUI_cls);
MACRO_FuncHookHideItem(FilterCenterOfMass);

void FilterCenterOfMassGUI_cls::
UpdateResult(){
    mpCParent->UpdateResult();
}

void FilterCenterOfMassGUI_cls::
FuncHookCombo(void *pArg, void *pEvent){
    ((FilterCenterOfMassGUI_cls*)DF3SeqCtrl_cls::GetPFilterGUI(pArg))
	->mpCParent->UpdateResult();
}

// ---------------------------------

FilterCenterOfMassData_cls::
FilterCenterOfMassData_cls(){
    mCConfList.Add(&mCType, wxT("CenterOfMassType="), L"0");
}

void FilterCenterOfMassData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCType, wxT("CenterOfMassType="));
}

MACRO_MakeInstanceAndAddToList(FilterCenterOfMass)

// ---------------------------------

FilterCenterOfMass_cls::
FilterCenterOfMass_cls(FilterCtrl_cls *pCParentCtrl,
		       FilterCenterOfMassData_cls *pCInitData) 
: FilterBase_cls(pCParentCtrl) {
    if (pCInitData != NULL) mCData = *pCInitData;
    ConstructorCommon(pCParentCtrl, _("CenterOfMass"), _FILTER_CENTEROFMASS);
    mpCGUIPanel = new FilterCenterOfMassGUI_cls(this);

    LOOP_XYZ mCM[xyz] = 0;
    mAverageDensity = 0;
    mstrMessage = wxT("<-0.00000, -0.00000, -0.00000>\nNone");
}

MACRO_Filter_Destructor_Basic(FilterCenterOfMass);

FilterData_cls* FilterCenterOfMass_cls::
GetPData(){
    return &mCData;
}

void FilterCenterOfMass_cls::
AddFilterItemPolymorph(wxBoxSizer *pBoxSizer, FilterPanels_cls *pCFilterPanels,
		       unsigned int index){
    ((FilterCenterOfMassGUI_cls*)mpCGUIPanel)->
	SetMNumPanel(pBoxSizer, pCFilterPanels, index);
}

FilterCenterOfMassData_cls* FilterCenterOfMass_cls::
GetPCData(){
    return (FilterCenterOfMassData_cls*)GetPData();
}

// ----

ThreadCenterOfMass::
ThreadCenterOfMass() : wxThread(wxTHREAD_JOINABLE) {}

void ThreadCenterOfMass::
SetParameters(FilterCenterOfMass_cls *pCParent, unsigned char *pSource,
	      int iniX, int finX, int iniY, int finY, int iniZ, int finZ,
	      unsigned int size[3],
	      double Amomentum[3], double *pMass){
    mpSource = pSource;
    mIniX = iniX; mFinX = finX;
    mIniY = iniY; mFinY = finY;
    mIniZ = iniZ; mFinZ = finZ;
    LOOP_XYZ mSize[xyz] = size[xyz];
    mpMomentum = Amomentum;
    mpMass = pMass;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadCenterOfMass::
Entry(){
    double sumZ = 0;
    double momentumZZ = 0, momentumZY = 0, momentumZX = 0;
    unsigned char *pSource = mpSource;
    for (unsigned int z = mIniZ; z < mFinZ; z++){
	int indexZ = z * mSize[X] * mSize[Y];
	double sumY = 0, momentumYY = 0, momentumYX = 0;
	for (unsigned int y = mIniY; y < mFinY; y++){
	    int indexY = indexZ + y * mSize[X];
	    double sumX = 0, momentumXX = 0;
	    int index = indexY + mIniX;
	    for (unsigned int x = mIniX; x < mFinX; x++){
		sumX += pSource[index];
		momentumXX += (/*0.5+*/x) * pSource[index];
		index++;
	    }
	    momentumXX += 0.5 * sumX; // ́iǰvZ܂Ƃ߂
	    sumY += sumX;
	    momentumYX += momentumXX;
	    momentumYY += (0.5 + y) * sumX;
	}
	sumZ += sumY;
	momentumZX += momentumYX;
	momentumZY += momentumYY;
	momentumZZ += (0.5 + z) * sumY;
    }

    mpMomentum[X] = momentumZX;
    mpMomentum[Y] = momentumZY;
    mpMomentum[Z] = momentumZZ;
    *mpMass = sumZ;

    return 0;
}

// ----

void FilterCenterOfMass_cls::
DoCalc(int frame, unsigned char *pSource, unsigned int size[3]){
    if (GetPCData()->mCbValid.mbBody == FALSE) return;

    double min[3] = {0, 0 ,0}, max[3] = {1, 1, 1};
    GetClippingInformation(min, max);

    unsigned int ini[3], fin[3];
    LOOP_XYZ {
	ini[xyz] = min[xyz] * size[xyz];
	fin[xyz] = max[xyz] * size[xyz];
    }
    
    GbGLLockForMultithread = TRUE;

    ThreadCenterOfMass *pCThread[_MAX_THREAD_NUM];
    int threadNum = mspCPreferenceLink->GetThreadNumFilter();

    int threadIndex = 0;
    int lineStep = 8;
    int blockNum = (fin[Z] - ini[Z] - 1) / lineStep + 1; // 9ɂȂ2
    
    double (*pAMomentum)[3]= new double [blockNum][3];
    double *pMass = new double [blockNum];

    int blockIndex = 0;
    for (unsigned int z = ini[Z]; z < fin[Z]; z += lineStep){
	int threadEndZ = z + lineStep;
	if (threadEndZ > fin[Z]) threadEndZ = fin[Z];
	
	pCThread[threadIndex] = new ThreadCenterOfMass();
	pCThread[threadIndex]->SetParameters
	    (this, pSource,
	     ini[X], fin[X], ini[Y], fin[Y], z, threadEndZ, 
	     size, pAMomentum[blockIndex], &pMass[blockIndex]);
	
	threadIndex++;
	blockIndex++;

	if (threadIndex >= threadNum ||
	    /*z >= fin[Z]*/
	    threadEndZ == fin[Z]){
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Run();
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Wait();
	    for (int i = 0; i < threadIndex; i++) delete pCThread[i];
	    threadIndex = 0;
	}
    }

    double totalMass = 0;
    double totalMomentum[3] = {0, 0, 0};

    for (blockIndex = 0; blockIndex < blockNum; blockIndex++){
	totalMass += pMass[blockIndex];
	LOOP_XYZ totalMomentum[xyz] += pAMomentum[blockIndex][xyz];
    }

    delete [] pAMomentum;
    delete pMass;
	
    // FIX_ME!! ̈悪0̂Ƃ0
    
    if (totalMass == 0)
	LOOP_XYZ mCM[xyz] = 0;
    else
	LOOP_XYZ 
	    mCM[xyz] = totalMomentum[xyz] / totalMass / size[xyz];
    LOOP_XYZ
	mAverageDensity = totalMass / ((fin[Z]-ini[Z]) * (fin[Y]-ini[Y]) * (fin[X]-ini[X]));
    
    LOOP_XYZ mACM[frame][xyz] = mCM[xyz];
    mAAverage[frame] = mAverageDensity;
    
    GbGLLockForMultithread = FALSE;
}

void FilterCenterOfMass_cls::
UpdateResult(){
    if (GbGLLockForMultithread == TRUE){
	mstrMessage = wxT("now calculating...");
	mpCGUIPanel->mpCMNumPanel->UpdateValues();
    }
    
    wxString strLabel  = wxT("Center Of Mass");
    wxString strCoordFormat = wxT("<%0.5f, %0.5f, %0.5f>\n");
    wxString strDataCoord = 
	wxString::Format(strCoordFormat,
			 mCM[X], mCM[Y], mCM[Z]);
    wxString strDisplayCoord = 
	wxString::Format(strCoordFormat,
			 (mCM[X] - 0.5) * mScale[X] + mCenter[X],
			 (mCM[Y] - 0.5) * mScale[Y] + mCenter[Y],
			 (mCM[Z] - 0.5) * mScale[Z] + mCenter[Z]);
    wxString strAverageFormat = wxT("%0.2f (%0.3f%%)");
    wxString strAverage = 
	wxString::Format(strAverageFormat,
			 mAverageDensity, mAverageDensity/255 *100);
    switch(GetPCData()->mCType.mBody){
    case 0:
	mstrMessage = strDataCoord + strAverage; break;
    case 1:
	mstrMessage = strDisplayCoord + strAverage; break;
    }
    mpCGUIPanel->mpCMNumPanel->UpdateValues();
}

void FilterCenterOfMass_cls::
DrawBox(int frame, double center[3], double scale[3]){
    glPushMatrix();

    glColor4f(0.5, 1, 1, 1);

    glTranslated(-0.5, -0.5, -0.5);
    glBegin(GL_LINES);
    glVertex3d(mACM[frame][X]-0.03, mACM[frame][Y], mACM[frame][Z]);
    glVertex3d(mACM[frame][X]+0.03, mACM[frame][Y], mACM[frame][Z]);
    glVertex3d(mACM[frame][X], mACM[frame][Y]-0.03, mACM[frame][Z]);
    glVertex3d(mACM[frame][X], mACM[frame][Y]+0.03, mACM[frame][Z]);
    glVertex3d(mACM[frame][X], mACM[frame][Y], mACM[frame][Z]-0.03);
    glVertex3d(mACM[frame][X], mACM[frame][Y], mACM[frame][Z]+0.03);
    glEnd();

    glPopMatrix();

    LOOP_XYZ mCM[xyz] = mACM[frame][xyz];
    mAverageDensity = mAAverage[frame];

    LOOP_XYZ mCenter[xyz] = center[xyz];
    LOOP_XYZ mScale[xyz] = scale[xyz];
}

// ----

wxString FilterCenterOfMass_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("CenterOfMass::");
    return strStatus;
}

MACRO_SerializeConf(FilterCenterOfMass)
MACRO_ReadDataBlock(FilterCenterOfMass)
