// -*- c -*- 
// (setq c-basic-offset 4)
// "clsMorphOpenClose.h"

#include "clsMorphOpenClose.h"

#include "../../DF3/clsDF3SeqGUI.h"
#include "../../FILTER/clsFilterCtrl.h"
#include "../../MAINCANVAS/clsPreference.h"
#include "../../MAINCANVAS/clsPreviewPanel.h"
#include "../../MYAPI/MsgBox.h"
#include "../../WIDGETCTRL/clsMultiNumericalInputs.h"

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

MACRO_FilterConstructor(MorphOpenClose,,)

void MorphOpenCloseGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

    pCurr->mTextWidth = 45;
    pCurr->mItemWidth = -1;
#ifdef WIN32
    pCurr->mComboWidth = 65;
#else
    pCurr->mComboWidth = 65;
#endif
    
    int boxWidth = 2;

    wxString listType[6] = { _("Dilation(Expand)"), _("Erosion(Shirink)"), _("Opening(Remove Noise)"), _("Closing(Fill Gaps)"), wxT("Opening(EEDD)"), wxT("Closing(DDEE)") };

    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mpCParent->mCData.mCType.mBody,
			  wxT(""), TRUE, 6, listType);
    
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Kernel"));
    
    wxString listKernel[2] = { _("Sphere"), _("Ellipsoid") };
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mpCParent->mCData.mCKernelType.mBody,
			  wxT(""), TRUE, 2, listKernel);
    mIndexKernelSize = pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mCData.mCKernelSize.mBody);
    pCurr->SetRange(1, 12);

    boxWidth = 3;
    mIndexAsymSize = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mCData.mCKernelAssimSize.mpBody[0]);
    pCurr->SetRange(0, 12);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mCData.mCKernelAssimSize.mpBody[1]);
    pCurr->SetRange(0, 12);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mCData.mCKernelAssimSize.mpBody[2]);
    pCurr->SetRange(0, 12);
    
    pCurr->mProportion = 1;
    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    pCurr->SetFuncHook(&FuncHookHideItem, _MN_HOOK_COMBOBOX_CHANGE);
    //pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookChanged, _MN_HOOK_CHANGE);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);

    HideAndShowItems();
}

MACRO_SetCustomFunctions(MorphOpenCloseGUI_cls,,,)

// -----

MorphOpenCloseData_cls::
MorphOpenCloseData_cls(){
    mCConfList.Add(&mCType, wxT("MorphOpenCloseType="), L"0");
    mCConfList.Add(&mCKernelType, wxT("MorphOpenCloseKernelType="), L"0");
    mCConfList.Add(&mCKernelSize, wxT("MorphOpenCloseKernelSize="), L"2");
    mCConfList.Add(&mCKernelAssimSize, wxT("MorphOpenCloseAsymSize="), L"2 2 2");
}

void MorphOpenCloseData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCType, wxT("MorphOpenCloseType="));
    mCConfList.Add(&mCKernelType, wxT("MorphOpenCloseKernelType="));
    mCConfList.Add(&mCKernelSize, wxT("MorphOpenCloseKernelSize="));
    mCConfList.Add(&mCKernelAssimSize, wxT("MorphOpenCloseAsymSize="));
}

MACRO_MakeInstanceAndAddToList(MorphOpenClose)

void MorphOpenCloseGUI_cls::
HideAndShowItems(){
    if (mpCMNumPanel->mbPanelOpen == FALSE) return;
    
    int kernelType = mpCParent->mCData.mCKernelType.mBody;

    bool bAsym = (kernelType == _KERNEL_ELLIPSOID);
    mpCMNumPanel->ShowItems(mIndexAsymSize, 4, bAsym);
    mpCMNumPanel->mVCPointers[mIndexKernelSize].Enable(!bAsym);

    mpCParent->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
}

MACRO_FuncHookHideItem(MorphOpenClose);

// -----

MorphOpenClose_cls::
MorphOpenClose_cls(FilterCtrl_cls *pCParentCtrl, MorphOpenCloseData_cls *pCInitData)
: FilterBase_cls(pCParentCtrl){
    if (pCInitData != NULL) mCData = *pCInitData;
    mbProcessData = TRUE;
    ConstructorCommon(pCParentCtrl, _("Morph Filter"), _FILTER_MORPH_OPENCLOSE);
    mpCGUIPanel = new MorphOpenCloseGUI_cls(this);
}

MACRO_Filter_Destructor_Basic(MorphOpenClose);

FilterData_cls* MorphOpenClose_cls::
GetPData(){
    return &mCData;
}

MACRO_AddFilterItemPolymorph(MorphOpenClose);

MorphOpenCloseData_cls* MorphOpenClose_cls::
GetPCData(){
    return &mCData;
}

// ---------

ThreadMorphDilation::
ThreadMorphDilation(wxThreadKind kind) : MultiThreadThread_cls (kind) {}

void ThreadMorphDilation::
SetParameters(MorphOpenClose_cls *pCParent, 
	      unsigned char ***pFrom, unsigned char ***pTo, int size[3],
	      int seSize_x, int seSize_y, int seSize_z,
	      char se[_SE_MAX][_SE_MAX][_SE_MAX],
	      char *pseList, int seListNum, char *pseDiff, int seDiffNum,
	      int iniZ, int finZ){
    mpCParent = pCParent;
    mpFrom = pFrom;
    mpTo = pTo;
    LOOP_XYZ mSize[xyz] = size[xyz];
    mSeSize_x = seSize_x; mSeSize_y = seSize_y; mSeSize_z = seSize_z;
    mse = se;
    mIniZ = iniZ; mFinZ = finZ;

    mpseList = pseList;
    mseListNum = seListNum;
    mpseDiff = pseDiff;
    mseDiffNum = seDiffNum;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadMorphDilation::
Entry(){
    int seSize_x = mSeSize_x;
    int seSize_y = mSeSize_y;
    int seSize_z = mSeSize_z;

    int xc = seSize_x/2; 
    int yc = seSize_y/2;
    int zc = seSize_z/2;

    int currMax = 0;
    int seMaxPosX = 0, seMaxPosY = 0, seMaxPosZ = 0;
    
    unsigned char *pFrom = mpFrom[0][0];
    char *pse = mse[0][0];

    for(int z = mIniZ ; z < mFinZ; z++){
	mzOffset = z+zc; 
	if (z < zc || z >= mSize[Z] - zc)
	    for (int y = 0; y < mSize[Y]; y++){
		myOffset = y+yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopDilationEdge
		    (y, z, xc, yc, zc, 0, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	else{
	    for (int y = 0; y < yc; y++){
		myOffset = y+yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopDilationEdge
		    (y, z, xc, yc, zc, 0, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	    for (int y = yc; y < mSize[Y]-yc; y++){
		myOffset = y+yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopDilationEdge
		    (y, z, xc, yc, zc, 0, xc, pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
		KernelLoopDilationCenter
		    (y, z, xc, yc, zc, pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
		KernelLoopDilationEdge
		    (y, z, xc, yc, zc, mSize[X]-xc, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	    for (int y = mSize[Y] - yc; y < mSize[Y]; y++){
		myOffset = y+yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopDilationEdge
		    (y, z, xc, yc, zc, 0, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	}
    }
    
    *mpbReadyFlag = TRUE;
    return 0;
}

// Ő[if炷߂ɕ
unsigned char ThreadMorphDilation::
ThreadSeLoopDilationEdge
(unsigned char *pFrom, 
 unsigned char max, char *pList, int listNum,
 const int indexListMax,
 char pMaxPos[3]){
    const int sizeX = mSize[X], sizeY = mSize[Y], sizeZ = mSize[Z];
    const int zoffset = mzOffset, yoffset = myOffset, xoffset = mxOffset;
    int maxIndex = -1;
    for (int index = 0; index < indexListMax; index+=3){
	int zn = zoffset - pList[index+Z];
	if (zn < 0) continue;
	if (zn >= sizeZ) continue;
		
	int yn = yoffset - pList[index+Y];
	if (yn < 0) continue;
	if (yn >= sizeY) continue;
	
	int xn = xoffset - pList[index+X];
	if (xn < 0) continue;
	if (xn >= sizeX) continue;
	
	//炭core2ȍ~̓ANZXvẐ̕
	//҂̕B
	//gv = pFrom[zn][yn][xn];//+se[h][r][c];
	unsigned char gv = pFrom[sizeX * (zn * sizeY + yn) + xn];
	if(gv>max){ 
	    max=gv;
	    maxIndex = index;
	}
    }
    if (maxIndex != -1)
	memcpy(pMaxPos, pList + maxIndex, 3);
    else if (max == 0){ // 0̎͑S0Ȃ̂ō𑝂₷
	pMaxPos[X] = 0;
	pMaxPos[Y] = mSeSize_y / 2;
	pMaxPos[Z] = mSeSize_z / 2; 
    }
    return max;
}

unsigned char ThreadMorphDilation::
ThreadSeLoopDilationCenter
(unsigned char *pFrom, 
 unsigned char max, char *pList, int listNum,
 const int indexListMax,
 char pMaxPos[3]){
    const int sizeX = mSize[X], sizeY = mSize[Y], sizeZ = mSize[Z];
    const int zoffset = mzOffset, yoffset = myOffset, xoffset = mxOffset;
    int maxIndex = -1;
    for (int index = 0; index < indexListMax; index+=3){
	int zn = zoffset - pList[index+Z];
	int yn = yoffset - pList[index+Y];
	int xn = xoffset - pList[index+X];
	
	//炭core2ȍ~̓ANZXvẐ̕
	//҂̕B
	//gv = pFrom[zn][yn][xn];//+se[h][r][c];
	unsigned char gv = pFrom[sizeX * (zn * sizeY + yn) + xn];
	if(gv>max){ 
	    max=gv;
	    maxIndex = index;
	}
    }
    if (maxIndex != -1)
	memcpy(pMaxPos, pList + maxIndex, 3);
    else if (max == 0){ // 0̎͑S0Ȃ̂ō𑝂₷
	pMaxPos[X] = 0;
	pMaxPos[Y] = mSeSize_y / 2;
	pMaxPos[Z] = mSeSize_z / 2; 
    }
    return max;
}

void ThreadMorphDilation::
KernelLoopDilationEdge
(int y, int z, int xc, int yc, int zc, int iniX, int finX,
 unsigned char *pTo, unsigned char *pFrom, char *pse,
 char *pseList, int seListNum, char *pseDiff, int seDiffNum ){

    unsigned char max;
    unsigned char currMax = 0;
    char seMaxPos[3];
    seMaxPos[X] = xc*2; seMaxPos[Y] = 0; seMaxPos[Z] = 0;
    const int indexListMax = seListNum*3, indexDiffMax = seDiffNum*3;

    for (int x = iniX; x < finX; x++){
	seMaxPos[X]++; // sȇ傫قx̗ĂȂ̂
	mxOffset = x+xc;
	if (seMaxPos[X] > xc*2 || 
	    pse[_SE_MAX * (_SE_MAX * seMaxPos[Z] + seMaxPos[Y])
		+ seMaxPos[X]] == -1)
	    //ǑvZʂ̗płȂ̂Ńtɒׂ
	    max = ThreadSeLoopDilationEdge
		(pFrom, 0, pseList, seListNum,
		 indexListMax, seMaxPos);
	else 
	    //ǑvZʂpł̂ŁAׂ΂I
	    max = ThreadSeLoopDilationEdge
		(pFrom, currMax, pseDiff, seDiffNum,
		 indexDiffMax, seMaxPos);
	currMax = max;
	pTo[x] = max;
    }
}

void ThreadMorphDilation::
KernelLoopDilationCenter
(int y, int z, int xc, int yc, int zc, 
 unsigned char *pTo, unsigned char *pFrom, char *pse,
 char *pseList, int seListNum, char *pseDiff, int seDiffNum ){

    unsigned char max;
    unsigned char currMax = 0;
    char seMaxPos[3];
    seMaxPos[X] = xc*2; seMaxPos[Y] = 0; seMaxPos[Z] = 0;
    const int indexListMax = seListNum*3, indexDiffMax = seDiffNum*3;

    for (int x = xc; x < mSize[X] - xc; x++){
	seMaxPos[X]++; // sȇ傫قx̗ĂȂ̂
	mxOffset = x+xc;
	if (seMaxPos[X] > xc*2 || 
	    pse[_SE_MAX * (_SE_MAX * seMaxPos[Z] + seMaxPos[Y])
	       + seMaxPos[X]] == -1)
	    //tɒׂ
	    max = ThreadSeLoopDilationCenter
		(pFrom, 0, pseList, seListNum,
		 indexListMax, seMaxPos);
	else 
	    //ׂ΂I
	    max = ThreadSeLoopDilationCenter
		(pFrom, currMax, pseDiff, seDiffNum,
		 indexDiffMax, seMaxPos);
	currMax = max;
	pTo[x] = max;
    }
}

// ---------

ThreadMorphErosion::
ThreadMorphErosion(wxThreadKind kind) : MultiThreadThread_cls (kind) {}

void ThreadMorphErosion::
SetParameters(MorphOpenClose_cls *pCParent, 
	      unsigned char ***pFrom, unsigned char ***pTo, int size[3],
	      int seSize_x, int seSize_y, int seSize_z, 
	      char se[_SE_MAX][_SE_MAX][_SE_MAX],
	      char *pseList, int seListNum, char *pseDiff, int seDiffNum,
	      int iniZ, int finZ){
    mpCParent = pCParent;
    mpFrom = pFrom;
    mpTo = pTo;
    LOOP_XYZ mSize[xyz] = size[xyz];
    mSeSize_x = seSize_x; mSeSize_y = seSize_y; mSeSize_z = seSize_z;
    mse = se;
    mIniZ = iniZ; mFinZ = finZ;

    mpseList = pseList;
    mseListNum = seListNum;
    mpseDiff = pseDiff;
    mseDiffNum = seDiffNum;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadMorphErosion::
Entry(){
    int seSize_x = mSeSize_x;
    int seSize_y = mSeSize_y;
    int seSize_z = mSeSize_z;

    int xc = seSize_x/2; 
    int yc = seSize_y/2;
    int zc = seSize_z/2;

    int seMaxPosX = 0, seMaxPosY = 0, seMaxPosZ = 0;
    
    unsigned char *pFrom = mpFrom[0][0];
    char *pse = mse[0][0];

    for(int z = mIniZ ; z < mFinZ; z++){
	mzOffset = z-zc; 
	if (z < zc || z >= mSize[Z] - zc)
	    for (int y = 0; y < mSize[Y]; y++){
		myOffset = y-yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopErosionEdge
		    (y, z, xc, yc, zc, 0, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	else{
	    for (int y = 0; y < yc; y++){
		myOffset = y-yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopErosionEdge
		    (y, z, xc, yc, zc, 0, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	    for (int y = yc; y < mSize[Y]-yc; y++){
		myOffset = y-yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopErosionEdge
		    (y, z, xc, yc, zc, 0, xc, pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
		KernelLoopErosionCenter
		    (y, z, xc, yc, zc, pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
		KernelLoopErosionEdge
		    (y, z, xc, yc, zc, mSize[X]-xc, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	    for (int y = mSize[Y] - yc; y < mSize[Y]; y++){
		myOffset = y-yc;
		unsigned char *pTo = mpTo[z][y];
		KernelLoopErosionEdge
		    (y, z, xc, yc, zc, 0, mSize[X], pTo, pFrom, pse,
		     mpseList, mseListNum, mpseDiff, mseDiffNum );
	    }
	}
    }
    
    *mpbReadyFlag = TRUE;
    return 0;
}

unsigned char ThreadMorphErosion::
ThreadSeLoopErosionEdge
(unsigned char *pFrom, 
 unsigned char min, char *pList, int listNum,
 const int indexListMax,
 char pMaxPos[3]){
    const int sizeX = mSize[X], sizeY = mSize[Y], sizeZ = mSize[Z];
    const int zoffset = mzOffset, yoffset = myOffset, xoffset = mxOffset;
    int minIndex = -1;
    for (int index = 0; index < indexListMax; index+=3){
	int zn = zoffset + pList[index+Z];
	if (zn < 0) continue;
	if (zn >= sizeZ) continue;

	int yn = yoffset + pList[index+Y];
	if (yn < 0) continue;
	if (yn >= sizeY) continue;
	
	int xn = xoffset + pList[index+X];
	if (xn < 0) continue;
	if (xn >= sizeX) continue;
	
	//炭core2ȍ~̓ANZXvẐ̕
	//҂̕B
	//gv = pFrom[zn][yn][xn];//+se[h][r][c];
	unsigned char gv = pFrom[sizeX * (zn * sizeY + yn) + xn];
	if(gv<min){ 
	    min=gv;
	    minIndex = index;
	}
    }
    if (minIndex != -1)
	memcpy(pMaxPos, pList + minIndex, 3);
    return min;
}

unsigned char ThreadMorphErosion::
ThreadSeLoopErosionCenter
(unsigned char *pFrom, 
 unsigned char min, char *pList, int listNum,
 const int indexListMax,
 char pMaxPos[3]){
    const int sizeX = mSize[X], sizeY = mSize[Y], sizeZ = mSize[Z];
    const int zoffset = mzOffset, yoffset = myOffset, xoffset = mxOffset;
    int minIndex = -1;
    for (int index = 0; index < indexListMax; index+=3){
	int zn = zoffset + pList[index+Z];
	int yn = yoffset + pList[index+Y];
	int xn = xoffset + pList[index+X];
	
	//炭core2ȍ~̓ANZXvẐ̕
	//҂̕B
	//gv = pFrom[zn][yn][xn];//+se[h][r][c];
	unsigned char gv = pFrom[sizeX * (zn * sizeY + yn) + xn];
	if(gv<min){ 
	    min=gv;
	    minIndex = index;
	}
    }
    if (minIndex != -1)
	memcpy(pMaxPos, pList + minIndex, 3);
    return min;
}

void ThreadMorphErosion::
KernelLoopErosionEdge
(int y, int z, int xc, int yc, int zc, int iniX, int finX, 
 unsigned char *pTo, unsigned char *pFrom, char *pse, 
 char *pseList, int seListNum, char *pseDiff, int seDiffNum){

    unsigned char min = 255;
    unsigned char currMin = 255;
    char seMaxPos[3] = {0, 0, 0};
    const int indexListMax = seListNum*3, indexDiffMax = seDiffNum*3;

    for (int x = iniX; x < finX; x++){
	seMaxPos[X]--; // sȅقx̗ĂȂ̂
	mxOffset = x-xc;
	if (seMaxPos[X] < 0 || 
	    pse[_SE_MAX * (_SE_MAX * seMaxPos[Z] + seMaxPos[Y])
	       + seMaxPos[X]] == -1)
	    //tɒׂ
	    min = ThreadSeLoopErosionEdge
		(pFrom, 255, pseList, seListNum,
		 indexListMax, seMaxPos);
	else if (min != 0) // currMin=min=0Ȃ璲ׂȂėǂ 
	    //ׂ΂I
	    min = ThreadSeLoopErosionEdge
		(pFrom, currMin, pseDiff, seDiffNum,
		 indexDiffMax, seMaxPos);
	currMin = min;
	pTo[x] = min;
    }
}

void ThreadMorphErosion::
KernelLoopErosionCenter
(int y, int z, int xc, int yc, int zc,
 unsigned char *pTo, unsigned char *pFrom,  char *pse,
 char *pseList, int seListNum, char *pseDiff, int seDiffNum){

    unsigned char min = 255;
    unsigned char currMin = 255;
    char seMaxPos[3] = {0, 0, 0};
    const int indexListMax = seListNum*3, indexDiffMax = seDiffNum*3;

    for (int x = xc; x < mSize[X] - xc; x++){
	seMaxPos[X]--; // sȅقx̗ĂȂ̂
	mxOffset = x-xc;
	if (seMaxPos[X] < 0 || 
	    pse[_SE_MAX * (_SE_MAX * seMaxPos[Z] + seMaxPos[Y])
	       + seMaxPos[X]] == -1)
	    //tɒׂ
	    min = ThreadSeLoopErosionCenter
		(pFrom, 255, pseList, seListNum,
		 indexListMax, seMaxPos);
	else /*if (min != 0)*/ // currMin=min=0Ȃ璲ׂȂėǂ 
	    //ׂ΂I
	    min = ThreadSeLoopErosionCenter
		(pFrom, currMin, pseDiff, seDiffNum,
		 indexDiffMax, seMaxPos);
	currMin = min;
	pTo[x] = min;
    }
}

// ---------

void MorphOpenClose_cls::
Dilation3D(unsigned char ***pFrom, unsigned char ***pTo, 
	   unsigned char se_org[_SE_MAX][_SE_MAX][_SE_MAX]){
    
    int seSize_x = mSeSize_x, seSize_y = mSeSize_y, seSize_z = mSeSize_z;

    char se[_SE_MAX][_SE_MAX][_SE_MAX];
    int seListNum = 0;
    char seList[_SE_MAX * _SE_MAX * _SE_MAX * 3]; //char ŏ\Ȃ͂I
    int seDiffNum = 0;
    char seDiff[_SE_MAX * _SE_MAX * 4 * 3]; //قǕςȌ`łȂΑ\
    
    for(int z = 0; z < seSize_z; z++) // tseList
	for(int y = 0; y < seSize_y; y++)
	    for(int x = 0; x < seSize_x; x++)
		if(se_org[z][y][x] != 0){ // org  0
		    se[z][y][x] = 0;
		    seList[seListNum*3+X] = x;
		    seList[seListNum*3+Y] = y;
		    seList[seListNum*3+Z] = z;
		    seListNum++;
		}
		else
		    se[z][y][x] = -1;

    for (int z = 0; z < seSize_z; z++) // seList
	for(int y = 0; y < seSize_y; y++)
	    for(int x = 0; x < seSize_x; x++)
		if (x == 0 && se_org[z][y][x] != 0){
		    seDiff[seDiffNum*3+X] = x;
		    seDiff[seDiffNum*3+Y] = y;
		    seDiff[seDiffNum*3+Z] = z;
		    seDiffNum++;
		}
		else if(se_org[z][y][x] != 0 && se_org[z][y][x-1] == 0){
		    seDiff[seDiffNum*3+X] = x;
		    seDiff[seDiffNum*3+Y] = y;
		    seDiff[seDiffNum*3+Z] = z;
		    seDiffNum++;
		}

    InitializeThreadCtrl();

    int lineStep = 8;
    if (mThreadNum == 1) lineStep = mSize[Z]; 
    
    for (int z = 0; z < mSize[Z]; z += lineStep){
	int threadEndZ = z + lineStep;
	if (threadEndZ > mSize[Z]) threadEndZ = mSize[Z];

	ThreadMorphDilation *pCThread = 
	    new ThreadMorphDilation(wxTHREAD_DETACHED);
	pCThread->SetParameters
	    (this, pFrom, pTo, mSize, 
	     seSize_x, seSize_y, seSize_z, se,
	     seList, seListNum, seDiff, seDiffNum,
	     z, threadEndZ);
	WaitForThreadsAndRun(pCThread);
    }
    WaitAllThreadsEnd();
}

void MorphOpenClose_cls::
Erosion3D(unsigned char ***pFrom, unsigned char *** pTo, 
	  unsigned char se_org[_SE_MAX][_SE_MAX][_SE_MAX]){

    int seSize_x = mSeSize_x, seSize_y = mSeSize_y, seSize_z = mSeSize_z;
    
    char se[_SE_MAX][_SE_MAX][_SE_MAX];

    int seListNum = 0;
    char seList[_SE_MAX * _SE_MAX * _SE_MAX * 3]; //char ŏ\Ȃ͂I
    int seDiffNum = 0;
    char seDiff[_SE_MAX * _SE_MAX * 4 * 3]; //قǕςȌ`łȂΑ\

    for(int z = 0; z < seSize_z; z++) // tseList _CHECK_ME!!
	for(int y = 0; y < seSize_y; y++)
	    for(int x = seSize_x - 1; x >= 0; x--) //tB
		if(se_org[z][y][x] != 0){
		    se[z][y][x] = 0;
		    seList[seListNum*3+X] = x;
		    seList[seListNum*3+Y] = y;
		    seList[seListNum*3+Z] = z;
		    seListNum++;
		}
		else
		    se[z][y][x] = -1;

    for (int z = 0; z < seSize_z; z++) // seList _CHECK_ME!!
	for(int y = 0; y < seSize_y; y++)
	    for(int x = seSize_x - 1; x >= 0; x--)
		if ((x == seSize_x - 1) && (se_org[z][y][x] != 0)){
		    seDiff[seDiffNum*3+X] = x;
		    seDiff[seDiffNum*3+Y] = y;
		    seDiff[seDiffNum*3+Z] = z;
		    seDiffNum++;
		}
		else if((se_org[z][y][x] != 0) && (se_org[z][y][x+1] == 0)){
		    seDiff[seDiffNum*3+X] = x;
		    seDiff[seDiffNum*3+Y] = y;
		    seDiff[seDiffNum*3+Z] = z;
		    seDiffNum++;
		}
    
    InitializeThreadCtrl();

    int lineStep = 8;
    if (mThreadNum == 1) lineStep = mSize[Z]; 

    for (int z = 0; z < mSize[Z]; z += lineStep){
	int threadEndZ = z + lineStep;
	if (threadEndZ > mSize[Z]) threadEndZ = mSize[Z];

	ThreadMorphErosion *pCThread = 
	    new ThreadMorphErosion(wxTHREAD_DETACHED);
	pCThread->SetParameters
	    (this, pFrom, pTo, mSize, 
	     seSize_x, seSize_y, seSize_z, se,
	     seList, seListNum, seDiff, seDiffNum,
	     z, threadEndZ);
	WaitForThreadsAndRun(pCThread);
    }
    WaitAllThreadsEnd();
}

bool MorphOpenClose_cls::
ProcessXYZData(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3]){
    LOOP_XYZ mSize[xyz] = size[xyz];
    int totalSize = size[X] * size[Y] * size[Z];
    
    if (mCData.mCbValid.mbBody == FALSE){
	memcpy(pTo, pFrom, totalSize);
	return FALSE;
    }

    GbGLLockForMultithread = TRUE;
    
    // IŃIɘAȂOz̍쐬. ł̂ȁH
    mpObj = new unsigned char** [size[Z]];
    mpObj[0] = new unsigned char* [size[Z] * size[Y]];
    mpObj[0][0] = new unsigned char [size[Z] * size[Y] * size[X]];
    for (int z = 0; z < mSize[Z]; z++) {
        mpObj[z] = mpObj[0] + z * size[Y];
        for (int y = 0; y < mSize[Y]; y++)
            mpObj[z][y] = mpObj[0][0] + z * size[Y] * size[X] + y * size[X];
    }

    mpObjB = new unsigned char** [size[Z]];
    mpObjB[0] = new unsigned char* [size[Z] * size[Y]];
    mpObjB[0][0] = new unsigned char [size[Z] * size[Y] * size[X]];
    for (int z = 0; z < mSize[Z]; z++) {
        mpObjB[z] = mpObjB[0] + z * size[Y];
        for (int y = 0; y < mSize[Y]; y++)
            mpObjB[z][y] = mpObjB[0][0] + z * size[Y] * size[X] + y * size[X];
    }

    memcpy(mpObj[0][0], pFrom, totalSize);

    unsigned char se_org[_SE_MAX][_SE_MAX][_SE_MAX];

    switch(mCData.mCKernelType.mBody){
    case _KERNEL_SPHERE:
	mSeSize_x = mCData.mCKernelSize.mBody * 2 + 1;
	mSeSize_y = mCData.mCKernelSize.mBody * 2 + 1;
	mSeSize_z = mCData.mCKernelSize.mBody * 2 + 1;
	break;
    case _KERNEL_ELLIPSOID:
	mSeSize_x = mCData.mCKernelAssimSize.mpBody[X] * 2 + 1;
	mSeSize_y = mCData.mCKernelAssimSize.mpBody[Y] * 2 + 1;
	mSeSize_z = mCData.mCKernelAssimSize.mpBody[Z] * 2 + 1;
    }

    MakeSphericalSE(se_org);
    //load_SE(se_org);

    switch (mCData.mCType.mBody){
    case _MORPH_DILATION:
	Dilation3D(mpObj, mpObjB, se_org);
	memcpy(pTo, mpObjB[0][0], totalSize);
	break;
    case _MORPH_EROSION:
	Erosion3D(mpObj, mpObjB, se_org);
	memcpy(pTo, mpObjB[0][0], totalSize);
	break;
    case _MORPH_OPENING:
	Erosion3D(mpObj, mpObjB, se_org);
	Dilation3D(mpObjB, mpObj, se_org);
	memcpy(pTo, mpObj[0][0], totalSize);
	break;
    case _MORPH_CLOSING:
	Dilation3D(mpObj, mpObjB, se_org);
	Erosion3D(mpObjB, mpObj, se_org);
	memcpy(pTo, mpObj[0][0], totalSize);
	break;
    case _MORPH_OPENING_EEDD:
	Erosion3D(mpObj, mpObjB, se_org);
	Erosion3D(mpObjB, mpObj, se_org);
	Dilation3D(mpObj, mpObjB, se_org);
	Dilation3D(mpObjB, mpObj, se_org);
	memcpy(pTo, mpObj[0][0], totalSize);
	break;
    case _MORPH_CLOSING_DDEE:	
	Dilation3D(mpObj, mpObjB, se_org);
	Dilation3D(mpObjB, mpObj, se_org);
	Erosion3D(mpObj, mpObjB, se_org);
	Erosion3D(mpObjB, mpObj, se_org);
	memcpy(pTo, mpObj[0][0], totalSize);
	break;
    }
    
    delete[] mpObj[0][0];
    delete[] mpObj[0];
    delete[] mpObj;

    delete[] mpObjB[0][0];
    delete[] mpObjB[0];
    delete[] mpObjB;

    GbGLLockForMultithread = FALSE;
	
    return TRUE;
}

void MorphOpenClose_cls::
MakeSphericalSE(unsigned char se_org[_SE_MAX][_SE_MAX][_SE_MAX]){
    int xc = mSeSize_x / 2;
    int yc = mSeSize_y / 2;
    int zc = mSeSize_z / 2;
    // ~̃TCY͂肬傫Ȃ悤
    double kernel2 = ((double)xc+0.5) * ((double)xc+0.5);
    for (int z = 0; z < mSeSize_z; z++){
	int dz = (z - zc) * mSeSize_x / mSeSize_z;
	for (int y = 0; y < mSeSize_y; y++){
	    int dy = (y - yc) * mSeSize_x / mSeSize_y;
	    for (int x = 0; x < mSeSize_x; x++){
		int dx = x - xc;
		int distance2 = dx*dx + dy*dy + dz*dz;
		if (distance2 < kernel2)
		    se_org[z][y][x] = 1;
		else
		    se_org[z][y][x] = 0;
	    }
	}
    }
}

wxString MorphOpenClose_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("Morphology::") << GetPCData()->mCbValid.mbBody;

    if (GetPCData()->mCbValid.mbBody == FALSE)
	strStatus<< wxT("::");
    else
	strStatus << wxT(":") <<
	    GetPCData()->mCType.mBody << wxT(":") <<
	    GetPCData()->mCKernelType.mBody << wxT(":") <<
	    GetPCData()->mCKernelSize.mBody << wxT(":") <<
	    GetPCData()->mCKernelAssimSize.mpBody[0] << wxT(":") <<
	    GetPCData()->mCKernelAssimSize.mpBody[1] << wxT(":") <<
	    GetPCData()->mCKernelAssimSize.mpBody[2] << wxT("::");

    return strStatus;
}

MACRO_SerializeConf(MorphOpenClose)
MACRO_ReadDataBlock(MorphOpenClose)

