// -*- c -*- 
// (setq c-basic-offset 4)
// "clsBlurGauss.h"

#include "clsBlurGauss.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/EX/clsCheckBoxEX.h"
#include "../../clsFrameMain.h"
#include "../../wxWidgetEnum.h"

#include "../MYAPI/MsgBox.h"

//SSE2
#include <malloc.h>
#include <emmintrin.h>

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define _BLUR_DIRECTION_XY 0
#define _BLUR_DIRECTION_X  1
#define _BLUR_DIRECTION_Y  2
#define _BLUR_DIRECTION_Z  3
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

MACRO_FilterConstructor(BlurGauss,,)

void BlurGaussGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

#ifdef WIN32
    pCurr->mTextWidth = 60;
#else
    pCurr->mTextWidth = 70;
    pCurr->mComboWidth = 70;
#endif
    
    int boxWidth = 2;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Kernel"));
    
    wxString listKernel[2] = { _("Sphere"), _("Ellipsoid") };
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mpCParent->mCData.mCKernelType.mBody,
			  wxT(""), TRUE, 2, listKernel);
    mIndexKernelSize = pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCData.mCKernelSize.mBody);

    boxWidth = 3;
    mIndexAsymSize = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->CombinePointer
	(_GP_FLOAT, &mpCParent->mCData.mCKernelAssimSize.mpBody[X], wxT("X"));
    pCurr->CombinePointer
	(_GP_FLOAT, &mpCParent->mCData.mCKernelAssimSize.mpBody[Y], wxT("Y"));
    pCurr->CombinePointer
	(_GP_FLOAT, &mpCParent->mCData.mCKernelAssimSize.mpBody[Z], wxT("Z"));
    
    pCurr->mProportion = 1;
    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    pCurr->SetFuncHook(&FuncHookHideItem, _MN_HOOK_COMBOBOX_CHANGE);
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookChanged, _MN_HOOK_CHANGE);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);

    HideAndShowItems();
}

MACRO_SetCustomFunctions(BlurGaussGUI_cls,,,)

void BlurGaussGUI_cls::
HideAndShowItems(){
    if (mpCMNumPanel->mbPanelOpen == FALSE) return;

    int kernelType = mpCParent->mCData.mCKernelType.mBody;
    bool bAsym = (kernelType == _KERNEL_ELLIPSOID);
    mpCMNumPanel->ShowItems(mIndexAsymSize, 4, bAsym);
    mpCMNumPanel->mVCPointers[mIndexKernelSize].Enable(!bAsym);

    
    mpCParent->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
}

MACRO_FuncHookHideItem(BlurGauss);

// -----

ThreadBluringF::
ThreadBluringF(wxThreadKind kind) : MultiThreadThread_cls(kind) {}
    
void ThreadBluringF::
SetParameters(BlurGauss_cls *pCParent,
	      float *pTo, float *pFrom, unsigned int size[3],
	      int initZ, int finZ, int direction) {
    mpCParent = pCParent;
    mpTo = pTo;
    mpFrom = pFrom;
    LOOP_XYZ mSize[xyz] = size[xyz];
    mInitZ = initZ;
    mFinZ = finZ;
    mDirection = direction;
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

void ThreadBluringF::
BlurLineSSEIfInLoop(const int kernelSize, float *pKernel,
		    const int offsetBase, /*const int offsetScale, */
		    const int lineWidth,
		    float *pIn, float *pOut){
    __m128 simd_sum;
    __m128 simd_value;
    __m128 simd_tmp1;
    __m128 simd_tmp2; 
    const int lineWidthMod = lineWidth - kernelSize - (lineWidth - 2 * kernelSize) % 4;
    const float *pKernelEnd = pKernel + kernelSize * 2 + 1;
    
    int index    = offsetBase + kernelSize;
    int indexEnd = offsetBase + lineWidthMod;
    float *pOutIndex = pOut + index;
    float *pInIndex = pIn + index;
    float *pInIndexEnd = pIn + indexEnd;
    for (; pInIndex < pInIndexEnd; pOutIndex+=4, pInIndex+=4){
	simd_sum = _mm_setzero_ps();
	float *pInIndexOffset = pInIndex - kernelSize;
	for (float *pKernelPos = pKernel;
	     pKernelPos < pKernelEnd;
	     pKernelPos++, pInIndexOffset++){
	    simd_tmp1  = _mm_set1_ps( *pKernelPos );
	    simd_tmp2  = _mm_loadu_ps( pInIndexOffset );
	    simd_value = _mm_mul_ps(simd_tmp1, simd_tmp2);
	    simd_sum   = _mm_add_ps(simd_sum, simd_value);
	}
	_mm_storeu_ps(pOutIndex, simd_sum);
    }

    index    = offsetBase + 0;
    indexEnd = offsetBase + kernelSize;
    pInIndex      = pIn + index;
    float *pInIndexBegin = pIn + index;
    pInIndexEnd = pIn  + indexEnd;
    pOutIndex   = pOut + index;
    for (; pInIndex < pInIndexEnd; pOutIndex++, pInIndex++){
	float value = 0;
	float *pInIndexOffset = pInIndex - kernelSize;
	for (float *pKernelPos = pKernel; 
	     pKernelPos < pKernelEnd; 
	     pKernelPos++, pInIndexOffset++)
	    if (pInIndexOffset >= pInIndexBegin)
		value += (*pKernelPos) * (*pInIndexOffset);
	    else
		value += (*pKernelPos) * (*pInIndexBegin);
	
	*pOutIndex = value;
    }


    index    = offsetBase + lineWidthMod;
    indexEnd = offsetBase + lineWidth;
    pInIndex    = pIn  + index;
    pInIndexEnd = pIn  + indexEnd;
    pOutIndex   = pOut + index;
    for (; pInIndex < pInIndexEnd; pOutIndex++, pInIndex++){
	float value = 0;
	float *pInIndexOffset = pInIndex - kernelSize;
	for (float *pKernelPos = pKernel; 
	     pKernelPos < pKernelEnd; 
	     pKernelPos++, pInIndexOffset++)
	    value += (*pKernelPos) * (*pInIndexOffset);
	
	*pOutIndex = value;
    }
}

void ThreadBluringF::
BlurLineSSE(const int kernelSize, float *pKernel,
	    const int offsetBase, const int offsetScale,
	    const int lineWidth,
	    float *pIn, float *pOut){
    
    __m128 simd_sum;
    __m128 simd_value;
    __m128 simd_tmp1;
    __m128 simd_tmp2; 

    // 16oCgACȂΓǂݍ݂̂
    // œKȃAhX̃ItZbgvZċɗp
    long memoryIni = (long)pIn;
    long memoryOffsetIni16 = (16 - (memoryIni - 1)%16 - 1) ;
    long indexOffset = memoryOffsetIni16 / 4; 

    // offsetScale4̔{(16oCg)łȂƑʖ
    bool bUseAlign = TRUE;
    if (offsetScale % 4 != 0) {
	indexOffset = 0;
	bUseAlign = FALSE;
    }

    const int lkernelOffsetScale = kernelSize * offsetScale;

    // noCgc4ɑ΂]
    const int lineWidthMod = (lineWidth - indexOffset) - (lineWidth - indexOffset) % 4;
    const float *pKernelEnd = pKernel + kernelSize * 2 + 1;

    int index    = offsetBase + indexOffset;
    int indexEnd = offsetBase + lineWidthMod + indexOffset;
    float *pInIndex    = pIn + index;
    float *pInIndexEnd = pIn + indexEnd;
    float *pOutIndex   = pOut + index;
    if (bUseAlign == TRUE)
	for (; pInIndex < pInIndexEnd; pOutIndex+=4, pInIndex+=4){
	    simd_sum = _mm_setzero_ps();
	    float *pInIndexOffset = pInIndex - lkernelOffsetScale;
	    for (float *pKernelPos = pKernel; 
		 pKernelPos < pKernelEnd; 
		 pKernelPos++, pInIndexOffset += offsetScale){
		simd_tmp1  = _mm_set1_ps( *pKernelPos );
		simd_tmp2  = _mm_load_ps( pInIndexOffset);
		simd_value = _mm_mul_ps(simd_tmp1, simd_tmp2);
		simd_sum   = _mm_add_ps(simd_sum, simd_value);
	    }
	    _mm_storeu_ps(pOutIndex, simd_sum);
	}
    else
	for (; pInIndex < pInIndexEnd; pOutIndex+=4, pInIndex+=4){
	    simd_sum = _mm_setzero_ps();
	    float *pInIndexOffset = pInIndex - lkernelOffsetScale;
	    for (float *pKernelPos = pKernel; 
		 pKernelPos < pKernelEnd; 
		 pKernelPos++, pInIndexOffset += offsetScale){
		simd_tmp1  = _mm_set1_ps( *pKernelPos );
		simd_tmp2  = _mm_loadu_ps( pInIndexOffset);
		simd_value = _mm_mul_ps(simd_tmp1, simd_tmp2);
		simd_sum   = _mm_add_ps(simd_sum, simd_value);
	    }
	    _mm_storeu_ps(pOutIndex, simd_sum);
	}
    
    index    = offsetBase + 0;
    indexEnd = offsetBase + indexOffset;
    pInIndex    = pIn  + index;
    pInIndexEnd = pIn  + indexEnd;
    pOutIndex   = pOut + index;
    for (; pInIndex < pInIndexEnd; pOutIndex++, pInIndex++){
	float value = 0;
	float *pInIndexOffset = pInIndex - lkernelOffsetScale;
	for (float *pKernelPos = pKernel; 
	     pKernelPos < pKernelEnd; 
	     pKernelPos++, pInIndex += offsetScale)
	    value += (*pKernelPos) * (*pInIndexOffset);
	*pOutIndex = value;
    }

    index    = offsetBase + lineWidthMod + indexOffset;
    indexEnd = offsetBase + lineWidth;
    pInIndex    = pIn  + index;
    pInIndexEnd = pIn  + indexEnd;
    pOutIndex   = pOut + index;
    for (; pInIndex < pInIndexEnd; pOutIndex++, pInIndex++){
	float value = 0;
	float *pInIndexOffset = pInIndex - lkernelOffsetScale;
	for (float *pKernelPos = pKernel; 
	     pKernelPos < pKernelEnd; 
	     pKernelPos++, pInIndex += offsetScale)
	    value += (*pKernelPos) * (*pInIndexOffset);
	*pOutIndex = value;
    }
}

wxThread::ExitCode ThreadBluringF::
Entry(){
    int totalSize = mSize[X] * mSize[Y] * (mFinZ - mInitZ);

    std::vector < float > VKernel[3];
    int kSize[3];
    float kernelSize[3];

    if (mpCParent->mCData.mCKernelType.mBody == 0)
	LOOP_XYZ kernelSize[xyz] = mpCParent->mCData.mCKernelSize.mBody;
    else
	LOOP_XYZ kernelSize[xyz] = mpCParent->mCData.mCKernelAssimSize.mpBody[xyz];

    LOOP_XYZ{
	kSize[xyz] = (int)kernelSize[xyz] + 1;
	if (kernelSize[xyz] != 0){
	    float total = 0;
	    for (int k = -kSize[xyz]; k <= kSize[xyz]; k++){
		float value = exp(-(double)(k*k) / (2.0 * kernelSize[xyz]));
		VKernel[xyz].push_back(value);
		total += value;
	    }
	    for (int k = -kSize[xyz]; k <= kSize[xyz]; k++)
		VKernel[xyz][k + kSize[xyz]] /= total;
	}
	else{
	    VKernel[xyz].push_back(1);
	    kSize[xyz] = 0;
	}
    }

    if (mDirection == _BLUR_DIRECTION_XY){
	// pFrom->pTo;
	for (int z = mInitZ; z < mFinZ; z++){
	    int indexZ = z * mSize[X] * mSize[Y];
	    for (int y = 0; y < mSize[Y]; y++){
		int indexY = indexZ + y * mSize[X];
		BlurLineSSEIfInLoop(kSize[X], &VKernel[X][0],
				    indexY, mSize[X], mpFrom, mpTo);
	    }

	    // 1C[IiLbVɂ邤ɁjỸu[
	    // pTo->pFrom;
	    for (int y = 0; y < mSize[Y]; y++){
		int indexY = indexZ + y * mSize[X];
		if (y > kSize[Y] && y < mSize[Y] - kSize[Y])
		    BlurLineSSE(kSize[Y], &VKernel[Y][0],
				indexY, mSize[X], mSize[X], mpTo, mpFrom);
		else
		    memcpy(&mpFrom[indexY], &mpTo[indexY], sizeof(float) * mSize[X]);
	    }
	}
    }
    
    if (mDirection == _BLUR_DIRECTION_Z){
	// pFrom->pTo;
	for (int z = mInitZ; z < mFinZ; z++){
	    int indexZ = z * mSize[X] * mSize[Y];
	    if (z > kSize[Z] && z < mSize[Z] - kSize[Z])
		for (int y = 0; y < mSize[Y]; y++){
		    int indexY = indexZ + y * mSize[X];
		    BlurLineSSE(kSize[Z], &VKernel[Z][0],
				indexY, mSize[X] * mSize[Y], mSize[X], mpFrom, mpTo);
		}
	    else
		memcpy(&mpTo[indexZ], &mpFrom[indexZ], sizeof(float) * mSize[X] * mSize[Y]);
	}
    }
    
    return 0;
}

// -----

ThreadBluring::
ThreadBluring(wxThreadKind kind) : MultiThreadThread_cls(kind) {
}

void ThreadBluring::
SetParameters(BlurGauss_cls *pCParent,
	      unsigned char *pTo, unsigned char *pFrom, unsigned int size[3],
	      unsigned char *pTmp,
	      int initZ, int finZ, int direction){
    mpCParent = pCParent;
    mpTo = pTo;
    mpFrom = pFrom;
    LOOP_XYZ mSize[xyz] = size[xyz];
    mpTmp = pTmp;
    mInitZ = initZ;
    mFinZ = finZ;
    mDirection = direction;
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

// ͍gĂȂ
// œKmFĂȂ̃~XĂ邩H
void ThreadBluring::
BlurLineSSE(const int kernelSize, const float *pKernel,
	    const int offsetBase, const int offsetScale, const int lineWidth,
	    const unsigned char *pIn, unsigned char *pOut){
    
    __m128 simd_sum;
    __m128 simd_value;
    __m128 simd_tmp1;
    __m128 simd_tmp2; 
    
    const int lKernelIndexSize(kernelSize * 2);
    float tmp[4];

    unsigned char *pOutIndex = pOut;

    /*
    memcpy(&pOut[0], &pIn[0], sizeof(unsigned char)*kernelSize);
    */

    for (int x = kernelSize; x < lineWidth - kernelSize; x+= 4){
	const int index       = offsetBase + x;

	simd_sum = _mm_setzero_ps();
	int addr = index - kernelSize * offsetScale;
	//for (int k = -kernelSize; k <= kernelSize; k++){
	for (int k = 0; k <= lKernelIndexSize; k++){
	    simd_tmp2 = _mm_set_ps
		( pIn[addr], pIn[addr+1], pIn[addr+2], pIn[addr+3]);
	    simd_tmp1 = _mm_set1_ps( pKernel[k/* + kernelSize*/] );
	    simd_value = _mm_mul_ps(simd_tmp1, simd_tmp2);
	    simd_sum = _mm_add_ps(simd_sum, simd_value);
	    addr += offsetScale;
	}
	
	_mm_storeu_ps(tmp, simd_sum);
	//pOut[index  ] = simd_sum.m128_f32[0];
	//pOut[index+1] = simd_sum.m128_f32[1];
	//pOut[index+2] = simd_sum.m128_f32[2];
	//pOut[index+3] = simd_sum.m128_f32[3];
	
	pOutIndex[0] = tmp[0];
	pOutIndex[1] = tmp[1];
	pOutIndex[2] = tmp[2];
	pOutIndex[3] = tmp[3];
	/*
	  pOut[index  ] = tmp[0];
	  pOut[index+1] = tmp[1];
	  pOut[index+2] = tmp[2];
	  pOut[index+3] = tmp[3];
	*/
	pOut += 4;
    }

    /*
    memcpy(&pOut[lineWidth - kernelSize], &pIn[lineWidth - kernelSize],
	   sizeof(unsigned char)*kernelSize);
    */
    
}

void ThreadBluring::
BlurLineIfInLoop(const int kernelSize, float *pKernel,
		 const int offsetBase, const int offsetScale, const int lineWidth,
		 unsigned char *pIn, unsigned char *pOut){

    const int lkernelOffsetScale = kernelSize * offsetScale;

    unsigned char *pInIndex = pIn + offsetBase + kernelSize;
    unsigned char *pInIndexEnd = pInIndex + lineWidth - 2 * kernelSize;
    unsigned char *pOutIndex = pOut + offsetBase + kernelSize;
    float *pKernelIndexEnd = pKernel + 2 * kernelSize + 1;
    for (; pInIndex < pInIndexEnd; pInIndex++, pOutIndex++){
	float value = 0;
	unsigned char *pInIndexOffset = pInIndex - lkernelOffsetScale;
	for (float *pKernelIndex = pKernel;
	     pKernelIndex < pKernelIndexEnd; 
	     pKernelIndex++, pInIndexOffset += offsetScale)
	    value += (*pKernelIndex) * (*pInIndexOffset);
	*pOutIndex = value;
    }
    
    // Et߂͌떂Ă
    for (int x = 0; x < kernelSize; x++)
	pOut[offsetBase + x] = pIn[offsetBase + x];
    
    for (int x = lineWidth - kernelSize ; x < lineWidth; x++)
	pOut[offsetBase + x] = pIn[offsetBase + x];
}

void ThreadBluring::
BlurLine(const int kernelSize, float *pKernel,
	 const int offsetBase, const int offsetScale, const int lineWidth,
	 unsigned char *pIn, unsigned char *pOut){

    const int lkernelOffsetScale = kernelSize * offsetScale;

    unsigned char *pInIndex    = pIn + offsetBase;
    unsigned char *pInIndexEnd = pInIndex + lineWidth;
    unsigned char *pOutIndex   = pOut + offsetBase;
    float *pKernelIndexEnd = pKernel + 2 * kernelSize + 1;
    for (; pInIndex < pInIndexEnd; pInIndex++, pOutIndex++){
	float value = 0;
	unsigned char *pInIndexOffset = pInIndex - lkernelOffsetScale;
	for (float *pKernelIndex = pKernel;
	     pKernelIndex < pKernelIndexEnd; 
	     pKernelIndex++, pInIndexOffset += offsetScale)
	    value += (*pKernelIndex) * (*pInIndexOffset);
	*pOutIndex = value;
    }
}

wxThread::ExitCode ThreadBluring::
Entry(){
    const int totalSize = mSize[X] * mSize[Y] * (mFinZ - mInitZ);
    const int finZ = mFinZ;
    const int sizeX = mSize[X];
    const int sizeY = mSize[Y];
    const int sizeZ = mSize[Z];

    std::vector < float > VKernel[3];
    int kSize[3];
    float kernelSize[3];

    if (mpCParent->mCData.mCKernelType.mBody == 0)
	LOOP_XYZ kernelSize[xyz] = mpCParent->mCData.mCKernelSize.mBody;
    else
	LOOP_XYZ kernelSize[xyz] = mpCParent->mCData.mCKernelAssimSize.mpBody[xyz];

    LOOP_XYZ{
	kSize[xyz] = (int)kernelSize[xyz] + 1;
	if (kernelSize[xyz] != 0){
	    float total = 0;
	    for (int k = -kSize[xyz]; k <= kSize[xyz]; k++){
		float value = exp(-(double)(k*k) / (2.0 * kernelSize[xyz]));
		VKernel[xyz].push_back(value);
		total += value;
	    }
	    for (int k = -kSize[xyz]; k <= kSize[xyz]; k++)
		VKernel[xyz][k + kSize[xyz]] /= total;
	}
	else{
	    VKernel[xyz].push_back(1);
	    kSize[xyz] = 0;
	}
    }

    if (mDirection == _BLUR_DIRECTION_XY){
	// pFrom->pTo;
	for (int z = mInitZ; z < finZ; z++){
	    int indexZ = z * sizeX * sizeY;
	    for (int y = 0; y < sizeY; y++){
		int indexY = indexZ + y * sizeX;
		BlurLineIfInLoop(kSize[X], &VKernel[X][0],
				 indexY, 1, sizeX, mpFrom, mpTo);
	    }

	    // 1C[IiLbVɂ邤ɁjỸu[
	    // pTo->pTmp;
	    for (int y = 0; y < sizeY; y++){
		int indexY = indexZ + y * sizeX;
		if (y > kSize[Y] && y < sizeY - kSize[Y])
		    BlurLine(kSize[Y], &VKernel[Y][0],
			     indexY, sizeX, sizeX, mpTo, mpTmp);
		else
		    memcpy(&mpTmp[indexY], &mpTo[indexY], sizeof(unsigned char) * sizeX);
	    }
	}
    }
    
    if (mDirection == _BLUR_DIRECTION_Z){
	// pTmp->pTo;
	for (int z = mInitZ; z < finZ; z++){
	    int indexZ = z * sizeX * sizeY;
	    if (z > kSize[Z] && z < sizeZ - kSize[Z])
		for (int y = 0; y < sizeY; y++){
		    int indexY = indexZ + y * sizeX;
		    BlurLine(kSize[Z], &VKernel[Z][0],
			     indexY, sizeX * sizeY, sizeX,
			     mpTmp, mpTo);
		}
	    else
		memcpy(&mpTo[indexZ], &mpTmp[indexZ], sizeof(unsigned char) * sizeX * sizeY);
	}
	
    }
    return 0;
}

// -----

BlurGaussData_cls::
BlurGaussData_cls(){
    mCConfList.Add(&mCKernelType, wxT("GaussBlurKernelType="), L"0");
    mCConfList.Add(&mCKernelSize, wxT("GaussBlurKernelSize="), L"0");
    mCConfList.Add(&mCKernelAssimSize, wxT("GaussBlurAssymSize="), L"0 0 0");
}

void BlurGaussData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCKernelType, wxT("GaussBlurKernelType="));
    mCConfList.Add(&mCKernelSize, wxT("GaussBlurKernelSize="));
    mCConfList.Add(&mCKernelAssimSize, wxT("GaussBlurAssymSize="));
}

MACRO_MakeInstanceAndAddToList(BlurGauss)

// -----

#include "../MYAPI/MsgBox.h"

BlurGauss_cls::
BlurGauss_cls(FilterCtrl_cls *pCParentCtrl, BlurGaussData_cls *pCInitData) 
: FilterBase_cls(pCParentCtrl){
    if (pCInitData != NULL) mCData = *pCInitData;
    mbProcessData = TRUE;
    ConstructorCommon(pCParentCtrl, _("Gauss Blur"), _FILTER_BLURING);
    mpCGUIPanel = new BlurGaussGUI_cls(this);

    mbAllocated = FALSE;
    mAllocatedSize = 0;
}

BlurGauss_cls::
~BlurGauss_cls(){
    if (mspCDF3SeqGUI != NULL)				
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
    if (mbAllocated == TRUE)
	delete mpTmp;
}

MACRO_GetPData(BlurGauss);
MACRO_AddFilterItemPolymorph(BlurGauss);

// ----

bool BlurGauss_cls::
ProcessXYZDataF(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3]){
    int totalSize = size[X] * size[Y] * size[Z];
    
    if (mCData.mCbValid.mbBody == FALSE){
	memcpy(pTo, pFrom, totalSize);
	return FALSE;
    }

    float *pFromF;
    float *pToF;

    try{
	pFromF = new float [totalSize];
    }
    catch(std::bad_alloc e){
        MsgBox(mpCGUIPanel->mpCMNumPanel, e.what(),
	       "Memory Allocation for temporal buffer *pFrom fails. Ingnore Bluring");
        return FALSE;
    }

    try{
	pToF = new float [totalSize];
    }
    catch(std::bad_alloc e){
	MsgBox(mpCGUIPanel->mpCMNumPanel, e.what(), 
	       "Memory Allocation for temporal buffer *pTo fails. Ingnore Bluring");
	delete pFromF;
        return FALSE;
    }

    float *pEnd  = pFromF+totalSize;
    unsigned char *pCurrUC = pFrom;
    for (float *pCurrFloat = pFromF; pCurrFloat < pEnd; pCurrFloat++, pCurrUC++)
	*pCurrFloat = *pCurrUC;

    int threadNum = mspCPreferenceLink->GetThreadNumFilter();
    
    GbGLLockForMultithread = TRUE;

    ThreadBluringF *pThreadXY[8];
    ThreadBluringF *pThreadZ[8];
    
    for (int thread = 0; thread < threadNum; thread++){
	int ini = (thread * size[Z]) / threadNum;
	int fin = ((thread + 1) * size[Z]) / threadNum;
	pThreadXY[thread] = new ThreadBluringF(wxTHREAD_JOINABLE);
	pThreadXY[thread]->SetParameters(this, pToF, pFromF, size, 
					 ini, fin, _BLUR_DIRECTION_XY);
	pThreadZ[thread] = new ThreadBluringF(wxTHREAD_JOINABLE);
	pThreadZ[thread]->SetParameters(this, pToF, pFromF, size, 
					ini, fin, _BLUR_DIRECTION_Z);
    }

    for (int thread = 0; thread < threadNum; thread++) pThreadXY[thread]->Run();
    for (int thread = 0; thread < threadNum; thread++) pThreadXY[thread]->Wait();
    for (int thread = 0; thread < threadNum; thread++) pThreadZ[thread]->Run();
    for (int thread = 0; thread < threadNum; thread++) pThreadZ[thread]->Wait();

    for (int thread = 0; thread < threadNum; thread++){
	delete pThreadXY[thread];
	delete pThreadZ[thread];
    }

    GbGLLockForMultithread = FALSE;

    pEnd = pToF+totalSize;
    pCurrUC = pTo;
    for (float *pCurrFloat = pToF; pCurrFloat < pEnd; pCurrFloat++, pCurrUC++)
	*pCurrUC = *pCurrFloat;

    delete pFromF;
    delete pToF;

    return TRUE;
}

bool BlurGauss_cls::
ProcessXYZDataC(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3]){
    int totalSize = size[X] * size[Y] * size[Z];
    
    if (mCData.mCbValid.mbBody == FALSE){
	memcpy(pTo, pFrom, totalSize);
	return FALSE;
    }

    // mۂdeleteȂ̂ŏ͒fЉ΍    
    // ͖ʌǂc
    if (mbAllocated == TRUE && mAllocatedSize != totalSize){
	delete mpTmp;
	mbAllocated = FALSE;
    }
    
    if (mbAllocated == FALSE){
	mpTmp = new unsigned char [totalSize];
	mbAllocated = TRUE;
	mAllocatedSize = totalSize;
    }
    
    int threadNum = mspCPreferenceLink->GetThreadNumFilter();

    GbGLLockForMultithread = TRUE;

    ThreadBluring *pThreadXY[_MAX_THREAD_NUM];
    ThreadBluring *pThreadZ[_MAX_THREAD_NUM];
    
    for (int thread = 0; thread < threadNum; thread++){
	int ini = (thread * size[Z]) / threadNum;
	int fin = ((thread + 1) * size[Z]) / threadNum;
	pThreadXY[thread] = new ThreadBluring(wxTHREAD_JOINABLE);
	pThreadXY[thread]->SetParameters(this, pTo, pFrom, size, mpTmp,
					 ini, fin, _BLUR_DIRECTION_XY);
	pThreadZ[thread] = new ThreadBluring(wxTHREAD_JOINABLE);
	pThreadZ[thread]->SetParameters(this, pTo, pFrom, size, mpTmp,
					ini, fin, _BLUR_DIRECTION_Z);
    }
    
    for (int thread = 0; thread < threadNum; thread++) pThreadXY[thread]->Run();
    for (int thread = 0; thread < threadNum; thread++) pThreadXY[thread]->Wait();
    for (int thread = 0; thread < threadNum; thread++) pThreadZ[thread]->Run();
    for (int thread = 0; thread < threadNum; thread++) pThreadZ[thread]->Wait();
    
    for (int thread = 0; thread < threadNum; thread++){
	delete pThreadXY[thread];
	delete pThreadZ[thread];
    }

    GbGLLockForMultithread = FALSE;
    
    //delete pTmp;

    return TRUE;
}

void BlurGauss_cls::
ProcessXYZData(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3], int totalSize, bool bSSE){
    bool bSuccess = FALSE;
    if (bSSE == TRUE)
	bSuccess = ProcessXYZDataF(pTo, pFrom, size);
    else
	bSuccess = ProcessXYZDataC(pTo, pFrom, size);
    if (bSuccess == FALSE)
	memcpy(pTo, pFrom, totalSize);
}

// ----

wxString BlurGauss_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("BlurGauss::") << mCData.mCbValid.mbBody;
    
    if (mCData.mCbValid.mbBody == FALSE)
	strStatus<< wxT("::");
    else
	strStatus << wxT(":") <<
	    mCData.mCKernelType.mBody << wxT(":") <<
	    mCData.mCKernelSize.mBody << wxT(":") <<
	    mCData.mCKernelAssimSize.mpBody[X] << wxT(":") <<
	    mCData.mCKernelAssimSize.mpBody[Y] << wxT(":") <<
	    mCData.mCKernelAssimSize.mpBody[Z] << wxT(":") << wxT("::");
    
    return strStatus;
}

MACRO_SerializeConf(BlurGauss)
MACRO_ReadDataBlock(BlurGauss)
