
#include "clsFilterWithTexture.h"

#include "../MYGL/clsTextureProperty.h"
#include "../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../FILTER/clsFilterCtrl.h"
#include "../POVRAY/clsPovObject.h"
#include "../POVRAY/clsPovIsosurface.h"
#include "../POVRAY/clsPovCrossSection.h"
#include "../POVRAY/clsPovThreeView.h"
#include "../POVRAY/VELOCITY/clsPovStreamLine.h"
#include "../POVRAY/VELOCITY/clsPovVelocityField.h"

#include "../WIDGETCTRL/clsGraphPanel.h"

FilterWithTextureGUI_cls::
FilterWithTextureGUI_cls(FilterBase_cls *pCParent) :
    FilterGUI_cls(pCParent)
{
    mpCPanelColorBar = NULL;
    pCParent->mbWithTexture = TRUE;
}

void FilterWithTextureGUI_cls::
UpdateColorList(){
    wxString pColorList[_CLR_CUSTOM + _CLR_CUSTOM_NUM_MAX];

    pColorList[0] = wxT("Mono Color");
    for (int i = 0; i < (int)GCColoring.mpCCustomColor->mCItemNum.mBody; i++)
	pColorList[1 + i] << wxT("Color ") << i;
    mpCMNumPanel->mVCPointers[mIndexColorType].CombinePointer
	(_GP_UNSIGNED_INT_LIST, NULL, wxT(""), TRUE, 1 + GCColoring.mpCCustomColor->mCItemNum.mBody, pColorList);
}

void FilterWithTextureGUI_cls::
InitializeColorPanel(wxWindow *pCParentWindow, void *pClue, 
		     double (*pFuncGetR)(double, void*), 
		     double (*pFuncGetG)(double, void*), 
		     double (*pFuncGetB)(double, void*),
		     wxString strLabel){
    mpCPanelColorBar = new PanelColorBar_cls(pCParentWindow, wxID_ANY, wxDefaultPosition, wxSize(-1, 20));
    mpCPanelColorBar->mbDisplayCurrent = FALSE;
    mpCPanelColorBar->mBorderSizeOffset[0] = 1;
    mpCPanelColorBar->SetFunction(pFuncGetR);
    mpCPanelColorBar->SetFunction(pFuncGetG);
    mpCPanelColorBar->SetFunction(pFuncGetB);
    mpCPanelColorBar->mOffset[0] = 4;
    mpCPanelColorBar->mpClue = pClue;
    mpCPanelColorBar->mstrLabel = strLabel;
}

void FilterWithTextureGUI_cls::
RefreshColorBar(){
    if (mpCPanelColorBar != NULL) mpCPanelColorBar->Refresh();
}

TexProperty_cls* FilterWithTextureGUI_cls::
GetPCTexProperty(){
    TexProperty_cls* pCTexProperty;
    switch(mpCParentBase->mFilterType){
    case _FILTER_POV_OBJECT:
	pCTexProperty = &((PovObject_cls*)mpCParentBase)->
	    GetPCData()->mCTexProperty;
	break;
    case _FILTER_ISOSURFACE:
	pCTexProperty = &((PovIsosurface_cls*)mpCParentBase)->
	    GetPCData()->mCTexProperty;
	break;
    case _FILTER_CROSS_SECTION:
	pCTexProperty = &((PovCrossSection_cls*)mpCParentBase)->
	    GetPCData()->mCTexProperty;
	break;
    case _FILTER_THREEVIEW:
	pCTexProperty = &((PovThreeView_cls*)mpCParentBase)->
	    GetPCData()->mCTexProperty;
	break;
    case _FILTER_STREAMLINE:
	pCTexProperty = &((PovStreamLine_cls*)mpCParentBase)->
	    GetPCData()->mCTexProperty;
	break;
    case _FILTER_VELOCITY_FIELD:
	pCTexProperty = &((PovVelField_cls*)mpCParentBase)->
	    GetPCData()->mCTexProperty;
	break;
    }
    return pCTexProperty;
}

double FilterWithTextureGUI_cls::
GetColor(double value, int rgbIndex){
    TexProperty_cls* pCTexProperty = GetPCTexProperty();
    
    int coloring = pCTexProperty->mCColorType.mBody;
    if (coloring == 0)
	return pCTexProperty->mCColor.mpBody[rgbIndex];
    if (coloring >= 1)
	return GCColoring.mpCCustomColor->GetCustomColor(value, rgbIndex, TRUE, coloring - 1);
    
    float pRGB[3];
    GCColoring.GetColor(pRGB, value, coloring, TRUE);
    return pRGB[rgbIndex];
}

double FilterWithTextureGUI_cls::
GetColorR(double value, void *pClue){
    return ((FilterWithTextureGUI_cls*)pClue/*=this*/)->GetColor(value, 0);
}

double FilterWithTextureGUI_cls::
GetColorG(double value, void *pClue){
    return ((FilterWithTextureGUI_cls*)pClue/*=this*/)->GetColor(value, 1);
}

double FilterWithTextureGUI_cls::
GetColorB(double value, void *pClue){
    return ((FilterWithTextureGUI_cls*)pClue/*=this*/)->GetColor(value, 2);
}

void FilterWithTextureGUI_cls::
LinkTexturePropertyCtrl(TexPropertyCtrl_cls *pTexpropertyCtrl){
    mspCLinkTexpropertyCtrl = pTexpropertyCtrl;
}

void FilterWithTextureGUI_cls::
FuncHookUpdateTextureDLG(void *pArg, void *pEvent){
    FilterWithTextureGUI_cls *pCThis = ((FilterWithTextureGUI_cls*)DF3SeqCtrl_cls::GetPFilterGUI(pArg));
    pCThis->mpCPanelColorBar->Refresh(FALSE);
    mspCLinkTexpropertyCtrl->UpdateColorType();
    mspCLinkTexpropertyCtrl->mpCGUI->UpdateColorList();
}

void FilterWithTextureGUI_cls::
FuncHookOpenTextureDLG(void *pArg, void *pEvent){
    FilterWithTextureGUI_cls *pCThis = ((FilterWithTextureGUI_cls*)DF3SeqCtrl_cls::GetPFilterGUI(pArg));
    pCThis->OpenTextureDLG(TRUE);
}

void FilterWithTextureGUI_cls::
OpenTextureDLG(bool bForceOpen){
    mspCLinkTexpropertyCtrl->ConnectProperty
	(GetPCTexProperty(), mpCParentBase->GetPData()->mCstrCustomName.msBody/*mstrLabel*/, mpCMNumPanel);
    mspCLinkTexpropertyCtrl->UpdateColorList();
    if (bForceOpen == TRUE)
	mspCLinkTexpropertyCtrl->OpenWindow((wxWindow*)mpCParentBase->mpCParentCtrl->mpCFilterPanelsLink);
}


TexPropertyCtrl_cls* FilterWithTextureGUI_cls::mspCLinkTexpropertyCtrl = NULL;
