// -*- c -*- 
// (setq c-basic-offset 4)
// "clsBoxCopy.h"

#include "clsLayerColorCurve.h"

#include "../COMMON/MyVector.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../GLCANVAS/clsMainGLCanvas.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/EX/clsCheckBoxEX.h"
#include "../../clsFrameMain.h"
#include "../../wxWidgetEnum.h"

#include "../MYAPI/MsgBox.h"

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

ThreadMakeLayer::
ThreadMakeLayer(unsigned short *pTmpBuffer, int type, bool bDither,
		unsigned int size[3], int z, int indexZ, double distance2,
		double *pInPoint, double *pRelativeVector,
		int inValue, int outValue)
: wxThread(wxTHREAD_JOINABLE),
    mpTmpBuffer(pTmpBuffer), mType(type), mbDither(bDither),
    mIndexZ(indexZ), mZ(z), mDistance2(distance2),
    mpInPoint(pInPoint), mpRelativeVector(pRelativeVector),
    mInValue(inValue), mOutValue(outValue)
{
    LOOP_XYZ mSize[xyz] = size[xyz];
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadMakeLayer::
Entry(){
    double CurrentPointFromIn[3];
    CurrentPointFromIn[Z] = (double)mZ/mSize[Z] - mpInPoint[Z];

    double valueDiff = mOutValue - mInValue;
    for (int y = 0; y < mSize[Y]; y++){
	int indexY = mIndexZ + y * mSize[X];
	CurrentPointFromIn[Y] = (double)y/mSize[Y] - mpInPoint[Y];
	int index = indexY;
	for (int x = 0; x < mSize[X]; x++, index++){
	    CurrentPointFromIn[X] = (double)x/mSize[X] - mpInPoint[X];
	    
	    double coef = MyVector::GetScalarProduct
		(mpRelativeVector, CurrentPointFromIn) / mDistance2;
	    if (coef < 0) coef = 0;
	    if (coef > 1) coef = 1;
	    mpTmpBuffer[index] = mInValue + valueDiff * coef;
	}
    }

    if (mbDither == FALSE) return 0;

    // 덷gU
    short preDiff = 0;
    for (int y = 0; y < mSize[Y]; y++){
	int indexY = mIndexZ + y * mSize[X];
	int index = indexY;
	for (int x = 0; x < mSize[X]; x++, index++){
	    mpTmpBuffer[index] += preDiff;

	    unsigned char valueFloor = mpTmpBuffer[index] / 32;
	    preDiff = mpTmpBuffer[index] - valueFloor * 32;
	    //if (preDiff > 16) preDiff -= 32;
	}
    }
    
    return 0;
}

// ---------------------

MACRO_FilterConstructor(LayerColorCurve,,)

void LayerColorCurveGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);
    
#ifdef WIN32
    pCurr->mTextWidth = 60;
#else
    pCurr->mTextWidth = 80;
#endif
    
    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  wxT("Composite"));

    pCurr->mComboWidth = 48;
    wxString listAxis[4] = { _("Add"), _("Substract"), _("Multiply"), _("Overwrite")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &((LayerColorCurveData_cls*)mpCParent->mpCData)->mCCompositeType.mBody, wxT(""), TRUE, 4, listAxis, 0);
    pCurr->CombinePointer(_GP_BOOL, &((LayerColorCurveData_cls*)mpCParent->mpCData)->mCCompositeDither.mbBody, wxT("Dither"));
    
    //int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  wxT("In/Out"));

    pCurr->CombinePointer(_GP_FLOAT, &((LayerColorCurveData_cls*)mpCParent->mpCData)->mCDensityIn.mBody);
    pCurr->CombinePointer(_GP_FLOAT, &((LayerColorCurveData_cls*)mpCParent->mpCData)->mCDensityOut.mBody);
    
    pCurr->mProportion = 1;
    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    //mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;

    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookRetexture, _MN_HOOK_COMBOBOX_CHANGE);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);
}

MACRO_SetCustomFunctions(LayerColorCurveGUI_cls,,,)
MACRO_HideAndShowEmpty(LayerColorCurveGUI_cls);
MACRO_FuncHookHideItem(LayerColorCurve);

// -----

LayerColorCurveData_cls::
LayerColorCurveData_cls(FilterCtrl_cls* pCCtrl,
			bool bConnectToStaticParam) : 
FilterWithParameterData_cls(pCCtrl)
{
    // ʂ̃Xgɒǉ
    mCConfList.Add(&mCCompositeType,   wxT("LayerColorCurveType="), L"0");
    mCConfList.Add(&mCCompositeDither, wxT("LayerColorCurveDither="), L"1");
    mCConfList.Add(&mCDensityIn, wxT("LayerColorCurveDensityIns="), L"0");
    mCConfList.Add(&mCDensityOut, wxT("LayerColorCurveDensityOut="), L"0.1");

    Parameters_cls CParam(6, L"DensityCurvePoints");
    double pointsPos[6] = {0, 0, 0, 1, 1, 0};
    CParam.ExecCommandInitSet(pointsPos);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void LayerColorCurveData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCCompositeType, wxT("LayerColorCurveType="));
    mCConfList.Add(&mCCompositeDither, wxT("LayerColorCurveDither="));
    mCConfList.Add(&mCDensityIn, wxT("LayerColorCurveDensityIn="));
    mCConfList.Add(&mCDensityOut, wxT("LayerColorCurveDensityOut="));
}

wxString LayerColorCurveData_cls::
SerializeConfPolymorph(){
    return msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();
}

MACRO_MakeInstanceAndAddToList(LayerColorCurve)

void LayerColorCurveData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void LayerColorCurveData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* LayerColorCurveData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > LayerColorCurveData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > LayerColorCurveData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// -----

#include "../MYAPI/MsgBox.h"

LayerColorCurve_cls::
LayerColorCurve_cls(FilterCtrl_cls *pCParentCtrl, LayerColorCurveData_cls *pCInitData) :
FilterWithParameter_cls(pCParentCtrl)
{
    mpCData = new LayerColorCurveData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) *(LayerColorCurveData_cls*)mpCData = *pCInitData;
    mbProcessData = TRUE;
    
    ConstructorCommon(pCParentCtrl, _("Density Curve"), _FILTER_LAYER_COLOR_CURVE);
    mpCGUIPanel = new LayerColorCurveGUI_cls(this);
}

LayerColorCurve_cls::
~LayerColorCurve_cls(){
    if (mspCDF3SeqGUI != NULL)
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
    if (MainGLCanvas_cls::mspCSelectedItem == mpCData->GetPParamActive()) 
	MainGLCanvas_cls::SelectItem(NULL);
}

MACRO_AddFilterItemPolymorphP(LayerColorCurve,);
MACRO_InitializeStaticMemberP(LayerColorCurve);
MACRO_GetPCData(LayerColorCurve);

// ----

bool LayerColorCurve_cls::
ProcessXYZData(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3]){
    int totalSize = size[X] * size[Y] * size[Z];
    
    if (mpCData->mCbValid.mbBody == FALSE){
	memcpy(pTo, pFrom, totalSize);
	return FALSE;
    }

    int type = ((LayerColorCurveData_cls*)mpCData)->mCCompositeType.mBody;
    bool bDither = ((LayerColorCurveData_cls*)mpCData)->mCCompositeDither.mbBody;
    
    // [vifȂ߂Ƀĝ
    // H ANZX̑H

    /* tmpBufferditherɂĝŁAx32{ƂĂŎg*/
    const int unit = 32;
    unsigned short *tmpBuffer = new unsigned short [totalSize]; // {Ƃc
    
    float inValue = ((LayerColorCurveData_cls*)mpCData)->mCDensityIn.mBody * 255 * unit;
    float outValue = ((LayerColorCurveData_cls*)mpCData)->mCDensityOut.mBody * 255 * unit;
    double *InPoint = mpCData->GetPParamActive()->mpInterpolatedValue;
    double *OutPoint = mpCData->GetPParamActive()->mpInterpolatedValue+3;

    double relativeVector[3];
    LOOP_XYZ relativeVector[xyz] = OutPoint[xyz] - InPoint[xyz];

    double distance = MyVector::GetNormRelative(InPoint, OutPoint);
    double distance2 = distance * distance;

    GbGLLockForMultithread = TRUE;

    ThreadMakeLayer *pCThread[8];
    int threadNum = mspCPreferenceLink->mCThreadNumFilter.mBody;

    int threadIndex = 0;
    for (int z = 0; z < size[Z]; z++){
	int indexZ = z * size[X] * size[Y];

	pCThread[threadIndex] = new ThreadMakeLayer
	    (tmpBuffer, type, bDither, size, z, indexZ, distance2,
	     InPoint, relativeVector, inValue, outValue);
	threadIndex ++;

	if (threadIndex >= threadNum ||
	    z == size[Z] - 1){
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Run();
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Wait();
	    for (int i = 0; i < threadIndex; i++) delete pCThread[i];
	    threadIndex = 0;
	}
    }

    switch(type){
    case 0: // Add
	for (int index = 0; index < totalSize; index++){
	    int value = pFrom[index] + tmpBuffer[index]/unit;
	    if (value > 255) value = 255;
	    pTo[index] = value;
	}
	break;
    case 1: // Substract
	for (int index = 0; index < totalSize; index++){
	    int value = pFrom[index] - tmpBuffer[index]/unit;
	    if (value < 0) value = 0;
	    pTo[index] = value;
	}
	break;
    case 2: // Multiply
	for (int index = 0; index < totalSize; index++){
	    int value = pFrom[index] * ((double)tmpBuffer[index]/255/unit);
	    if (value > 255) value = 255;
	    pTo[index] = value;
	}
	break;
    case 3: // Overwrite
	for (int index = 0; index < totalSize; index++){
	    int value = tmpBuffer[index]/unit;
	    if (value > 255) value = 255;
	    pTo[index] = value;
	}
	break;
    }
    
    delete tmpBuffer;

    GbGLLockForMultithread = FALSE;

    return TRUE;
}

// ----

wxString LayerColorCurve_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("ColorCurve::") <<	GetPCData()->mCbValid.mbBody;

    if (GetPCData()->mCbValid.mbBody == FALSE)
	strStatus<< wxT("::");
    else
	strStatus << wxT(":") <<
	    GetPCData()->GetPInterpolated()[0] << wxT(":") <<
	    GetPCData()->GetPInterpolated()[1] << wxT(":") <<
	    GetPCData()->GetPInterpolated()[2] << wxT(":") <<
	    GetPCData()->GetPInterpolated()[3] << wxT(":") <<
	    GetPCData()->GetPInterpolated()[4] << wxT(":") <<
	    GetPCData()->GetPInterpolated()[5] << wxT("::");
    
    return strStatus;
}


MACRO_SerializeConfP(LayerColorCurve)
MACRO_GetNewPReadDataBlock(LayerColorCurve)
