// -*- c -*-
// "clsDolly.h"

#include "clsDolly.h"
#include "clsMouseCtrl.h"
#include "clsMainGLCanvas.h"
#include "../MYGL/clsFBO.h"
#include "../MYGL/clsGLGuides.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../SCENEOBJ/clsSceneObj.h"

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define X 0
#define Y 1
#define Z 2

Dolly_cls::
Dolly_cls(Preference_cls *pCPreferenceLink){
    mpCPreferenceLink = pCPreferenceLink;
}

void Dolly_cls::
GetGLMatrixInfo(){
    glGetDoublev(GL_MODELVIEW_MATRIX,  mMatrixModelView);
    glGetDoublev(GL_PROJECTION_MATRIX, mMatrixProjection);
    glGetIntegerv(GL_VIEWPORT, mMatrixViewport);
}

void Dolly_cls::
SetParameter(double sensitivity, MouseNormalized_cls *pCMP, 
	     wxSize sz, int displayMode){
    mSensitivity = sensitivity;
    mpCMP = pCMP;
    mMouseX = mpCMP->mCurrX;
    mMouseY = mpCMP->mCurrY;
    mAspectRatio = (double)sz.x / sz.y;
    mDisplayMode = displayMode;
}

bool Dolly_cls::
Gyration(PositionAndDirection_cls *pCCamPos){
    PositionAndDirection_cls CamPosBK = *pCCamPos;

    if (mpCPreferenceLink->mCbGyrationInvert.mbBody == TRUE) 
	mpCMP->Invert();

    pCCamPos->SetSensitivity(mSensitivity);
    if (mDisplayMode == _DIRECTION_FREE ||
	mDisplayMode >= _DOLLY_LINKED)
	pCCamPos->TrackBall(mpCMP->mPrevX, mpCMP->mPrevY, 
			    mpCMP->mCurrX, mpCMP->mCurrY);
    else{
	mSensitivity = mSensitivity * pCCamPos->GetProjectedDistanceToAnc();
	pCCamPos->SetSensitivity(mSensitivity);
	pCCamPos->SetAspectRatio(mAspectRatio);
	pCCamPos->Translate(mpCMP->GetDX(), mpCMP->GetDY());
    }

    return TRUE;
}

bool Dolly_cls::
Zoom(PositionAndDirection_cls *pCCamPos, float *pOrthoScale){
    if (mpCPreferenceLink->mCbZoomInvert.mbBody == TRUE) 
	mpCMP->Invert();
    
    if (mDisplayMode >= _DIRECTION_X_PLUS &&
	mDisplayMode <= _DIRECTION_Z_MINUS){
	*pOrthoScale *= (1.0 + mpCMP->GetDY() - mpCMP->GetDX() * 0.2);
    }
    else {
	pCCamPos->SetSensitivity(1);
	pCCamPos->Zoom(mpCMP->GetDX(), mpCMP->GetDY());
    }

    return TRUE;
}

bool Dolly_cls::
MoveCenter(PositionAndDirection_cls *pCCamPos,
	   Position_cls *pCAncPos){
    PositionAndDirection_cls CAncTmp(*pCAncPos, *pCCamPos);
    
    mSensitivity = 0.5 * mSensitivity * pCCamPos->GetProjectedDistanceToAnc();
    
    pCCamPos->SetSensitivity(mSensitivity);
    pCCamPos->SetAspectRatio(mAspectRatio);
    
    if (/*mpCLinkedPath == NULL*/TRUE)
	pCCamPos->Translate(mpCMP->GetDX(), mpCMP->GetDY());
    else mSensitivity *= -1;
    
    CAncTmp.SetAncPos(pCCamPos->mPos);
    CAncTmp.SetSensitivity(mSensitivity);
    CAncTmp.SetAspectRatio(mAspectRatio);
    CAncTmp.Translate(mpCMP->GetDX(), mpCMP->GetDY());
    pCAncPos->Set(CAncTmp);
    
    //*trigAnchor = TRUE;
    return TRUE;
}

void Dolly_cls::
MoveItemTranslate(PositionAndDirection_cls *pCCamPos,
		  double aspectRatio, double dragX, double dragY,
		  int dragMode){
    Position_cls CPosTmp(msSelectedItemPos[0],
			 msSelectedItemPos[1],
			 msSelectedItemPos[2], 0, 0, 0);
    PositionAndDirection_cls CPos(CPosTmp, *pCCamPos);
    PositionAndDirection_cls CParentCoord;
    
    LOOP_XYZ CPos.mATranslateScale[xyz] = 1;
    
    //double sensitivity = -0.5 * pCCamPos->GetProjectedDistanceToAnc();
    double sensitivity = -0.3*pCCamPos->GetProjectedDistance(msSelectedItemPos);
    CPos.SetAncPos(pCCamPos->mPos);
    CPos.SetSensitivity(sensitivity);
    CPos.SetAspectRatio(aspectRatio);
    CPos.Translate(dragX, dragY, 
		   _POS_AND_DIR_MOVE_FREE, &CParentCoord);
    LOOP_XYZ msSelectedItemPos[xyz] = CPos.mPos[xyz];
    
    if (dragMode == X){
	msSelectedItemPos[Y] = msSelectedItemPosOrg[Y];
	msSelectedItemPos[Z] = msSelectedItemPosOrg[Z];
    }
    if (dragMode == Y){
	msSelectedItemPos[Z] = msSelectedItemPosOrg[Z];
	msSelectedItemPos[X] = msSelectedItemPosOrg[X];
    }
    if (dragMode == Z){
	msSelectedItemPos[X] = msSelectedItemPosOrg[X];
	msSelectedItemPos[Y] = msSelectedItemPosOrg[Y];
    }
}

void Dolly_cls::
MoveItemRotate(double dragRot, int dragMode){
    if (dragMode >= 0)
	msSelectedItemDRot[dragMode] += dragRot * 75;
}

void Dolly_cls::
DrawMouseGuide(int moveMode, int dragMode, int panelWidth, int panelHeight,
	       int bufferWidth, int bufferHeight){
    if (moveMode != _DOLLY_ROTATE) return;
    if (dragMode == -1) return;

    switch(dragMode){
    case _ROTATE_MOVE_Y:
	glColor4d(0.2, 0.2, 1, 1); break;
    case _ROTATE_MOVE_P:
	glColor4d(0.2, 1, 0.2, 1); break;
    case _ROTATE_MOVE_R:
	glColor4d(1, 0.3, 0.3, 1); break;
    }
    
    glEnable(GL_LINE_STIPPLE);
    glLineStipple(1,0xf0f0);
    FBO_cls::Push2DMatrix(bufferWidth, bufferHeight);
    int orgX = mPosItemOrgOnW[X] / panelWidth * bufferWidth;
    int orgY = mPosItemOrgOnW[Y] / panelHeight * bufferHeight;
    int curX = (mMouseX/2+0.5) * bufferWidth;
    int curY = (mMouseY/2/mAspectRatio + 0.5) * bufferHeight;
    glBegin(GL_LINES); 
    glVertex2d(orgX, orgY);
    glVertex2d(curX, curY);
    glEnd();
    FBO_cls::Pop2DMatrix();
    glDisable(GL_LINE_STIPPLE);
}

bool Dolly_cls::
MoveItem(PositionAndDirection_cls *pCCamPos, bool bDragging, wxSize sz,
	 int dragMode, int moveMode, bool bInv){
    if (mspCSelectedObject == NULL) return FALSE;

    if (bDragging == FALSE){
	LOOP_XYZ msSelectedItemPos[xyz] = 
	    mspCSelectedObject->mpCTranslate->mVCParam[xyz].GetInterpolatedValue();
	LOOP_XYZ msSelectedItemPosOrg[xyz] = msSelectedItemPos[xyz];
	LOOP_XYZ msSelectedItemScale[xyz] = 1.0;
	LOOP_XYZ msSelectedItemDRot[xyz] = 0;
    }

    gluProject(msSelectedItemPos[X], 
	       msSelectedItemPos[Y], 
	       msSelectedItemPos[Z],
	       mMatrixModelView, mMatrixProjection, mMatrixViewport,
	       &mPosItemOnW[0], &mPosItemOnW[1], &mPosItemOnW[2]);
    gluProject(msSelectedItemPosOrg[X], 
	       msSelectedItemPosOrg[Y], 
	       msSelectedItemPosOrg[Z],
	       mMatrixModelView, mMatrixProjection, mMatrixViewport,
	       &mPosItemOrgOnW[0], &mPosItemOrgOnW[1], &mPosItemOrgOnW[2]);
    double normalPosX = (2*mPosItemOnW[X] - sz.x)/sz.x;
    double normalPosY = (2*mPosItemOnW[Y] - sz.y)/sz.y;
    double dragX = mpCMP->mCurrX - mpCMP->mPrevX;
    double dragY = mpCMP->mCurrY - mpCMP->mPrevY;
    double relX  = mpCMP->mCurrX - normalPosX;
    double relY  = mpCMP->mCurrY - normalPosY;
    //double aspectRatio = (double)sz.x / sz.y;

    switch (moveMode){
    case _DOLLY_TRANSLATE: 
	MoveItemTranslate(pCCamPos, mAspectRatio, dragX, dragY, dragMode);
	break; 
    case _DOLLY_ROTATE: {
	double dragRot = (dragX * relY - dragY * relX) / /*sqrt*/ (relX * relX + relY * relY);
	if (bInv == TRUE) dragRot *= -1;
	MoveItemRotate(dragRot, dragMode);
	break; }
    }

    return TRUE;
}

bool Dolly_cls::
ZoomWheel(PositionAndDirection_cls *pCCamPos,
	  double wheelRotation, float *pOrthoScale){
    if (mpCPreferenceLink->mCbZoomInvert.mbBody == TRUE) 
	wheelRotation *= -1;

    if (mDisplayMode >= _DIRECTION_X_PLUS &&
	mDisplayMode <= _DIRECTION_Z_MINUS){
	*pOrthoScale *= (1.0 + wheelRotation * 0.1);
    }
    else{
	pCCamPos->SetSensitivity(1);
	
	pCCamPos->Zoom(0, wheelRotation * 0.1);
    }
    return TRUE;
}

void Dolly_cls::
FinishDragging(){
    LOOP_XYZ mspCSelectedObject->mpCTranslate->mVCParam[xyz].mInterpolatedValue 
	= msSelectedItemPos[xyz];
    LOOP_XYZ mspCSelectedObject->mpCRotation->mVCParam[xyz].mInterpolatedValue 
	+= msSelectedItemDRot[xyz];
    LOOP_XYZ msSelectedItemDRot[xyz] = 0;
}

SceneObj_cls *Dolly_cls::mspCSelectedObject = NULL;
double Dolly_cls::msSelectedItemPos[3] = {0, 0 ,0};
double Dolly_cls::msSelectedItemPosOrg[3] = {0, 0 ,0};
double Dolly_cls::msSelectedItemScale[3] = {1, 1 ,1};
double Dolly_cls::msSelectedItemDRot[3] = {0, 0, 0};
