
#include "clsGLCanvasCtrl.h"

#include <wx/tglbtn.h>
#include <wx/valgen.h>

#include "clsMainGLCanvas.h"
#include "../MYGL/clsFBO.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

BEGIN_EVENT_TABLE(GLCanvasCtrl_cls, wxPanel)
     EVT_PAINT       (GLCanvasCtrl_cls::OnPaint)
     EVT_KEY_DOWN    (GLCanvasCtrl_cls::OnKeyDown)
     EVT_RIGHT_DOWN  (GLCanvasCtrl_cls::OnMouse)
     EVT_MENU        (GLMenuHidePanel, GLCanvasCtrl_cls::OnMenuHide)
END_EVENT_TABLE()

#define _POV_PERSPECTIVE  0
#define _POV_ORTHOGRAPHIC 1
#define _POV_FISHEYE      2
#define _POV_PANORAMIC    3

#define MACRO_SetButtons(POINTER, CHAR, FUNC, TIPS)			\
    POINTER = new wxToggleButton					\
	(this, wxID_ANY, CHAR, defPos, wxSize(24,-1));			\
    Connect(POINTER->GetId(), wxEVT_COMMAND_TOGGLEBUTTON_CLICKED,	\
	    wxCommandEventHandler(GLCanvasCtrl_cls::FUNC));		\
    POINTER->SetToolTip(TIPS);

GLCanvasCtrl_cls::
GLCanvasCtrl_cls(wxWindow *parent, FrameMain_cls *pCParentFrame,
		 long style, const wxString& name,
		 const wxGLContext *pContextToShare,
		 unsigned int selfIndex, int displayMode) :
wxPanel(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, style, name)
{
    mSelfIndex = selfIndex;
    mbPanelShown = TRUE;

    int attribList[5] = {WX_GL_DOUBLEBUFFER, 1, WX_GL_STENCIL_SIZE, 1, 0};
    mpCGLCanvas = new MainGLCanvas_cls
	(this, wxID_ANY, attribList, wxDefaultPosition, wxDefaultSize,
	 style, name, pContextToShare, selfIndex, displayMode);
    mpCGLCanvas->SetParentClass(pCParentFrame);
					 
    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    mpBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);
    
    wxString choice[_GL_CANVAS_LINKED_CAM0 + 1] = { 
	wxT("Free"), 
	wxT("From -X to +X"),	wxT("From +X to -X"),
	wxT("From -Y to +Y"),	wxT("From +Y to -Y"),
	wxT("From -Z to +Z"),	wxT("From +Z to -Z")
    };

    for (int i = 0; i < _GL_CANVAS_LINKED_CAM0; i++) mAChoiceStr[i] = choice[i];
    for (int i = 0; i < _PATH_NUM_MAX; i++) 
	mAChoiceStr[_GL_CANVAS_LINKED_CAM0 + i] << wxT("Cam") << i;
    mstrChoiceCopyOfView0 = wxT("Copy of View 0");
    
    mpCComboDirection = new wxComboBox(this, wxID_ANY, wxT("Free"), wxDefaultPosition, wxSize(108, -1), 8, mAChoiceStr, wxCB_DROPDOWN | wxCB_READONLY);
   
    this->Connect(mpCComboDirection->GetId(), wxEVT_COMMAND_COMBOBOX_SELECTED,
		  wxCommandEventHandler(GLCanvasCtrl_cls::OnComboBox));
    mpCComboDirection->SetToolTip(_("Point of View"));

    wxPoint defPos = wxDefaultPosition;

    mpCBoolFlip = new wxCheckBoxEX
	(this, wxID_ANY, wxT("Flip"), 
	 wxGenericValidator(&mpCGLCanvas->mCbStereoFlip.mbBody));
    Connect(mpCBoolFlip->GetId(), wxEVT_COMMAND_CHECKBOX_CLICKED,
	    wxCommandEventHandler(GLCanvasCtrl_cls::OnCheckBox));
    //mpCBoolFlip->SetToopTip(wxT("Flip in Stereo Gram"));

    mpCFBO = new wxCheckBoxEX
	(this, wxID_ANY, _("FBO(HDR)"), 
	 wxGenericValidator(&mpCGLCanvas->mbCFBO.mbBody));
    Connect(mpCFBO->GetId(), wxEVT_COMMAND_CHECKBOX_CLICKED,
	    wxCommandEventHandler(GLCanvasCtrl_cls::OnCheckBox));
    
    MACRO_SetButtons(mpCToggleItem, wxT("I"), OnToggleItem,
		     _("Show/Hide Items"));
    MACRO_SetButtons(mpCToggleWire, wxT("W"), OnToggleWire,
		     _("Show/Hide Data Box by Wire"));
    MACRO_SetButtons(mpCToggleBox, wxT("B"), OnToggleBox, 
		     _("Show/Hide Data Box"));
    MACRO_SetButtons(mpCToggleGrid, wxT("G"), OnToggleGrid,
		     _("Show/Hide Grid"));
    MACRO_SetButtons(mpCToggleData, wxT("D"), OnToggleData,
		     _("Show/Hide Data"));

    UpdateWindow();
    UpdateToggleButtonsAndParams();

    mpBoxSizerBar->Add(mpCComboDirection, 0, 0);
    mpBoxSizerBar->Add(-1, -1, 1, wxEXPAND);
    mpBoxSizerBar->Add(mpCFBO, 0, wxALIGN_CENTER_VERTICAL);
    mpBoxSizerBar->Add(mpCBoolFlip, 0, wxALIGN_CENTER_VERTICAL);
    mpBoxSizerBar->Add(mpCToggleItem, 0, 0);
    mpBoxSizerBar->Add(mpCToggleWire, 0, 0);
    mpBoxSizerBar->Add(mpCToggleBox, 0, 0);
    mpBoxSizerBar->Add(mpCToggleGrid, 0, 0);
    mpBoxSizerBar->Add(mpCToggleData, 0, 0);
    pBoxSizerSrc->Add(mpBoxSizerBar, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCGLCanvas, 1, wxEXPAND);

    SetSizer(pBoxSizerSrc);

    mpCPopupMenu = new wxMenu;
    mpCPopupMenu->Append(GLMenuHidePanel, _("Hide Control Panel"));
}

GLCanvasCtrl_cls::
~GLCanvasCtrl_cls(){
    delete mpCPopupMenu;
}

void GLCanvasCtrl_cls::
SetParentClass(FrameMain_cls* pCParentFrame){
    mpCGLCanvas->SetParentClass(pCParentFrame);
}

int GLCanvasCtrl_cls::
GetDisplayMode(){
    return mpCGLCanvas->mCDisplayMode.mBody;
}

bool GLCanvasCtrl_cls::
IsPanelShown(){
    return mbPanelShown;
}

void GLCanvasCtrl_cls::
ShowPanel(){
    size_t itemNum = mpBoxSizerBar->GetChildren().GetCount();
    for (size_t index = 0; index < itemNum; index++)
	mpBoxSizerBar->Show(index);
    mbPanelShown = TRUE;
    Layout();
}

void GLCanvasCtrl_cls::
UpdateWindow(){
    switch (mpCGLCanvas->mCDisplayMode.mBody){
    case _GL_CANVAS_DISPLAY_FREE:
	mpCComboDirection->SetValue(wxT("Free")); break;
    case _GL_CANVAS_DISPLAY_X_PLUS:
	mpCComboDirection->SetValue(wxT("From -X to +X")); break;
    case _GL_CANVAS_DISPLAY_X_MINUS:
	mpCComboDirection->SetValue(wxT("From +X to -X")); break;
    case _GL_CANVAS_DISPLAY_Y_PLUS:
	mpCComboDirection->SetValue(wxT("From -Y to +Y")); break;
    case _GL_CANVAS_DISPLAY_Y_MINUS:
	mpCComboDirection->SetValue(wxT("From +Y to -Y")); break;
    case _GL_CANVAS_DISPLAY_Z_PLUS:
	mpCComboDirection->SetValue(wxT("From -Z to +Z")); break;
    case _GL_CANVAS_DISPLAY_Z_MINUS:
	mpCComboDirection->SetValue(wxT("From +Z to -Z")); break;
    case _GL_CANVAS_COPY_OF_VIEW_0:
	mpCComboDirection->SetValue(mstrChoiceCopyOfView0); break;
    default: {
	for (int i = 0; i < _PATH_NUM_MAX; i++)
	    if (mpCGLCanvas->mCDisplayMode.mBody == _GL_CANVAS_LINKED_CAM0 + i){
		wxString strCam; strCam << wxT("Cam") << i;
		mpCComboDirection->SetValue(strCam);
	    }
	break; }
    }

    mpCFBO->GetValidator()->TransferToWindow();
    mpCBoolFlip->GetValidator()->TransferToWindow();
}

void GLCanvasCtrl_cls::
UpdateComboBoxList(int camNum){
    mpCComboDirection->Freeze();
    mpCComboDirection->Clear();
    for (int i = 0; i < _GL_CANVAS_LINKED_CAM0; i++) 
	mpCComboDirection->Append(mAChoiceStr[i]);
    for (int i = 0; i < camNum; i++) 
	mpCComboDirection->Append(mAChoiceStr[_GL_CANVAS_LINKED_CAM0 + i]);
    if (mSelfIndex == 4)
	mpCComboDirection->Append(mstrChoiceCopyOfView0);
    mpCComboDirection->Thaw();
    UpdateWindow();

    mpCFBO->Enable(FBO_cls::msbSupported);
}

void GLCanvasCtrl_cls::
OnPaint(wxPaintEvent& event){
    event.Skip();
    //SkipAԈړȂǂinvalidateŖ̂AppendB
    //@\邯ǖʂeventłHƖʂȂłH
    mpCGLCanvas->AddPendingEvent(event);
    mpCComboDirection->SetValue(mAChoiceStr[GetDisplayMode()]);
}

void GLCanvasCtrl_cls::OnSize(wxSizeEvent& event){
    event.Skip();
    //UpdateWindowSizeInChildClass();
}

#include "../MYAPI/MsgBox.h"

void GLCanvasCtrl_cls::
OnKeyDown(wxKeyEvent& event){
    event.SetId(wxGetTopLevelParent(this)->GetId());
    wxGetTopLevelParent(this)->AddPendingEvent(event);
}

void GLCanvasCtrl_cls::
OnBGColor(wxColourPickerEvent& event){
    float pColorFloat[3];
    pColorFloat[0] = (float)event.GetColour().Red() / 255;
    pColorFloat[1] = (float)event.GetColour().Green() / 255;
    pColorFloat[2] = (float)event.GetColour().Blue() / 255;
    mpCGLCanvas->ChangeBGColor(pColorFloat, 0);
}

void GLCanvasCtrl_cls::
OnComboBox(wxCommandEvent& event){
    if (mpCComboDirection->GetValue() == wxT("Free"))
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_FREE);
    if (mpCComboDirection->GetValue() == wxT("From -X to +X"))
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_X_PLUS);
    if (mpCComboDirection->GetValue() == wxT("From +X to -X"))
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_X_MINUS);
    if (mpCComboDirection->GetValue() == wxT("From -Y to +Y"))
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_Y_PLUS);
    if (mpCComboDirection->GetValue() == wxT("From +Y to -Y"))
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_Y_MINUS);
    if (mpCComboDirection->GetValue() == wxT("From -Z to +Z"))
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_Z_PLUS);
    if (mpCComboDirection->GetValue() == wxT("From +Z to -Z"))
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_Z_MINUS);
    if (mpCComboDirection->GetValue() == mstrChoiceCopyOfView0)
	mpCGLCanvas->SetDisplayMode(_GL_CANVAS_COPY_OF_VIEW_0);

    for (int i = 0; i < _PATH_NUM_MAX; i++){
	wxString strIndex; strIndex << wxT("Cam") << i;
	if (mpCComboDirection->GetValue() == strIndex){
	    mpCGLCanvas->LinkageConnect(i);
	    mpCGLCanvas->mCPathIndex.mBody = i;
	    mpCGLCanvas->Refresh(FALSE);
	    wxCommandEvent command = wxCommandEvent(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshPanels);
	    wxGetTopLevelParent(mpCGLCanvas)->AddPendingEvent(command);
	}
    }
}

void GLCanvasCtrl_cls::
OnCheckBox(wxCommandEvent& event){
    wxCommandEvent command = wxCommandEvent(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    wxGetTopLevelParent(mpCGLCanvas)->AddPendingEvent(command);
}

void GLCanvasCtrl_cls::
OnToggleItem(wxCommandEvent& event){
    mpCGLCanvas->mCbShowItems.mbBody = ! mpCToggleItem->GetValue();
    mpCGLCanvas->Refresh(FALSE);
}

void GLCanvasCtrl_cls::
OnToggleWire(wxCommandEvent& event){
    mpCGLCanvas->mCbShowWire.mbBody = ! mpCToggleWire->GetValue();
    mpCGLCanvas->Refresh(FALSE);
}

void GLCanvasCtrl_cls::
OnToggleBox(wxCommandEvent& event){
    mpCGLCanvas->mCbShowBox.mbBody = ! mpCToggleBox->GetValue();
    mpCGLCanvas->Refresh(FALSE);
}

void GLCanvasCtrl_cls::
OnToggleGrid(wxCommandEvent& event){
    mpCGLCanvas->mCbShowGrid.mbBody = ! mpCToggleGrid->GetValue();
    mpCGLCanvas->Refresh(FALSE);
}

void GLCanvasCtrl_cls::
OnToggleData(wxCommandEvent& event){
    mpCGLCanvas->mCbShowData.mbBody = ! mpCToggleData->GetValue();
    mpCGLCanvas->Refresh(FALSE);
}

void GLCanvasCtrl_cls::
OnMenuHide(wxCommandEvent& event){
    size_t itemNum = mpBoxSizerBar->GetChildren().GetCount();
    for (size_t index = 0; index < itemNum; index++)
	mpBoxSizerBar->Hide(index);
    Layout();
    mbPanelShown = FALSE;
}

void GLCanvasCtrl_cls::
OnMouse(wxMouseEvent& event){
    PopupMenu(mpCPopupMenu, wxDefaultPosition);
}

void GLCanvasCtrl_cls::
UpdateToggleButtonsAndParams(){
    mpCToggleItem->SetValue(!mpCGLCanvas->mCbShowItems.mbBody);
    mpCToggleWire->SetValue(!mpCGLCanvas->mCbShowWire.mbBody);
    mpCToggleBox ->SetValue(!mpCGLCanvas->mCbShowBox.mbBody);
    mpCToggleGrid->SetValue(!mpCGLCanvas->mCbShowGrid.mbBody);
    mpCToggleData->SetValue(!mpCGLCanvas->mCbShowData.mbBody);
    if (mpCGLCanvas->mCGLGeometry.mBody == _POV_ORTHOGRAPHIC)
	mpCGLCanvas->mCGLPerspective.mbPerspective = FALSE;
    else
	mpCGLCanvas->mCGLPerspective.mbPerspective = TRUE;
}

wxString GLCanvasCtrl_cls::
GetPovStrCamPos(){
    return mpCGLCanvas->GetPovStrCamPos();
}

wxString GLCanvasCtrl_cls::
GetPovStrCamMatrix(){
    return mpCGLCanvas->GetPovStrCamMatrix();
}

wxString GLCanvasCtrl_cls::
GetPovStrMask(int start, int end, int height, double theta){
    return mpCGLCanvas->GetPovStrMask(start, end, height, theta);
}

void GLCanvasCtrl_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    mpCGLCanvas->ReadStrLinePanel(istream, strLine);
}

wxString GLCanvasCtrl_cls::
SerializeConf(){
    return mpCGLCanvas->SerializePanel();
}
