// "clsMainGLCanvas.h"
// (setq c-basic-offset 4)

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
PFNGLTEXIMAGE3DPROC glTexImage3D;
PFNGLCOMPRESSEDTEXIMAGE3DPROC glTexCompressedImage3D;
PFNGLBLENDCOLORPROC glBlendColor;
PFNGLBLENDEQUATIONEXTPROC glBlendEquationEXT;
extern PFNGLBINDBUFFERPROC glBindBuffer; //!!
//PFNGLCOLORTABLEEXTPROC* glColorTableEXT
PFNGLCOLORTABLEEXTPROC glColorTableEXT;
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "clsMainGLCanvas.h"
#include "clsHitTest.h"

// dll$B$N%X%C%@(B
#include "../GLOBJ/GLOBJECTS.h"
// ---
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include "../COMMON/MyVector.h"
#include "../DF3/clsDF3SeqCtrl.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/GL/clsPreviewProperty.h"
#include "../DRAWBOX/clsBoxCopy.h"
#include "../GLCANVAS/clsGLCanvasCtrl.h"
#include "../GLOBJ/GLOBJLightSource.h"
#include "../GLOBJ/GLOBJTransCube.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../FILTER/clsClipping.h"
#include "../MYAPI/MsgBox.h"
#include "../MYGL/clsFBO.h"
#include "../MYGL/clsGLGuides.h"
#include "../MYGL/clsGLInfo.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../MAINCANVAS/clsDrawItem.h"
#include "../POVRAY/clsLightSource.h"
#include "../POVRAY/clsPovDiffObj.h"
#include "../POVRAY/clsPovRay.h"
#include "../POVRAY/clsPovVolume.h"
#include "../POVRAY/clsPovCrossSection.h"
#include "../POVRAY/clsPovObject.h"
#include "../POVRAY/clsStereoGram.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../../clsCameraPath.h"
#include "../../clsTimeCtrl.h"
#include <wx/tglbtn.h>

int displayListSphere;
int displayListCylinder;
int displayListCylinderOpen;
int displayListCone;
int displayListConeSimple;
int displayListTransCube[4];
int displayListWireBox[5];
int displayListCamera;

bool GbGLLockForMultithread = FALSE;
bool GbGLLockDrawing = FALSE;
bool GbGLForceWaitThreadEnd = FALSE;
bool GbGLLockForLoading = FALSE;

long GTestVariableLong = 0;
int GTestVariableInt = 0;

extern MainGLCanvas_cls* GpCurrentGLPanel;
extern MainGLCanvas_cls* GpGLPanelJustHandled;

#include "../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;

#include "../GRAPHIC/clsImageIO.h"
#include "../GRAPHIC/clsImgSequence.h"
extern ImageIO_cls GCImgIO;

#include "../../clsFrameMain.h"
#include "../../clsMain.h"
#include "../POVRAY/clsTreeBoxRule.h"

#include <algorithm>
#include <functional>
#include <sstream>

#include <math.h>

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

#define R 0
#define G 1
#define B 2
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)

#define LOOP_YPRY for (int ypry = 0; ypry < 4; ypry++)

#define _STENCIL_MODE_0 0
#define _STENCIL_MODE_1 1
#define _STENCIL_MODE_2 2

#define ENDL std::endl
#define ENDLWX wxT("\n")

#define _MAINGL_NO_MIRROR -1

#ifdef WIN32
#define _FLAG_SET_CURRENT
#endif
#ifdef __WXGTK__
#define _FLAG_SET_CURRENT
#endif

#define ENDL std::endl

// -----------------------------------------

BEGIN_EVENT_TABLE(MainGLCanvas_cls, wxGLCanvas)
     EVT_SIZE        (MainGLCanvas_cls::OnSize)
     EVT_PAINT       (MainGLCanvas_cls::OnPaint)
     EVT_MOUSE_EVENTS(MainGLCanvas_cls::OnMouse)
     EVT_MOUSE_CAPTURE_LOST(MainGLCanvas_cls::OnMouseRelease)
     EVT_MOTION(MainGLCanvas_cls::OnMouseMotion)
     EVT_MENU (GLMenuMoveToCamPos,   MainGLCanvas_cls::OnMenuMoveToCamPos)
     EVT_MENU (GLMenuSetKeyPos,      MainGLCanvas_cls::OnMenuSetKeyPos)
     EVT_MENU (GLMenuSetKeyAnc,      MainGLCanvas_cls::OnMenuSetKeyAnc)
     EVT_MENU (GLMenuToggleLink,     MainGLCanvas_cls::OnMenuToggleLink)
     EVT_MENU (menuItemAddLight,     MainGLCanvas_cls::OnMenuPassToParent)
     EVT_MENU (menuItemAddHotSpot,   MainGLCanvas_cls::OnMenuPassToParent)
     EVT_MENU (menuItemAddPrimitive, MainGLCanvas_cls::OnMenuAddPrimitive)
     EVT_MENU (GLMenuShowPanel,      MainGLCanvas_cls::OnMenuShowPanel)
     EVT_KEY_DOWN  (MainGLCanvas_cls::OnKeyDown)
END_EVENT_TABLE()

MainGLCanvas_cls::
MainGLCanvas_cls(GLCanvasCtrl_cls *parent, wxWindowID id, int* attribList,
		 const wxPoint& pos, const wxSize& size,
		 long style, const wxString& name,
		 const wxGLContext *pContextToShare,
		 unsigned int selfIndex, int displayMode) : 
    //wxGLCanvas(parent, id, attribList, pos, size, style|wxFULL_REPAINT_ON_RESIZE, name), //$BK\Ev$O$3$C$A$r;H$&5$$,$9$k$1$I$&$^$/$$$+$J$$(B
    wxGLCanvas(parent, id, pos, size, style | wxFULL_REPAINT_ON_RESIZE, name, attribList)

{
    mpCGLCtrl = parent;
    mpCParentFrameLink = NULL;
    
    SetBackgroundColour(*wxBLACK);
    mSelfIndex = selfIndex;
    mbInitialized    = FALSE;
    mPrevFrame       = -1;
    mbMouseDragging  = FALSE;
    mbItemDragging   = FALSE;
    mVHitObj.push_back(0); mVHitObj.push_back(0);
    mbSaveImageQueue = FALSE;
    mstrWarning = wxT("");
    mWarningLine = 0;

    LOOP_XYZ mCCameraPos.mPos[xyz] = 3.5 * GLPerspective_cls::sDefaultCameraPos[displayMode][xyz];
    LOOP_XYZ mCCameraPos.mTgt[xyz] = - GLPerspective_cls::sDefaultCameraPos[displayMode][xyz];
    LOOP_XYZ mCCameraPos.mTop[xyz] = GLPerspective_cls::sDefaultCameraTop[displayMode][xyz];
    mCCameraPos.NormalizeAndOrthogonalize();
    mCCameraPosCached = mCCameraPos;
    
    std::wstringstream ssTag;
    ssTag << "Panel" << selfIndex;
    
    mCConfList.Add(&mCColorBG0 , ssTag.str() + L"BGColor0=", L"0 0 0");
    mCConfList.Add(&mCColorBG1 , ssTag.str() + L"BGColor1=", L"0.5 0.5 0.5");
    mCConfList.Add(&mCbShowBox , ssTag.str() + L"ShowBox=", L"1");
    mCConfList.Add(&mCbShowWire, ssTag.str() + L"ShowWire=", L"1");
    mCConfList.Add(&mCbShowData, ssTag.str() + L"ShowData=", L"1");
    mCConfList.Add(&mCbShowGrid, ssTag.str() + L"ShowGrid=", L"1");
    mCConfList.Add(&mCbShowItems , ssTag.str() + L"ShowItems=", L"1");
    mCConfList.Add(&mCbStereoFlip, ssTag.str() + L"StereoFlip=", L"0");
    
    mCConfList.Add(&mCGLPersAngle, ssTag.str() + L"GLPersAngle=", L"45");
    mCConfList.Add(&mCbShowOtherCam, ssTag.str() + L"ShowOtherCam=", L"0");
    mCConfList.Add(&mCbXOR, ssTag.str() + L"DrawItemByXOR=", L"0");
    mCConfList.Add(&mCGLGeometry, ssTag.str() + L"GLGeometry=", L"0");
    mCConfList.Add(&mbCFBO, ssTag.str() + L"UseFBO=", L"0");
    mCConfList.Add(&mCFBOSizeIndex, ssTag.str() + L"FBOSize=", L"2");
    mCConfList.Add(&mCDisplayMode, ssTag.str() + L"DisplayMode=", L"0");
    mCConfList.Add(&mCPathIndex, ssTag.str() + L"GLPathIndex=", L"0");
    mCConfList.Add(&mCGridType,  ssTag.str() + L"GridType", L"0");
    mCConfList.Add(&mCbGridAnti, ssTag.str() + L"GridAntialias", L"0");
    mCConfList.Add(&mCbShowGridTag, ssTag.str() + L"GridShowTag", L"0");

    SetDisplayMode(displayMode);

    //mCGLPersAngle.mBody = mCGLPerspective.mDeg;
        
#ifdef _FLAG_SET_CURRENT
    mpCGLContext = new wxGLContext(this, pContextToShare);
#else
    wxPalette palette;
    mpCGLContext = new wxGLContext(true, this, palette, pContextToShare);
#endif
    
    mpCGLGuide = new GLGuides_cls;
    mpCGLGuideScrew = new GLGuides_cls(1);
    mpCGLGuideYPR = new GLGuides_cls(3);
    mpCLinkedPath = NULL;
    mpCLinkedImgSeq = NULL;

    //--- SetSubWindows;

    // Status bar $B$,L5$$$N$G(B helpString$B$OI=<($5$l$J$$$N$G>JN,(B
    mpPopupMenu = new wxMenu; // $B%a%$%s1&%/%j%C%/%a%K%e!<(B

    mpPopupMenu->Append(GLMenuMoveToCamPos, _("Move to Cam Pos"));
    mpPopupMenu->Append(GLMenuSetKeyPos   , _("Sey Key Cam Pos"));
    mpPopupMenu->Append(GLMenuSetKeyAnc   , _("Sey Key Anc Pos"));
    mpPopupMenu->Append(GLMenuToggleLink  , _("Link/Unlink to Cam"));

    mpPopupMenu->AppendSeparator();
    mpMenuAdd = new wxMenu;
    mpMenuAdd->Append(menuItemAddPrimitive, _("Add Primitive"));
    mpMenuAdd->Append(menuItemAddLight, _("Add Light"));
    mpMenuAdd->Append(menuItemAddHotSpot, _("Add HotSpot"));
    
    mpPopupMenu->Append(GLMenuAdd, _("Add Item"), mpMenuAdd);
    
    mpPopupMenu->AppendSeparator();
    mpMenuItemShowPanel = mpPopupMenu->Append(GLMenuShowPanel   , _("Show Control Panel"));

    mpPopupMenu->UpdateUI(GetEventHandler());
}

MainGLCanvas_cls::
~MainGLCanvas_cls(){
    delete mpCGLContext;
    delete mpCGLGuide;
    delete mpCGLGuideScrew;
    delete mpCGLGuideYPR;
    delete mpPopupMenu;
    delete mpCMouseCtrl;
}

// ----------

void MainGLCanvas_cls::
SetParentClass(FrameMain_cls* pCParentFrame){
    mpCParentFrameLink = pCParentFrame;
}

bool MainGLCanvas_cls::
SetGLContext(){
#ifndef __WXMOTIF__
    if (!GetContext()) return FALSE; 
#endif
    
#ifdef _FLAG_SET_CURRENT
    mpCGLContext->SetCurrent(*(wxGLCanvas*)this);
#else
    mpCGLContext->SetCurrent();
#endif
    return TRUE;
}

void MainGLCanvas_cls::
InitGL(){
    glClearColor( 0, 0, 0, 1.0f );
    glClearStencil(0);

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    
    glEnable(GL_DEPTH_TEST);
    
    GLfloat light0Pos[4]    = { 1.0f, 1.0f, 1.0f, 0.0f};
    GLfloat light0Color[4]  = { 1.0f, 1.0f, 1.0f, 1.0f};
    GLfloat light0ColorAmbient[4]  = { 1.0f, 1.0f, 1.0f, 1.0f};
    glLightfv(GL_LIGHT0, GL_POSITION, light0Pos);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0Color);
    glLightfv(GL_LIGHT0, GL_AMBIENT,  light0Color);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0Color);
    
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);

    //glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glDisable(GL_COLOR_MATERIAL);

    glLineStipple(2, 0xAAAA);
}

void MainGLCanvas_cls::
InitializeGLIfNot(){
    if (mbInitialized == TRUE) return;
    
    InitGL();
    mpCParentFrameLink->mpCGLInfo->Initialize();
    UpdateWindowSizeInChildClass();
    
    displayListSphere = GLOBJ::Sphere::DListDraw(3, GL_COMPILE);
    displayListCylinder = 
	GLOBJ::Cone::DListDraw(90, 4, 1, FALSE, GL_COMPILE);
    displayListCylinderOpen = 
	GLOBJ::Cone::DListDraw(90, 4, 1, TRUE, GL_COMPILE);
    displayListCone       = GLOBJ::Cone::DListDraw(90, 4, 0, FALSE, GL_COMPILE);
    displayListConeSimple = GLOBJ::Cone::DListDraw(12, 1, 0, FALSE, GL_COMPILE);
    
    for (unsigned int i = 0; i < 4; i++)
	displayListTransCube[i] = 
	    GLOBJ::TransCube::DListDraw(GL_COMPILE, i);

    for (int i = 0; i < 5; i++)
	displayListWireBox[i] = 
	    GLOBJ::WireBox::DListDraw(GL_COMPILE, i);
        
    if (mpCParentFrameLink->mpCGLInfo->mMajorVersion >= 2) 
	RedrawPBO_cls::msbSupported = TRUE;
    if (mpCParentFrameLink->mpCGLInfo->mMajorVersion == 1 &&
	mpCParentFrameLink->mpCGLInfo->mMinorVersion >= 5) 
	RedrawPBO_cls::msbSupported = TRUE;

    if (1){// $B8:;;=hM}$,2DG=$@$C$?$i(B
#ifdef WIN32
	glBlendEquationEXT = (PFNGLBLENDEQUATIONEXTPROC) wglGetProcAddress("glBlendEquationEXT");
	glBlendColor = (PFNGLBLENDCOLORPROC) wglGetProcAddress("glBlendColor");
	
	glColorTableEXT = (PFNGLCOLORTABLEEXTPROC)wglGetProcAddress("glColorTableEXT");
#endif 
    }
    
    /*
    wxImage CImageRot(wxT("ICONS/Rotation.png"));
    wxImage CImageTrans(wxT("ICONS/Translation.png"));
    wxImage CImageScale(wxT("ICONS/Scale.png"));
    CImageRot = CImageRot.Mirror(FALSE);
    CImageTrans = CImageTrans.Mirror(FALSE);
    CImageScale = CImageScale.Mirror(FALSE);
    */

    glEnable(GL_TEXTURE_2D);
    /*
    glGenTextures(1, &msTexIDRotation);
    MakeTextureFromImage(&CImageRot, 24, 24, msTexIDRotation);
    glGenTextures(1, &msTexIDTranslation);
    MakeTextureFromImage(&CImageTrans, 24, 24, msTexIDTranslation);
    glGenTextures(1, &msTexIDScale);
    MakeTextureFromImage(&CImageScale, 24, 24, msTexIDScale);
    */

    msCGLBTNRot.MakeTextureFromImage(wxT("ICONS/Rotation.png"));
    msCGLBTNTrans.MakeTextureFromImage(wxT("ICONS/Translation.png"));
    msCGLBTNScale.MakeTextureFromImage(wxT("ICONS/Scale.png"));
    
    glDisable(GL_TEXTURE_2D);
    mbInitialized = TRUE;
}

void MainGLCanvas_cls::
UpdateWindowSizeInChildClass(){
    int w, h;
    GetClientSize(&w, &h);
    mCGLPerspective.SetWindowSize(w, h);
}

// ----

DF3SeqCtrl_cls* MainGLCanvas_cls::
GetPDF3SeqCtrl(){
    return mpCParentFrameLink->mpCDF3SeqCtrl;
}

TimeCtrl_cls* MainGLCanvas_cls::
GetPTimeCtrl(){
    return mpCParentFrameLink->mpCParent->mpCTime;
}

TreeBoxRule_cls* MainGLCanvas_cls::
GetPTreeBoxRule(){
    return mpCParentFrameLink->mpCParent->mpCTreeBoxRule;
}

void MainGLCanvas_cls::
UpdateCameraPos(){
    if (mpCLinkedPath != NULL){
	mCCameraPos = mpCLinkedPath->mCCurrentCam;
    }
}

CameraPath_cls* MainGLCanvas_cls::
GetPLinkedPath(){
    if (mpCLinkedPath == NULL) 
	return mpCParentFrameLink->mpCPathCtrl->GetActivePath();
    return mpCLinkedPath;
}

bool MainGLCanvas_cls::
IsStereo(){
    if (GetPLinkedPath()->mpCStereoGramLinked->mCType.mBody != _STEREO_NONE){
	if (GetPLinkedPath()->mpCStereoGramLinked->mCEye.mBody == _STEREO_NONE)
	    return FALSE;
	if (mCDisplayMode.mBody == _DIRECTION_FREE || 
	    mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	    return TRUE;
    }
    return FALSE;
}

bool MainGLCanvas_cls::
IsOrtographic(){
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0 &&
	mCDisplayMode.mBody < _GL_CANVAS_COPY_OF_VIEW_0)
	return (GetPLinkedPath()->mCCamGeometry.mBody == _POV_ORTHOGRAPHIC);
    
    return (mCGLGeometry.mBody == _POV_ORTHOGRAPHIC);
}

bool MainGLCanvas_cls::
IsFishEye(){
    if (mCDisplayMode.mBody == _DIRECTION_FREE ||
	mCDisplayMode.mBody == _GL_CANVAS_COPY_OF_VIEW_0)
	return (mCGLGeometry.mBody == _POV_FISHEYE);
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	return (GetPLinkedPath()->mCCamGeometry.mBody == _POV_FISHEYE);
    return FALSE;
}

bool MainGLCanvas_cls::
IsPanoramic(){
    if (mCDisplayMode.mBody == _DIRECTION_FREE)
	return (mCGLGeometry.mBody == _POV_CYLINDER);
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	return (GetPLinkedPath()->mCCamGeometry.mBody == _POV_CYLINDER);
    return FALSE;
}

bool MainGLCanvas_cls::
IsFBOUse(){
    // !!FIX_ME!! FBO$B$r;H$($J$$$H$-$N=hM}$,L5$$(B
    if (mbCFBO.mbBody == TRUE) return TRUE;
    if (IsFishEye() == TRUE) return TRUE;
    if (IsPanoramic() == TRUE) return TRUE;
    return FALSE;
}

void MainGLCanvas_cls::
SetAngleIfLinked(){
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0 &&
	mCDisplayMode.mBody < _GL_CANVAS_COPY_OF_VIEW_0){
	mCGLPerspective.mDeg = GetPLinkedPath()->mCVAngle.mBody;
	if (GetPLinkedPath()->mCCamGeometry.mBody == _POV_PERSPECTIVE &&
	    mCGLPerspective.mDeg >= 120) {
	    mCGLPerspective.mDeg = 120;
	    mstrWarning << wxT("View angle is limited to 120 deg in perspective view.") << ENDLWX;
	    mWarningLine++;
	}
    }
    else{
	mCGLPerspective.mDeg = mCGLPersAngle.mBody;
	if (mCGLGeometry.mBody == _POV_PERSPECTIVE &&
	    mCGLPerspective.mDeg >= 120) {
	    mCGLPerspective.mDeg = 120;
	    mstrWarning << wxT("View angle is limited to 120 deg in perspective view.") << ENDLWX;
	    mWarningLine++;
	}
    }
}

void MainGLCanvas_cls::
ModifyByStereogram(int forceType){
    int orgType = GetPLinkedPath()->mpCStereoGramLinked->mCEye.mBody;
    if (forceType != -1)
	GetPLinkedPath()->mpCStereoGramLinked->mCEye.mBody = forceType;
    
    GetPLinkedPath()->ModifyStereogramPos(mCCameraPos);
    mCCameraPos.RotateYPR(GetPLinkedPath()->GetStereogramYawAngle(), 0, 0);
    
    GetPLinkedPath()->mpCStereoGramLinked->mCEye.mBody = orgType;
}

void MainGLCanvas_cls::
ModifyByDomeTilt(PositionAndDirection_cls* pCCamPos){
    if (pCCamPos == NULL)
	mCCameraPos.RotateYPR(0, GetPLinkedPath()->GetDomePitchAngle(), 0);
    else
	pCCamPos->RotateYPR(0, GetPLinkedPath()->GetDomePitchAngle(), 0);
}

void MainGLCanvas_cls::
ModifyByDome(int forceType){
    switch(forceType){
    case 0:
	mCCameraPos.RotateYPR(-45, 0 , 0); break;
    case 1:
	mCCameraPos.RotateYPR(45, 0, 0);   break;
    case 2:
	mCCameraPos.RotateYPR(-45, 90, 0); break;
    case 3:
	mCCameraPos.RotateYPR(45, -90, 0); break;
    }
}

void MainGLCanvas_cls::
ModifyByPanorama(int forceType){
    switch(forceType){
    case 0:
	mCCameraPos.RotateYPR(-45, 0 , 0); break;
    case 1:
	mCCameraPos.RotateYPR(-135, 0, 0);   break;
    case 2:
	mCCameraPos.RotateYPR(+45, 0, 0); break;
    case 3:
	mCCameraPos.RotateYPR(+135, 0, 0); break;
    }
}

void MainGLCanvas_cls::
UpdateDrawInfo(int width, int height){
    mCDrawInfoPrev = mCDrawInfo;
    
    mCDrawInfo.mbForceNewDraw = FALSE;
    mCDrawInfo.mCCamPos = mCCameraPos;
    mCDrawInfo.mCAncPos = msCAnchorPos;
    mCDrawInfo.mRemappedIndex = 
	GetPDF3SeqCtrl()->GetRemappedIndex(0/*FIX_ME!!*/);
    mCDrawInfo.mWidth = width;
    mCDrawInfo.mHeight = height;
    mCDrawInfo.mDisplayMode = mCDisplayMode.mBody;
    mCDrawInfo.SetBoolList(mCbShowData.mbBody,   mCbShowGrid.mbBody,
			   mCbShowBox.mbBody,    mCbShowWire.mbBody,
			   mCbShowItems.mbBody,  mCbShowOtherCam.mbBody,
			   mCbGridAnti.mbBody,   mbCFBO.mbBody,
			   mCGLPerspective.mbPerspective);
    mCDrawInfo.SetBoolPolygon(mpCParentFrameLink->mpCDF3SeqCtrl->mAbShow);


    bool bMirr[3] = {FALSE, FALSE, FALSE};
    bool bCopy[3] = {FALSE, FALSE, FALSE};
    int copyFrom[3] = {0, 0, 0};
    int copyTo[3]   = {0, 0, 0};
    if (mspCDF3SeqCtrl->DoesFilterExistInAllDF3(_FILTER_BOXCOPY) == TRUE){
	BoxCopyData_cls* pCData =
	    (BoxCopyData_cls*)mspCDF3SeqCtrl->GetFirstPFilterDataInAllDF3(_FILTER_BOXCOPY);
	if (pCData != NULL && pCData->mCbValid.mbBody == TRUE){
	    LOOP_XYZ{
		bMirr[xyz] = pCData->mCbMirroring.mpbBody[xyz];
		bCopy[xyz] = pCData->mCbCopying.mpbBody[xyz];
		copyFrom[xyz] = pCData->mCCopyFrom.mpBody[xyz];
		copyTo  [xyz] = pCData->mCCopyTo.mpBody[xyz];
	    }
	}
    }

    mCDrawInfo.SetBoolMirror (bMirr);
    mCDrawInfo.SetBoolCopying(bCopy);
    
    mCDrawInfo.SetCopyFrom(copyFrom);
    mCDrawInfo.SetCopyTo  (copyTo);
    
    mCDrawInfo.SetClipInfo(DrawingBlock_cls::msClipMin,
			   DrawingBlock_cls::msClipMax);

    double dummymin[3], dummymax[3];
    LOOP_XYZ dummymin[xyz] = 0;
    LOOP_XYZ dummymax[xyz] = 1;
    std::vector < std::vector < double > > VVClipInfoMin, VVClipInfoMax;
    std::vector < bool > VClipInfoActive;
    mspCFilterCtrl->GetClippingInformation(wxT("_FILTER_ALLINFO"),
					   dummymin, dummymax,
					   &VVClipInfoMin, &VVClipInfoMax,
					   &VClipInfoActive);
    mCDrawInfo.SetClipInfoNew(VVClipInfoMin, VVClipInfoMax, VClipInfoActive);
    
    mCDrawInfo.SetBoxInfo(&DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[0],
			  &DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[3],
			  &DrawingBlock_cls::mspCRotate->mpInterpolatedValue[0]);
    mCDrawInfo.SetLightInfo(mLightNumber, mALightPos, mALightColor);

    double pShadeOffset[3] = {0.001, 0.001, -0.001};
    double doubleIndex = GetPTimeCtrl()->GetDoubleIndex();

    mpCParentFrameLink->mpCParent->mpCLightSRC->
	GetShadeOffset(pShadeOffset, doubleIndex);
    
    mCDrawInfo.SetShadeInfo(pShadeOffset);
    
    mCDrawInfo.SetFilenames(GetPDF3SeqCtrl()->GetCurrentFWXStrAllChannel());
}

double* MainGLCanvas_cls::
GetAnchorPosForRender(){
    static double ancPosForRender[3];
    LOOP_XYZ ancPosForRender[xyz] = 
	mCCameraPos.mPos[xyz] + mCCameraPos.mTgt[xyz];
    return ancPosForRender;
}

void MainGLCanvas_cls::
SetViewport(){
    mCGLPerspective.SetViewportOnGL();
}

void MainGLCanvas_cls::
SetPerspective(){
    
    // $B2hA|=PNO;~$O%Q!<%9%Q%i%a%?$,0c$&(B
    if (mbSaveImageQueue == TRUE){
	if (mbSaveImgSeqQueue == TRUE && mpCLinkedImgSeq != NULL &&
	    IsFishEye() == FALSE && IsPanoramic() == FALSE){
	    float deg = mCGLPerspective.mDeg;
	    float ClipRadian = 2.0 * atan ( tan(mpCLinkedImgSeq->mClipAngle * M_PI/180 / 2.0) / mpCLinkedImgSeq->mClipHeight * mCGLPerspective.mHeight);
	    mCGLPerspective.mDeg = ClipRadian * 180 / M_PI;
	    mCGLPerspective.SetPerspectiveOnGL(&mCCameraPos);
	    mCGLPerspective.mDeg = deg;
	    return;
	}
    }
    mCGLPerspective.SetPerspectiveOnGL(&mCCameraPos);
}

void MainGLCanvas_cls::
SetModelView(){
    mCGLPerspective.SetModelOrProjectionViewOnGL(&mCCameraPos, GL_MODELVIEW);

    FilterBase_cls::mspCCamPos = &mCCameraPos;
    glGetDoublev(GL_MODELVIEW_MATRIX, mMatrixModelView);
    glGetDoublev(GL_PROJECTION_MATRIX, mMatrixProjection);
    glGetIntegerv(GL_VIEWPORT, mMatrixViewport);
}

void MainGLCanvas_cls::
SetLight(){
    // $B8w8;$O:GDc(B8$B$D$^$G$O;H$($k(B
    // $BD4@0:Q$_$N0LCV$r(BmALightPos$B$K;}$C$F$$$k$N$G$3$C$A$G4pK\=hM}(B
    int lightNum = mLightNumber;
    if (lightNum == 0) lightNum = 1; //$B8w8;(B0$B$G$b2>%i%$%H(B
    if (lightNum > 8) lightNum = 8;

    for (int i = 0; i < 8; i++)
	if (i < lightNum) glEnable(msGLLightIndex[i]);
	else glDisable(msGLLightIndex[i]);

    // 8$B$D$a$N%i%$%H$OFC<l(B
    for (int i = 0; i < 7; i++){
	glLightfv(msGLLightIndex[i], GL_POSITION, mALightPos[i]);
	glLightfv(msGLLightIndex[i], GL_DIFFUSE,  mALightColor[i]);
	glLightfv(msGLLightIndex[i], GL_AMBIENT,  mALightColor[i]);
	glLightfv(msGLLightIndex[i], GL_SPECULAR, mALightColor[i]);
	glLightfv(msGLLightIndex[i], GL_LINEAR_ATTENUATION, &mALightAttenuateL[i]);
	glLightfv(msGLLightIndex[i], GL_QUADRATIC_ATTENUATION, &mALightAttenuateQ[i]);

	// 8$B$D$a$N%i%$%H$K$b(BAttenuation
	glLightfv(msGLLightIndex[7], GL_LINEAR_ATTENUATION, &mALightAttenuateL[i]);
	glLightfv(msGLLightIndex[7], GL_QUADRATIC_ATTENUATION, &mALightAttenuateQ[i]);
	mpCParentFrameLink->mpCParent->mpCLightSRC->
	    SetGLLightOptional(msGLLightIndex[i], i, mALightPos[i],
			       mALightColor[i]);
    }
}

void MainGLCanvas_cls::
DrawDataDrawObject(double scale, double scale10, 
		   float fogDistance, float fogColor[4],
		   std::vector < DrawingBlock_cls > &VCDrawingCls, 
		   int stencilMode){

    glDisable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    
    GetPDF3SeqCtrl()->DrawDataInitSetting(); //GL_ATTRIB $B$N=i4|2=(B

    if (mCbGridAnti.mbBody == TRUE){
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);
    }

    //DrawGrid(scale, scale10); // $B%0%j%C%II=<((B

    // $B8w8;J}8~$rF@$k(B
    double pShadeOffset[3] = {0.003, 0.003, -0.003};
    double doubleIndex = GetPTimeCtrl()->GetDoubleIndex();
    LightSRC_cls::SetParentObjPos
	(&mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCCurrentCam);
    mpCParentFrameLink->mpCParent->mpCLightSRC->GetShadeOffset
	(pShadeOffset, doubleIndex);

    SetLight();

    if (fogDistance != -1 && fogDistance < 9998){
	glEnable(GL_FOG);
	glFogf(GL_FOG_DENSITY, 1.0f/fogDistance);
	glFogfv(GL_FOG_COLOR, fogColor);
    }
    else
	glDisable(GL_FOG);
    
    // Object$B$O(Bglobal$B:BI8$G@h$KIA$/(B
    glEnable(GL_CULL_FACE);
    if (stencilMode >= 1){
	// $BN"(B($B1|B&(B)$B$rIA<L(B 
	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS, 1, ~0); 
	glCullFace(GL_FRONT); 
	glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
	//$B:9J,%*%V%8%'%/%H$N$_IA$/(B
	GetPDF3SeqCtrl()->DrawObject(scale, FALSE, TRUE); 
	if (stencilMode == 1) DrawGrid(scale, scale10); // $B%0%j%C%II=<((B
	
	if (stencilMode == 2){
	    // $BI=!J<jA0B&(B)$B$rIA<L(B depth$BH=Dj$O(B2$B$N;~$ODL>oDL$j(B
	    glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
	    DrawGrid(scale, scale10); // $B%0%j%C%II=<((B
	    glCullFace(GL_BACK); 
	    glStencilOp(GL_KEEP,GL_KEEP,GL_DECR);
	    GetPDF3SeqCtrl()->DrawObject(scale, TRUE, TRUE);
	    // $B%+%a%i$,JD6JLL$NCf$K$$$k$H!"%9%F%s%7%k$,(B+$B$K$J$k(B
	}

    }
    else{
	DrawGrid(scale, scale10); // $B%0%j%C%II=<((B
	glCullFace(GL_FRONT);
	GetPDF3SeqCtrl()->DrawObject(scale, FALSE, TRUE);
	glCullFace(GL_BACK);
	GetPDF3SeqCtrl()->DrawObject(scale, TRUE, FALSE);
    }

    glDisable(GL_CULL_FACE);

    glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
    if (stencilMode == 1){
	glStencilFunc(GL_LEQUAL, 1, ~0); //1$B$,%9%F%s%7%k%P%C%U%!0J2<(B:$B$D$^$j%9%F%s%7%k$,(B1$B0J>e$N;~=q$/(B
    }

    if (stencilMode == 2){
	glStencilFunc(GL_EQUAL, 0, ~0); //1$B$,%9%F%s%7%k$h$j$bBg$-$$!'$D$^$j%9%F%s%7%k$,(B0$B$N$H$3$m$N$_IA2h(B
    }
    
    mCDataParentCoordinate.SetZero();
    mCDataParentCoordinate.SetvUnitZ(-1);
    mCDataParentCoordinate.RotateRoll(-DrawingBlock_cls::mspCRotate->mpInterpolatedValue[0]);
    mCDataParentCoordinate.RotatePitch(DrawingBlock_cls::mspCRotate->mpInterpolatedValue[1]);
    mCDataParentCoordinate.RotateYaw (-DrawingBlock_cls::mspCRotate->mpInterpolatedValue[2]);
    mCDataParentCoordinate.RotateRoll(-DrawingBlock_cls::mspCRotate->mpInterpolatedValue[3]);
    mCDataParentCoordinate.NormalizeAndOrthogonalize();

    glPushMatrix();
    // $B:BI8$N2sE>(B
    mCDataParentCoordinate.GLSetMatrixToThisCoordinate();

    // --- Box$B$NI=<((B ---
    int layerNum = GetPDF3SeqCtrl()->GetMaxColorLayerNum();
    for (unsigned int cindex = 0; cindex < layerNum; cindex++){
	
	GetPDF3SeqCtrl()->SetRGBAtoGLAllChannel(cindex);
	
	for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
	    if (VCDrawingCls[i].mbVisibility == TRUE ||
		(GetPTreeBoxRule()->mCType.mBody == _TREE_NESTED &&
		 VCDrawingCls[i].mDepth == 0)){
		if (mpCParentFrameLink->mParamForDebugginng == i)
		    DrawingBlock_cls::msParamForDebugging = TRUE;
		else
		    DrawingBlock_cls::msParamForDebugging = FALSE;
		VCDrawingCls[i].DrawDataBlock
		    (mCCameraPos, mCbShowData.mbBody, 
		     cindex == 0, pShadeOffset);
	    }
    }

    glPopMatrix();

    // --- Box $B$NI=<(=*N;(B ---
    GetPDF3SeqCtrl()->DrawDataEndSetting();

    if (mCbGridAnti.mbBody == FALSE){
	glDisable(GL_BLEND);
	glDisable(GL_LINE_SMOOTH);
    }

    //$B4D6-$N2sI|(B
    glDisable(GL_FOG);
    glDepthMask(GL_TRUE);
    glDisable(GL_STENCIL_TEST);
}

void MainGLCanvas_cls::
ClearBuffer(){
    float fogColor[4] = {0, 0, 0, 0};
    float fogDistance = 
	mpCParentFrameLink->mpCDF3SeqCtrl->GetFogInfoAll(fogColor, _("Global"));
    
    //Clear
    if (fogDistance == -1)
	glClearColor( mCColorBG0.mpBody[R], 
		      mCColorBG0.mpBody[G],
		      mCColorBG0.mpBody[B], 1.0f);
    else
	glClearColor( fogColor[R], fogColor[G], fogColor[B], 1.0f);

    glClear( GL_COLOR_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void MainGLCanvas_cls::
DrawDataMain(double scale, double scale10, 
	     std::vector < DrawingBlock_cls > &VCDrawingCls){

    float fogColor[4] = {0, 0, 0, 0};
    float fogDistance = 
	mpCParentFrameLink->mpCDF3SeqCtrl->GetFogInfoAll(fogColor, _("Global"));
    
    bool bDiffObject = mpCParentFrameLink->mpCDF3SeqCtrl->
	DoesFilterExistInAllDF3(_FILTER_POV_DIFFOBJECT);

    if (bDiffObject == TRUE){

	if (IsFBOUse() == TRUE && FALSE){
	    // CHECK_ME!! OpenGL$B$N%P!<%8%g%s$K$h$C$F$O(BFBO$B;~$O%9%F%s%7%k$r;H$($J$$!#(B
	    glClearDepth(0);
	    glClear(GL_DEPTH_BUFFER_BIT);
	    glDepthFunc(GL_GREATER); 
	    DrawDataDrawObject(scale, scale10, fogDistance, fogColor,
			       VCDrawingCls, _STENCIL_MODE_0);
	    // $B$9$k$HDL>o%*%V%8%'%/%H$r$^$H$b$K$+$1$J$$!D(B
	}
	else {
	    glClearDepth(0);
	    glClear(GL_DEPTH_BUFFER_BIT);
	    glDepthFunc(GL_GREATER); // $B1s$/$rIA$/(B
	    DrawDataDrawObject(scale, scale10, fogDistance, fogColor,
			       VCDrawingCls, _STENCIL_MODE_1);
	    
	    glClearDepth(1); // $B6a$/$rIA$/(B
	    glDepthFunc(GL_LESS); 
	    glClear( GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	    DrawDataDrawObject(scale, scale10, fogDistance, fogColor,
			       VCDrawingCls, _STENCIL_MODE_2);
	}
    }
    else {
	glClearDepth(1);
	glClear(GL_DEPTH_BUFFER_BIT);
	glDepthFunc(GL_LESS); 
	
	DrawDataDrawObject(scale, scale10, fogDistance, fogColor,
			   VCDrawingCls, _STENCIL_MODE_0);
    }

    glClearDepth(1);
    glDepthFunc(GL_LESS);
}


void MainGLCanvas_cls::
DrawCameraPos(double scale, PositionAndDirection_cls CPos){
    glColor4f(0.8f, 0.2f, 0.2f, 0.5f);
    glPushMatrix();
    GLOBJ::SetMatrix3dv3(CPos.mPos, CPos.mTgt, CPos.mTop);
    glScalef(scale, scale, scale);
    GLOBJ::Object::Draw(_CAMERA_1);
    GLOBJ::Object::Draw(_CAMERA_2);
    glPopMatrix();
}

void MainGLCanvas_cls::  

DrawGrid(double scale, double scale10){

    if (mCbShowGrid.mbBody == FALSE) return;
    glColor4f(0.3f, 0.3f, 0.3f, 1.f);
    if (mCGridType.mBody != _GL_CANVAS_GRID_POLAR)
	GLOBJ::GridFlat::Draw(scale10);
    if (mCGridType.mBody == _GL_CANVAS_GRID_BASIC_CIRCULAR) 
	GLOBJ::GridCircularFlat::Draw(scale10);
    if (mCGridType.mBody == _GL_CANVAS_GRID_POLAR){
	GLOBJ::GridPolar::Draw(scale10);
    }

    // $B:BI8<4$NI=<(0LCV(B
    double obj[2][3][3] 
	= {{{+scale * 0.35, 0, 0}, {0, +scale * 0.35, 0}, {0, 0, +scale * 0.35}},
	   {{-scale * 0.35, 0, 0}, {0, -scale * 0.35, 0}, {0, 0, -scale * 0.35}}};
    
    for (int pm = 0; pm < 2; pm++){
	LOOP_XYZ{
	    double cosAngle = 
		MyVector::GetScalarProduct(obj[pm][xyz], mCCameraPos.mTgt) 
		/ ( 0.35 * scale); 
	    if (cosAngle > -0.9 && cosAngle < 0.9){
		mbShowCoordinateTagXYZ[xyz] = TRUE;
		gluProject(obj[pm][xyz][X],obj[pm][xyz][Y],obj[pm][xyz][Z],
			   mMatrixModelView, mMatrixProjection, mMatrixViewport,
			   &mPosCoordinateTag[pm][xyz][X],
			   &mPosCoordinateTag[pm][xyz][Y],
			   &mPosCoordinateTag[pm][xyz][Z]); }
	    else
		mbShowCoordinateTagXYZ[xyz] = FALSE; 
	}
    }
}

void MainGLCanvas_cls::
DrawModeAxis(double scale){
    if (msMoveMode == _POS_AND_DIR_MOVE_FREE ||
	msManipulatorMode == _MANIP_ROTATE) return;
    
    switch(msMoveMode){
    case _POS_AND_DIR_MOVE_X:
	glColor4f(0.8, 0.2, 0.2, 1);
	glBegin(GL_LINES);
	glVertex3d(-scale*10, 0, 0); glVertex3d(scale*10, 0, 0);
	glEnd();
	break;
    case _POS_AND_DIR_MOVE_Y:
	glColor4f(0.2, 0.8, 0.2, 1);
	glBegin(GL_LINES);
	glVertex3d(0, -scale*10, 0); glVertex3d(0, scale*10, 0);
	glEnd();
	break;
    case _POS_AND_DIR_MOVE_Z:
	glColor4f(0.4, 0.4, 1.0, 1);
	glBegin(GL_LINES);
	glVertex3d(0, 0, -scale*10); glVertex3d(0, 0, scale*10);
	glEnd();
	break;
    }
}

void MainGLCanvas_cls::
DrawItems(double scale){
    if (mCbShowItems.mbBody == FALSE) return;

    if (mCbGridAnti.mbBody == TRUE){
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);
    }

    if (mCbXOR.mbBody == TRUE){
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE_MINUS_SRC_COLOR);
    }

    DrawModeAxis(scale);
    
    int framenum = GetPTimeCtrl()->mCMaxFrame.mBody + 1;
    int interval = TimeCtrl_cls::GetInterval(framenum);
    DrawItem_cls::SetParams(framenum, 
			    GetPTimeCtrl()->GetDoubleIndexWAMin(),
			    GetPTimeCtrl()->GetDoubleIndexWAMax(),
			    scale*0.01, interval);
    
    mpCParentFrameLink->mpCPathCtrl->DrawPath(mCCameraPos);  // $B%+%a%i%Q%9(B
    
    mpCParentFrameLink->mpCStereo->DrawFocus
	(mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCCurrentCam,
	 GetPTimeCtrl()->GetDoubleIndex(),
	 scale); // $B%9%F%l%*%0%i%`(B

    // $B8w8;$H%[%C%H%9%]%C%H(B
    mpCParentFrameLink->mpCParent->mpCLightSRC->DrawGLPath
	(_CAMERA_CROSSHAIR, 0.015 * scale, mCCameraPos);
    
    DrawingBlock_cls::msCHotSpot.DrawGLPath(_GLOBJ_WIREBOX, 1, -1); 

    // $BB>$N%+%a%i(B
    if (mCbShowOtherCam.mbBody){
	for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++)
	    DrawCameraPos(scale*0.01, mVPOtherCanvas[i]->mCCameraPos);
	for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++){
	    gluProject(mVPOtherCanvas[i]->mCCameraPos.mPos[X],
		       mVPOtherCanvas[i]->mCCameraPos.mPos[Y],
		       mVPOtherCanvas[i]->mCCameraPos.mPos[Z],
		       mMatrixModelView, mMatrixProjection, mMatrixViewport,
		       &mPosCameraTag[i][X],
		       &mPosCameraTag[i][Y],
		       &mPosCameraTag[i][Z]); }
	
    }
    
    glDisable(GL_LINE_SMOOTH);
}

void MainGLCanvas_cls::
DrawAnchor(double scale){
    if (mCbShowItems.mbBody == FALSE) return;

    glColor4f(0.2f, 1.0f, 1.0f, 1.f);
    glPushMatrix();
    glTranslated(msCAnchorPos.mPos[0],
		 msCAnchorPos.mPos[1],
		 msCAnchorPos.mPos[2]);

    glScaled(scale*0.02, scale*0.02, scale*0.02);
    GLOBJ::Object::Draw(_CAMERA_CROSSHAIR);
    glPopMatrix();
}

void MainGLCanvas_cls::
DrawSelectedItem(double scale){
    if (mspCSelectedItem == NULL) return;
    glColor4f(1.0, 1.0, 0.2, 1.0);
    bool bSelectKey = FALSE;
    double dpos[3] = {
	msSelectedItemPos[0] - msSelectedItemPosOrg[0],
	msSelectedItemPos[1] - msSelectedItemPosOrg[1],
	msSelectedItemPos[2] - msSelectedItemPosOrg[2]};

    if (mspCSelectedItem->mLabelStr == L"Light"){
	int selected = mpCParentFrameLink->mpCParent->mpCLightSRC->GetSelectedIndex();
	if (selected != -1)
	    mpCParentFrameLink->mpCParent->mpCLightSRC->DrawSelectedPath
		(mpCGLGuide, dpos, _CAMERA_CROSSHAIR, 
		 0.03, selected);
    }

    if (mspCSelectedItem->mLabelStr == L"HSpot"){
	int selected = DrawingBlock_cls::msCHotSpot.GetSelectedIndex();
	if (selected != -1){
	    PositionAndDirection_cls CTmp;
	    DrawingBlock_cls::msCHotSpot.mpCGUI->DrawSelectedPath(mpCGLGuide, dpos, _GLOBJ_WIREBOX, 1, selected, CTmp); 
	}
    }
    
    if (mspCSelectedItem->mLabelStr == L"CamPos" && 
	!(mCCameraPos == mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCCurrentCam))
	mpCParentFrameLink->mpCPathCtrl->GetActivePathGLGUI()->DrawCamPath
	    (mpCGLGuideScrew, dpos, msManipulatorMode,
	     msSelectedItemScale, msSelectedItemDYPRY);
    
    // $B0LCV$,(B CamPos $B$r;H$&4X78$G!"(BSelected Key $B$NI=<($,:$Fq$K$J$C$F$$$k(B
    if (mspCSelectedItem->mLabelStr == L"CamRotYPR" &&
	!(mCCameraPos == mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCCurrentCam)){
	double DYPR[4];
	LOOP_YPRY DYPR[ypry] = msSelectedItemDYPRY[ypry];
	mpCParentFrameLink->mpCPathCtrl->GetActivePathGLGUI()->DrawCamPath
	    (mpCGLGuideYPR, dpos, _MANIP_ROTATE,
	     msSelectedItemScale, DYPR);
    }
    
    if (mspCSelectedItem->mLabelStr == L"AncPos")
	mpCParentFrameLink->mpCPathCtrl->GetActivePathGLGUI()->DrawAncPath
	    (mpCGLGuide, dpos, _MANIP_TRANSLATE);
    
    if (mspCSelectedItem->mLabelStr == L"DensityCurvePoints"){
	glPushMatrix();
	double boxScale[3];
	LOOP_XYZ boxScale[xyz] = 0.5 * 
	    (DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[xyz+3] -
	     DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[xyz]);
	GLOBJ::AdjustCoordinateMinMax
	    (DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue,
	     DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue+3);

	mspCSelectedItem->DrawGLTwoGuides(mpCGLGuide, 0.25*scale, boxScale, mCDataParentCoordinate);
	glPopMatrix();
    }

    if (mspCSelectedItem->mLabelStr.Matches(L"PovObject*")){
	glPushMatrix();
	FilterBase_cls* pCSelectedFilter = mspCFilterCtrl->GetPSelectedItem();
	if (pCSelectedFilter != NULL)
	    ((PovObject_cls*)pCSelectedFilter)->DrawGLGuide
		(mpCGLGuide, dpos, 0.5 * scale, msManipulatorMode, 
		 msSelectedItemScale, msSelectedItemDYPRY);
	mspCSelectedItem->DrawGLPath3dv
	    (DrawItem_cls::mStepNum, 0, 1, 0.1,
	     DrawItem_cls::mInterval, FALSE, FALSE);
	glPopMatrix();
    }

    if (mspCSelectedItem->mLabelStr.Matches(L"PovDiffObject*")){
	glPushMatrix();
	FilterBase_cls* pCSelectedFilter = mspCFilterCtrl->GetPSelectedItem();
	if (pCSelectedFilter != NULL)
	    ((PovDiffObject_cls*)pCSelectedFilter)->DrawGLGuide
		(mpCGLGuide, dpos, 0.5 * scale, msManipulatorMode, 
		 msSelectedItemScale, msSelectedItemDYPRY);
	mspCSelectedItem->DrawGLPath3dv
	    (DrawItem_cls::mStepNum, 0, 1, 0.1,
	     DrawItem_cls::mInterval, FALSE, FALSE);
	glPopMatrix();
    }
    
    glPushMatrix();
    // $B:BI8$N2sE>(B

    if (mspCSelectedItem->mLabelStr == L"BoxRotate"){
	glPushMatrix();
	double posBox[6] = {0, 0, 0, 0, 0, 0};
	PositionAndDirection_cls CParentObjPos;
	mpCGLGuide->DrawGuideRotation(posBox, scale, 1, NULL, CParentObjPos, 
				      mspCSelectedItem->mpInterpolatedValue);
	
	double pZero[3] = {0, 0, 0}, pOne[3] = {1, 1, 1};
	GLOBJ::AdjustCoordinateMinMax (pZero, pOne);
	double boxScale[3] = {0.5, 0.5, 0.5};
	double dpos[6] = {0, 0, 0, 0, 0, 0};
	glRotated(mspCSelectedItem->mpInterpolatedValue[0] + msSelectedItemDYPRY[0], 0, 0 ,1);
	glRotated(mspCSelectedItem->mpInterpolatedValue[1] + msSelectedItemDYPRY[1], 1, 0, 0);
	glRotated(mspCSelectedItem->mpInterpolatedValue[2] + msSelectedItemDYPRY[2], 0, 1, 0);
	glRotated(mspCSelectedItem->mpInterpolatedValue[3] + msSelectedItemDYPRY[3], 0, 0, 1);
	glColor4f(1, 1, 0, 1);

	GLOBJ::DrawWireBoxMinMax(DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue,
				 DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue+3);
	glPopMatrix();
    }

    mCDataParentCoordinate.GLSetMatrixToThisCoordinate();

    if (mspCSelectedItem->mLabelStr == L"BoxSize"){
	glPushMatrix();
	double pZero[3] = {0, 0, 0}, pOne[3] = {1, 1, 1};
	GLOBJ::AdjustCoordinateMinMax (pZero, pOne);
	double boxScale[3] = {0.5, 0.5, 0.5};
	double posBox[6];
	LOOP_XYZ posBox[xyz] =   mspCSelectedItem->mpInterpolatedValue[xyz];
	LOOP_XYZ posBox[xyz+3] = mspCSelectedItem->mpInterpolatedValue[xyz+3];
	
	mCDataParentCoordinate.RotTransformToThisCoordinate(dpos);
	Parameters_cls::MovePosForBox(posBox, dpos, msSelectedItemScale, mVHitObj[0]); // hitObj$B$KBP1~$7$F0LCV$N0\F0(B
	
	mpCGLGuide->DrawGuideArrowBox(posBox, scale, boxScale,
				      _GL_GUIDES_NOLIMIT_FOR_BOX);
	glPopMatrix();
    }

    

    if (mspCSelectedItem->mLabelStr.Matches(L"ClippingBox")){
	glPushMatrix();

	GLOBJ::AdjustCoordinateMinMax
	    (DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue,
	     DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue+3);
	if (FilterClipping_cls::msCConfClippingFollowTracking.mbBody == TRUE)
	    glTranslated(DrawingBlock_cls::msTranslate[X],
			 DrawingBlock_cls::msTranslate[Y],
			 DrawingBlock_cls::msTranslate[Z]);

	double clipScaleMin[3] = {0, 0, 0}, clipScaleMax[3] = {1, 1, 1};
	wxString strTagActivated = GetPDF3SeqCtrl()->mpCGUI->mpCPanelList->GetTagActivated(); //($B<+J,$,7W;;$KF~$C$A$cBLL\$J$N$GI,MW(B)
	
	// Clipping$B=hM}$rAv::(B
	FilterClipping_cls *pClipping = (FilterClipping_cls*)(GetPDF3SeqCtrl()->GetPFilterFromTag(strTagActivated));
	wxString targetStr = pClipping->GetTargetLabelStr();
	GetPDF3SeqCtrl()->GetClippingInformation(clipScaleMin, clipScaleMax, 
						 targetStr, strTagActivated);
	GLOBJ::AdjustCoordinateMinMax(clipScaleMin, clipScaleMax);

	double boxScale[3];
	LOOP_XYZ
	    boxScale[xyz] = 0.5 * 
	    (DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[xyz+3] -
	     DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[xyz]) *
	    (clipScaleMax[xyz]-clipScaleMin[xyz]);
	
	double posBox[6];
	LOOP_XYZ posBox[xyz] =   mspCSelectedItem->mpInterpolatedValue[xyz];
	LOOP_XYZ posBox[xyz+3] = mspCSelectedItem->mpInterpolatedValue[xyz+3];

	mCDataParentCoordinate.RotTransformToThisCoordinate(dpos);
	LOOP_XYZ mdPosTmp[xyz] = dpos[xyz];

	Parameters_cls::MovePosForBox(posBox, dpos, msSelectedItemScale, mVHitObj[0]); // hitObj$B$KBP1~$7$F0LCV$N0\F0(B
	mpCGLGuide->DrawGuideArrowBox(posBox, scale, boxScale,
				      _GL_GUIDES_WITH_LIMIT);
	glPopMatrix();
    }

    glPopMatrix();
}

void MainGLCanvas_cls::
ConstructDrawingBoxes(std::vector < DrawingBlock_cls > *pVCDrawingCls,
		      PositionAndDirection_cls CCamPos){
    // $B?F(BBlock$B$N@_Dj(B
    mpCParentFrameLink->mpCParent->mpCDBoxMain->
	SetShowParam(mCbShowWire.mbBody, mCbShowBox.mbBody,
		     mpCParentFrameLink->IsPanelActive(wxT("[Box Size]")));
    
    // --- $BJ#?t$N(BBox$B$r:n@.(B ---
    //std::vector < DrawingBlock_cls > VCDrawingCls;
    mpCParentFrameLink->GainDrawingBlockCopies(*pVCDrawingCls);
    
    // distance$B$K1~$8$F%=!<%H$9$k(B
    for (unsigned int i = 0; i < pVCDrawingCls->size(); i++)
	(*pVCDrawingCls)[i].CalcDistance(mCCameraPos.mPos);
    std::sort(pVCDrawingCls->begin(), pVCDrawingCls->end(), 
	      std::greater<DrawingBlock_cls>());

    //$BGX8e$N(BBox$B$K8+$($J$$%U%i%0$r(B
    /*
    PositionAndDirection_cls CCamPos = 
	mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCCurrentCam;
    */
        
    for (unsigned int i = 0; i < pVCDrawingCls->size(); i++)
	(*pVCDrawingCls)[i].CheckVis
	    (CCamPos, 
	     mpCParentFrameLink->mpCPovRay->GetCameraGeometry(),
	     mpCParentFrameLink->mpCPovRay->GetRadianForVisCheck());
}

void MainGLCanvas_cls::
ShowNavigator(PositionAndDirection_cls CCamPos, int width, int height){
    FBO_cls *pCNaviFBO = new FBO_cls(256,256);
    pCNaviFBO->SetGeometryBuffer(256, 256);
    pCNaviFBO->Begin(_FBO_USE_RGBA);
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glClearColor(1, 0, 0, 0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);

    GLPerspective_cls CGLPerspective(256, 256, 15);
    CGLPerspective.SetWindowSize(256, 256);
    PositionAndDirection_cls CNaviCameraPos;
    CNaviCameraPos = CCamPos;//mCCameraPos;
    LOOP_XYZ CNaviCameraPos.mPos[xyz] = -6 * CCamPos.mTgt[xyz];

    CGLPerspective.SetViewportOnGL();
    CGLPerspective.SetPerspectiveOnGL(&CNaviCameraPos);

    CGLPerspective.SetModelOrProjectionViewOnGL
	(&CNaviCameraPos, GL_MODELVIEW);

    glLineWidth(5);

    glTranslated(-0.25, -0.25, -0.25);

    glColor4f(1, 0.1, 0.1, 1);
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0);
    glVertex3d(+1, 0, 0);
    glEnd();
    glPushMatrix();
    glRotatef(90, 0, 1, 0);
    glTranslated(0, 0, 0.8);
    glScaled(0.1, 0.1, 0.2);
    glCallList(displayListConeSimple);
    glPopMatrix();

    glColor4f(0.1, 1.0, 0.1, 1);
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0);
    glVertex3d(0, +1, 0);
    glEnd();
    glPushMatrix();
    glRotatef(-90, 1, 0, 0);
    glTranslated(0, 0, 0.8);
    glScaled(0.1, 0.1, 0.2);
    glCallList(displayListConeSimple);
    glPopMatrix();

    glColor4f(0.1, 0.1, 1, 1);
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0);
    glVertex3d(0, 0, +1);
    glEnd();
    glTranslated(0, 0, 0.8);
    glScaled(0.1, 0.1, 0.2);
    glCallList(displayListConeSimple);

    glPopAttrib();
    pCNaviFBO->End();

    FBO_cls::DrawOnScreen2D(pCNaviFBO, 
			    width, height, 
			    64, 64, 8, 8);

    delete pCNaviFBO;
}

void MainGLCanvas_cls::
ShowInformation(int width, int height, bool bReuse, int InfoDiff,
		std::vector < DrawingBlock_cls > *pVCDrawingCls){
    void* font = GLUT_BITMAP_HELVETICA_12;
    
    if (mCbShowGrid.mbBody == TRUE && mCbShowGridTag.mbBody == TRUE){
    	glColor4f(0.4f, 0.4f, 0.4f, 1.f);
	std::string strTag[2][3] = {{"+x", "+y", "+z"}, {"-x", "-y", "-z"}};
	LOOP_XYZ
	    if (mbShowCoordinateTagXYZ[xyz] == TRUE)
		for (int pm = 0; pm < 2; pm++)
		    GLOBJ::Strings2D::Draw(font, strTag[pm][xyz].c_str(),
					   mPosCoordinateTag[pm][xyz][X] + 1,
					   mPosCoordinateTag[pm][xyz][Y] + 1,
					   _STRINGS_BOTTOM_LEFT);
    }
    
    if (mbSaveImgSeqQueue == TRUE) return;

    if (mCbShowOtherCam.mbBody){
	glColor4f(0.6f, 0.1f, 0.1f, 1.f);
	for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++){
	    wxString strBuf = wxT("cam "); strBuf << i;
	    GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
				   mPosCameraTag[i][X] + 24,
				   mPosCameraTag[i][Y] + 4,
				   _STRINGS_BOTTOM_LEFT);
	}
    }

    glColor4f(1.0f, 1.0f, 1.0f, 1.f);
    
    wxString strBuf;
    if (mCbShowItems.mbBody == TRUE){
	int boxnum = 0, boxnumFileExist = 0;
	DrawingBlock_cls::msVSAlreadyCountedTag.clear();
	for (unsigned int i = 0; i < pVCDrawingCls->size(); i++)
	    boxnum += (*pVCDrawingCls)[i].GetBoxNumber(FALSE);
	DrawingBlock_cls::msVSAlreadyCountedTag.clear();
	for (unsigned int i = 0; i < pVCDrawingCls->size(); i++)
	    boxnumFileExist += (*pVCDrawingCls)[i].GetBoxNumber(TRUE);
    
	if (GetPTreeBoxRule()->mCbCheckFile.mbBody == TRUE &&
	    boxnum != boxnumFileExist)
	    strBuf = wxString::Format
		(wxT("Box=%d %lu MB (%d %lu MB) "),
		 boxnumFileExist, GetPDF3SeqCtrl()->GetCurrentTotalSizeAllChannel() / 1024 * boxnumFileExist / 1024,
		 boxnum, GetPDF3SeqCtrl()->GetCurrentTotalSizeAllChannel() / 1024 * boxnum / 1024);
	else
	    strBuf = wxString::Format
		(wxT("Box=%d %lu MB !! %d %lu"), 
		 boxnum, GetPDF3SeqCtrl()->GetCurrentTotalSizeAllChannel() / 1024  * boxnum / 1024, 
		 GTestVariableInt, GTestVariableLong);

	GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
			       05, 05, _STRINGS_TOP_LEFT);

	strBuf = wxString::Format
	    (wxT("%d x %d %d %d KB" ), width, height, 
	     ((DF3GL_cls*)GetPDF3SeqCtrl()->GetPDF3Active())->m3DTexCompressed, 
	     ((DF3GL_cls*)GetPDF3SeqCtrl()->GetPDF3Active())->m3DTexUsedByte / 1024);
	GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
			       05, 25, _STRINGS_TOP_LEFT);
	
	strBuf = wxString::Format(wxT("%d x %d x %d "), 
				  GetPDF3SeqCtrl()->GetCurrentSizeOf1Data(X),
				  GetPDF3SeqCtrl()->GetCurrentSizeOf1Data(Y),
				  GetPDF3SeqCtrl()->GetCurrentSizeOf1Data(Z));
	
	GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
			       width - (int)strBuf.Len()*6, 05, _STRINGS_TOP_LEFT);

	//int warningOffset = 32;
	glColor4f(1.0f, 0.4f, 0.4f, 1.f);

	if (mspCFilterCtrl->GetVolumeColorType() == _CLR_CUSTOM &&
	    DF3GL_cls::mspCProperty->mCTextureType.mBody != _PREVIEW_RGBA){
	    mstrWarning << wxT("*NOTE* preview with current txture format is pseudo-color. Use RGBA-Mode") << ENDLWX;
	    mWarningLine++;
	    /*
	    GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
				   05, height-warningOffset, _STRINGS_TOP_LEFT);
	    warningOffset += 16;
	    */
	}

	if (mpCParentFrameLink->mpCDF3SeqCtrl->
	    DoesFilterExistInAllDF3(_FILTER_POV_DIFFOBJECT) == TRUE){
	    mstrWarning << wxT("*NOTE* Differential object is experimental. Display of other object breaks down.") << ENDLWX;
	    mWarningLine++;
	    /*
	    GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
				   05, height-warningOffset, _STRINGS_TOP_LEFT);
	    */
	}

	if (mstrWarning != wxT("")){
	    int warningOffset = 32 + mWarningLine * 16;
	    GLOBJ::Strings2D::Draw(font, mstrWarning.ToAscii(),
				   05, height-warningOffset, _STRINGS_TOP_LEFT);
	    mstrWarning = wxT("");
	    mWarningLine = 0;
	}

	wxString strItem;
	if (mspCSelectedItem == NULL) strItem = wxT("");
	else strItem = mspCSelectedItem->mLabelStr;

	if (this == GpCurrentGLPanel)
	    glColor4f(0.3f, 0.8f, 0.3f, 1.0f);
	else
	    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	strBuf = wxT("View ");
	strBuf << mSelfIndex << wxString::Format(wxT(": %ld ms : %ld ms"), mCSW.Time(), mpCParentFrameLink->mCSWMain.Time());
	mCSW.Start();

#ifdef _DEBUG
	if (bReuse == FALSE)
	    strBuf << wxT(" Diff:") << InfoDiff; 
#endif
	
	GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
			       05, height-17, _STRINGS_TOP_LEFT);

	strBuf = wxT("");
	if (this == GpCurrentGLPanel && msCGLBTNTrans.mbMouseOn == TRUE)
	    strBuf << wxT("Translate(G)");
	if (this == GpCurrentGLPanel && msCGLBTNScale.mbMouseOn == TRUE)
	    strBuf << wxT("Scaling(S)");
	if (this == GpCurrentGLPanel && msCGLBTNRot.mbMouseOn == TRUE)
	    strBuf << wxT("Rotation(R)");
	
	glColor4f(0.3f, 0.8f, 0.3f, 1.0f);
	GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
			       width - 160, height-17, _STRINGS_TOP_LEFT);
    }
}

void MainGLCanvas_cls::
ProcessXYZDataIfRequired(){
    for (int df3index = 0; df3index < _DF3_MAX_SLOT; df3index++){
	if (GetPDF3SeqCtrl()->IsReprocessRequired(df3index, _CHECK_VOLUME_ONLY) == TRUE){
	    GetPDF3SeqCtrl()->
		ReleaseStepDF3(df3index, 
			       GetPDF3SeqCtrl()->GetRemappedIndex(df3index),
			       _DF3SEQ_TEXTURE);
	    GetPDF3SeqCtrl()->ClearTreeInfo(-1);
	}
    }

    if (GetPDF3SeqCtrl()->ProcessXYZDataAllChannel() == TRUE)
	mCDrawInfo.mbForceNewDraw = TRUE;
}

bool MainGLCanvas_cls::
IsMouseOnClient(){
    wxRect CScreenRect = GetScreenRect();
    wxPoint CMousePos = wxGetMousePosition();

    return CScreenRect.Contains(CMousePos);
}

bool MainGLCanvas_cls::
CheckFocusOnTherCanvas(){
    for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++){
	if (mVPOtherCanvas[i]->IsMouseOnClient() == TRUE){
	    GpCurrentGLPanel = mVPOtherCanvas[i];
	    return TRUE;
	}
    }
    return FALSE;
}

bool MainGLCanvas_cls::
CheckGLButton(int canvasWidth, int canvasHeight,
	      int mouseX, int mouseY){
    if (msCGLBTNScale.IsMouseOnButton(canvasWidth, canvasHeight,
				      mouseX, mouseY) == TRUE){
	ScaleMode();
	return TRUE; }
    if (msCGLBTNRot.IsMouseOnButton(canvasWidth, canvasHeight,
				      mouseX, mouseY) == TRUE){
	RotationMode();
	return TRUE; }
    if (msCGLBTNTrans.IsMouseOnButton(canvasWidth, canvasHeight,
				      mouseX, mouseY) == TRUE){
	TranslationMode();
	return TRUE; }

    return FALSE;
}

void MainGLCanvas_cls::
OnPaint(wxPaintEvent& event){
    if (IsShownOnScreen() == FALSE) return;
    if (mpCParentFrameLink->mpCPovRay->mbExportingAnim == TRUE) {event.Skip(); return; }
    if (GbGLLockForLoading == TRUE) { event.Skip(); return; }
    if (GbGLLockDrawing == TRUE) {event.Skip(); return; }
    if (FilterCtrl_cls::msOmitUpdateList != 0) { event.Skip(); return; }
    if (GbGLLockForMultithread == TRUE){
	//RGBA$B$N%^%k%A%9%l%C%ICf$J$iBT$D(B
	if (GbGLForceWaitThreadEnd == FALSE) return;
	wxMilliSleep(25);
	Refresh(FALSE);
	return;
    }

    if (mCDisplayMode.mBody == _GL_CANVAS_COPY_OF_VIEW_0)
	mCCameraPos = mpCParentFrameLink->GetPCGLCanvas(0)->mCCameraPos;
    
    GbGLForceWaitThreadEnd = FALSE;
    GbGLLockDrawing = TRUE;

    wxPaintDC dc(this); // must always be here
    
    mpCParentFrameLink->mClockLastDraw = clock();
    
    // Update$B;~$N%^%k%A%9%l%C%I$NM>GH$G(BContext$B$,JQ$o$k2DG=@-$"$j(B 
    // $B%^%k%A%9%l%C%I$,$i$_$N=hM}$r:G=i$K(B
    ProcessXYZDataIfRequired();
    
    bool bReadyForDrawing = FALSE;
    if (mCbShowData.mbBody == FALSE) bReadyForDrawing = TRUE;
    else bReadyForDrawing = GetPDF3SeqCtrl()->CheckUpdate();
    
    if (bReadyForDrawing == FALSE){
	AddPendingEvent(event);
	GbGLLockDrawing = FALSE; 
	return;
    }
    
    PositionAndDirection_cls CCamPos = 
	mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCCurrentCam;
    if (IsFishEye() == TRUE)
	ModifyByDomeTilt(&CCamPos);
    std::vector < DrawingBlock_cls > VCDrawingCls;
    ConstructDrawingBoxes(&VCDrawingCls, CCamPos);
    
    if (SetGLContext() == FALSE){ GbGLLockDrawing = FALSE; return; }
    InitializeGLIfNot();
    
    int width, height, shortRim, longRim;
    GetClientSize(&width, &height);
    if (width > height) {
	shortRim = height; longRim = width;
    }
    else {
	shortRim = width; longRim = height;
    }
    double scale10 = 0.0001, scale = CalcAutoScale();
    while (scale10*10 < scale) scale10 *= 10;

    GLOBJ::Strings2D::SetWindowGeometry(width, height);
    void* font = GLUT_BITMAP_HELVETICA_12;

    if (IsOrtographic() == TRUE)
	mCGLPerspective.mbPerspective = FALSE;
    else
	mCGLPerspective.mbPerspective = TRUE;

    SetAngleIfLinked();

    // Transformationsm ($B%S%e!<%]!<%H$,$"$k$N$G(BFBO$BA0$K(B)
    mpCParentFrameLink->UpdateDrawingData();
    mLightNumber = 
	mpCParentFrameLink->mpCParent->mpCLightSRC->
	GetLightInfo(mALightPos, mALightColor, 
		     mALightAttenuateL, mALightAttenuateQ);
    
    if (mPrevFrame != GetPTimeCtrl()->mFrame) UpdateCameraPos();
    if (mpCLinkedPath != NULL && mCDisplayMode.mBody == _DIRECTION_FREE)
	mCCameraPos.PointAt(msCAnchorPos.mPos);
    PositionAndDirection_cls CCameraPosBase = mCCameraPos;
        
    double doubleIndex = GetPTimeCtrl()->GetDoubleIndex();
    DrawingBlock_cls::SetDefaultClipping();
    
    UpdateDrawInfo(width, height);
    mPrevFrame = GetPTimeCtrl()->mFrame;
    bool bReuse = FALSE;

    int InfoDiff = mCDrawInfo.DiffForDebug(mCDrawInfoPrev);

    if (mCDrawInfo == mCDrawInfoPrev){ 
	if (IsStereo() == FALSE)
	    mCRedrawPBO[0].RedrawPBO(width, height);
	else{
	    switch(GetPLinkedPath()->mpCStereoGramLinked->mCEye.mBody){
	    case _STEREO_RED_CYAN:
		glClear( GL_COLOR_BUFFER_BIT );
		mCRedrawPBO[0].RedrawPBO(width, height, _REDRAW_PBO_ADDITIVE);
		break;
	    case _STEREO_SCANLINE:
		mCRedrawPBO[1].RedrawPBO(width, height);
		mCRedrawPBO[0].RedrawPBO(width, height, _REDRAW_PBO_SCANLINE,
					 mCbStereoFlip.mbBody);
		break;
	    case _STEREO_LFT:
		mCRedrawPBO[0].RedrawPBO(width, height);
		break;
	    case _STEREO_RGT:
		mCRedrawPBO[1].RedrawPBO(width, height);
		break;
	    }
	}

	bReuse = TRUE;
    }
    else{
	//StereoLoop;
	int stereoMax = 0;
	if (IsStereo() == TRUE) stereoMax = 1;

	for (int stereo = 0; stereo <= stereoMax; stereo++){

	    // $BIA2h@h(B
	    FBO_cls *pCFBO[4];
	    for (int i = 0; i < 4; i++) pCFBO[i] = new FBO_cls(width, height);
	    
	    mCCameraPos = CCameraPosBase;
	    if (IsStereo() == TRUE)
		mspCStereogramLink->ModifyByStereogram(&mCCameraPos,
						       _STEREO_LFT + stereo);
	    
	    int fboMax = 1;
	    if (IsFBOUse() == TRUE && 
		(IsFishEye() == TRUE || IsPanoramic() == TRUE)) fboMax = 4;
	    
	    int orgDeg = mCGLPerspective.mDeg;
	    for (int fboIdx = 0; fboIdx < fboMax; fboIdx++){
		
		int lwidth = width, lheight = height;
		if (IsFishEye() == TRUE || IsPanoramic() == TRUE){
		    mCCameraPos = CCameraPosBase;
		    mspCStereogramLink->
			ModifyByStereogram(&mCCameraPos, _STEREO_LFT + stereo);
		    ModifyByDomeTilt(&mCCameraPos);
		    if (IsPanoramic() == TRUE)
			ModifyByPanorama(fboIdx);
		    else
			ModifyByDome(fboIdx);
		    mCGLPerspective.mDeg = 90;
		    mCGLPerspective.SetWindowSize(256, 256);
		}

		SetViewport();
		SetPerspective();
		UpdateWindowSizeInChildClass(); //w x h $B$,>e=q$-$N2DG=@-$G99?7(B
		SetModelView();

		if (IsFBOUse() == TRUE){
		    pCFBO[fboIdx]->
			SetGeometryBuffer(msAFBOSize[mCFBOSizeIndex.mBody], 
					  msAFBOSize[mCFBOSizeIndex.mBody]);
		    pCFBO[fboIdx]->Begin();
		    glPushAttrib(GL_ALL_ATTRIB_BITS);
		}
		
		float lineWidth = msLineWidth;
		if (IsFBOUse() == TRUE)
		    lineWidth *= pCFBO[stereo]->GetBestLineWidth(longRim);
		glLineWidth(lineWidth);
		FilterBase_cls::msBestLineWidth = lineWidth;
		
		glColorMask(1,1,1,1);
		ClearBuffer(); // $B$3$l$O(BColorMask1$B$G$d$i$J$$$H$$$1$J$$(B

		mspCStereogramLink->
		    SetColorMaskBegin(stereo, mCbStereoFlip.mbBody);

		// -- $B<B:]$N%l%s%@%j%s%03+;O(B
		DrawDataMain(scale, scale10, VCDrawingCls);

		if (IsFBOUse() == TRUE){
		    glPopAttrib();
		    pCFBO[fboIdx]->End(/*mCGLPerspective*/);
		}
		
		mspCStereogramLink->SetColorMaskEnd();
	    }

	    //glColorMask(1,1,1,1);
	    
	    mCGLPerspective.mDeg = orgDeg;

	    if (IsFBOUse() == TRUE){
		glClearColor( mCColorBG1.mpBody[0],
			      mCColorBG1.mpBody[1],
			      mCColorBG1.mpBody[2], 1.0f);
		if (IsFishEye() == TRUE || IsPanoramic() == TRUE){
		    if (mbSaveImageQueue == TRUE &&
			mbSaveImgSeqQueue == TRUE)
			FBO_cls::DrawOnScreen(pCFBO, 4, mCGLPerspective,
					      mpCLinkedImgSeq->mClipWidth,
					      mpCLinkedImgSeq->mClipHeight,
					      IsPanoramic());
		    else
			FBO_cls::DrawOnScreen(pCFBO, 4, mCGLPerspective,
					      width, height,
					      IsPanoramic());
		    
		}
		else
		    FBO_cls::DrawOnScreen(pCFBO, 1, mCGLPerspective, 
					  width, height, FALSE);
	    }

	    for (int i = 0; i < 4; i++) delete pCFBO[i];
	    
	    mCRedrawPBO[stereo].SavePBO(width,height);
	}
   
	SetViewport();
	SetPerspective();
	SetModelView();

	if (IsStereo() == TRUE)
	    switch(mspCStereogramLink->mCEye.mBody){
	    case _STEREO_RED_CYAN:
	    case _STEREO_YELLOW_BLUE:
		mspCStereogramLink->SetColorMaskBegin(1, mCbStereoFlip.mbBody);
		mCRedrawPBO[1].RedrawPBO(width, height);
		mspCStereogramLink->SetColorMaskBegin(0, mCbStereoFlip.mbBody);
		mCRedrawPBO[0].RedrawPBO(width, height, _REDRAW_PBO_ADDITIVE);
		mspCStereogramLink->SetColorMaskEnd();
		break;
	    case _STEREO_SCANLINE:
		mCRedrawPBO[0].RedrawPBO(width, height, 
					 _REDRAW_PBO_SCANLINE,
					 mCbStereoFlip.mbBody);
		break;
	    case _STEREO_LFT:
		mCRedrawPBO[0].RedrawPBO(width, height);
		break;
	    }
    }

    // -- $B%J%S%2!<%?(B
    if (mCbShowItems.mbBody == TRUE)
	ShowNavigator(CCameraPosBase, width, height);

    mCCameraPos = CCameraPosBase;

    // FIX_ME!! 
    // $B%"%$%F%`IA2h$O$^$@(BGeometry$B$KBP1~$7$F$$$J$$!#(B180$BEY$H$+:$$k$N$G;CDj(B
    int orgDeg = mCGLPerspective.mDeg;
    if (IsFishEye() == TRUE || IsPanoramic() == TRUE) 
	mCGLPerspective.mDeg = orgDeg - 0.5 * orgDeg * orgDeg / 180;
    SetViewport();
    SetPerspective();
    SetModelView();
    mCGLPerspective.mDeg = orgDeg;

    glLineWidth(1);
    // $B%"%$%F%`I=<((B

    //glDisable(GL_DEPTH_TEST);
    glClearDepth(1.0);
    glClear(GL_DEPTH_BUFFER_BIT);
    glDepthFunc(GL_LEQUAL); 
    DrawItems(scale);
    
    // $BI=<($N%"%s%+!<I=<((B
    DrawAnchor(scale);

    // $BA*Br$5$l$?%"%$%F%`$N%I%i%C%0(B
    if (mspCSelectedItem != NULL) DrawSelectedItem(scale);

    glDisable(GL_BLEND);

    // $B>pJsI=<((B
    ShowInformation(width, height, bReuse, InfoDiff, &VCDrawingCls);
    
    glFlush();
        
    if (mbSaveImageQueue == TRUE){
	glReadBuffer( GL_BACK );
	if (mbSaveImgSeqQueue == TRUE && mpCLinkedImgSeq != NULL){
	    // $BJL%9%l%C%I2=$7$?$N$G8=>u@.H]$N>pJs$r<h$l$J$$(B
	    int currentFrame = GetPTimeCtrl()->mFrame;
	    
	    //int offset = GetPTimeCtrl()->mFrame - GetPTimeCtrl()->mCWAMin.mBody;
	    int offset = GetPTimeCtrl()->GetOffsetForImgSeq();
	    if (IsStereo() == TRUE){
		mCRedrawPBO[0].RedrawPBO(width, height);
		mpCLinkedImgSeq->SaveImageFromGL
		    (&GCImgIO, width, height, offset, _STEREO_LFT);
		mCRedrawPBO[1].RedrawPBO(width, height);
		mpCLinkedImgSeq->SaveImageFromGL
		    (&GCImgIO, width, height, offset, _STEREO_RGT);
	    }
	    else{
		mpCLinkedImgSeq->SaveImageFromGL
		    (&GCImgIO, width, height, offset, _STEREO_NONE);
	    }
	    if (FALSE){ // $B<:GT$7$?$i$N%k!<%A%s$,2?=h$+$KMW$k(B
		mbSaveImgSeqQueue = FALSE;
		mbSaveImageQueue = FALSE;
	    }
	
	    mpCLinkedImgSeq->Increment(offset);
	    
	    mpCParentFrameLink->mCLastOpenedFileIMG.msBody = 
		GCImgIO.mstrLastSavedFile;
	}
	else{
	    mbSaveImageQueue = FALSE; // TRUE$BCf$K(BDLG$B$O$^$:$$$N$G@h$K(B
	    if (GCImgIO.SelectFile
		(mpCParentFrameLink->mCLastOpenedFileIMG.msBody, mpCParentFrameLink) == TRUE){
		mpCParentFrameLink->mCLastOpenedFileIMG.msBody = 
		    GCImgIO.mstrFile;
	    
		// $BJL%9%l%C%I2=$7$?$N$G8=>u@.H]$N>pJs$r<h$l$J$$(B
		if (IsStereo() == TRUE){
		    mCRedrawPBO[0].RedrawPBO(width, height);
		    GCImgIO.SetFilenameStereo(_STEREO_LFT);
		    GCImgIO.SaveImageFromGL(width, height);
		    mCRedrawPBO[1].RedrawPBO(width, height);
		    GCImgIO.SetFilenameStereo(_STEREO_RGT);
		    GCImgIO.SaveImageFromGL(width, height);
		}
		else
		    GCImgIO.SaveImageFromGL(width, height);
	    }
	}
	
    }
    
    if (mbSaveImgSeqQueue == TRUE){
	mpCLinkedImgSeq->RenderFrameBoxInRender(width, height, mCGLPerspective);

	glColor4f(1.0f, 0.4f, 0.4f, 1.0f);
	wxString strBuf = wxT("@ Now Recording...");
	GLOBJ::Strings2D::Draw(font, strBuf.ToAscii(),
			       05, height-34, _STRINGS_TOP_LEFT);
    }

    if (GpCurrentGLPanel == this){
	GLPerspective_cls::Push2DMatrix(width, height);
	//glPushAttrib(GL_TEXTURE_BIT);
	//glEnable(GL_TEXTURE_2D);
	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //$B%"%k%U%!9g@.(B
	msCGLBTNScale.DrawButton(width, msManipulatorMode == _MANIP_SCALE);
	msCGLBTNRot.DrawButton(width, msManipulatorMode == _MANIP_ROTATE);
	msCGLBTNTrans.DrawButton(width, msManipulatorMode == _MANIP_TRANSLATE);
	
	//glPopAttrib();
	GLPerspective_cls::Pop2DMatrix();
    }

    glFlush();

    SwapBuffers();

    mPrevFrame = GetPTimeCtrl()->mFrame;

    event.Skip();

    GbGLLockDrawing = FALSE;
}


void MainGLCanvas_cls::
UpdateAnc(){
    Position_cls CCurrAnc;
    CCurrAnc = ModifyPosAndGetAnc();
    
    if (mpCLinkedPath != NULL)
	mCCameraPos.SetAncPos(mpCLinkedPath->mCurrentAnc);
    else
	mCCameraPos.SetAncPos(CCurrAnc.mPos);

    mCCameraPos.PointAtAnc();
}

void MainGLCanvas_cls::OnSize(wxSizeEvent& event)
{
    UpdateWindowSizeInChildClass();
    mCGLPerspective.SetPerspectiveOnGL(&mCCameraPos);
    if (mpCLinkedImgSeq != NULL){
	int width, height;
	GetClientSize(&width, &height);
	mpCLinkedImgSeq->SetGeom(width, height, mCGLPerspective.mDeg);
    }
    event.Skip();
}

Position_cls MainGLCanvas_cls::
ModifyPosAndGetAnc(){
    if (mCDisplayMode.mBody == _GL_CANVAS_DISPLAY_FREE || 
	mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0 ){
	return msCAnchorPos;
    }
    
    Position_cls CAnc = mCCameraPos;
    switch (mCDisplayMode.mBody){
    case _GL_CANVAS_DISPLAY_X_PLUS:
	CAnc.mPos[X] += +1;
	if (msCAnchorPos.mPos[X] > mCCameraPos.mPos[X]) 
	    CAnc.mPos[X] = msCAnchorPos.mPos[X];
	break;
    case _GL_CANVAS_DISPLAY_X_MINUS:
	CAnc.mPos[X] += -1;
	if (msCAnchorPos.mPos[X] < mCCameraPos.mPos[X])
	    CAnc.mPos[X] = msCAnchorPos.mPos[X];
	break;
    case _GL_CANVAS_DISPLAY_Y_PLUS:
	CAnc.mPos[Y] += +1;
	if (msCAnchorPos.mPos[Y] > mCCameraPos.mPos[Y])
	    CAnc.mPos[Y] = msCAnchorPos.mPos[Y];
	break;
    case _GL_CANVAS_DISPLAY_Y_MINUS:
	CAnc.mPos[Y] += -1;
	if (msCAnchorPos.mPos[Y] < mCCameraPos.mPos[Y])
	    CAnc.mPos[Y] = msCAnchorPos.mPos[Y];
	break;
    case _GL_CANVAS_DISPLAY_Z_PLUS:
	CAnc.mPos[Z] += +1;
	if (msCAnchorPos.mPos[Z] > mCCameraPos.mPos[Z])
	    CAnc.mPos[Z] = msCAnchorPos.mPos[Z];
	break;
    case _GL_CANVAS_DISPLAY_Z_MINUS:
	CAnc.mPos[Z] += -1;
	if (msCAnchorPos.mPos[Z] < mCCameraPos.mPos[Z])
	    CAnc.mPos[Z] = msCAnchorPos.mPos[Z];
	break;
    }
    mCCameraPos.SetDirection(mCDisplayMode.mBody);
    return CAnc;
}

double MainGLCanvas_cls::
CalcAutoScale(){
    double scale = MyVector::GetNormRelative(mCCameraPos.mPos,
					     msCAnchorPos.mPos);
    return scale;
}

bool MainGLCanvas_cls::
GyrationMain(double sensitivity, MouseNormalized_cls MP, wxSize sz){
    PositionAndDirection_cls mCCameraPosBK = mCCameraPos;
    
    if (mspCPreference->mCbGyrationInvert.mbBody == TRUE){
	MP.mPrevX *= -1; MP.mCurrX *= -1;
	MP.mPrevY *= -1; MP.mCurrY *= -1;
    }
    mCCameraPos.SetSensitivity(sensitivity);
    if (mCDisplayMode.mBody == _GL_CANVAS_DISPLAY_FREE ||
	mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	mCCameraPos.TrackBall(MP.mPrevX, MP.mPrevY, MP.mCurrX, MP.mCurrY);
    else{
	sensitivity = sensitivity * mCCameraPos.GetProjectedDistanceToAnc();
	mCCameraPos.SetSensitivity(sensitivity);
	mCCameraPos.SetAspectRatio((double)sz.x / sz.y);
	mCCameraPos.Translate(MP.mPrevX, MP.mPrevY, MP.mCurrX, MP.mCurrY);
    }
    
    if (MP.mGyrationMode == _GYRATION_FIXPOS){
	LOOP_XYZ mCCameraPos.mPos[xyz] = mCCameraPosBK.mPos[xyz];
	if (mCDisplayMode.mBody == _GL_CANVAS_DISPLAY_FREE)
	    mCCameraPos.PointAt(msCAnchorPos.mPos);
	else if (mCDisplayMode.mBody == _GL_CANVAS_LINKED_CAM0) {		
	    mCCameraPos.PointAt(GetPLinkedPath()->mCurrentAnc);
	}
    }
    
    return TRUE;
}

bool MainGLCanvas_cls::
ZoomMain(double sensitivity, MouseNormalized_cls MP, wxSize sz){
    if (mspCPreference->mCbZoomInvert.mbBody == TRUE){
	MP.mPrevY *=-1 ; MP.mCurrY *=-1 ;
    }
    mCCameraPos.SetSensitivity(1);
    mCCameraPos.Zoom(MP.mPrevX, MP.mPrevY, MP.mCurrX, MP.mCurrY);
    return TRUE;
}

bool MainGLCanvas_cls::
ZoomWheel(float delta){
    UpdateAnc();
    mCCameraPos.SetSensitivity(1);
    if (mspCPreference->mCbZoomInvert.mbBody == TRUE) delta *= -1;
    mCCameraPos.Zoom(0, 0, 0, delta);
    return TRUE;
}

bool MainGLCanvas_cls::
MoveCenter(double sensitivity, MouseNormalized_cls MP, wxSize sz,
	   bool *trigAnchor){
    PositionAndDirection_cls CAnc(msCAnchorPos, mCCameraPos);
    sensitivity = sensitivity * mCCameraPos.GetProjectedDistanceToAnc();
    mCCameraPos.SetSensitivity(sensitivity);
    mCCameraPos.SetAspectRatio((double)sz.x / sz.y);
    if (mpCLinkedPath == NULL)
	mCCameraPos.Translate(MP.mPrevX, MP.mPrevY, MP.mCurrX, MP.mCurrY);
    else sensitivity *= -1;
    CAnc       .SetAncPos(mCCameraPos.mPos);
    CAnc       .SetSensitivity(sensitivity);
    CAnc       .SetAspectRatio((double)sz.x / sz.y);
    CAnc       .Translate(MP.mPrevX, MP.mPrevY, MP.mCurrX, MP.mCurrY);
    msCAnchorPos.Set(CAnc);
    
    *trigAnchor = TRUE;
    return TRUE;
}

bool MainGLCanvas_cls::
MoveItem(double sensitivity, MouseNormalized_cls MP, wxSize sz,
	 bool *trigAnchor){
    double doubleIndex = GetPTimeCtrl()->GetDoubleIndex();

    if (mbMouseDragging == FALSE){
	if (mspCSelectedItem->mLabelStr == L"BoxRotate")
	    LOOP_XYZ msSelectedItemPos[xyz] = 
		DrawingBlock_cls::mspCBoxSize->GetPos(doubleIndex, xyz);
	else if (mspCSelectedItem->mLabelStr == L"CamRotYPR")
	    LOOP_XYZ msSelectedItemPos[xyz] = 
		mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCPosition.GetPos(doubleIndex, xyz);
	else
	    LOOP_XYZ msSelectedItemPos[xyz] =
		mspCSelectedItem->GetPos(doubleIndex, xyz);
	LOOP_XYZ msSelectedItemPosOrg[xyz] = msSelectedItemPos[xyz];
	LOOP_XYZ msSelectedItemScale[xyz] = 1.0;
	LOOP_YPRY msSelectedItemDYPRY[ypry] = 0;
	mbItemDragging = TRUE;
    }

    gluProject(msSelectedItemPos[X], msSelectedItemPos[Y], msSelectedItemPos[Z],
	       mMatrixModelView, mMatrixProjection, mMatrixViewport,
	       &mPosItemOnW[0], &mPosItemOnW[1], &mPosItemOnW[2]);
    double normalPosX = (2*mPosItemOnW[X] - sz.x)/sz.x;
    double normalPosY = (2*mPosItemOnW[Y] - sz.y)/sz.y;
    double dragX = MP.mCurrX - MP.mPrevX, dragY = MP.mCurrY - MP.mPrevY;
    double relX = MP.mCurrX - normalPosX, relY = MP.mCurrY - normalPosY;

    switch (msManipulatorMode){
    case _MANIP_ROTATE: {
	double drag = (dragX * relY - dragY * relX) / sqrt (relX * relX + relY * relY);
	if (mspCSelectedItem->mLabelStr == L"CamPos" &&
	    mpCGLGuideScrew->mbDirectionInv == TRUE) drag *= -1;
	else if (mspCSelectedItem->mLabelStr == L"CamRotYPR" &&
		 mpCGLGuideYPR->mbDirectionInv == TRUE) drag *= -1;
	else if (mpCGLGuide->mbDirectionInv == TRUE) drag *= -1;
	if (msMoveMode <= _ROTATE_MOVE_Y2)
	    msSelectedItemDYPRY[msMoveMode] += drag * 100;
	break; }
    case _MANIP_SCALE: {
	double drag = (dragX * relX + dragY * relY) / sqrt (relX * relX + relY * relY);
	double scale[3];
	LOOP_XYZ scale[xyz] = 1 + drag * PositionAndDirection_cls::msbMatrixMove[msMoveMode][xyz];
	LOOP_XYZ msSelectedItemScale[xyz] *= scale[xyz];
	break; }
    case _MANIP_TRANSLATE:
	Position_cls CPosTmp(msSelectedItemPos[0],
			     msSelectedItemPos[1],
			     msSelectedItemPos[2], 0, 0, 0);
	PositionAndDirection_cls CPos(CPosTmp, mCCameraPos);
	PositionAndDirection_cls CParentCoord;

	if (mspCSelectedItem->mLabelStr.Matches(L"ClippingBox") || 
	    mspCSelectedItem->mLabelStr.Matches(L"BoxSize") ||
	    mspCSelectedItem->mLabelStr.Matches(L"DensityCurvePoints")){
	    LOOP_XYZ CPos.mATranslateScale[xyz] = 0.5 
		/ DrawingBlock_cls::GetRootBoxScale(xyz);
	    CParentCoord = mCDataParentCoordinate;
	}
	else {
	    LOOP_XYZ CPos.mATranslateScale[xyz] = 1;
	}
	
	sensitivity = sensitivity * mCCameraPos.GetProjectedDistanceToAnc();
	CPos.SetAncPos(mCCameraPos.mPos);
	CPos.SetSensitivity(sensitivity);
	CPos.SetAspectRatio((double)sz.x / sz.y);
	CPos.Translate(MP.mPrevX, MP.mPrevY, MP.mCurrX, MP.mCurrY, msMoveMode, &CParentCoord);
	LOOP_XYZ msSelectedItemPos[xyz] = CPos.mPos[xyz];
	break;
    }
    
    *trigAnchor = TRUE;
    
    return TRUE;
}

bool MainGLCanvas_cls::
ShowMenu(/*bool bAMouseBTN[4], bool bAModifier[3],*/ int x, int y){
    /*
    if (bAMouseBTN[mspCPreference->mCMenuMouse.mBody] == FALSE) return FALSE;
    for (unsigned int i = 0; i < 3; i++)
	if (bAModifier[i] != mspCPreference->mCbMenuMod.mpbBody[i]) 
	    return FALSE;
    */

    ShowMenuPanel(0, x, y);
    
    if (HasCapture() == TRUE) ReleaseMouse();

    return TRUE;
}

bool MainGLCanvas_cls::
ProcViewOrItemDragging(/*wxMouseEvent& event, wxSize sz,
		       bool bAModifier[4], bool bAMouseBTN[4],
		       */bool& trigAnchor, wxSize sz){

    bool trigRefresh = FALSE;

    MouseNormalized_cls CMP(mpCMouseCtrl, sz);
    CMP.LimitMove();
    
    UpdateAnc();
    
    double sensitivity = 
	(mCDisplayMode.mBody == _GL_CANVAS_DISPLAY_FREE || 
	 mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0) ? 1 : 0.5;
    
    // Clipping $B$N$H$-$N%9%1!<%k(B

    if (mpCMouseCtrl->IsGyration())
	trigRefresh |= GyrationMain(sensitivity, CMP, sz); 
    if (mpCMouseCtrl->IsZoom())
	trigRefresh |= ZoomMain(sensitivity, CMP, sz);
    if (mpCMouseCtrl->IsMoveCenter())
	trigRefresh |= MoveCenter(0.5, CMP, sz, &trigAnchor);
    if (mspCSelectedItem != NULL && 
	mpCMouseCtrl->IsMoveItem())
	trigRefresh |= MoveItem(-0.5, CMP, sz, &trigAnchor);
    
    GpGLPanelJustHandled = this;
    
    mpCMouseCtrl->SetPreviousPosition();
    
    mbMouseDragging = TRUE; // $B:G=i$N#12s$O(B FALSE $B$GH=JL(B

    return trigRefresh;
}

void MainGLCanvas_cls::
ProcSelectItemForDragging(){
    int mouseX = mpCMouseCtrl->mMouseX;
    int mouseY = mpCMouseCtrl->mMouseY;

    mpCGLGuide->InitializeProjection(); //$B%R%C%H%F%9%H3+;O(B
		
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    
    SetViewport();
    gluPickMatrix(mouseX, viewport[3] - mouseY+1, 15., 15., viewport);
    SetPerspective();
    
    SetModelView();
    double pos[3] = {0, 0, 0}; 
    double scale = CalcAutoScale();
    DrawSelectedItem(scale);
    //$B%F%9%HMQ%+%a%i%Q%9IA2h(B
    
    //mpCParentFrameLink->mpCPathCtrl->GetActivePathGLGUI()->DrawPath();  // $B$3$l$OL5$/$FJ?5$!)(B
    mVHitObj = mpCGLGuide->GetSelection(); // $B%R%C%H%F%9%H=*N;(B
    
    if (mVHitObj[1] == 1) msMoveMode = _POS_AND_DIR_MOVE_X;
    if (mVHitObj[1] == 2) msMoveMode = _POS_AND_DIR_MOVE_Y;
    if (mVHitObj[1] == 3) msMoveMode = _POS_AND_DIR_MOVE_Z;
    if (mVHitObj[1] == 4) msMoveMode = _ROTATE_MOVE_Y2;
}

bool MainGLCanvas_cls::
ProcSelectItemForSelection(){

    int mouseX = mpCMouseCtrl->mMouseX;
    int mouseY = mpCMouseCtrl->mMouseY;

    mpCGLGuide->InitializeProjection(); //$B%R%C%H%F%9%H3+;O(B
		
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    
    SetViewport();
    gluPickMatrix(mouseX, viewport[3] - mouseY + 1, 15., 15., viewport);
    SetPerspective();
    
    SetModelView();
    double pos[3] = {0, 0, 0}; 
    double scale = CalcAutoScale();
    DrawItems(scale);
    //$B%F%9%HMQ%+%a%i%Q%9IA2h(B
    
    //mpCParentFrameLink->mpCPathCtrl->GetActivePathGLGUI()->DrawPath();  // $B$3$l$OL5$/$FJ?5$!)(B
    mVHitObj = mpCGLGuide->GetSelection(); // $B%R%C%H%F%9%H=*N;(B
    
    if (mVHitObj.size() > 1){
	wxString strInfo;
	switch(mVHitObj[0]){
	case _HITTEST_LIGHT:
	    mpCParentFrameLink->mpCParent->mpCLightSRC->mpCGUI->mActiveItem = mVHitObj[1];
	    mpCParentFrameLink->mpCParent->mpCLightSRC->UpdateActiveOnTL();
	    strInfo << mVHitObj.size() << wxT(" Hit Light ") << mVHitObj[1];
	    break;
	case _HITTEST_CAMERA:
	    mpCParentFrameLink->mpCDF3SeqCtrl->mpCGUI->ActivatePanel(wxT("[Cam XYZ]"), TRUE);
	    strInfo << mVHitObj.size() << wxT(" Hit Camera ") << mVHitObj[1];
	    break;
	case _HITTEST_ANCHOR:
	    mpCParentFrameLink->mpCDF3SeqCtrl->mpCGUI->ActivatePanel(wxT("[Anc XYZ]"), TRUE);
	    strInfo << mVHitObj.size() << wxT(" Hit Anchor ") << mVHitObj[1];
	    break;
	default:
	    // $B$s!)2?$K$b%R%C%H$7$J$/$F$b6u$8$c$J$$!)(B
	    mpCParentFrameLink->mpCDF3SeqCtrl->mpCGUI->DeactivateAllPanel(FALSE);
	    strInfo << mVHitObj.size() << wxT(" Nothing ") << mVHitObj[1];
	    break;
	}
	mpCParentFrameLink->PushStatusTextMain(strInfo, 2);
	return TRUE;
    }
    else
	// $B$3$C$A$K$OMh$J$$!)(B
	mpCParentFrameLink->mpCDF3SeqCtrl->mpCGUI->DeactivateAllPanel(FALSE);
    return FALSE;
}

bool MainGLCanvas_cls::
ProcCaptureMouse(){
    bool trigRefresh = FALSE;
    
    CaptureMouse();
    
    for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++)
	if (mVPOtherCanvas[i]->HasCapture() == TRUE)
	    mVPOtherCanvas[i]->ReleaseMouse();
    
    trigRefresh |= SetThisToBeActiveGL();
    mpCParentFrameLink->ReconnectCameraPathGUI();
    
    return trigRefresh;
}

void MainGLCanvas_cls::
StartHitTest(){
    int bHitting = FALSE;
    // $B%I%i%C%0MQ$N%R%C%H%F%9%H(B
    if (mpCMouseCtrl->IsMoveItem()) ProcSelectItemForDragging();
    else
	if (msManipulatorMode != _MANIP_ROTATE) 
	    msMoveMode = _POS_AND_DIR_MOVE_FREE;
    
    if (mpCMouseCtrl->IsHitTest()) bHitting = ProcSelectItemForSelection();
    
    if (bHitting == TRUE) ReleaseMouse();
}

void MainGLCanvas_cls::
RefreshEX(){
    wxPaintEvent com(GetId());
    AddPendingEvent(com);
    if (mSelfIndex == 0){
	MainGLCanvas_cls *pCExt = mpCParentFrameLink->GetPCGLCanvas(4);
	if (pCExt->mCDisplayMode.mBody != _GL_CANVAS_COPY_OF_VIEW_0) return;
	wxPaintEvent com(pCExt->GetId());
	pCExt->AddPendingEvent(com);
    }
}

void MainGLCanvas_cls::
OnMouse(wxMouseEvent& event){
    
    bool trigRefresh = FALSE;
    static bool trigAnchor = FALSE;
    static bool sbLockDragging = FALSE;

    wxSize sz(GetClientSize()); 
    int width = sz.GetWidth(), height = sz.GetHeight();

    mpCMouseCtrl->SetMouseInformation(event);
    int mouseX = event.GetX(), mouseY = event.GetY();
    
    // $B%a%K%e!<(B
    if (mpCMouseCtrl->IsShowMenu() == TRUE){
	ShowMenu(mpCMouseCtrl->mMouseX, mpCMouseCtrl->mMouseY);
	event.Skip();
	return; }
	
    // $B2hLL>e$N%\%?%s(B
    if (mpCMouseCtrl->IsButton(_MOUSE_BTN_LEFT) == TRUE &&
	CheckGLButton(width, height, 
		      mpCMouseCtrl->mMouseX, mpCMouseCtrl->mMouseY) == TRUE){
	RefreshEX();
	sbLockDragging = TRUE;
	event.Skip();
	return; }
    
    // $B%^%&%9$rN%$9(B
    if (mpCMouseCtrl->IsButtonUp() == TRUE){
	sbLockDragging = FALSE;
	if (HasCapture()) ReleaseMouse();
    }
        
    // $B%S%e!<$^$?$O%"%$%F%`$N%I%i%C%0(B
    if (mpCMouseCtrl->IsMouseDragging() == TRUE && sbLockDragging == FALSE)
	trigRefresh |= ProcViewOrItemDragging(trigAnchor, sz);
    else
	if (HasCapture() == FALSE)
	    mbMouseDragging = FALSE;
    
    bool bHitting = FALSE;
    if (event.ButtonDown(wxMOUSE_BTN_ANY)){
	SetFocus();
	if (HasCapture() == FALSE){
	    trigRefresh |= ProcCaptureMouse();
	    trigRefresh = TRUE;
	    StartHitTest();
	    mpCMouseCtrl->SetPreviousPosition();
	    mpCMouseCtrl->SetGyrationMode(sz);
	}
    }

    if (event.GetEventType() == wxEVT_MOUSEWHEEL){
	if (IsMouseOnClient() == TRUE){
	    trigRefresh |= ZoomWheel((float)(event.m_wheelRotation)/
				     event.GetWheelDelta() * 0.1);
#ifndef WIN32
	    SetFocus();
	    if (GpCurrentGLPanel != this){
		GpCurrentGLPanel->Refresh(TRUE);
		GpCurrentGLPanel = this;
	    }
#endif
	}
	else
	    if (CheckFocusOnTherCanvas() == TRUE){
		Refresh(FALSE);
		event.Skip();
		GpCurrentGLPanel->SetFocus();
		GpCurrentGLPanel->Refresh(FALSE);
		return;
	    }
    }

    if (event.ButtonUp(wxMOUSE_BTN_LEFT) ||
	event.ButtonUp(wxMOUSE_BTN_MIDDLE) ||
	event.GetEventType() == wxEVT_MOUSEWHEEL) {
	//msMoveMode = _POS_AND_DIR_MOVE_FREE;
	if (mbItemDragging == TRUE && mspCSelectedItem != NULL){
	    double dpos[3];
	    LOOP_XYZ dpos[xyz] = 
		msSelectedItemPos[xyz] - msSelectedItemPosOrg[xyz];
	    
	    // local$B:BI8$N>l9g$b$"$k$N$G(Bdpos$B$N:BI8JQ49(B
	    if (mspCSelectedItem->mLabelStr == L"Light"){
		int selected = mpCParentFrameLink->mpCParent->mpCLightSRC->GetSelectedIndex();
		if (selected != -1)
		    mpCParentFrameLink->mpCParent->mpCLightSRC->
			TransformCoordinate(selected, dpos);
	    }
	    
	    if (mspCSelectedItem->mLabelStr.Matches(L"BoxSize") ||
		mspCSelectedItem->mLabelStr.Matches(L"ClippingBox")){
		
		// !! $B2sE>$K$"$o$;$?:BI8JQ49(B !!
		mCDataParentCoordinate.RotTransformToThisCoordinate(dpos);
	    }
	    
	    GCUndoList.OpenBlock();
	    
	    // $B%-!<$N%;%l%/%7%g%sL5$7(B
	    if (mspCSelectedItem->GetSelectedKeyNum() == 0){
		mspCSelectedItem->
		    GetPInterpolatedValue(GetPTimeCtrl()->GetDoubleIndex());
		mspCSelectedItem->MoveInterpolatedPos(dpos, msSelectedItemScale, msSelectedItemDYPRY, mVHitObj[0]);

		if (mspCSelectedItem->mLabelStr == L"CamPos" &&
		    msSelectedItemDYPRY[0] != 0){
		    // TOP$B$N2sE>$r(B
		    double axis[3];
		    LOOP_XYZ axis[xyz] = 
			mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCurrentAnc[xyz] - 
			mpCParentFrameLink->mpCPathCtrl->GetActivePath()->mCCurrentCam.mPos[xyz];
		    MyVector::Normalize(axis, 1);
		    
		    MyVector::RotateAround
			(mspCSelectedItem->mpCSlaveParam->mpInterpolatedValue,
			 axis, msSelectedItemDYPRY[0]);

		    mspCSelectedItem->mpCSlaveParam->SetKeyIfChanged();
		    mspCSelectedItem->SetKeyIfChanged(_PARAM_CURRENT_DINDEX, TRUE);
		    trigAnchor = TRUE;
		    trigRefresh = TRUE;
		    GetPDF3SeqCtrl()->mpCGUI->UpdateDisplay();
		    mpCParentFrameLink->UpdateDrawingData();
		}
		
		if (mspCSelectedItem->SetKeyIfChanged() == TRUE){
		    if (mspCSelectedItem->mLabelStr == L"CamPos"){
			mpCParentFrameLink->mpCPathCtrl->GetActivePath()->ApplySky();
			mspCSelectedItem->mpCSlaveParam->SetKeyIfChanged();
		    }
		    
		    trigAnchor = TRUE;
		    trigRefresh = TRUE;
		    GetPDF3SeqCtrl()->mpCGUI->UpdateDisplay();
		    mpCParentFrameLink->UpdateDrawingData();
		}
	    }
	    // $B%-!<$N%;%l%/%7%g%s$"$j(B
	    else{
		if (mspCSelectedItem->mLabelStr == L"CamPos" &&
		    msSelectedItemDYPRY[0] != 0){
		    mpCParentFrameLink->mpCPathCtrl->GetActivePath()->
			RotateAroundSelectedKey(msSelectedItemDYPRY[0]);
		}
		else
		    mspCSelectedItem->MovePosSelectedKey(dpos, msSelectedItemScale, mVHitObj[0]);
		if (mspCSelectedItem->mLabelStr == L"CamPos")
		    mpCParentFrameLink->mpCPathCtrl->GetActivePath()->ApplySkyToSelectedKey();
		
		mpCParentFrameLink->UpdateDrawingData();
		trigRefresh = TRUE;
	    }

	    if (mspCSelectedItem->mLabelStr == L"BoxSize" ||
		mspCSelectedItem->mLabelStr == L"BoxRotate"){
		wxCommandEvent
		    com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshPanels);
		mpCParentFrameLink->AddPendingEvent(com);
	    }

	    if (mspCSelectedItem->mLabelStr.Matches(L"PovObject*") ||
		mspCSelectedItem->mLabelStr.Matches(L"PovDiffObject*")){
		mCDrawInfo.ForceNewDraw();
		trigRefresh = TRUE;
	    }

	    if (mspCSelectedItem->mLabelStr == L"CamPos"){
		if (mspCSelectedItem == &(mpCLinkedPath->mCPosition))
		    SetDisplayMode(_DIRECTION_FREE, TRUE);
		mpCParentFrameLink->NormalizeActiveCameraPath();
	    }

	    GCUndoList.CloseBlock();

	    // $BLrL\=*N;(B
	    LOOP_XYZ msSelectedItemPosOrg[xyz] = msSelectedItemPos[xyz];
	    LOOP_XYZ msSelectedItemScale[xyz] = 1;
	    LOOP_YPRY msSelectedItemDYPRY[ypry] = 0;
	    	
	    wxCommandEvent 
		com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshCamera);
	    mpCParentFrameLink->AddPendingEvent(com);
	    mbItemDragging = FALSE;
	}

	// $BB>$N%&%#%s%I%&$N99?7(B
	// $B%H%j%,!<$H$7$F%@%_!<$N(BwxChildFocusEvent$B$d(BwxSizeEvent $B$r%]%9%H(B
	PendEventLists();
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshCamera);
	mpCParentFrameLink->AddPendingEvent(com);
	 
	if (HasCapture() == TRUE)
	    ReleaseMouse();

	for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++)
	    if (trigAnchor == TRUE || mVPOtherCanvas[i]->mCbShowOtherCam.mbBody == TRUE){
		wxPaintEvent dummy;
		mVPOtherCanvas[i]->UpdateCameraPos();
		if (trigRefresh)
		    mVPOtherCanvas[i]->mCDrawInfo.ForceNewDraw();
		mVPOtherCanvas[i]->AddPendingEvent(dummy);
	    }
	trigAnchor = FALSE;
    }

    if (trigRefresh == TRUE) {
	
	if (mCDisplayMode.mBody == _DIRECTION_FREE || 
	    mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	    mpCParentFrameLink->mpCPathCtrl->GetActivePath()->ApplySky();
	
	RefreshEX();

	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemPopStatusBar);
	mpCParentFrameLink->AddPendingEvent(com);
    }

    if (event.GetEventType() == wxEVT_MOUSEWHEEL) return;

    if (HasCapture() == FALSE)
	event.Skip();
    else
	/*event.Skip()*/;
}

void MainGLCanvas_cls::
OnMouseRelease(wxMouseCaptureLostEvent& event){
    if (HasCapture() == TRUE)
	ReleaseMouse();
    event.Skip();
}

void MainGLCanvas_cls::
OnMouseMotion(wxMouseEvent& event){

    if (GpCurrentGLPanel == this){
	int width, height;
	GetClientSize(&width, &height);
    
	bool bRefresh = FALSE;
	int mouseX = event.GetX(), mouseY = event.GetY();
	
	if (msCGLBTNScale.mbMouseOn != 
	    msCGLBTNScale.IsMouseOnButton(width, height, mouseX, mouseY))
	    RefreshEX();
	if (msCGLBTNRot.mbMouseOn != 
	    msCGLBTNRot.IsMouseOnButton(width, height, mouseX, mouseY))
	    RefreshEX();
	if (msCGLBTNTrans.mbMouseOn != 
	    msCGLBTNTrans.IsMouseOnButton(width, height, mouseX, mouseY))
	    RefreshEX();
    }

    event.Skip();
}

void MainGLCanvas_cls::
ProcShortcutKey(int modifier, int code){
    if (modifier == wxMOD_ALT){
	if (code == WXK_RETURN){
	    mpCGLCtrl->GetSizer()->Hide(mpCGLCtrl->mpBoxSizerBar);
	    ((wxFrame*)wxGetTopLevelParent(this))->ShowFullScreen(TRUE);
	    
	}
    }

    if (code == WXK_ESCAPE){
	mpCGLCtrl->GetSizer()->Show(mpCGLCtrl->mpBoxSizerBar);
	((wxFrame*)wxGetTopLevelParent(this))->ShowFullScreen(FALSE);
    }
}

void MainGLCanvas_cls::
OnKeyDown(wxKeyEvent& event){

    if (mSelfIndex == 4)
	ProcShortcutKey(event.GetModifiers(), event.GetKeyCode());

    event.SetId(mpCParentFrameLink->GetId());
    mpCParentFrameLink->AddPendingEvent(event);
}

void MainGLCanvas_cls::
RotationMode(){
    msManipulatorMode = _MANIP_ROTATE;
    msMoveMode = _POS_AND_DIR_MOVE_X;
    if (mspCSelectedItem == NULL) return;
    if (mspCSelectedItem->mLabelStr == L"BoxSize"){
	SelectItem(DrawingBlock_cls::mspCRotate);
	mpCParentFrameLink->mpCDF3SeqCtrl->mpCGUI->ActivatePanel(wxT("[Box Rotate]"), TRUE);
    }
}

void MainGLCanvas_cls::
TranslationMode(){
    msManipulatorMode = _MANIP_TRANSLATE;
    msMoveMode = _POS_AND_DIR_MOVE_FREE;
    if (mspCSelectedItem == NULL) return;
    if (mspCSelectedItem->mLabelStr == L"BoxRotate"){
	SelectItem(DrawingBlock_cls::mspCBoxSize);
	mpCParentFrameLink->mpCDF3SeqCtrl->mpCGUI->ActivatePanel(wxT("[Box Size]"), TRUE);
    }
}

void MainGLCanvas_cls::
ScaleMode(){
    msManipulatorMode = _MANIP_SCALE;
    msMoveMode = _POS_AND_DIR_MOVE_FREE;
    if (mspCSelectedItem == NULL) return;
    if (mspCSelectedItem->mLabelStr == L"BoxRotate"){
	SelectItem(DrawingBlock_cls::mspCBoxSize);
	mpCParentFrameLink->mpCDF3SeqCtrl->mpCGUI->ActivatePanel(wxT("[Box Size]"), TRUE);
    }
}

void MainGLCanvas_cls::
ShowMenuPanel(int type, int x, int y){
    wxMenu* pMenu = NULL;
    switch(type){
    case 0: pMenu = mpPopupMenu;     break;
    }

    // LINUX::$B$I$&$b%a%K%e!<I=<($r<:GT$9$k>l9g$,$"$C$F!"(B
    // $B$=$NJU$,IT0BDj$K$J$k860x$N0l$D$C$]$$$,!#$I$&$bITL@(B
    if (pMenu != NULL) {
	mpMenuItemShowPanel->Enable((mpCGLCtrl->IsPanelShown() == FALSE));
#ifdef WIN32
	bool bStatus = PopupMenu(pMenu, x, y);
#else
	bool bStatus = PopupMenu(pMenu);
#endif
	//printf ("popup!! status=%d\n", bStatus);
    }
}

bool MainGLCanvas_cls::
SetThisToBeActiveGL(){
    if (GpCurrentGLPanel != this) {
	GpCurrentGLPanel->Refresh(FALSE); 
	GpCurrentGLPanel = this;
	return TRUE;
    }
    return FALSE;
}

void MainGLCanvas_cls::
RegistPOtherCanvas(MainGLCanvas_cls* pGLCanvas){
    mVPOtherCanvas.push_back(pGLCanvas);
}

void MainGLCanvas_cls::
InitializeStaticMembers(FrameMain_cls *pCFrame, DF3SeqCtrl_cls *pCDF3SeqCtrl, FilterCtrl_cls *pCFilterCtrl){
    mspCDF3SeqCtrl = pCDF3SeqCtrl;
    mspCFilterCtrl = pCFilterCtrl;
    mspCPreference = new Preference_cls(pCFrame);
}

bool MainGLCanvas_cls::
IsGLBlendEquationEXOK(){
    return (glBlendEquationEXT != NULL);
}

void MainGLCanvas_cls::
DeleteInstancePreference(){
    delete mspCPreference;
}

void MainGLCanvas_cls::
OpenPreference(wxWindow *pC){
    mspCPreference->OpenWindow(pC);
    mspCPreference->mpCGUI->GrayOutItems();
}

void MainGLCanvas_cls::
SetDefaultPosition(int displayMode){
    if (displayMode == _GL_CANVAS_DISPLAY_FREE) return;
    double distance = mCCameraPos.GetDistanceToAnc();
    LOOP_XYZ mCCameraPos.mPos[xyz] = distance *
	GLPerspective_cls::sDefaultCameraPos[displayMode][xyz];
}

void MainGLCanvas_cls::
TogglePerspective(){
    if (mCGLGeometry.mBody == _POV_PERSPECTIVE)
	mCGLGeometry.mBody = _POV_ORTHOGRAPHIC;
    else
	mCGLGeometry.mBody = _POV_PERSPECTIVE;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshCamera);
    GetParent()->AddPendingEvent(com);
}

void MainGLCanvas_cls::
SetDisplayMode(int displayMode, bool bUpdateCombobox){
    if (mCDisplayMode.mBody == _GL_CANVAS_DISPLAY_FREE &&
	displayMode != _GL_CANVAS_DISPLAY_FREE)
	mCCameraPosCached = mCCameraPos;
    if (mCDisplayMode.mBody != _GL_CANVAS_DISPLAY_FREE &&
	mCDisplayMode.mBody != _GL_CANVAS_LINKED_CAM0 &&
	displayMode == _GL_CANVAS_DISPLAY_FREE)
	mCCameraPos = mCCameraPosCached;

    mCDisplayMode.mBody = displayMode;
    mpCLinkedPath = NULL;

    if (displayMode == _GL_CANVAS_DISPLAY_FREE && 
	mCGLGeometry.mBody == _POV_ORTHOGRAPHIC)
	mCGLGeometry.mBody = _POV_PERSPECTIVE;
    
    if (displayMode > _GL_CANVAS_DISPLAY_FREE &&
	displayMode < _GL_CANVAS_LINKED_CAM0)
	mCGLGeometry.mBody = _POV_ORTHOGRAPHIC;

    /*
    if (mCGLGeometry.mBody == _POV_PERSPECTIVE)
	if (mCGLPerspective.mDeg > 170) mCGLPerspective.mDeg = 170;
    */
    
    SetDefaultPosition(mCDisplayMode.mBody);
    ModifyPosAndGetAnc();
    PendEventLists();

    if (mpCParentFrameLink != NULL){
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshCamera);
	mpCParentFrameLink->AddPendingEvent(com);
    }

    if (bUpdateCombobox == TRUE)
	mpCGLCtrl->UpdateWindow();
    Refresh(false);
}

void MainGLCanvas_cls::
LinkageConnect(int pathIndex){
    if (pathIndex == -1) pathIndex = mCPathIndex.mBody ;
    else mCPathIndex.mBody = pathIndex;
    mpCLinkedPath = mpCParentFrameLink->mpCPathCtrl->mVpCPath[pathIndex];
    mCGLPerspective.mbPerspective = TRUE;
    mCDisplayMode.mBody = _GL_CANVAS_LINKED_CAM0 + pathIndex;
    mpCGLCtrl->UpdateWindow();
    UpdateCameraPos();
    //ModifyByStereogram();
    mspCStereogramLink->ModifyByStereogram(&mCCameraPos, -1);
}

void MainGLCanvas_cls::
OnMenuPassToParent(wxCommandEvent& event){
    wxGetTopLevelParent(this)->AddPendingEvent(event);
}

void MainGLCanvas_cls::
OnMenuAddPrimitive(wxCommandEvent& event){
    mpCParentFrameLink->
	PopupMenu(mpCParentFrameLink->mpCMenuFilterPrimitive, wxDefaultPosition);
}

void MainGLCanvas_cls::
OnMenuShowPanel(wxCommandEvent& event){
    mpCGLCtrl->ShowPanel();
}

void MainGLCanvas_cls::
OnMenuMoveToCamPos(wxCommandEvent& WXUNUSED(event)){
    LinkageConnect(mpCParentFrameLink->mpCPathCtrl->mActivePath);
    wxGetTopLevelParent(this)->Refresh(false);
}

void MainGLCanvas_cls::
OnMenuSetKeyPos(wxCommandEvent& event){
    wxCommandEvent dummy(wxEVT_COMMAND_MENU_SELECTED, menuItemSetCamPos);
    wxGetTopLevelParent(this)->AddPendingEvent(dummy);
}

void MainGLCanvas_cls::
OnMenuSetKeyAnc(wxCommandEvent& event){
    wxCommandEvent dummy(wxEVT_COMMAND_MENU_SELECTED, menuItemSetAncPos);
    wxGetTopLevelParent(this)->AddPendingEvent(dummy);
}

void MainGLCanvas_cls::
OnMenuToggleLink(wxCommandEvent& event){
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	SetDisplayMode(_DIRECTION_FREE);
    else
	LinkageConnect(-1);
    GetParent()->Refresh(false);
}

void MainGLCanvas_cls::
HookGridComboBox(void *pArg, void *pEvent){
    MultiNumericalInputs_cls* pCDLG = (MultiNumericalInputs_cls*)pArg;
    pCDLG->GetParent()->Refresh(FALSE);
}

void MainGLCanvas_cls::
CopySettingToAll(){
    // Serialize $B<~$j$r9)IW$9$l$P0lEY$K%3%T!<$b$G$-$k$h$&$K$J$k$+$b!)(B
    for (size_t i = 0; i < mVPOtherCanvas.size(); i++){
	MainGLCanvas_cls* pC = mVPOtherCanvas[i];

	pC->mCGridType.mBody = mCGridType.mBody;
	pC->mCbShowOtherCam.mbBody = mCbShowOtherCam.mbBody;
	pC->mCbXOR.mbBody      = mCbXOR.mbBody;
	pC->mCbGridAnti.mbBody = mCbGridAnti.mbBody;
	pC->mCbShowGridTag.mbBody = mCbShowGridTag.mbBody;
	pC->mCDrawInfo.mbForceNewDraw = TRUE;
	
	pC->ChangeBGColor(mCColorBG0.mpBody, 0);
	pC->ChangeBGColor(mCColorBG1.mpBody, 1);
	pC->mCGLPerspective.mDepthMIN = mCGLPerspective.mDepthMIN;
	pC->mCGLPerspective.mDepthMAX = mCGLPerspective.mDepthMAX;
    }
}

wxColour MainGLCanvas_cls::
GetColourBG(int index){
    if (index == 0)
	return wxColour(mCColorBG0.mpBody[R] * 255,
			mCColorBG0.mpBody[G] * 255,
			mCColorBG0.mpBody[B] * 255);
    else
	return wxColour(mCColorBG1.mpBody[R] * 255,
			mCColorBG1.mpBody[G] * 255,
			mCColorBG1.mpBody[B] * 255);
}

void MainGLCanvas_cls::
ChangeBGColor(float *pRGB, int index){
    if (index == 0)
	LOOP_RGB mCColorBG0.mpBody[rgb] = pRGB[rgb];
    else
	LOOP_RGB mCColorBG1.mpBody[rgb] = pRGB[rgb];
    SetBackgroundColour(GetColourBG(0));
    Refresh(FALSE);
}

void MainGLCanvas_cls::
SelectItem(Parameters_cls *pCParam){
    mspCSelectedItem = pCParam;
    if (mspCSelectedItem == NULL)
	LOOP_XYZ msSelectedItemPosOrg[xyz] = msSelectedItemPos[xyz];
    else{
	LOOP_XYZ msSelectedItemPos[xyz] = mspCSelectedItem->mpInterpolatedValue[xyz];
	LOOP_XYZ msSelectedItemPosOrg[xyz] = msSelectedItemPos[xyz];
    }
}

// ----------------------------------------------

wxString MainGLCanvas_cls::
GetPovStrCamPos(){
    wxString strCam;
    strCam << PovOut3d(L"\tlocation", mCCameraPos.mPos) << ENDLWX;
    strCam << PovOut3d(L"\tlook_at" , GetAnchorPosForRender()) << ENDLWX;
    strCam << PovOut3d(L"\tsky"     , mCCameraPos.mTop) << ENDLWX;
    return strCam;
}

wxString MainGLCanvas_cls::
GetPovStrCamMatrix(){
    wxString strMatrix;
    strMatrix << wxT("#macro MatrixYZ2Cam()") << ENDLWX
	      << wxT("matrix < ") << mCCameraPos.mTgt[X] << wxT(",") 
	      << mCCameraPos.mTgt[Y] << wxT(",") 
	      << mCCameraPos.mTgt[Z] << wxT(",") << ENDLWX
	      << wxT("\t") << mCCameraPos.mRgt[X] << wxT(",") 
	      << mCCameraPos.mRgt[Y] << wxT(",")
	      << mCCameraPos.mRgt[Z] << wxT(",") << ENDLWX
	      << wxT("\t") << mCCameraPos.mTop[X] << wxT(",")
	      << mCCameraPos.mTop[Y] << wxT(",")
	      << mCCameraPos.mTop[Z] << wxT(",") << ENDLWX
	      << wxT("0, 0, 0>") << ENDLWX
	      << wxT("#end") << ENDLWX;
    wxString strScale;
    strScale << wxT("#declare SCALE=") << CalcAutoScale() << wxT(";") << ENDLWX;
    return strMatrix + strScale;
}

wxString MainGLCanvas_cls::
GetPovStrMask(int start, int end, int height, double theta){
    wxString strMask;
    
    float Ystart = (float)(height - (start - 1) * 2) / height * tan (theta * M_PI / 180 / 2);
    float Yend   = (float)(height - end * 2) / height * tan (theta * M_PI / 180 / 2);
    float Xstart = -2; //$B8e$G(B
    float Xend   = 2;
    
    strMask << wxT("//") << start << wxT(",") << end << wxT(",") << height << wxT(",") << theta << ENDLWX;

    strMask << wxT("/*") << ENDLWX;
    strMask << wxT("polygon {") << ENDLWX;
    strMask << wxT("\t10, <0, -3, -2>, <0, 3, -2>, <0, 3, 2>, <0, -3, 2>, <0, -3, -2>,") << ENDLWX;
    strMask << PovOut3(L"\t", 0, Xstart, Yend);
    strMask << PovOut3(L",", 0,  Xend, Yend);
    strMask << PovOut3(L",", 0,  Xend, Ystart);
    strMask << PovOut3(L",", 0,  Xstart, Ystart);
    strMask << PovOut3(L",", 0,  Xstart, Yend) << ENDLWX;
    strMask << wxT("\thollow") << ENDLWX;
    strMask << wxT("\ttexture { pigment { color rgb <0, 0, 0> } }") << ENDLWX;
    strMask << wxT("\tMatrixYZ2Cam()") << ENDLWX;
    strMask << wxT("\tscale 0.00001") << ENDLWX;
    
    double pos[3];
    LOOP_XYZ pos[xyz] = mCCameraPos.mPos[xyz] + 0.00001 * mCCameraPos.mTgt[xyz];
    
    strMask << PovOut3d(L"\ttranslate", pos) << ENDLWX << wxT("}") << ENDLWX << ENDLWX;
    
    strMask << wxT("plane {") << ENDLWX;
    strMask << wxT("\t-x, 0") << ENDLWX;
    strMask << wxT("\tMatrixYZ2Cam()") << ENDLWX;
    LOOP_XYZ pos[xyz] = mCCameraPos.mPos[xyz] + 0.00002 * mCCameraPos.mTgt[xyz];
    strMask << PovOut3d(L"\ttranslate", pos) << ENDLWX << wxT("}") << ENDLWX << ENDLWX;
    strMask << wxT("*/") << ENDLWX;
    
    return strMask;
}

void MainGLCanvas_cls::
ReadStrLinePanel(std::wistream &istream, std::wstring strLine){
    std::wstringstream sstrTag; 
    sstrTag << L"--GLCanvas" << mSelfIndex;
    std::wstring strTag = sstrTag.str();
    std::wstring::size_type loc = strLine.find(strTag,0);
    if (loc != std::string::npos){
	mCCameraPos.ReadStrLine(istream);
	std::getline(istream, strLine);
    }
    if (mCConfList.ReadStrLineAll(strLine) == TRUE)
	mCGLPerspective.mDeg = mCGLPersAngle.mBody;
}

wxString MainGLCanvas_cls::
SerializePanel(){
    //mCGLDepthMin.mBody = mCGLPerspective.mDepthMIN;
    //mCGLDepthMax.mBody = mCGLPerspective.mDepthMAX;
    //mCbGLPerspective.mbBody = mCGLPerspective.mbPerspective;
    mCGLPersAngle.mBody = mCGLPerspective.mDeg;
    wxString strOut;
    strOut << wxT("--GLCanvas") << mSelfIndex << ENDLWX;
    strOut << mCCameraPos.Serialize();
    strOut << mCConfList.SerializeAll();
    return strOut;
}

void MainGLCanvas_cls::
ReadStrLineGUI(std::wstring strLine){
    mspCPreference->ReadStrLine(strLine);
}

void MainGLCanvas_cls::
ReadStrLineAnchor(std::wistream &istream, std::wstring strLine){
    std::wstring strTag = L"--AnchorPos";
    std::wstring::size_type loc = strLine.find(strTag,0);
    if (loc != std::string::npos){
	std::getline(istream, strLine);
	msCAnchorPos.ReadStrLine(strLine);
    }
}

wxString MainGLCanvas_cls::
SerializeGUI(){
    wxString strOut;
    strOut << mspCPreference->Serialize();
    return strOut;
}

wxString MainGLCanvas_cls::
SerializeAnchor(){
    wxString strOut;
    strOut << wxT("--AnchorPos") << ENDLWX;
    strOut << msCAnchorPos.Serialize();
    return strOut;
}

Position_cls MainGLCanvas_cls::msCAnchorPos(0, 0, 0);
int    MainGLCanvas_cls::msMoveMode = _POS_AND_DIR_MOVE_FREE;
int    MainGLCanvas_cls::msManipulatorMode = _MANIP_TRANSLATE;
double MainGLCanvas_cls::msSelectedItemPos[3] = {0, 0 ,0};
double MainGLCanvas_cls::msSelectedItemPosOrg[3] = {0, 0 ,0};
double MainGLCanvas_cls::msSelectedItemScale[3] = {1, 1 ,1};
double MainGLCanvas_cls::msSelectedItemDYPRY[4] = {0, 0, 0, 0};
Parameters_cls* MainGLCanvas_cls::mspCSelectedItem = NULL;
Preference_cls* MainGLCanvas_cls::mspCPreference = NULL;
StereoGram_cls* MainGLCanvas_cls::mspCStereogramLink = NULL;
FilterCtrl_cls* MainGLCanvas_cls::mspCFilterCtrl = NULL;
DF3SeqCtrl_cls* MainGLCanvas_cls::mspCDF3SeqCtrl = NULL;
int MainGLCanvas_cls::msAFBOSize[5] = {256, 512, 1024, 2048, 4096};
float MainGLCanvas_cls::msLineWidth = 1;
int MainGLCanvas_cls::msGLLightIndex[8] = {
    GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3, 
    GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7
};

GLButton_cls MainGLCanvas_cls::msCGLBTNTrans(_BOTTOM_RIGHT, 80, 4);
GLButton_cls MainGLCanvas_cls::msCGLBTNRot  (_BOTTOM_RIGHT, 56, 4);
GLButton_cls MainGLCanvas_cls::msCGLBTNScale(_BOTTOM_RIGHT, 32, 4);
