// -*- c -*-
// "clsMainGLCanvas.cpp"

#ifndef _MAIN_GL_CANVAS_H_
#define _MAIN_GL_CANVAS_H_

#define GL_GLEXT_PROTOTYPES 
#define _MANIP_TRANSLATE 0
#define _MANIP_SCALE 1
#define _MANIP_ROTATE 2

#include <wx/wx.h>
#include <wx/glcanvas.h>
#include <wx/colordlg.h>
#include <wx/stopwatch.h>

class DF3SeqCtrl_cls;
class DF3Sequence_cls;
class TreeBoxRule_cls;

class CameraPath_cls; 
class Canvas_cls; 
class DrawingBlock_cls; 
class FilterCtrl_cls; 
class GLGuides_cls;
class ImgSequence_cls;
class MainGLCanvas_cls; 
class MouseCtrl_cls;
class MultiNumericalInputs_cls; 
class Preference_cls; 
class StereoGram_cls; 
class TimeCtrl_cls;

#include "../CONFITEM/clsConfItem.h"
#include "../GLCANVAS/clsStructDrawInfo.h"
#include "../MYGL/clsGLPerspective.h"
#include "../MYGL/clsRedrawPBO.h"
#include "../PARAMETER/clsMultiParameters.h"
#include "../POVRAY/clsPovSerialize.h"

#include "clsGLButton.h"
#include "clsMouseCtrl.h"

#include <vector>

class FrameMain_cls; 

#include "../WIDGETCTRL/clsPendEventList.h"

class LightSRC_cls; 
class Parameters_cls; 
class GLCanvasCtrl_cls;

#define _GL_CANVAS_DISPLAY_FREE    _DIRECTION_FREE
#define _GL_CANVAS_DISPLAY_X_PLUS  _DIRECTION_X_PLUS
#define _GL_CANVAS_DISPLAY_X_MINUS _DIRECTION_X_MINUS
#define _GL_CANVAS_DISPLAY_Y_PLUS  _DIRECTION_Y_PLUS
#define _GL_CANVAS_DISPLAY_Y_MINUS _DIRECTION_Y_MINUS
#define _GL_CANVAS_DISPLAY_Z_PLUS  _DIRECTION_Z_PLUS
#define _GL_CANVAS_DISPLAY_Z_MINUS _DIRECTION_Z_MINUS
#define _GL_CANVAS_LINKED_CAM0 7

#define _GL_CANVAS_GRID_BASIC 0
#define _GL_CANVAS_GRID_BASIC_CIRCULAR 1
#define _GL_CANVAS_GRID_POLAR 2

#define _GL_CANVAS_NOASSIN  0
#define _GL_CANVAS_GYRATION 1

/*
#define _GL_PERSPECTIVE 0
#define _GL_ORTHOGRAPHIC 1
#define _GL_FISHEYE 2
#define _GL_PANORAMIC 3
*/

#define _GYRATION_FREE 0
#define _GYRATION_FIXPOS 1

/*
class MousePos_cls{ 
 public:
    long mPrevX, mPrevY; // $B%^%&%9$N0LCV(B
    bool mbMode; // $B=i4|0LCV$K4X$7$F%b!<%I@_Dj(B

    MousePos_cls(long prevX = 0, long prevY = 0);
    void SetInitMode(wxMouseEvent event, wxSize size);
};

class MousePosNormal_cls{
 public:
    MousePosNormal_cls(MousePos_cls CMousePos, wxMouseEvent event, wxSize sz);
    
    float prevX;
    float prevY;
    float currX;
    float currY;
    bool mbMode;

    void LimitMove();
};
*/

class MainGLCanvas_cls : 
public wxGLCanvas,
public PendEventList_cls,
public PovSerialize_cls{
 public:
    MainGLCanvas_cls(GLCanvasCtrl_cls *parent, wxWindowID id = wxID_ANY,
		     int* attribList = 0,
		     const wxPoint& pos = wxDefaultPosition,
		     const wxSize& size = wxDefaultSize, 
		     long style = 0,
		     const wxString& name = wxT("MainGLCanvas"),
		     const wxGLContext *contextToShare = NULL,
		     unsigned int selfIndex = 0,
		     int displayMode = _GL_CANVAS_DISPLAY_FREE);

    ~MainGLCanvas_cls();
    
    GLCanvasCtrl_cls *mpCGLCtrl;
    MouseCtrl_cls    *mpCMouseCtrl;
    
    wxString     mstrWarning;
    int          mWarningLine;
    unsigned int mSelfIndex;
    wxStopWatch  mCSW;
    double mMatrixModelView[16];
    double mMatrixProjection[16];
    int    mMatrixViewport[4];
    
    double mPosItemOnW[3];
    double mPosCoordinateTag[2][3][3];
    double mPosCameraTag[4][3];
    bool   mbShowCoordinateTagXYZ[3];
    int    mLightNumber;
    static int msGLLightIndex[8];
    //static GLuint msTexIDRotation, msTexIDTranslation, msTexIDScale;
    static GLButton_cls msCGLBTNRot, msCGLBTNTrans, msCGLBTNScale;
    float  mALightPos[_MAX_ITEM_NUMBER][4];
    float  mALightColor[_MAX_ITEM_NUMBER][4];
    float  mALightAttenuateL[_MAX_ITEM_NUMBER];
    float  mALightAttenuateQ[_MAX_ITEM_NUMBER];

    void SetParentClass(FrameMain_cls* pCParentFrame);

    DF3SeqCtrl_cls* GetPDF3SeqCtrl();
    TimeCtrl_cls*   GetPTimeCtrl();
    TreeBoxRule_cls* GetPTreeBoxRule();

 private:
    bool SetGLContext();
    void InitGL();
    void InitializeGLIfNot();
    void UpdateWindowSizeInChildClass();
    
    // ----

 public:
    void UpdateDrawInfo(int width, int height);

    void UpdateCameraPos();
    CameraPath_cls* GetPLinkedPath();
    bool IsStereo();
    bool IsOrtographic();
    bool IsFishEye();
    bool IsPanoramic();
    bool IsFBOUse();
    void SetAngleIfLinked();
    void ModifyByStereogram(int forceType = -1);
    void ModifyByDomeTilt(PositionAndDirection_cls *pCCamPos);
    void ModifyByDome(int index);
    void ModifyByPanorama(int index);
    double* GetAnchorPosForRender();

    void SetViewport();
    void SetPerspective();
    void SetModelView();
    void SetLight();
    void DrawDataDrawObject(double scale, double scale10, 
			    float fogDistance, float fogColor[4],
			    std::vector < DrawingBlock_cls > &VCDrawingCls,
			    int stencilMode);
    void ClearBuffer();
    void DrawDataMain(double scale, double scale10, 
		      std::vector < DrawingBlock_cls > &VCDrawingCls);
    
    void DrawCameraPos(double scale, PositionAndDirection_cls CPos);
    void DrawGrid(double scale, double scale10);
    void DrawModeAxis(double scale);
    void DrawItems(double scale);
    void DrawAnchor(double scale);
    void DrawSelectedItem(double scale);
    void ShowNavigator(PositionAndDirection_cls CCamPos,
		       int width, int height);
    
    void ConstructDrawingBoxes(std::vector < DrawingBlock_cls > *pVCDrawingCls,
			       PositionAndDirection_cls CCamPos);
    
    void ShowInformation(int width, int height, bool bReuse, int InfoDiff,
			 std::vector < DrawingBlock_cls > *pVCDrawingCls);
    void ProcessXYZDataIfRequired();
    void RefreshEX();
    bool IsMouseOnClient();
    bool CheckFocusOnTherCanvas();
    bool CheckGLButton(int canvasWidth, int canvasHeight,
		       int mouseX, int mouseY);
    void OnPaint(wxPaintEvent& event);
    void OnSize (wxSizeEvent& event);

    bool (*mApFuncMove[3])(double);
    bool GyrationMain(double seinsitivity, MouseNormalized_cls MP, wxSize sz);
    bool ZoomMain(double seinsitivity, MouseNormalized_cls MP, wxSize sz);

    bool ZoomWheel(float delta);

    bool MoveCenter(double seinsitivity, MouseNormalized_cls MP, wxSize sz,
		    bool *trigAnchor);
    bool MoveItem(double seinsitivity, MouseNormalized_cls MP, wxSize sz,
		  bool *trigAnchor);
    bool ShowMenu(int x, int y);
    void ShowMenuPanel(int type, int x, int y);
    
    bool SetThisToBeActiveGL();
    bool ProcViewOrItemDragging(bool& trigAnchor, wxSize sz);
    void ProcSelectItemForDragging();
    bool ProcSelectItemForSelection();
    bool ProcCaptureMouse();
    void StartHitTest();
    void OnMouse(wxMouseEvent& event);
    void OnMouseRelease(wxMouseCaptureLostEvent& event);
    void OnMouseMotion(wxMouseEvent& event);
    void ProcShortcutKey(int modifier, int code);
    void OnKeyDown(wxKeyEvent& event);
    void RotationMode();
    void TranslationMode();
    void ScaleMode();



    std::vector < MainGLCanvas_cls* > mVPOtherCanvas;
    void RegistPOtherCanvas(MainGLCanvas_cls* pGLCanvas);

    static void InitializeStaticMembers
	(FrameMain_cls *pCFrame, DF3SeqCtrl_cls *pCDF3SeqCtrl, FilterCtrl_cls *pCFilterCtrl);
    static bool IsGLBlendEquationEXOK();
    static void DeleteInstancePreference();
    static void OpenPreference(wxWindow* pC);

 private:
    static FilterCtrl_cls* mspCFilterCtrl;
    static DF3SeqCtrl_cls* mspCDF3SeqCtrl;

 public:
    static Preference_cls* mspCPreference;
    static StereoGram_cls* mspCStereogramLink;
    
    void SetDefaultPosition(int displayMode);
    void TogglePerspective();
    void SetDisplayMode(int displayMode, bool bUpdateCombobox = FALSE);
    void LinkageConnect(int pathIndex);
    void OnPanelConfigure(wxCommandEvent &event);
    void OnMenuPers(wxCommandEvent& event);
    void OnMenuGridAnti(wxCommandEvent& event);
    void OnMenuOtherCam(wxCommandEvent& event);
    void OnMenuMoveToCamPos(wxCommandEvent& event);
    void OnMenuSetKeyPos   (wxCommandEvent& event);
    void OnMenuSetKeyAnc   (wxCommandEvent& event);
    void OnMenuToggleLink  (wxCommandEvent& event);
    void OnMenuPassToParent(wxCommandEvent& event);
    void OnMenuAddPrimitive(wxCommandEvent& event);
    void OnMenuShowPanel   (wxCommandEvent& event);

    void CopySettingToAll();
    wxColour GetColourBG(int index);

    void ChangeBGColor(float *pRGB, int index);
    
    static void SelectItem(Parameters_cls *pCParam);
    static void HookGridComboBox(void* pArg, void *pEvent);

    wxString GetPovStrCamPos();
    wxString GetPovStrCamMatrix();
    wxString GetPovStrMask(int start, int end, int height, double theta);
    void ReadStrLinePanel(std::wistream &istream, std::wstring strLine);
    wxString SerializePanel();
    static void ReadStrLineGUI(std::wstring strLine);
    static void ReadStrLineAnchor(std::wistream &istream, std::wstring strLine);
    static wxString SerializeGUI();
    static wxString SerializeAnchor();

 public:
    void UpdateAnc();
    Position_cls ModifyPosAndGetAnc();
 private:
    double CalcAutoScale();

    FrameMain_cls* mpCParentFrameLink;
    //MousePos_cls   mCMousePos;

    wxMenu *mpPopupMenu;
    wxMenu *mpMenuConfigure;
    wxMenu *mpMenuAdd;

    wxMenuItem *mpMenuItemShowPanel;

    bool mbInitialized;
    unsigned int mPrevFrame;
    
 public:
    bool mbMouseDragging;
    bool mbItemDragging;
    std::vector < int > mVHitObj; //
    
    int  mIndexForDebug;
    
 public:
    ConfItemList_cls      mCConfList;
    ConfItemFLOATARRAY_cls mCColorBG0;
    ConfItemFLOATARRAY_cls mCColorBG1; // outsideColor

    ConfItemBOOL_cls      mCbShowData, mCbShowGrid;
    ConfItemBOOL_cls      mCbShowBox,  mCbShowWire;
    ConfItemBOOL_cls      mCbShowItems;
    ConfItemBOOL_cls      mCbStereoFlip;
    ConfItemFLOAT_cls     mCGLPersAngle;
    
    ConfItemINT_cls       mCGLGeometry;
    ConfItemINT_cls       mCDisplayMode;
    ConfItemINT_cls       mCPathIndex;
    ConfItemUINT_cls      mCGridType;
    ConfItemBOOL_cls      mCbShowOtherCam;
    ConfItemBOOL_cls      mCbXOR;
    ConfItemBOOL_cls      mCbGridAnti;
    ConfItemBOOL_cls      mCbShowGridTag;
    bool mbSaveImageQueue;
    bool mbSaveImgSeqQueue;

 public:
    wxGLContext*              mpCGLContext;

    RedrawPBO_cls             mCRedrawPBO[4];
    StructDrawInfo            mCDrawInfo, mCDrawInfoPrev;
    
    GLPerspective_cls         mCGLPerspective;
    GLGuides_cls              *mpCGLGuide, *mpCGLGuideScrew, *mpCGLGuideYPR;
    PositionAndDirection_cls  mCCameraPos, mCCameraPosCached;
    PositionAndDirection_cls  mCDataParentCoordinate;
    static Position_cls       msCAnchorPos; 
    static int                msMoveMode;
    static int                msManipulatorMode;

    static Parameters_cls     *mspCSelectedItem;
    static double             msSelectedItemPos[3];
    static double             msSelectedItemPosOrg[3];
    static double             msSelectedItemScale[3];
    static double             msSelectedItemDYPRY[4];
    double                    mdPosTmp[3];
    
    static int                msAFBOSize[5];
        
    ConfItemBOOL_cls mbCFBO;
    ConfItemUINT_cls mCFBOSizeIndex;

    static float              msLineWidth;

    // -- $B%j%s%/(B --
    CameraPath_cls*           mpCLinkedPath;
    ImgSequence_cls*          mpCLinkedImgSeq;
    
    DECLARE_EVENT_TABLE()
};

#endif //_MAIN_GL_CANVAS_H_
