// (setq c-basic-offset 4)

#ifndef _GLOBJECTS_H_
#define _GLOBJECTS_H_

#ifdef _MAKE_DLL

// 쐬ꍇ
#ifdef WIN32

#ifdef __cplusplus
#define PORT extern "C" __declspec(dllexport)
#else
#define PORT __declspec(dllexport)
#endif
#define PORTCLS __declspec(dllexport)

#else //WIN32
#define PORT
#define PORTCLS 
#define __stdcall
#endif //WIN32

#else 

// pꍇ
#ifdef WIN32

#ifdef WIN64
#pragma comment(lib, "GLOBJECTS64.lib")  // C|[gCuN
#else
#pragma comment(lib, "GLOBJECTS32.lib")  // C|[gCuN
#endif

#ifdef __cplusplus
#define PORT extern "C" __declspec(dllimport)
#else
#define PORT __declspec(dllimport)
#endif
#define PORTCLS __declspec(dllimport)

#else //WIN32
#define PORT
#define PORTCLS 
#define __stdcall
#endif //WIN32

#endif

#include <vector>
#include <string>

#define _CAMERA_1 0
#define _CAMERA_2 1
#define _CAMERA_POLYGON 2
#define _CAMERA_CROSSHAIR 3
#define _GLOBJ_WIREBOX 4
#define _GLOBJ_LIGHTSRC 5

#define _STRINGS_BOTTOM_LEFT 0
#define _STRINGS_TOP_LEFT    1

namespace GLOBJ{

    void Normalize(double vec[3], double norm);
    void Normalize(float vec[3], float norm);
    void ChangeCoordinate3Dto2D(double vec3D[3], double vec2D[2]);
    void AdjustVec2DOverBoundary(double vector[2], double vectormem[2]);
    PORT void SetMatrix3dv3(double loc[3], double tgt[3], double top[3]);
    PORT void SetMatrix3dv3Inv(double loc[3], double tgt[3], double top[3]);
    
    class PORTCLS Sphere {
    private:
	static void DrawTriangle(double v1[3], double v2[3], double v3[3]);
	static void DivideTriangle(double v1[3], double v2[3], double v3[3], int depth);
    public:
	static int DListDraw(int depth, int iMode);

	static void MakeMeshSphereDivide
	    (std::vector < float > *pVVertices,  
	     std::vector < float > *pVNormals,
	     std::vector < unsigned int > *pVIndex,
	     std::vector < unsigned int > *pVIndexNormal,
	     float v1[3], float v2[3], float v3[3], 
	     int index1, int index2, int index3, int *pCurrentIndex, int depth);
	
	static void  MakeMeshSphere
	    (int depth, 
	     std::vector < float > *pVVertices,
	     std::vector < float > *pVNormals,
	     std::vector < unsigned int > *pVIndex,
	     std::vector < unsigned int > *pVIndexNormal);
    };

    class PORTCLS Cone {
    public:
	static int DListDraw(int facets, int fineness, float topScale = 1, 
			     bool bOpen = false, int iMode = 0);

	static void MakeMeshDisk(std::vector < float > *pVVertices,
				 std::vector < float > *pVNormals,
				 std::vector < unsigned int > *pVIndex,
				 std::vector < unsigned int > *pVIndexNormal,
				 int facets, int fineness, float scale,
				 float z, float nz);

	static void MakeMeshCone
	    (std::vector < float > *pVVertices,
	     std::vector < float > *pVNormals,
	     std::vector < unsigned int > *pVIndex,
	     std::vector < unsigned int > *pVIndexNormal,
	     int facets, int fineness, float topScale = 1, bool bOpen = false);
    };
    
    class PORTCLS GridFlat {
    private:
	static bool mbInitialized;
	static bool mbColor;
	static int mDListInner; // gȂׂ̍Obh
	static int mDListOuter; // gObh
	static int mFineness;
	
	static int  DListDraw(int fineness, int GLMode, 
			      int gridMode = 0);
	static void InitializeIfNot();
	
    public:
	static void Draw(double scale);
	static void ColorModeOn(void);
    };

    class PORTCLS GridCircularFlat {
	static bool mbInitialized;
	static bool mbColor;
	static int mDListCircular;
	static int mFineness;

	static int DListDraw(int fineness, int GLMode);

	static void InitializeIfNot();
    public:
	static void Draw(double scale);
    };

    class PORTCLS GridPolar {
	static bool mbInitialized;
	static bool mbColor;
	static int mDListPolar;
	static int mFineness;

	static int DListDraw(int fineness, int GLMode);

	static void InitializeIfNot();
    public:
	static void Draw(double scale);
    };

    class PORTCLS Object{
    private:
	static bool mbInitialized;
	static int  mDList[6];

	static int  DListDraw(int type, int GLMode);
	static void InitializeIfNot();

    public:
	static void Draw(int type);
	static int  GetList(int type);
    };

    class PORTCLS Raster2D{
    public:
	static void Push2DMatrix(int width, int height);
	static void Pop2DMatrix();
    };

    class PORTCLS Strings2D{
    private:
	static bool mbInitGLUT;
	static void MyInitiGLUT();
	static int msWidth; //ȂƃoO1Â錾łȂ
	static int msHeight;
    public:

	static void SetWindowGeometry(int width, int height);
	static void Draw(void *fonttype, const char *chars, int posx, int posy,
			 int mode = _STRINGS_TOP_LEFT);
	static void Draw3DPos(void *fonttype, const char *chars, 
			      float posx, float posy, float posz);
	static void FDrawFloat(void *fonttype, std::string format, float val,
			       int posx, int posy,
			       int mode = _STRINGS_TOP_LEFT);
    };
}

#endif
