// "clsMainGLCanvas.h"
// (setq c-basic-offset 4)

#include "clsImageIO.h"
#include "../MYAPI/MsgBox.h"
#include "../COMMON/clsFilename.h"
#include "../POVRAY/clsStereoGram.h"
#include "../WIDGETCTRL/EX/clsFileDialogEX.h"
#include <wx/wx.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>

extern wxFileDialogEX*   GpCFileDialog;

ThreadSave_cls::
ThreadSave_cls() : wxThread(wxTHREAD_JOINABLE){
}

void ThreadSave_cls::
SetParameters(ImageIO_cls *pCParent,
	      int width, int height,
	      wxString strFileOut, unsigned char *pBuffer, 
	      bool bFreeDataAtEnd){
    mpCParent = pCParent;
    mWidth = width; mHeight = height;
    mstrFileOut = strFileOut;
    
    if (pBuffer == NULL){
	mpBuffer = (unsigned char*)malloc(mWidth * mHeight * 3);
	mbAllocLocal = TRUE;
    }
    else{
	mpBuffer = pBuffer;
	mbAllocLocal = bFreeDataAtEnd;
    }

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode 
ThreadSave_cls::
Entry(){
    wxImage CImage(mWidth, mHeight, mpBuffer, TRUE);

    SplitPathWX_cls CPath(mstrFileOut);
    if (CPath.GetPathWXStr(_FN_EXT) == wxT("")) mstrFileOut += wxT(".png");
    
    mpCParent->mstrLastSavedFile = mstrFileOut; //Xbh瑀댯
    
    wxImage CMirror(CImage.Mirror(FALSE)); // ]ftH
    CMirror.SetOption(wxT("quality"), mpCParent->mJPEGQuality);
    CMirror.SaveFile(mstrFileOut);

    if (mbAllocLocal == TRUE) free(mpBuffer);
    return 0;
}

// -----

ImageIO_cls::
ImageIO_cls(){
    mJPEGQuality = 80;
    for (int i = 0; i < _IMAGE_OUT_MAX_THREAD; i++)
	mbSaveThreadActive[i] = FALSE;
    mActiveThread = 0;
}

ImageIO_cls::
~ImageIO_cls(){
    for (int i = 0; i < _IMAGE_OUT_MAX_THREAD; i++)
	if (mbSaveThreadActive[i] == TRUE){
	    mpCThreadSave[i]->Wait();
	    delete mpCThreadSave[i]; }
}

bool ImageIO_cls::
SaveImageFromData(int width, int height, wxString strFileOut, 
		  unsigned char* pBuffer){
    if (mbSaveThreadActive[mActiveThread] == TRUE){
	mpCThreadSave[mActiveThread]->Wait();
	delete mpCThreadSave[mActiveThread]; // 
	mbSaveThreadActive[mActiveThread] = FALSE;
    }

    bool retVal = FALSE;

    mpCThreadSave[mActiveThread] = new ThreadSave_cls;
    mpCThreadSave[mActiveThread]->SetParameters(this, width, height, strFileOut,
						pBuffer, TRUE);
 
    mbSaveThreadActive[mActiveThread] = TRUE;    
    mpCThreadSave[mActiveThread]->Run();
    mActiveThread++;
    if (mActiveThread >= _IMAGE_OUT_MAX_THREAD) mActiveThread = 0;
    
    return TRUE;
}

bool ImageIO_cls::
SaveImageFromGL(int width, int height){
    return SaveImageFromGL(width, height, mstrFile);
}

bool ImageIO_cls::
SaveImageFromGL(int width, int height, wxString strFileOut,
		int offsetX, int offsetY){

    if (mbSaveThreadActive[mActiveThread] == TRUE){
	mpCThreadSave[mActiveThread]->Wait();
	delete mpCThreadSave[mActiveThread]; // 
	mbSaveThreadActive[mActiveThread] = FALSE;
    }

    bool retVal = FALSE;
    int trimWidth = width - (width % 8);
    
    mpCThreadSave[mActiveThread] = new ThreadSave_cls;
    mpCThreadSave[mActiveThread]->SetParameters(this, trimWidth, height, strFileOut);

    glReadPixels(offsetX, offsetY, trimWidth, height, GL_RGB, GL_UNSIGNED_BYTE, mpCThreadSave[mActiveThread]->mpBuffer);

    mbSaveThreadActive[mActiveThread] = TRUE;    
    mpCThreadSave[mActiveThread]->Run();
    mActiveThread++;
    if (mActiveThread >= _IMAGE_OUT_MAX_THREAD) mActiveThread = 0;
    
    return TRUE;
}

bool ImageIO_cls::
SelectFile(wxString strDefaultDir, wxWindow *pC){
    static bool sbState = FALSE; // Ƀ_CAOoĂ邩
    if (sbState == TRUE) {sbState = FALSE; return FALSE;}
    sbState = TRUE; 

    GpCFileDialog->SetMessage(_("Select a image file to save"));
    GpCFileDialog->SetWildcard(wxT("image(*.png;*.jpg;*.tiff)|*.png;*.jpg;*.tiff|png (*.png)|*.png|jpg (*.jpg)|*.jpg|tiff (*.tiff)|*.tiff"));
    GpCFileDialog->SetStyle(wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
    
    if (strDefaultDir != wxT("")){
	SplitPathWX_cls CPath(strDefaultDir);
	GpCFileDialog->SetInitPath(CPath.GetPathWXStr(_FN_DRIVE_DIR));
    }

    bool bRetval = GpCFileDialog->ShowModal();
    if (bRetval == TRUE) mstrFile = GpCFileDialog->GetPath();

    sbState = FALSE;
    return bRetval;
}

void ImageIO_cls::
SetFilenameStereo(int eyeType){
    SplitPathWX_cls CSplit(mstrFile);

    wxString strLast = CSplit.mwsFilename.Right(1);
    if (strLast.IsSameAs(wxT("L")) || strLast.IsSameAs(wxT("R")))
	CSplit.mwsFilename.RemoveLast();
    
    switch(eyeType){
    case _STEREO_LFT:
	CSplit.mwsFilename.Append(wxT("L"));
	break;
    case _STEREO_RGT:
	CSplit.mwsFilename.Append(wxT("R"));
	break;
    }
			 
    mstrFile = CSplit.GetPathWXStr();
}

ImageIO_cls GCImgIO;
