// -*- c -*- 
// (setq c-basic-offset 4)
// "clsImgSequence.h"

#include "clsImgSequence.h"
#include "clsImageIO.h"
#include "../MYGL/clsPositionAndDirection.h"
#include "../WIDGETCTRL/EX/clsFileDialogEX.h"

extern wxFileDialogEX*   GpCFileDialog;

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#ifdef WIN32
#define DELIM L"\\"
#define DELIMWX wxT("\\")
#else
#define DELIM L"/"
#define DELIMWX wxT("/")
#endif

BEGIN_EVENT_TABLE(ImgSequenceGUI_cls, wxDialog)
    EVT_BUTTON(buttonImgSeqRender,    ImgSequenceGUI_cls::OnButtonRender)
    EVT_BUTTON(buttonImgSeqReference, ImgSequenceGUI_cls::OnButtonReference)
    EVT_BUTTON(wxID_CANCEL, ImgSequenceGUI_cls::OnButtonClose)
END_EVENT_TABLE()

ImgSequenceGUI_cls::
ImgSequenceGUI_cls(ImgSequence_cls *pCParent, wxWindow *pC,
    		   wxPoint pos, wxSize size) :
    wxDialogEXKey(pC, wxID_ANY, _("Image Sequcnce Output"),
		  pos, size){
    mpCParent = pCParent;
    SetScreenRectPos(&pCParent->mCScreenRect);

#ifdef WIN32
    wxIcon icon(wxT("ICONS/ImgSequence.ico"), wxBITMAP_TYPE_ICO);
#else
    wxIcon icon(wxT("ICONS/ImgSequence.png"), wxBITMAP_TYPE_PNG);
#endif
    SetIcon(icon);

    // TCU[
    wxBoxSizer       *pBoxSizerSrc     = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerParam   = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerButtons = new wxBoxSizer(wxHORIZONTAL);
 
    SetMNumImgSeqParam(pBoxSizerParam);
    
#ifdef WIN32
    int buttonWidth=100;
#else
    int buttonWidth=140;
#endif
    mpCButtonRender =
	new wxButtonEX(this, buttonImgSeqRender, wxT("Render Anim"),
		       wxDefaultPosition, wxSize(buttonWidth, -1));

    wxButtonEX *pCButtonCancel =
	new wxButtonEX(this, wxID_CANCEL, wxT("Close"),
		       wxDefaultPosition, wxSize(buttonWidth, -1));


    pBoxSizerButtons->Add(mpCButtonRender, 0, wxEXPAND);
    pBoxSizerButtons->Add( -1, -1, 1, wxEXPAND);
    pBoxSizerButtons->Add(pCButtonCancel, 0, wxEXPAND);
    
    pBoxSizerSrc->Add(pBoxSizerParam, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerButtons, 0, wxEXPAND);

    SetSizerAndFit(pBoxSizerSrc);

    mpCButtonRender->SetFocus();
}

void ImgSequenceGUI_cls::
SetMNumImgSeqParam(wxBoxSizer *pBoxSizer){
    mpCMNumImgSeqParam = new MultiNumericalInputsPanel_cls(&mpCMNumImgSeqParam, this, wxID_ANY, wxDefaultPosition, wxSize(-1, -1), wxBORDER_SIMPLE);

    MultiNumericalInputsPanel_cls *pCurr = mpCMNumImgSeqParam;
    int boxWidth;
    pCurr->mTextWidth = 160/*-1*/;

    boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Preview (W, H, V)"));
    pCurr->CombinePointer(_GP_INTEGER, &mpCParent->mClipWidthMax, wxT(""), FALSE);
    pCurr->CombinePointer(_GP_INTEGER, &mpCParent->mClipHeightMax, wxT(""), FALSE);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mAngleOrg, wxT(""), FALSE);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Image (W, H, V)"));
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mClipWidth);
    pCurr->SetRange(0, 2048);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mClipHeight);
    pCurr->SetRange(0, 2048);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mClipAngle);

    boxWidth = 3;
    pCurr->mDefaultButton = _MN_NO_BUTTONS;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Initial sequence"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &mpCParent->mCInitNum.mBody);
    pCurr->SetRange(0, 9999);
    pCurr->CombinePointer(_GP_INTEGER, &mpCParent->mCInitNum.mBody);
    pCurr->mButtonWidth = -1;
    pCurr->CombinePointerButton(buttonImgSeqReference, _("Reference"));

    pCurr->CombinePointer(_GP_WXSTRING, &mpCParent->mCFileSequence.mwsFormat);

#ifdef WIN32    
    pCurr->mMinimumWidth = 480;
#else
    pCurr->mMinimumWidth = 520;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);
}

#include "../MYAPI/MsgBox.h"

void ImgSequenceGUI_cls::
OnButtonReference(wxCommandEvent &event){
    GpCFileDialog->SetMessage(_("Select a image file to save"));
    GpCFileDialog->SetWildcard(wxT("image(*.png;*.jpg;*.tiff)|*.png;*.jpg;*.tiff|png (*.png)|*.png|jpg (*.jpg)|*.jpg|tiff (*.tiff)|*.tiff"));
    GpCFileDialog->SetStyle(wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
    GpCFileDialog->SetInitPath(mpCParent->mCFileSequence.GetPathWXStr(_FN_DRIVE_DIR));

    if (GpCFileDialog->ShowModal() == wxID_OK){
	mpCParent->mCFileSequence.Set(GpCFileDialog->GetPath());
	if (mpCParent->mCFileSequence.mDigitLength != -1)
	    mpCParent->mCInitNum.mBody = mpCParent->mCFileSequence.mDigit;
	mpCMNumImgSeqParam->UpdateValues();
    }
}

void ImgSequenceGUI_cls::
OnButtonRender(wxCommandEvent &event){
    if (mpCParent->mbInRendering == TRUE) 
	{ AbortRender(); return; }

    mpCMNumImgSeqParam->UpdateFromText();
    mpCMNumImgSeqParam->AssignBase();
    
    mpCParent->mCFileSequence.Set(mpCParent->mCFileSequence.GetFormat());
    mpCMNumImgSeqParam->UpdateValues();

    if (mpCParent->mCFileSequence.mDigitLength == -1){
	if (MsgBoxYesno(this, _("Sequence Format error"), 
			_("Not Sequence. Append 4-digit tag ?")) == FALSE)
	    return;
	mpCParent->mCFileSequence.AppendDefaultSequenceTag();
	mpCMNumImgSeqParam->UpdateValues();
    }

    wxCommandEvent cmd(wxEVT_COMMAND_MENU_SELECTED, menuItemSaveImgSeqBegin);
    GetParent()->AddPendingEvent(cmd);
    mpCButtonRender->SetLabel(wxT("Abort Anim"));
    mpCButtonRender->SetBackgroundColour(wxColour(wxT("RED")));
    mpCButtonRender->SetForegroundColour(wxColour(wxT("WHITE")));
    mpCParent->mbInRendering = TRUE;
}

void ImgSequenceGUI_cls::
OnButtonClose(wxCommandEvent &event){
    mpCMNumImgSeqParam->UpdateValues();
    CloseEXKey();
}

void ImgSequenceGUI_cls::
AbortRender(){
    wxCommandEvent cmd(wxEVT_COMMAND_MENU_SELECTED, menuItemPlayStop);
    GetParent()->AddPendingEvent(cmd);
    OutputFinish();
}

void ImgSequenceGUI_cls::
OutputFinish(){
    mpCButtonRender->SetLabel(wxT("Render Anim"));
    mpCButtonRender->SetBackgroundColour(wxNullColour);
    mpCButtonRender->SetForegroundColour(wxNullColour);
    mpCParent->mbInRendering = FALSE;
}

// ----

ImgSequence_cls::
ImgSequence_cls(wxWindow *pC) :
    mCFileSequence(wxGetCwd()+DELIMWX+wxT("Preview%04d.png"))
{
    mClipWidthMax = 640;
    mClipHeightMax = 480;
    mAngleOrg = 45;
    mClipWidth = 640;
    mClipHeight = 480;
    mClipAngle = 30;
    mOffsetX = 0;
    mOffsetY = 0;
    mbInRendering = FALSE;
    mpCGUI = NULL;
    SetToDefault();
    InitGUI(pC); // GUĨCX^X
}

void ImgSequence_cls::
SetGeom(int width, int height, float angle, bool bSetTrim){
    mClipWidthMax  = width;
    mClipHeightMax = height;
    mClipAngle = angle;
    if (bSetTrim == TRUE){
	mClipWidth = width;
	mClipHeight = height;
    }
    if (mpCGUI != NULL && mpCGUI->IsShownOnScreen())
	UpdateItems();
}

void ImgSequence_cls::
CalcOffset(int widthOrg, int heightOrg){
    mOffsetX = (widthOrg - mClipWidth) / 2;
    mOffsetY = (heightOrg - mClipHeight) / 2;
}

wxString ImgSequence_cls::
GetWXStrModifiedFilename(){
    return mCFileSequence.GetModifiedWXStr(mCurrentOffset + mCInitNum.mBody,
					   mCurrentStereo);
}

void ImgSequence_cls::
SetDefaultPath(wxString strPath){
    if (strPath == wxT("") || strPath == wxT("NONE")) return;
    mCFileSequence.Set(strPath);
}

bool ImgSequence_cls::
SaveImageFromGL(ImageIO_cls* pCImageIO, int widthOrg, int heightOrg, int offset, int stereo){
    // CHECK_ME!! łQ`݂̖łĂH
    if (mCurrentStereo == stereo && 
	offset < mCurrentOffset) return TRUE;

    mCurrentStereo = stereo;
    bool retVal = FALSE;
    CalcOffset(widthOrg, heightOrg);
    retVal = pCImageIO->SaveImageFromGL(mClipWidth, mClipHeight,
					GetWXStrModifiedFilename(),
					mOffsetX, mOffsetY);
    //Increment();
    return retVal;
}

void ImgSequence_cls::
Increment(int offset){
    if (offset == mCurrentOffset)
	mCurrentOffset++;
}

void ImgSequence_cls::
SetToDefault(){
    mCInitNum.SetItem(L"ImgSequenceInitNum", L"0");
}

void ImgSequence_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new ImgSequenceGUI_cls(this, pC,
				    mCScreenRect.GetPosition(),
				    mCScreenRect.GetSize());
    SetPDialog(mpCGUI);
}

void ImgSequence_cls::
UpdateItems(bool bForce){
    mpCGUI->mpCMNumImgSeqParam->UpdateValuesAll(bForce);
}

void ImgSequence_cls::
RenderFrameBoxInRender(int width, int height, GLPerspective_cls CGLPerspective){
    glPushMatrix();
    PositionAndDirection_cls CPos(0, 0, -1);
    //GLPerspective_cls CGLPerspective = mCGLPerspective;
    CGLPerspective.mbPerspective = FALSE;
    CGLPerspective.SetPerspectiveOnGL(&CPos);
    double rateX = mClipWidth/width;
    double rateY = mClipHeight/height;
    double aspect = CGLPerspective.GetAspectRatio();
    glBegin(GL_LINE_LOOP);
    glVertex2f(-0.5 * aspect * rateX, -0.5 * rateY);
    glVertex2f(+0.5 * aspect * rateX, -0.5 * rateY);
    glVertex2f(+0.5 * aspect * rateX, +0.5 * rateY);
    glVertex2f(-0.5 * aspect * rateX, +0.5 * rateY);
    glEnd();
    glPopMatrix();
}
