// "clsRedrawPBO.h"
// (setq c-basic-offset 4)

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
PFNGLGENBUFFERSPROC glGenBuffers;
PFNGLISBUFFERPROC glIsBuffer;
PFNGLBINDBUFFERPROC glBindBuffer;
PFNGLBUFFERDATAPROC glBufferData;
PFNGLBUFFERSUBDATAPROC glBufferSubData;
PFNGLMAPBUFFERPROC glMapBuffer;
PFNGLUNMAPBUFFERPROC glUnmapBuffer;
PFNGLDELETEBUFFERSPROC glDeleteBuffers;
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "clsRedrawPBO.h"

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

RedrawPBO_cls::
RedrawPBO_cls(){
    mbInitialized = FALSE;
}

#include "../MYAPI/MsgBox.h"

void RedrawPBO_cls::
Initialize(){
    if (msbSupported == FALSE) return;
#ifdef WIN32
    glGenBuffers =
	(PFNGLGENBUFFERSARBPROC)wglGetProcAddress("glGenBuffers");
    glIsBuffer =
	(PFNGLISBUFFERPROC)wglGetProcAddress("glIsBuffer");
    glBindBuffer =
	(PFNGLBINDBUFFERPROC)wglGetProcAddress("glBindBuffer");
    glBufferData =
	(PFNGLBUFFERDATAPROC)wglGetProcAddress("glBufferData");
    glBufferSubData =
	(PFNGLBUFFERSUBDATAPROC)wglGetProcAddress("glBufferSubData");
    glMapBuffer =
	(PFNGLMAPBUFFERPROC)wglGetProcAddress("glMapBuffer");
    glUnmapBuffer =
	(PFNGLUNMAPBUFFERPROC)wglGetProcAddress("glUnmapBuffer");
    glDeleteBuffers =
	(PFNGLDELETEBUFFERSPROC)wglGetProcAddress("glDeleteBuffers");
    //if (glGenBuffers == NULL) msbSupported = FALSE;
#endif
    //if (msbSupported == TRUE)
    //glGenBuffers(1, &mPBOStock);

    mbInitialized = TRUE;
}

void RedrawPBO_cls::
MakeStripePatternOnStencil(int width, int height, bool bOdd){
    
    // XeṼNA
    glClear( GL_STENCIL_BUFFER_BIT);

    //ꂩ`悷̂̃XeVlɂׂĂP^O    
    glStencilFunc( GL_ALWAYS, 1, ~0); 
    glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
    
    //XeVobt@ȊÕobt@ɂ͏܂Ȃ
    glColorMask(0,0,0,0);
    glDepthMask(0);
    
    glLineWidth(1);
    for (int y = (int)bOdd; y < height; y+=2){
	glBegin(GL_LINES);
	glVertex2f(     0, y);
	glVertex2f( width, y);
	glEnd();
    }

    //̉
    glColorMask(1,1,1,1);
    glDepthMask(1);
    glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
}

bool RedrawPBO_cls::
RedrawPBO(int width, int height, int renderType, bool bOdd){
    if (msbSupported == FALSE) return FALSE;
    if (mbInitialized == FALSE) Initialize();
    
    if (mPBOStock == 0) return FALSE;
        
    glClear(GL_DEPTH_BUFFER_BIT);
	
    int alignment = 8;
    if (width % 8 == 0) alignment = 8;
    else if (width % 4 == 0) alignment = 4;
    else if (width % 2 == 0) alignment = 2;
    else alignment = 1;
    glPixelStorei(GL_UNPACK_ALIGNMENT, alignment);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0, width, 0, height);
    glRasterPos2i(0, 0);
    
    if (renderType == _REDRAW_PBO_SCANLINE){
	glEnable(GL_STENCIL_TEST);
	MakeStripePatternOnStencil(width, height, bOdd);
	glStencilFunc( GL_EQUAL, 1, ~0); //1̃sNZɕ`
    }

    glBindBuffer( GL_PIXEL_UNPACK_BUFFER_ARB, mPBOStock);
    if (renderType == _REDRAW_PBO_ADDITIVE){
	glEnable (GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
    }
    glDrawPixels(width, height, GL_BGRA, GL_UNSIGNED_BYTE, 0);
    //glDrawPixels(width, height, GL_LUMINANCE, GL_UNSIGNED_BYTE, 0);

    //XeVeXgI
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_BLEND);

    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
    
    return TRUE;
}

bool RedrawPBO_cls::
SavePBO(int width, int height){
    if (msbSupported == FALSE) return FALSE;
    if (mbInitialized == FALSE) Initialize();

    if (mPBOStock != 0)
	glDeleteBuffers(1, &mPBOStock);
    glGenBuffers(1, &mPBOStock);
    
    glBindBuffer( GL_PIXEL_PACK_BUFFER_ARB, mPBOStock);
    glBufferData( GL_PIXEL_PACK_BUFFER_ARB, width * height * 4, 0, GL_STREAM_READ_ARB);
    //݂̃t[obt@PBOɃRs[
    glBindBuffer( GL_PIXEL_PACK_BUFFER_ARB, mPBOStock);

    // ƂĂxF PBOH ObsoluteH
    //glPixelTransferi(GL_MAP_COLOR, GL_FALSE);
    //glPixelTransferf(GL_RED_SCALE, 0.299f);
    //glPixelTransferf(GL_GREEN_SCALE, 0.587f);
    //glPixelTransferf(GL_BLUE_SCALE, 0.114f);

    glReadPixels( 0, 0, width, height, GL_BGRA, GL_UNSIGNED_BYTE, 0);
    //glReadPixels( 0, 0, width, height, GL_LUMINANCE, GL_UNSIGNED_BYTE, 0);

    glPixelTransferi(GL_MAP_COLOR, GL_FALSE);
    glPixelTransferf(GL_RED_SCALE, 1.0f);
    glPixelTransferf(GL_GREEN_SCALE, 1.0f);
    glPixelTransferf(GL_BLUE_SCALE, 1.0f);

    glBindBuffer( GL_PIXEL_PACK_BUFFER_ARB, 0);
    
    return TRUE;
}

bool RedrawPBO_cls::msbSupported = FALSE;

