// "clsTextureMapping.h"
// (setq c-basic-offset 4)

#include "../POVRAY/POVTEXTURE/texture.h"
#include "../POVRAY/POVTEXTURE/pattern.h"
#include "../POVRAY/POVTEXTURE/clsHexagon.h"

#include "clsTextureMapping.h"
#include "clsTextureProperty.h"

#include "../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;

#define R 0
#define G 1
#define B 2
#define A 3
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)
#define LOOP_RGBA for (int rgba = 0; rgba < 4; rgba++)
#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

#define _RAD 3.14159 / 180

TextureMapping_cls::
TextureMapping_cls(){
    mPrevColorType = -1;
    mPrevPatternType = -1;
    mColorRefreshId = -1;
    mTexId = -1;
    mPrevPos[X] = mPrevPos[Y] = mPrevPos[Z] = 9999;
    mPrevMapScale = -9999;
    mPrevMapTurb = -9999;
    mPrevAlpha = -1;
}

TextureMapping_cls::
~TextureMapping_cls(){
    if (mTexId != -1)
	glDeleteTextures(1, &mTexId);
}

void TextureMapping_cls::
CopyCurrToPrev(TexProperty_cls *pCTexProp){
    mPrevDoubleIndex = mCurrDoubleIndex;
    mPrevMapScale = mCurrMapScale;
    mPrevMapTurb = mCurrMapTurb;
    mPrevMappingType = mCurrMappingType;
    mPrevPatternType = mCurrPatternType;
    mPrevMapAxis = mCurrMapAxis;
    mPrevMappingCoord = mCurrMappingCoord;
    mPrevAlpha = pCTexProp->GetLinkedParam();
    LOOP_XYZ mPrevPos[xyz] = mCurrPos[xyz];
    LOOP_XYZ mPrevOffset[xyz] = mCurrOffset[xyz];
    LOOP_XYZ mPrevScale[xyz] = mCurrScale[xyz];
    for (int i = 0; i < 4; i++)	mPrevRotate[i] = mCurrRotate[i];
}

bool TextureMapping_cls::
IsParameterIsNewOne(double doubleIndex, int mappingType, 
		    int patternType, int mapAxis, int mappingCoord,
		    float pos[3], float mapscale, float mapturb,
		    double offset[3], double scale[3],
		    double rotation[4]){
    mCurrDoubleIndex = doubleIndex;
    mCurrMapScale = mapscale;
    mCurrMapTurb = mapturb;
    mCurrMappingType = mappingType;
    mCurrPatternType = patternType;
    mCurrMapAxis = mapAxis;
    mCurrMappingCoord = mappingCoord;
    LOOP_XYZ mCurrPos[xyz] = pos[xyz];
    LOOP_XYZ mCurrOffset[xyz] = offset[xyz];
    LOOP_XYZ mCurrScale[xyz] = scale[xyz];
    for (int i = 0; i < 4; i++) mCurrRotate[i] = rotation[i];

    mCosZ1 = cos(rotation[3] * _RAD); mSinZ1 = sin(rotation[3] * _RAD);
    mCosY0 = cos(rotation[2] * _RAD); mSinY0 = sin(rotation[2] * _RAD);
    mCosX0 = cos(rotation[1] * _RAD); mSinX0 = sin(rotation[1] * _RAD);
    mCosZ0 = cos(rotation[0] * _RAD); mSinZ0 = sin(rotation[0] * _RAD);
    
    bool bDiff = FALSE;
    if (mappingType == _TEX_MAPPING_VALUE && 
	mPrevDoubleIndex != mCurrDoubleIndex) bDiff = TRUE;
    if (mPrevMapScale != mCurrMapScale) bDiff = TRUE;
    if (mPrevMapTurb != mCurrMapTurb) bDiff = TRUE;
    if (mPrevMappingType != mCurrMappingType) bDiff = TRUE;
    if (mPrevPatternType != mCurrPatternType) bDiff = TRUE;
    if (mPrevMapAxis != mCurrMapAxis) bDiff = TRUE;
    if (mPrevMappingCoord != mCurrMappingCoord) bDiff = TRUE;
    LOOP_XYZ if (mPrevPos[xyz] != mCurrPos[xyz]) bDiff = TRUE;
    LOOP_XYZ if (mPrevOffset[xyz] != mCurrOffset[xyz]) bDiff = TRUE;
    LOOP_XYZ if (mPrevScale[xyz] != mCurrScale[xyz]) bDiff = TRUE;
    for (int i = 0; i < 4; i++)
	if (mPrevRotate[i] != mCurrRotate[i]) bDiff = TRUE;
    
    return bDiff;
}

void TextureMapping_cls::
MakeTexture1D(int colorType, bool bMultiply){
    if (colorType == 0) return;
    
    if (mTexId == -1)
	glGenTextures(1, &mTexId);

    mColorRefreshId = GCColoring.mpCCustomColor->mRefreshId;
    mPrevColorType = colorType;
        
    unsigned char ATexData1D[256 * 4];
    for (int i = 0; i < 256; i++){
	int offset = i * 4;
	float value = (float)i/256;
	LOOP_RGB
	    ATexData1D[offset + rgb] = 255 * GCColoring.mpCCustomColor->
	    GetCustomColor(value, rgb, TRUE, colorType - 1);
	ATexData1D[offset + 3] = mpCTexPropLink->GetLinkedParam() * 255;//128;//255;
	if (bMultiply == TRUE)
	    LOOP_RGB
		ATexData1D[offset + rgb] *= value;
    }
    
    glBindTexture(GL_TEXTURE_1D, mTexId);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA, 256, 0, 
		 GL_RGBA, GL_UNSIGNED_BYTE, ATexData1D);
}

void TextureMapping_cls::
Bind1DTexture(int patternType, int colorType, bool bShade, GLuint wrapType,
	      bool bMultiply){
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_TEXTURE_1D);
    
    if (mPrevColorType != colorType ||
	mPrevPatternType != patternType || 
	mPrevAlpha != mpCTexPropLink->GetLinkedParam() ||
	GCColoring.mpCCustomColor->IsRefreshed(mColorRefreshId) == TRUE)
	MakeTexture1D(colorType, bMultiply);
    
    glBindTexture(GL_TEXTURE_1D, mTexId);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S , wrapType);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    if (bShade == FALSE) glEnable(GL_COLOR_MATERIAL);
    glColor4f(1, 1, 1, 1);
}

void TextureMapping_cls::
SetMaterial1D(double diffuseFactor, double ambientFactor){
    float diffuse[4] = {1, 1, 1, 1}, ambient[4] = {1, 1 ,1 ,1};
    LOOP_RGB diffuse[rgb] *= diffuseFactor;
    LOOP_RGB ambient[rgb] *= ambientFactor;
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
}

void TextureMapping_cls::
SetGLColoring(TexProperty_cls *pCTexProp, float value){
    
    float shininess = 0.04 / pCTexProp->mCRoughness.mBody * 255;
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, 
		 &pCTexProp->GetV4AmbientColor(value)[0]);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, 
		 &pCTexProp->GetV4DiffuseColor(value)[0]);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, 
		 &pCTexProp->GetV4SpecularColor(value)[0]);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shininess);
}

void TextureMapping_cls::
Configure1DMapping(TexProperty_cls *pCTexProp,
		   double offset[3], double scale[3],
		   bool bShade, bool bUseMapping, bool bMultiply){

    mpCTexPropLink = pCTexProp;
    mColorType = pCTexProp->mCColorType.mBody;
    
    if (bUseMapping == FALSE) return;
    
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_TEXTURE_3D);
    glDisable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, 0);
    glBindTexture(GL_TEXTURE_3D, 0);

    if (mColorType != 0)
	if (pCTexProp->mCMappingType.mBody == _TEX_MAPPING_VALUE ||
	    (pCTexProp->mCMappingType.mBody == _TEX_MAPPING_PATTERN &&
	     (pCTexProp->IsUVPatternReady() == TRUE))){
	    Bind1DTexture( pCTexProp->mCPatternType.mBody, 
			   mColorType, bShade, /*GL_REPEAT*/GL_CLAMP_TO_EDGE,
			   bMultiply);
	    
	    MakeUV1D(pCTexProp, offset, scale);
	    
	    if (bShade == TRUE)
		SetMaterial1D(pCTexProp->mCDiffuse.mBody,
			      pCTexProp->mCAmbient.mBody);
	}
}

void  TextureMapping_cls::
GetGlobalPos(float globPos[3], float *pVertex){
    float tmpPos0[3];

    double cosZ1 = mCosZ1, sinZ1 = -mSinZ1;
    double cosY0 = mCosY0, sinY0 = -mSinY0;
    double cosX0 = mCosX0, sinX0 = -mSinX0;
    double cosZ0 = mCosZ0, sinZ0 = -mSinZ0;
    
    globPos[X] = pVertex[X] * mCurrScale[X] + mCurrOffset[X];
    globPos[Y] = pVertex[Y] * mCurrScale[Y] + mCurrOffset[Y]; 
    //globPos[Z] = mVVVerticesOpt[index][i+Z] * scale[Z] + offset[Z]; 
    
    tmpPos0[X] = globPos[X] * cosZ1 + globPos[Y] * sinZ1;
    tmpPos0[Y] = - globPos[X] * sinZ1 + globPos[Y] * cosZ1;
    tmpPos0[Z] = pVertex[Z] * mCurrScale[Z] + mCurrOffset[Z]; 
    
    globPos[X] = - tmpPos0[Z] * sinY0 + tmpPos0[X] * cosY0;
    globPos[Y] = tmpPos0[Y];
    globPos[Z] = tmpPos0[Z] * cosY0 + tmpPos0[X] * sinY0;
    
    tmpPos0[X] = globPos[X];
    tmpPos0[Y] = globPos[Y] * cosX0 + globPos[Z] * sinX0;
    tmpPos0[Z] = - globPos[Y] * sinX0 + globPos[Z] * cosX0;
    
    globPos[X] = tmpPos0[X] * cosZ0 + tmpPos0[Y] * sinZ0;
    globPos[Y] = - tmpPos0[X] * sinZ0 + tmpPos0[Y] * cosZ0;
    globPos[Z] = tmpPos0[Z];

    if (mCurrMapTurb != 0){
	// ??? ϊ@́H ???
	double fpos[3] = { globPos[X], globPos[Y], globPos[Z] };
	TURB Turb;
	LOOP_XYZ Turb.Turbulence[xyz] = 1;
	Turb.Warp_Type = CLASSIC_TURB_WARP;
	Turb.Octaves = 6;
	Turb.Lambda = 2.0;
	Turb.Omega = 0.5;
	
	double result[3];
	DTurbulence(result, fpos, &Turb);
	globPos[X] += result[X] * mCurrMapTurb;
	globPos[Y] += result[Y] * mCurrMapTurb;
	globPos[Z] += result[Z] * mCurrMapTurb;
    }
}

#define MACRO_SETFPOS_GLOBAL()						\
    double fpos[3] = { (fieldPos[X] - pivot[X]) / mCurrMapScale,	\
		       (fieldPos[Y] - pivot[Y]) / mCurrMapScale,	\
		       (fieldPos[Z] - pivot[Z]) / mCurrMapScale};
#define MACRO_SETFPOS_LOCAL()						\
    double fpos[3] = { (fieldPos[X] - pivot[X]) * mCurrScale[X] / mCurrMapScale, \
		       (fieldPos[Y] - pivot[Y]) * mCurrScale[Y] / mCurrMapScale, \
		       (fieldPos[Z] - pivot[Z]) * mCurrScale[Z] / mCurrMapScale};
    
float TextureMapping_cls::
GetMappingValueGlobalAgate(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_GLOBAL();

    TURB Turb;
    LOOP_XYZ Turb.Turbulence[xyz] = 1;
    Turb.Warp_Type = CLASSIC_TURB_WARP;
    Turb.Octaves = 6;
    Turb.Lambda = 2.0;
    Turb.Omega = 0.5;
    
    return agate_pattern_MOD(fpos, &Turb, 2);
}

float TextureMapping_cls:: //FIX_ME!!
GetMappingValueLocalAgate(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_LOCAL();

    TURB Turb;
    LOOP_XYZ Turb.Turbulence[xyz] = 1;
    Turb.Warp_Type = CLASSIC_TURB_WARP;
    Turb.Octaves = 6;
    Turb.Lambda = 2.0;
    Turb.Omega = 0.5;
    
    return agate_pattern_MOD(fpos, &Turb, 2);
}

float TextureMapping_cls::
GetMappingValueGlobalBozo(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_GLOBAL();
    return Noise(fpos, 2);
}

float TextureMapping_cls:: //FIX_ME!!
GetMappingValueLocalBozo(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_LOCAL();
    return Noise(fpos, 2);
}

float TextureMapping_cls::
GetMappingValueGlobalChecker(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_GLOBAL();
    float value;
    if (((int)fpos[X]) % 2 == 0) value = 0.01; else value = 0.99;
    if (fpos[X] < 0) value = 1.0 - value;
    if (((int)fpos[Y]) % 2 == 0) value = 1.0 - value;
    if (fpos[Y] < 0) value = 1.0 - value;
    if (((int)fpos[Z]) % 2 == 0) value = 1.0 - value;
    if (fpos[Z] < 0) value = 1.0 - value;
    return value;
}

float TextureMapping_cls::
GetMappingValueLocalChecker(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_LOCAL();
    float value;
    if (((int)fpos[X]) % 2 == 0) value = 0.01; else value = 0.99;
    if (fpos[X] < 0) value = 1.0 - value;
    if (((int)fpos[Y]) % 2 == 0) value = 1.0 - value;
    if (fpos[Y] < 0) value = 1.0 - value;
    if (((int)fpos[Z]) % 2 == 0) value = 1.0 - value;
    if (fpos[Z] < 0) value = 1.0 - value;
    return value;
}

float TextureMapping_cls::
GetMappingValueGlobalCylindrical(float fieldPos[3], float pivot[3], int axis){
    MACRO_SETFPOS_GLOBAL();
    float value;
    switch(axis){
    case X:
	value = sqrt(fpos[Y] * fpos[Y] + fpos[Z] * fpos[Z]);
	break;
    case Y:
	value = sqrt(fpos[Z] * fpos[Z] + fpos[X] * fpos[X]);
	break;
    case Z:
	value = sqrt(fpos[X] * fpos[X] + fpos[Y] * fpos[Y]);
	break;
    }
    return value - floor(value);
}

float TextureMapping_cls:: 
GetMappingValueLocalCylindrical(float fieldPos[3], float pivot[3], int axis){
    MACRO_SETFPOS_LOCAL();
    float value;
    switch(axis){
    case X:
	value = sqrt(fpos[Y] * fpos[Y] + fpos[Z] * fpos[Z]);
	break;
    case Y:
	value = sqrt(fpos[Z] * fpos[Z] + fpos[X] * fpos[X]);
	break;
    case Z:
	value = sqrt(fpos[X] * fpos[X] + fpos[Y] * fpos[Y]);
	break;
    }
    return value - floor(value);
}

float TextureMapping_cls::
GetMappingValueGlobalDents(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_GLOBAL();
    double value = Noise(fpos, 2);
    return value * value * value;
}

float TextureMapping_cls:: 
GetMappingValueLocalDents(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_LOCAL();
    double value = Noise(fpos, 2);
    return value * value * value;
}

float TextureMapping_cls::
GetMappingValueGlobalGranite(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_GLOBAL();
    return granite_pattern(fpos, 2);
}

float TextureMapping_cls:: 
GetMappingValueLocalGranite(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_LOCAL();
    return granite_pattern(fpos, 2);
}

float TextureMapping_cls::
GetMappingValueGlobalHexagon(float fieldPos[3], float pivot[3], int axis){
    MACRO_SETFPOS_GLOBAL();
    if (axis == X){
	double tmp = fpos[X];
	fpos[X] = fpos[Y];
	fpos[Y] = tmp;
	fpos[Z] *= -1;
    }
    if (axis == Y)
	fpos[Z] *= -1;
    if (axis == Z){
	double tmp = fpos[Z];
	fpos[Z] = fpos[Y];
	fpos[Y] = tmp;
    }

    return Hexagon_cls::GetPattern(fpos);
    //return hexagon_pattern(fpos) * 0.49 + 0.01;
}

float TextureMapping_cls:: 
GetMappingValueLocalHexagon(float fieldPos[3], float pivot[3], int axis){
    MACRO_SETFPOS_LOCAL();
    if (axis == X){
	double tmp = fpos[X];
	fpos[X] = fpos[Y];
	fpos[Y] = tmp;
	fpos[Z] *= -1;
    }
    if (axis == Z){
	double tmp = fpos[Z];
	fpos[Z] = fpos[Y];
	fpos[Y] = tmp;
    }
    
    return Hexagon_cls::GetPattern(fpos);
    //return hexagon_pattern(fpos) * 0.49 + 0.01;
}

float TextureMapping_cls::
GetMappingValueGlobalGradient(float fieldPos[3], float pivot[3], int axis){
    MACRO_SETFPOS_GLOBAL();
    float value = fpos[axis];
    return value - floor(value);
}

float TextureMapping_cls::
GetMappingValueLocalGradient(float fieldPos[3], float pivot[3], int axis){
    MACRO_SETFPOS_LOCAL();
    float value = fpos[axis];
    return value - floor(value);
}

float TextureMapping_cls::
GetMappingValueGlobalRadial(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_GLOBAL();
    float value = sqrt(fpos[X] * fpos[X] +
		       fpos[Y] * fpos[Y] +
		       fpos[Z] * fpos[Z]);
    return value - floor(value);
}

float TextureMapping_cls::
GetMappingValueLocalRadial(float fieldPos[3], float pivot[3]){
    MACRO_SETFPOS_LOCAL();
    float value = sqrt(fpos[X] * fpos[X] +
		       fpos[Y] * fpos[Y] +
		       fpos[Z] * fpos[Z]);
    return value - floor(value);
}

float TextureMapping_cls::
GetMappingValueGlobal(float fieldPos[3]){
    // ̎ŏ̓o֐ɓĂ͂

    double value = 0.5;
    switch(mCurrPatternType){
    case _TEX_PATTERN_AGATE:
	value = GetMappingValueGlobalAgate(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_BOZO:
	value = GetMappingValueGlobalBozo(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_CHECKER:
	value = GetMappingValueGlobalChecker(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_CYLINDRICAL:
	value = GetMappingValueGlobalCylindrical(fieldPos, mCurrPos, mCurrMapAxis);
	break;
    case _TEX_PATTERN_DENTS:
	value = GetMappingValueGlobalDents(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_GRADIENT:
	value = GetMappingValueGlobalGradient(fieldPos, mCurrPos, mCurrMapAxis);
	break;
    case _TEX_PATTERN_GRANITE:
	value = GetMappingValueGlobalGranite(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_HEXAGON:
	value = GetMappingValueGlobalHexagon(fieldPos, mCurrPos, mCurrMapAxis);
	break;
    case _TEX_PATTERN_RADIAL:
	value = GetMappingValueGlobalRadial(fieldPos, mCurrPos);
	break;
    }
    return value;
}

float TextureMapping_cls::
GetMappingValueLocal(float fieldPos[3]){
    // ̎ŏ̓o֐ɓĂ͂
    float localPivot[3];
    LOOP_XYZ localPivot[xyz] = (mCurrPos[xyz] - mCurrOffset[xyz]) / mCurrScale[xyz];
    double value = 0.5;
    switch(mCurrPatternType){
    case _TEX_PATTERN_AGATE:
	value = GetMappingValueLocalAgate(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_BOZO:
	value = GetMappingValueLocalBozo(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_CHECKER:
	value = GetMappingValueLocalChecker(fieldPos, localPivot);
	break;
    case _TEX_PATTERN_CYLINDRICAL:
	value = GetMappingValueLocalCylindrical(fieldPos, localPivot, mCurrMapAxis);
	break;
    case _TEX_PATTERN_DENTS:
	value = GetMappingValueLocalDents(fieldPos, localPivot);
	break;
    case _TEX_PATTERN_GRADIENT:
	value = GetMappingValueLocalGradient(fieldPos, localPivot, mCurrMapAxis);
	break;
    case _TEX_PATTERN_GRANITE:
	value = GetMappingValueLocalGranite(fieldPos, mCurrPos);
	break;
    case _TEX_PATTERN_HEXAGON:
	value = GetMappingValueLocalHexagon(fieldPos, mCurrPos, mCurrMapAxis);
	break;
    case _TEX_PATTERN_RADIAL:
	value = GetMappingValueLocalRadial(fieldPos, localPivot);
	break;
    }
    return value;
}

double TextureMapping_cls::
GetAveragedValue(unsigned int size[3], double pos[3], unsigned char *pData){

    if (pos[X] < 0 || pos[Y] < 0 || pos[Z] < 0) return 0;
    
    unsigned int sizeX = size[X];
    unsigned int sizeY = size[Y];
    unsigned int sizeZ = size[Z];

    int iX = pos[X] * sizeX;
    int iY = pos[Y] * sizeY;
    int iZ = pos[Z] * sizeZ;
    
    if (iX > sizeX - 1 || iY > sizeY - 1 || iZ > sizeZ - 1)
	return 0;

    double fx = pos[X] * sizeX - iX;
    double fy = pos[Y] * sizeY - iY;
    double fz = pos[Z] * sizeZ - iZ;
    
    int lBaseIndex = (iZ * sizeY + iY ) * sizeX + iX;

    if (iX == sizeX - 1 || iY == sizeY - 1 || iZ == sizeZ - 1)
	return pData[lBaseIndex];
    
    int lOffsetZ = sizeX * sizeY;
    //int lOffsetY = sizeX;
    
    double P000(pData[lBaseIndex]);
    double P001(pData[lBaseIndex+1]);
    double P010(pData[lBaseIndex+sizeX]);
    double P011(pData[lBaseIndex+sizeX+1]);
    double P100(pData[lBaseIndex+lOffsetZ]); 
    double P101(pData[lBaseIndex+lOffsetZ+1]);
    double P110(pData[lBaseIndex+lOffsetZ+sizeX]);
    double P111(pData[lBaseIndex+lOffsetZ+sizeX+1]);
    
    return  
	(1.0f-fz) * ( (1.0f-fy) * ((1.0f-fx)*P000 + fx*P001)  +
		      fy        * ((1.0f-fx)*P010 + fx*P011)) +
	fz        * ( (1.0f-fy) * ((1.0f-fx)*P100 + fx*P101)  +
		      fy        * ((1.0f-fx)*P110 + fx*P111));
}
