// "clsTextureMapping.h"
// (c-basic-offset 4)

#ifndef _POV_TEXTURE_MAPPING_H_
#define _POV_TEXTURE_MAPPING_H_

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

class TexProperty_cls;

class TextureMapping_cls{
 public:
    TextureMapping_cls();
    ~TextureMapping_cls();

    int mColorRefreshId;
    GLuint  mTexId;
    double mPrevDoubleIndex;
    int mPrevMappingType, mPrevColorType, mPrevMappingCoord;
    int mPrevPatternType, mPrevMapAxis, mPrevAlpha;
    float mPrevPos[3], mPrevMapScale, mPrevMapTurb;
    float mPrevOffset[3], mPrevScale[3];
    double mPrevRotate[4];

    double mCurrDoubleIndex;
    int mCurrMappingType, mCurrColorType, mCurrMappingCoord;
    int mCurrPatternType, mCurrMapAxis;
    float mCurrPos[3], mCurrMapScale, mCurrMapTurb;
    float mCurrOffset[3], mCurrScale[3];
    double mCurrRotate[4];

    double mCosZ0, mSinZ0;
    double mCosX0, mSinX0;
    double mCosY0, mSinY0;
    double mCosZ1, mSinZ1;

    TexProperty_cls *mpCTexPropLink;
    int  mColorType;
    
    void CopyCurrToPrev(TexProperty_cls *pCTexProp);
    bool IsParameterIsNewOne(double doubleIndex, int mappingType, 
			     int patternType, int mapAxis, int mappingCoord,
			     float pos[3], float mapscale, float mapturb, 
			     double offset[3], double scale[3],
			     double rotation[4]);
    virtual void MakeUV1D(TexProperty_cls *pCTexProp,
			  double offset[3], double scale[3]) = 0;
    void MakeTexture1D(int colorType, bool bMultiply);
    void Bind1DTexture(int patternType, int colorType, bool bShade, 
		       GLuint wrapType, bool bMultiply);
    void SetMaterial1D(double diffuseFactor, double ambientFactor);
    void SetGLColoring(TexProperty_cls *pCTexProp, float value = 0);
    void Configure1DMapping(TexProperty_cls *pCTexProp,
			    double offset[3], double scale[3],
			    bool bShade = TRUE, bool bUseMapping = TRUE,
			    bool bMultiply = FALSE);
    virtual bool DoesUseMapping() = 0;
    
    void GetGlobalPos(float globPos[3], float *pVertex);
    
    float GetMappingValueGlobalAgate(float fieldPos[3], float pivot[3]);
    float GetMappingValueLocalAgate(float fieldPos[3], float pivot[3]);
    
    float GetMappingValueGlobalBozo(float fieldPos[3], float pivot[3]);
    float GetMappingValueLocalBozo(float fieldPos[3], float pivot[3]);
    
    float GetMappingValueGlobalChecker(float fieldPos[3], float pivot[3]);
    float GetMappingValueLocalChecker(float fieldPos[3], float pivot[3]);

    float GetMappingValueGlobalDents(float fieldPos[3], float pivot[3]);
    float GetMappingValueLocalDents(float fieldPos[3], float pivot[3]);

    float GetMappingValueGlobalCylindrical(float fieldPos[3], float pivot[3], int axis);
    float GetMappingValueLocalCylindrical(float fieldPos[3], float pivot[3], int axis);
    
    float GetMappingValueGlobalGradient(float fieldPos[3], float pivot[3], int axis);
    float GetMappingValueLocalGradient(float fieldPos[3], float pivot[3], int axis);

    float GetMappingValueGlobalGranite(float fieldPos[3], float pivot[3]);
    float GetMappingValueLocalGranite(float fieldPos[3], float pivot[3]);

    float GetMappingValueGlobalHexagon(float fieldPos[3], float pivot[3], int axis);
    float GetMappingValueLocalHexagon(float fieldPos[3], float pivot[3], int axis);

    float GetMappingValueGlobalRadial(float fieldPos[3], float pivot[3]);
    float GetMappingValueLocalRadial(float fieldPos[3], float pivot[3]);
    
    
    float GetMappingValueGlobal(float fieldPos[3]);
    float GetMappingValueLocal(float fieldPos[3]);

    static double GetAveragedValue
	(unsigned int size[3], double pos[3], unsigned char *pData);
};

#endif
