// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsTextureProperty.h"

#include "clsTextureProperty.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include "../COMMON/clsFilename.h"
#include "../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../FILTER/clsFilterWithParameter.h"
#include "../PARAMETER/clsParameters.h"
#include "../POVRAY/clsPovRay.h"
#include "../POVRAY/clsPovVolume.h"
#include "../WIDGETCTRL/clsAtoiComboCtrl.h"
#include "../WIDGETCTRL/clsAtoWXSTextCtrl.h"
#include "../WIDGETCTRL/clsGraphPanel.h"
#include "../WIDGETCTRL/EX/clsButtonEX.h"
#include "../WIDGETCTRL/EX/clsStaticTextEX.h"
#include "../WIDGETCTRL/EX/clsColourDialogEX.h"

#include "../MYAPI/MsgBox.h"

extern wxColourDialogEX* GpCClrDialog;

#include <wx/notebook.h>

#define ENDLWX wxT("\n")
#define X 0
#define Y 1
#define Z 2
#define R 0
#define G 1
#define B 2
#ifdef WIN32
#define _MEM_PANEL_OFFSET 14
#else
#define _MEM_PANEL_OFFSET 10
#endif

#define LOOP_XYZ for (int xyz=0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb=0; rgb < 3; rgb++)
#define wxCmdEvt wxCommandEvent

#define _RAD 3.14159 / 180

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

// ÓIɃCxgт
BEGIN_EVENT_TABLE(TexPropertyGUI_cls, wxDialog)
     EVT_SIZE(wxDialogEXKey::OnSize)
     EVT_BUTTON(btnTexDefault, TexPropertyGUI_cls::OnButtonDefault)
END_EVENT_TABLE()

TexPropertyGUI_cls::
TexPropertyGUI_cls(TexPropertyCtrl_cls* pCParent,
		   wxWindow *pCWindow,
		   wxPoint pos, wxSize size) :
    wxDialogEXKey(pCWindow, wxID_ANY, _("Texture Property Configure"),
		  pos, size, pCParent->mpBitmapLink)
{
    mspCParent = pCParent;
    mspCThis = this;
    //SetCloseButtonID(btnTexClose);
    SetScreenRectPos(&pCParent->mCScreenRect);

    // TCU[
    wxBoxSizer       *pBoxSizerSRC   = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerTOPMain = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerBTM   = new wxBoxSizer(wxHORIZONTAL);

    SetMNumMain(pBoxSizerTOPMain);

    mpCBook1 = new wxNotebook(this, ID_BOOK_TEXTURE1);
			      
    SetMNumBump(NULL, mpCBook1);
    SetMNumMirror(NULL, mpCBook1);
    SetMNumInterior(NULL, mpCBook1);
    SetMNumOptions(NULL, mpCBook1);

    mpCBook1->AddPage(mpCMNumBump, _("Bump Map"));
    mpCBook1->AddPage(mpCMNumMirror, _("Mirror"));
    mpCBook1->AddPage(mpCMNumInterior, _("Interior"));
    mpCBook1->AddPage(mpCMNumOptions, _("Options"));

    pBoxSizerBTM->Add( -1, -1, 1, wxEXPAND);
    
    pBoxSizerSRC->Add (pBoxSizerTOPMain, 0, wxEXPAND);
    pBoxSizerSRC->Add (mpCBook1, 0, wxEXPAND);

    pBoxSizerSRC->Add (pBoxSizerBTM, 0, wxEXPAND);

    SetSizerAndFit(pBoxSizerSRC);

    // -----

    int minimumWidth = GetClientSize().GetWidth() - 8;
    minimumWidth = mpCMNumMain->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumBump->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumMirror->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumInterior->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumOptions->GetPaddingWidth(minimumWidth);
    mpCMNumMain->SetPaddingWidth(minimumWidth);
    mpCMNumBump->SetPaddingWidth(minimumWidth);
    mpCMNumMirror->SetPaddingWidth(minimumWidth);
    mpCMNumInterior->SetPaddingWidth(minimumWidth);
    mpCMNumOptions->SetPaddingWidth(minimumWidth);

    //pCButtonClose->SetFocus();
}

TexPropertyGUI_cls::
~TexPropertyGUI_cls(){
    mspCParent->mpCGUI = NULL;
}

void TexPropertyGUI_cls::
SetMNumMain(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumMain, this);
    pCurr->mComboWidth = 80;
    pCurr->mButtonWidth = 80;

    pCurr->mTextWidth = 60;
    wxString colorList[1] = {wxT("Mono Color")};

    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth/*, _("Base Color")*/);
    mIndexColorType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyColorType,
	 wxT(""), TRUE, 1, colorList);
       
    mpCPanelColorBar = new PanelColorBar_cls(pCurr, wxID_ANY, wxDefaultPosition, wxSize(-1, 20));
    mpCPanelColorBar->mbDisplayCurrent = FALSE;
    mpCPanelColorBar->mBorderSizeOffset[0] = 1;
    mpCPanelColorBar->SetFunction(&GetColorR);
    mpCPanelColorBar->SetFunction(&GetColorG);
    mpCPanelColorBar->SetFunction(&GetColorB);
    mpCPanelColorBar->mOffset[0] = 4;
    mpCPanelColorBar->mpClue = this;

    mIndexColorBar = pCurr->CombinePointer(_GP_WXPANEL, mpCPanelColorBar);

    pCurr->mTextWidth = 80;
    mIndexMappingType = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    wxString mappingList[4] = 
	{_("by value"), _("by pattern"), _("by imagefile"), wxT("Plugins")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyMappingType,
	 wxT(""), TRUE, 4, mappingList);
    wxString coordList[2] = 
	{_("global mappng"), _("local mapping")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyMappingCoord,
	 wxT(""), TRUE, 2, coordList);

    mIndexImagePath = pCurr->CombinePointer
	(_GP_WXSTRING_PATH, &mspCParent->mstrDummyImagePath, wxT(""), TRUE);
    mIndexImageType = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    wxString imageMappingList[3] = 
	{_("planer map"), _("spherical map"), _("cylingder map") };
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyImageMappingType,
	 wxT(""), TRUE, 3, imageMappingList);
    wxString imageColoringList[2] = 
	{_("use image color"), _("use color curve") };
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyImageColoringType,
	 wxT(""), TRUE, 2, imageColoringList);
        
    wxString patternList[13] = 
	{wxT("+agate"), _("aoi"), _("+bozo"), wxT("brick"), wxT("cells"), _("*checker"), wxT("crackle"), _("*cylindrical"), _("+dents"), _("*gradient"), _("+granite"), _("*hexagon"), _("*radial")};
    mIndexPatternType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyPatternType,
	 _("Pattern"), TRUE, 13, patternList);
    
    wxString pluginList[1] = 
	{wxT("grasstex.inc")};
    mIndexPluginType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyPluginType,
	 _("Plugin Pattern"), TRUE, 1, pluginList);
    
    boxWidth = 3;
    mIndexMappingPos = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Origin"));
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyMappingPos[X]);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyMappingPos[Y]);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyMappingPos[Z]);
    
    boxWidth = 2;
    mIndexMappingScaleAndTurb = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Scale/Turb"));
    mIndexMappingScale = pCurr->CombinePointer
	(_GP_FLOAT, &mspCParent->mDummyMappingScale, wxT("Scale"));
    mIndexMappingTurb = pCurr->CombinePointer
	(_GP_FLOAT, &mspCParent->mDummyMappingTurb, wxT("Turb"));

    mIndexMappingAxisAndRepeat = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Axis/Repeat"));
    //wxString patternAxis[3] = { wxT("x"), wxT("y"), wxT("z") };
    mIndexMappingScale = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyMappingAxis, wxT("Axis"),
	 TRUE, 3, /*patternAxis*/PovSerialize_cls::mAstrAxis);
    mIndexMappingRepeat = pCurr->CombinePointer
	(_GP_BOOL, &mspCParent->mbDummyMappingRepeat, wxT("Repeat"));
    
    mIndexPluginMappingOptional = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Optional"));

    pCurr->CombinePointer(_GP_BOOL, &mspCParent->mbDummyMappingOptional);
    
    mIndexPluginHelp = pCurr->CombinePointerButton(wxID_ANY, _("Plugin Help"));
    
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Specular"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummySpecular);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummySpecular);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Diffuse"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyDiffuse);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyDiffuse);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Ambient"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyAmbient);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyAmbient);

    pCurr->CombinePointer(_GP_STATIC_LINE);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Roughness"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyRoughness);
    pCurr->SetRange(1, 100, _MN_LASTITEM, 200);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyRoughness);
    
    wxString transmitList[3] = 
	{_("no transmit"), _("transmit"), _("filtered")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyTransmitType,
	 _("Transmit Type"), TRUE, 3, transmitList);

    mIndexAlpha = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Alpha"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyAlpha);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyAlpha);

    wxString renderTypeList[2] = {_("0: Alpha Blend"), _("1: Add") };
    mIndexRenderType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyRenderType,
	 _("Render Type"), TRUE, 2, renderTypeList);

#ifdef WIN32
    pCurr->mMinimumWidth = 240;
#else
    pCurr->mMinimumWidth = 300;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);
    
    ((AtoWXSTextPathCtrl_cls*)pCurr->mVCPointers[mIndexImagePath].mpPointerCtrl)->
	mpstrDefaultDirLink = &TexProperty_cls::mspCPovRayLink->mCConfWorkingDir.msBody;
    ((AtoWXSTextPathCtrl_cls*)pCurr->mVCPointers[mIndexImagePath].mpPointerCtrl)->
	mstrFilter = wxT("Image File|*.png;*.PNG;*.jpeg;*.JPEG;*.jpg;*.JPG;*.gif;*.GIF;*.hdr;*.HDR;*.exr;*.EXR;*.tga;*.TGA;*.tif;*.TIF;*.tiff;*.TIFF)|all|*");
    
    Connect(pCurr->mVCPointers[mIndexPluginHelp].mID, 
	    wxEVT_COMMAND_BUTTON_CLICKED, 
	    wxCommandEventHandler(TexPropertyGUI_cls::OnButtonHelp));

    pCurr->SetFuncHook(&HookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&HookAssign, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&HookCombo, _MN_HOOK_COMBOBOX);
    pCurr->SetFuncHook(&HookPanelClick, _MN_HOOK_PANELCLICK);
}

void TexPropertyGUI_cls::
SetMNumInterior(wxBoxSizer *pBoxSizer, wxWindow *pCWindow){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumInterior, pCWindow);

    pCurr->mTextWidth = -1;
    pCurr->mTextWidth = 80;
    pCurr->mButtonWidth = 80;

    int boxWidth = 2;
    mIndexSSS = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("SSS Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummySubsurfaceFactor);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummySubsurfaceFactor);
    
    mIndexBackDiffuse = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Back Diffuse"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyDiffuseBack);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyDiffuseBack);

    mIndexFadeClrBtn = pCurr->CombinePointerButton(wxID_ANY, _("Fade_Color"));
    pCurr->AlignRight();

    mIndexFadeDistance = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Fade_Distance"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyFadeDistance);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyFadeDistance);

    mIndexIOR = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("IOR"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyIOR);
    pCurr->SetRange(0, 250, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyIOR);

    mIndexDispersion = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Dispersion"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyDispersion);
    pCurr->SetRange(0, 250, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyDispersion);
    
    pCurr->FinishSetting(pBoxSizer, TRUE);

    Connect(pCurr->mVCPointers[mIndexFadeClrBtn].mID, 
	    wxEVT_COMMAND_BUTTON_CLICKED, 
	    wxCommandEventHandler(TexPropertyGUI_cls::OnButtonFadeColor));
    
    pCurr->SetFuncHook(&HookAssign, _MN_HOOK_ASSIGN);

    pCurr->mVCPointers[mIndexFadeDistance].
	mpCTextLabel->SetToolTip(_("Diffuse for the back side of the surface."));
    pCurr->mVCPointers[mIndexFadeDistance].
	mpCTextLabel->SetToolTip(_("Lower value means denser color, but 0 is special for no color"));
    pCurr->mVCPointers[mIndexIOR].
	mpCTextLabel->SetToolTip(_("Default value is 1, where no effect apprears"));
    pCurr->mVCPointers[mIndexDispersion].
	mpCTextLabel->SetToolTip(_("Default value is 1, where no effect apprears"));
}

void TexPropertyGUI_cls::
SetMNumBump(wxBoxSizer *pBoxSizer, wxWindow *pCParent){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumBump, pCParent);

    pCurr->mTextWidth = -1;
    pCurr->mComboWidth = 100;

    wxString normalPatternList[12] = 
	{_("no normal pattern"), wxT("agate"), _("bozo"), wxT("bumps"), wxT("brick"), wxT("cells"), _("checker"), wxT("crackle"), /*wxT("cylindrical"), */wxT("dents"), wxT("facets"), /*wxT("mandel"), wxT("julia"), wxT("hexagon"),*/ wxT("marble"), wxT("ridged_mf")};
    
    int boxWidth = 2;
    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, 
			      &mspCParent->mDummyNormalPattern[i],
			      _("Normal"), TRUE, 12, normalPatternList);
	
	mIndexNormalDegree[i] = pCurr->CombinePointer
	    (_GP_BOXSIZERHORIZONTAL, &boxWidth, _("degree"));
	pCurr->CombinePointer
	    (_GP_FLOAT_SLIDER, &mspCParent->mDummyNormalDegree[i]);
	pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
	pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyNormalDegree[i]);
	
	mIndexNormalScale[i] = pCurr->CombinePointer
	    (_GP_BOXSIZERHORIZONTAL, &boxWidth, _("scale"));
	pCurr->CombinePointer
	    (_GP_FLOAT_SLIDER, &mspCParent->mDummyNormalScale[i]);
	pCurr->SetRange(1, 20000, _MN_LASTITEM, 10000);
	pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyNormalScale[i]);

	mIndexNormalTurb[i] = pCurr->CombinePointer
	    (_GP_BOXSIZERHORIZONTAL, &boxWidth, _("turbulence"));
	pCurr->CombinePointer
	    (_GP_FLOAT_SLIDER, &mspCParent->mDummyNormalTurb[i]);
	pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
	pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyNormalTurb[i]);
    }

    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(&HookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&HookCombo, _MN_HOOK_COMBOBOX);
}

void TexPropertyGUI_cls::
SetMNumMirror(wxBoxSizer *pBoxSizer, wxWindow *pCWindow){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumMirror, pCWindow);

    pCurr->mTextWidth = 80;
    pCurr->mComboWidth = 100;
    int boxWidth = 2;
    wxString reflexType[3] = 
	{_("no mirror"), _("mirror like"), _("metal like")};

    mIndexReflexType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mDummyReflectionType,
	 _("Reflex Type"), TRUE, 3, reflexType);
    
    mIndexReflexMin = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("reflex min"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyReflectionMin);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyReflectionMin);
    
    mIndexReflexMax = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("reflex max"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mspCParent->mDummyReflectionMax);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mspCParent->mDummyReflectionMax);
    
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(&HookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&HookCombo, _MN_HOOK_COMBOBOX);

    pCurr->mVCPointers[mIndexReflexType].
	mpCTextLabel->SetToolTip(_("Metal like reflection is affected by the color of material itself."));
    pCurr->mVCPointers[mIndexReflexMin].
	mpCTextLabel->SetToolTip(_("Reflection when viewed from a direction parallel to its normal."));
    pCurr->mVCPointers[mIndexReflexMax].
	mpCTextLabel->SetToolTip(_("Reflection when viewed from shallow angle.\nMany materials are more reflective in shallow angle."));
    
}

void TexPropertyGUI_cls::
SetMNumOptions(wxBoxSizer *pBoxSizer, wxWindow *pCWindow){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumOptions, pCWindow);

    pCurr->mTextWidth = 80;
    pCurr->mComboWidth = 100;
    int boxWidth = 2;
    
    pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->CombinePointer
	(_GP_BOOL, &mspCParent->mDummyImageOn, _("Show Object"));
    pCurr->CombinePointer
	(_GP_BOOL, &mspCParent->mDummyShadowOn, _("Cast Shadow"));
    pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->CombinePointer
	(_GP_BOOL, &mspCParent->mDummyRadiosityOn, _("Calc Radiosity"));
    pCurr->CombinePointer
	(_GP_BOOL, &mspCParent->mDummyReflectionOn, _("Show in Mirror"));
    
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(&HookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&HookAssign, _MN_HOOK_CHECKBOX);
}


void TexPropertyGUI_cls::
OnButtonDefault(wxCommandEvent &event){
    if (MsgBoxYesno(this, _("Query"), _("Set to default ?")) == FALSE) return;
    mspCParent->SetToDefault();
    mspCParent->UpdateItems();
}

void TexPropertyGUI_cls::
OnButtonFadeColor(wxCommandEvent &event){
    mspCParent->OpenFadeColorDLG();
    mspCParent->Apply();
    mspCParent->UpdateItems();
    if (mspCParent->mpCParentMNumPanel != NULL)
	mspCParent->mpCParentMNumPanel->UpdateValues();
}

void TexPropertyGUI_cls::
OnButtonHelp(wxCommandEvent &event){
    wxString strMessage;
    switch(mspCParent->mDummyPluginType){
    case _TEX_PLUGINGS_GRASSTEX:
	strMessage = _("Description of grasstex.inc");
	break;
    }
    MsgBox(this, strMessage, _("Plugin description"));
}

void TexPropertyGUI_cls::
UpdateColorList(){
    wxString pColorList[_CLR_CUSTOM + _CLR_CUSTOM_NUM_MAX];

    // 
    if (((AtoiComboCtrl_cls*)mpCMNumMain->mVCPointers[mIndexColorType].GetPCtrlWindow())->mVChoiceList.size() != GCColoring.mpCCustomColor->mCItemNum.mBody + 1){
	pColorList[0] = wxT("Mono Color");
	for (int i = 0; i < (int)GCColoring.mpCCustomColor->mCItemNum.mBody; i++)
	    pColorList[1 + i] << wxT("Color ") << i;
	mpCMNumMain->mVCPointers[mIndexColorType].CombinePointer
	    (_GP_UNSIGNED_INT_LIST, NULL, wxT(""), TRUE, 1 + GCColoring.mpCCustomColor->mCItemNum.mBody, pColorList);
	mpCMNumMain->mVCPointers[mIndexColorType].UpdateValue(TRUE);
    }

    if (mspCParent->mpCConnected == NULL) return;

    
    if (mspCParent->mpCConnected->mCColorType.mBody == 0)
	mpCPanelColorBar->mstrLabel = _("Mono Color");
    else
	mpCPanelColorBar->mstrLabel = wxT("Color Curve");

    mpCMNumInterior->mVCPointers[mIndexFadeClrBtn].mpConnectColorFloat
	= mspCParent->mDummyFadeColor;
    
    //mpCMNumMain->mVCPointers[mIndexClrBtn].UpdateValue();
    mpCMNumInterior->mVCPointers[mIndexFadeClrBtn].UpdateValue();

    HideAndShow();
}

void TexPropertyGUI_cls::
HideAndShow(){
    bool bMapping = FALSE, bPos = FALSE, bFile = FALSE;
    if (mspCParent->mpCConnected != NULL && 
	mspCParent->mpCConnected->mCColorType.mBody > 0) bMapping = TRUE;
    
    mpCMNumMain->ShowItems(mIndexMappingType, 3, bMapping);
    if (mspCParent->mDummyMappingType == _TEX_MAPPING_PATTERN &&
	(mspCParent->mDummyPatternType == _TEX_PATTERN_RADIAL ||
	 mspCParent->mDummyPatternType == _TEX_PATTERN_GRADIENT))
	bPos = TRUE;
    
    mpCMNumMain->mVCPointers[mIndexMappingPos].Show(bMapping & bPos);
    LOOP_XYZ mpCMNumMain->mVCPointers[mIndexMappingPos+1+xyz].Show(bMapping & bPos);

    if (bMapping == TRUE &&
	mspCParent->mDummyMappingType == _TEX_MAPPING_IMAGE) bFile = TRUE;
    mpCMNumMain->ShowItems(mIndexImagePath, 1, bFile);
    mpCMNumMain->ShowItems(mIndexImageType, 3, bFile);

    bool bPlugin = FALSE;
    if (bMapping == TRUE &&
	mspCParent->mDummyMappingType == _TEX_MAPPING_PLUGINS) bPlugin = TRUE;
    mpCMNumMain->ShowItems(mIndexPluginType, 1, bPlugin);
    mpCMNumMain->ShowItems(mIndexPluginMappingOptional, 3, bPlugin);
    
    bool bPattern = FALSE, bScaling = TRUE;
    if (mspCParent->mDummyMappingType == _TEX_MAPPING_PATTERN) bPattern = TRUE;
    if (mspCParent->mDummyPatternType == _TEX_PATTERN_AOI) bScaling = FALSE;
    mpCMNumMain->ShowItems(mIndexPatternType, 1, (bMapping & bPattern));

    bool bAxis = FALSE;
    if (mspCParent->mDummyPatternType == _TEX_PATTERN_GRADIENT ||
	mspCParent->mDummyPatternType == _TEX_PATTERN_CYLINDRICAL ||
	mspCParent->mDummyPatternType == _TEX_PATTERN_HEXAGON) bAxis = TRUE;
    
    mpCMNumMain->ShowItems(mIndexMappingAxisAndRepeat, 3, (bPattern & bAxis));

    if (bPlugin == TRUE && mspCParent->mDummyPluginType == _TEX_PLUGINGS_GRASSTEX)
	mpCMNumMain->ShowItems(mIndexMappingScaleAndTurb, 3, TRUE);
    else
	mpCMNumMain->ShowItems(mIndexMappingScaleAndTurb, 3, bMapping & (bPos || bPattern) & bScaling);
    
    bool bTransmit = TRUE;
    if (mspCParent->mDummyTransmitType == 0) bTransmit = FALSE;
    mpCMNumMain->ShowItems(mIndexAlpha, 3, bTransmit);
    mpCMNumMain->ShowItems(mIndexRenderType, 1, bTransmit);

    bool bMirror = TRUE;
    if (mspCParent->mDummyReflectionType == 0) bMirror = FALSE;
    mpCMNumMirror->ShowItems(mIndexReflexMin, 3, bMirror);
    mpCMNumMirror->ShowItems(mIndexReflexMax, 3, bMirror);
    
    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	bool bState = TRUE;
	if (mspCParent->mDummyNormalPattern[i] == 0) bState = FALSE;
	if (mspCParent->mpCConnected == NULL) bState = FALSE;
	mpCMNumBump->ShowItems(mIndexNormalDegree[i], 3, bState);
	mpCMNumBump->ShowItems(mIndexNormalScale[i], 3, bState);
	mpCMNumBump->ShowItems(mIndexNormalTurb[i], 3, bState);
    }

    bTransmit = (mspCParent->mDummyTransmitType != 0);
    mpCMNumInterior->mVCPointers[mIndexFadeClrBtn].Show(bTransmit);
    mpCMNumInterior->ShowItems(mIndexFadeDistance, 3, bTransmit);
    mpCMNumInterior->ShowItems(mIndexIOR, 3, bTransmit);
    mpCMNumInterior->ShowItems(mIndexDispersion, 3, bTransmit);

    RecalcSize();
    ResizeDialog();
    Refresh(TRUE);
}

void TexPropertyGUI_cls::
RecalcSize(){
    int minSize = 0;
    for (int i = 0; i < mpCBook1->GetPageCount(); i++){
	mpCBook1->GetPage(i)->Layout();
	wxSize size = mpCBook1->GetPage(i)->GetBestSize();
	if (minSize < size.GetHeight())
	    minSize = size.GetHeight();
    }
    mpCBook1->SetSizeHints(-1, minSize + 30);
}

void TexPropertyGUI_cls::
HookCombo(void *pArg, void *pEvent){
    mspCParent->Apply();
    mspCThis->UpdateColorList();
    mspCThis->HideAndShow();

    wxCommandEvent com1(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshColorBar);
    mspCThis->GetParent()->AddPendingEvent(com1);
    wxCommandEvent com2(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    mspCThis->GetParent()->AddPendingEvent(com2);
}

void TexPropertyGUI_cls::
HookPanelClick(void *pArg, void *pEvent){

    int scrx, scry;
    mspCThis->mpCMNumMain->mVCPointers[mspCThis->mIndexColorBar].GetPCtrlWindow()->GetScreenPosition(&scrx, &scry);
    int x = scrx + ((wxMouseEvent*)pEvent)->GetX();

    if (mspCParent->mDummyColorType != 0){
	GCColoring.mpCCustomColor->mpCGUI->mpCParent->mActiveItem = mspCParent->mDummyColorType - 1;
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemCustomColor);
	mspCThis->GetParent()->AddPendingEvent(com);
	return ;
    }
    else
	mspCParent->OpenColorDLG(x-240, scry+36);
    mspCParent->Apply();
    mspCParent->UpdateItems();
    if (mspCParent->mpCParentMNumPanel != NULL)
	mspCParent->mpCParentMNumPanel->UpdateValues();
}

void TexPropertyGUI_cls::
HookAssign(void *pArg, void *pEvent){
    mspCParent->Apply();
}

double TexPropertyGUI_cls::
GetColorR(double value, void *pClue){
    if (mspCParent->mDummyColorType == 0) return mspCParent->mDummyColor[R];
    return GCColoring.mpCCustomColor->GetCustomColor(value, R, TRUE, mspCParent->mDummyColorType - 1);
}

double TexPropertyGUI_cls::
GetColorG(double value, void *pClue){
    if (mspCParent->mDummyColorType == 0) return mspCParent->mDummyColor[G];
    return GCColoring.mpCCustomColor->GetCustomColor(value, G, TRUE, mspCParent->mDummyColorType - 1);
}

double TexPropertyGUI_cls::
GetColorB(double value, void *pClue){
    if (mspCParent->mDummyColorType == 0) return mspCParent->mDummyColor[B];
    return GCColoring.mpCCustomColor->GetCustomColor(value, B, TRUE, mspCParent->mDummyColorType - 1);
}

TexPropertyCtrl_cls *TexPropertyGUI_cls::mspCParent = NULL;
TexPropertyGUI_cls *TexPropertyGUI_cls::mspCThis = NULL;

// ---------------------------

TexPropertyCtrl_cls::
TexPropertyCtrl_cls(wxWindow *pC, wxBitmap *pBitmap){
    mpBitmapLink = pBitmap;
    mpCConnected = NULL;
    mpCGUI  = NULL;
    mpCParentMNumPanel = NULL;
    mCScreenRect = wxRect(wxDefaultPosition, wxDefaultSize);
    mDummySpecular = 0;
    mDummyDiffuse = 0;
    mDummyDiffuseBack = 0;
    mDummyAmbient = 0;
    mDummyRoughness = 0;
    mDummyTransmitType = 0;
    mDummyAlpha = 1;
    mDummySubsurfaceFactor = 0;
    mDummyRenderType = 0;
    mDummyIOR = 1;
    mDummyDispersion = 1;
    mDummyFadeDistance = 0;

    mDummyReflectionType = 0;
    mDummyReflectionMin = 0;
    mDummyReflectionMax = 0;

    LOOP_RGB mDummyFadeColor[rgb] = 0.5;
    LOOP_RGB mDummySubsurfaceColor[rgb] = 0.5;
    LOOP_RGB mDummyColor[rgb] = 0.5;

    mDummyMappingType = 0;
    mDummyMappingCoord = 0;
    mDummyPatternType = 0;
    mDummyPluginType = 0;

    mDummyImageOn = TRUE; mDummyShadowOn = TRUE;
    mDummyRadiosityOn = TRUE; mDummyReflectionOn = TRUE;
    
    //SetToDefault();
    InitGUI(pC); // GUI ̃CX^X
}

TexPropertyCtrl_cls::
~TexPropertyCtrl_cls(){
    //if (mpCGUI != NULL) delete mpCGUI;
}

void TexPropertyCtrl_cls::
UpdateColorType(){
    if (mpCConnected == NULL) return;
    mDummyColorType = mpCConnected->mCColorType.mBody;
    /*
    if (mpCConnected->mpColorIndexLinked != NULL)
	mDummyColorType = *(mpCConnected->mpColorIndexLinked);
    else
	mDummyColorType = 0;
    */
}

void TexPropertyCtrl_cls::
UpdateColorBar(){
    if (mpCGUI != NULL)	mpCGUI->mpCPanelColorBar->Refresh(FALSE);
}

void TexPropertyCtrl_cls::
ConnectProperty(TexProperty_cls *pCConnected, wxString strItemName,
		MultiNumericalInputsPanel_cls *pCParentMNumPanel){

    mpCConnected = pCConnected;
    mstrItemName = strItemName;
    mpCParentMNumPanel = pCParentMNumPanel;
    if (pCConnected == NULL){
	if (mpCGUI != NULL){
	    mpCGUI->mpCMNumMain->Disable();
	    mpCGUI->mpCMNumBump->Disable();
	    mpCGUI->mpCMNumMirror->Disable();
	    mpCGUI->mpCMNumOptions->Disable();
	    mpCGUI->HideAndShow();
	    UpdateItems(); }
	return;
    }
    UpdateColorType();
    
    mDummySpecular = mpCConnected->mCSpecular.mBody;
    mDummyDiffuse  = mpCConnected->mCDiffuse.mBody;
    mDummyDiffuseBack = mpCConnected->mCDiffuseBack.mBody;
    mDummySubsurfaceFactor = mpCConnected->mCSubsurfaceFactor.mBody;
    mDummyAmbient  = mpCConnected->mCAmbient.mBody;
    mDummyRoughness = mpCConnected->mCRoughness.mBody;
    mDummyTransmitType = mpCConnected->mCTransmitType.mBody;
    mDummyIOR = mpCConnected->mCIOR.mBody;
    mDummyDispersion = mpCConnected->mCDispersion.mBody;
    mDummyFadeDistance = mpCConnected->mCFadeDistance.mBody;

    mDummyReflectionType = mpCConnected->mCReflectionType.mBody;
    mDummyReflectionMin = mpCConnected->mCReflectionMin.mBody;
    mDummyReflectionMax = mpCConnected->mCReflectionMax.mBody;

    UpdateLinkedProperty();
    mDummyRenderType = mpCConnected->mCRenderType.mBody;

    LOOP_RGB{
	mGrayColor[rgb] = 0.7f;
	mDummyFadeColor[rgb] = mpCConnected->mCFadeColor.mpBody[rgb];
	mDummyColor[rgb] = mpCConnected->mCColor.mpBody[rgb];
    }

    mDummyMappingType  = mpCConnected->mCMappingType.mBody;
    mDummyMappingCoord = mpCConnected->mCMappingCoord.mBody;
    mDummyPatternType = mpCConnected->mCPatternType.mBody;
    mDummyPluginType = mpCConnected->mCPluginType.mBody;
    LOOP_XYZ mDummyMappingPos[xyz] = mpCConnected->mCMappingPos.mpBody[xyz];
    mDummyMappingScale = mpCConnected->mCMappingScale.mBody;
    mDummyMappingTurb = mpCConnected->mCMappingTurb.mBody;
    mDummyMappingAxis = mpCConnected->mCMappingAxis.mBody;
    mbDummyMappingRepeat = mpCConnected->mCbMappingRepeat.mbBody;
    mbDummyMappingOptional = mpCConnected->mCbMappingOptional.mbBody;
    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	mDummyNormalPattern[i] = mpCConnected->mCNormalPattern[i].mBody;
	mDummyNormalDegree[i]  = mpCConnected->mCNormalDegree[i].mBody;
	mDummyNormalScale[i]   = mpCConnected->mCNormalScale[i].mBody;
	mDummyNormalTurb[i]    = mpCConnected->mCNormalTurb[i].mBody;
    }
    mstrDummyImagePath = mpCConnected->mCStrImagePath.msBody;
    mDummyImageMappingType = mpCConnected->mCImageMappingType.mBody;
    mDummyImageColoringType = mpCConnected->mCImageColoringType.mBody;

    mDummyImageOn = mpCConnected->mCbImageOn.mbBody;
    mDummyShadowOn = mpCConnected->mCbShadowOn.mbBody;
    mDummyRadiosityOn = mpCConnected->mCbRadiosityOn.mbBody;
    mDummyReflectionOn = mpCConnected->mCbReflectionOn.mbBody;
    
    mpCGUI->mpCMNumMain->Enable();
    mpCGUI->mpCMNumBump->Enable();
    mpCGUI->mpCMNumMirror->Enable();
    mpCGUI->mpCMNumOptions->Enable();
    mpCGUI->HideAndShow();
    
    wxString strTitle;
    strTitle << strItemName << wxT(":") << _("Texture Property Configure");
    mpCGUI->SetTitle(strTitle);
    UpdateColorList();
    UpdateItems();
}

void TexPropertyCtrl_cls::
UpdateLinkedProperty(){
    if (mpCConnected == NULL) return;
    mDummyAlpha = mpCConnected->GetLinkedParam(); return;
}

void TexPropertyCtrl_cls::
Apply(){
    if (mpCConnected == NULL) return;

    mpCConnected->mCColorType.mBody = mDummyColorType;
    if (mpCConnected->mpCFilterParamLinked != NULL)
	mpCConnected->mpCFilterParamLinked->mpCGUIPanel->mpCMNumPanel->UpdateValues();
    
    mpCConnected->mCSpecular.mBody = mDummySpecular;
    mpCConnected->mCDiffuse.mBody  = mDummyDiffuse;
    mpCConnected->mCDiffuseBack.mBody  = mDummyDiffuseBack;
    mpCConnected->mCSubsurfaceFactor.mBody = mDummySubsurfaceFactor;
    mpCConnected->mCAmbient.mBody  = mDummyAmbient;
    mpCConnected->mCRoughness.mBody = mDummyRoughness;
    mpCConnected->mCTransmitType.mBody = mDummyTransmitType;
    mpCConnected->mCIOR.mBody = mDummyIOR;
    mpCConnected->mCDispersion.mBody = mDummyDispersion;
    mpCConnected->mCFadeDistance.mBody = mDummyFadeDistance;
    mpCConnected->mCReflectionType.mBody = mDummyReflectionType;
    mpCConnected->mCReflectionMin.mBody = mDummyReflectionMin;
    mpCConnected->mCReflectionMax.mBody = mDummyReflectionMax;

    LOOP_RGB{
	mpCConnected->mCFadeColor.mpBody[rgb] = mDummyFadeColor[rgb];
	mpCConnected->mCColor.mpBody[rgb] = mDummyColor[rgb];
    }

    mpCConnected->mCMappingType.mBody  = mDummyMappingType;
    mpCConnected->mCMappingCoord.mBody = mDummyMappingCoord;
    mpCConnected->mCPatternType.mBody = mDummyPatternType;
    mpCConnected->mCPluginType.mBody = mDummyPluginType;
    LOOP_XYZ mpCConnected->mCMappingPos.mpBody[xyz] = mDummyMappingPos[xyz];
    mpCConnected->mCMappingScale.mBody = mDummyMappingScale;
    mpCConnected->mCMappingTurb.mBody = mDummyMappingTurb;
    mpCConnected->mCMappingAxis.mBody = mDummyMappingAxis;
    mpCConnected->mCbMappingRepeat.mbBody = mbDummyMappingRepeat;
    mpCConnected->mCbMappingOptional.mbBody = mbDummyMappingOptional;
    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	mpCConnected->mCNormalPattern[i].mBody = mDummyNormalPattern[i];
	mpCConnected->mCNormalDegree[i].mBody = mDummyNormalDegree[i];
	mpCConnected->mCNormalScale[i].mBody = mDummyNormalScale[i];
	mpCConnected->mCNormalTurb[i].mBody = mDummyNormalTurb[i];
    }
    mpCConnected->mCStrImagePath.msBody = mstrDummyImagePath;
    mpCConnected->mCImageMappingType.mBody = mDummyImageMappingType ;
    mpCConnected->mCImageColoringType.mBody = mDummyImageColoringType;

    mpCConnected->mCbImageOn.mbBody = mDummyImageOn;
    mpCConnected->mCbShadowOn.mbBody = mDummyShadowOn;
    mpCConnected->mCbRadiosityOn.mbBody = mDummyRadiosityOn;
    mpCConnected->mCbReflectionOn.mbBody = mDummyReflectionOn;
    
    if ( mpCConnected->mpCFilterParamLinked != NULL )
	mpCConnected->mpCFilterParamLinked->SetParam
	    (mpCConnected->mLinkedParamIndex, mDummyAlpha);
    if ( mpCConnected->mpAlphaLinked != NULL)
	*mpCConnected->mpAlphaLinked = mDummyAlpha;
    mpCConnected->mCRenderType.mBody = mDummyRenderType;
    
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshTL);
    mpCGUI->GetParent()->AddPendingEvent(com);
}

void TexPropertyCtrl_cls::
UpdateItems(bool bForce){
    if (mpCGUI != NULL){
	mpCGUI->mpCMNumMain->UpdateValuesAll(bForce);
	mpCGUI->mpCMNumInterior->UpdateValuesAll(bForce);
	mpCGUI->mpCMNumBump->UpdateValuesAll(bForce);
	mpCGUI->mpCMNumMirror->UpdateValuesAll(bForce);
	mpCGUI->mpCMNumOptions->UpdateValuesAll(bForce);
    }
}

void TexPropertyCtrl_cls::
UpdateColorList(){
    if (mpCGUI != NULL)
	mpCGUI->UpdateColorList();
}

void TexPropertyCtrl_cls::
SetToDefault(){
    if (mpCConnected != NULL)
	mpCConnected->SetToDefault();
}

void TexPropertyCtrl_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new TexPropertyGUI_cls
	(this, pC, mCScreenRect.GetPosition(), mCScreenRect.GetSize());
    SetPDialog(mpCGUI);
}

void TexPropertyCtrl_cls::
OpenColorDLG(int x, int y){
    wxColour CColor(mDummyColor[0]*255, mDummyColor[1]*255, mDummyColor[2]*255);
    GpCClrDialog->SetColour(CColor);
    GpCClrDialog->mpCDialog->Move(x, y);
    
    if (GpCClrDialog->ShowModal() == wxID_OK){
	mDummyColor[0] = (float)GpCClrDialog->GetR()/255;
	mDummyColor[1] = (float)GpCClrDialog->GetG()/255;
	mDummyColor[2] = (float)GpCClrDialog->GetB()/255;

	Apply();
	wxCmdEvt com0(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	mpCGUI->GetParent()->AddPendingEvent(com0);
	wxCmdEvt com1(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshColorBar);
	mpCGUI->GetParent()->AddPendingEvent(com1);
    }
}

void TexPropertyCtrl_cls::
OpenFadeColorDLG(){
    wxColour CColor(mDummyFadeColor[0]*255, mDummyFadeColor[1]*255, mDummyFadeColor[2]*255);
    wxColourDialogEX CDialog(mpCGUI, CColor, TRUE);
    
    if (CDialog.ShowModal() == wxID_OK){
	mDummyFadeColor[0] = (float)CDialog.GetR()/255;
	mDummyFadeColor[1] = (float)CDialog.GetG()/255;
	mDummyFadeColor[2] = (float)CDialog.GetB()/255;

	Apply();
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	mpCGUI->GetParent()->AddPendingEvent(com);
    }
}

bool TexPropertyCtrl_cls::
IsGUIShownOnScreen(){
    if (mpCGUI == NULL) return FALSE;
    return mpCGUI->IsShownOnScreen();
}

// ------------------------------

TexProperty_cls::
TexProperty_cls(){
    SetToDefault();
    mpCFilterParamLinked = NULL;
    mLinkedParamIndex = 0;
    mpAlphaLinked = NULL;
}

void TexProperty_cls::
LinkFilterParameter(FilterWithParameter_cls *pCFilterParamLinked,
		    int linkedParamIndex){
    mpCFilterParamLinked = pCFilterParamLinked;
    mLinkedParamIndex = linkedParamIndex;
    mpAlphaLinked = NULL;
}

void TexProperty_cls::
LinkAlphaParameter(double *pAlpha){
    mpAlphaLinked = pAlpha;
    mpCFilterParamLinked = NULL;
    mLinkedParamIndex = 0;
}

void TexProperty_cls::
SetToDefault(){
    mCConfList.mbSetWithInitValue = TRUE;
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
}

void TexProperty_cls::
ReDefineList(){
    mCConfList.Clear();
    mCConfList.Add(&mCColorType,  wxT("TexColorType="), wxT("0"));
    mCConfList.Add(&mCColor,  wxT("TexColorArray="), wxT("1 1 0.5"));
    mCConfList.Add(&mCMappingType , wxT("TexMappingType="),  wxT("0"));
    mCConfList.Add(&mCMappingCoord, wxT("TexMappingCoord="), wxT("0"));
    mCConfList.Add(&mCPatternType, wxT("TexPatternType="), wxT("0"));
    mCConfList.Add(&mCPluginType, wxT("TexPluginType="), wxT("0"));
    mCConfList.Add(&mCMappingPos, wxT("TexMappingPos="), wxT("0 0 0"));
    mCConfList.Add(&mCMappingScale, wxT("TexMappingScale="), wxT("1"));
    mCConfList.Add(&mCMappingTurb, wxT("TexMappingTurb="), wxT("0"));
    mCConfList.Add(&mCMappingAxis, wxT("TexMappingAxis="), wxT("2"));
    mCConfList.Add(&mCbMappingRepeat, wxT("TexMappingRepeat="), wxT("1"));
    mCConfList.Add(&mCbMappingOptional, wxT("TexMappingOptional="), wxT("0"));
    mCConfList.Add(&mCSpecular, wxT("TexColorSpecular="), wxT("0.1"));
    mCConfList.Add(&mCDiffuse,  wxT("TexColorDiffuse="), wxT("0.8"));
    mCConfList.Add(&mCDiffuseBack,  wxT("TexColorDiffuseBack="), wxT("0"));
    mCConfList.Add(&mCAmbient,  wxT("TexColorAmbient="), wxT("0.2"));
    mCConfList.Add(&mCRoughness,wxT("TexColorRoughness="), wxT("0.02"));
    mCConfList.Add(&mCTransmitType, wxT("TexTransmitType="), wxT("0"));
    mCConfList.Add(&mCRenderType, wxT("TexRenderType="), wxT("0"));
    mCConfList.Add(&mCAlpha, wxT("TexAlpha="), wxT("0"));
    mCConfList.Add(&mCSubsurfaceFactor, wxT("TexSubsurfaceFactor="), wxT("0"));
    mCConfList.Add(&mCIOR,wxT("TexIOR="), wxT("1"));
    mCConfList.Add(&mCDispersion,wxT("TexDispersion="), wxT("1"));
    mCConfList.Add(&mCFadeDistance,wxT("TexFadeDistance="), wxT("0"));
    mCConfList.Add(&mCFadeColor, wxT("TexFadeColorArray="), wxT("0.5 1 1"));
    mCConfList.Add(&mCSubsurfaceColor, wxT("TexSubsurfaceColorArray="), wxT("1 1 1"));
    mCConfList.Add(&mCReflectionType,wxT("TexColorReflectionType="), wxT("0"));
    mCConfList.Add(&mCReflectionMin,wxT("TexColorReflectionMin="), wxT("0"));
    mCConfList.Add(&mCReflectionMax,wxT("TexColorReflectionMax="), wxT("0"));
    mCConfList.Add(&mCNormalPattern[0], wxT("TexNormalPattern="), wxT("0"));
    mCConfList.Add(&mCNormalDegree[0] , wxT("TexNormalDegree="), wxT("0.2"));
    mCConfList.Add(&mCNormalScale[0]  , wxT("TexNormalScale="), wxT("0.1"));
    mCConfList.Add(&mCNormalTurb[0]   , wxT("TexNormalTurbulence="), wxT("0"));
    mCConfList.Add(&mCNormalPattern[1], wxT("TexNormalPattern1="), wxT("0"));
    mCConfList.Add(&mCNormalDegree[1] , wxT("TexNormalDegree1="), wxT("0.2"));
    mCConfList.Add(&mCNormalScale[1]  , wxT("TexNormalScale1="), wxT("0.1"));
    mCConfList.Add(&mCNormalTurb[1]   , wxT("TexNormalTurbulence1="), wxT("0"));
    mCConfList.Add(&mCStrImagePath    , wxT("TexImagePath="), wxT(""));
    mCConfList.Add(&mCImageMappingType, wxT("TexImageMapping="), wxT("0"));
    mCConfList.Add(&mCImageColoringType, wxT("TexImageColoring="), wxT("0"));
    mCConfList.Add(&mCbShadowOn       , wxT("TexShadowOn="), wxT("1"));
    mCConfList.Add(&mCbRadiosityOn    , wxT("TexRadiosityOn="), wxT("1"));
    mCConfList.Add(&mCbImageOn        , wxT("TexImageOn="), wxT("1"));
    mCConfList.Add(&mCbReflectionOn   , wxT("TexReflectionOn="), wxT("1"));
}

void TexProperty_cls::
SetBoxGeom(double min[3], double max[3], double rotParam[4]){
    LOOP_XYZ msBoxMinWorld[xyz] = min[xyz];
    LOOP_XYZ msBoxMaxWorld[xyz] = max[xyz];
    for (int i = 0; i < 4; i++) msBoxRotParam[i] = rotParam[i];
    
    LOOP_XYZ g2boxA[xyz] = 1.0 / (max[xyz] - min[xyz]);
    LOOP_XYZ g2boxB[xyz] = -min[xyz] / (max[xyz] - min[xyz]);
    LOOP_XYZ gcenter[xyz] = (max[xyz] + min[xyz]) / 2.0;
   
    msRotZ0 = msBoxRotParam[0];
    msRotX0 = msBoxRotParam[1];
    msRotY0 = msBoxRotParam[2];
    msRotZ1 = msBoxRotParam[3];
    msCosZ0 = cos(msRotZ0 * _RAD); msSinZ0 = sin(msRotZ0  * _RAD);
    msCosX0 = cos(msRotX0 * _RAD); msSinX0 = sin(msRotX0 * _RAD);
    msCosY0 = cos(msRotY0 * _RAD); msSinY0 = sin(msRotY0 * _RAD);
    msCosZ1 = cos(msRotZ1 * _RAD); msSinZ1 = sin(msRotZ1 * _RAD);
}

bool TexProperty_cls::
IsRotationRequired(){
    if (msBoxRotParam[0] == 0 && msBoxRotParam[1] == 0  &&
	msBoxRotParam[2] == 0 && msBoxRotParam[3] == 0)
	return FALSE;
    return TRUE;
}

void TexProperty_cls::
ConvertGlobal2LocalNoRot(double posGlobal[3], double posLocal[3]){
    LOOP_XYZ posLocal[xyz] = g2boxA[xyz] * posGlobal[xyz] + g2boxB[xyz];
}

void TexProperty_cls::
ConvertGlobal2Local(double posGlobal[3], double posLocal[3]){
    double tmpPos0[3], tmpPos1[3];

    double rotZ0 = msBoxRotParam[0] * _RAD;
    double rotX0 = msBoxRotParam[1] * _RAD;
    double rotY0 = msBoxRotParam[2] * _RAD;
    double rotZ1 = msBoxRotParam[3] * _RAD;

    tmpPos0[X] = posGlobal[X] - gcenter[X];
    tmpPos0[Y] = posGlobal[Y] - gcenter[Y];
    tmpPos1[Z] = posGlobal[Z] - gcenter[Z];

    tmpPos1[X] = tmpPos0[X] * msCosZ0 + tmpPos0[Y] * msSinZ0;
    tmpPos1[Y] = - tmpPos0[X] * msSinZ0 + tmpPos0[Y] * msCosZ0;
    //tmpPos1[Z] = tmpPos0[Z];

    tmpPos0[X] = tmpPos1[X];
    tmpPos0[Y] = tmpPos1[Y] * msCosX0 + tmpPos1[Z] * msSinX0;
    tmpPos0[Z] = - tmpPos1[Y] * msSinX0 + tmpPos1[Z] * msCosX0;


    tmpPos1[X] = - tmpPos0[Z] * msSinY0 + tmpPos0[X] * msCosY0;
    tmpPos1[Y] = tmpPos0[Y];
    tmpPos1[Z] = tmpPos0[Z] * msCosY0 + tmpPos0[X] * msSinY0;
    
    tmpPos0[X] = tmpPos1[X] * msCosZ1 + tmpPos1[Y] * msSinZ1;
    tmpPos0[Y] = - tmpPos1[X] * msSinZ1 + tmpPos1[Y] * msCosZ1;
    //tmpPos0[Z] = tmpPos1[Z];

    tmpPos0[X] += gcenter[X];
    tmpPos0[Y] += gcenter[Y];
    tmpPos0[Z] = tmpPos1[Z] + gcenter[Z];
    
    LOOP_XYZ posLocal[xyz] = g2boxA[xyz] * tmpPos0[xyz] + g2boxB[xyz];
}



double TexProperty_cls::
GetLinkedParam(){
    if (mpCFilterParamLinked == NULL){
	if (mpAlphaLinked == NULL) return 0;
	else return *mpAlphaLinked;
    }
    return mpCFilterParamLinked->mpCData->GetPParamActive()->mpInterpolatedValue[mLinkedParamIndex];
}

bool TexProperty_cls::
IsHalfMapping(){
    return (mCPatternType.mBody == _TEX_PATTERN_AOI);
}

bool TexProperty_cls::
IsUVPatternReady(){
  if (mCPatternType.mBody == _TEX_PATTERN_AGATE ||
      mCPatternType.mBody == _TEX_PATTERN_BOZO ||
      mCPatternType.mBody == _TEX_PATTERN_CHECKER ||
      mCPatternType.mBody == _TEX_PATTERN_CHECKER ||
      mCPatternType.mBody == _TEX_PATTERN_CYLINDRICAL ||
      mCPatternType.mBody == _TEX_PATTERN_DENTS ||
      mCPatternType.mBody == _TEX_PATTERN_GRADIENT ||
      mCPatternType.mBody == _TEX_PATTERN_GRANITE || 
      mCPatternType.mBody == _TEX_PATTERN_HEXAGON || 
      mCPatternType.mBody == _TEX_PATTERN_RADIAL ) return TRUE;
  return FALSE;
}

std::vector < float > TexProperty_cls::
GetV4AmbientColor(float value){
    std::vector < float > VRet;
    if (mCColorType.mBody == 0){
	LOOP_RGB VRet.push_back(mCColor.mpBody[rgb] * mCAmbient.mBody);
	VRet.push_back(GetLinkedParam());
    }
    else{
	VRet = GCColoring.mpCCustomColor->
	    GetV4CustomColor(value, TRUE, mCColorType.mBody - 1);
	LOOP_RGB VRet[rgb] *= mCAmbient.mBody;
    }
    
    return VRet;
}

std::vector < float > TexProperty_cls::
GetV4DiffuseColor(float value){
    std::vector < float > VRet;
    if (mCColorType.mBody == 0){
	LOOP_RGB VRet.push_back(mCColor.mpBody[rgb] * mCDiffuse.mBody);
	VRet.push_back(GetLinkedParam()); // alpha
    }
    else{
	VRet = GCColoring.mpCCustomColor->
	    GetV4CustomColor(value, TRUE, mCColorType.mBody - 1);
	LOOP_RGB VRet[rgb] *= mCDiffuse.mBody;
    }
    return VRet;
}

std::vector < float > TexProperty_cls::
GetV4SpecularColor(float value){
    std::vector < float > VRet;
    if (mCColorType.mBody == 0){
	LOOP_RGB VRet.push_back(mCColor.mpBody[rgb] * mCSpecular.mBody);
	VRet.push_back(GetLinkedParam());
    }
    else{
	VRet = GCColoring.mpCCustomColor->
	    GetV4CustomColor(value, TRUE, mCColorType.mBody - 1);
	LOOP_RGB VRet[rgb] *= mCSpecular.mBody;
    }
    return VRet;
}

wxString TexProperty_cls::
Serialize(){
    ReDefineList();
    if (mpAlphaLinked != NULL) mCAlpha.mBody = *mpAlphaLinked;
    return mCConfList.SerializeAll();
}

void TexProperty_cls::
ReadStrLine(std::wstring strLine){
    mCConfList.ReadStrLineAll(strLine);
    if (mpAlphaLinked != NULL) *mpAlphaLinked = mCAlpha.mBody;
}

void TexProperty_cls::
CalcScale(float scale[3], double normScale,
	  bool bGlobalCoord, int rotAxis){
    if (bGlobalCoord == TRUE)
	LOOP_XYZ scale[xyz] = normScale;
    else{
	switch (rotAxis){ // HeightField͍W]Ă
	case -1:
	case Y:
	    LOOP_XYZ scale[xyz] =
		normScale / (msBoxMaxWorld[xyz] - msBoxMinWorld[xyz]);
	    break;
	case X:
	    scale[X] = normScale / (msBoxMaxWorld[Y] - msBoxMinWorld[Y]);
	    scale[Y] = normScale / (msBoxMaxWorld[X] - msBoxMinWorld[X]);
	    scale[Z] = normScale / (msBoxMaxWorld[Z] - msBoxMinWorld[Z]); 
	    break;
	case Z:
	    scale[X] = normScale / (msBoxMaxWorld[X] - msBoxMinWorld[X]);
	    scale[Y] = normScale / (msBoxMaxWorld[Z] - msBoxMinWorld[Z]);
	    scale[Z] = normScale / (msBoxMaxWorld[Y] - msBoxMinWorld[Y]); 
	    break;
	}
    }
}

wxString TexProperty_cls::
GetPovStrFinish(bool bShading){
    wxString strFinish;

    if (bShading == TRUE){
	strFinish << L"\tambient rgb " << mCAmbient.mBody << PovOut3(L"*", 1, 1, 1) << ENDLWX
		  << L"\tspecular " << mCSpecular.mBody << ENDLWX
		  << L"\troughness " << mCRoughness.mBody << ENDLWX;
	if (mspCPovRayLink->mCPovRayVersion.mBody == _POVRAY_37)
	    // 0łp[^ݒ肵ق_O(Pen4)
	    strFinish << wxT("\tdiffuse ") << mCDiffuse.mBody << wxT(",") << mCDiffuseBack.mBody << ENDLWX; 
	else
	    strFinish << wxT("\tdiffuse ") << mCDiffuse.mBody << ENDLWX;
	if (mspCPovRayLink->mCPovRayVersion.mBody == _POVRAY_37 &&
	    mCSubsurfaceFactor.mBody != 0)
	    strFinish << PovOut3f(L"\tsubsurface { translucency ", mCSubsurfaceColor.mpBody) << wxT("*") << mCSubsurfaceFactor.mBody << wxT("}") << ENDLWX;
    }
    
    else
	strFinish << L"\tambient rgb < 1, 1, 1>" << ENDLWX
		  << L"\tspecular 0" << ENDLWX
		  << L"\troughness 0" << ENDLWX
		  << L"\tdiffuse 0" << ENDLWX;

    
    if (mCReflectionType.mBody != _TEX_NO_MIRROR && 
	(mCReflectionMin.mBody != 0 || mCReflectionMax.mBody != 0)){
	strFinish << L"\treflection {" << mCReflectionMin.mBody << wxT(",") 
		  << mCReflectionMax.mBody;
	if (mCReflectionType.mBody == _TEX_MIRROR_METAL)
	    strFinish << wxT("\tmetallic");
	strFinish << wxT(" }") << ENDLWX;
    }
    
    return strFinish;
}

std::vector < wxString > TexProperty_cls::
GetVPovStrColorMap(int df3index, float trans){
    std::vector < wxString > strV(4);
    strV[0] = GCColoring.GetPovStrColorMap
	(df3index, mCColorType.mBody - 1 + _CLR_CUSTOM, FALSE, 
	 0, NULL, FALSE, trans, mCTransmitType.mBody, 
	 IsHalfMapping(), FALSE);
    strV[1] = GCColoring.GetPovStrClr
	(df3index, mCColorType.mBody - 1 + _CLR_CUSTOM, 0,
	 trans, mCTransmitType.mBody);
    strV[2] = GCColoring.GetPovStrClr
	(df3index, mCColorType.mBody - 1 + _CLR_CUSTOM, 1,
	 trans, mCTransmitType.mBody);
    strV[3] = GCColoring.GetPovStrClr
	(df3index, mCColorType.mBody - 1 + _CLR_CUSTOM, 0.5,
	 trans, mCTransmitType.mBody);
    return strV;
}

std::vector < wxString > TexProperty_cls::
GetVPovStrColors(int df3index, float trans){
    std::vector < wxString > strV(4);

    switch (mCMappingType.mBody){
    case _TEX_MAPPING_VALUE:
	strV[0] = GCColoring.GetPovStrColorMap
	    (df3index, mCColorType.mBody - 1 + _CLR_CUSTOM, FALSE, 
	     0, NULL, FALSE, trans, mCTransmitType.mBody, 
	     FALSE, TRUE, TRUE);
	break;
    case _TEX_MAPPING_PATTERN: 
    case _TEX_MAPPING_IMAGE:
    case _TEX_MAPPING_PLUGINS: {
	strV = GetVPovStrColorMap(df3index, trans);
	break; }
    }
    
    return strV;
}

void TexProperty_cls::
SetDensityFile(wxString strDensityFile){
    mstrDensityFile = strDensityFile;
}

void TexProperty_cls::
SetTranslationForDensityFile(wxString strTranslate){
    mstrTranslationForDensityFile = strTranslate;
}

wxString TexProperty_cls::
GetPovStrNormalAll(float scaleFactorX, float scaleFactorY, float scaleFactorZ,
		   bool bShading, bool bGlobalCoord, int rotAxis){
    wxString strNormal;
    
    if (mCNormalPattern[0].mBody != 0 && mCNormalPattern[1].mBody != 0){
	float scale0[3], scale1[3];
	CalcScale(scale0, mCNormalScale[0].mBody, bGlobalCoord, rotAxis);
	CalcScale(scale1, mCNormalScale[1].mBody, bGlobalCoord, rotAxis);
	
	scale0[X] *= scaleFactorX;
	scale0[Y] *= scaleFactorY;
	scale0[Z] *= scaleFactorZ;
	scale1[X] *= scaleFactorX;
	scale1[Y] *= scaleFactorY;
	scale1[Z] *= scaleFactorZ;
	
	strNormal << L"\tnormal {" << ENDLWX
		  << L"\t\taverage" << ENDLWX
		  << L"\t\t\tnormal_map {" << ENDLWX
		  << L"\t\t\t [ 1.00 " << GetPovStrNormal(0, scale0) << L" ]" << ENDLWX
		  << L"\t\t\t [ 1.00 " << GetPovStrNormal(1, scale1) << L" ]" << ENDLWX
		  << L"\t\t}" << ENDLWX
		  << L"\t}" << ENDLWX;
    }
    else{
	for (int i = 0; i < _TEX_NORMAL_MAX; i++)
	    if (mCNormalPattern[i].mBody != 0){
		float scale[3];
		CalcScale(scale, mCNormalScale[0].mBody, bGlobalCoord, rotAxis);
		
		scale[X] *= scaleFactorX;
		scale[Y] *= scaleFactorY;
		scale[Z] *= scaleFactorZ;
		strNormal << L"\tnormal {" << GetPovStrNormal(i, scale) <<  L" }" << ENDLWX;
	    }
    }
    
    return strNormal;
}

wxString TexProperty_cls::
GetPovStrTex(float alpha, wxString strFinishId, wxString strColorMap,
	     float scaleFactorX, float scaleFactorY, float scaleFactorZ,
	     bool bShading, bool bGlobalCoord, int rotAxis,
	     wxString strClr1, wxString strClr2, wxString strClr3){
    wxString strTex;
    
    int transmitType = mCTransmitType.mBody;
    wxString strRGBA;
    switch (transmitType){
    case 0:
	strRGBA = wxT("color rgb"); break; // FIX_ME ꂾOK?
    case _TRANSMIT_TRANSMIT:
	strRGBA = wxT("color rgbt"); break;
    case _TRANSMIT_FILTER:
	strRGBA = wxT("color rgbf"); break;
    }

    wxString strColor;
    float colorModified[4];
    LOOP_RGB colorModified[rgb] = mCColor.mpBody[rgb];
    if (alpha > 1) {
	LOOP_RGB colorModified[rgb] *= alpha;
	alpha = 1; 
    }
    colorModified[3] = 1.0 - alpha;
    if (transmitType == 0 || alpha == 1)
	strColor << PovOut3f(strRGBA, colorModified);
    else
	strColor << PovOut4f(strRGBA, colorModified);
    
    if (strColorMap == wxT(""))
	strTex << L"\tpigment {" << strColor << L" }" << ENDLWX;
    else
	switch (mCMappingType.mBody){
	case _TEX_MAPPING_VALUE:
	    if (strColorMap != wxT("MONOCOLOR")){
		strTex << L"\tpigment {" << ENDLWX;
		if (mstrDensityFile != wxT("")) {
		    strTex << wxT("\tdensity_file df3 \"") << mstrDensityFile << wxT("\"") << ENDLWX
			   << wxT("\tinterpolate ") << PovVolume_cls::msCInterpolate.mBody << ENDLWX 
			   << mstrTranslationForDensityFile << ENDLWX;
		}
		strTex << strColorMap << L"}" << ENDLWX;
	    }
	    else
		strTex << L"\tpigment {" << strClr1 << wxT("}") << ENDLWX;
	    break;
	case _TEX_MAPPING_PATTERN:
	    strTex << L"\tpigment { " << GetPovStrPattern(strColorMap, strClr1, strClr2, strClr3);
	    if (mCMappingTurb.mBody != 0)
		strTex << wxT("\t\twarp { turbulence ") << mCMappingTurb.mBody << wxT("}") << ENDLWX;
	    strTex << wxT("\t}") << ENDLWX;
	    strTex << PovOut3(wxT("\tscale"), 
			      mCMappingScale.mBody, mCMappingScale.mBody, mCMappingScale.mBody) << ENDLWX;
	    strTex << PovOut3f(wxT("\ttranslate"),
			       mCMappingPos.mpBody) << ENDLWX;
	    if (mCPatternType.mBody == _TEX_PATTERN_BRICK ||
		mCPatternType.mBody == _TEX_PATTERN_HEXAGON)
		switch (mCMappingAxis.mBody){
		case X:
		    strTex << PovOut3(wxT("\trotate"), 0, 0, 90) << ENDLWX;
		    break;
		case Z:
		    strTex << PovOut3(wxT("\trotate"), 90, 0, 0) << ENDLWX;
		    break;
		}
	    break;
	case _TEX_MAPPING_IMAGE: {
	    SplitPathWX_cls CPath(mCStrImagePath.msBody);
	    wxString strImageType = CPath.GetImageTypeStr();
	    strTex << wxT("\tpigment {") << ENDLWX;
	    if (mCImageColoringType.mBody == 0)
		strTex << wxT("\t\timage_map { jpeg \"") << mCStrImagePath.msBody << wxT("\"") << ENDLWX
		       << wxT("\t\tgamma 1") << ENDLWX
		       << wxT("\t\tmap_type ") << mCImageMappingType.mBody << wxT("}") << ENDLWX
		       << wxT("\t}") << ENDLWX;
	    else
		strTex << wxT("\t\timage_pattern { jpeg \"") << mCStrImagePath.msBody << wxT("\"") << ENDLWX
		       << wxT("\t\tgamma 1") << ENDLWX
		       << wxT("\t\tmap_type " ) << mCImageMappingType.mBody << wxT("}") << ENDLWX
		       << strColorMap << ENDLWX
		       << wxT("\t}") << ENDLWX;
	    switch(mCImageMappingType.mBody){
	    case _IMAGEMAP_PLANER:
		strTex << wxT("\ttranslate <-0.5, -0.5, 0>");
		break;
	    case _IMAGEMAP_SPHERICAL:
		strTex << wxT("\trotate <90, 0, 0>") << ENDLWX;
		break;
	    case _IMAGEMAP_CYLINDER:
		// FIX_ME!! ]ƈړzȊOK
		switch(mCMappingAxis.mBody){
		case X: break;
		case Y: 
		    strTex << wxT("\trotate <0, 0, 90>") << ENDLWX;
		    break;
		case Z:
		    strTex << wxT("\trotate <90, 0, 0>") << ENDLWX;
		    break;
		}
		strTex << wxT("\ttranslate <0, 0, -0.5>");
		break; }
	    break; }
	case _TEX_MAPPING_PLUGINS:
	    break;
	}
    
    strTex << GetPovStrNormalAll
	(scaleFactorX, scaleFactorY, scaleFactorZ,
	 bShading, bGlobalCoord, rotAxis);

    strTex << wxT("\tfinish { ") << strFinishId << wxT("}") << ENDLWX;
    
    return strTex;
}

wxString TexProperty_cls::
GetPovStrObjectOption(){
    wxString strOption;
    if (mCbImageOn.mbBody == FALSE) strOption << wxT("no_image ");
    if (mCbShadowOn.mbBody == FALSE) strOption << wxT("no_shadow ");
    if (mCbRadiosityOn.mbBody == FALSE) strOption << wxT("no_radiosity ");
    if (mCbReflectionOn.mbBody == FALSE) strOption << wxT("no_reflection ");
    return strOption;
}

wxString TexProperty_cls::
GetPovStrPluginsParameters(PositionAndDirection_cls CCamPos, wxString strClr1, wxString strClr2, wxString strClr3){
    wxString strPlugins;

    switch(mCPluginType.mBody){
    case _TEX_PLUGINGS_GRASSTEX: {
	float camPosForGrass[3];
	camPosForGrass[X] = CCamPos.mPos[X];
	camPosForGrass[Y] = CCamPos.mPos[Z];
	camPosForGrass[Z] = -CCamPos.mPos[Y];
	strPlugins << wxT("#include \"grasstex/grasstex.inc\"") << ENDLWX;
	strPlugins << wxT("#declare camera_location=") << PovOut3f(wxT(""), camPosForGrass) << wxT(";") << ENDLWX;
	strPlugins << wxT("#declare gt_height =") << 0.075 * mCMappingScale.mBody << wxT(";") << ENDLWX
		   << wxT("#declare gt_scale =") << 0.03 * mCMappingScale.mBody << wxT(";") << ENDLWX
		   << wxT("#declare gt_color =") << strClr3 << wxT(";") << ENDLWX
		   << wxT("#declare gt_flowercolor =") << strClr2 << wxT(";") << ENDLWX
		   << wxT("#declare gt_groundcolors = color_map {") <<ENDLWX
		   << wxT("\t[1.0 ") << strClr1 << wxT("]") << ENDLWX 
		   << wxT("}") << ENDLWX
		   << wxT("#declare gt_offset = 0.05;") << ENDLWX
		   << wxT("#declare gt_samples = 25;") << ENDLWX
		   << wxT("#declare gt_turb = ") << mCMappingTurb.mBody << wxT(";") << ENDLWX
		   << wxT("#declare gt_heightmap=function {0.3 + 0.7 * f_bozo(x*5,y*5,z*5)}") << ENDLWX;
	break; }
    }
    return strPlugins;
}

wxString TexProperty_cls::
GetPovStrPluginsInObject(wxString strFinishId){
    wxString strPlugins;
    
    switch(mCPluginType.mBody){
    case _TEX_PLUGINGS_GRASSTEX:
	strPlugins << wxT("\ttexture { gt_ground() rotate x*90 normal {bumps 0.5 scale 0.01}") << ENDLWX
		   << wxT("\t\tfinish {") << strFinishId <<  wxT("} }") << ENDLWX
		   << wxT("\ttexture { gt_grass () rotate x*90 normal {bumps 0.5 scale 2 warp {turbulence 2} scale 1}") << ENDLWX
		   << wxT("\t\tfinish {") << strFinishId <<  wxT("} }") << ENDLWX;
	if (mCbMappingOptional.mbBody == TRUE){
	    strPlugins << wxT("\ttexture { gt_flowers () rotate x*90 scale 0.3") << ENDLWX;
	    strPlugins << wxT("\t\tfinish {") << strFinishId <<  wxT("} }") << ENDLWX;
	}
	break;
    }
    return strPlugins;
}

wxString TexProperty_cls::
GetPovStrInterior(){
    wxString strInterior;
    if (mCIOR.mBody != 1){
	strInterior << wxT("ior ") << mCIOR.mBody;
	if (mCDispersion.mBody != 1)
	    strInterior << wxT(" dispersion ") << mCDispersion.mBody;
    }
    if (mCFadeDistance.mBody != 0){
	strInterior << wxT(" fade_power ") << 1;
	strInterior << wxT(" fade_distance ") << mCFadeDistance.mBody << ENDLWX;
	strInterior << PovOut3f(wxT("\t\t\tfade_color"), mCFadeColor.mpBody);
    }
    return strInterior;
}

wxString TexProperty_cls::
GetPovStrNormal(int index, float scale[3]){

    // ͓{ʖځFFIX_ME !! ĂƂ
    wxString normalPatternList[12] = 
	{wxT("no normal pattern"), wxT("agate"), wxT("bozo"), wxT("bumps"), wxT("brick"), wxT("cells"), wxT("checker"), wxT("crackle"), /*wxT("cylindrical"),*/ wxT("dents"), wxT("facets"), /*wxT("mandel"), wxT("julia"), wxT("hexagon"),*/ wxT("marble"), 
	 wxT("function { f_ridged_mf(x, y, z, 0.07, 2.2,  7, 0.6, 0.9, 1) }")
};
    
    wxString strNormal;

    strNormal << normalPatternList[mCNormalPattern[index].mBody];
    strNormal << wxT(" ") << mCNormalDegree[index].mBody;
    
    strNormal << PovOut3f(wxT(" scale"), scale);
    strNormal << wxT(" turbulence ") << mCNormalTurb[index].mBody;
    
    return strNormal;
}

wxString TexProperty_cls::
GetPovStrPattern(wxString strClrMap, wxString strClr1, wxString strClr2, wxString strClr3){
    wxString strPattern;
    switch (mCPatternType.mBody){
    case _TEX_PATTERN_AGATE:
	strPattern << wxT("agate ") << strClrMap;
	break;
    case _TEX_PATTERN_AOI:
	strPattern << wxT("aoi ") << strClrMap;
	break;
    case _TEX_PATTERN_BOZO:
	strPattern << wxT("bozo ") << strClrMap;
	break;
    case _TEX_PATTERN_BRICK:
	strPattern << wxT("brick ") << strClr1 << wxT(",") << strClr2 << ENDLWX;
	break;
    case _TEX_PATTERN_CELLS:
	strPattern << wxT("cells ") << strClrMap << ENDLWX;
	break;
    case _TEX_PATTERN_CHECKER:
	strPattern << wxT("checker ") << strClr1 << wxT(",") << strClr2 << ENDLWX;
	break;
    case _TEX_PATTERN_CRACKLE:
	strPattern << wxT("crackle ") << strClrMap;
	break;
    case _TEX_PATTERN_CYLINDRICAL: 
	switch (mCMappingAxis.mBody){
	case X:
	    strPattern << wxT("function { pow(y*y + z*z, 0.5) }") << ENDLWX;
	    break;
	case Y:
	    strPattern << wxT("function { pow(z*z + x*x, 0.5) }") << ENDLWX;
	    break;
	case Z:
	    strPattern << wxT("function { pow(x*x + y*y, 0.5) }") << ENDLWX;
	    break;
	}
	strPattern << strClrMap << ENDLWX;
	break; 
    case _TEX_PATTERN_DENTS:
	strPattern << wxT("dents ") << strClrMap;
	break;
    case _TEX_PATTERN_GRADIENT: //FIX_ME!!
	strPattern << wxT("gradient ") << mAstrAxis[mCMappingAxis.mBody] << wxT(" ") << strClrMap;
	break;
    case _TEX_PATTERN_GRANITE: //FIX_ME!!
	strPattern << wxT("granite ") << strClrMap;
	break;
    case _TEX_PATTERN_HEXAGON:
	strPattern << wxT("hexagon ") << strClr1 << wxT(",") << strClr2 << wxT(",") << strClr3 << ENDLWX;
	break;
    case _TEX_PATTERN_RADIAL: {
	strPattern << wxT("function { pow(x*x + y*y + z*z, 0.5) }") << ENDLWX
		   << strClrMap << ENDLWX;
	break; }
    }

    return strPattern;
}

PovRay_cls *TexProperty_cls::mspCPovRayLink;
double TexProperty_cls::msBoxMinWorld[3];
double TexProperty_cls::msBoxMaxWorld[3];
double TexProperty_cls::msBoxRotParam[4];
double TexProperty_cls::g2boxA[3];
double TexProperty_cls::g2boxB[3];
double TexProperty_cls::gcenter[3];
double TexProperty_cls::msRotZ0, TexProperty_cls::msRotX0;
double TexProperty_cls::msRotY0, TexProperty_cls::msRotZ1;
double TexProperty_cls::msCosZ1, TexProperty_cls::msSinZ1;
double TexProperty_cls::msCosX0, TexProperty_cls::msSinX0;
double TexProperty_cls::msCosY0, TexProperty_cls::msSinY0;
double TexProperty_cls::msCosZ0, TexProperty_cls::msSinZ0;

