// -*- c -*-
// "clsMultiParameters.h"

#include "clsMultiParameters.h"

#include "../DF3/clsDF3SeqGUI.h"
#include "../MYAPI/MsgBox.h"
#include "../MYGL/clsGLGuides.h"
#include "../UNDO/clsUndoManager.h"

#include "../WIDGETCTRL/clsAtoiComboCtrl.h"
#include "../WIDGETCTRL/clsIconSelector.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"

#include <GL/glut.h>
#include "../GLOBJ/GLOBJECTS.h"

#include <sstream>

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define _MANIP_TRANSLATE 0

#ifdef WIN32
#define ENDLWX wxT("\r\n")
#else
#define ENDLWX wxT("\n")
#endif

extern UndoManager_cls GCUndoList;

MultiParametersGUI_cls::
MultiParametersGUI_cls(MultiParameters_cls *pCParent, 
		       wxWindow *pC, wxString strTitle,
		       wxPoint pos, wxSize size, wxBitmap *pBitmap, long style) : 
    wxDialogEXKey(pC, wxID_ANY, strTitle,
		  pos, size, pBitmap, style)    
{
    mpCParent = pCParent;
}

void MultiParametersGUI_cls::
SetMatrix(Parameters_cls *pParam, int keyIndex,
	  double dpos[3], int flag){
    glTranslated(pParam->mVCKeys[keyIndex].mpValue[0],
		 pParam->mVCKeys[keyIndex].mpValue[1],
		 pParam->mVCKeys[keyIndex].mpValue[2]);
    glTranslated(dpos[0], dpos[1], dpos[2]);
    glScaled(pParam->mVCKeys[keyIndex].mpValue[3],
	     pParam->mVCKeys[keyIndex].mpValue[3],
	     pParam->mVCKeys[keyIndex].mpValue[3]);
}

void MultiParametersGUI_cls::
DrawSelectedPath(GLGuides_cls *pCGLGuide, double dpos[3], unsigned int listGL, double scale, unsigned int pathIndex, PositionAndDirection_cls CParentObjPos){
    if (DrawSelectedKey(pCGLGuide, &mpCParent->mVCParam[pathIndex], NULL, NULL, listGL, 
			dpos, scale, 0, _MANIP_TRANSLATE, NULL) == FALSE){
	PositionAndDirection_cls CPosDir;
	LOOP_XYZ CPosDir.mPos[xyz] = mpCParent->mVCParam[pathIndex].mpInterpolatedValue[xyz];
	double logScale = log10(1+mpCParent->mVCParam[pathIndex].mpInterpolatedValue[3]) * 3 + 1; //clsParameters.cpp::SetMatirixCurrentScale ɂ
	
	scale *= logScale;
	DrawCurrent(pCGLGuide, CPosDir, mVCDisplayList,
		    dpos, scale, NULL, CParentObjPos);
    }
}

void MultiParametersGUI_cls::
SetMNumMain(wxBoxSizer *pBoxSizer, 
	    unsigned int idAdd, unsigned int idRemove){
    mpCMNumMain = new MultiNumericalInputsPanel_cls(&mpCMNumMain, this, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxBORDER_SIMPLE);
    
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumMain;
    
    pCurr->SetDefaultButtonType(_MN_NO_BUTTONS);

#ifdef WIN32
    pCurr->mTextWidth  = 80;
    pCurr->mComboWidth = 60;
#else
    pCurr->mTextWidth  = 100;
    pCurr->mComboWidth = 100;
#endif
    
    int boxWidth = 3;
    pCurr->mButtonWidth = 24;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);

    mIndexIcon = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_ICON, &mpCParent->mActiveItem);
    pCurr->mVCPointers[mIndexIcon].mpVIcon = &mVIcon;
    pCurr->mVCPointers[mIndexIcon].SetRange(0,mpCParent->mCItemNum.mBody,1);

    mIndexAdd    = pCurr->CombinePointerButton(idAdd   , wxT("+"));
    mIndexDelete = pCurr->CombinePointerButton(idRemove, wxT("-"));
    
#ifdef WIN32
    pCurr->mMinimumWidth = 180;
#else
    pCurr->mMinimumWidth = 220;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);
    //delete [] pstrList;

    Connect(mpCMNumMain->mVCPointers[mIndexIcon].mID, 
	    wxEVT_COMMAND_COMBOBOX_SELECTED,
	    wxCommandEventHandler(MultiParametersGUI_cls::OnSelector));

    pCurr->SetFuncHook(&MultiParameters_cls::FuncHookItemNumberChanged,
		       _MN_HOOK_ASSIGN, wxID_ANY,
		       (MultiParameters_cls*)mpCParent);
}

void MultiParametersGUI_cls::
OnSelector(wxCommandEvent &event){    
    UpdateActiveOnTL();
}

void MultiParametersGUI_cls::
UpdateComboBox(){
    wxButtonEX *pBtnRemove = (wxButtonEX*)mpCMNumMain->mVCPointers[mIndexDelete].mpPointerCtrl;
    
    if (mpCParent->mCItemNum.mBody > 0)
	pBtnRemove->Enable(TRUE);
    else 
	pBtnRemove->Disable();

    ((IconSelector_cls*)mpCMNumMain->mVCPointers[mIndexIcon].mpPointerCtrl)->
	mMax = mpCParent->mCItemNum.mBody;
    mpCMNumMain->mVCPointers[mIndexIcon].UpdateValue();

    ((IconSelector_cls*)mpCMNumMain->mVCPointers[mIndexIcon].mpPointerCtrl)->
	Refresh(FALSE);
    
    UpdateComboBoxPropagate();
}

void MultiParametersGUI_cls::
OnButtonAdd(wxCommandEvent &event){
    mpCParent->AddItem();
    //mpCParent->mActiveItem++;
    mpCParent->FuncHookItemNumberChanged(NULL, mpCParent);
    UpdateActiveOnTL();
    UpdateComboBox();
    //mpCMNumMain->UpdateValues();
}

void MultiParametersGUI_cls::
OnButtonRemove(wxCommandEvent& WXUNUSED(event)){
    RemoveItem();
}

bool MultiParametersGUI_cls::
RemoveItem(){
    if (mpCParent->mCItemNum.mBody == 0) return FALSE;
    char buf[64];
    unsigned int current = mpCParent->mActiveItem;
    sprintf(buf, "Remove Item %d ?\n", current);
    if (MsgBoxYesno(this, "Remove Item", buf) == FALSE)
	return FALSE;

    mpCParent->RemoveItem();
    UpdateActiveOnTL();
    UpdateComboBox();
    return TRUE;
}

void MultiParametersGUI_cls::
OnButtonClose(wxCommandEvent &event){
    mpCMNumParam->UpdateValues();
    CloseEXKey();
}

//----------------

MultiParameters_cls::
MultiParameters_cls(std::wstring strItemNum, std::wstring strTag,
		    unsigned int dimention, bool bConnectToTLPanel){
    mCItemNum.SetItem(strItemNum, L"0");
    mActiveItem = 0;

    mstrTag = strTag;
    mstrLabelForPanel = strTag;
    mDimension = dimention;
    mbConnectToTLPanel = bConnectToTLPanel;
    mpInit = new double [mDimension];
    for (unsigned int i = 0; i < mDimension; i++) mpInit[i] = 0;

    // |C^ĊmۂňړȂ悤ɂB(UNDO݂ł܂)
    // *todo _MAX_ITEM_NUMBERɂ鏈
    mVCParam.reserve(_MAX_ITEM_NUMBER); 
}

MultiParameters_cls::
~MultiParameters_cls(){
    delete [] mpInit;
}

void MultiParameters_cls::
InitializeStaticMembers(DF3SeqGUI_cls* pCDF3SeqGUI){
    mspCDF3SeqGUI = pCDF3SeqGUI;
}

void MultiParameters_cls::
SetInitParam(double *pInit){
    for (unsigned int i = 0; i < mDimension; i++) mpInit[i] = pInit[i];
}

unsigned int MultiParameters_cls::
GetActiveItemKeyNum(){
    return (unsigned int)mVCParam[mActiveItem].mVCKeys.size();
}

Parameters_cls* MultiParameters_cls::
GetPActiveItem(){
    if (mCItemNum.mBody == 0) return NULL;
    return &mVCParam[mActiveItem];
}

/*
wxString* MultiParameters_cls::
GetPItemListStringVolatile(){
    if (mCItemNum.mBody != 0){
	wxString *pstrList = new wxString [mCItemNum.mBody];
	for (unsigned int i = 0; i < mCItemNum.mBody; i++)
	    pstrList[i] << wxConvFile.cWC2WX(mstrTag.c_str()) << i;
	return pstrList;
    }
    wxString *pstrList = new wxString [1];
    *pstrList = wxT("no Item");
    return pstrList;
}
*/

void MultiParameters_cls::
AddItem(){
    Parameters_cls CParam(mDimension, mstrTag);
    CParam.ExecCommandInitSet(mpInit);

    GCUndoList.SetNewCommand
	(new ComMultiParamAdd_cls(this, &CParam, mCItemNum.mBody),
	 new ComMultiParamRemove_cls(this, mCItemNum.mBody));
    
    GCUndoList.ExecCurrent();

    mActiveItem = mCItemNum.mBody;
}

void MultiParameters_cls::
RemoveItem(){
    //if (mCItemNum.mBody == 0) return;

    GCUndoList.SetNewCommand
	(new ComMultiParamRemove_cls(this, mActiveItem),
	 new ComMultiParamAdd_cls(this, &mVCParam[mActiveItem], mActiveItem));
    GCUndoList.ExecCurrent();
}

void MultiParameters_cls::
AdjustAutoControl(){
    for (unsigned int index = 0; index < mCItemNum.mBody; index++)
	mVCParam[index].AdjustAutoControl();
}

void MultiParameters_cls::
AddItemProc(Parameters_cls *pCNewParam, unsigned int index){
    mVCParam.push_back(Parameters_cls(mDimension, mstrTag));
    mCItemNum.mBody++;
			  
    for (unsigned int i = mCItemNum.mBody - 1; i > index; i--){
	mVCParam[i].CopyFromP(&mVCParam[i-1]);
	if (mbConnectToTLPanel == TRUE){
	    wxString strLabelFrom, strLabelTo; 
	    strLabelFrom << wxT("[") <<  mstrLabelForPanel << wxT("]") << i-1;
	    strLabelTo << wxT("[") <<  mstrLabelForPanel << wxT("]") << i;
	    mspCDF3SeqGUI->mpCPanelList->Reconnect(strLabelFrom,
						   &mVCParam[i]);
	    mspCDF3SeqGUI->mpCPanelList->Rename(strLabelFrom, strLabelTo);
	}
    }
    
    if (pCNewParam != NULL)
	mVCParam[index].CopyFromP(pCNewParam);

    if (mbConnectToTLPanel == TRUE){
	mspCDF3SeqGUI->mpCPanelList->DeactivateAllPanel(TRUE);
	wxString strLabel;
	strLabel << wxT("[") << mstrLabelForPanel << wxT("]") << index;
	mspCDF3SeqGUI->mpCPanelList->Add(&mVCParam[index], strLabel);
	mspCDF3SeqGUI->ConnectMouseEvent();
	mspCDF3SeqGUI->UpdatePanelLayout();
    }

    AddItemPropagate();
}

void MultiParameters_cls::
RemoveItemProc(unsigned int index){

    if (mbConnectToTLPanel == TRUE){
	wxString strLabel;
	strLabel << wxT("[") <<  mstrLabelForPanel << wxT("]") << index;
	mspCDF3SeqGUI->mpCPanelList->Remove(strLabel);
	for (unsigned int i = index + 1; i < mCItemNum.mBody; i++){
	    wxString strLabelFrom, strLabelTo; 
	    strLabelFrom << wxT("[") <<  mstrLabelForPanel << wxT("]") << i;
	    strLabelTo << wxT("[") <<  mstrLabelForPanel << wxT("]") << i-1;
	    mspCDF3SeqGUI->mpCPanelList->Reconnect(strLabelFrom,
						   &mVCParam[i-1]);
	    mspCDF3SeqGUI->mpCPanelList->Rename(strLabelFrom, strLabelTo);
	}
	mspCDF3SeqGUI->UpdatePanelLayout();
    }

    for (unsigned int i = index + 1; i < mCItemNum.mBody; i++)
	mVCParam[i - 1].CopyFromP(&mVCParam[i]);
        
    mVCParam.resize(mCItemNum.mBody - 1);
    mCItemNum.mBody--;
    FuncHookItemNumberChanged(NULL, this);
    
    RemoveItemPropagate();
}

// -----

void MultiParametersGUI_cls::
DrawPath(unsigned int i, bool bSelectedPanel, 
	 unsigned int listGL, double scale){
    if (mpCParent->mVCParam[i].mCbVisible.mbBody == FALSE) return;
    
    bool bSelected = bSelectedPanel /*&& (i == mActiveItem)*/;
    
    glColor4fv(mpCParent->mVCParam[i].GetPColor(bSelected, TRUE));

    glPushMatrix();
    mpCParent->mVCParam[i].SetMatrixCurrentPos();
    mpCParent->mVCParam[i].SetMatrixCurrentScale(3);
    glScaled(scale, scale, scale);
    GLOBJ::Object::Draw(listGL);
    glPopMatrix();

    mpCParent->mVCParam[i].DrawGLPath3dv
	(mpCParent->mStepNum, 
	 mpCParent->mDobWAMin,
	 mpCParent->mDobWAMax,
	 mpCParent->mScale,
	 mpCParent->mInterval, bSelected);

    double pZero[3] = {0, 0, 0};
    if (bSelected == TRUE)
	glColor4f(1.0f, 1.0f, 0.2f, 1.0f);
    else
	glColor4f(0.7f, 0.7f, 0.1f, 1.0f);
    for (unsigned int key = 0; key < mpCParent->mVCParam[i].GetKeyNumber(); key++){
	if (mpCParent->mVCParam[i].mVCKeys[key].mbSelected == FALSE) continue;
	glPushMatrix();
	SetMatrix(&mpCParent->mVCParam[i], key, pZero, 0);
	glScaled(scale, scale, scale);
	GLOBJ::Object::Draw(listGL);
	glPopMatrix();
    }
}

// -----

void MultiParameters_cls::
HideTLPanel(int index){
    wxString strLabel;
    strLabel << wxT("[") << mstrLabelForPanel << wxT("]") << index;
    mspCDF3SeqGUI->mpCPanelList->DeactivateAllPanel(TRUE);
    mspCDF3SeqGUI->mpCPanelList->Activate(strLabel, TRUE);
    mspCDF3SeqGUI->mpCPanelList->HidePanel();
}

void MultiParameters_cls::
ReadStrLinePropagate(){
    MsgBox(NULL, "!!", "!!");
}

void MultiParameters_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    std::wstring strTag = std::wstring(L"--MultiParameter:") + mstrTag;
    std::wstring::size_type loc = strLine.find(strTag);
    if (loc != std::wstring::npos){
	std::wstring strLine;
	std::getline(istream, strLine);
	// UNAĂ炶ȂƂȂ̂mCItemNum𒼐ژMĂ̂
	// ʓ|Ƃ
	int itemNumberOrg = mCItemNum.mBody;
	if (mCItemNum.ReadStrLine(strLine)){
	    int itemNumberNew = mCItemNum.mBody;
	    mCItemNum.mBody = itemNumberOrg;
	    for (int i = itemNumberOrg - 1; i >= 0; i--){
		RemoveItemProc(i);
	    }
	    
	    for (unsigned int i = 0; i < itemNumberNew; i++){
		// RR͌ߑł
		AddItemProc(NULL, i);
		std::string wstrLine;
		std::getline(istream, strLine);
		mVCParam[i].ReadStrLine(istream, strLine);

		if (mVCParam[i].mCbShow.mbBody == FALSE)
		    HideTLPanel(i);
	    }
	}
	mspCDF3SeqGUI->mpCPanelList->DeactivateAllPanel(TRUE);
    }

    ReadStrLinePropagate();
}

wxString MultiParameters_cls::
Serialize(){
    wxString sOut;
    sOut << wxT("--MultiParameter:") << mstrTag << ENDLWX;
    sOut << mCItemNum.Serialize();
    for (unsigned int i = 0; i < mCItemNum.mBody; i++){
	sOut << mVCParam[i].Serialize();
    }
    return sOut;
}

void MultiParameters_cls::
FuncHookItemNumberChanged(void *WXUNUSED(pArg), void *pEventOvwr){
    MultiParameters_cls* pThis = (MultiParameters_cls*)pEventOvwr;
    if (pThis->mCItemNum.mBody == 0){
	pThis->mActiveItem = 0;
	return;
    }

    if (pThis->mActiveItem >= pThis->mCItemNum.mBody) 
	pThis->mActiveItem = pThis->mCItemNum.mBody - 1;
}

//ParametersPanelList_cls* MultiParameters_cls::mspCPanelList;
DF3SeqGUI_cls* MultiParameters_cls::mspCDF3SeqGUI;

// -----------------------------------------

ComMultiParamBase_cls::
ComMultiParamBase_cls(MultiParameters_cls *pCTgt){
    mpCTgt = pCTgt;
}

ComMultiParamAdd_cls::
ComMultiParamAdd_cls(MultiParameters_cls *pCTgt, Parameters_cls *pCNewParam, 
		     unsigned int index) :
    ComMultiParamBase_cls(pCTgt)
{
    mIndex = index;
    mpCParamMemento = new Parameters_cls(pCNewParam->mCDimension.mBody);
    mpCParamMemento->CopyFromP(pCNewParam);
}

ComMultiParamAdd_cls::
~ComMultiParamAdd_cls(){
    delete mpCParamMemento;
}

bool ComMultiParamAdd_cls::
Execute(bool bFlag){
    mpCTgt->AddItemProc(mpCParamMemento, mIndex);
    return _COMMAND_NODEPENDENT;
}

ComMultiParamRemove_cls::
ComMultiParamRemove_cls(MultiParameters_cls *pCTgt, unsigned int index) :
    ComMultiParamBase_cls(pCTgt)
{
    mIndex = index;
}

bool ComMultiParamRemove_cls::
Execute(bool bFlag){
    mpCTgt->RemoveItemProc(mIndex);
    return _COMMAND_NODEPENDENT;
}
