// -*- c -*-
// "clsParameters.h"
// (setq c-basic-offset 4)

#include "clsParameters.h"
#include "clsParamCommands.h"

#include <math.h>
#include <algorithm>
#include <fstream>
#include <sstream>

#include "../GLCANVAS/clsMainGLCanvas.h"
#include "../MYAPI/MsgBox.h"
#include "../MYGL/clsGLGuides.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/EX/clsStringEX.h"

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include <GL/glut.h>
#include "../GLOBJ/GLOBJECTS.h"

#include <wx/clipbrd.h>

extern UndoManager_cls GCUndoList;
extern std::vector < Parameters_cls* > GpVPInstancePrim;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGBA for (int rgba = 0; rgba < 4; rgba++)
#define LOOP_YPRY for (int ypry = 0; ypry < 4; ypry++)

#ifdef WIN32
#define ENDLWX wxT("\r\n")
#else
#define ENDLWX wxT("\n")
#endif

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

ParamGL_cls::
ParamGL_cls(){
    SetColor(_PARAMGL_SELECTED_IN  , 1  , 1  , 1,   1);
    SetColor(_PARAMGL_SELECTED_OUT , 0.8, 0.8, 0.8, 1);
    SetColor(_PARAMGL_NOTSELECT_IN , 0.5, 0.5, 0.5, 1);
    SetColor(_PARAMGL_NOTSELECT_OUT, 0.3, 0.3, 0.3, 1);
}

float* ParamGL_cls::
GetPColor(bool bSelected, bool bInWA){
    if ((bSelected == TRUE) && (bInWA == TRUE))
	return mColorSelectedIn;
    if ((bSelected == TRUE) && (bInWA == FALSE))
	return mColorSelectedOut;
    if ((bSelected == FALSE) && (bInWA == TRUE))
	return mColorNotSelectIn;
    if ((bSelected == FALSE) && (bInWA == FALSE))
	return mColorNotSelectOut;
    return mColorNotSelectIn;
};

void ParamGL_cls::
SetColor(int tag, float r, float g, float b, float a){
    float *pTgt;
    switch(tag){
    case _PARAMGL_SELECTED_IN:
	pTgt = mColorSelectedIn; break;
    case _PARAMGL_SELECTED_OUT:
	pTgt = mColorSelectedOut; break;
    case _PARAMGL_NOTSELECT_IN:
	pTgt = mColorNotSelectIn; break;
    case _PARAMGL_NOTSELECT_OUT:
	pTgt = mColorNotSelectOut; break;
    }
    pTgt[0] = r; pTgt[1] = g; pTgt[2] = b; pTgt[3] = a;
}

// ------

Parameters_cls::
Parameters_cls(unsigned int dimension, std::wstring labelStr):
    mLabelStr(labelStr),
    mpCMasterParam(NULL), mpCSlaveParam(NULL),
    mVCKeys(0)
{
    mpInterpolatedValue = new double [dimension];
    mVstrItemLabel.resize(dimension);
    mVstrItemLabelD1.resize(dimension);
    wchar_t buf[8];
    swprintf(buf, 8, L"%d", dimension);
    mCConfList.Add(&mCDimension, L"dimension ", buf);
    mCConfList.Add(&mCbTimeDependent, L"timeDepend ", L"0");
    mCConfList.Add(&mCbVisible, L"visible ", L"1");
    mCConfList.Add(&mCbShow, L"show ", L"1");
    mCConfList.Add(&mCbSelectable, L"selectable ", L"1");
    mbUpdateCamera = FALSE;

    bool bPrim = FALSE;

    if (labelStr == L"TimeRemap"){
	mCbSelectable.mbBody = FALSE;
    }

    if (labelStr == L"Light"){
	SetColor(_PARAMGL_SELECTED_IN  , 0.7, 1.0, 0, 1);
	SetColor(_PARAMGL_SELECTED_OUT , 0.4, 0.6, 0, 1);
	SetColor(_PARAMGL_NOTSELECT_IN , 0.3, 0.5, 0, 1);
	SetColor(_PARAMGL_NOTSELECT_OUT, 0.2, 0.3, 0, 1);
	mVstrItemLabel[0] = wxT("x");
	mVstrItemLabel[1] = wxT("y");
	mVstrItemLabel[2] = wxT("z");
	mVstrItemLabel[3] = wxT("Energy");
	mVstrItemLabelD1[0] = wxT("vx");
	mVstrItemLabelD1[1] = wxT("vy");
	mVstrItemLabelD1[2] = wxT("vz");
	mVstrItemLabelD1[3] = wxT("dEnergy");
	bPrim = TRUE;
    }

    if (labelStr == L"CamPos" || labelStr == L"AncPos"){
	SetColor(_PARAMGL_SELECTED_IN  , 0.2, 1.0, 0.2, 1);
	SetColor(_PARAMGL_SELECTED_OUT , 0.1, 0.5, 0.1, 1);
	SetColor(_PARAMGL_NOTSELECT_IN , 0.1, 0.8, 0.1 , 1);
	SetColor(_PARAMGL_NOTSELECT_OUT, 0.0, 0.3, 0.0, 1);
	mVstrItemLabel[0] = wxT("x");
	mVstrItemLabel[1] = wxT("y");
	mVstrItemLabel[2] = wxT("z");
	mVstrItemLabelD1[0] = wxT("vx");
	mVstrItemLabelD1[1] = wxT("vy");
	mVstrItemLabelD1[2] = wxT("vz");
	bPrim = TRUE;
    }

    if (labelStr == L"CamRotYPR"){
	mVstrItemLabel[0] = wxT("Yaw");
	mVstrItemLabel[1] = wxT("Pitch");
	mVstrItemLabel[2] = wxT("Row");
	bPrim = TRUE;
    }

    if (labelStr == L"HSpot"){
	SetColor(_PARAMGL_SELECTED_IN  , 1.0, 0.5, 0.5, 1);
	SetColor(_PARAMGL_SELECTED_OUT , 0.5, 0.1, 0.1, 1);
	SetColor(_PARAMGL_NOTSELECT_IN , 0.6, 0.2, 0.2 , 1);
	SetColor(_PARAMGL_NOTSELECT_OUT, 0.3, 0.0, 0.0, 1);
	mVstrItemLabel[0] = wxT("x");
	mVstrItemLabel[1] = wxT("y");
	mVstrItemLabel[2] = wxT("z");
	mVstrItemLabel[3] = wxT("size");
	mVstrItemLabel[4] = wxT("pow");
	mVstrItemLabelD1[0] = wxT("vx");
	mVstrItemLabelD1[1] = wxT("vy");
	mVstrItemLabelD1[2] = wxT("vz");
	mVstrItemLabelD1[3] = wxT("size");
	mVstrItemLabelD1[4] = wxT("pow");
	bPrim = TRUE;
    }
    
    if (labelStr == L"Color"){
	mCbTimeDependent.mbBody = TRUE;
    }

    if (labelStr == L"BoxSize" || labelStr == L"Clipping"){
	mVstrItemLabel[0] = wxT("minX");
	mVstrItemLabel[1] = wxT("minY");
	mVstrItemLabel[2] = wxT("minZ");
	mVstrItemLabel[3] = wxT("maxX");
	mVstrItemLabel[4] = wxT("maxY");
	mVstrItemLabel[5] = wxT("maxZ");
	mVstrItemLabelD1[0] = wxT("dminX");
	mVstrItemLabelD1[1] = wxT("dminY");
	mVstrItemLabelD1[2] = wxT("dminZ");
	mVstrItemLabelD1[3] = wxT("dmaxX");
	mVstrItemLabelD1[4] = wxT("dmaxY");
	mVstrItemLabelD1[5] = wxT("dmaxZ");
	bPrim = TRUE;
    }

    if (labelStr == L"BoxRotate"){
	mVstrItemLabel[0] = wxT("z");
	mVstrItemLabel[1] = wxT("x");
	mVstrItemLabel[2] = wxT("y");
	mVstrItemLabel[3] = wxT("z");
	mVstrItemLabelD1[0] = wxT("z");
	mVstrItemLabelD1[1] = wxT("x");
	mVstrItemLabelD1[2] = wxT("y");
	mVstrItemLabelD1[3] = wxT("z");
	bPrim = TRUE;
    }

    if (labelStr == L"Isosurface"){
	mVstrItemLabel[0] = wxT("Critical");
	mVstrItemLabel[1] = wxT("Alpha");
    }

    if (labelStr == L"DataRotate"){
	mVstrItemLabel[0] = wxT("degree");
    }

    if (labelStr == L"CrossSection"){
	mVstrItemLabel[0] = wxT("Z layer");
	mVstrItemLabel[1] = wxT("Emboss");
	mVstrItemLabel[2] = wxT("Alpha");
	mVstrItemLabel[3] = wxT("Crit");
    }

    if (labelStr == L"PovObject"){
	mVstrItemLabel[0] = wxT("X");
	mVstrItemLabel[1] = wxT("Y");
	mVstrItemLabel[2] = wxT("Z");
	mVstrItemLabel[3] = wxT("scaleX");
	mVstrItemLabel[4] = wxT("scaleY");
	mVstrItemLabel[5] = wxT("scaleZ");
	mVstrItemLabel[6] = wxT("rotY");
	mVstrItemLabel[7] = wxT("rotP");
	mVstrItemLabel[8] = wxT("rotR");
	mVstrItemLabel[9] = wxT("rotY2");
	mVstrItemLabel[10] = wxT("Alpha");
    }

    if (labelStr == L"PovDiffObject"){
	mVstrItemLabel[0] = wxT("X");
	mVstrItemLabel[1] = wxT("Y");
	mVstrItemLabel[2] = wxT("Z");
	mVstrItemLabel[3] = wxT("scaleX");
	mVstrItemLabel[4] = wxT("scaleY");
	mVstrItemLabel[5] = wxT("scaleZ");
	mVstrItemLabel[6] = wxT("rotY");
	mVstrItemLabel[7] = wxT("rotP");
	mVstrItemLabel[8] = wxT("rotR");
	mVstrItemLabel[9] = wxT("rotY2");
    }
    
    if (labelStr == L"BaseAlpha"){
	mVstrItemLabel[0] = wxT("Alpha");
	mVstrItemLabelD1[0] = wxT("dAlpha");
	bPrim = TRUE;
    }

    if (labelStr == L"DensityCurvePoints"){
	mVstrItemLabel[0] = wxT("x(in)");
	mVstrItemLabel[1] = wxT("y");
	mVstrItemLabel[2] = wxT("z");
	mVstrItemLabel[3] = wxT("x(out)");
	mVstrItemLabel[4] = wxT("y");
	mVstrItemLabel[5] = wxT("z");
    }

    if (labelStr == L"LevelAndGamma"){
	mVstrItemLabel[0] = wxT("InMin");
	mVstrItemLabel[1] = wxT("InMax");
	mVstrItemLabel[2] = wxT("OutMin");
	mVstrItemLabel[3] = wxT("OutMax");
	mVstrItemLabel[4] = wxT("gamma");
	mVstrItemLabel[5] = wxT("offset");
    }

    if (labelStr == L"ThreeView"){
	mVstrItemLabel[0] = wxT("LayerX");
	mVstrItemLabel[1] = wxT("LayerY");
	mVstrItemLabel[2] = wxT("LayerZ");
	mVstrItemLabel[3] = wxT("Emboss");
	mVstrItemLabel[4] = wxT("Alpha");
	mVstrItemLabel[5] = wxT("Crit");
    }

    if (labelStr == L"StreamLine"){
	mVstrItemLabel[0] = wxT("L Alpha");
	mVstrItemLabel[1] = wxT("P Alpha");
	mVstrItemLabel[2] = wxT("P Size");
    }

    if (labelStr == L"VelField"){
	mVstrItemLabel[0] = wxT("alpha");
	mVstrItemLabel[1] = wxT("length");
    }
    
    //if (bPrim == TRUE && mspVPInstancePrim != NULL) mspVPInstancePrim->push_back(this);
}

Parameters_cls:: // $B%3%T!<%3%s%9%H%i%/%?(B
Parameters_cls(const Parameters_cls &obj){

    mLabelStr = obj.mLabelStr;
    mVstrItemLabel = obj.mVstrItemLabel;
    mVstrItemLabelD1 = obj.mVstrItemLabelD1;

    mCDimension      = obj.mCDimension;
    mCbTimeDependent = obj.mCbTimeDependent;
    mCbVisible       = obj.mCbVisible;
    mCbShow          = obj.mCbShow;
    mCbSelectable    = obj.mCbSelectable;
    mbUpdateCamera   = obj.mbUpdateCamera;

    LOOP_RGBA{
	mColorSelectedIn[rgba] = obj.mColorSelectedIn[rgba];
	mColorSelectedOut[rgba] = obj.mColorSelectedOut[rgba];
	mColorNotSelectIn[rgba] = obj.mColorNotSelectIn[rgba];
	mColorNotSelectOut[rgba] = obj.mColorNotSelectOut[rgba];
    }

    DefineConfList();
    
    // $B?7$7$$$[$&$,<BBN$H$9$k(B
    mpCSlaveParam = obj.mpCSlaveParam;
    if (mpCSlaveParam != NULL) mpCSlaveParam->SetMasterParam(this);
    mpCMasterParam = obj.mpCMasterParam;
    if (mpCMasterParam != NULL)	mpCMasterParam->SetSlaveParam(this);

    mpInterpolatedValue = new double [mCDimension.mBody];
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	mpInterpolatedValue[i] = obj.mpInterpolatedValue[i];
    
    mVCKeys.clear();
    size_t keySize = obj.mVCKeys.size();
    for (size_t index = 0; index < keySize; index++)
	mVCKeys.push_back(obj.mVCKeys[index]);

    for (unsigned int i = 0; i < mspVPInstancePrim->size(); i++)
	if ((*mspVPInstancePrim)[i] == &obj)
	    (*mspVPInstancePrim)[i] = this;
}    

void Parameters_cls::
CopyFromP(const Parameters_cls *pObj){

    mLabelStr = pObj->mLabelStr;

    //_ASSERT(mCDimension.mBody == pObj->mCDimension.mBody);    
    mCDimension      = pObj->mCDimension;
    mCbTimeDependent = pObj->mCbTimeDependent;
    mCbVisible       = pObj->mCbVisible;
    mCbShow          = pObj->mCbShow;
    mCbSelectable    = pObj->mCbSelectable;
    mbUpdateCamera   = pObj->mbUpdateCamera;
    
    LOOP_RGBA{
	mColorSelectedIn[rgba] = pObj->mColorSelectedIn[rgba];
	mColorSelectedOut[rgba] = pObj->mColorSelectedOut[rgba];
	mColorNotSelectIn[rgba] = pObj->mColorNotSelectIn[rgba];
	mColorNotSelectOut[rgba] = pObj->mColorNotSelectOut[rgba];
    }

    DefineConfList();

    // Mement$B$NJ}$,?7$7$$$N$G!"(B
    // $B%j%s%/$,(BMemento$BB&$K=q$-49$o$C$F$O:$$k(B
    mpCSlaveParam = pObj->mpCSlaveParam;
    //if (mpCSlaveParam != NULL) mpCSlaveParam->SetMasterParam(this);
    mpCMasterParam = pObj->mpCMasterParam;
    //if (mpCMasterParam != NULL) mpCMasterParam->SetSlaveParam(this);

    // !! $B<!85$N0c$&%G!<%?$N%3%T!<;2>H$7$F$$$k%]%$%s%?$,DI?o$G$-$J$/$F4m81!#(B
    // $B7Y9p$r=P$7$F=*N;$G$b$9$k$h$&$K$7$H$/!)(B
    if (mCDimension.mBody != pObj->mCDimension.mBody){
	delete mpInterpolatedValue;
	mpInterpolatedValue = new double [mCDimension.mBody];
    }
    
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	mpInterpolatedValue[i] = pObj->mpInterpolatedValue[i];
    
    mVCKeys.clear();
    size_t keySize = pObj->mVCKeys.size();
    for (size_t index = 0; index < keySize; index++)
	mVCKeys.push_back(pObj->mVCKeys[index]);
}

void Parameters_cls::
DefineConfList(){
    mCConfList.Clear();
    mCConfList.Add(&mCDimension, L"dimension ", mCDimension.GetItemStr());
    mCConfList.Add(&mCbTimeDependent, L"timeDepend ", mCbTimeDependent.GetItemStr());
    mCConfList.Add(&mCbVisible, L"visible ", mCbVisible.GetItemStr());
    mCConfList.Add(&mCbShow, L"show ", mCbShow.GetItemStr());
    mCConfList.Add(&mCbSelectable, L"selectable ", mCbSelectable.GetItemStr());
}

Parameters_cls::
~Parameters_cls(){
    if (mpInterpolatedValue != NULL)
	delete mpInterpolatedValue;

    if (MainGLCanvas_cls::mspCSelectedItem == this)
	MainGLCanvas_cls::SelectItem(NULL);
}

void Parameters_cls::
SetMasterParam(Parameters_cls *pCMasterParam){
    mpCMasterParam = pCMasterParam;
}

void Parameters_cls::
SetSlaveParam(Parameters_cls *pCSlaveParam){
    mpCSlaveParam = pCSlaveParam;
    mpCSlaveParam->SetMasterParam(this);
}

void Parameters_cls::
RemapKeyFrame(int oldMax, int newMax){
    for (unsigned int i = 0; i < GetKeyNumber(); i++)
	mVCKeys[i].mDoubleIndex *= (double) oldMax / newMax;
}

void Parameters_cls::
RemapKeyFrameAll(int oldMax, int newMax){
    for (unsigned int i = 0; i < mspVPInstancePrim->size(); i++){
	//MsgBox(NULL, (*mspVPInstancePrim)[i]->mLabelStr, L"!!");
	(*mspVPInstancePrim)[i]->RemapKeyFrame(oldMax, newMax);
    }
}

void Parameters_cls::
SetFrameNumber(int frame, unsigned int quantum){
    msFrameNumber = frame;
    if (frame >= 1)
	msResolution = 1.0/msFrameNumber;
    else
	msResolution = 1.0;
    //msStepQuantum = quantum;
}

void Parameters_cls::
SetCurrentDoubleIndex(double doubleIndex){
    msCurrentDoubleIndex = doubleIndex;
}

double Parameters_cls::
Quantize(double doubleIndex, unsigned int frameNumber){
    if (frameNumber == _PARAM_USE_COMMON) frameNumber = msFrameNumber;
    int index = doubleIndex * frameNumber * 2;
    index = (index + 1) / 2;
    if (frameNumber == 0) 
	doubleIndex = 0;
    else
	doubleIndex = (double)index/ frameNumber;
    return doubleIndex;
};

void Parameters_cls::
SetTimeDependent(bool bTimeDependent, bool bSlave){
    if (bSlave == FALSE)
	GCUndoList.OpenBlock();

    if (mpCSlaveParam != NULL)
	mpCSlaveParam->SetTimeDependent(bTimeDependent, TRUE);

    double *pTmp = new double [mCDimension.mBody];
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	pTmp[i] = mpInterpolatedValue[i];
    GCUndoList.SetNewCommand
	(new ComParamClearKey_cls(this, pTmp),
	 new ComParamMemento_cls(this));
    GCUndoList.ExecCurrent();
    GCUndoList.PushNewCommand
	(new ComParamSetTimeDependent_cls(this, bTimeDependent),
	 new ComParamSetTimeDependent_cls(this, !bTimeDependent));
    GCUndoList.ExecCurrent();
    if (bSlave == FALSE) 
	GCUndoList.CloseBlock();
    delete pTmp;
}

void Parameters_cls::
CopyParameterFromP(Parameters_cls *pCParamFrom, bool bSlave){
    if (bSlave == FALSE)
	GCUndoList.OpenBlock();
    
    if (mpCSlaveParam != NULL)
	mpCSlaveParam->SetTimeDependent(TRUE);

    GCUndoList.SetNewCommand
	(new ComParamMemento_cls(this, pCParamFrom),
	 new ComParamMemento_cls(this));
    GCUndoList.ExecCurrent();

    if (bSlave == FALSE) 
	GCUndoList.CloseBlock();
}

void Parameters_cls::
ExecCommandTimeDependent(bool bTimeDependent){
    mCbTimeDependent.mbBody = bTimeDependent;
}

unsigned int Parameters_cls::
GetKeyNumber(){
    return (unsigned int)mVCKeys.size();
}

void Parameters_cls::
SetKeyToValueAt(KeyFrame_cls *pCKey, double doubleIndex){
    double *pInitD1 = new double [mCDimension.mBody];
    //double *pInitD2 = new double [mCDimension.mBody];
    
    int prev = GetAdjacentKey(doubleIndex, -1);
    int next = GetAdjacentKey(doubleIndex, +1);
    CalcVelocity(prev, next, pInitD1);
    //for (unsigned int i = 0; i < mCDimension.mBody; i++) pInitD2[i] = 0;

    pCKey->Set(doubleIndex, GetInterpTypeFromDouble(doubleIndex),
	       GetPInterpolatedValue(doubleIndex), pInitD1);
    
    delete pInitD1;
    //delete pInitD2;
}

void Parameters_cls::
SetSnippingTag(std::vector < bool > &mVBSnipping, int maxFrame){
    for (unsigned int i = 0; i < GetKeyNumber(); i++){
	int frame = (mVCKeys[i].mDoubleIndex * maxFrame + 0.1);
	mVBSnipping[frame] = TRUE;
    }
}

void Parameters_cls::
SortKey(){
    if (mpCSlaveParam != NULL)
	mpCSlaveParam->SortKey();

    int status = 1;
    // $BD>@\%/%i%9$r%=!<%H$9$k$N$OF0E*%a%b%j$N4X78$GLq2p!#<jF0$G(B
    while (status == 1){
	status = 0;
	size_t keySize = mVCKeys.size();
	for (size_t index = 0; index < keySize - 1; index++){
	    if (mVCKeys[index] > mVCKeys[index + 1]){
		status = 1;
		KeyFrame_cls tmpKey(mCDimension.mBody);
		tmpKey.CopyWithoutConstruct(mVCKeys[index]);
		mVCKeys[index].CopyWithoutConstruct(mVCKeys[index+1]);
		mVCKeys[index+1].CopyWithoutConstruct(tmpKey);
		index = 0;
	    }
	}
    }
}

void Parameters_cls::
CalcVelocity(int prev, int next, double *pArg){
    if (prev == next)
	for (unsigned int i = 0; i < mCDimension.mBody; i++)
	    pArg[i] = 0;
    else
	for (unsigned int i = 0; i < mCDimension.mBody; i++)
	    pArg[i] = 
		(mVCKeys[next].mpValue[i]-mVCKeys[prev].mpValue[i])/
		(mVCKeys[next].mDoubleIndex - mVCKeys[prev].mDoubleIndex);
}

void Parameters_cls::
AdjustAutoControl(){
    size_t keySize = mVCKeys.size();
    if (keySize == 1) return;
    unsigned int dimension = mCDimension.mBody;

    for (unsigned int index = 0; index < keySize; index++){
	unsigned int prev = 
	    (index != 0)                ? index - 1 : 0;
	unsigned int next = 
	    (index != keySize-1) ? index + 1 : keySize-1;
	switch (mVCKeys[index].mType){
	case _PARAM_HERMITE_AUTO:
	    CalcVelocity(prev, next, mVCKeys[index].mpDValue1);
	    for (unsigned int i = 0; i < dimension; i++)
		mVCKeys[index].mpDValue2[i] = mVCKeys[index].mpDValue1[i];
	    break;
	case _PARAM_HERMITE_ADJ:
	    CalcVelocity(prev, next, mVCKeys[index].mpDValue1);
	    break;
	default:
	    break;
	}
    }
}

void Parameters_cls::
InitSet(double *pFrom){
    GCUndoList.SetNewCommand
	(new ComParamClearKey_cls(this, pFrom),
	 new ComParamMemento_cls(this));
    GCUndoList.ExecCurrent();
}

void Parameters_cls::
ExecCommandInitSet(double *pFrom){
    mVCKeys.clear();
    KeyFrame_cls newKey(mCDimension.mBody);
    //double *pInitD = new double [mCDimension.mBody];
    //for (unsigned int i = 0; i < mCDimension.mBody; i++) pInitD[i] = 0;
    newKey.Set(0, _PARAM_HERMITE_AUTO, pFrom);
    mVCKeys.push_back(newKey);
    //delete [] pInitD;
}

void Parameters_cls::
InsertKey(double doubleIndex, double *pValue){
    
    KeyFrame_cls CKey(mCDimension.mBody);//, CKeySlave(0);
    SetKeyToValueAt(&CKey, doubleIndex);
    
    if (pValue != NULL)	CKey.SetValue(pValue);
    
    SetKey(CKey);
}

int Parameters_cls::
CheckOverWrite(double doubleIndex, double resolution){
    int overwrite = -1;
    size_t keySize = mVCKeys.size();
    for (size_t index = 0; index < keySize; index++){
	double difference = doubleIndex - mVCKeys[index].mDoubleIndex;
	difference = sqrt(difference * difference);
	if (difference < resolution / 2.0)
	    overwrite = (int)index;
    }
    return overwrite;
}

void Parameters_cls::
SetKey(KeyFrame_cls CKey, KeyFrame_cls CKeySlave, double resolution){

    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;

    if (mpCSlaveParam != NULL){
	if (CKeySlave.GetDimension() != 0)
	    mpCSlaveParam->SetKey(CKeySlave, CKeySlave, resolution);
	else
	    mpCSlaveParam->InsertKey(CKey.mDoubleIndex);
    }

    //Check Overwrite
    int overwrite = CheckOverWrite(CKey.mDoubleIndex, resolution);

    if (overwrite == -1)
	GCUndoList.SetNewCommand
	    (new ComParamSetKeyC_cls(this, CKey),
	     new ComParamDelKey_cls(this, 
				    GetKeyFrameIndexToInsert(CKey.mDoubleIndex)));
    else
	GCUndoList.SetNewCommand
	    (new ComParamOvwtKeyC_cls(this, overwrite, CKey),
	     new ComParamOvwtKeyC_cls(this, overwrite, mVCKeys[overwrite]));
    GCUndoList.ExecCurrent();
}

/*
void Parameters_cls::
SetKey(double doubleIndex, 
       double *pFrom, double *pDValue1, double *pDValue2,
       unsigned int type,
       double resolution,
       double *pFromSlave, double *pDValue1Slave, double *pDValue2Slave){
    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;

    if (mpCSlaveParam != NULL){
	if (pFromSlave != NULL)
	    mpCSlaveParam->SetKey(doubleIndex, pFromSlave, 
				  pDValue1Slave, pDValue2Slave, type);
	else
	    mpCSlaveParam->InsertKey(doubleIndex);
    }

    if (mCbTimeDependent.mbBody == FALSE) doubleIndex = 0;
    
    // Check Overwrite
    int overwrite = CheckOverWrite(doubleIndex, resolution);

   
    if (overwrite == -1)
	GCUndoList.SetNewCommand
	    (new ComParamSetKey_cls(this, doubleIndex, 
				    pFrom, pDValue1, pDValue2, type),
	     new ComParamDelKey_cls(this, GetKeyFrameIndexToInsert(doubleIndex)));
    else{
	KeyFrame_cls CKey(mCDimension.mBody);
	CKey.Set(mVCKeys[overwrite].mDoubleIndex,
		 pFrom, pDValue1, pDValue2, type, 
		 mVCKeys[overwrite].mbSelected);
	
	GCUndoList.SetNewCommand
	    (new ComParamOvwtKeyC_cls(this, overwrite, CKey),
	     new ComParamOvwtKeyC_cls(this, overwrite, mVCKeys[overwrite]));
    }
    GCUndoList.ExecCurrent();
}
*/

void Parameters_cls::
ExecCommandSetKey(KeyFrame_cls CKey){
    mVCKeys.push_back(CKey);
    SortKey();
    AdjustAutoControl();
}

void Parameters_cls::
ExecCommandSetKey(double doubleIndex, double *pFrom, unsigned int type,
		  double *pDValue1, double *pDValue2){
    KeyFrame_cls newKey(mCDimension.mBody);
    newKey.Set(doubleIndex, type, pFrom, pDValue1, pDValue2);
    mVCKeys.push_back(newKey);
    SortKey();
    AdjustAutoControl();
}

void Parameters_cls::
ExecCommandOvwtKey(unsigned int index, KeyFrame_cls CKey){
    //CKey.mDoubleIndex = mVCKeys[index].mDoubleIndex;
    mVCKeys[index].CopyWithoutConstruct(CKey);
    SortKey();
    AdjustAutoControl();
}

/*
void Parameters_cls::
ExecCommandOvwtKey(unsigned int index, unsigned int type, double *pFrom,
		   double *pDValue1, double *pDValue2, bool bSelected){
    KeyFrame_cls newKey(mCDimension.mBody);
    newKey.Set(mVCKeys[index].mDoubleIndex, pFrom, pDValue1, pDValue2, type, bSelected);
    mVCKeys[index].CopyWithoutConstruct(newKey);
    SortKey();
    AdjustAutoControl();
}
*/

bool Parameters_cls::
SetKeyIfChanged(double doubleIndex, bool bForce){
    if (doubleIndex == _PARAM_CURRENT_DINDEX)
	doubleIndex = msCurrentDoubleIndex;
    
    if (mCbTimeDependent.mbBody == FALSE)
	doubleIndex = 0;

    double *pCache = new double [mCDimension.mBody];
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	pCache[i] = mpInterpolatedValue[i];
    GetPInterpolatedValue(doubleIndex);
    bool bChanged = FALSE;
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	if (pCache[i] != mpInterpolatedValue[i])
	    bChanged = TRUE;
    if (bForce == TRUE) bChanged = TRUE;

    if (bChanged == TRUE){
	int keyframe = GetKeyFrameIndexFromDouble(doubleIndex, _PARAM_USE_COMMON, FALSE);
	bool bSelected = FALSE;
	if (keyframe != -1)
	    bSelected = mVCKeys[keyframe].mbSelected;
	InsertKey(doubleIndex, pCache);
	if (keyframe != -1)
	    mVCKeys[keyframe].mbSelected = bSelected;
	GetPInterpolatedValue(doubleIndex);
    }
    delete pCache;
    return bChanged;
}

bool Parameters_cls::
MoveTimeSelectedKey(double doubleIndexDiff){
    bool retval = FALSE;

    if ((doubleIndexDiff < msResolution * 0.5) && (doubleIndexDiff > -msResolution * 0.5))
	return FALSE;
    if (mVCKeys.size() == 0) return FALSE;

    GCUndoList.OpenBlock();
    if (doubleIndexDiff < 0){
	for (int i = 0; i < (int)mVCKeys.size(); i++)
	    if (mVCKeys[i].mbSelected == TRUE){
		int processBit = MoveKey(i, doubleIndexDiff);
		if ((processBit & 1) != 0) retval = TRUE;
		if ((processBit & _PARAM_PROCESS_OVRW) != 0) i--; 
		// key$B$,(B1$B$D8:$C$?(B
	    }
    }
    else{
	for (int i = (int)mVCKeys.size() - 1; i >= 0; i--)
	    if (mVCKeys[i].mbSelected == TRUE)
		if ((MoveKey(i, doubleIndexDiff) & 1) == TRUE) retval = TRUE;
    }
    GCUndoList.CloseBlock();
    if (retval == FALSE) GCUndoList.UndoCurrent(); 

    return retval;
}

void Parameters_cls::
MovePosForBox(double pPos[6], double *dpos, double dscale[3], int hitObj){
   
    switch(hitObj){
    case 1:
	LOOP_XYZ pPos[xyz]   += dpos[xyz];
	LOOP_XYZ pPos[xyz+3] += dpos[xyz];
	break;
    case 2:
	pPos[X] += dpos[X]; break;
    case 3:
	pPos[X+3] += dpos[X]; break;
    case 4:
	pPos[Y] += dpos[Y]; break;
    case 5:
	pPos[Y+3] += dpos[Y]; break;
    case 6:
	pPos[Z] += dpos[Z]; break;
    case 7:
	pPos[Z+3] += dpos[Z]; break;
    }

    LOOP_XYZ
	if (dscale[xyz] > 1 && (pPos[xyz] == pPos[xyz+3])){
	    pPos[xyz] -= 0.001;
	    pPos[xyz+3] += 0.001;
	}
    
    double center[3];
    LOOP_XYZ center[xyz] = (pPos[xyz] + pPos[xyz+3]) / 2;
    LOOP_XYZ pPos[xyz] *= dscale[xyz];
    LOOP_XYZ pPos[xyz] += (1.0 - dscale[xyz]) * center[xyz];
    LOOP_XYZ pPos[xyz+3] *= dscale[xyz];
    LOOP_XYZ pPos[xyz+3] += (1.0 - dscale[xyz]) * center[xyz];
}

void Parameters_cls::
MovePosForTwoPoints(double pPos[6], double *dpos, double dscale[3], int hitObj){
    switch(hitObj){
    case 1:
	LOOP_XYZ pPos[xyz]   += dpos[xyz];
	break;
    case 2:
	LOOP_XYZ pPos[xyz+3] += dpos[xyz];
	break;
    }
}

bool Parameters_cls::
MovePosSelectedKey(double *dpos, double dscale[3], int hitObj){
    bool retval = FALSE;
    GCUndoList.OpenBlock();
    double *pos = new double [mCDimension.mBody];
    size_t keySize = mVCKeys.size();
    int dimension = (int)mCDimension.mBody;
    for (size_t i = 0; i < keySize; i++){
	if (mVCKeys[i].mbSelected == FALSE) continue;
    
	for (int j = 0; j < dimension; j++)
	    pos[j] = mVCKeys[i].mpValue[j];
	
	if (mLabelStr == L"BoxSize"){
	    LOOP_XYZ dscale[3] = 1;
	    MovePosForBox(pos, dpos, dscale, hitObj);
	}
	else if (mLabelStr.Matches(L"ClippingBox"))
	    MovePosForBox(pos, dpos, dscale, hitObj);
	else
	    for (int j = 0; j < 3; j++)
		pos[j] += dpos[j];
	
	if (mLabelStr.Matches(L"PovObject") || 
	    mLabelStr.Matches(L"PovDiffObject"))
	    LOOP_XYZ pos[xyz+3] *= dscale[xyz];
	
	
	if (mLabelStr.Matches(L"ClippingBox")){
	    LOOP_XYZ if (pos[xyz] < 0)   pos[xyz] = 0;
	    LOOP_XYZ if (pos[xyz+3] < 0) pos[xyz+3] = 0;
	    LOOP_XYZ if (pos[xyz] > 1)   pos[xyz] = 1;
	    LOOP_XYZ if (pos[xyz+3] > 1) pos[xyz+3] = 1;
	}
	
	KeyFrame_cls CKey(mVCKeys[i]);//, CKeySlave(0);
	CKey.SetValue(pos);
	SetKey(CKey);
	
	mVCKeys[i].mbSelected = TRUE;
	retval = TRUE;
    }
    delete [] pos;
    
    GCUndoList.CloseBlock();
    return FALSE;
}

void Parameters_cls::
RemapKeys(double rate){
    size_t keySize = mVCKeys.size();
    if (rate < 1){
	for (size_t i = 0; i < keySize; i++)
	    if (mVCKeys[i].mDoubleIndex != 0)
		MoveKey((int)i, mVCKeys[i].mDoubleIndex * (rate - 1.0));
    }
    else{
	for (int i = (int)keySize - 1; i >= 0; i--)
	    if (mVCKeys[i].mDoubleIndex != 0)
		MoveKey(i, mVCKeys[i].mDoubleIndex * (rate - 1.0));
    }
}

int Parameters_cls::
MoveKey(int indexToMove, double doubleIndexDiff,
	unsigned int frameNumber, bool bSlave){

    double resolution;
    if (frameNumber == _PARAM_USE_COMMON){
	frameNumber = msFrameNumber;
	resolution = msResolution;
    }
    else
	resolution = 1.0 / frameNumber;

    double orgDoubleIndex = mVCKeys[indexToMove].mDoubleIndex;
    unsigned int orgType = mVCKeys[indexToMove].mType;
    double newDoubleIndex = orgDoubleIndex + doubleIndexDiff;
    newDoubleIndex = Quantize(newDoubleIndex, frameNumber);
    if (newDoubleIndex < 0) newDoubleIndex = 0;
    if (newDoubleIndex > 1) newDoubleIndex = 1;

    if (newDoubleIndex == orgDoubleIndex) return 0;

    // Check Overwrite
    int overwrite = CheckOverWrite(newDoubleIndex, resolution);
    /*
    int overwrite = -1;
    for (unsigned int index = 0; index < mVCKeys.size(); index++){
	double difference = newDoubleIndex - mVCKeys[index].mDoubleIndex;
	difference = sqrt(difference * difference);
	if (difference < resolution / 2.0 && index != indexToMove)
	    overwrite = index;
    }
    */

    if (bSlave == FALSE) GCUndoList.OpenBlock();
    if (mpCSlaveParam != NULL)
	mpCSlaveParam->MoveKey(indexToMove, doubleIndexDiff,
			       frameNumber, TRUE);

    int retFlag = 1;

    if (overwrite == -1){
	GCUndoList.SetNewCommand
	    (new ComParamMoveKey_cls(this, orgDoubleIndex, newDoubleIndex, resolution),
	     new ComParamMoveKey_cls(this, newDoubleIndex, orgDoubleIndex, resolution));
	GCUndoList.ExecCurrent();
    }
    else{
	retFlag += _PARAM_PROCESS_OVRW;
	int typeOvrwted = mVCKeys[overwrite].mType;
	
	KeyFrame_cls CKey(mVCKeys[indexToMove]);
	CKey.mDoubleIndex = mVCKeys[overwrite].mDoubleIndex;

	GCUndoList.SetNewCommand
	    (new ComParamOvwtKeyC_cls(this, overwrite, CKey),
	     new ComParamOvwtKeyC_cls(this, overwrite, mVCKeys[overwrite]));

	GCUndoList.PushNewCommand
	    (new ComParamDelKey_cls(this, indexToMove),
	     new ComParamSetKey_cls(this, orgDoubleIndex, 
				    GetPKeyFrameValue(indexToMove),
				    GetPKeyFrameValueD1(indexToMove),
				    GetPKeyFrameValueD2(indexToMove),
				    orgType));
	GCUndoList.ExecCurrent();
	//$B>e=q$-$5$l$?(Bkey$B$OA*Br:Q$_(B
	if (bSlave == FALSE){
	    if (indexToMove < overwrite)
		mVCKeys[overwrite-1].mbSelected = TRUE;
	    else
		mVCKeys[overwrite].mbSelected = TRUE;
	}
    }

    if (bSlave == FALSE) GCUndoList.CloseBlock();

    return retFlag;
}

void Parameters_cls::
ExecCommandMoveKey(double doubleIndexFrom, double doubleIndexTo, double resolution){
    int index = GetKeyFrameIndexFromDouble(doubleIndexFrom, resolution);
    if (index == -1) return;
    mVCKeys[index].mDoubleIndex = doubleIndexTo;
    SortKey();
    AdjustAutoControl();
}

void Parameters_cls::
SetKeyType(unsigned int index, unsigned int mode){
    if (mpCSlaveParam != NULL)
	mpCSlaveParam->SetKeyType(index, mode);    

    if (mVCKeys[index].mType == mode) return;
   
    GCUndoList.SetNewCommand
	(new ComParamSetKeyType_cls(this, index, mode),
	 new ComParamSetKeyType_cls(this, index, mVCKeys[index].mType));
    GCUndoList.ExecCurrent();
}

void Parameters_cls::
SetKeyTypeMulti(unsigned int mode){
    if (mpCSlaveParam != NULL){
	for (unsigned int i = 0; i < mpCSlaveParam->GetKeyNumber(); i++)
	    mpCSlaveParam->mVCKeys[i].mbSelected = mVCKeys[i].mbSelected;
	mpCSlaveParam->SetKeyTypeMulti(mode);
    }

    unsigned int num = 0;
    for (unsigned int i = 0; i < GetKeyNumber(); i++)
	if (mVCKeys[i].mbSelected == TRUE) num++;

    if (num == 0) return;

    unsigned int *listIndexFront = new unsigned int [num];
    unsigned int *listTypeFront  = new unsigned int [num];
    unsigned int *listIndexBack  = new unsigned int [num];
    unsigned int *listTypeBack   = new unsigned int [num];
    // UndoList$B$K<u$1EO$7$F!"8~$&$G(Bdelte$B$7$F$b$i$&(B

    unsigned int index = 0;
    for (unsigned int i = 0; i < GetKeyNumber(); i++)
	if (mVCKeys[i].mbSelected == TRUE){
	    listIndexFront[index] = i;
	    listIndexBack[index] = i;
	    listTypeFront[index] = mode;
	    listTypeBack[index] = mVCKeys[i].mType;
	    index++;
	}
   
    GCUndoList.SetNewCommand
	(new ComParamSetKeyTypeMulti_cls(this, num, listIndexFront, listTypeFront),
	 new ComParamSetKeyTypeMulti_cls(this, num, listIndexBack,  listTypeBack));
    GCUndoList.ExecCurrent();
}


void Parameters_cls::
ExecCommandSetKeyType(unsigned int index, unsigned int mode){
    mVCKeys[index].mType = mode;
    AdjustAutoControl();
}

void Parameters_cls::
ExecCommandSetKeyTypeMulti(unsigned int num, 
			   unsigned int *pIndex, unsigned int *pType){
    for (unsigned int i = 0; i < num; i++)
	mVCKeys[pIndex[i]].mType = pType[i];
    AdjustAutoControl();
}

bool Parameters_cls::
DeleteSelectedKey(){
    bool retval = FALSE;
    GCUndoList.OpenBlock();
    // FIX_ME $BMWAG$,5M$^$k:o=|$O5U=g$G%k!<%W$9$l$P$9$s$J$j!)(B
    for (int i = 0; i < (int)mVCKeys.size(); i++)
	if (mVCKeys[i].mbSelected == TRUE)
	    if (DeleteKey(mVCKeys[i].mDoubleIndex) == TRUE){
		retval = TRUE;
		i--;
	    }
    GCUndoList.CloseBlock();
    return retval;
}

#include "../MYAPI/MsgBox.h"

bool Parameters_cls::
DeleteKey(double doubleIndex, double resolution){
    if (mpCSlaveParam != NULL)
	mpCSlaveParam->DeleteKey(doubleIndex);
    if (mVCKeys.size() == 1) {
	mVCKeys[0].mDoubleIndex = 0;
	return FALSE; 
    }

    int index = GetKeyFrameIndexFromDouble(doubleIndex, resolution);
    if (index == -1) return FALSE;
    GCUndoList.SetNewCommand
	(new ComParamDelKey_cls(this, index),
	 new ComParamSetKey_cls(this, doubleIndex, 
				GetPKeyFrameValue(index),
				GetPKeyFrameValueD1(index), 
				GetPKeyFrameValueD2(index), 
				mVCKeys[index].mType)
	 );
    GCUndoList.ExecCurrent();
    return TRUE;
}

void Parameters_cls::
ExecCommandDeleteKey(int index){
    if (index == -1) return;
    mVCKeys[index].mDoubleIndex = 255;
    SortKey();
    AdjustAutoControl();
    mVCKeys.pop_back();
}

bool Parameters_cls::
IsKeyFrame(double doubleIndex, double resolution){
    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;
    bool isTrue = FALSE;
    for (unsigned int index = 0; index < GetKeyNumber(); index++)
	if (mVCKeys[index].mDoubleIndex < doubleIndex + resolution / 2.0 &&
	    mVCKeys[index].mDoubleIndex >= doubleIndex - resolution / 2.0)
	    isTrue = TRUE;
    return isTrue;
}

double Parameters_cls::
GetNextKeyPosDoubleIndex(double cur, double resolution){
    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;
    cur = cur + resolution / 2.0;
    if (cur < mVCKeys[0].mDoubleIndex)
	return mVCKeys[0].mDoubleIndex;
    for (unsigned int index = 0; index < GetKeyNumber() - 1; index++)
	if (cur < mVCKeys[index + 1].mDoubleIndex) 
	    return mVCKeys[index + 1].mDoubleIndex;
    return mVCKeys[GetKeyNumber() - 1].mDoubleIndex;
}

double Parameters_cls::
GetPrevKeyPosDoubleIndex(double cur, double resolution){
    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;
    cur = cur - resolution / 2.0;
    if (cur > mVCKeys[GetKeyNumber()-1].mDoubleIndex)
	return mVCKeys[GetKeyNumber()-1].mDoubleIndex;
    for (unsigned int index = GetKeyNumber(); index > 0; index--) 
	if (cur > mVCKeys[index - 1].mDoubleIndex)
	    return mVCKeys[index - 1].mDoubleIndex; 
    return 0;
}

int Parameters_cls::
GetKeyFrameIndexFromDouble(double doubleIndex, double resolution,
			   bool bFlagGetKeyEvenNotKeyFrame){
    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;
    unsigned int index;
    
    for (index = 0; index < GetKeyNumber(); index++){
	if (mVCKeys[index].mDoubleIndex < doubleIndex + resolution / 2.0 &&
	    mVCKeys[index].mDoubleIndex >= doubleIndex - resolution / 2.0)
	    return (int)index;
    }

    if (bFlagGetKeyEvenNotKeyFrame == TRUE){
	for (index = 0; index < GetKeyNumber(); index++)
	    if (mVCKeys[index].mDoubleIndex > doubleIndex + resolution / 2.0)
		if (index == 0)
		    return 0;
		else
		    return (int)index - 1;
	return GetKeyNumber() - 1;
    }

    return -1;
}

int Parameters_cls::
GetKeyFrameIndexToInsert(double doubleIndex, double resolution){
    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;

    for (unsigned int index = 0; index < GetKeyNumber(); index++)
	if (mVCKeys[index].mDoubleIndex > doubleIndex + resolution / 2.0)
	    return (int)index;
    return GetKeyNumber();
}

int Parameters_cls::
GetAdjacentKey(double doubleIndex, int direction){
    int index = GetKeyFrameIndexFromDouble(doubleIndex);
    if (index != -1){
	index += direction;
	if (index < 0) return 0;
	if (index >= (int)GetKeyNumber()) return GetKeyNumber() - 1;
	return index;
    }
    
    switch(direction){
    case 1:
	for (index = 0; index < (int)GetKeyNumber(); index++)
	    if (mVCKeys[index].mDoubleIndex > doubleIndex)
		return (int)index;
	return GetKeyNumber() - 1;
    case -1:
	for (index = GetKeyNumber() - 1; index >= 0; index--)
	    if (mVCKeys[index].mDoubleIndex < doubleIndex)
		return (int)index;
	return 0;
    }
    
    return 0;
}

int Parameters_cls::
GetSelectedKeyNum(){
    int num = 0;
    size_t keySize = mVCKeys.size();
    for (size_t i = 0; i < keySize; i++)
	if (mVCKeys[i].mbSelected == TRUE) num++;
    return num;
}

int Parameters_cls::
GetSelectedKeyIndex(){
    size_t keySize = mVCKeys.size();
    for (size_t i = 0; i < keySize; i++)
	if (mVCKeys[i].mbSelected == TRUE) return (int)i;
    return -1;
}

double Parameters_cls::
GetKeyFrameDoubleIndex(unsigned int index){
    if (index >= GetKeyNumber()) return -1;
    return mVCKeys[index].mDoubleIndex;
}

double* Parameters_cls::
GetPKeyFrameValue(unsigned int index){
    if (index >= GetKeyNumber()) return NULL;
    return mVCKeys[index].mpValue;
}

double* Parameters_cls::
GetPKeyFrameValueD1(unsigned int index){
    if (index >= GetKeyNumber()) return NULL;
    return mVCKeys[index].mpDValue1;
}

double* Parameters_cls::
GetPKeyFrameValueD2(unsigned int index){
    if (index >= GetKeyNumber()) return NULL;
    return mVCKeys[index].mpDValue2;
}

unsigned int Parameters_cls::
GetInterpTypeFromDouble(double doubleIndex){
    unsigned int type = _PARAM_LINEAR;
    int candid = GetKeyFrameIndexFromDouble(doubleIndex, _PARAM_USE_COMMON,
					    TRUE);
    if (candid != -1) type = mVCKeys[candid].mType;
    return type;
}

void Parameters_cls::
CalcInterpolateStep(double interpolateRate, int index){
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	mpInterpolatedValue[i] = mVCKeys[index].mpValue[i];
}

void Parameters_cls::
CalcInterpolateLinear(double interpolateRate, int index){
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	mpInterpolatedValue[i] = mVCKeys[index].mpValue[i] + 
	    interpolateRate * (mVCKeys[index+1].mpValue[i] - mVCKeys[index].mpValue[i]);
}

void Parameters_cls::
CalcInterpolateHermite(double interpolateRate, int index){
    double scale = mVCKeys[index+1].mDoubleIndex - mVCKeys[index].mDoubleIndex;
    for (unsigned int i = 0; i < mCDimension.mBody; i++){
	double vin, vout;
	
	if (mVCKeys[index].mType == _PARAM_HERMITE_ADJ)
	    vin  = mVCKeys[index].mpDValue1[i] * mVCKeys[index].mpDValue2[0];
	else
	    vin  = mVCKeys[index].mpDValue1[i];
	if (mVCKeys[index+1].mType == _PARAM_HERMITE_ADJ)
	    vout = mVCKeys[index+1].mpDValue1[i] * mVCKeys[index+1].mpDValue2[0];
	else
	    vout = mVCKeys[index+1].mpDValue1[i];

	mpInterpolatedValue[i] = 
	    mVCKeys[index].mpValue[i] + 
	    vin * scale * interpolateRate + 
	    (-3 * mVCKeys[index].mpValue[i] - 2 * vin * scale + 3 * mVCKeys[index+1].mpValue[i] - vout * scale) * interpolateRate * interpolateRate + 
	    (2 * mVCKeys[index].mpValue[i] + vin * scale - 2 * mVCKeys[index+1].mpValue[i] + vout * scale) * interpolateRate * interpolateRate * interpolateRate;
    }
}

double* Parameters_cls::
GetPInterpolatedValue(double doubleIndex, double resolution){
    if (resolution == _PARAM_USE_COMMON) resolution = msResolution;

    if (mVCKeys.size() == 0) {
	for (unsigned int i = 0; i < mCDimension.mBody; i++)
	    mpInterpolatedValue[i] = 0;
	return mpInterpolatedValue;
    }
    if (mCbTimeDependent.mbBody == FALSE ||
	doubleIndex <= mVCKeys[0].mDoubleIndex){
	for (unsigned int i = 0; i < mCDimension.mBody; i++)
	    mpInterpolatedValue[i] = mVCKeys[0].mpValue[i];
	return mpInterpolatedValue;
    }

    if (doubleIndex >= mVCKeys[mVCKeys.size() - 1].mDoubleIndex){
	for (unsigned int i = 0; i < mCDimension.mBody; i++)
	    mpInterpolatedValue[i] = mVCKeys[mVCKeys.size() - 1].mpValue[i];
	return mpInterpolatedValue;
    }
    
    for (unsigned int index = 0; index < mVCKeys.size() - 1; index++){
	if (doubleIndex < mVCKeys[index + 1].mDoubleIndex - resolution / 2){
	    double interpolateRate = 
		(doubleIndex - mVCKeys[index].mDoubleIndex) /
		(mVCKeys[index + 1].mDoubleIndex - mVCKeys[index].mDoubleIndex);
	    switch(mVCKeys[index].mType){
	    case _PARAM_LINEAR:
		CalcInterpolateLinear(interpolateRate, index);
		break;
	    case _PARAM_HERMITE_AUTO:
	    case _PARAM_HERMITE:
	    case _PARAM_HERMITE_ADJ:
		CalcInterpolateHermite(interpolateRate, index);
		break;
	    case _PARAM_STEP:
		CalcInterpolateStep(interpolateRate, index);
		break;
	    default:
		CalcInterpolateLinear(interpolateRate, index);
		break;
	    }
	    return mpInterpolatedValue;
	}
    }
    
    for (unsigned int i = 0; i < mCDimension.mBody; i++)
	mpInterpolatedValue[i] = mVCKeys[mVCKeys.size()-1].mpValue[i];
    return mpInterpolatedValue;
}

double Parameters_cls::
GetPos(double doubleIndex, int axis){
    double retval;
    if (mLabelStr == L"BoxSize" || mLabelStr.Matches(L"ClippingBox"))
	retval = (GetPInterpolatedValue(doubleIndex)[axis] + GetPInterpolatedValue(doubleIndex)[axis + 3]) / 2;
    else 
	retval = GetPInterpolatedValue(doubleIndex)[axis];
    return retval;
}

void Parameters_cls::
MoveInterpolatedPos(double dpos[3], double dscale[3], double drot[4], int hitObj){
    if (mLabelStr == L"BoxSize" || mLabelStr.Matches(L"ClippingBox"))
	MovePosForBox(mpInterpolatedValue, dpos, dscale, hitObj);
    else if (mLabelStr == L"DensityCurvePoints")
	MovePosForTwoPoints(mpInterpolatedValue, dpos, dscale, hitObj);
    else
	LOOP_XYZ mpInterpolatedValue[xyz]   += dpos[xyz];

    if (mLabelStr.Matches(L"CamRotYPR"))
	LOOP_YPRY mpInterpolatedValue[ypry] += drot[ypry];
    
    if (mLabelStr.Matches(L"BoxRotate"))
	LOOP_YPRY mpInterpolatedValue[ypry] += drot[ypry];
    
    if (mLabelStr.Matches(L"PovObject")){
	LOOP_XYZ mpInterpolatedValue[xyz+3] *= dscale[xyz];
	LOOP_YPRY mpInterpolatedValue[ypry+6] += drot[ypry];
    }
    if (mLabelStr.Matches(L"PovDiffObject")){
	LOOP_XYZ mpInterpolatedValue[xyz+3] *= dscale[xyz];
	LOOP_YPRY mpInterpolatedValue[ypry+6] += drot[ypry];
    }

    if (mLabelStr.Matches(L"ClippingBox")){
	LOOP_XYZ if (mpInterpolatedValue[xyz] < 0) mpInterpolatedValue[xyz] = 0;
	LOOP_XYZ if (mpInterpolatedValue[xyz+3] < 0) mpInterpolatedValue[xyz+3] = 0;
	LOOP_XYZ if (mpInterpolatedValue[xyz] > 1) mpInterpolatedValue[xyz] = 1;
	LOOP_XYZ if (mpInterpolatedValue[xyz+3] > 1) mpInterpolatedValue[xyz+3] = 1;
    }
}

bool Parameters_cls::
SelectKeyFrameRegion(double doubleIndexA, double doubleIndexB, bool bSelected){
    bool bRet = FALSE;
    if (doubleIndexB < doubleIndexA){
	double tmp = doubleIndexB;
	doubleIndexB = doubleIndexA;
	doubleIndexA = tmp;}
    size_t keySize = mVCKeys.size();
    for (size_t i = 0; i < keySize; i++)
	if (mVCKeys[i].mDoubleIndex >= doubleIndexA &&
	    mVCKeys[i].mDoubleIndex <= doubleIndexB){
	    mVCKeys[i].mbSelected = bSelected;
	    bRet = TRUE;
	}
    return bRet;
}

int Parameters_cls::
OnWhichKeyFrame(double doubleIndex,  double resolutionByPixel){
    int  selected = -1;
    size_t keySize = mVCKeys.size();
    for (size_t i = 0; i < keySize; i++)
	if (mVCKeys[i].Select(doubleIndex, resolutionByPixel, TRUE) == TRUE)
	    selected = (int)i;
    return selected;
}

int Parameters_cls::
SelectKeyFrame(double doubleIndex,  double resolutionByPixel,
	       bool   bMultiSelect, bool bDeselect){
    int  selected = -1;
    bool bSelectSelected = FALSE;
    
    size_t keySize = mVCKeys.size();
    for (size_t i = 0; i < keySize; i++){
	bool state = mVCKeys[i].mbSelected;
	if (mVCKeys[i].Select(doubleIndex, resolutionByPixel) == TRUE){
	    selected = (int)i;
	    if (state == TRUE) bSelectSelected = TRUE;
	    i = mVCKeys.size();
	}
    }
    
    if (bMultiSelect == TRUE){
	if (bDeselect == FALSE || selected == -1)
	    return selected;
    }

    if (bDeselect == TRUE && bSelectSelected == FALSE)
	for (size_t i = 0; i < keySize; i++)
	    if ((int)i != selected)
		mVCKeys[i].mbSelected = FALSE;
    
    return selected;
}

// ----
void Parameters_cls::
SetMatrixCurrentPos(){
    glTranslated(mpInterpolatedValue[0],
		 mpInterpolatedValue[1],
		 mpInterpolatedValue[2]);
}

void Parameters_cls::
SetMatrixCurrentScale(unsigned int index){
    double scale = mpInterpolatedValue[index];
    double logScale = log10(1 + scale) * 3 + 1;
    glScaled(logScale, logScale, logScale);
}

void Parameters_cls::
DrawGLPath3dv(unsigned int stepNum, double dobWAMin, double dobWAMax,
	      float scale, int interval,
	      bool bSelected, bool drawKeyPos){
    
    if (mCDimension.mBody < 3) return;
    glBegin(GL_POINTS);
    for (double dobIndex = 0; 
	 dobIndex < 1; dobIndex += (double)interval/stepNum){
	if (dobIndex >= dobWAMin && dobIndex <= dobWAMax + 1e-5)
	    if (bSelected == TRUE)
		glColor4fv(mColorSelectedIn);
	    else
		glColor4fv(mColorNotSelectIn);
	else
	    if (bSelected == TRUE)
		glColor4fv(mColorSelectedOut);
	    else
		glColor4fv(mColorNotSelectOut);
	
	glVertex3dv(GetPInterpolatedValue(dobIndex));
    }
    glEnd();

    if (drawKeyPos == FALSE) return;

    for (unsigned int i = 0; i < GetKeyNumber(); i++){
	double dobIndexAtKey = mVCKeys[i].mDoubleIndex;
	if (dobIndexAtKey >= dobWAMin && dobIndexAtKey <= dobWAMax + 1e-5)
	    if (bSelected == TRUE)
		glColor4fv(mColorSelectedIn);
	    else
		glColor4fv(mColorNotSelectIn);
	else
	    if (bSelected == TRUE)
		glColor4fv(mColorSelectedOut);
	    else
		glColor4fv(mColorNotSelectOut);

	if (mVCKeys[i].mbSelected == TRUE && bSelected == TRUE)
	    glColor4f(1.0f, 1.0f, 0.2f, 1.0f);

	glPushMatrix();
	glTranslated(mVCKeys[i].mpValue[0],
		     mVCKeys[i].mpValue[1],
		     mVCKeys[i].mpValue[2]);
	glScalef(scale, scale, scale);
	GLOBJ::Object::Draw(_CAMERA_CROSSHAIR);
	glPopMatrix();
    }
}

void Parameters_cls::
DrawGLTwoGuides(GLGuides_cls *pCGLGuides, double scale, 
		double *boxScale, PositionAndDirection_cls CParentObjPos){
    double scaleInv[3] = {1, 1, 1};
    LOOP_XYZ 
	if (boxScale[xyz] != 0)
	    scaleInv[xyz] = 1.0/boxScale[xyz];

    glEnable(GL_LINE_STIPPLE);
    pCGLGuides->DrawGuide/*Translate*/
	(mpInterpolatedValue, scale, 1, scaleInv, CParentObjPos);

    glColor4f(0.1f, 0.5f, 0.1f, 0.5f);
    glBegin(GL_LINES);
    glVertex3dv(mpInterpolatedValue);
    glVertex3dv(mpInterpolatedValue+3);
    glEnd();

    glDisable(GL_LINE_STIPPLE);

    pCGLGuides->DrawGuide/*Translate*/
	(mpInterpolatedValue+3, scale, 2, scaleInv, CParentObjPos);
}

// -----
void Parameters_cls::
CopyKeyFrameToClipBoard(){

    int itemnum = 0;
    size_t keySize = mVCKeys.size();
    for (size_t i = 0; i < keySize; i++)
	if (mVCKeys[i].mbSelected == TRUE) itemnum++;
    if (itemnum == 0) return;
    
    wxString strInfo;
    strInfo << wxT("Parameter_") << mLabelStr << ENDLWX;
    strInfo << wxT("dimension ") << mCDimension.mBody << ENDLWX;
    strInfo << wxT("itemnum ") << itemnum << ENDLWX;
	    
    for (size_t i = 0; i < keySize; i++)
	if (mVCKeys[i].mbSelected == TRUE){
	    strInfo << mVCKeys[i].Serialize();
	    if (mpCSlaveParam != NULL)
		strInfo << mpCSlaveParam->mVCKeys[i].Serialize();
	}
    
    if (wxTheClipboard->Open()){
	// This data objects are held by the clipboard,
	// so do not delete them in the app.
	wxTheClipboard->SetData( new wxTextDataObject(strInfo) );
	wxTheClipboard->Close();
    }
}

void Parameters_cls::
PasteKeyFrameFromClipBoard(){
    wxStringEX strData;
    wxString strLine;
    if (wxTheClipboard->Open())	{
	if (wxTheClipboard->IsSupported( wxDF_TEXT )){
	    wxTextDataObject data;
	    wxTheClipboard->GetData( data );
	    strData = data.GetText();
	}
	wxTheClipboard->Close();
    }
    
    strData.GetLine(strLine);
    wxString tagStr; tagStr << wxT("Parameter_") << mLabelStr;
    int  loc = strLine.Find(tagStr);
    if (loc == wxNOT_FOUND) return;
    
    //MsgBoxf(NULL, "!!", "Find\n%s", strLine.ToAscii());

    strData.GetLine(strLine);
    wxString strDimension = strLine.c_str() + strlen("dimension ");
    unsigned int dimension = atoi(strDimension.ToAscii());

    strData.GetLine(strLine);
    wxString strItemNum = strLine.c_str() + strlen("itemnum ");
    unsigned int itemnum = atoi(strItemNum.ToAscii());
    
    //MsgBoxf(NULL, "!!", "Find\n%s dim=%d item=%d", strLine.ToAscii(), dimension, itemnum);

    double initDoubleIndex;
    GCUndoList.OpenBlock();
    for (unsigned int i = 0; i < itemnum; i++){
	strData.GetLine(strLine);
	KeyFrame_cls CKey(mCDimension.mBody);
	CKey.ReadStrLine(strLine.c_str());
	if (i == 0) initDoubleIndex = CKey.mDoubleIndex;
	CKey.mDoubleIndex += -initDoubleIndex + msCurrentDoubleIndex;

	if (CKey.mDoubleIndex > 1) continue;

	if (mpCSlaveParam == NULL){
	    //arg2 $B$O(B $B%@%_!<(B
	    SetKey(CKey, CKey);
	}
	else{
	    strData.GetLine(strLine);
	    KeyFrame_cls CKeySlave(mpCSlaveParam->mCDimension.mBody);
	    CKeySlave.ReadStrLine(strLine.c_str());
	    CKeySlave.mDoubleIndex += -initDoubleIndex + msCurrentDoubleIndex;
	    SetKey(CKey, CKeySlave);
	}
	
    }
    GCUndoList.CloseBlock();
    AdjustAutoControl();
}

wxString Parameters_cls::
Serialize(){
    wxString strOut;
    strOut << wxT("Parameter_") << mLabelStr << ENDLWX;
    strOut << mCConfList.SerializeAll();
    strOut << wxT("itemnum ") << (int)mVCKeys.size() << ENDLWX;
    size_t keySize = mVCKeys.size();
    for (size_t i = 0; i < keySize; i++)
	strOut << mVCKeys[i].Serialize();

    return strOut;
}

void Parameters_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    std::wstring tagStr = ((std::wstring)L"Parameter_").append(mLabelStr);
    std::wstring::size_type loc = strLine.find(tagStr);
    if (loc != std::wstring::npos){
	mVCKeys.clear();
	while (TRUE){
	    std::getline(istream, strLine);
	    if (mCConfList.ReadStrLineAll(strLine) == FALSE) break;
	}
	// $B%3%3$O7h$aBG$A(B
#ifdef WIN32
	unsigned int itemnum = 
	    _wtoi(strLine.c_str() + strlen("itemnum "));
#else
	wxString wxItemNum = strLine.c_str() + strlen("itemnum ");
	unsigned int itemnum = 
	    atoi(wxItemNum.ToAscii());
#endif
	for (unsigned int i = 0; i < itemnum; i++){
	    std::getline(istream, strLine);
	    mVCKeys.push_back(KeyFrame_cls(mCDimension.mBody));
	    mVCKeys[mVCKeys.size() - 1].ReadStrLine(strLine);
	}
	AdjustAutoControl();
    }
}

int    Parameters_cls::msFrameNumber = 1;
double Parameters_cls::msResolution = 1.0;
//unsigned int Parameters_cls::msStepQuantum = 1;
double Parameters_cls::msCurrentDoubleIndex = 0;
std::vector < Parameters_cls* > *Parameters_cls::mspVPInstancePrim;

// ------------------------

ComParamBase_cls::
ComParamBase_cls(Parameters_cls* pCTgt)  
{
    mpCTgt = pCTgt;
}

bool ComParamBase_cls::
CheckRepeat(bool bFlag){
    if (bFlag == _COMMAND_UNDO && mpCTgt->mpCSlaveParam != NULL)
	return TRUE;
    if (bFlag == _COMMAND_REDO && mpCTgt->mpCMasterParam != NULL)
	return TRUE;
    return FALSE;
}

ComParamSetKey_cls::
ComParamSetKey_cls(Parameters_cls* pCTgt, double doubleIndex,
		   double *pFrom, double *pDValue1, double *pDValue2,
		   unsigned int type) : 
    ComParamBase_cls(pCTgt)
{
    mDoubleIndex = doubleIndex;
    mType = type;
    mAFrom = new double [pCTgt->mCDimension.mBody];
    mADValue1 = new double [pCTgt->mCDimension.mBody];
    mADValue2 = new double [pCTgt->mCDimension.mBody];
    for (unsigned int i = 0; i < pCTgt->mCDimension.mBody; i++){
	mAFrom[i] = pFrom[i];
	mADValue1[i] = pDValue1[i];
	mADValue2[i] = pDValue2[i];
    }
    if (pCTgt->mLabelStr.Matches(wxT("Color"))){
	mCommandType |= _COMMAND_TYPE_REQUIRE_RETEX;
	mCommandType |= _COMMAND_TYPE_REQUIRE_RECOLOR;
    }
}

ComParamSetKeyC_cls::
ComParamSetKeyC_cls(Parameters_cls* pCTgt, KeyFrame_cls CKey) :
    ComParamBase_cls(pCTgt)
{
    mCKey.FullCopy(CKey);
    if (pCTgt->mLabelStr.Matches(wxT("Color"))){
	mCommandType |= _COMMAND_TYPE_REQUIRE_RETEX;
	mCommandType |= _COMMAND_TYPE_REQUIRE_RECOLOR;
    }
}

ComParamSetKey_cls::
~ComParamSetKey_cls(){
    delete mAFrom;
    delete mADValue1;
    delete mADValue2;
}

bool ComParamSetKey_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandSetKey(mDoubleIndex, mAFrom, mType,
			      mADValue1, mADValue2);
    return CheckRepeat(bFlag);
}

bool ComParamSetKeyC_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandSetKey(mCKey);
    return CheckRepeat(bFlag);
}

ComParamDelKey_cls::
ComParamDelKey_cls(Parameters_cls *pCTgt, int unsigned index) :
    ComParamBase_cls(pCTgt)
{
    mIndex = index;
    if (pCTgt->mLabelStr.Matches(wxT("Color"))){
	mCommandType |= _COMMAND_TYPE_REQUIRE_RETEX;
	mCommandType |= _COMMAND_TYPE_REQUIRE_RECOLOR;
    }
}

bool ComParamDelKey_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandDeleteKey(mIndex);
    return CheckRepeat(bFlag);
}

ComParamOvwtKeyC_cls::
ComParamOvwtKeyC_cls(Parameters_cls* pCTgt, unsigned int index,
		     KeyFrame_cls CKey):
    ComParamBase_cls(pCTgt)
{
    mIndex = index;
    mCKey.FullCopy(CKey);
    if (pCTgt->mLabelStr.Matches(wxT("Color"))){
	mCommandType |= _COMMAND_TYPE_REQUIRE_RETEX;
	mCommandType |= _COMMAND_TYPE_REQUIRE_RECOLOR;
    }
}

bool ComParamOvwtKeyC_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandOvwtKey(mIndex, mCKey);
    return CheckRepeat(bFlag);
}

ComParamMoveKey_cls::
ComParamMoveKey_cls(Parameters_cls* pCTgt, 
		    double doubleIndexFrom, double doubleIndexTo, 
		    double resolution) :
    ComParamBase_cls(pCTgt)
{
    mDoubleIndexFrom = doubleIndexFrom;
    mDoubleIndexTo   = doubleIndexTo;
    mResolution      = resolution;
    if (pCTgt->mLabelStr.Matches(wxT("Color"))){
	mCommandType |= _COMMAND_TYPE_REQUIRE_RETEX;
	mCommandType |= _COMMAND_TYPE_REQUIRE_RECOLOR;
    }
}

bool ComParamMoveKey_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandMoveKey(mDoubleIndexFrom, mDoubleIndexTo, mResolution);
    return CheckRepeat(bFlag);
}

ComParamSetKeyType_cls::
ComParamSetKeyType_cls(Parameters_cls *pCTgt, 
		       unsigned int index, unsigned int type) : 
    ComParamBase_cls(pCTgt)
{
    mIndex = index;
    mType = type;
}

bool ComParamSetKeyType_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandSetKeyType(mIndex, mType);
    return CheckRepeat(bFlag);
}

ComParamSetKeyTypeMulti_cls::
ComParamSetKeyTypeMulti_cls(Parameters_cls *pCTgt,
			    unsigned int listnum,
			    unsigned int *listIndex,
			    unsigned int *listType) :
    ComParamBase_cls(pCTgt)
{
    mListnum = listnum;
    mAIndex = listIndex;
    mAType = listType;
}

ComParamSetKeyTypeMulti_cls::
~ComParamSetKeyTypeMulti_cls(){
    delete mAIndex;
    delete mAType;
}

bool ComParamSetKeyTypeMulti_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandSetKeyTypeMulti(mListnum, mAIndex, mAType);
    return CheckRepeat(bFlag);
}

ComParamClearKey_cls::
ComParamClearKey_cls(Parameters_cls* pCTgt, double *pFrom) :
    ComParamBase_cls(pCTgt)
{
    mAFrom = new double [pCTgt->mCDimension.mBody];
    //mVPDeleteListAtDestruct.push_back(mAFrom);
    for (unsigned int i = 0; i < pCTgt->mCDimension.mBody; i++)
	mAFrom[i] = pFrom[i];
}

ComParamClearKey_cls::
~ComParamClearKey_cls(){
    delete mAFrom;
}

bool ComParamClearKey_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandInitSet(mAFrom);
    return CheckRepeat(bFlag);
}

ComParamSetTimeDependent_cls::
ComParamSetTimeDependent_cls(Parameters_cls* pCTgt, bool bDependent) :
    ComParamBase_cls(pCTgt)
{
    mbDependent = bDependent;
}

bool ComParamSetTimeDependent_cls::
Execute(bool bFlag){
    mpCTgt->ExecCommandTimeDependent(mbDependent);
    return CheckRepeat(bFlag);
}

ComParamMemento_cls::
ComParamMemento_cls(Parameters_cls* pCTgt, Parameters_cls *pCFrom) :
    ComParamBase_cls(pCTgt)
{
    if (pCFrom == NULL){
	mpCParamMemento = new Parameters_cls(pCTgt->mCDimension.mBody);
	mpCParamMemento->CopyFromP(pCTgt);
    }
    else{
	mpCParamMemento = new Parameters_cls(pCFrom->mCDimension.mBody);
	mpCParamMemento->CopyFromP(pCFrom);
    }
    
    if (pCTgt->mLabelStr.Matches(wxT("Color"))){
	mCommandType |= _COMMAND_TYPE_REQUIRE_RETEX;
	mCommandType |= _COMMAND_TYPE_REQUIRE_RECOLOR;
    }
}

ComParamMemento_cls::
~ComParamMemento_cls(){
    delete mpCParamMemento;
}

bool ComParamMemento_cls::
Execute(bool bFlag){
    mpCTgt->CopyFromP(mpCParamMemento);
    return CheckRepeat(bFlag);
}
