// -*- c -*-
// "clsParameters.cpp"

#ifndef _PARAMETERS_H
#define _PARAMETERS_H

#include "clsKeyFrame.h"
#include "../CONFITEM/clsConfItem.h"
#include "../MYGL/clsPositionAndDirection.h"

#include <vector>
#include <list>
#include <string>

#define _PARAMGL_SELECTED_IN   0
#define _PARAMGL_SELECTED_OUT  1
#define _PARAMGL_NOTSELECT_IN  2
#define _PARAMGL_NOTSELECT_OUT 3

#define _PARAM_USE_COMMON 0

#define _PARAM_CURRENT_DINDEX -1
#define _PARAM_PROCESS_OVRW 2

class GLGuides_cls;

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

class ParamGL_cls {
 public:
    ParamGL_cls();
    
    float mColorSelectedIn[4];
    float mColorSelectedOut[4];
    float mColorNotSelectIn[4];
    float mColorNotSelectOut[4];
    float *GetPColor(bool bSelected, bool bInWA);

    void SetColor(int tag, float r, float g, float b, float a);
};

class Parameters_cls : public ParamGL_cls {
    friend class TimeLinePanel_cls; 
    friend class ComParamBase_cls; 
 public:
    Parameters_cls(unsigned int dimension = 0, std::wstring labelStr = L"Untitled");
    // new $B$H(B std::vector $B$NAH$_9g$o$;$O%3%T!<%3%s%9%H%i%/%?I,?\(B
    Parameters_cls(const Parameters_cls &obj); 
    void CopyFromP(const Parameters_cls *pObj);
    void DefineConfList();
    ~Parameters_cls();
    
    // $B;~4VJT=8$K1F6A$r<u$1$k%$%s%9%?%s%9%j%9%H(B
    static std::vector < Parameters_cls* > *mspVPInstancePrim;

 public:
    ConfItemList_cls mCConfList;
    ConfItemUINT_cls mCDimension;
    ConfItemBOOL_cls mCbTimeDependent;
    ConfItemBOOL_cls mCbVisible;
    ConfItemBOOL_cls mCbShow;
    ConfItemBOOL_cls mCbSelectable;

    wxString mLabelStr;
    // $B?tCMF~NO%@%$%d%m%0$G$N%"%$%F%`$N%i%Y%k(B
    std::vector < wxString > mVstrItemLabel;
    std::vector < wxString > mVstrItemLabelD1;
    
    bool mbUpdateCamera; // $B99?7$N:]$K%+%a%i$b%"%C%W%G!<%H(B
    
 public: // protected:
    Parameters_cls *mpCMasterParam;
    Parameters_cls *mpCSlaveParam; // Delete$B$H(BInsert$B$rO"F0$5$;$k(B
    
 protected:
    static int             msFrameNumber;
    static double          msResolution; // = 1.0/mStepNumber;
    //static unsigned int    msStepQuantum;
    static double          msCurrentDoubleIndex;
    
 public:
    double *mpInterpolatedValue;

    void SetMasterParam(Parameters_cls *pCMasterParam);
    void SetSlaveParam(Parameters_cls *pCSlaveParam);

    void RemapKeyFrame(int oldMax, int newMax);
    static void RemapKeyFrameAll(int oldMax, int newMax);
    static void SetFrameNumber(int frame, unsigned int quantum);
    static void SetCurrentDoubleIndex(double doubleIndex);
    double Quantize(double doubleIndex, 
		    unsigned int frameNumber = _PARAM_USE_COMMON);

    void SetTimeDependent(bool bTimeDependent, bool bSlave = FALSE);
    void CopyParameterFromP(Parameters_cls *pCParamFrom, bool bSlave = FALSE);
    void ExecCommandTimeDependent(bool bTimeDependent);
    std::vector < KeyFrame_cls > mVCKeys;
    unsigned int GetKeyNumber();
    void SetKeyToValueAt(KeyFrame_cls *pCKey, double doubleIndex);
    void SetSnippingTag(std::vector < bool > &VBSnipping, int maxFrame);
        
    void SortKey();
    void CalcVelocity(int prev, int next, double *pArg);
    void AdjustAutoControl();

    void InitSet(double *pFrom);
    void ExecCommandInitSet(double *pFrom);
    void InsertKey(double doubleIndex, double *pValue = NULL);
    
    int  CheckOverWrite(double doubleIndex, double Resolution);
    void SetKey(KeyFrame_cls CKey, KeyFrame_cls CKeySlave = KeyFrame_cls(0),
		double resolution = _PARAM_USE_COMMON);
    void ExecCommandSetKey(KeyFrame_cls CKey);
    void ExecCommandSetKey(double doubleIndex, double *pFrom, unsigned int type, 
			   double *pDValue1, double *pDValue2);
    void ExecCommandOvwtKey(unsigned int index, KeyFrame_cls CKey);
    bool SetKeyIfChanged(double doubleIndex = _PARAM_CURRENT_DINDEX,
			 bool bForce = FALSE);
    bool MoveTimeSelectedKey(double doubleIndexDiff);
    static void MovePosForBox(double pPos[6], double dpos[3], double dscale[3], int hitObj);
    void MovePosForTwoPoints(double pPos[6], double dpos[3], double dscale[3], int hitObj);
    bool MovePosSelectedKey(double dpos[3], double dscale[3], int hitObj);
    void RemapKeys(double rate);
    int  MoveKey(int indexToMove, double doubleIndexDiff,
		 unsigned int frameNumber = _PARAM_USE_COMMON,
		 bool bSlave = FALSE);
    void ExecCommandMoveKey(double doubleIndexFrom, double doubleIndexTo,
			    double resolution);
    void SetKeyType(unsigned int index, unsigned int type);
    void ExecCommandSetKeyType(unsigned int index, unsigned int type);
    void SetKeyTypeMulti(unsigned int mode);
    void ExecCommandSetKeyTypeMulti(unsigned int num,
				    unsigned int *pIndex,
				    unsigned int *pType);
    
    bool DeleteSelectedKey();
    bool DeleteKey(double doubleIndex, 
		   double resolution = _PARAM_USE_COMMON);
    void ExecCommandDeleteKey(int index);
    bool IsKeyFrame(double doubleIndex, double resolution= _PARAM_USE_COMMON);
    
    double GetNextKeyPosDoubleIndex(double cur, 
				    double resolution = _PARAM_USE_COMMON);
    double GetPrevKeyPosDoubleIndex(double cur, 
				    double resolution = _PARAM_USE_COMMON);

    // flag$B$,(BFALSE$B$GBP1~$9$k(BKey$B$,L5$1$l$P(B -1
    int GetKeyFrameIndexFromDouble(double doubleIndex,
				   double resolution = _PARAM_USE_COMMON,
				   bool bFlagGetKeyEvenNotKeyFrame = FALSE);
    int GetKeyFrameIndexToInsert(double doubleIndex, 
				 double resolution = _PARAM_USE_COMMON);   
    int GetAdjacentKey(double doubleIndex, int direction);
    int GetSelectedKeyNum();
    int GetSelectedKeyIndex();

    double GetKeyFrameDoubleIndex(unsigned int index);
    double* GetPKeyFrameValue(unsigned int index);
    double* GetPKeyFrameValueD1(unsigned int index);
    double* GetPKeyFrameValueD2(unsigned int index);

    unsigned int GetInterpTypeFromDouble(double doubleIndex);

    void    CalcInterpolateStep(double interpolateRate, int index);
    void    CalcInterpolateLinear(double interpolateRate, int index);
    void    CalcInterpolateHermite(double interpolateRate, int index);
    double* GetPInterpolatedValue(double doubleIndex, 
				  double resolution = _PARAM_USE_COMMON);
    double  GetPos(double doubleIndex, int axis);
    void    MoveInterpolatedPos(double dpos[3], double dscale[3], double drot[4], int hitObj);
    void    SetInterpolatedPos(double dpos[3]);
    
    bool    SelectKeyFrameRegion(double doubleIndexA, double doubleIndexB, 
				 bool bSelect = TRUE);
    int     OnWhichKeyFrame(double doubleIndex, double resolutionByPixel);
    int     SelectKeyFrame(double doubleIndex, double resolutionByPixel,
			   bool bMultiSelect = FALSE,
			   bool bDeselect = FALSE);
  
    //--
    void    SetMatrixCurrentPos();
    void    SetMatrixCurrentScale(unsigned int index);
    void    DrawGLPath3dv(unsigned int stepNum,
			  double dobWAMin, double dobWAMax,
			  float scale, int interval = 1,
			  bool bSelected = TRUE, 
			  bool drawKeyPos = TRUE);
    void    DrawGLTwoGuides(GLGuides_cls *pCGLGuides, double scale,
			    double *boxScale,
			    PositionAndDirection_cls CParentObjPos);
    
    //--
    void CopyKeyFrameToClipBoard();
    void PasteKeyFrameFromClipBoard();
    
    wxString Serialize();
    void ReadStrLine(std::wistream &istream, std::wstring strLine);
};

#endif
