// -*- c -*-
// "clsParametersPanel.h"
// (setq c-basic-offset 4)

#include "clsParametersPanel.h"
#include "clsParameters.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "clsTimeLinePanel.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../../clsTimeCtrl.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include <wx/artprov.h>

#ifdef WIN32
#define _MEM_PANEL_OFFSET 14
#define _CHECKBOX_WIDTH 13
#else
#define _MEM_PANEL_OFFSET 10
#define _CHECKBOX_WIDTH 17
#endif


BEGIN_EVENT_TABLE(ParametersPanel_cls, wxPanel)
     EVT_PAINT       (ParametersPanel_cls::OnPaint)
     EVT_KEY_DOWN    (ParametersPanel_cls::OnKeyDown)
END_EVENT_TABLE()

ParametersPanel_cls::
     ParametersPanel_cls(Parameters_cls *pCParam, wxWindow *parent, /*wxString strTitle,*/ wxString strCustomName, bool bShow, wxPoint pos, wxSize size, wxWindowID id) :
    wxPanel(parent, id, pos, size),
    mpCParam(pCParam)
{
    //SetBackgroundColour(wxColour(216, 0, 216));
    //mstrSubTitle = L"";
    mOffset = _MEM_PANEL_OFFSET;

    if (mbInitialized == FALSE){
	msCBitmapTimeDependent = wxBitmap(wxT("ICONS/timeP.png"),wxBITMAP_TYPE_PNG);
	msCBitmapConstant      = wxBitmap(wxT("ICONS/timeN.png"),wxBITMAP_TYPE_PNG);
	msCBitmapTimeDisable   = wxBitmap(wxT("ICONS/timeD.png"),wxBITMAP_TYPE_PNG);
	msCBitmapVisible   = wxBitmap(wxT("ICONS/eyeP.png"),wxBITMAP_TYPE_PNG);
	msCBitmapInvisible = wxBitmap(wxT("ICONS/eyeN.png"),wxBITMAP_TYPE_PNG);
#ifndef WIN32
	msCBitmapLeft  = wxBitmap(wxT("ICONS/left.png"), wxBITMAP_TYPE_PNG);
	msCBitmapLeftDisabled  = wxBitmap(wxT("ICONS/leftDisabled.png"), wxBITMAP_TYPE_PNG);
	msCBitmapRight = wxBitmap(wxT("ICONS/right.png"), wxBITMAP_TYPE_PNG);
	msCBitmapRightDisabled = wxBitmap(wxT("ICONS/rightDisabled.png"), wxBITMAP_TYPE_PNG);
#endif
	mbInitialized = TRUE;
    }
    
    // TCU[
    wxBoxSizer  *pBoxSizerSrc = new wxBoxSizer(wxHORIZONTAL);
    
    mpCBPanelTime = new BitmapPanel_cls(this, wxID_ANY, wxDefaultPosition, wxSize(14,16));
    mpCBPanelTime->AddBitmapP(&msCBitmapConstant);
    mpCBPanelTime->AddBitmapP(&msCBitmapTimeDependent);
    mpCBPanelTime->SetBitmapPDisable(&msCBitmapTimeDisable);
    Connect(mpCBPanelTime->GetId(), wxEVT_LEFT_UP,
	    wxMouseEventHandler(ParametersPanel_cls::OnMouseL));
    Connect(mpCBPanelTime->GetId(), wxEVT_RIGHT_UP,
	    wxMouseEventHandler(ParametersPanel_cls::OnMouseR));
    
#ifdef WIN32
    mpButtonLFT   = new wxBitmapButtonEX
      (this, wxID_ANY, wxArtProvider::GetBitmap(wxT("wxART_LEFT")),
       wxDefaultPosition, wxSize(14, 16), wxBORDER_NONE);
    mpButtonLFT->SetBitmapHover
	(wxArtProvider::GetBitmap(wxT("wxART_LEFT_HOVER")));
    mpButtonLFT->SetBitmapDisabled
	(wxArtProvider::GetBitmap(wxT("wxART_LEFT_DISABLED")));

    
    mpButtonRGT   = new wxBitmapButtonEX
	(this, wxID_ANY, wxArtProvider::GetBitmap(wxT("wxART_RIGHT")),
	 wxDefaultPosition, wxSize(14, 16), 0);
    mpButtonRGT->SetBitmapHover
	(wxArtProvider::GetBitmap(wxT("wxART_RIGHT_HOVER")));
    mpButtonRGT->SetBitmapDisabled
	(wxArtProvider::GetBitmap(wxT("wxART_RIGHT_DISABLED")));
    
    Connect(mpButtonLFT->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
	    wxCommandEventHandler(ParametersPanel_cls::OnButtonKeyLFT));
    Connect(mpButtonRGT->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
	    wxCommandEventHandler(ParametersPanel_cls::OnButtonKeyRGT));
#else
    mpButtonLFT = new BitmapPanel_cls(this, wxID_ANY, wxDefaultPosition, wxSize(14,16));
    mpButtonLFT->AddBitmapP(&msCBitmapLeft);
    mpButtonLFT->SetBitmapPDisable(&msCBitmapLeftDisabled);
    mpButtonRGT = new BitmapPanel_cls(this, wxID_ANY, wxDefaultPosition, wxSize(14,16));
    mpButtonRGT->AddBitmapP(&msCBitmapRight);
    mpButtonRGT->SetBitmapPDisable(&msCBitmapRightDisabled);

    Connect(mpButtonLFT->GetId(), wxEVT_LEFT_UP,
	    wxCommandEventHandler(ParametersPanel_cls::OnButtonKeyLFT));
    Connect(mpButtonRGT->GetId(), wxEVT_LEFT_UP,
	    wxCommandEventHandler(ParametersPanel_cls::OnButtonKeyRGT));
    
#endif
    //mpButtonLFT->SetBackgroundColour(wxColour(216, 216, 216));
    //mpButtonRGT->SetBackgroundColour(wxColour(216, 216, 216));
    
    mpButtonLFT->Connect(wxEVT_KEY_DOWN,
			 wxKeyEventHandler(ParametersPanel_cls::OnKeyDown));
    mpButtonRGT->Connect(wxEVT_KEY_DOWN,
			 wxKeyEventHandler(ParametersPanel_cls::OnKeyDown));
    
    mpCheckBoxKey = new wxCheckBox(this, wxID_ANY, wxT(""),
				   wxDefaultPosition, wxSize(_CHECKBOX_WIDTH,14));
    Connect(mpCheckBoxKey->GetId(), wxEVT_COMMAND_CHECKBOX_CLICKED,
	    wxCommandEventHandler(ParametersPanel_cls::OnCheckBoxKey));
    

    mpCBPanelVisible = new BitmapPanel_cls(this, wxID_ANY, wxDefaultPosition, wxSize(14,16));
    mpCBPanelVisible->AddBitmapP(&msCBitmapInvisible);
    mpCBPanelVisible->AddBitmapP(&msCBitmapVisible);
    Connect(mpCBPanelVisible->GetId(), wxEVT_LEFT_UP,
	    wxMouseEventHandler(ParametersPanel_cls::OnMouseL));
		  
    mpCTimeLine = new TimeLinePanel_cls(mpCParam, this, /*strTitle,*/ strCustomName,
					wxDefaultPosition, wxSize(-1, 16));
    Connect(mpCTimeLine->GetId(), wxEVT_LEFT_DOWN,
	    wxMouseEventHandler(ParametersPanel_cls::OnMouseL));
    Connect(mpCTimeLine->GetId(), wxEVT_RIGHT_DOWN,
	    wxMouseEventHandler(ParametersPanel_cls::OnMouseR));

    //pBoxSizerSrc->Add(mpButtonHide, 0, 0);
    pBoxSizerSrc->Add(mpCBPanelTime, 0, 0);
    pBoxSizerSrc->Add(mpButtonLFT  , 0, 0);
    pBoxSizerSrc->Add(mpCheckBoxKey, 0, wxALIGN_CENTER_VERTICAL);
    pBoxSizerSrc->Add(mpButtonRGT  , 0, 0);
    pBoxSizerSrc->Add(mpCBPanelVisible, 0, 0);
    pBoxSizerSrc->Add(mpCTimeLine  , 1, wxEXPAND);
    
    SetSizer(pBoxSizerSrc);
    
    UpdatePanelBitmapState();
    Show(bShow);
}

void ParametersPanel_cls::
LinkParameter(Parameters_cls *pCParam){
    mpCParam = pCParam;
    mpCTimeLine->LinkParameter(pCParam);
    UpdatePanelBitmapState();
    Refresh(TRUE);
}

void ParametersPanel_cls::
UpdatePanelBitmapState(){
    if (mpCParam != NULL) {
	mpCBPanelTime->mState = 
	    (unsigned int)mpCParam->mCbTimeDependent.mbBody;
	mpCBPanelVisible->mState = 
	    (unsigned int)mpCParam->mCbVisible.mbBody;
    }
    else {
	mpCBPanelTime->mState = 0;
	mpCBPanelVisible->mState = 0;
    }
}

void ParametersPanel_cls::
SetDoubleIndex(int maxFrame, int AAMin, int AAMax, double doubleIndex,
	       double doubleIndexAAMin, double doubleIndexAAMax,
	       double doubleIndexWAMin, double doubleIndexWAMax){
    msMaxFrame = maxFrame;
    msAAMin = AAMin;
    msAAMax = AAMax;
    msCurrentDoubleIndex = doubleIndex;
    msDoubleIndexAAMin   = doubleIndexAAMin;
    msDoubleIndexAAMax   = doubleIndexAAMax;
    msDoubleIndexWAMin   = doubleIndexWAMin;
    msDoubleIndexWAMax   = doubleIndexWAMax;
    TimeLinePanel_cls::SetDoubleIndex
	(maxFrame, AAMin, AAMax, doubleIndex, 
	 doubleIndexAAMin, doubleIndexAAMax,
	 doubleIndexWAMin, doubleIndexWAMax);
}

void ParametersPanel_cls::
HidePanel(){
    mspCDF3SeqGUILink->HidePanel();
    mspCDF3SeqGUILink->UpdateItemsVisivility();
    /*
    mspCTimeLink->UpdateDF3SeqGUI(TRUE);
    wxCommandEvent dummy2(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    wxGetTopLevelParent(this)->AddPendingEvent(dummy2);
    */
}

void ParametersPanel_cls::
OnButtonKeyLFT(wxCommandEvent& event){
    if (mpCParam == NULL) return;
    mspCTimeLink->SetIndexFromDouble
	(mpCParam->GetPrevKeyPosDoubleIndex(msCurrentDoubleIndex));
    mspCTimeLink->UpdateDF3SeqGUI(TRUE);
    event.Skip();
}

void ParametersPanel_cls::
OnButtonKeyRGT(wxCommandEvent& event){
    if (mpCParam == NULL) return;
    mspCTimeLink->SetIndexFromDouble
	(mpCParam->GetNextKeyPosDoubleIndex(msCurrentDoubleIndex));
    mspCTimeLink->UpdateDF3SeqGUI(TRUE);
    event.Skip();
}

void ParametersPanel_cls::
OnCheckBoxKey(wxCommandEvent& event){
    
    if (mpCParam == NULL) return;
    if (mpCheckBoxKey->GetValue() == FALSE){
	mspCDF3SeqGUILink->DeleteKeyFrame(msCurrentDoubleIndex);
    }
    else{
	mspCDF3SeqGUILink->SelectKeyFrame(msCurrentDoubleIndex, 1e-12);
    }
    
    wxMouseEvent com(wxEVT_LEFT_DOWN);
    com.SetId(mpCTimeLine->GetId());
    com.m_shiftDown = TRUE;
    GetParent()->ProcessEvent(com);
    
    mspCTimeLink->UpdateDF3SeqGUI(TRUE);
    event.Skip();
}

void ParametersPanel_cls::
EmuCheckBox(){
    if (mpCParam == NULL) return;
    mpCheckBoxKey->SetValue(!mpCheckBoxKey->GetValue());
    wxCommandEvent 
	event(wxEVT_COMMAND_CHECKBOX_CLICKED, mpCheckBoxKey->GetId());
#ifdef WIN32
    AddPendingEvent(event);
#else
    OnCheckBoxKey(event);
#endif
}

void ParametersPanel_cls::
UpdateDisplay(){
    if (mpCTimeLine != NULL)
	mpCTimeLine->Refresh(FALSE);
    if (mpCBPanelTime != NULL)
	mpCBPanelTime->Refresh(FALSE);
    if (mpCBPanelVisible != NULL)
	mpCBPanelVisible->Refresh(FALSE);
    if (mpCParam == NULL) {
	mpCheckBoxKey->Disable();
	mpButtonLFT->Disable();
	mpButtonRGT->Disable();
	mpCBPanelTime->Disable();
	mpCBPanelVisible->Disable();
	return;
    } 
    else{
	if (mpCParam->mCbTimeDependent.mbBody == TRUE){
	    mpCheckBoxKey->Enable();
	    mpButtonLFT->Enable();
	    mpButtonRGT->Enable();
	    mpCBPanelTime->Enable();
	    mpCBPanelVisible->Enable();
	}
	else{
	    mpCheckBoxKey->Disable();
	    mpButtonLFT->Disable();
	    mpButtonRGT->Disable();
	    mpCBPanelTime->Enable();
	    mpCBPanelVisible->Enable();
	}
    }
    mpCheckBoxKey->SetValue(IsKeyFrame(msCurrentDoubleIndex));
}

bool ParametersPanel_cls::
IsKeyFrame(double doubleIndex){
    return mpCParam->IsKeyFrame(doubleIndex);
}

void ParametersPanel_cls::
SetMemberSize(wxSize size){
    mWidth = size.GetWidth();
    mHeight = size.GetHeight();
}

int ParametersPanel_cls::
GetDeviceX(float x){
    return (int)(x * (mWidth - 2 * mOffset) + mOffset);
}

void ParametersPanel_cls::
OnPaint(wxPaintEvent& event){
    // must always be here
    wxPaintDC dc(this);
    SetMemberSize(dc.GetSize());
    UpdateDisplay();
    event.Skip();
}

void ParametersPanel_cls::
OnKeyDown(wxKeyEvent& event){
    int code = event.GetKeyCode();
    GetParent()->AddPendingEvent(event);
}

void ParametersPanel_cls::
OnMouseL(wxMouseEvent& event){
    if (event.GetId() == mpCBPanelTime->GetId()){
	if (mpCTimeLine->mbSelected == FALSE){
	    if (mpCParam == NULL) return;
	    mpCParam->SetTimeDependent(mpCBPanelTime->mState == 1);
	}
	else
	    mspCDF3SeqGUILink->SetTimeDependent(mpCBPanelTime->mState == 1);
	wxCommandEvent dummy(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	GetGrandParent()->AddPendingEvent(dummy);
    }

    if (event.GetId() == mpCBPanelVisible->GetId()){
	if (mpCTimeLine->mbSelected == FALSE){
	    if (mpCParam == NULL) return;
	    mpCParam->mCbVisible.mbBody = (mpCBPanelVisible->mState == 1);
	}
	else
	    mspCDF3SeqGUILink->SetVisibility(mpCBPanelVisible->mState == 1);
	wxCommandEvent dummy(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	GetGrandParent()->AddPendingEvent(dummy);
    }

    GetParent()->AddPendingEvent(event);
    event.Skip();
}

void ParametersPanel_cls::
OnMouseR(wxMouseEvent& event){
    GetParent()->AddPendingEvent(event);
    event.Skip();
}

DF3SeqGUI_cls* ParametersPanel_cls::mspCDF3SeqGUILink;
TimeCtrl_cls* ParametersPanel_cls::mspCTimeLink;

double ParametersPanel_cls::msCurrentDoubleIndex;
double ParametersPanel_cls::msDoubleIndexAAMin;
double ParametersPanel_cls::msDoubleIndexAAMax;
double ParametersPanel_cls::msDoubleIndexWAMin;
double ParametersPanel_cls::msDoubleIndexWAMax;
int    ParametersPanel_cls::msMaxFrame;
int    ParametersPanel_cls::msAAMin;
int    ParametersPanel_cls::msAAMax;

wxBitmap ParametersPanel_cls::msCBitmapTimeDependent;
wxBitmap ParametersPanel_cls::msCBitmapConstant;
wxBitmap ParametersPanel_cls::msCBitmapTimeDisable;
wxBitmap ParametersPanel_cls::msCBitmapVisible;
wxBitmap ParametersPanel_cls::msCBitmapInvisible;
#ifndef WIN32
wxBitmap ParametersPanel_cls::msCBitmapLeft;
wxBitmap ParametersPanel_cls::msCBitmapLeftDisabled;
wxBitmap ParametersPanel_cls::msCBitmapRight;
wxBitmap ParametersPanel_cls::msCBitmapRightDisabled;
#endif
bool    ParametersPanel_cls::mbInitialized = FALSE;
