// "clsMarchingCube.cpp"
// (setq c-basic-offset 4)

#ifndef _ISOSURFICE_MARCHINGCUBE_H_
#define _ISOSURFICE_MARCHINGCUBE_H_

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif


#include <vector>
#include <wx/thread.h>

// The main routine of this class used
// Marching Cubes Example Program 
// by Cory Bloyd (corysama@yahoo.com)
// which is publicshd as public domain 

class MarchingCube_cls;

/*
struct GLvector{
    GLfloat fX;
    GLfloat fY;
    GLfloat fZ;     
};
*/

/*
class VerticesVector_cls{
 public:
    std::vector < float > mVVertices;
    std::vector < float > mVNormals;
};
*/

class ThreadMarching : public wxThread {
 public:
    ThreadMarching(unsigned char *pData, 
		   float *pTmpVertices, float *pTmpNormal,
		   int *pArrayIndex);
    void SetParameters(/*float *pTmpVertices, float *pTmpNormal,
		       int *pArrayIndex,
		       unsigned char *pData,*/
		       unsigned int size[3], int initZ, int finZ,
		       float targetValue,
		       bool bUseSSE);

    GLfloat GetOffset(GLfloat fValue1, GLfloat fValue2, GLfloat fValueDesired);
    inline float GetInterpolatedValue
	(const unsigned char *pData,
	 const int lBaseIndex, const float fx, const float fy, const float fz);
    void GetNormalPrecise
	(GLfloat ANorm[3], const unsigned char *pData,
	 const int x, const GLfloat fX, 
	 const int y, const GLfloat fY,
	 const int z, const GLfloat fZ);
    void GetNormalPreciseSSE
	(GLfloat ANorm[3], const unsigned char *pData,
	 const int x, const GLfloat fX, 
	 const int y, const GLfloat fY,
	 const int z, const GLfloat fZ);

    int MarchCubeMainThread(const unsigned char *pData,
			    const int x, const int y, const int z);
    
    
    void MarchCubeLOOPX1(const unsigned char *pData, int &currStateY,
			 int y, int z);
    void MarchCubeLOOPX2(const unsigned char *pData, 
			 int y, int z);
    void MarchCubeLOOPY(const unsigned char *pData,
			int &currStateZ, int z);
   
    virtual ExitCode Entry();

    unsigned char * const mpData;
    float * const mpTmpVertices;
    float * const mpTmpNormal;
    int   * const mpArrayIndex;
    
    int mSize[3];
    int mInitZ, mFinZ;

    int mOffsetZ, mOffsetY;
    int mBaseIndex;
    float mTargetValue, mTargetValue255;
    bool mbUseSSE;

    //GLfloat afCubeValue[8];
    //GLfloat AAEdgeVertex[12][3];
    //GLfloat AAEdgeNorm[12][4]; // SSÊ߂4߂̓_~[

    static const GLint   VOffset[8][3];
    static const GLint   a2iEdgeConnection[12][2];
    static const GLfloat a2fEdgeDirection[12][3];

    static const GLint aiCubeEdgeFlags[256];
    static const GLint a2iTriangleConnectionTable[256][16];
};

class MarchingCube_cls{
 public:

    std::vector < std::vector < float > > mVV_Vertices;
    std::vector < std::vector < float > > mVV_Normals;

    //int mSizeX, mSizeY, mSizeZ;
    unsigned int mSize[3];
    double mScalePrev[3];
    
    float mTargetValue, mTargetValue255;
    
    void Clear(int sizeZ);
    bool ThreadControlMain(unsigned char *pData, int zLayerBase,
			   bool bUseSSE, int threadNum, const int layerFactor);
    void ConstructVertices
	(unsigned char *pData, int sizeX, int sizeY, int sizeZ,
	 bool bUseSSE, int threadNum, const int layerFactor);
    void GLDraw(double scale[3], bool bCullFace, GLuint showType);
}; 

#endif

