// "clsMarchingCube.h"
// (setq c-basic-offset 4)

#include "clsIsosurfOptimize.h"
#include "clsMarchingCube.h"
#include "clsMarchMainThread.h"
#include <math.h>

#include "../../MYAPI/MsgBox.h"
#include "../../../wxWidgetString.h"

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

extern bool GbGLLockForMultithread;
#define X 0
#define Y 1
#define Z 2

#define ENDLWX wxT("\n")

// ----

ThreadExportMesh2PovSub::
ThreadExportMesh2PovSub(wxString *pstrOut, unsigned int indexSize,
			unsigned int polygonNumber,
			std::vector < std::vector < unsigned int > > *pVVVertexIndicesOptFull,
			std::vector < unsigned int > *pVvertexOffset) 
: wxThread(wxTHREAD_JOINABLE),
    mpStrOut(pstrOut), mIndexSize(indexSize), mPolygonNumber(polygonNumber), mpVVVertexIndicesOptFull(pVVVertexIndicesOptFull),
    mpVvertexOffset(pVvertexOffset)
{
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadExportMesh2PovSub::
Entry(){

    wxString *pstrOut = mpStrOut;
    pstrOut->Alloc(2 * mPolygonNumber * 40 + 1000);
    
    *pstrOut << wxT("\tface_indices{") << ENDLWX
	     << wxT("\t\t") << mPolygonNumber << wxT (",") << ENDLWX;

    std::vector < std::vector < unsigned int > > 
	*pVVVertexIndicesOptFull = mpVVVertexIndicesOptFull;
    std::vector < unsigned int >
	*pVvertexOffset = mpVvertexOffset;

    for (unsigned int index = 0; index < mIndexSize; index++){
	size_t vertexIndexSize = (*mpVVVertexIndicesOptFull)[index].size();
	if (vertexIndexSize == 0) continue;
	unsigned int *pIndex = &((*pVVVertexIndicesOptFull)[index][0]);
	unsigned int indexOffset = (*pVvertexOffset)[index];
	for (size_t v = 0; v < vertexIndexSize; v+= 3){
	    *pstrOut << 
		wxString::Format(wxT("\t\t<%d,%d,%d>,"),
				 *(pIndex + v + X) + indexOffset,
				 *(pIndex + v + Y) + indexOffset,
				 *(pIndex + v + Z) + indexOffset) << ENDLWX;
	}
    }

    pstrOut->RemoveLast(); pstrOut->RemoveLast();
    
    // Ō͂ł͕Ȃ
    
    return 0;
}

// ----

MarchingCube_cls::
MarchingCube_cls(){
    mbMemoryReadyOpt = FALSE;
    mbMemoryReadyOptFull = FALSE;
}

MarchingCube_cls::
~MarchingCube_cls(){
    for (unsigned int i = 0; i < mVpVertices.size(); i++){
	if (mVAllocSize[i] != 0){
	    free(mVpVertices[i]);
	    free(mVpNormals[i]);
	    mVAllocSize[i] = 0;
	}
    }
}

void MarchingCube_cls::
Clear(int sizeZ){
    for (unsigned int i = 0; i < mVpVertices.size(); i++){
	if (mVAllocSize[i] != 0){
	    free(mVpVertices[i]);
	    free(mVpNormals[i]);
	    mVAllocSize[i] = 0;
	    mVAllocSizeOrg[i] = 0;
	}
    }

    mVpVertices.clear();
    mVpNormals.clear();
    mVAllocSize.clear();
    mVAllocSizeOrg.clear();

    // ̐܂K
    mVpVertices.resize(2 * sizeZ);
    mVpNormals.resize(2 * sizeZ);
    mVAllocSize.resize(2 * sizeZ);
    mVAllocSizeOrg.resize(2 * sizeZ);
    mVbOptimized.resize(2 * sizeZ);
    mVbOptimizedFull.resize(2 * sizeZ);
    mVOptVertexNumOrg.resize(2 * sizeZ);

    for (unsigned int i = 0; i < mVbOptimized.size(); i++){
	mVbOptimized[i] = FALSE;
	mVOptVertexNumOrg[i] = 0;
	mVbOptimizedFull[i] = FALSE;
    }

    mbMemoryReadyOpt = FALSE;
    mbMemoryReadyOptFull = FALSE;
}

void MarchingCube_cls::
ClearOpt(){
    for (unsigned int i = 0; i < mVVVerticesOpt.size(); i++){
	// ̊mۗ̈悲Clear@(TCY0̈ꎞxNgƌ)
	std::vector<float>().swap(mVVVerticesOpt[i]);
	std::vector<float>().swap(mVVNormalsOpt[i]);
	std::vector<float>().swap(mVVUV1DOpt[i]);
	std::vector<unsigned int>().swap(mVVVertexIndicesOpt[i]);
    }

    mVVVerticesOpt.resize(mVpVertices.size());
    mVVNormalsOpt.resize(mVpVertices.size());
    mVVUV1DOpt.resize(mVpVertices.size());
    mVVVertexIndicesOpt.resize(mVpVertices.size());
    mbMemoryReadyOpt = TRUE;
}

void MarchingCube_cls::
ClearOptFull(){
    for (unsigned int i = 0; i < mVVVerticesOptFull.size(); i++){
	std::vector<float>().swap(mVVVerticesOptFull[i]);
	std::vector<float>().swap(mVVNormalsOptFull[i]);
	std::vector<float>().swap(mVVUV1DOptFull[i]); 
	std::vector<unsigned int>().swap(mVVVertexIndicesOptFull[i]);
    }
    
    mVVVerticesOptFull.resize(mVpVertices.size());
    mVVNormalsOptFull.resize(mVpVertices.size());
    mVVUV1DOptFull.resize(mVpVertices.size());
    mVVVertexIndicesOptFull.resize(mVpVertices.size());
    mbMemoryReadyOptFull = TRUE;
}

unsigned int MarchingCube_cls::
GetVertexNumber(){
    unsigned int number = 0;
    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	number += mVAllocSizeOrg[i];
    return number / 3;
}

unsigned int MarchingCube_cls::
GetVertexNumberOpt(){
    unsigned int number = 0;
    for (unsigned int i = 0; i < mVVVerticesOpt.size(); i++)
	number += mVOptVertexNumOrg[i];
    return number / 3;
}

int MarchingCube_cls::
GetVertexNumberOptFull(){
    int number = 0;
    for (unsigned int i = 0; i < mVVVerticesOpt.size(); i++)
	number += (int)mVVVerticesOptFull[i].size();
    return number / 3;
}

float MarchingCube_cls::
GetVertexDensity(){
    int number = 0;
    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	number += mVAllocSizeOrg[i];
    float density = (float)number / (mSize[X] * mSize[Y] * mSize[Z]);
    return density;
}

float MarchingCube_cls::
GetMaxVertexDensity(int layerFactor){
    int number = 0;
    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	if (mVAllocSize[i] > number)
	    number = mVAllocSize[i];
    float density = (float)number / (mSize[X] * mSize[Y] * layerFactor);
    return density;
}

// ThreadԍǗ layerԍǗ̏B
// OptimizeVertices ł̂̕YB
bool MarchingCube_cls::
ThreadControlMain(unsigned char *pData, int *pSkip,
		  int xLayerBase, int zLayerBase,
		  bool bUseSSE, int threadNum, const int layerFactor,
		  bool bOptimize, bool bPrecise){
    
    float *pTmpVertices[8];
    float *pTmpNormal[8];
    int arrayIndex[8];
    ThreadMarching *pCThread[8];
    bool AbAllocationVertices[8] = {FALSE};
    bool AbAllocationNormal[8] = {FALSE};
    bool bAllocation = TRUE;

    int xIni = xLayerBase;
    int xFin = xLayerBase + mSize[X]/1;
    if (xFin == mSize[X]) xFin = mSize[X] - 1;
    int xSize = xFin - xIni;

    // Kvobt@vZB_}bNX1box5*3(*3)iႤHjǁA
    // ɂ܂ł͂ȂƉ
    int bufferSize = sizeof(float)* xSize * mSize[Y] * layerFactor * 10;
    
    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	int zLayerIndex = zLayerBase + threadIndex * layerFactor;
	int zIni = zLayerIndex;
	int zFin = zLayerIndex + layerFactor;
	
	if (zIni >= (int)mSize[Z]-1) continue;
	if (zFin >= (int)mSize[Z]-1) zFin = mSize[Z] - 1;
	
	if ((pTmpVertices[threadIndex] = (float*)malloc(bufferSize)) == NULL){
	    printf("malloc fail Vertices %d %d\n", zIni, threadIndex);
	    bAllocation = FALSE; threadIndex = threadNum; continue; }
	else
	    AbAllocationVertices[threadIndex] = TRUE;

	if ((pTmpNormal[threadIndex] = (float*)malloc(bufferSize)) == NULL){
	    printf("malloc fail Normal %d %d\n", zIni, threadIndex);
	    bAllocation = FALSE; threadIndex = threadNum; continue; }
	else
	    AbAllocationNormal[threadIndex] = TRUE;
		
	pCThread    [threadIndex] = 
	    new ThreadMarching(pData,
			       pTmpVertices[threadIndex], 
			       pTmpNormal[threadIndex],
			       &arrayIndex[threadIndex]);
	arrayIndex  [threadIndex] = 0;
	
	pCThread[threadIndex]->
	    SetParameters(mSize, xIni, xFin, zIni, zFin, pSkip,
			  mTargetValue, bUseSSE, bPrecise);
    }

    if (bAllocation == FALSE) {
	for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	    if (AbAllocationVertices[threadIndex] == TRUE)
		free(pTmpVertices[threadIndex]);
	    if (AbAllocationNormal[threadIndex] == TRUE)
		free(pTmpNormal[threadIndex]);
	}
	return FALSE;
    }
        
    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	if (zLayerBase + threadIndex * layerFactor >= mSize[Z]-1) continue;
	pCThread[threadIndex]->Run();
    }
    
    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	if (zLayerBase + threadIndex * layerFactor >= mSize[Z]-1) continue;
	pCThread[threadIndex]->Wait(); 
    }

    // RR猋ʂ̃Rs[
    int verticesIndexOptInit = mVerticesIndex;
    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	int zLayerIndex = zLayerBase + threadIndex * layerFactor;
	if (zLayerIndex >= mSize[Z]-1) continue;

	bool bMallocVertex = TRUE, bMallocNormal = TRUE;
	if (arrayIndex[threadIndex] != 0){
	    if (arrayIndex[threadIndex] > bufferSize/4 ||
		bOptimize == TRUE){
		// ǃfĂڂڂɁc
		if ((mVpVertices[mVerticesIndex] = (float*)realloc((void*)pTmpVertices[threadIndex], sizeof(float)*arrayIndex[threadIndex]))== NULL){
		    printf("raalloc fail Vertices %d \n", arrayIndex[threadIndex]);
		    bAllocation = FALSE; }
		if ((mVpNormals[mVerticesIndex] = (float*)realloc((void*)pTmpNormal[threadIndex], sizeof(float)*arrayIndex[threadIndex])) == NULL){
		    printf("raalloc fail Normals %d \n", arrayIndex[threadIndex]);
		    bAllocation = FALSE; }
		
	    }
	    else{
		// xǁA̍ĔzuĂB
		mVpVertices[mVerticesIndex] = 
		    (float*)malloc(sizeof(float)*arrayIndex[threadIndex]);
		if (mVpVertices[mVerticesIndex] == NULL)
		    bMallocVertex = FALSE;
		else
		    for (int i = 0; i < arrayIndex[threadIndex]; i++)
			mVpVertices[mVerticesIndex][i] = 
			    pTmpVertices[threadIndex][i]; 
		delete pTmpVertices[threadIndex];
		
		mVpNormals[mVerticesIndex]  = 
		    (float*)malloc(sizeof(float)*arrayIndex[threadIndex]);
		if (mVpNormals[mVerticesIndex] == NULL)
		    bMallocNormal = FALSE;
		else
		    for (int i = 0; i < arrayIndex[threadIndex]; i++)
			mVpNormals[mVerticesIndex][i] = 
			    pTmpNormal[threadIndex][i]; 
		delete pTmpNormal[threadIndex];
	    }
	    if (bMallocVertex == TRUE && bMallocNormal == TRUE)
		// arrayIndex = float̐ = triangle̐*3
		mVAllocSize[mVerticesIndex] = arrayIndex[threadIndex]; 
	    else{
		mVAllocSize[mVerticesIndex] = 0;
		if (bMallocVertex == TRUE) delete mVpVertices[mVerticesIndex];
		if (bMallocNormal == TRUE) delete mVpNormals[mVerticesIndex];
	    }
	    
	}
	else{
	    delete pTmpVertices[threadIndex];
	    delete pTmpNormal[threadIndex];
	    mVAllocSize[mVerticesIndex] = 0;
	}
	
	mVAllocSizeOrg[mVerticesIndex] = mVAllocSize[mVerticesIndex];
	delete pCThread[threadIndex];
	
	mVerticesIndex++;
    }
    
    if (bOptimize == TRUE)
	OptimizeVerticesMainBlock(threadNum);
    
    return TRUE;
}

void MarchingCube_cls::
ConstructVertices(unsigned char *pData, int sizeX, int sizeY, int sizeZ,
		  double zMin, double zMax,
		  bool bUseSSE, int threadNum, const int layerFactor,
		  bool bOptimize, bool bTrimEdge, bool bPrecise){
    GbGLLockForMultithread = TRUE;

    int ASkip[3] = {0, 0 ,0};

    Clear(sizeZ);
    ClearOpt();
    ClearOptFull();
    mSize[X] = sizeX; mSize[Y] = sizeY; mSize[Z] = sizeZ;
    int zLayerBaseMin = sizeZ * zMin;
    int zLayerBaseMax = sizeZ * zMax; 
    if (zLayerBaseMax == sizeZ) zLayerBaseMax --;
    mTargetValue255 = mTargetValue * 255;
    
    unsigned char *pDataLink = pData;
    unsigned char *pDataTmp;

    if (bTrimEdge == TRUE){ // E0ɂ
	pDataTmp = new unsigned char [sizeX * sizeY * sizeZ];
	memcpy(pDataTmp, pData, sizeX * sizeY * sizeZ);
	int offsetY = 0;
	int offsetZ = 0;
	for (int z = 0; z < sizeZ; z+= sizeZ - 1){
	    offsetZ = z * sizeX * sizeY;
	    for (int y = 0; y < sizeY; y++){
		offsetY = offsetZ + y * sizeX;
		for (int x = 0; x < sizeX; x++) pDataTmp[offsetY + x] = 0;
	    } }
	for (int y = 0; y < sizeY; y+= sizeY - 1){
	    offsetY = y * sizeX;
	    for (int z = 0; z < sizeZ; z++){
		offsetZ = offsetY + z * sizeX * sizeY;
		for (int x = 0; x < sizeX; x++) pDataTmp[offsetZ + x] = 0;
	    } }
	for (int x = 0; x < sizeX; x+= sizeX - 1)
	    for (int z = 0; z < sizeZ; z++){
		offsetZ = z * sizeX * sizeY;
		for (int y = 0; y < sizeY; y++){
		    offsetY = offsetZ + y * sizeX;
		    pDataTmp[offsetY + x] = 0;
		} }
	pDataLink = pDataTmp;
    }

    mVerticesIndex = 0;
    for (int xLayerBase = 0; xLayerBase < mSize[X]; xLayerBase += mSize[X]/1)
	for (int zLayerBase = zLayerBaseMin; zLayerBase < zLayerBaseMax;
	     zLayerBase += threadNum * layerFactor){
	    
	    if (ThreadControlMain
		(pDataLink, ASkip, xLayerBase, zLayerBase,
		 bUseSSE, threadNum, layerFactor, bOptimize, bPrecise) == FALSE){
		
		if (bTrimEdge == TRUE) delete pDataTmp;
		GbGLLockForMultithread = FALSE;
		return;
	    }
	}
	
    //LOOP_XYZ { mScalePrev[xyz] = 1; }

    if (bTrimEdge == TRUE) delete pDataTmp;
    GbGLLockForMultithread = FALSE;
}

void MarchingCube_cls::
OptimizeVerticesMainBlock(int threadNum){
    if (mbMemoryReadyOpt == FALSE) ClearOpt();

    ThreadOptimizeA *pCThread[8];
    int optedIndex[8];
    
    int threadIndex = 0;
    for (unsigned int index = 0; index < mVpVertices.size(); index++){
	
	if (mVbOptimized[index] == FALSE &&
	    mVAllocSize[index] != 0){
	    // 1/3x܂ŗ\Allocămۂ̃I[o[wbh҂
	    mVVVerticesOpt[index].reserve(mVAllocSize[index] / 3);
	    mVVNormalsOpt[index].reserve(mVAllocSize[index] / 3);
	    pCThread[threadIndex] = new ThreadOptimizeA
		(mVpVertices[index], mVpNormals[index],
		 mVAllocSize[index],
		 &mVVVerticesOpt[index], &mVVNormalsOpt[index],
		 &mVVVertexIndicesOpt[index]);
	    optedIndex[threadIndex] = index;
	    threadIndex ++;
	}
	
	if (threadIndex >= threadNum ||
	    index == mVpVertices.size() - 1){
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Run();
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Wait();
	    for (int i = 0; i < threadIndex; i++) delete pCThread[i];
	    for (int i = 0; i < threadIndex; i++){
		int opted = optedIndex[i];
		if (mVAllocSize[opted] != 0){
		    free(mVpVertices[opted]);
		    free(mVpNormals[opted]);
		    mVAllocSize[opted] = 0;
		    mVbOptimized[opted] = TRUE;
		}
		mVOptVertexNumOrg[opted] = (int)mVVVerticesOpt[opted].size();
	    }
	    threadIndex = 0;
	}
    }
}

void MarchingCube_cls::
OptimizeVerticesMainBlockFull(int threadNum){
    if (mbMemoryReadyOptFull == FALSE) ClearOptFull();

    ThreadOptimizeFull *pCThread[8];
    int optedIndex[8];

    int threadIndex = 0;
    for (unsigned int index = 0; index < mVpVertices.size(); index++){
	
	if (mVbOptimizedFull[index] == FALSE &&
	    mVVVerticesOpt[index].size() != 0){
	    // 1/3x܂ŗ\Allocămۂ̃I[o[wbh҂
	    mVVVerticesOptFull[index].reserve(mVVVerticesOpt[index].size()/ 3);
	    mVVNormalsOptFull[index].reserve(mVVNormalsOpt[index].size()/ 3);
	    pCThread[threadIndex] = new ThreadOptimizeFull
		(&mVVVerticesOpt[index], &mVVNormalsOpt[index],
		 &mVVVertexIndicesOpt[index],
		 mVVVerticesOpt[index].size(),
		 &mVVVerticesOptFull[index], &mVVNormalsOptFull[index],
		 &mVVVertexIndicesOptFull[index], mSize[Y]);
	    optedIndex[threadIndex] = index;
	    threadIndex ++;
	}

	if (threadIndex >= threadNum ||
	    index == mVpVertices.size() - 1){
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Run();
	    for (int i = 0; i < threadIndex; i++) pCThread[i]->Wait();
	    for (int i = 0; i < threadIndex; i++) delete pCThread[i];
	    for (int i = 0; i < threadIndex; i++){
		int opted = optedIndex[i];
		if (mVVVerticesOpt[opted].size() != 0){
		    mVVVerticesOpt[opted].clear();
		    mVVNormalsOpt[opted].clear();
		    mVVVertexIndicesOpt[opted].clear();
		    mVbOptimizedFull[opted] = TRUE;
		}
	    }
	    threadIndex = 0;
	}
    }
    ClearOpt();
}

void MarchingCube_cls::
OptimizeVertices(int threadNum){
    // Xbh҂Ƀ{^ƃRRɍēxĂ܂B
    if (GbGLLockForMultithread == TRUE) return; 
    GbGLLockForMultithread = TRUE;
    OptimizeVerticesMainBlock(threadNum);
    GbGLLockForMultithread = FALSE;
}

void MarchingCube_cls::
OptimizeVerticesFull(int threadNum){
    // Xbh҂Ƀ{^ƃRRɍēxĂ܂B
    if (GbGLLockForMultithread == TRUE) return; 
    GbGLLockForMultithread = TRUE;
    OptimizeVerticesMainBlockFull(threadNum);
    GbGLLockForMultithread = FALSE;
}

void MarchingCube_cls::
MakeUV1D(float pos[3], float mapscale,
	 double offset[3], double scale[3]){

    size_t size = 0; // VKlʂɂȂĂ0ɂȂ
    for (unsigned int index = 0; index < mVpNormals.size(); index++){
	size += mVVUV1DOpt[index].size();
	size += mVVUV1DOptFull[index].size();
    }

    if (IsParameterIsNewOne(pos, mapscale, offset, scale) == FALSE && 
	size != 0)
	return;

    // ̊mۗ̈悲Clear@(TCY0̈ꎞxNgƌ)
    for (unsigned int index = 0; index < mVpNormals.size(); index++){
	std::vector<float>().swap(mVVUV1DOpt[index]);
	std::vector<float>().swap(mVVUV1DOptFull[index]);
    }

    for (unsigned int index = 0; index < mVpNormals.size(); index++){
	int iSize = (int)mVVVerticesOpt[index].size();
	double localPos[3];
	LOOP_XYZ localPos[xyz] = (pos[xyz] - offset[xyz]) / scale[xyz];

	// FIX_ME!! ̗]n
	for (int i = 0; i < iSize; i+=3){
	    double distance = sqrt
		((mVVVerticesOpt[index][i+X] - localPos[X]) * (mVVVerticesOpt[index][i+X] - localPos[X]) * scale[X] * scale[X] +
		 (mVVVerticesOpt[index][i+Y] - localPos[Y]) * (mVVVerticesOpt[index][i+Y] - localPos[Y]) * scale[Y] * scale[Y] +
		 (mVVVerticesOpt[index][i+Z] - localPos[Z]) * (mVVVerticesOpt[index][i+Z] - localPos[Z]) * scale[Z] * scale[Z]) / mapscale;
	    mVVUV1DOpt[index].push_back(distance);
	}

	iSize = (int)mVVVerticesOptFull[index].size();
	// FIX_ME!! ̗]n
	for (int i = 0; i < iSize; i+=3){
	    double distance = sqrt
		((mVVVerticesOptFull[index][i+X] - localPos[X]) * (mVVVerticesOptFull[index][i+X] - localPos[X]) * scale[X] * scale[X] +
		 (mVVVerticesOptFull[index][i+Y] - localPos[Y]) * (mVVVerticesOptFull[index][i+Y] - localPos[Y]) * scale[Y] * scale[Y] +
		 (mVVVerticesOptFull[index][i+Z] - localPos[Z]) * (mVVVerticesOptFull[index][i+Z] - localPos[Z]) * scale[Z] * scale[Z]) / mapscale;
	    mVVUV1DOptFull[index].push_back(distance);
	}
    }

    mPrevMapScale = mapscale;
    LOOP_XYZ mPrevPos[xyz] = pos[xyz];
    LOOP_XYZ mPrevOffset[xyz] = offset[xyz];
    LOOP_XYZ mPrevScale[xyz] = scale[xyz];
}

void MarchingCube_cls::
GLDraw(double center[3], double scale[3], bool bCullFace, GLuint showType,
       TexProperty_cls *pCTexProp, int colorType){

    if (bCullFace == TRUE)
	glEnable(GL_CULL_FACE);

    double offset[3];
    LOOP_XYZ offset[xyz] = center[xyz] - scale[xyz]/2;

    Configure1DMapping(pCTexProp, colorType,
		       offset, scale);

    glEnable(GL_NORMALIZE);
    if (DoesUseMapping() == TRUE)
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);

    for (unsigned int index = 0; index < mVpNormals.size(); index++){

	if (mVbOptimizedFull[index] == TRUE){
	    if (mVVVertexIndicesOptFull[index].size() != 0){
		glVertexPointer(3, GL_FLOAT, 0, &mVVVerticesOptFull[index][0]);
		glNormalPointer(GL_FLOAT, 0, &mVVNormalsOptFull[index][0]);
		if (DoesUseMapping() == TRUE)
		    glTexCoordPointer(1, GL_FLOAT, 0, &mVVUV1DOptFull[index][0]);
		glDrawElements(GL_TRIANGLES, (GLsizei)mVVVertexIndicesOptFull[index].size() , GL_UNSIGNED_INT, &mVVVertexIndicesOptFull[index][0]);
	    }
	}
	else if (mVbOptimized[index] == TRUE){
	    if (mVVVertexIndicesOpt[index].size() != 0){
		glVertexPointer(3, GL_FLOAT, 0, &mVVVerticesOpt[index][0]);
		glNormalPointer(GL_FLOAT, 0, &mVVNormalsOpt[index][0]);
		if (DoesUseMapping() == TRUE)
		    glTexCoordPointer(1, GL_FLOAT, 0, &mVVUV1DOpt[index][0]);
		glDrawElements(GL_TRIANGLES, (GLsizei)mVVVertexIndicesOpt[index].size() , GL_UNSIGNED_INT, &mVVVertexIndicesOpt[index][0]);
	    }
	}
	else{ 
	    if (mVAllocSize[index] != 0){
		glVertexPointer(3, GL_FLOAT, 0, mVpVertices[index]);
		glNormalPointer(GL_FLOAT, 0, mVpNormals[index]);
		glDrawArrays(showType, 0, mVAllocSize[index]/3);
	    }
	}
	
    }

    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisable(GL_NORMALIZE);

    glDisable(GL_CULL_FACE);
}

void MarchingCube_cls::
GetStrWaveFrontObj(wxString *pstrOut, bool bOffset, bool bNormal){

    // UNormalɖ߂
    //double scaleOrg[3], scale1[3] = {1, 1, 1};
    //LOOP_XYZ scaleOrg[xyz] = mScalePrev[xyz];
    //RecalcNormal(size1);
    
    // ̒_Triangleőz
    pstrOut->Alloc(2 * (GetVertexNumberOptFull() * (27 + bNormal * 27) + GetVertexNumber()/3 * 25 + 1000));
    
    *pstrOut << wxT("o Isosurface") << ENDLWX;

    // Full̂݁F
    std::vector < unsigned int > VvertexOffset;
    unsigned int offset = 1; // ŏ1
    VvertexOffset.push_back(offset); 
    size_t indexSize = mVpNormals.size();

    std::vector < std::vector < float > >
	*pVVVerticesOptFull = &mVVVerticesOptFull;
    std::vector < std::vector < float > >
	*pVVNormalsOptFull = &mVVNormalsOptFull;

    double vertOffset = 0;
    if (bOffset == TRUE) vertOffset = -0.5;
    for (size_t index = 0; index < indexSize; index++){
	size_t verticesSize = mVVVerticesOptFull[index].size();
	if (verticesSize == 0) continue;
	float *pVertex = &((*pVVVerticesOptFull)[index][0]);
	for (size_t v = 0; v < verticesSize; v+= 3){
	    *pstrOut << wxT("v ")
		     << *(pVertex + v + X) + vertOffset << wxT(" ")
		     << *(pVertex + v + Y) + vertOffset << wxT(" ")
		     << *(pVertex + v + Z) + vertOffset << ENDLWX;
	}
	offset += (int)verticesSize / 3;
	VvertexOffset.push_back(offset);
    }

    if (bNormal == TRUE){
	for (size_t index = 0; index < indexSize; index++){
	    size_t verticesSize = mVVVerticesOptFull[index].size();
	    if (verticesSize == 0) continue;
	    float *pNormal = &((*pVVNormalsOptFull)[index][0]);
	    for (size_t v = 0; v < verticesSize; v+= 3){
		*pstrOut << wxT("vn ")
			 << *(pNormal + v + X) << wxT(" ")
			 << *(pNormal + v + Y) << wxT(" ")
			 << *(pNormal + v + Z) << ENDLWX;
	    }
	    offset += (int)verticesSize / 3;
	    VvertexOffset.push_back(offset);
	}
    }

    *pstrOut << wxT("s off") << ENDLWX;

    for (unsigned int index = 0; index < indexSize; index++){
	size_t vertexIndexSize = mVVVertexIndicesOptFull[index].size();
	for (size_t v = 0; v < vertexIndexSize; v+= 3)
	    *pstrOut << wxT("f ")
		     << mVVVertexIndicesOptFull[index][v+X] + VvertexOffset[index] << wxT(" ")
		     << mVVVertexIndicesOptFull[index][v+Y] + VvertexOffset[index] << wxT(" ")
		     << mVVVertexIndicesOptFull[index][v+Z] + VvertexOffset[index] << ENDLWX;
    }

    //RecalcNormal(scaleOrg);
}

void MarchingCube_cls::
GetStrPovMesh2(wxString *pstrOut, wxString strID, wxString strTag, bool bNormal){

    // ̒_Triangleőz
    // wideȂ̂2H
    pstrOut->Alloc(2 * (GetVertexNumberOptFull() * 40 + GetVertexNumber()/3 * 40 + 1000));
    
    std::vector < unsigned int > VvertexOffset;
    unsigned int offset = 0;
    VvertexOffset.push_back(offset); // ŏ0
    size_t indexSize = mVpNormals.size();
    
    //*pstrOut << wxT("//scaling ") << scaleOrg[X] << wxT(" ") << scaleOrg[Y] << wxT(" ") << scaleOrg[Z] << ENDLWX;
    *pstrOut << wxT("//Mesh2file generated by ") << _FRAME_MAIN_APPNAME << ENDLWX;

    *pstrOut << wxT("//normal_start_at=");
    int posNormalTag = pstrOut->Len();
    *pstrOut << wxT("##########") << ENDLWX;

    *pstrOut << wxT("//index_start_at=");
    int posIndexTag = pstrOut->Len();
    *pstrOut << wxT("##########") << ENDLWX;


    *pstrOut << wxT("#declare ") << wxT("ISOSURF_") << strID << strTag << L"=mesh2{" << ENDLWX
	     << wxT("\tvertex_vectors{") << ENDLWX
	     << wxT("\t\t") << GetVertexNumberOptFull() << wxT (",") << ENDLWX;
    
    for (size_t index = 0; index < indexSize; index++){
	size_t verticesSize = mVVVerticesOptFull[index].size();
	offset += (int)verticesSize / 3;
	VvertexOffset.push_back(offset);
    }

    wxString strOutTriangles;
    ThreadExportMesh2PovSub CThreadSub(&strOutTriangles, indexSize,
				       GetVertexNumber()/3,
				       &mVVVertexIndicesOptFull,
				       &VvertexOffset);
    CThreadSub.Run();

    std::vector < std::vector < float > >
	*pVVVerticesOptFull = &mVVVerticesOptFull;
    std::vector < std::vector < float > >
	*pVVNormalsOptFull = &mVVNormalsOptFull;
    
    for (size_t index = 0; index < indexSize; index++){
	size_t verticesSize = mVVVerticesOptFull[index].size();
	if (verticesSize == 0) continue;
	float *pVertex = &((*pVVVerticesOptFull)[index][0]);
	for (size_t v = 0; v < verticesSize; v+= 3)
	    *pstrOut << 
		wxString::Format(wxT("\t\t<%0.6f,%0.6f,%0.6f>,"),
				 *(pVertex + v + X),
				 *(pVertex + v + Y),
				 *(pVertex + v + Z)) << ENDLWX;
    }
    pstrOut->RemoveLast(); pstrOut->RemoveLast();

    int posNormal = pstrOut->Len();

    if (bNormal == TRUE){
	*pstrOut << ENDLWX  << wxT("\t}") << ENDLWX
		 << wxT("\tnormal_vectors{") << ENDLWX
		 << wxT("\t\t") << GetVertexNumberOptFull() << wxT (",") 
		 << ENDLWX;
	
	wxString strNormal;
	strNormal.Alloc(2 * (GetVertexNumberOptFull() * 40));

	for (size_t index = 0; index < indexSize; index++){
	    size_t verticesSize = mVVVerticesOptFull[index].size();
	    if (verticesSize == 0) continue;
	    float *pNormal = &((*pVVNormalsOptFull)[index][0]);
	    for (size_t v = 0; v < verticesSize; v+= 3)
		strNormal << 
		    wxString::Format(wxT("\t\t<%0.4f,%0.4f,%0.4f>,"),
				     *(pNormal + v + X),
				     *(pNormal + v + Y),
				     *(pNormal + v + Z)) << ENDLWX;
	}
	strNormal.RemoveLast(); strNormal.RemoveLast();
	*pstrOut << strNormal;
    }

    *pstrOut << ENDLWX 
	     << wxT("\t}") << ENDLWX;

    CThreadSub.Wait();
    
    /*
    *pstrOut << wxT("\tface_indices{") << ENDLWX
	     << wxT("\t\t") << GetVertexNumber()/3 << wxT (",") << ENDLWX;
    
    for (unsigned int index = 0; index < indexSize; index++){
	size_t vertexIndexSize = mVVVertexIndicesOptFull[index].size();
	if (vertexIndexSize == 0) continue;
	unsigned int *pIndex = &(mVVVertexIndicesOptFull[index][0]);
	unsigned int indexOffset = VvertexOffset[index];
	for (size_t v = 0; v < vertexIndexSize; v+= 3){
	    *pstrOut << wxT("\t\t<")
		     << *(pIndex + v + X) + indexOffset << wxT(",")
		     << *(pIndex + v + Y) + indexOffset << wxT(",")
		     << *(pIndex + v + Z) + indexOffset << wxT(">")
		     << wxT(",") << ENDLWX;
	}
    }
    
    pstrOut->RemoveLast(); pstrOut->RemoveLast();

    *pstrOut << ENDLWX 
	     << wxT("\t}\n}") << ENDLWX;
    */

    int posIndex = pstrOut->Len();

    *pstrOut += strOutTriangles;

    // inside_vector͈Ȃi|SƓjȂ牽łǂ
    *pstrOut << ENDLWX 
	     << wxT("\t}") << ENDLWX
	     << wxT("\tinside_vector <0.01, 0.01, 0.99>") << ENDLWX
	     << ENDLWX << wxT("}") << ENDLWX;
				 
    wxString strNormalTag = wxString::Format(wxT("%010d"), posNormal);
    wxString strIndexTag = wxString::Format(wxT("%010d"), posIndex);

    for (int i = 0; i < 10; i++)
	pstrOut->SetChar(posNormalTag + i, strNormalTag.GetChar(i));
    for (int i = 0; i < 10; i++)
	pstrOut->SetChar(posIndexTag + i, strIndexTag.GetChar(i));
}

