// "clsMarchMainThread.cpp"
// (setq c-basic-offset 4)

#ifndef _ISOSURFACE_MARCHING_MAINTHREAD_H_
#define _ISOSURFACE_MARCHING_MAINTHREAD_H_

#include "../../COMMON/clsMultiThreadCtrl.h"

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

//#include <wx/wx.h>
//#include <wx/thread.h>

class MarchingCube_cls;

// The main routine of this class used
// Marching Cubes Example Program 
// by Cory Bloyd (corysama@yahoo.com)
// which is publicshd as public domain 

class ThreadMarching : public MultiThreadThread_cls {
 public:
    ThreadMarching(wxThreadKind kind, unsigned char *pData, 
		   int *pAllocSize,
		   float **ppVertices, float **ppNormals,
		   int bufferSize);

    void SetParameters(int size[3], 
		       int initZ, int finZ, 
		       int initY, int finY,
		       int initX, int finX,
		       float targetValue, 
		       bool bOptimize, bool bUseSSE, bool bPrecise);
    
    GLfloat GetOffset(GLfloat fValue1, GLfloat fValue2, GLfloat fValueDesired);
    float GetInterpolatedValue
	(const unsigned char *pData,
	 const int lBaseIndex, const float fx, const float fy, const float fz);
    void GetNormalPrecise2
	(GLfloat ANorm[3], const unsigned char *pData,
	 const int x, const GLfloat fX, 
	 const int y, const GLfloat fY,
	 const int z, const GLfloat fZ);
    void GetNormalPrecise
	(GLfloat ANorm[3], const unsigned char *pData,
	 const int x, const GLfloat fX, 
	 const int y, const GLfloat fY,
	 const int z, const GLfloat fZ);
    void GetNormalPreciseSSE
	(GLfloat ANorm[3], const unsigned char *pData,
	 const int x, const GLfloat fX, 
	 const int y, const GLfloat fY,
	 const int z, const GLfloat fZ);

    int MarchCubeMainThread(const unsigned char *pData,
			       const int x, const int y, const int z);

    int MarchCubeMainThreadSSE(const unsigned char *pData,
			    const int x, const int y, const int z);
    
    
    int MarchCubeLOOPX1(const unsigned char *pData, int y, int z,
			int startX, int finX);
    //void MarchCubeLOOPX2(const unsigned char *pData, int y, int z);
    int MarchCubeLOOPY(const unsigned char *pData, int z,
		       int startY, int finY, int startX, int finX);
   
    bool CopyResults();
    virtual ExitCode Entry();

    unsigned char * const mpData;
    float *mpTmpVertices;
    float *mpTmpNormal;
    int   mArrayIndex;
    int   * const mpAllocSize;
    float ** mppVertices;
    float ** mppNormals;

    int mBufferSize;
    int mSize[3];
    int mInitX, mFinX;
    int mInitY, mFinY;
    int mInitZ, mFinZ;

    int mOffsetZ, mOffsetY;
    int mBaseIndex;
    float mTargetValue, mTargetValue255;
    bool mbOptimize, mbUseSSE, mbPrecise;

    static const GLint   VOffset[8][3];
    static const GLint   a2iEdgeConnection[12][2];
    static const GLfloat a2fEdgeDirection[12][3];

    static const GLint aiCubeEdgeFlags[256];
    static const GLint a2iTriangleConnectionTable[256][16];
};

#endif
