// "clsMarchingCube.h"
// (setq c-basic-offset 4)

#include "clsIsosurfOptimize.h"
#include "clsMarchingCube.h"
#include "clsMarchMainThread.h"
#include <math.h>

#include "../../DF3/clsDF3SeqCtrl.h"
#include "../../MYAPI/MsgBox.h"
#include "../../POVRAY/clsPovRay.h"
#include "../../../wxWidgetString.h"

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

// FIX_ME!! Ƃ肠ł`
#define _MAX_THREAD_NUM 16

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2

#ifdef WIN32
#define ENDLWX wxT("\r\n")
#else
#define ENDLWX wxT("\n")
#endif

// ----

ThreadExportMesh2PovSubI::
ThreadExportMesh2PovSubI(wxString *pstrOut, unsigned int indexSize,
			unsigned int polygonNumber,
			std::vector < std::vector < unsigned int > > *pVVVertexIndicesOptFull,
			std::vector < unsigned int > *pVvertexOffset) 
    : wxThread(wxTHREAD_JOINABLE),
      mpStrOut(pstrOut), mIndexSize(indexSize), mPolygonNumber(polygonNumber), mpVVVertexIndicesOptFull(pVVVertexIndicesOptFull),
      mpVvertexOffset(pVvertexOffset)
{
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadExportMesh2PovSubI::
Entry(){

    wxString *pstrOut = mpStrOut;
    pstrOut->Alloc(2 * mPolygonNumber * 40 + 1000);
    
    *pstrOut << wxT("\tface_indices{") << ENDLWX
	     << wxT("\t\t") << mPolygonNumber << wxT (",") << ENDLWX;

    std::vector < std::vector < unsigned int > > 
	*pVVVertexIndicesOptFull = mpVVVertexIndicesOptFull;
    std::vector < unsigned int >
	*pVvertexOffset = mpVvertexOffset;

    for (unsigned int index = 0; index < mIndexSize; index++){
	size_t vertexIndexSize = (*mpVVVertexIndicesOptFull)[index].size();
	size_t vertexIndexNum = vertexIndexSize / 3;
	if (vertexIndexSize == 0) continue;
	unsigned int *pIndex = &((*pVVVertexIndicesOptFull)[index][0]);
	unsigned int indexOffset = (*pVvertexOffset)[index];
	for (size_t v = 0; v < vertexIndexNum; v++){
	    if (v % 4 == 0)
		*pstrOut << wxT("\t\t");
	    *pstrOut << 
		wxString::Format(wxT("<%d,%d,%d>,"),
				 *(pIndex + v*3 + X) + indexOffset,
				 *(pIndex + v*3 + Y) + indexOffset,
				 *(pIndex + v*3 + Z) + indexOffset);
	    if (v % 4 == 3 || v == vertexIndexNum - 1)
		*pstrOut << ENDLWX;
	}
    }

    pstrOut->RemoveLast(); pstrOut->RemoveLast();
    
    // Ō͂ł͕Ȃ
    
    return 0;
}

ThreadExportMesh2PovSubN::
ThreadExportMesh2PovSubN(wxString *pstrOut, unsigned int indexSize,
			 unsigned int normalNumber,
			 std::vector < std::vector < float > > *pVVNormalsOptFull,
			 bool bValid) 
    : wxThread(wxTHREAD_JOINABLE),
      mpStrOut(pstrOut), mIndexSize(indexSize), mNormalNumber(normalNumber), mpVVNormalsOptFull(pVVNormalsOptFull)
{
    if (bValid)
	if ( Create() != wxTHREAD_NO_ERROR) 
	    MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadExportMesh2PovSubN::
Entry(){
    
    wxString *pstrOut = mpStrOut;
    wxString strNormal;
    strNormal.Alloc(2 * mNormalNumber * 40);
    
    strNormal << wxT("\tnormal_vectors{") << ENDLWX
	      << wxT("\t\t") << mNormalNumber << wxT (",") << ENDLWX;

    std::vector < std::vector < float > > 
	*pVVNormalsOptFull = mpVVNormalsOptFull;
        
    
    for (unsigned int index = 0; index < mIndexSize; index++){
	size_t normalSize = (*mpVVNormalsOptFull)[index].size();
	size_t normalNum = normalSize / 3;
	if (normalSize == 0) continue;
	float *pNormal = &((*pVVNormalsOptFull)[index][0]);
	for (size_t n = 0; n < normalNum; n++){
	    char buf[64];
	    if (n % 3 == 0 && n == normalNum - 1)
		sprintf(buf, "\t\t<%0.2f,%0.2f,%0.2f>,\n",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z)); 
	    else if (n % 3 == 2)
		sprintf(buf, "<%0.2f,%0.2f,%0.2f>,\n",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z));
	    else if (n % 3 == 0)
		sprintf(buf, "\t\t<%0.2f,%0.2f,%0.2f>",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z));
	    else
		sprintf(buf, "<%0.2f,%0.2f,%0.2f>",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z));
	    strNormal << wxString::From8BitData(buf);
	}
    }

    while (strNormal.Last() != wxT('>'))
	strNormal.RemoveLast(); //strNormal.RemoveLast();
    
    *pstrOut << strNormal;
    
    return 0;
}

// ----

MarchingCube_cls::
MarchingCube_cls(){
    mbMemoryReadyOpt = FALSE;
    mbMemoryReadyOptFull = FALSE;
}

MarchingCube_cls::
~MarchingCube_cls(){
    for (unsigned int i = 0; i < mVpVertices.size(); i++){
	if (mVAllocSize[i] != 0){
	    free(mVpVertices[i]);
	    free(mVpNormals[i]);
	    mVAllocSize[i] = 0;
	}
    }
}

void MarchingCube_cls::
Clear(int sizeZ){
    for (unsigned int i = 0; i < mVpVertices.size(); i++){
	if (mVAllocSize[i] != 0){
	    free(mVpVertices[i]);
	    free(mVpNormals[i]);
	    mVAllocSize[i] = 0;
	    mVVertexNumOrg[i] = 0;
	}
    }

    mVpVertices.clear();
    mVpNormals.clear();
    mVAllocSize.clear();
    mVVertexNumOrg.clear();

    // ̐܂K
    /*
    mVpVertices.resize(2 * sizeZ);
    mVpNormals.resize(2 * sizeZ);
    mVAllocSize.resize(2 * sizeZ);
    mVVertexNumOrg.resize(2 * sizeZ);
    mVbOptimized.resize(2 * sizeZ);
    mVbOptimizedFull.resize(2 * sizeZ);
    mVOptVertexNumOrg.resize(2 * sizeZ);
    */

    mVpVertices.resize(512*2);
    mVpNormals.resize(512*2);
    mVAllocSize.resize(512*2);
    mVVertexNumOrg.resize(512*2);
    mVbOptimized.resize(512*2);
    mVbOptimizedFull.resize(512*2);
    mVOptVertexNumOrg.resize(512*2);

    for (unsigned int i = 0; i < mVbOptimized.size(); i++){
	mVbOptimized[i] = FALSE;
	mVOptVertexNumOrg[i] = 0;
	mVbOptimizedFull[i] = FALSE;
    }

    mbMemoryReadyOpt = FALSE;
    mbMemoryReadyOptFull = FALSE;
}

void MarchingCube_cls::
ClearOpt(){
    for (unsigned int i = 0; i < mVVVerticesOpt.size(); i++){
	// ̊mۗ̈悲Clear@(TCY0̈ꎞxNgƌ)
	std::vector<float>().swap(mVVVerticesOpt[i]);
	std::vector<float>().swap(mVVNormalsOpt[i]);
	std::vector<float>().swap(mVVUV1DOpt[i]);
	std::vector<unsigned int>().swap(mVVVertexIndicesOpt[i]);
    }

    mVVVerticesOpt.resize(mVpVertices.size());
    mVVNormalsOpt.resize(mVpVertices.size());
    mVVUV1DOpt.resize(mVpVertices.size());
    mVVVertexIndicesOpt.resize(mVpVertices.size());
    mbMemoryReadyOpt = TRUE;
}

void MarchingCube_cls::
ClearOptFull(){
    for (unsigned int i = 0; i < mVVVerticesOptFull.size(); i++){
	std::vector<float>().swap(mVVVerticesOptFull[i]);
	std::vector<float>().swap(mVVNormalsOptFull[i]);
	std::vector<float>().swap(mVVUV1DOptFull[i]); 
	std::vector<unsigned int>().swap(mVVVertexIndicesOptFull[i]);
    }
    
    mVVVerticesOptFull.resize(mVpVertices.size());
    mVVNormalsOptFull.resize(mVpVertices.size());
    mVVUV1DOptFull.resize(mVpVertices.size());
    mVVVertexIndicesOptFull.resize(mVpVertices.size());
    mbMemoryReadyOptFull = TRUE;
}

unsigned int MarchingCube_cls::
GetVertexNumber(){
    unsigned int number = 0;
    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	number += mVVertexNumOrg[i];
    return number / 3;
}

unsigned int MarchingCube_cls::
GetVertexNumberOpt(){
    unsigned int number = 0;
    for (unsigned int i = 0; i < mVVVerticesOpt.size(); i++)
	number += mVOptVertexNumOrg[i];
    return number / 3;
}

int MarchingCube_cls::
GetVertexNumberOptFull(){
    int number = 0;
    for (unsigned int i = 0; i < mVVVerticesOpt.size(); i++)
	number += (int)mVVVerticesOptFull[i].size();
    return number / 3;
}

float MarchingCube_cls::
GetVertexDensity(){
    int number = 0;
    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	number += mVVertexNumOrg[i];
    float density = (float)number / (mSize[X] * mSize[Y] * mSize[Z]);
    return density;
}

float MarchingCube_cls::
GetMaxVertexDensity(int layerFactor){
    int number = 0;
    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	if (mVAllocSize[i] > number)
	    number = mVAllocSize[i];
    float density = (float)number / (mSize[X] * mSize[Y] * layerFactor);
    return density;
}

bool MarchingCube_cls::
ThreadControlMain(unsigned char *pData, bool bUseSSE,
		  const int lineStep, bool bOptimize, bool bPrecise){
    
    InitializeThreadCtrl();

    //int intervalZ = mSize[Z] / 8;
    //int intervalY = mSize[Y] / 8;
    //int intervalX = mSize[X] / 8;

    //int indexX = 0;
    //int indexY = 0;
    //int indexZ = 0;

    int bufferSize = 
	sizeof(float) * (mSize[X]/8)*(mSize[Y]/8)*(mSize[Z]/8)* 12;

    int verticesIndex = 0;
    for (int indexZ = 0; indexZ < 8; indexZ++){
	int initZ = mSize[Z] * indexZ / 8;
	int finZ  = mSize[Z] * (indexZ+1) / 8;
	if (finZ > mSize[Z] - 1) finZ = mSize[Z] - 1; 
	// Ō̃C[͏ƃG[
	
	for (int indexY = 0; indexY < 8; indexY++){
	    int initY = mSize[Y] * indexY /8;
	    int finY  = mSize[Y] * (indexY+1) /8;
	    if (finY > mSize[Y] - 1) finY = mSize[Y] - 1;
	    
	    for (int indexX = 0; indexX < 8; indexX++){
		
		int min = mSectionMinValue[indexZ][indexY][indexX];
		int max = mSectionMaxValue[indexZ][indexY][indexX];
		int tgt = mTargetValue * 255;
		if (tgt < min || tgt > max){
		    verticesIndex++;
		    continue;
		}

		int initX = mSize[X] * indexX / 8;
		int finX  = mSize[X] * (indexX+1) / 8;
		if (finX > mSize[X] - 1) finX = mSize[X] - 1;
		
		ThreadMarching *pCThread = new ThreadMarching
		    (wxTHREAD_DETACHED, pData, 
		     &mVAllocSize[verticesIndex],  &mVpVertices[verticesIndex],
		     &mVpNormals[verticesIndex],  bufferSize);
		pCThread->SetParameters
		    (mSize, 
		     initZ, finZ, initY, finY, initX, finX,
		     mTargetValue, bOptimize, bUseSSE, bPrecise);
		WaitForThreadsAndRun(pCThread);
		
		verticesIndex++;
	    }
	}
    }
	
    // Kvobt@ _}bNX1box5*3(*3)iႤHjǁA
    // ɂ܂ł͂ȂƉ
    //int bufferSize = sizeof(float)* mSize[X] * mSize[Y] * lineStep * 12;
    
    /*
    // XbhPłobt@̖肪̂ŕ͕ς炸B
    int verticesIndex = 0;
    for (int z = 0; z < mSize[Z]; z += lineStep){
	int threadEndZ = z + lineStep;
	if (threadEndZ > mSize[Z] - 1) threadEndZ = mSize[Z] - 1; 
	// Ō̃C[͏ƃG[

	// Kvobt@ _}bNX1box5*3(*3)iႤHjǁA
	// ɂ܂ł͂ȂƉ
	int bufferSize = sizeof(float)* mSize[X] * mSize[Y] * lineStep * 12;

	ThreadMarching *pCThread = new ThreadMarching
	    (wxTHREAD_DETACHED, pData, 
	     &mVAllocSize[verticesIndex],  &mVpVertices[verticesIndex],
	     &mVpNormals[verticesIndex],  bufferSize);
	pCThread->SetParameters
	    (mSize, z, threadEndZ, 
	     mTargetValue, bOptimize, bUseSSE, bPrecise);

	WaitForThreadsAndRun(pCThread);

	verticesIndex++;
    }
    */
    
    WaitAllThreadsEnd(); 

    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	mVVertexNumOrg[i] = mVAllocSize[i];

    return TRUE;
}

std::vector < unsigned char > MarchingCube_cls::
GetMinMaxValue(unsigned char *pData, 
	       int initZ, int finZ, int initY, int finY, int initX, int finX){

    unsigned char max = 0;
    unsigned char min = 255;

    int sizeX = mSize[X];
    int sizeY = mSize[Y];
    int sizeZ = mSize[Z];

    for (int z = initZ; z < finZ; z++){
	int offsetZ = z * sizeX * sizeY;
	for (int y = initY; y < finY; y++){
	    int offsetY = offsetZ + y * sizeX;
	    for (int x = initZ; x < finX; x++){
		int offset = offsetY + x;
		if (pData[offset] > max) max = pData[offset];
		if (pData[offset] < min) min = pData[offset];
	    }
	}
    }

    std::vector < unsigned char > retVec (2);
    retVec[0] = min;
    retVec[1] = max;
    return retVec;
}

void MarchingCube_cls::
ConstructVertices(unsigned char *pData, int sizeX, int sizeY, int sizeZ,
		  double zMin, double zMax,
		  bool bUseSSE, const int layerFactor, 
		  bool bRecalcMinMax,
		  bool bOptimize, bool bTrimEdge, bool bPrecise){

    Clear(sizeZ);
    ClearOpt();
    ClearOptFull();
    mSize[X] = sizeX; mSize[Y] = sizeY; mSize[Z] = sizeZ;
    int zLayerBaseMin = sizeZ * zMin;
    int zLayerBaseMax = sizeZ * zMax; 
    if (zLayerBaseMax == sizeZ) zLayerBaseMax --;
    mTargetValue255 = mTargetValue * 255;
    
    unsigned char *pDataLink = pData;
    unsigned char *pDataTmp;

    if (bTrimEdge == TRUE){ // E0ɂ
	pDataTmp = new unsigned char [sizeX * sizeY * sizeZ];
	memcpy(pDataTmp, pData, sizeX * sizeY * sizeZ);
	int offsetY = 0;
	int offsetZ = 0;
	for (int z = 0; z < sizeZ; z+= sizeZ - 1){
	    offsetZ = z * sizeX * sizeY;
	    for (int y = 0; y < sizeY; y++){
		offsetY = offsetZ + y * sizeX;
		for (int x = 0; x < sizeX; x++) pDataTmp[offsetY + x] = 0;
	    } }
	for (int y = 0; y < sizeY; y+= sizeY - 1){
	    offsetY = y * sizeX;
	    for (int z = 0; z < sizeZ; z++){
		offsetZ = offsetY + z * sizeX * sizeY;
		for (int x = 0; x < sizeX; x++) pDataTmp[offsetZ + x] = 0;
	    } }
	for (int x = 0; x < sizeX; x+= sizeX - 1)
	    for (int z = 0; z < sizeZ; z++){
		offsetZ = z * sizeX * sizeY;
		for (int y = 0; y < sizeY; y++){
		    offsetY = offsetZ + y * sizeX;
		    pDataTmp[offsetY + x] = 0;
		} }
	pDataLink = pDataTmp;
    }
    
    // őlƍŏl𓾂
    if (bRecalcMinMax == TRUE){

	int verticesIndex = 0;
	for (int indexZ = 0; indexZ < 8; indexZ++){
	    int initZ = mSize[Z] * indexZ / 8;
	    int finZ  = mSize[Z] * (indexZ+1) / 8;
	    if (indexZ != 7) finZ++;
	    if (finZ > mSize[Z] - 1) finZ = mSize[Z] - 1; 
	    
	    for (int indexY = 0; indexY < 8; indexY++){
		int initY = mSize[Y] * indexY /8;
		int finY  = mSize[Y] * (indexY+1) /8;
		if (indexY != 7) finY++;
		if (finY > mSize[Y] - 1) finY = mSize[Y] - 1;
		
		for (int indexX = 0; indexX < 8; indexX++){
		    int initX = mSize[X] * indexX / 8;
		    int finX  = mSize[X] * (indexX+1) / 8;
		    if (indexX != 7) finX++;
		    if (finX > mSize[X] - 1) finX = mSize[X] - 1;
		    
		    std::vector < unsigned char> VMinMax = GetMinMaxValue
			(pDataLink, initZ, finZ, initY, finY, initX, finX);
		    mSectionMinValue[indexZ][indexY][indexX] = VMinMax[0];
		    mSectionMaxValue[indexZ][indexY][indexX] = VMinMax[1];
		}
	    }
	}
    }
    
    mVerticesIndex = 0;

    ThreadControlMain
	(pDataLink, bUseSSE, layerFactor, bOptimize, bPrecise);

    if (bOptimize == TRUE)
	OptimizeVerticesMainBlock();
    

    //LOOP_XYZ { mScalePrev[xyz] = 1; }

    if (bTrimEdge == TRUE) delete pDataTmp;
}

void MarchingCube_cls::
OptimizeVerticesMainBlock(){
    if (mbMemoryReadyOpt == FALSE) ClearOpt();

    InitializeThreadCtrl();

    for (unsigned int index = 0; index < mVpVertices.size(); index++){

	if (mVbOptimized[index] == TRUE ||
	    mVAllocSize[index] == 0) continue;

	//1/3x܂ŗ\Allocămۂ̃I[o[wbh҂
	mVVVerticesOpt[index].reserve(mVAllocSize[index] / 3);
	mVVNormalsOpt[index].reserve(mVAllocSize[index] / 3);
	ThreadOptimizeA *pCThread = new ThreadOptimizeA
	    (wxTHREAD_DETACHED, 
	     mVpVertices[index], mVpNormals[index],
	     &mVAllocSize[index],
	     &mVVVerticesOpt[index], &mVVNormalsOpt[index],
	     &mVVVertexIndicesOpt[index],
	     &mVbOptimized[index]);
	WaitForThreadsAndRun(pCThread);
    }

    WaitAllThreadsEnd(); 

    for (unsigned int i = 0; i < mVpVertices.size(); i++)
	mVOptVertexNumOrg[i] = (int)mVVVerticesOpt[i].size();
}

void MarchingCube_cls::
OptimizeVerticesMainBlockFull(){
    if (mbMemoryReadyOptFull == FALSE) ClearOptFull();

    InitializeThreadCtrl();

    for (unsigned int index = 0; index < mVpVertices.size(); index++){

	if (mVbOptimizedFull[index] == TRUE ||
	    mVVVerticesOpt[index].size() == 0) continue;

	//1/3x܂ŗ\Allocămۂ̃I[o[wbh҂
	mVVVerticesOptFull[index].reserve
	    (mVVVerticesOpt[index].size()/ 3);
	mVVNormalsOptFull[index].reserve
	    (mVVNormalsOpt[index].size()/ 3);
	
	ThreadOptimizeFull *pCThread = new ThreadOptimizeFull
	    (wxTHREAD_DETACHED, 
	     &mVVVerticesOpt[index], &mVVNormalsOpt[index],
	     &mVVVertexIndicesOpt[index],
	     mVVVerticesOpt[index].size(),
	     &mVVVerticesOptFull[index], &mVVNormalsOptFull[index],
	     &mVVVertexIndicesOptFull[index], mSize[Y],
	     &mVbOptimizedFull[index]);
	WaitForThreadsAndRun(pCThread);
    }

    WaitAllThreadsEnd(); 

    ClearOpt();
}

void MarchingCube_cls::
OptimizeVertices(){
    // Xbh҂Ƀ{^ƃRRɍēxĂ܂B
    if (GbGLLockForMultithread == TRUE) return; 
    GbGLLockForMultithread = TRUE;
    OptimizeVerticesMainBlock();
    GbGLLockForMultithread = FALSE;
}

void MarchingCube_cls::
OptimizeVerticesFull(){
    // Xbh҂Ƀ{^ƃRRɍēxĂ܂B
    if (GbGLLockForMultithread == TRUE) return; 
    GbGLLockForMultithread = TRUE;
    OptimizeVerticesMainBlockFull();
    GbGLLockForMultithread = FALSE;
}

bool MarchingCube_cls::
DoesUseMapping(){
    if (mpCTexPropLink->mCColorType.mBody == 0) return FALSE;
    if (mpCTexPropLink->mCMappingType.mBody == _TEX_MAPPING_VALUE) return FALSE;
    return mpCTexPropLink->IsUVPatternReady();
}

#define MACRO_PATTERN(pattern, TAG)\
    if (patternType == _TEX_PATTERN_##TAG){				\
	for (unsigned int index = 0; index < mVpNormals.size(); index++){ \
	    int iSize = (int)mVVVerticesOpt[index].size();		\
	    float globPos[3];						\
	    if (mappingCoord == _MAPPING_GLOBAL){			\
		for (int i = 0; i < iSize; i+=3){			\
		    GetGlobalPos(globPos, &mVVVerticesOpt[index][i]);	\
		    mVVUV1DOpt[index].push_back(GetMappingValueGlobal##pattern(globPos, pos)); \
		}							\
		iSize = (int)mVVVerticesOptFull[index].size();		\
		for (int i = 0; i < iSize; i+=3){			\
		    GetGlobalPos(globPos, &mVVVerticesOptFull[index][i]); \
		    mVVUV1DOptFull[index].push_back(GetMappingValueGlobal##pattern(globPos, pos)); \
		}							\
	    }								\
	    else{							\
		for (int i = 0; i < iSize; i+=3)			\
		    mVVUV1DOpt[index].push_back(GetMappingValueLocal##pattern(&mVVVerticesOpt[index][i], localPivot)); \
		iSize = (int)mVVVerticesOptFull[index].size();		\
		for (int i = 0; i < iSize; i+=3)			\
		    mVVUV1DOptFull[index].push_back(GetMappingValueLocal##pattern(&mVVVerticesOptFull[index][i], localPivot)); \
	    }								\
	}								\
    }

#define MACRO_PATTERNB(pattern, TAG)\
    if (patternType == _TEX_PATTERN_##TAG){				\
	for (unsigned int index = 0; index < mVpNormals.size(); index++){ \
	    int iSize = (int)mVVVerticesOpt[index].size();		\
	    float globPos[3];						\
	    if (mappingCoord == _MAPPING_GLOBAL){			\
		for (int i = 0; i < iSize; i+=3){			\
		    GetGlobalPos(globPos, &mVVVerticesOpt[index][i]);	\
		    mVVUV1DOpt[index].push_back(GetMappingValueGlobal##pattern(globPos, pos, mCurrMapAxis)); \
		}							\
		iSize = (int)mVVVerticesOptFull[index].size();		\
		for (int i = 0; i < iSize; i+=3){			\
		    GetGlobalPos(globPos, &mVVVerticesOptFull[index][i]); \
		    mVVUV1DOptFull[index].push_back(GetMappingValueGlobal##pattern(globPos, pos, mCurrMapAxis)); \
		}							\
	    }								\
	    else{							\
		for (int i = 0; i < iSize; i+=3)			\
		    mVVUV1DOpt[index].push_back(GetMappingValueLocal##pattern(&mVVVerticesOpt[index][i], localPivot, mCurrMapAxis)); \
		iSize = (int)mVVVerticesOptFull[index].size();		\
		for (int i = 0; i < iSize; i+=3)			\
		    mVVUV1DOptFull[index].push_back(GetMappingValueLocal##pattern(&mVVVerticesOptFull[index][i], localPivot, mCurrMapAxis)); \
	    }								\
	}								\
    }

void MarchingCube_cls::
MakeUV1D(TexProperty_cls *pCTexProp, double offset[3], double scale[3]){
    int mappingType = pCTexProp->mCMappingType.mBody;
    int patternType = pCTexProp->mCPatternType.mBody;
    int mappingCoord = pCTexProp->mCMappingCoord.mBody;
    float *pos = pCTexProp->mCMappingPos.mpBody;
    float mapscale = pCTexProp->mCMappingScale.mBody;
    float mapturb = pCTexProp->mCMappingTurb.mBody;
    int axis = pCTexProp->mCMappingAxis.mBody;

    size_t size = 0; // VKlʂɂȂĂ0ɂȂ
    for (unsigned int index = 0; index < mVpNormals.size(); index++){
	size += mVVUV1DOpt[index].size();
	size += mVVUV1DOptFull[index].size();
    }

    double rotate[4] = 
	{TexProperty_cls::msRotZ0, TexProperty_cls::msRotX0,
	 TexProperty_cls::msRotY0, TexProperty_cls::msRotZ1};

    double doubleIndex = TexProperty_cls::mspCPovRayLink->mspCDF3SeqCtrlLink->GetRemappedIndexDouble(0);
    if (IsParameterIsNewOne(doubleIndex, mappingType, patternType, axis, mappingCoord, 
			    pos, mapscale, mapturb, offset, scale, rotate) == FALSE && 
	size != 0){
	return;
    }

    // ̊mۗ̈悲Clear@(TCY0̈ꎞxNgƌ)
    for (unsigned int index = 0; index < mVpNormals.size(); index++){
	std::vector<float>().swap(mVVUV1DOpt[index]);
	std::vector<float>().swap(mVVUV1DOptFull[index]);
    }

    float localPivot[3];
    LOOP_XYZ localPivot[xyz] = (pos[xyz] - offset[xyz]) / scale[xyz];

    // FIX_ME!! indexƂɃXbhō̗]n
    MACRO_PATTERN(Agate, AGATE);
    MACRO_PATTERN(Bozo, BOZO);
    MACRO_PATTERN(Checker, CHECKER);
    MACRO_PATTERNB(Cylindrical, CYLINDRICAL);
    MACRO_PATTERN(Dents, DENTS);
    MACRO_PATTERNB(Gradient, GRADIENT);
    MACRO_PATTERN(Granite, GRANITE);
    MACRO_PATTERNB(Hexagon, HEXAGON);
    MACRO_PATTERN(Radial, RADIAL);
    
    CopyCurrToPrev(pCTexProp);
}

void MarchingCube_cls::
GLDraw(double center[3], double scale[3], bool bCullFace, GLuint showType,
       TexProperty_cls *pCTexProp, float critValue){
    
    if (bCullFace == TRUE)
	glEnable(GL_CULL_FACE);

    double offset[3];
    LOOP_XYZ offset[xyz] = center[xyz] - scale[xyz]/2;

    bool bMapping = TRUE; // lʂ̎͒lŃ}bsOPFȂ̂FALSE
    if (pCTexProp->mCMappingType.mBody == _TEX_MAPPING_VALUE) bMapping = FALSE;
    if (bMapping == FALSE)
	SetGLColoring(pCTexProp, critValue);
    else
	SetGLColoring(pCTexProp, 0.5);
    Configure1DMapping(pCTexProp, offset, scale, TRUE, bMapping);
    
    glEnable(GL_NORMALIZE);
    if (DoesUseMapping() == TRUE)
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);

    for (unsigned int index = 0; index < mVpNormals.size(); index++){

	if (mVbOptimizedFull[index] == TRUE){
	    if (mVVVertexIndicesOptFull[index].size() != 0){
		glVertexPointer(3, GL_FLOAT, 0, &mVVVerticesOptFull[index][0]);
		glNormalPointer(GL_FLOAT, 0, &mVVNormalsOptFull[index][0]);
		if (DoesUseMapping() == TRUE)
		    glTexCoordPointer(1, GL_FLOAT, 0, &mVVUV1DOptFull[index][0]);
		glDrawElements(GL_TRIANGLES, (GLsizei)mVVVertexIndicesOptFull[index].size() , GL_UNSIGNED_INT, &mVVVertexIndicesOptFull[index][0]);
	    }
	}
	else if (mVbOptimized[index] == TRUE){
	    if (mVVVertexIndicesOpt[index].size() != 0){
		glVertexPointer(3, GL_FLOAT, 0, &mVVVerticesOpt[index][0]);
		glNormalPointer(GL_FLOAT, 0, &mVVNormalsOpt[index][0]);
		if (DoesUseMapping() == TRUE)
		    glTexCoordPointer(1, GL_FLOAT, 0, &mVVUV1DOpt[index][0]);
		glDrawElements(GL_TRIANGLES, (GLsizei)mVVVertexIndicesOpt[index].size() , GL_UNSIGNED_INT, &mVVVertexIndicesOpt[index][0]);
	    }
	}
	else{ 
	    if (mVAllocSize[index] != 0){
		glVertexPointer(3, GL_FLOAT, 0, mVpVertices[index]);
		glNormalPointer(GL_FLOAT, 0, mVpNormals[index]);
		glDrawArrays(showType, 0, mVAllocSize[index]/3);
	    }
	}
	
    }

    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisable(GL_NORMALIZE);

    glDisable(GL_CULL_FACE);
}

void MarchingCube_cls::
GetStrWaveFrontObj(wxString *pstrOut, bool bOffset, bool bNormal){

    // UNormalɖ߂
    //double scaleOrg[3], scale1[3] = {1, 1, 1};
    //LOOP_XYZ scaleOrg[xyz] = mScalePrev[xyz];
    //RecalcNormal(size1);
    
    // ̒_Triangleőz
    pstrOut->Alloc(2 * (GetVertexNumberOptFull() * (27 + bNormal * 27) + GetVertexNumber()/3 * 25 + 1000));
    
    *pstrOut << wxT("o Isosurface") << ENDLWX;

    // Full̂݁F
    std::vector < unsigned int > VvertexOffset;
    unsigned int offset = 1; // ŏ1
    VvertexOffset.push_back(offset); 
    size_t indexSize = mVpNormals.size();

    std::vector < std::vector < float > >
	*pVVVerticesOptFull = &mVVVerticesOptFull;
    std::vector < std::vector < float > >
	*pVVNormalsOptFull = &mVVNormalsOptFull;

    double vertOffset = 0;
    if (bOffset == TRUE) vertOffset = -0.5;
    for (size_t index = 0; index < indexSize; index++){
	size_t verticesSize = mVVVerticesOptFull[index].size();
	if (verticesSize == 0){ 
	    VvertexOffset.push_back(offset);
	    continue; }
	float *pVertex = &((*pVVVerticesOptFull)[index][0]);
	for (size_t v = 0; v < verticesSize; v+= 3){
	    *pstrOut << wxT("v ")
		     << *(pVertex + v + X) + vertOffset << wxT(" ")
		     << *(pVertex + v + Y) + vertOffset << wxT(" ")
		     << *(pVertex + v + Z) + vertOffset << ENDLWX;
	}
	offset += (int)verticesSize / 3;
	VvertexOffset.push_back(offset);
    }

    if (bNormal == TRUE){
	for (size_t index = 0; index < indexSize; index++){
	    size_t verticesSize = mVVVerticesOptFull[index].size();
	    if (verticesSize == 0) {
		VvertexOffset.push_back(offset);
		continue;
	    }
	    float *pNormal = &((*pVVNormalsOptFull)[index][0]);
	    for (size_t v = 0; v < verticesSize; v+= 3){
		*pstrOut << wxT("vn ")
			 << *(pNormal + v + X) << wxT(" ")
			 << *(pNormal + v + Y) << wxT(" ")
			 << *(pNormal + v + Z) << ENDLWX;
	    }
	    offset += (int)verticesSize / 3;
	    VvertexOffset.push_back(offset);
	}
    }

    *pstrOut << wxT("s off") << ENDLWX;

    for (unsigned int index = 0; index < indexSize; index++){
	size_t vertexIndexSize = mVVVertexIndicesOptFull[index].size();
	for (size_t v = 0; v < vertexIndexSize; v+= 3)
	    *pstrOut << wxT("f ")
		     << mVVVertexIndicesOptFull[index][v+X] + VvertexOffset[index] << wxT(" ")
		     << mVVVertexIndicesOptFull[index][v+Y] + VvertexOffset[index] << wxT(" ")
		     << mVVVertexIndicesOptFull[index][v+Z] + VvertexOffset[index] << ENDLWX;
    }

    //RecalcNormal(scaleOrg);
}

void MarchingCube_cls::
GetStrPovMesh2(wxString *pstrOut, wxString strID, wxString strTag, bool bNormal){

    // ̒_Triangleőz
    // wideȂ̂2H
    pstrOut->Alloc(2 * (GetVertexNumberOptFull() * 40 + GetVertexNumber()/3 * 40 + 1000));
    
    std::vector < unsigned int > VvertexOffset;
    unsigned int offset = 0;
    VvertexOffset.push_back(offset); // ŏ0
    size_t indexSize = mVpNormals.size();
    
    *pstrOut << wxT("//Mesh2file generated by ") << _FRAME_MAIN_APPNAME << ENDLWX;

    *pstrOut << wxT("//normal_start_at=");
    int posNormalTag = pstrOut->Len();
    *pstrOut << wxT("##########") << ENDLWX;

    *pstrOut << wxT("//index_start_at=");
    int posIndexTag = pstrOut->Len();
    *pstrOut << wxT("##########") << ENDLWX;

    *pstrOut << wxT("#declare ") << wxT("ISOSURF_") << strID << strTag << L"=mesh2{" << ENDLWX
	     << wxT("\tvertex_vectors{") << ENDLWX
	     << wxT("\t\t") << GetVertexNumberOptFull() << wxT (",") << ENDLWX;
    
    for (size_t index = 0; index < indexSize; index++){
	size_t verticesSize = mVVVerticesOptFull[index].size();
	offset += (int)verticesSize / 3;
	VvertexOffset.push_back(offset);
    }

    wxString strOutTriangles;
    ThreadExportMesh2PovSubI CThreadSub(&strOutTriangles, indexSize,
					GetVertexNumber()/3,
					&mVVVertexIndicesOptFull,
					&VvertexOffset);
    CThreadSub.Run();

    wxString strOutNormal;
    ThreadExportMesh2PovSubN CThreadSubN(&strOutNormal, indexSize,
					 GetVertexNumberOptFull(),
					 &mVVNormalsOptFull,
					 bNormal);

    if (bNormal) CThreadSubN.Run();
    
    std::vector < std::vector < float > >
	*pVVVerticesOptFull = &mVVVerticesOptFull;
    std::vector < std::vector < float > >
	*pVVNormalsOptFull = &mVVNormalsOptFull;
    
    wxString strVertices;
    strVertices.Alloc(2 * (GetVertexNumberOptFull() * 40));

    for (size_t index = 0; index < indexSize; index++){
	size_t verticesSize = mVVVerticesOptFull[index].size();
	size_t verticesNum = verticesSize / 3;
	if (verticesSize == 0) continue;
	float *pVertex = &((*pVVVerticesOptFull)[index][0]);
	for (size_t v = 0; v < verticesNum; v++){
	    char buf[64];
	    if (v % 3 == 0 && v == verticesNum - 1)
		sprintf(buf, "\t\t<%0.5f,%0.5f,%0.5f>,\n",
			*(pVertex + v*3 + X),
			*(pVertex + v*3 + Y),
			*(pVertex + v*3 + Z));
	    else if (v % 3 == 2)
		sprintf(buf, "<%0.5f,%0.5f,%0.5f>,\n",
			*(pVertex + v*3 + X),
			*(pVertex + v*3 + Y),
			*(pVertex + v*3 + Z));
	    else if (v % 3 == 0)
		sprintf(buf, "\t\t<%0.5f,%0.5f,%0.5f>,",
			*(pVertex + v*3 + X),
			*(pVertex + v*3 + Y),
			*(pVertex + v*3 + Z));
	    else
		sprintf(buf, "<%0.5f,%0.5f,%0.5f>,",
			*(pVertex + v*3 + X),
			*(pVertex + v*3 + Y),
			*(pVertex + v*3 + Z));
	    strVertices << wxString::From8BitData(buf);
	}
    }
    strVertices.RemoveLast(); //strVertices.RemoveLast();
    *pstrOut << strVertices << ENDLWX 
	     << wxT("\t}") << ENDLWX;

    int posNormal = pstrOut->Len();

    CThreadSub.Wait();
   
    if (bNormal) {
	CThreadSubN.Wait();
	*pstrOut << strOutNormal << ENDLWX
		 << wxT("\t}") << ENDLWX;
    }
    
    int posIndex = pstrOut->Len();

    *pstrOut += strOutTriangles;

    // inside_vector͈Ȃi|SƓjȂ牽łǂ
    *pstrOut << ENDLWX 
	     << wxT("\t}") << ENDLWX
	     << wxT("\tinside_vector <0.01, 0.01, 0.99>") << ENDLWX
	     << ENDLWX << wxT("}") << ENDLWX;
				 
    wxString strNormalTag = wxString::Format(wxT("%010d"), posNormal);
    wxString strIndexTag = wxString::Format(wxT("%010d"), posIndex);

    for (int i = 0; i < 10; i++)
	pstrOut->SetChar(posNormalTag + i, strNormalTag.GetChar(i));
    for (int i = 0; i < 10; i++)
	pstrOut->SetChar(posIndexTag + i, strIndexTag.GetChar(i));
}

