// "clsMarchingCube.cpp"
// (setq c-basic-offset 4)

#ifndef _ISOSURFACE_MARCHINGCUBE_H_
#define _ISOSURFACE_MARCHINGCUBE_H_

#include "../../COMMON/clsMultiThreadCtrl.h"
#include "../../MYGL/clsTextureProperty.h"
#include "../../MYGL/clsTextureMapping.h"

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include <vector>

class ThreadExportMesh2PovSubI : public wxThread {
 public:
    ThreadExportMesh2PovSubI
	(wxString *pstrOut, unsigned int indexSize,
	 unsigned int polygonNumber,
	 std::vector < std::vector < unsigned int > > *pVVVertexIndicesOptFull,
	 std::vector < unsigned int > *pVvertexOffset);
    
    wxString *mpStrOut;
    unsigned int mIndexSize;
    unsigned int mPolygonNumber;
    std::vector < std::vector < unsigned int > > *mpVVVertexIndicesOptFull;
    std::vector < unsigned int > *mpVvertexOffset;

    virtual ExitCode Entry();
};

class ThreadExportMesh2PovSubN : public wxThread {
 public:
    ThreadExportMesh2PovSubN
	(wxString *pstrOut, unsigned int indexSize,
	 unsigned int normalNumber,
	 std::vector < std::vector < float > > *pVVNormalsOptFull,
	 bool bValid);
    
    wxString *mpStrOut;
    unsigned int mIndexSize;
    unsigned int mNormalNumber;
    std::vector < std::vector < float > > *mpVVNormalsOptFull;
    //std::vector < unsigned int > *mpVnormalOffset;

    virtual ExitCode Entry();
};

class MarchingCube_cls : public TextureMapping_cls, public MultiThreadCtrl_cls {
 public:
    MarchingCube_cls();
    ~MarchingCube_cls();

    unsigned char mSectionMinValue[8][8][8];
    unsigned char mSectionMaxValue[8][8][8];
    
    // f̃|Sf[^
    std::vector < float* > mVpVertices;
    std::vector < float* > mVpNormals;
    std::vector < char > mVbOptimized; // vector<bool>͎[@Ⴄ̂őp
    std::vector < int > mVAllocSize;
    std::vector < int > mVVertexNumOrg;

    // yx̍œKꂽ|Sf[^
    std::vector < std::vector < float > > mVVVerticesOpt;
    std::vector < std::vector < float > > mVVNormalsOpt;
    std::vector < std::vector < float > > mVVUV1DOpt;
    std::vector < std::vector < unsigned int > > mVVVertexIndicesOpt;
    std::vector < char > mVbOptimizedFull; // vector<bool>͎[@Ⴄ̂őp
    std::vector < unsigned int > mVOptVertexNumOrg;
    bool mbMemoryReadyOpt;
    
    // tɍœKꂽ|Sf[^
    std::vector < std::vector < float > > mVVVerticesOptFull;
    std::vector < std::vector < float > > mVVNormalsOptFull;
    std::vector < std::vector < float > > mVVUV1DOptFull;
    std::vector < std::vector < unsigned int > > mVVVertexIndicesOptFull;
    bool mbMemoryReadyOptFull;

    int mSize[3];
    int mVerticesIndex;

    float mTargetValue, mTargetValue255;
    
    void Clear(int sizeZ);
    void ClearOpt();
    void ClearOptFull();
    unsigned int GetVertexNumber();
    unsigned int GetVertexNumberOpt();
    int   GetVertexNumberOptFull();
    float GetVertexDensity();
    float GetMaxVertexDensity(int layerFactor);

    bool AllocateBuffer(float **ppTmpVertices, float **ppTmpNormal,
			int bufferSize);
    bool ThreadControlMain
	(unsigned char *pData, bool bUseSSE,
	 const int layerFactor, bool bOptimize, bool bPrecise);

    std::vector < unsigned char > GetMinMaxValue
	(unsigned char *pData, 
	 int initZ, int finZ, int initY, int finY, int initX, int finX);
    void ConstructVertices
	(unsigned char *pData, int sizeX, int sizeY, int sizeZ,
	 double zMin, double zMax,
	 bool bUseSSE, const int layerFactor,
	 bool bRecalcMinMax,
	 bool bOptimize, bool bTrimEdge, bool bPrecise);

    void OptimizeVerticesMainBlock();
    void OptimizeVerticesMainBlockFull();
    void OptimizeVertices();
    void OptimizeVerticesFull();

    bool DoesUseMapping();
    /*
    void GetGlobalPos
      (float globPos[3], std::vector < float > *pVVertices, int i,
       double offset[3], double scale[3]);
    */
    void MakeUV1D
	(TexProperty_cls *pCTexProp,
       /*int mappingType, int patternType, int mappingCoord,
       float pos[3], float mapscale, float mapturb,
       */
	 double offset[3], double scale[3]);

    void GLDraw(double center[3], double scale[3], bool bCullFace, 
		GLuint showType,
		TexProperty_cls *pCTexProp, float critValue);

    void GetStrWaveFrontObj(wxString *pstrOut, bool bOffset, bool bNormal);
    void GetStrPovMesh2(wxString *pstrOut, wxString strID, wxString strTag,
			bool bNormal);
}; 

#endif

