// -*- c -*- 
// (setq c-basic-offset 4)
// "clsPovFocalBlur.h"

#include "clsPovFocalBlur.h"

#include "../clsPovRay.h"
#include "../../../wxWidgetEnum.h"
#include "../../WIDGETCTRL/clsDialogEXKey.h"
#include "../../GLCANVAS/clsMainGLCanvas.h"

#define ENDLWX wxT("\n")

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

void PovFocalBlur_cls::
SetToDefault(){
    mCConfList.Clear();

    mCConfList.Add(&mCType, wxT("PovFocalBlur="), L"0");
    mCConfList.Add(&mCFocalBlurScale, wxT("PovFocalBlurScale="), L"0");
    mCConfList.Add(&mCFocalBlurQuality, wxT("PovFocalBlurQuality="), L"1");
    mCConfList.Add(&mCFocusLocationType, wxT("PovFocusLocationType="), L"0");
    mCConfList.Add(&mCFocusDistance, wxT("PovFocusDistance="), L"1");
    mCConfList.Add(&mCFocusLocation, wxT("PovFocusLocation="), L"0 0 0");
}

void PovFocalBlur_cls::
SetMNumFocalBlur(wxBoxSizer *pBoxSizer, wxWindow* pCParentPanel){
    MultiNumericalInputsPanel_cls* pCurr = 
	wxDialogEXKey::InitializeMNumPanel(&mpCMNumFocalBlur, pCParentPanel);
    /*
    pCurr->AssignMinimizeButtonBoth(mpCBTNFocalBlur, btnPovFocalBlur,
				    _("[+]Focal Blur Settings"));
    */

    pCurr->mTextWidth     = -1;
    pCurr->mComboWidth = 108;

    wxString focalBlurType[2] =
	{_("No Focal Blur"), _("Use Focal Blur") };
    mIndexType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mCType.mBody,
	 _("Focal Blur"), TRUE, 2, focalBlurType);

    int boxWidth = 2;
    mIndexScale = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("scale"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mCFocalBlurScale.mBody);
    pCurr->SetRange(0, 500, _MN_LASTITEM, 1000);
    pCurr->CombinePointer(_GP_FLOAT, &mCFocalBlurScale.mBody);
    pCurr->SetRange(0, 500, _MN_LASTITEM, 1000);

    mIndexQuality = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Quality"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &mCFocalBlurQuality.mBody);
    pCurr->SetRange(0, 5);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mCFocalBlurQuality.mBody);
    pCurr->SetRange(0, 5);

    wxString focusLocationType[5] =
	{_("Origin"), _("Anchor Pos"), _("Stereo Focus"), _("Distance"),
	 _("Numeric") };
    mIndexFocusLocationType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mCFocusLocationType.mBody,
	 _("Focus location"), TRUE, 5, focusLocationType);

    mIndexFocusDistance = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("distance"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mCFocusDistance.mBody);
    pCurr->SetRange(0, 1000, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mCFocusDistance.mBody);
    pCurr->SetRange(0, 1000, _MN_LASTITEM, 100);

    boxWidth = 3;
    mIndexFocusLocation = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("location"));
    pCurr->CombinePointer(_GP_FLOAT, &mCFocusLocation.mpBody[0]);
    pCurr->CombinePointer(_GP_FLOAT, &mCFocusLocation.mpBody[1]);
    pCurr->CombinePointer(_GP_FLOAT, &mCFocusLocation.mpBody[2]);

    #ifdef WIN32
    pCurr->mMinimumWidth = 320;
#else
    pCurr->mMinimumWidth = 340;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(&PovRay_cls::HookFocalBlur, _MN_HOOK_COMBOBOX);
    
    HideAndShow();
}

void PovFocalBlur_cls::
HideAndShow(){
    MultiNumericalInputsPanel_cls *pCurr = mpCMNumFocalBlur;
    bool bShow = (mCType.mBody > _POV_FOCALBLUR_NOUSE);
    
    pCurr->ShowItems(mIndexScale, 3, bShow);
    pCurr->ShowItems(mIndexQuality, 3, bShow);
    pCurr->ShowItems(mIndexFocusLocationType, 1, bShow);

    bool bDistance = (mCFocusLocationType.mBody == _POV_FOCUSPOS_DISTANCE);
    pCurr->ShowItems(mIndexFocusDistance, 3, bShow & bDistance);

    bool bLocation = (mCFocusLocationType.mBody == _POV_FOCUSPOS_NUMERIC);
    pCurr->ShowItems(mIndexFocusLocation, 4, bShow & bLocation);

    pCurr->Layout();
}

wxString PovFocalBlur_cls::
GetPovStrFocalBlur(double activeAnchorPos[3], double apatureFactor){
    if (mCType.mBody == _POV_FOCALBLUR_NOUSE) return wxT("");
    wxString strFocalBlur;

    strFocalBlur << wxT("\taperture ") << mCFocalBlurScale.mBody * apatureFactor << ENDLWX;
    
    // FIX_ME!! Ad_Hoc
    int quality = mCFocalBlurQuality.mBody;
    int blurSamples = 8 * quality * quality;
    //float confidence = 0.86 + quality * 0.03;
    int varianceInv = 128 + 256 * quality * (quality - 1);
    if (quality == 0) varianceInv = 64;
    if (blurSamples == 0) blurSamples = 3;
    //if (confidence >= 0.99) confidence = 0.99;
    float confidence = 1.0f - 1.0f/varianceInv;
    strFocalBlur << wxT("\tblur_samples ") << blurSamples << ENDLWX;
    strFocalBlur << wxT("\tconfidence ") << confidence << ENDLWX;
    strFocalBlur << wxT("\tvariance 1/") << varianceInv << ENDLWX;

    double anchorPos[3] = {0, 0, 0};
    switch(mCFocusLocationType.mBody){
    case _POV_FOCUSPOS_ORIGIN:
	break;
    case _POV_FOCUSPOS_ANCHORPOS: 
    case _POV_FOCUSPOS_STEREOFOCUS:
    case _POV_FOCUSPOS_DISTANCE:
	LOOP_XYZ anchorPos[xyz] = activeAnchorPos[xyz];
	break; 
    case _POV_FOCUSPOS_NUMERIC:
	LOOP_XYZ anchorPos[xyz] = mCFocusLocation.mpBody[xyz];
	break;
    }

    strFocalBlur << PovSerialize_cls::PovOut3d(wxT("\tfocal_point "), anchorPos) << ENDLWX;
    
    return strFocalBlur;
}
    
void PovFocalBlur_cls::
ReadStrLine(std::wstring strLine){
    mCConfList.ReadStrLineAll(strLine);
}

wxString PovFocalBlur_cls::
SerializeConf(){
    return mCConfList.SerializeAll();
}
