// "clsPovStreamLine.h"
// (setq c-basic-offset 4)

#include "clsPovStreamLine.h"

#include "../../DF3/clsDF3SeqCtrl.h"
#include "../../DF3/clsDF3SeqGUI.h"
#include "../../DF3/clsDF3Sequence.h"
#include "../../DF3/clsDF3.h"
#include "../../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../../FILTER/clsFilterCtrl.h"
#include "../../MAINCANVAS/clsFilterPanels.h"
#include "../../MYAPI/MsgBox.h"

#include "../../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../../WIDGETCTRL/EX/clsColourDialogEX.h"

#include "../../../clsDrawingBox.h"

#include "../../../wxWidgetEnum.h"

#include <wx/wx.h>
#include <wx/wfstream.h> 

#ifdef WIN32
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define R 0
#define G 1
#define B 2
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)
#define ENDLWX wxT("\n")
#define ENDL L"\n"

ThreadPovoutStream::
ThreadPovoutStream(wxThreadKind kind, PovStreamLine_cls *pCParent)
    : MultiThreadThread_cls(kind)
{
    mpCParent = pCParent;
}

void ThreadPovoutStream::
SetParameters(int currentFrame, int currMaxFrame,
	      int clrType, double *pScale,
	      int lineNumber, unsigned int *pIndexForDraw,
	      float **ppPos, float *pColorCurvePov, float width,
	      bool bInside, wxString *pstrTarget) {
    mCurrentFrame = currentFrame;  mCurrMaxFrame = currMaxFrame;
    mClrType = clrType;
    mLineNumber = lineNumber;
    mpIndexForDraw = pIndexForDraw;
    mppPos = ppPos;
    mpScale = pScale;
    mpColorCurvePov = pColorCurvePov;
    mWidth = width;
    mpstrTarget = pstrTarget;
    mbInside = bInside;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

void ThreadPovoutStream::
PushBackStreamLineElements0(wxString *pstrTgt,
			    float *pPos0, float *pPos1, 
			    wxString strWidth, bool bFinish){
    if (bFinish == FALSE)
	PushBackPovOut3f(pstrTgt, wxT("\tTUBE("), pPos0, 6);
    else
	PushBackPovOut3f(pstrTgt, wxT("\tTUBEFIN("), pPos0, 6);
    PushBackPovOut3f(pstrTgt, wxT(","), pPos1, 6);
    *pstrTgt << wxT(",") << strWidth << wxT(")") << ENDLWX;
}

void ThreadPovoutStream::
PushBackStreamLineElements0Intersect(wxString *pstrTgt,
				     float *pPos0, float *pPos1, 
				     wxString strWidth, bool bFinish){

    *pstrTgt << wxT("\tintersection {") << ENDLWX;
    if (bFinish == FALSE)
	PushBackPovOut3f(pstrTgt, wxT("\t\tTUBE("), pPos0, 6);
    else
	PushBackPovOut3f(pstrTgt, wxT("\t\tTUBEFIN("), pPos0, 6);
    PushBackPovOut3f(pstrTgt, wxT(","), pPos1, 6);
    *pstrTgt << wxT(",") << strWidth << wxT(")") << ENDLWX;
    *pstrTgt << wxT("\t\tbox { <0, 0, 0>, <1, 1, 1> } }") << ENDLWX;
}

void ThreadPovoutStream::
PushBackStreamLineElements1(wxString *pstrTgt,
			    float *pPos0, float *pPos1, float *pClr,
			    wxString strWidth, bool bFinish){
    if (bFinish == FALSE)
	PushBackPovOut3f(pstrTgt, wxT("\tTUBECLR("), pPos0, 6);
    else
	PushBackPovOut3f(pstrTgt, wxT("\tTUBECLRFIN("), pPos0, 6);
    PushBackPovOut3f(pstrTgt, wxT(","), pPos1, 6);
    *pstrTgt << wxT(",") << strWidth << wxT(",");
    PushBackPovOut3f(pstrTgt, wxT(""), pClr, 3);
    *pstrTgt << wxT(")") << ENDLWX;
}

void ThreadPovoutStream::
PushBackStreamLineElements1Intersect(wxString *pstrTgt,
				     float *pPos0, float *pPos1, float *pClr,
				     wxString strWidth,
				     wxString strScaleBox, bool bFinish){

    *pstrTgt << wxT("\tintersection {") << ENDLWX;
    if (bFinish == FALSE)
	PushBackPovOut3f(pstrTgt, wxT("\t\tTUBE("), pPos0, 6);
    else
	PushBackPovOut3f(pstrTgt, wxT("\t\tTUBEFIN("), pPos0, 6);
    PushBackPovOut3f(pstrTgt, wxT(","), pPos1, 6);
    *pstrTgt << wxT(",") << strWidth << wxT(")") << ENDLWX;
    *pstrTgt << wxT("\t\tbox { <0, 0, 0>,") << strScaleBox << wxT("}") << ENDLWX;
    PushBackPovOut3f(pstrTgt, wxT("\t\tTEXPIG("), pClr, 3);
    *pstrTgt << wxT(")") << ENDLWX << wxT("\t}") << ENDLWX;
}

int ThreadPovoutStream::
CheckWithinBox(float pos0[3], float pos1[3],
	       float width, float oneMinusWidth, float onePlusWidth){
    int type = 0;
    if (pos0[X] > width && pos0[X] < oneMinusWidth &&
	pos0[Y] > width && pos0[Y] < oneMinusWidth &&
	pos0[Z] > width && pos0[Z] < oneMinusWidth) type++;
    if (pos1[X] > width && pos1[X] < oneMinusWidth &&
	pos1[Y] > width && pos1[Y] < oneMinusWidth &&
	pos1[Z] > width && pos1[Z] < oneMinusWidth) type++;
    if (pos0[X] > -width && pos0[X] < onePlusWidth &&
	pos0[Y] > -width && pos0[Y] < onePlusWidth &&
	pos0[Z] > -width && pos0[Z] < onePlusWidth) type++;
    if (pos1[X] > -width && pos1[X] < onePlusWidth &&
	pos1[Y] > -width && pos1[Y] < onePlusWidth &&
	pos1[Z] > -width && pos1[Z] < onePlusWidth) type++;
    
    return type; // intersect̐
}


void ThreadPovoutStream::
PushBackSingleStreamLine(wxString *pstrTgt, 
			 float *pPos, float *pClr, int clrType, float width,
			 int currentFrame, double scale[3]){
    wxString strStreamLine;
    
    int maxFrame;
    if (mpCParent->GetPCData()->mCAnimation.mBody == 0 && mbInside)
	maxFrame = mCurrMaxFrame;
    else
	maxFrame = currentFrame - mpCParent->GetPCData()->mCInitAnim.mBody;
    if (maxFrame > mCurrMaxFrame) maxFrame = mCurrMaxFrame;
    if (maxFrame < 0) return;

    if (clrType == 0) 
	strStreamLine.Alloc(maxFrame * 70);
    else
	strStreamLine.Alloc(maxFrame * 120);
	
    float oneMinusWidth = 1.0f - width;
    float onePlusWidth = 1.0f + width;
    
    wxString strWidth = GetShortString(width);
    int skip = mpCParent->GetPCData()->mCSkip.mBody;
    wxString strScaleBox;
    strScaleBox << PovOut3d(wxT(""), scale);
    
    int indexIni = currentFrame % skip;
    int lastIndex = 0,indexMax = maxFrame - skip;
    
    float pos0[3], pos1[3];
    if (indexIni != 0 && indexIni < indexMax){
	LOOP_XYZ pos0[xyz] = pPos[xyz];
	LOOP_XYZ pos1[xyz] = pPos[indexIni*3 + xyz];
	int drawType = CheckWithinBox(pos0, pos1, 
				      width, oneMinusWidth, onePlusWidth);
	LOOP_XYZ pos0[xyz] *= scale[xyz];
	LOOP_XYZ pos1[xyz] *= scale[xyz];

	switch(drawType){
	case 4:
	    if (clrType == 0)
		PushBackStreamLineElements0(&strStreamLine, pos0, pos1, strWidth);
	    else 
		PushBackStreamLineElements1(&strStreamLine, pos0, pos1, &mpColorCurvePov[0], strWidth);
	    break;
	case 3: case 2: case 1:
	    if (clrType == 0)
		PushBackStreamLineElements0Intersect(&strStreamLine, pos0, pos1, strWidth);
	    else 
		PushBackStreamLineElements1Intersect(&strStreamLine, pos0, pos1, &mpColorCurvePov[0], strWidth, strScaleBox);
	    break;
	}
    }

    // [v̍Đ[Ȃ̂ŏo邾
    int skip3 = skip * 3;
    double scaleX = scale[X], scaleY = scale[Y], scaleZ = scale[Z];
    if (clrType == 0)
	for (int index = indexIni; index < indexMax; index+=skip){
	    int index3 = index*3;
	    int index3next = index3 + skip3;
	    pos0[X] = pPos[index3 + X];
	    pos0[Y] = pPos[index3 + Y];
	    pos0[Z] = pPos[index3 + Z];
	    pos1[X] = pPos[index3next + X];
	    pos1[Y] = pPos[index3next + Y];
	    pos1[Z] = pPos[index3next + Z];
	    int drawType = CheckWithinBox(pos0, pos1, 
					  width, oneMinusWidth, onePlusWidth);
	    pos0[X] *= scaleX; pos0[Y] *= scaleY; pos0[Z] *= scaleZ;
	    pos1[X] *= scaleX; pos1[Y] *= scaleY; pos1[Z] *= scaleZ;
	    
	    switch(drawType){
	    case 4:
		PushBackStreamLineElements0
		    (&strStreamLine, pos0, pos1, strWidth);
		break;
	    case 3: case 2: case 1:
		PushBackStreamLineElements0Intersect
		    (&strStreamLine, pos0, pos1, strWidth);
		break;
	    }
	    lastIndex = index;
	}
    else
	for (int index = indexIni; index < indexMax; index+=skip){
	    int index3 = index*3;
	    int index3next = index3 + skip3;
	    pos0[X] = pPos[index3 + X];
	    pos0[Y] = pPos[index3 + Y];
	    pos0[Z] = pPos[index3 + Z];
	    pos1[X] = pPos[index3next + X];
	    pos1[Y] = pPos[index3next + Y];
	    pos1[Z] = pPos[index3next + Z];
	    int drawType = CheckWithinBox(pos0, pos1, 
					  width, oneMinusWidth, onePlusWidth);
	    pos0[X] *= scaleX; pos0[Y] *= scaleY; pos0[Z] *= scaleZ;
	    pos1[X] *= scaleX; pos1[Y] *= scaleY; pos1[Z] *= scaleZ;
	    
	    switch(drawType){
	    case 4:
		PushBackStreamLineElements1
		    (&strStreamLine, pos0, pos1, 
		     &mpColorCurvePov[index*3], strWidth);
		break;
	    case 3: case 2: case 1:
		PushBackStreamLineElements1Intersect
		    (&strStreamLine, pos0, pos1, 
		     &mpColorCurvePov[index*3], strWidth, strScaleBox);
		break;
	    }
	    lastIndex = index;
	}
    
    if (lastIndex != maxFrame - 1 &&
	maxFrame != 0){
	LOOP_XYZ pos0[xyz] = pPos[lastIndex*3 + xyz];
	LOOP_XYZ pos1[xyz] = pPos[(maxFrame-1)*3 + xyz];
	int drawType = CheckWithinBox(pos0, pos1, 
				      width, oneMinusWidth, onePlusWidth);
	LOOP_XYZ pos0[xyz] *= scale[xyz];
	LOOP_XYZ pos1[xyz] *= scale[xyz];

	if (clrType == 0)
	    switch(drawType){
	    case 4:
		PushBackStreamLineElements0
		    (&strStreamLine, pos0, pos1, strWidth, TRUE);
		break;
	    case 3: case 2: case 1:
		PushBackStreamLineElements0Intersect
		    (&strStreamLine, pos0, pos1, strWidth, TRUE);
		break;
	    }
	else
	    switch(drawType){
	    case 4:
		PushBackStreamLineElements1
		    (&strStreamLine, pos0, pos1, 
		     &mpColorCurvePov[lastIndex*3], strWidth, TRUE);
		break;
	    case 3: case 2: case 1:
		PushBackStreamLineElements1Intersect
		    (&strStreamLine, pos0, pos1, 
		     &mpColorCurvePov[lastIndex*3], strWidth, strScaleBox, TRUE);
		break;
	    }
    }
    
    *pstrTgt << strStreamLine;
}

wxThread::ExitCode ThreadPovoutStream::
Entry(){

    for (int i = 0; i < mLineNumber; i++){
	unsigned int index = mpIndexForDraw[i];
	float *pPos = mppPos[index];
	PushBackSingleStreamLine
	    (mpstrTarget, pPos, mpColorCurvePov, mClrType, mWidth,
	     mCurrentFrame, mpScale);
    }
	
    *mpbReadyFlag = TRUE;
    return 0;
}

// -------------------------------------

ThreadPovoutPoint::
ThreadPovoutPoint(wxThreadKind kind, PovStreamLine_cls *pCParent)
    : MultiThreadThread_cls(kind)
{
    mpCParent = pCParent;
}

void ThreadPovoutPoint::
SetParameters(int currentFrame, int currMaxFrame,
	      int lineNumber, int lineIni, double *pScale,
	      float **ppPos, unsigned char *pDensity,
	      int sizeX, int sizeY, int sizeZ,
	      float pointSize, double *pCamPosLocal,
	      wxString *pstrTarget){
    
    mCurrentFrame = currentFrame; mCurrMaxFrame = currMaxFrame;
    mLineNumber = lineNumber;
    mLineIni = lineIni;
    mpstrTarget = pstrTarget;
    mpScale = pScale;
    mppPos = ppPos;
    mpDensity = pDensity;
    mSizeX = sizeX; mSizeY = sizeY; mSizeZ = sizeZ;
    mPointSize = pointSize;
    mpCamPos = pCamPosLocal;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

bool ThreadPovoutPoint::
CheckWithinBoxPoint(float pos[3]){
    if (pos[X] < 0 || pos[X] > 1 ||
	pos[Y] < 0 || pos[Y] > 1 ||
	pos[Z] < 0 || pos[Z] > 1) return FALSE;
    return TRUE;
}

wxThread::ExitCode ThreadPovoutPoint::
Entry(){

    wxString strPointSize = GetShortString(mPointSize);
    double scaleX = mpScale[X];
    double scaleY = mpScale[Y];
    double scaleZ = mpScale[Z];
    
    int pointType = mpCParent->GetPCData()->mCPointType.mBody;
    int pointNumber = mpCParent->GetPCData()->mCPointNumber.mBody;
    float nominalInterval = mCurrMaxFrame / pointNumber;
    float *pRandTable = mpCParent->mpRandTable;
    int offsetZ = mSizeX * mSizeY;
    
    (*mpstrTarget) << wxT("//") << mpCamPos[X] << wxT(",") << mpCamPos[Y] << wxT(",") << mpCamPos[Z] << ENDLWX;
    (*mpstrTarget) << wxT("//") << scaleX << wxT(",") << scaleY << wxT(",") << scaleZ << ENDLWX;

    for (unsigned int i = 0; i < mLineNumber; i++){
	float *pPos = mppPos[i];
	int ix = pPos[X] * mSizeX;
	int iy = pPos[Y] * mSizeY;
	int iz = pPos[Z] * mSizeZ;
	unsigned char density;
	if (mpDensity != NULL)
	    density = mpDensity[iz * offsetZ + iy * mSizeX + ix];
	else
	    density = 16;
	float factor = /*16384.0*/8.0f / (density + 0.01);
	int interval = nominalInterval * factor;
	if (interval <= 30) interval = 30;
	int initframe = (mCurrentFrame + (int)(pRandTable[i+mLineIni] / RAND_MAX * interval)) % interval;
	float pointPos[3];
	
	for (int frame = initframe; frame < mCurrMaxFrame; frame+=interval){
	    pointPos[X] = pPos[frame * 3 + X];
	    pointPos[Y] = pPos[frame * 3 + Y];
	    pointPos[Z] = pPos[frame * 3 + Z];
	    if (CheckWithinBoxPoint(pointPos) == FALSE) continue;

	    double distance = 
		sqrt(((mpCamPos[X] + 0.5) - pointPos[X]) * ((mpCamPos[X] + 0.5) - pointPos[X]) * scaleX * scaleX +
		     ((mpCamPos[Y] + 0.5) - pointPos[Y]) * ((mpCamPos[Y] + 0.5) - pointPos[Y]) * scaleY * scaleY +
		     ((mpCamPos[Z] + 0.5) - pointPos[Z]) * ((mpCamPos[Z] + 0.5) - pointPos[Z]) * scaleZ * scaleZ );
		
	    pointPos[X] *= scaleX;
	    pointPos[Y] *= scaleY;
	    pointPos[Z] *= scaleZ;

	    PushBackPovOut3f(mpstrTarget, wxT("\tsphere {"), pointPos);
	    (*mpstrTarget) << wxT(",");

	    float pointSize = mPointSize;
	    if (frame < 9) 
		pointSize *= ((float)frame + 1) / 10;
	    else if (frame > mCurrMaxFrame - 10) 
		pointSize *= ((float)mCurrMaxFrame - frame) / 10;
	    if (pointType == 1)
		pointSize *= distance;
	    else if (pointType == 2)
		pointSize *= sqrt(distance);
	    if (pointSize == mPointSize)
		(*mpstrTarget) << strPointSize;
	    else
		(*mpstrTarget) << pointSize;
	    
	    (*mpstrTarget) << L"}" << ENDLWX;
	}
    }

    *mpbReadyFlag = TRUE;
    return 0;
}

// -------------------------------------

ThreadConstructStream::
ThreadConstructStream(wxThreadKind kind, PovStreamLine_cls *pCParent) 
    : MultiThreadThread_cls(kind)
{
    mCurrMaxFrame = pCParent->mCurrMaxFrame;
    mStepFactorX = pCParent->GetPCData()->mCStepFactor.mBody / pCParent->msABoxScale[X];
    mStepFactorY = pCParent->GetPCData()->mCStepFactor.mBody / pCParent->msABoxScale[Y];
    mStepFactorZ = pCParent->GetPCData()->mCStepFactor.mBody / pCParent->msABoxScale[Z];
    mSizeX = pCParent->mSizeX;
    mSizeY = pCParent->mSizeY;
    mSizeZ = pCParent->mSizeZ;
}

void ThreadConstructStream::
FollowField(float initX, float initY, float initZ,
	    float *pStream, 
	    unsigned char *pX, unsigned char *pY, unsigned char *pZ){
    
    pStream[X] = initX;
    pStream[Y] = initY;
    pStream[Z] = initZ;

    mOffsetZ = mSizeX * mSizeY;
    mOffsetY = mSizeX;

    float stepFactorX = mStepFactorX;
    float stepFactorY = mStepFactorY;
    float stepFactorZ = mStepFactorZ;

    for (int i = 1; i < mCurrMaxFrame; i++){
	int baseIndex = i*3;
	float posX = pStream[baseIndex-3 + X] * mSizeX;
	float posY = pStream[baseIndex-3 + Y] * mSizeY;
	float posZ = pStream[baseIndex-3 + Z] * mSizeZ;
	int iX = (int)posX;
	int iY = (int)posY;
	int iZ = (int)posZ;
	float fX = posX - iX;
	float fY = posY - iY;
	float fZ = posZ - iZ;
	
	if (iX < 0) iX = 0;
	if (iX >= mSizeX - 1) iX = mSizeX-2;
	if (iY < 0) iY = 0;
	if (iY >= mSizeY - 2) iY = mSizeY-2;
	if (iZ < 0) iZ = 0;
	if (iZ >= mSizeZ - 3) iZ = mSizeZ-2;

	int baseOffset = iZ * mOffsetZ + iY * mOffsetY + iX;

	float dx = (GetInterpolatedValue(pX, baseOffset, fX, fY, fZ)-128) / 255;
	float dy = (GetInterpolatedValue(pY, baseOffset, fX, fY, fZ)-128) / 255;
	float dz = (GetInterpolatedValue(pZ, baseOffset, fX, fY, fZ)-128) / 255;
	
	dx *= stepFactorX;
	dy *= stepFactorY;
	dz *= stepFactorZ;
	
	pStream[baseIndex + X] = pStream[baseIndex - 3 + X] + dx;
	pStream[baseIndex + Y] = pStream[baseIndex - 3 + Y] + dy;
	pStream[baseIndex + Z] = pStream[baseIndex - 3 + Z] + dz;
	// ---܂ŗ\q

	posX = pStream[baseIndex + X] * mSizeX;
	posY = pStream[baseIndex + Y] * mSizeY;
	posZ = pStream[baseIndex + Z] * mSizeZ;
	iX = (int)posX;
	iY = (int)posY;
	iZ = (int)posZ;
	fX = posX - iX;
	fY = posY - iY;
	fZ = posZ - iZ;

	if (iX < 0) iX = 0;
	if (iX >= mSizeX - 1) iX = mSizeX-2;
	if (iY < 0) iY = 0;
	if (iY >= mSizeY - 2) iY = mSizeY-2;
	if (iZ < 0) iZ = 0;
	if (iZ >= mSizeZ - 3) iZ = mSizeZ-2;

	baseOffset = iZ * mOffsetZ + iY * mOffsetY + iX;
	float dx2 = (GetInterpolatedValue(pX, baseOffset, fX, fY, fZ)-127) / 255;
	float dy2 = (GetInterpolatedValue(pY, baseOffset, fX, fY, fZ)-127) / 255;
	float dz2 = (GetInterpolatedValue(pZ, baseOffset, fX, fY, fZ)-127) / 255;
	
	dx2 *= stepFactorX;
	dy2 *= stepFactorY;
	dz2 *= stepFactorZ;

	pStream[baseIndex + X] += 0.5 * (dx2 - dx);
	pStream[baseIndex + Y] += 0.5 * (dy2 - dy);
	pStream[baseIndex + Z] += 0.5 * (dz2 - dz); 
	// ---Cq
    }
}

void ThreadConstructStream::
SetParameters(int initIndex, int finIndex,
	      unsigned char *pFieldX, unsigned char *pFieldY, unsigned char *pFieldZ,
	      float *pInitPosX, float *pInitPosY, float *pInitPosZ,
	      std::vector < float* > VpStreamPos){
    mInitIndex = initIndex;
    mFinIndex = finIndex;
    mpFieldX   = pFieldX;   mpFieldY   = pFieldY;   mpFieldZ   = pFieldZ;
    mpInitPosX = pInitPosX; mpInitPosY = pInitPosY; mpInitPosZ = pInitPosZ;
    mVpStreamPos = VpStreamPos;

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadConstructStream::
Entry(){
    for (int index = mInitIndex; index < mFinIndex; index++)
	FollowField(mpInitPosX[index], mpInitPosY[index], mpInitPosZ[index],
		    mVpStreamPos[index],
		    mpFieldX, mpFieldY, mpFieldZ);
    
    *mpbReadyFlag = TRUE;
    return 0;
}

// -----------------------------------------------

MACRO_FilterConstructor(PovStreamLine, WithTexture, mpCMNumCustomA = NULL);

void PovStreamLineGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    PovStreamLineData_cls *pCData = mpCParent->GetPCData();
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

    pCurr->mTextWidth = 75;
    pCurr->mItemWidth = -1;
    pCurr->mComboWidth = 70;
    pCurr->mButtonWidth = 70;
    
    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    wxString listColor[1] = { wxT("Mono Color") };
    mIndexColorType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &pCData->mCTexProperty.mCColorType.mBody,
	 wxT(""), TRUE, 1, listColor);

    InitializeColorPanel(pCurr, this, &GetColorR, &GetColorG, &GetColorB,
			 _("Texture"));
    mIndexColorBar = pCurr->CombinePointer(_GP_WXPANEL, mpCPanelColorBar);
    
    
    mIndexPointNum = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Line Intensity"));
    wxString intensityList[9] = { _("Very High"), _("High"), _("Middle High"), _("Middle"), _("Middle Low"), _("Low"), _("Very Low"), _("Verry Low"), _("No Line")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &pCData->mCLineIntensityIn.mBody,
	 wxT(""), TRUE, 9, intensityList);
    pCurr->CombinePointer(_GP_FLOAT, &pCData->mCLineWidthIn.mBody);

    // lݒ肵Ă
    pCData->GetPInterpolated()[2] = 3;

    pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Point Size"));
    pCurr->CombinePointer
	(_GP_BOOL, &pCData->mCbShowPoints.mbBody, _("Show Points"));
    pCurr->CombinePointer(_GP_DOUBLE, &pCData->GetPInterpolated()[2]);
    
    mIndexPointColor = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Point Color"));
    pCurr->CombinePointerButton
	(buttonCtrlPanelItemsCustomB, wxT("Color"));
    pCurr->mVCPointers[mIndexPointColor+1].mpConnectColorFloat = 
	pCData->mCPointColor.mpBody;

    pCurr->CombinePointer(_GP_DOUBLE, &pCData->GetPInterpolated()[1]);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 1);

    mIndexPointNum = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Point Dens."));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &pCData->mCPointNumber.mBody);
    pCurr->SetRange(0, 100);
    pCurr->CombinePointer(_GP_INTEGER, &pCData->mCPointNumber.mBody);
    pCurr->SetRange(0, 100);

    pCurr->mButtonWidth = -1;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    int indexCustomC = pCurr->CombinePointerButton
	(buttonCtrlPanelItemsCustomC, _("Set VxVyVz"));
    int indexCustomA = pCurr->CombinePointerButton
	(buttonCtrlPanelItemsCustomA, _("Construct Streamlines"));
    
    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    pCurr->UpdateValues();

    mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;
    mpObjectForCustomB = pCurr->mVCPointers[mIndexPointColor+1].mpPointerCtrl;
    mpObjectForCustomC = pCurr->mVCPointers[indexCustomC].mpPointerCtrl;

    UpdateColorList();

    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookOpenTextureDLG, _MN_HOOK_PANELCLICK);
    pCurr->SetFuncHook(&FilterWithTextureGUI_cls::FuncHookUpdateTextureDLG,
		       _MN_HOOK_COMBOBOX_CHANGE);
    pCurr->SetFuncHook(&FuncHookHideItem, _MN_HOOK_CHECKBOX);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);

    HideAndShowItems();
}

MACRO_SetCustomFunctions(PovStreamLineGUI_cls,ConstructStreamLine(),SelectColor(),SetVxVyVz());

void PovStreamLineGUI_cls::
HideAndShowItems(){
    if (mpCMNumPanel->mbPanelOpen == FALSE) return;
    
    bool bPoint = mpCParent->GetPCData()->mCbShowPoints.mbBody;

    mpCMNumPanel->ShowItems(mIndexPointNum, 3, bPoint);
    mpCMNumPanel->ShowItems(mIndexPointColor, 3, bPoint);

    //RefreshColorBar();
    mpCMNumPanel->GetSizer()->Layout();
    mpCParent->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
}

void PovStreamLineGUI_cls::
ConstructStreamLine(){
    mpCParent->ConstructStreamLine();
}

void PovStreamLineGUI_cls::
SelectColor(){
    wxColour CColor(mpCParent->GetPCData()->mCPointColor.mpBody[R]*255,
		    mpCParent->GetPCData()->mCPointColor.mpBody[G]*255,
		    mpCParent->GetPCData()->mCPointColor.mpBody[B]*255);
    wxColourDialogEX CDialog
	(mpCMNumPanel, CColor, TRUE);
    if (CDialog.ShowModal() == wxID_OK){
	LOOP_RGB mpCParent->GetPCData()->mCPointColor.mpBody[rgb] = 
	    CDialog.GetPColorFloat()[rgb];
    }
    mpCMNumPanel->UpdateValues();
    FilterCtrl_cls::FuncHookRefreshTop(mpCMNumPanel, NULL);
}

void PovStreamLineGUI_cls::
SetVxVyVz(){
    if (mpCMNumCustomA == NULL)	SetMNumCustomA(mpCMNumPanel);
    mpCMNumCustomA->UpdateValues();
    mpCMNumCustomA->ShowModal();
}

MACRO_FuncHookHideItem(PovStreamLine);

// -----

PovStreamLineData_cls::
PovStreamLineData_cls(FilterCtrl_cls* pCCtrl,
		      bool bConnectToStaticParam) :
    FilterWithParameterData_cls(pCCtrl)
{
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
    
    Parameters_cls CParam(3, L"StreamLine");
    double initParam[3] = {1, 1, 1};
    CParam.ExecCommandInitSet(initParam);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void PovStreamLineData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCCompX, wxT("PovStreamLineCompX="), L"0");
    mCConfList.Add(&mCCompY, wxT("PovStreamLineCompY="), L"1");
    mCConfList.Add(&mCCompZ, wxT("PovStreamLineCompZ="), L"2");
    mCConfList.Add(&mCLineWidthIn, wxT("PovStreamLineWidthIn="), L"1");
    mCConfList.Add(&mCbShowPoints, wxT("PovStreamShowPoints="), L"0");
    mCConfList.Add(&mCPointNumber, wxT("PovStreamPointNum="), L"1");
    mCConfList.Add(&mCPointColor, wxT("PovStreamPointColor="), L"1 0.5 0.5");
    mCConfList.Add(&mCLineIntensityIn, wxT("PovStreamLineIntensityIn="), L"3");
    mCConfList.Add(&mCLength,     wxT("PovStreamLength="), L"200");
    mCConfList.Add(&mCSkip,       wxT("PovStreamSkip="), L"2");
    mCConfList.Add(&mCStepFactor, wxT("PovStreamStepFactor="), L"0.005");
    mCConfList.Add(&mCClrStep,    wxT("PovStreamClrStep="), L"40");
    mCConfList.Add(&mCAnimation,  wxT("PovStreamAnimation="), L"0");
    mCConfList.Add(&mCPointType,  wxT("PovStreamPointType="), L"0");
    mCConfList.Add(&mCRandom,     wxT("PovStreamRandom="), L"0");
    mCConfList.Add(&mCInitAnim,   wxT("PovStreamInitAnim="), L"0");
    mCConfList.Add(&mCLineNumber, wxT("PovStreamLineNumber="), L"40 40 40");
}

MACRO_MakeInstanceAndAddToList(PovStreamLine);

wxString PovStreamLineData_cls::
SerializeConfPolymorph(){
    return mCTexProperty.Serialize() + msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();;
}

void PovStreamLineData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    mCTexProperty.ReadStrLine(strLine);
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void PovStreamLineData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* PovStreamLineData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > PovStreamLineData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > PovStreamLineData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// -----

PovStreamLine_cls::
PovStreamLine_cls(FilterCtrl_cls *pCParentCtrl, 
		  PovStreamLineData_cls* pCInitData) 
    : FilterWithParameter_cls(pCParentCtrl) 
{
    mpCData = new PovStreamLineData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) {
	*GetPCData() = *pCInitData; /*GetPCData()->ReDefineList(); */
    }
    GetPCData()->mCTexProperty.LinkFilterParameter(this, 0);

    ConstructorCommon(pCParentCtrl, _("Streamline"), _FILTER_STREAMLINE);
    mpCGUIPanel = new PovStreamLineGUI_cls(this);
    ((PovStreamLineGUI_cls*)mpCGUIPanel)->mpCDataCommon = GetPCData();
    
    mCurrMaxFrame = 0;
    mCurrLineNumber = 0;
    LOOP_XYZ mCurrLinePerAxis[xyz] = 0;
    mCurrClrStep = 0;
    mLineSkipIn = 0; mLineSkipInPrev = 0;
    //mLineSkipOut = 0; mLineSkipOutPrev = 0;
    mbNewStream = FALSE;
    LOOP_XYZ mMinRootPrev[xyz] = 0;
    LOOP_XYZ mMaxRootPrev[xyz] = 1;

    mbColorCurveAllocated = FALSE;
}
    
PovStreamLine_cls::
~PovStreamLine_cls(){
    for (unsigned int i = 0; i < mVpPosition.size(); i++)
	delete [] mVpPosition[i];
    
    if (mbColorCurveAllocated == TRUE){
	delete mpColorCurve;
	delete mpColorCurvePov;
    }
    if (mCurrLineNumber != 0){
	delete mpInitPosX; delete mpInitPosY; delete mpInitPosZ;
	delete mpRandTable;
    }

    if (FilterWithTextureGUI_cls::mspCLinkTexpropertyCtrl->mpCConnected == &GetPCData()->mCTexProperty)
	FilterWithTextureGUI_cls::mspCLinkTexpropertyCtrl->ConnectProperty(NULL, wxT("No Item is Linked"));
    
    if (mspCDF3SeqGUI != NULL)
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
}

MACRO_GetPCData(PovStreamLine);

MACRO_AddFilterItemPolymorphP(PovStreamLine,);
MACRO_InitializeStaticMemberP(PovStreamLine);

// ----

void PovStreamLine_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls* pCurr =
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));

    pCurr->mTextWidth = 100;
    pCurr->mComboWidth = 70;

    wxString listRandom[2] = { _("Orderly"), _("Random")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GetPCData()->mCRandom.mBody,
	 _("Start Point"), TRUE, 2, listRandom);

    int boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Stream Number"));
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCLineNumber.mpBody[X]);
    pCurr->SetRange(0, 80);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCLineNumber.mpBody[Y]);
    pCurr->SetRange(0, 80);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCLineNumber.mpBody[Z]);
    pCurr->SetRange(0, 80);

    boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Stream Length"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &GetPCData()->mCLength.mBody);
    pCurr->SetRange(0, 800);
    pCurr->CombinePointer(_GP_INTEGER, &GetPCData()->mCLength.mBody);
    pCurr->SetRange(0, 800);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("integration step"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &GetPCData()->mCStepFactor.mBody);
    pCurr->SetRange(0, 300, _MN_LASTITEM, 10000);
    pCurr->CombinePointer(_GP_FLOAT, &GetPCData()->mCStepFactor.mBody);
    pCurr->SetRange(0, 0.1);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("color step"));
    pCurr->CombinePointer(_GP_INTEGER_SLIDER, &GetPCData()->mCClrStep.mBody);
    pCurr->SetRange(0, 200);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCClrStep.mBody);
    pCurr->SetRange(0, 200);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Line Animation"));
    wxString streamAnim[2] = { _("No Animation"), _("Stream Line Anim")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GetPCData()->mCAnimation.mBody,
	 wxT(""), TRUE, 2, streamAnim);

    pCurr->CombinePointer
	(_GP_INTEGER_SPIN, &GetPCData()->mCInitAnim.mBody);
    pCurr->SetRange(0, 9999);
    
    pCurr->CombinePointer
	(_GP_INTEGER_SPIN, &GetPCData()->mCSkip.mBody, _("Skip in .pov"));
    pCurr->SetRange(1, 8);

    wxString listPointType[3] = { _("Consant size"), _("Constant in view"), _("mixed")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GetPCData()->mCPointType.mBody,
	 wxT("Point size"), TRUE, 3, listPointType);
    
    pCurr->FinishSettingDLG(FALSE);
    //pCurr->SetScreenRectPos(&msCRectMNumStaticMembers);
}

void PovStreamLine_cls::
AdvancedSetting(){
    AdvancedSettingCommon();
    FilterCtrl_cls::FuncHookRefreshTop(mpCGUIPanel->mpCMNumPanel, NULL);
}

void PovStreamLine_cls::
UpdateColors(){
    ((PovStreamLineGUI_cls*)mpCGUIPanel)->UpdateColorList();
}

// ----

void PovStreamLine_cls::
Allocate(int maxFrame, int linePerAxis[3], int clrStep){
    if (linePerAxis[X] == mCurrLinePerAxis[X] &&
	linePerAxis[Y] == mCurrLinePerAxis[Y] &&
	linePerAxis[Z] == mCurrLinePerAxis[Z] &&
	maxFrame == mCurrMaxFrame &&
	clrStep == mCurrClrStep)
	return;

    printf("ALLOCATION!!\n");

    for (unsigned int i = 0; i < mVpPosition.size(); i++){
	delete [] mVpPosition[i];
    }
    if (mbColorCurveAllocated == TRUE){
	delete mpColorCurve;
	delete mpColorCurvePov;
    }
    if (mCurrLineNumber != 0){
	delete mpInitPosX; delete mpInitPosY; delete mpInitPosZ;
	delete mpRandTable;
    }

    mCurrMaxFrame = maxFrame;
    mCurrLineNumber = linePerAxis[X] * linePerAxis[Y] * linePerAxis[Z];
    LOOP_XYZ mCurrLinePerAxis[xyz] = linePerAxis[xyz];
    mCurrClrStep = clrStep;
    
    mVpPosition.reserve(mCurrLineNumber);
    mVpPosition.resize(mCurrLineNumber);
    
    mVIndexForDrawIn.reserve(mCurrLineNumber);
    //mVIndexForDrawOut.reserve(mCurrLineNumber);
    mpInitPosX = new float [mCurrLineNumber];
    mpInitPosY = new float [mCurrLineNumber];
    mpInitPosZ = new float [mCurrLineNumber];
    mpRandTable = new float [mCurrLineNumber];
    
    for (unsigned int i = 0; i < mVpPosition.size(); i++)
	mVpPosition[i] = new float [ 3 * mCurrMaxFrame ];
    mpColorCurve = new float [mCurrMaxFrame * 2];
    mpColorCurvePov = new float [mCurrMaxFrame * 3];
    mbColorCurveAllocated = TRUE;

    for (int i = 0; i < mCurrMaxFrame * 2; i++){
	mpColorCurve[i] = (float)(i % mCurrClrStep) / mCurrClrStep;
    }
}

void PovStreamLine_cls::
ConstructInitPosition(){
    double minRoot[3] = {0, 0, 0};
    double maxRoot[3] = {1, 1, 1};
    GetClippingInformation(minRoot, maxRoot,
			   GetPCData()->mCstrCustomName.msBody + wxT("_Root"));
    double size[3] = {maxRoot[X] - minRoot[X],
		      maxRoot[Y] - minRoot[Y],
		      maxRoot[Z] - minRoot[Z]};
    PovStreamLineData_cls *pCData = GetPCData();
		      
    float initX = size[X] * 0.5f/pCData->mCLineNumber.mpBody[X] + minRoot[X];
    float initY = size[Y] * 0.5f/pCData->mCLineNumber.mpBody[Y] + minRoot[Y];
    float initZ = size[Z] * 0.5f/pCData->mCLineNumber.mpBody[Z] + minRoot[Z];
    
    srand(0);
    for (unsigned int i = 0; i < mVpPosition.size(); i++){
	mpInitPosX[i] = initX * (mSizeX-1) / mSizeX;
	mpInitPosY[i] = initY * (mSizeY-1) / mSizeY;
	mpInitPosZ[i] = initZ * (mSizeZ-1) / mSizeZ;
	initX += size[X] * 1.0f/pCData->mCLineNumber.mpBody[X];
	
	if (GetPCData()->mCRandom.mBody == 1){
	    mpInitPosX[i] += size[X] * ((float)rand() / RAND_MAX - 0.5) / 
		GetPCData()->mCLineNumber.mpBody[X];
	    mpInitPosY[i] += size[Y] * ((float)rand() / RAND_MAX - 0.5) /
		GetPCData()->mCLineNumber.mpBody[Y];
	    mpInitPosZ[i] += size[Z] * ((float)rand() / RAND_MAX - 0.5) / 
		GetPCData()->mCLineNumber.mpBody[Z];
	}
	
	if (initX > maxRoot[X]){
	    initX = size[X] * 0.5f/pCData->mCLineNumber.mpBody[X] + minRoot[X];
	    initY += size[Y] * 1.0f/pCData->mCLineNumber.mpBody[Y];}
	if (initY > maxRoot[Y]){
	    initY = size[Y] * 0.5f/pCData->mCLineNumber.mpBody[Y] + minRoot[Y];
	    initZ += size[Z] * 1.0f/pCData->mCLineNumber.mpBody[Z]; }
	mpRandTable[i] = rand();
    }
}

void PovStreamLine_cls::
ConstructStreamLine(){
    PovStreamLineData_cls *pCData = GetPCData();
    
    int slotIndex[3];
    slotIndex[X] = pCData->mCCompX.mBody;
    slotIndex[Y] = pCData->mCCompY.mBody;
    slotIndex[Z] = pCData->mCCompZ.mBody;
    
    int tmpsizeX[3], tmpsizeY[3], tmpsizeZ[3];
    LOOP_XYZ {
	tmpsizeX[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(X, slotIndex[xyz]);
	tmpsizeY[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Y, slotIndex[xyz]);
	tmpsizeZ[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Z, slotIndex[xyz]);
    }

    // TCYႤ̂͌
    if (! (tmpsizeX[X] == tmpsizeX[Y] && tmpsizeX[X] == tmpsizeX[Z]))
	return;
    if (! (tmpsizeY[X] == tmpsizeY[Y] && tmpsizeY[X] == tmpsizeY[Z]))
	return;
    if (! (tmpsizeZ[X] == tmpsizeZ[Y] && tmpsizeZ[X] == tmpsizeZ[Z]))
	return;

    mSizeX = tmpsizeX[X]; mSizeY = tmpsizeY[X]; mSizeZ = tmpsizeZ[X];
    mOffsetZ = mSizeX * mSizeY;
    mOffsetY = mSizeX;

    unsigned char *pX = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[X])->GetPDF3()->mpMainXYZ[Z];
    unsigned char *pY = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[Y])->GetPDF3()->mpMainXYZ[Z];
    unsigned char *pZ = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[Z])->GetPDF3()->mpMainXYZ[Z];
    
    if (pX == NULL || pY == NULL || pZ == NULL) return;

    Allocate(pCData->mCLength.mBody, pCData->mCLineNumber.mpBody,
	     pCData->mCClrStep.mBody);

    ConstructInitPosition();

    InitializeThreadCtrl();
    
    int increment = 
	pCData->mCLineNumber.mpBody[X] * pCData->mCLineNumber.mpBody[Y] * 2;
    for (unsigned int i = 0; i < mVpPosition.size(); i += increment){
	int indexEnd = i + increment;
	if (indexEnd > mVpPosition.size()) indexEnd = mVpPosition.size();
	ThreadConstructStream *pCThread = new ThreadConstructStream
	    (wxTHREAD_DETACHED, this);
	pCThread->SetParameters
	    (i, indexEnd, pX, pY, pZ, 
	     mpInitPosX, mpInitPosY, mpInitPosZ, mVpPosition);

	WaitForThreadsAndRun(pCThread);
    }

    WaitAllThreadsEnd(); 
    mbNewStream = TRUE;
    mpCParentCtrl->FuncHookRefreshTop(mpCGUIPanel->mpCMNumPanel, NULL);
}

void PovStreamLine_cls::
MakeIndexForDraw(double min[3], double max[3]){
    mVIndexForDrawIn.clear();
    int lineSkipIn = mLineSkipIn;
    int lineModIn = 0;
    if (lineSkipIn >= 4) lineModIn = lineSkipIn / 2;
    if (lineSkipIn == 0) return;

    for (int z = lineModIn; z < mCurrLinePerAxis[Z]; z+=lineSkipIn){
	int indexZ = z * mCurrLinePerAxis[X] * mCurrLinePerAxis[Y];
	for (int y = lineModIn; y < mCurrLinePerAxis[Y]; y+=lineSkipIn){
	    int indexY = indexZ + y * mCurrLinePerAxis[X];
	    for (int x = lineModIn; x < mCurrLinePerAxis[X]; x+=lineSkipIn){
		int index = indexY + x;
		mVIndexForDrawIn.push_back(index);
	    }
	}
    }
}

void PovStreamLine_cls::
DrawLinesMonoColorA(int maxFrame){
    for (unsigned int i = 0; i < mVIndexForDrawIn.size(); i++){
	unsigned int index = mVIndexForDrawIn[i];
	float *pPos = mVpPosition[index];
	glVertexPointer(3, GL_FLOAT, 0, mVpPosition[index]);
	glDrawElements(GL_LINE_STRIP, (GLsizei)maxFrame, GL_UNSIGNED_INT, msAIndexList);
    }
}

void PovStreamLine_cls::
DrawLinesColorCurveA(int currentFrame, int maxFrame){
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    for (unsigned int i = 0; i < mVIndexForDrawIn.size(); i++){
	unsigned int index = mVIndexForDrawIn[i];
	float *pPos = mVpPosition[index];

	glVertexPointer(3, GL_FLOAT, 0, mVpPosition[index]);
	glTexCoordPointer(1, GL_FLOAT, 0, 
			  mpColorCurve + mCurrMaxFrame - (currentFrame % mCurrMaxFrame));
	glDrawElements(GL_LINE_STRIP, (GLsizei)maxFrame, GL_UNSIGNED_INT, msAIndexList);
    }

    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}
    
void PovStreamLine_cls::
DrawBox(int currentFrame){
    if (GetPCData()->mCbValid.mbBody == FALSE) return;

    double minRoot[3] = {0, 0, 0};
    double maxRoot[3] = {1, 1, 1};
    GetClippingInformation(minRoot, maxRoot,
			   GetPCData()->mCstrCustomName.msBody + wxT("_Root"));
    double minClip[3] = {0, 0, 0};
    double maxClip[3] = {1, 1, 1};
    GetClippingInformation(minClip, maxClip,
			   GetPCData()->mCstrCustomName.msBody + wxT("_Clip"));
 
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    float alpha;

    if (GetPCData()->mCTexProperty.mCTransmitType.mBody != 0)
	alpha = GetPCData()->mCTexProperty.GetLinkedParam();
    else
	alpha = 1;

    unsigned char *pDensity = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(mpCParentCtrl->mSelfDF3Index)->GetPDF3()->mpMainXYZ[Z];
    
    float clr[3];
    LOOP_RGB clr[rgb] = GetPCData()->mCTexProperty.mCColor.mpBody[rgb];
    glColor4f(clr[0], clr[1], clr[2], alpha);
    
    switch(GetPCData()->mCLineIntensityIn.mBody){
    case 0: mLineSkipIn = 1; break;
    case 1: mLineSkipIn = 2; break;
    case 2: mLineSkipIn = 3; break;
    case 3: mLineSkipIn = 4; break;
    case 4: mLineSkipIn = 6; break;
    case 5: mLineSkipIn = 8; break;
    case 6: mLineSkipIn = 10; break;
    case 7: mLineSkipIn = 12; break;
    case 8: mLineSkipIn = 0; break;

    }
    
    if (minRoot[X] != mMinRootPrev[X] || maxRoot[X] != mMaxRootPrev[X] ||
	minRoot[Y] != mMinRootPrev[Y] || maxRoot[Y] != mMaxRootPrev[Y] ||
	minRoot[Z] != mMinRootPrev[Z] || maxRoot[X] != mMaxRootPrev[Z] ||
	mLineSkipIn != mLineSkipInPrev || mbNewStream == TRUE){
	MakeIndexForDraw(minRoot, maxRoot);
	mLineSkipInPrev = mLineSkipIn;
	mbNewStream = FALSE;
	LOOP_XYZ mMinRootPrev[xyz] = minRoot[xyz];
	LOOP_XYZ mMaxRootPrev[xyz] = maxRoot[xyz];
    }

    double dummy[3] = {0, 0, 0};
    SetGLColoring(&GetPCData()->mCTexProperty);
    Configure1DMapping(&GetPCData()->mCTexProperty, dummy, dummy,
		       TRUE, TRUE);

    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);

    glPushMatrix();

    glTranslated(0.5 / mSizeX, 0.5 / mSizeY, 0.5 / mSizeZ);
    // E炭 1boxel̍ŴAŏIIɂǂc
    glScaled((double)mSizeX/(mSizeX-1),
	     (double)mSizeY/(mSizeY-1),
	     (double)mSizeZ/(mSizeZ-1)); 
		 
    glTranslated(-0.5, -0.5, -0.5);

    if (maxClip[X] > ((double)mSizeX-1)/mSizeX) 
	maxClip[X] = ((double)mSizeX-1)/mSizeX;
    if (maxClip[Y] > ((double)mSizeY-1)/mSizeY) 
	maxClip[Y] = ((double)mSizeY-1)/mSizeY;
    if (maxClip[Z] > ((double)mSizeZ-1)/mSizeZ) 
	maxClip[Z] = ((double)mSizeZ-1)/mSizeZ;

    double PlaneEqXmin[]={+1.0, 0, 0, -minClip[X]};
    double PlaneEqXmax[]={-1.0, 0, 0, maxClip[X]};
    glClipPlane(GL_CLIP_PLANE0, PlaneEqXmin);
    glClipPlane(GL_CLIP_PLANE1, PlaneEqXmax); 
    double PlaneEqYmin[]={0, +1.0, 0, -minClip[Y]};
    double PlaneEqYmax[]={0, -1.0, 0, maxClip[Y]};
    glClipPlane(GL_CLIP_PLANE2, PlaneEqYmin);
    glClipPlane(GL_CLIP_PLANE3, PlaneEqYmax); 
    double PlaneEqZmin[]={0, 0, +1.0, -minClip[Z]};
    double PlaneEqZmax[]={0, 0, -1.0, maxClip[Z]};
    glClipPlane(GL_CLIP_PLANE4, PlaneEqZmin);
    glClipPlane(GL_CLIP_PLANE5, PlaneEqZmax); 

    glEnable(GL_CLIP_PLANE0);
    glEnable(GL_CLIP_PLANE1); 
    glEnable(GL_CLIP_PLANE2);
    glEnable(GL_CLIP_PLANE3); 
    glEnable(GL_CLIP_PLANE4);
    glEnable(GL_CLIP_PLANE5); 

    glLineWidth(GetPCData()->mCLineWidthIn.mBody);
    
    glEnableClientState(GL_VERTEX_ARRAY);

    int maxFrame;
    if (GetPCData()->mCAnimation.mBody == 0)
	maxFrame = mCurrMaxFrame;
    else
	maxFrame = currentFrame - GetPCData()->mCInitAnim.mBody;
    if (maxFrame > mCurrMaxFrame) maxFrame = mCurrMaxFrame;
    if (maxFrame < 0) maxFrame = 0;

    // STREAMLINE
    if (alpha != 0 && maxFrame > 0){
	if (GetPCData()->mCTexProperty.mCColorType.mBody == 0)
	    DrawLinesMonoColorA(maxFrame);
	else
	    DrawLinesColorCurveA(currentFrame, maxFrame);
    }

    glDisableClientState(GL_VERTEX_ARRAY);
    
    glDisable(GL_TEXTURE_1D);
    glPointSize(GetPCData()->GetPInterpolated()[2]);
    float pointAlpha = GetPCData()->GetPParamActive()->mpInterpolatedValue[1];
    float pointClr[3];
    LOOP_RGB pointClr[rgb] = GetPCData()->mCPointColor.mpBody[rgb];
    glColor4f(pointClr[R], pointClr[G], pointClr[B], pointAlpha);
    
    glBegin(GL_POINTS);
    int frameOffset = mCurrMaxFrame - currentFrame;
    if (GetPCData()->mCPointNumber.mBody != 0 && pointAlpha > 0 &&
	GetPCData()->mCbShowPoints.mbBody == TRUE){
	if (pDensity != NULL)
	    for (unsigned int i = 0; i < mVpPosition.size(); i++){
		float *pPos = mVpPosition[i];
		int ix = pPos[X] * mSizeX;
		int iy = pPos[Y] * mSizeY;
		int iz = pPos[Z] * mSizeZ;
		unsigned char density = 
		    pDensity[iz * mOffsetZ + iy * mSizeX + ix];
		float factor = 8.0f / (density + 0.01);
		int interval = mCurrMaxFrame / 
		    GetPCData()->mCPointNumber.mBody * factor;
		if (interval <= 30) interval = 30;
		
		int initframe = (currentFrame + (int)(mpRandTable[i] / RAND_MAX * interval )) % interval;
		for (int frame = initframe; frame < mCurrMaxFrame; frame+=interval)
		    glVertex3fv(&pPos[frame * 3]);
	    }
	else
	    for (unsigned int i = 0; i < mVpPosition.size(); i++){
		float *pPos = mVpPosition[i];
		unsigned char density = 16;
		float factor = 8.0f / (density + 0.01);
		int interval = mCurrMaxFrame /
		    GetPCData()->mCPointNumber.mBody * factor;
		srand(i);
		int initframe = (currentFrame + rand() * interval) % interval;
		for (int frame = initframe; frame < mCurrMaxFrame; frame+=interval)
		    glVertex3fv(&pPos[frame * 3]);
	    }
    }
    glEnd();
    glLineWidth(1);
    glPointSize(1);

    glDisable(GL_CLIP_PLANE5);
    glDisable(GL_CLIP_PLANE4);
    glDisable(GL_CLIP_PLANE3);
    glDisable(GL_CLIP_PLANE2);
    glDisable(GL_CLIP_PLANE1);
    glDisable(GL_CLIP_PLANE0);

    glPopMatrix();

    glPopAttrib();
}

bool PovStreamLine_cls::
DoesUseMapping(){
    return TRUE;
}

void PovStreamLine_cls::
MakeUV1D(TexProperty_cls *pCTexProp,
	 double offset[3], double scale[3]){
    int mappingType = pCTexProp->mCMappingType.mBody;
    int patternType = pCTexProp->mCPatternType.mBody;
    int mappingCoord = pCTexProp->mCMappingCoord.mBody;
    float *pos = pCTexProp->mCMappingPos.mpBody;
    float mapscale = pCTexProp->mCMappingScale.mBody;
    float mapturb = pCTexProp->mCMappingTurb.mBody;
    int axis = pCTexProp->mCMappingAxis.mBody;

    double rotate[4] = 
	{TexProperty_cls::msRotZ0, TexProperty_cls::msRotX0,
	 TexProperty_cls::msRotY0, TexProperty_cls::msRotZ1};

    if (IsParameterIsNewOne(mpCParentCtrl->GetRemappedDoubleIndex(),
			    mappingType, patternType, axis, mappingCoord, pos,
			    mapscale, mapturb, offset, scale, rotate) == FALSE
	&& mappingType == _TEX_MAPPING_PATTERN)
	return; 
    
    CopyCurrToPrev(pCTexProp);
}

// ----

wxString PovStreamLine_cls::
GetPovStrStreamLineTex(wxString strTag){
    wxString strTex, strFinishTag;

    strFinishTag << wxT("STREAMLINE_FINISH_") << strTag;

    strTex << ENDLWX;
    strTex << L"#declare " 
	   << wxT("STREAMLINE_TEX_") << strTag << L"=texture{" << ENDLWX;
    
    int colorType = -1;
    float trans = 1.0f - GetPCData()->GetPInterpolated()[0];
    int transmitType = GetPCData()->mCTexProperty.mCTransmitType.mBody;
    
    strTex << GetPCData()->mCTexProperty.GetPovStrTex
	(GetPCData()->GetPInterpolated()[0], strFinishTag);

    strTex << wxT("}") < ENDLWX;

    return strTex;
}

void PovStreamLine_cls::
MakeColorCurvePov(int clrType, int currentFrame){
    for (int index = 0; index < mCurrMaxFrame; index++){
	float coord = (mCurrMaxFrame + index - (currentFrame % mCurrMaxFrame)) % mCurrClrStep ;
	coord /= mCurrClrStep;
	LOOP_RGB
	    mpColorCurvePov[index*3+rgb] = GCColoring.mpCCustomColor->GetCustomColor(coord, rgb, TRUE, clrType - 1);
    }
}

void PovStreamLine_cls::
GetPovStrStreamLines(std::wfstream &ofstream,
		     wxString strTag, wxString strGeometry, 
		     int currentFrame, DrawingBlock_cls* pCDrawingBlock){

    int clrType = GetPCData()->mCTexProperty.mCColorType.mBody;

    double scale[3], scaleInv[3];
    LOOP_XYZ scale[xyz] = pCDrawingBlock->GetBoxScaleWorldCoord(xyz);
    LOOP_XYZ scaleInv[xyz] = 1.0/scale[xyz];
    
    if (mpCData->mCbValid.mbBody == FALSE) return;
    if (pCDrawingBlock->mDepth > 0) return;

    for (unsigned int i = 0; i < mVSAlreadyOutputedTag.size(); i++)
	if (strTag == mVSAlreadyOutputedTag[i]) return;
    mVSAlreadyOutputedTag.push_back(strTag);

    wxString strDeclare;

    strDeclare << wxT("#declare STREAMLINE_FINISH_") << mstrId + strTag << wxT("=finish{") << ENDLWX;
    strDeclare << GetPCData()->mCTexProperty.GetPovStrFinish(TRUE)
		  << wxT("}") << ENDLWX;

    strDeclare << GetPovStrStreamLineTex(mstrId + strTag) << ENDLWX;
    
    strDeclare << wxT("#macro TUBECLR(P0,P1,R,CLR)") << ENDLWX
	       << wxT("union{") << ENDLWX
	       << wxT("\tcylinder{P0,P1,R}") << ENDLWX
	       << wxT("\tsphere{P0,R}") << ENDLWX
	       << wxT("\tfinish {STREAMLINE_FINISH_") << mstrId + strTag << wxT("}") << ENDLWX 
	       << GetPCData()->mCTexProperty.GetPovStrNormalAll()
	       << wxT("\tpigment{ rgb CLR}") << ENDLWX
	       << wxT("}") << ENDLWX << wxT("#end") << ENDLWX;

    strDeclare << wxT("#macro TUBECLRFIN(P0,P1,R,CLR)") << ENDLWX
	       << wxT("union{") << ENDLWX
	       << wxT("\tcylinder{P0,P1,R}") << ENDLWX
	       << wxT("\tsphere{P0,R}") << ENDLWX
	       << wxT("\tsphere{P1,R}") << ENDLWX
	       << wxT("\tfinish {STREAMLINE_FINISH_") << mstrId + strTag << wxT("}") << ENDLWX 
	       << GetPCData()->mCTexProperty.GetPovStrNormalAll()
	       << wxT("\tpigment{ rgb CLR}") << ENDLWX
	       << wxT("}") << ENDLWX << wxT("#end") << ENDLWX;

    strDeclare << wxT("#macro TUBE(P0,P1,R)") << ENDLWX
	       << wxT("union{") << ENDLWX
	       << wxT("\tcylinder{P0,P1,R}") << ENDLWX
	       << wxT("\tsphere{P0,R}") << ENDLWX
	       << wxT("}") << ENDLWX << wxT("#end") << ENDLWX;

    strDeclare << wxT("#macro TUBEFIN(P0,P1,R)") << ENDLWX
	       << wxT("union{") << ENDLWX
	       << wxT("\tcylinder{P0,P1,R}") << ENDLWX
	       << wxT("\tsphere{P0,R}") << ENDLWX
	       << wxT("\tsphere{P1,R}") << ENDLWX
	       << wxT("}") << ENDLWX << wxT("#end") << ENDLWX;
    
    strDeclare << wxT("#macro TEXPIG(CLR)") << ENDLWX
	       << wxT("\tfinish {STREAMLINE_FINISH_") << mstrId + strTag << wxT("}") << ENDLWX
	       << GetPCData()->mCTexProperty.GetPovStrNormalAll()
	       << wxT("\tpigment{ rgb CLR}") << ENDLWX 
	       << wxT("#end") << ENDLWX;

    ofstream << wxConvFile.cWX2WC(strDeclare);
    
    ofstream << L"#declare OBJECT_STREAM_LINES_" << wxConvFile.cWX2WC(mstrId + strTag) << L"=union{" << ENDLWX;

    if (clrType != 0) MakeColorCurvePov(clrType, currentFrame);
        
    InitializeThreadCtrl();
    int corenum = 1; 
    int streamNum = mVIndexForDrawIn.size();
    
    if (streamNum > 8) corenum = 2;
    if (streamNum > 16) corenum = 4;
       
    wxString VstrStream[4];
    if (corenum > mThreadNum) corenum = mThreadNum;
    
    float width = 0.001 * GetPCData()->mCLineWidthIn.mBody;
    if (streamNum > 0){
	for (int i = 0; i < corenum; i++){
	    ThreadPovoutStream *pCThread = 
		new ThreadPovoutStream (wxTHREAD_DETACHED, this);
	    
	    int lineIni = i * streamNum / corenum;
	    int lineFin = (i + 1) * streamNum / corenum;
	    if (i == corenum - 1) lineFin = streamNum;
	    
	    // K؂ȐlH
	    if (clrType == 0)
		VstrStream[i].Alloc((lineFin - lineIni) * mCurrMaxFrame * 70);
	    else
		VstrStream[i].Alloc((lineFin - lineIni) * mCurrMaxFrame * 120);
	    pCThread->SetParameters
		(currentFrame, mCurrMaxFrame, clrType, scale,
		 lineFin - lineIni, &mVIndexForDrawIn[lineIni], 
		 &mVpPosition[0], mpColorCurvePov, width,
		 TRUE, &VstrStream[i]);
	    
	    WaitForThreadsAndRun(pCThread);
	}
    
	WaitAllThreadsEnd(); 

	for (int i = 0; i < corenum; i++)
	    ofstream << (const char*)VstrStream[i].ToAscii();
    }
    
    ofstream << wxConvFile.cWX2WC(PovOut3d(wxT("\tscale"), scaleInv)) << ENDL;
    // E炭 1boxel̍ŴAŏIIɂǂc
    ofstream << wxConvFile.cWX2WC(GetPovStrPeriodicOffset());
			    
    ofstream << L"\ttranslate <-0.5, -0.5, -0.5>" << ENDL;
    wxString strOption = GetPCData()->mCTexProperty.GetPovStrObjectOption();
    if (strOption != wxT(""))
	ofstream << L"\t" << wxConvFile.cWX2WC(strOption) << ENDL;
    ofstream << L"}" << ENDL;
    
    // --POINTS--
    float pointSize = 0.001 * GetPCData()->/*mCPointSize.mBody*/GetPInterpolated()[2];
    wxString strPointSize = GetShortString(pointSize);
    unsigned char *pDensity = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(mpCParentCtrl->mSelfDF3Index)->GetPDF3()->mpMainXYZ[Z];
    float pointAlpha = GetPCData()->GetPParamActive()->mpInterpolatedValue[1];
    PositionAndDirection_cls CCamLocalCoord;
    CCamLocalCoord = pCDrawingBlock->GetInvertedPosDir(DrawingBlock_cls::msCCamGlobalCoord, pCDrawingBlock->mbMirroring, FALSE, TRUE);
    
    int frameOffset = mCurrMaxFrame - currentFrame;
    int pointNumber = GetPCData()->mCPointNumber.mBody;
    if (mCurrLineNumber != 0 && 
	GetPCData()->mCPointNumber.mBody != 0 && pointAlpha > 0 &&
	GetPCData()->mCbShowPoints.mbBody == TRUE){
	ofstream << L"#declare OBJECT_STREAM_POINTS_" << wxConvFile.cWX2WC(mstrId + strTag) << L"=union{" << ENDL;
	
	InitializeThreadCtrl();
	corenum = 4;
	for (int i = 0; i < corenum; i++) VstrStream[i].Clear();
	
	streamNum = mVpPosition.size();
	for (int i = 0; i < corenum; i++){
	    ThreadPovoutPoint *pCThread = 
		new ThreadPovoutPoint(wxTHREAD_DETACHED, this);
	    
	    int lineIni = i * streamNum / corenum;
	    int lineFin = (i + 1) * streamNum / corenum;
	    if (i == corenum - 1) lineFin = streamNum;
	    
	    // K؂ȐlH
	    VstrStream[i].Alloc((lineFin - lineIni) * mCurrMaxFrame * 60);
	    
	    pCThread->SetParameters
		(currentFrame, mCurrMaxFrame, lineFin - lineIni, lineIni, 
		 scale, &mVpPosition[lineIni], pDensity, 
		 mSizeX, mSizeY, mSizeZ, pointSize, 
		 CCamLocalCoord.mPos/*DrawingBlock_cls::msCCamGlobalCoord.mPos*/,
		 &VstrStream[i]);
	
	    WaitForThreadsAndRun(pCThread);
	}

	WaitAllThreadsEnd();

	for (int i = 0; i < corenum; i++)
	    ofstream << (const char*)VstrStream[i].ToAscii();
	
	float alpha = GetPCData()->GetPParamActive()->mpInterpolatedValue[1];
	float pointClr[4];
	LOOP_RGB pointClr[rgb] = GetPCData()->mCPointColor.mpBody[rgb];
	if (alpha > 1) {
	    LOOP_RGB pointClr[rgb] *= alpha;
	    alpha = 1; 
	}
	
	pointClr[3] = 1.0f - alpha;
	
	ofstream << L"\tfinish { STREAMLINE_FINISH_" << wxConvFile.cWX2WC(mstrId + strTag) << L"}" << ENDL;
	ofstream << wxConvFile.cWX2WC(PovOut4f(wxT("\tpigment { rgbt"), pointClr, 3)) << L"}" << ENDL;
	ofstream << wxConvFile.cWX2WC(PovOut3d(wxT("\tscale"), scaleInv)) << ENDL;

	// E炭 1boxel̍ŴAŏIIɂǂc
	ofstream << wxConvFile.cWX2WC(GetPovStrPeriodicOffset());
	
	ofstream << L"\ttranslate <-0.5, -0.5, -0.5>" << ENDL;
	wxString strOption = GetPCData()->mCTexProperty.GetPovStrObjectOption();
	if (strOption != wxT(""))
	    ofstream << L"\t" << wxConvFile.cWX2WC(strOption) << ENDL;
	ofstream << L"}" << ENDL;
    }
    
    wxString strBoxLabel = mstrId + strTag;
    if (clrType ==0)
	ofstream << L"#declare STREAMLINETEXGLOBAL_" << wxConvFile.cWX2WC(strBoxLabel) << L"=1;" << ENDL << ENDL;
}

// ----

wxString PovStreamLine_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("PovStreamLine::");
    
    return strStatus;
}

MACRO_SerializeConfP(PovStreamLine)
MACRO_GetNewPReadDataBlock(PovStreamLine)

int    PovStreamLine_cls::msAIndexList[_DF3_SEQ_MAX];
