// "clsPovVelocityField.h"
// (setq c-basic-offset 4)

#include "clsPovVelocityField.h"

#include "../../COMMON/MyVector.h"
#include "../../DF3/clsDF3SeqCtrl.h"
#include "../../DF3/clsDF3SeqGUI.h"
#include "../../DF3/clsDF3Sequence.h"
#include "../../DF3/clsDF3.h"
#include "../../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../../FILTER/clsFilterCtrl.h"
#include "../../MAINCANVAS/clsFilterPanels.h"
#include "../../MYAPI/MsgBox.h"

#include "../../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../../WIDGETCTRL/EX/clsColourDialogEX.h"

#include "../../../clsDrawingBox.h"

#include "../../../wxWidgetEnum.h"

#include <wx/wx.h>
#include <wx/wfstream.h> 

#ifdef WIN32
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define R 0
#define G 1
#define B 2
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)
#define ENDLWX wxT("\n")
#define ENDL L"\n"

MACRO_FilterConstructor(PovVelField, WithTexture, mpCMNumCustomA = NULL);

void PovVelFieldGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    PovVelFieldData_cls *pCData = mpCParent->GetPCData();
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

    pCurr->mTextWidth = 75;
    pCurr->mItemWidth = -1;
    pCurr->mComboWidth = 70;
    pCurr->mButtonWidth = 70;
    
    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    wxString listColor[1] = { wxT("Mono Color") };
    mIndexColorType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &pCData->mCTexProperty.mCColorType.mBody,
	 wxT(""), TRUE, 1, listColor);

    InitializeColorPanel(pCurr, this, &GetColorR, &GetColorG, &GetColorB,
			 _("Texture"));
    mIndexColorBar = pCurr->CombinePointer(_GP_WXPANEL, mpCPanelColorBar);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Length"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &pCData->mCArrowLength.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &pCData->mCArrowLength.mBody);
    pCurr->SetRange(0, 2);
    
    int indexCustomA = pCurr->CombinePointerButton
	(buttonCtrlPanelItemsCustomA, _("Set VxVyVz"));

    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    pCurr->UpdateValues();

    mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;

    UpdateColorList();

    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookOpenTextureDLG, _MN_HOOK_PANELCLICK);
    pCurr->SetFuncHook(&FilterWithTextureGUI_cls::FuncHookUpdateTextureDLG,
		       _MN_HOOK_COMBOBOX_CHANGE);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);
    
    HideAndShowItems();
}

MACRO_SetCustomFunctions(PovVelFieldGUI_cls,SetVxVyVz(),,);

MACRO_HideAndShowEmpty(PovVelFieldGUI_cls);
MACRO_FuncHookHideItem(PovVelField);

void PovVelFieldGUI_cls::
SetVxVyVz(){
    if (mpCMNumCustomA == NULL)	SetMNumCustomA(mpCMNumPanel);
    mpCMNumCustomA->UpdateValues();
    mpCMNumCustomA->ShowModal();
}

// ----

PovVelFieldData_cls::
PovVelFieldData_cls(FilterCtrl_cls* pCCtrl,
		    bool bConnectToStaticParam) :
    FilterWithParameterData_cls(pCCtrl)
{
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
    
    Parameters_cls CParam(2, L"VelField");
    double initParam[2] = {1, 1}; // alpha, length
    CParam.ExecCommandInitSet(initParam);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void PovVelFieldData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCCompX, wxT("PovVelocityFiledX="), L"0");
    mCConfList.Add(&mCCompY, wxT("PovVelocityFiledY="), L"1");
    mCConfList.Add(&mCCompZ, wxT("PovVelocityFiledZ="), L"2");
    mCConfList.Add(&mCArrowWidth, wxT("PovVelFieldArrowWidth="), L"1");
    mCConfList.Add(&mCArrowHeadSize, wxT("PovVelFieldArrowHeadSize="), L"1");
    mCConfList.Add(&mCArrowLength, wxT("PovVelFieldArrowLength="), L"0.2");
    mCConfList.Add(&mCArrowNumber, wxT("PovVelFieldArrowNumber="), L"10 10 10");
}

MACRO_MakeInstanceAndAddToList(PovVelField);

wxString PovVelFieldData_cls::
SerializeConfPolymorph(){
    return mCTexProperty.Serialize() + msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();;
}

void PovVelFieldData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    mCTexProperty.ReadStrLine(strLine);
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void PovVelFieldData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* PovVelFieldData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > PovVelFieldData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > PovVelFieldData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// ----

// -----

PovVelField_cls::
PovVelField_cls(FilterCtrl_cls *pCParentCtrl, 
		PovVelFieldData_cls* pCInitData) 
    : FilterWithParameter_cls(pCParentCtrl) 
{
    mpCData = new PovVelFieldData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) {
	*GetPCData() = *pCInitData; /*GetPCData()->ReDefineList(); */
    }
    GetPCData()->mCTexProperty.LinkFilterParameter(this, 0);

    ConstructorCommon(pCParentCtrl, _("VelocityField"), _FILTER_VELOCITY_FIELD);
    mpCGUIPanel = new PovVelFieldGUI_cls(this);
    ((PovVelFieldGUI_cls*)mpCGUIPanel)->mpCDataCommon = GetPCData();
}

PovVelField_cls::
~PovVelField_cls(){
    if (FilterWithTextureGUI_cls::mspCLinkTexpropertyCtrl->mpCConnected == &GetPCData()->mCTexProperty)
	FilterWithTextureGUI_cls::mspCLinkTexpropertyCtrl->ConnectProperty(NULL, wxT("No Item is Linked"));
    
    if (mspCDF3SeqGUI != NULL)
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
}

MACRO_GetPCData(PovVelField);

MACRO_AddFilterItemPolymorphP(PovVelField,);
MACRO_InitializeStaticMemberP(PovVelField);

// ----
    
void PovVelField_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls* pCurr =
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));

    pCurr->mTextWidth = 100;
    pCurr->mComboWidth = 70;

    int boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Arrow Number"));
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCArrowNumber.mpBody[X]);
    pCurr->SetRange(0, 40);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCArrowNumber.mpBody[Y]);
    pCurr->SetRange(0, 40);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCArrowNumber.mpBody[Z]);
    pCurr->SetRange(0, 40);

    boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Arrow Width"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &GetPCData()->mCArrowWidth.mBody);
    pCurr->SetRange(0, 50, _MN_LASTITEM, 10);
    pCurr->CombinePointer(_GP_FLOAT, &GetPCData()->mCArrowWidth.mBody);
    pCurr->SetRange(0, 5);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Arrow Head Size"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &GetPCData()->mCArrowHeadSize.mBody);
    pCurr->SetRange(0, 40, _MN_LASTITEM, 10);
    pCurr->CombinePointer(_GP_FLOAT, &GetPCData()->mCArrowHeadSize.mBody);
    pCurr->SetRange(0, 4);
    
    pCurr->FinishSettingDLG(FALSE);
}

void PovVelField_cls::
AdvancedSetting(){
    AdvancedSettingCommon();
    FilterCtrl_cls::FuncHookRefreshTop(mpCGUIPanel->mpCMNumPanel, NULL);
}

void PovVelField_cls::
UpdateColors(){
    ((PovVelFieldGUI_cls*)mpCGUIPanel)->UpdateColorList();
}

bool PovVelField_cls::
DoesUseMapping(){
    return TRUE;
}

void PovVelField_cls::
MakeUV1D(TexProperty_cls *pCTexProp,
	 double offset[3], double scale[3]){
    int mappingType = pCTexProp->mCMappingType.mBody;
    int patternType = pCTexProp->mCPatternType.mBody;
    int mappingCoord = pCTexProp->mCMappingCoord.mBody;
    float *pos = pCTexProp->mCMappingPos.mpBody;
    float mapscale = pCTexProp->mCMappingScale.mBody;
    float mapturb = pCTexProp->mCMappingTurb.mBody;
    int axis = pCTexProp->mCMappingAxis.mBody;

    double rotate[4] = 
	{TexProperty_cls::msRotZ0, TexProperty_cls::msRotX0,
	 TexProperty_cls::msRotY0, TexProperty_cls::msRotZ1};

    if (IsParameterIsNewOne(mpCParentCtrl->GetRemappedDoubleIndex(),
			    mappingType, patternType, axis, mappingCoord, pos,
			    mapscale, mapturb, offset, scale, rotate) == FALSE
	&& mappingType == _TEX_MAPPING_PATTERN)
	return; 
    
    CopyCurrToPrev(pCTexProp);
}

// ----

wxString PovVelField_cls::
GetPovStrVelFieldTex(wxString strTag){
    wxString strTex, strFinishTag;

    strTex << wxT("#declare VELOCITYFIELD_FINISH_") << strTag << wxT("=finish{") << ENDLWX;
    strTex << GetPCData()->mCTexProperty.GetPovStrFinish(TRUE)
	   << wxT("}") << ENDLWX;
    
    strFinishTag << wxT("VELOCITYFIELD_FINISH_") << strTag;
    
    strTex << ENDLWX;
    strTex << L"#declare " 
	   << wxT("VELOCITYFIELD_TEX_") << strTag << L"=texture{" << ENDLWX;

    int colorType = -1;
    float trans = 1.0f - GetPCData()->GetPInterpolated()[0];
    int transmitType = GetPCData()->mCTexProperty.mCTransmitType.mBody;
    
    strTex << GetPCData()->mCTexProperty.GetPovStrTex
	(GetPCData()->GetPInterpolated()[0], strFinishTag);

    strTex << wxT("}") << ENDLWX;

    strTex << L"#macro VF_MACRO_TEX" << strTag << L"(CLR)" << ENDLWX
	   << L"texture{" << ENDLWX
	   << L"\tpigment {color rgb CLR}" << ENDLWX
	   << L"\tfinish { VELOCITYFIELD_FINISH_000}" << ENDLWX
	   << L"}" << ENDLWX
	   << L"#end" << ENDLWX;
    
    return strTex;
}

void PovVelField_cls::
GetPovStrVelocityField(std::wfstream &ofstream,
		       wxString strTag, wxString strGeometry, 
		       int currentFrame, DrawingBlock_cls* pCDrawingBlock){
    int clrType = GetPCData()->mCTexProperty.mCColorType.mBody;

    double scale[3], scaleInv[3];
    LOOP_XYZ scale[xyz] = pCDrawingBlock->GetBoxScaleWorldCoord(xyz);
    LOOP_XYZ scaleInv[xyz] = 1.0/scale[xyz];
    
    if (mpCData->mCbValid.mbBody == FALSE) return;
    if (pCDrawingBlock->mDepth > 0) return;

    for (unsigned int i = 0; i < mVSAlreadyOutputedTag.size(); i++)
	if (strTag == mVSAlreadyOutputedTag[i]) return;
    mVSAlreadyOutputedTag.push_back(strTag);


    int slotIndex[3];
    slotIndex[X] = GetPCData()->mCCompX.mBody;
    slotIndex[Y] = GetPCData()->mCCompY.mBody;
    slotIndex[Z] = GetPCData()->mCCompZ.mBody;

    int tmpsizeX[3], tmpsizeY[3], tmpsizeZ[3];
    LOOP_XYZ {
	tmpsizeX[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(X, slotIndex[xyz]);
	tmpsizeY[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Y, slotIndex[xyz]);
	tmpsizeZ[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Z, slotIndex[xyz]);
    }

    // TCYႤ̂͌
    if (! (tmpsizeX[X] == tmpsizeX[Y] && tmpsizeX[X] == tmpsizeX[Z]))
	return;
    if (! (tmpsizeY[X] == tmpsizeY[Y] && tmpsizeY[X] == tmpsizeY[Z]))
	return;
    if (! (tmpsizeZ[X] == tmpsizeZ[Y] && tmpsizeZ[X] == tmpsizeZ[Z]))
	return;

    mSize[X] = tmpsizeX[X]; mSize[Y] = tmpsizeY[X]; mSize[Z] = tmpsizeZ[X];
    mOffsetZ = mSize[X] * mSize[Y];
    mOffsetY = mSize[X];

    unsigned char *pX = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[X])->GetPDF3()->mpMainXYZ[Z];
    unsigned char *pY = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[Y])->GetPDF3()->mpMainXYZ[Z];
    unsigned char *pZ = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[Z])->GetPDF3()->mpMainXYZ[Z];

    if (pX == NULL || pY == NULL || pZ == NULL) return;

    double minRoot[3] = {0, 0, 0};
    double maxRoot[3] = {1, 1, 1};
    GetClippingInformation(minRoot, maxRoot,
			   GetPCData()->mCstrCustomName.msBody);
    double size[3] = {maxRoot[X] - minRoot[X],
		      maxRoot[Y] - minRoot[Y],
		      maxRoot[Z] - minRoot[Z]};
    PovVelFieldData_cls *pCData = GetPCData();

    float init[3]; // o_̈ʒu
    LOOP_XYZ
	init[xyz] = size[xyz] * 0.5f/pCData->mCArrowNumber.mpBody[xyz] + minRoot[xyz];
    
    int arrowNum = 
	pCData->mCArrowNumber.mpBody[X] *
	pCData->mCArrowNumber.mpBody[Y] *
	pCData->mCArrowNumber.mpBody[Z];

    float arrowLength = pCData->mCArrowLength.mBody;

    ofstream << wxConvFile.cWX2WC(GetPovStrVelFieldTex(mstrId + strTag)) << ENDLWX;
    ofstream << L"#declare OBJECT_VELOCITYFIELD_" << wxConvFile.cWX2WC(mstrId + strTag) << L"=union{" << ENDLWX;
    
    for (int i = 0; i < arrowNum; i++){
	// o_̈ʒuł̑x𓾂(ܕ)
	float pos[3], f[3], v[3], vunit[3], end[3], relpos[3]; 
	float arrowend[3], capend[3];
	int iPos[3];
	LOOP_XYZ pos[xyz] = init[xyz] * mSize[xyz];
	LOOP_XYZ iPos[xyz] = (int)pos[xyz];
	LOOP_XYZ f[xyz] = pos[xyz] - iPos[xyz];
	
	LOOP_XYZ if (iPos[xyz] < 0) iPos[xyz] = 0;
	LOOP_XYZ if (iPos[X] >= mSize[X] - 1) iPos[X] = mSize[X]-2;
	    
	int baseOffset = iPos[Z] * mOffsetZ + iPos[Y] * mOffsetY + iPos[X];

	v[X] = (GetInterpolatedValue(pX, baseOffset, f[X], f[Y], f[Z])-128) / 255;
	v[Y] = (GetInterpolatedValue(pY, baseOffset, f[X], f[Y], f[Z])-128) / 255;
	v[Z] = (GetInterpolatedValue(pZ, baseOffset, f[X], f[Y], f[Z])-128) / 255;

	LOOP_XYZ pos[xyz] = init[xyz];
	LOOP_XYZ end[xyz] = pos[xyz] + v[xyz] * arrowLength / msABoxScale[xyz];
	LOOP_XYZ vunit[xyz] = v[xyz];
	if (MyVector::GetNorm(vunit) == 0) continue;
	MyVector::Normalize(vunit, 1);
	
	LOOP_XYZ capend[xyz]   = end[xyz] - vunit[xyz] * GetPCData()->mCArrowHeadSize.mBody * GetPCData()->mCArrowWidth.mBody * 0.01 ;
	LOOP_XYZ arrowend[xyz] = end[xyz] - vunit[xyz] * GetPCData()->mCArrowHeadSize.mBody * GetPCData()->mCArrowWidth.mBody * 0.01 * 0.99;
	
	LOOP_XYZ relpos[xyz] = capend[xyz] - pos[xyz];
	
	LOOP_XYZ pos[xyz] *= scale[xyz];
	LOOP_XYZ end[xyz] *= scale[xyz];
	LOOP_XYZ capend[xyz] *= scale[xyz];
	LOOP_XYZ arrowend[xyz] *= scale[xyz];
	
	wxString strColor;
	if (clrType != 0){
	    double normv = MyVector::GetNorm(v);
	    strColor << PovOut3f(wxT(" VF_MACRO_TEX000("), 
				 &GCColoring.mpCCustomColor->GetV4CustomColor
				 (normv, TRUE, clrType - 1)[0], 3)
		     << wxT(")");
	}
	
	wxString strCylin;
	if (MyVector::GetScalarProduct(vunit, relpos) > 0)
	    strCylin << wxT("cylinder") << PovOut3f(wxT("{"), pos, 6) << PovOut3f(wxT(","), arrowend) << wxT(",") << GetPCData()->mCArrowWidth.mBody * 0.01 << strColor << wxT("}") << ENDLWX;

	wxString strCap;
	strCap << wxT("cone") << PovOut3f(wxT("{"), end, 6) << wxT(",0") << PovOut3f(wxT(","), capend) << wxT(",") << GetPCData()->mCArrowWidth.mBody * (0.015 + GetPCData()->mCArrowHeadSize.mBody * 0.01) << strColor << wxT("}") << ENDLWX;
	ofstream << wxConvFile.cWX2WC(strCylin);
	ofstream << wxConvFile.cWX2WC(strCap);
    
	init[X] += size[X] * 1.0f/pCData->mCArrowNumber.mpBody[X];
	if (init[X] > maxRoot[X]){
	    init[X] = size[X] * 0.5f/pCData->mCArrowNumber.mpBody[X] + minRoot[X];
	    init[Y] += size[Y] * 1.0f/pCData->mCArrowNumber.mpBody[Y];}
	if (init[Y] > maxRoot[Y]){
	    init[Y] = size[Y] * 0.5f/pCData->mCArrowNumber.mpBody[Y] + minRoot[Y];
	    init[Z] += size[Z] * 1.0f/pCData->mCArrowNumber.mpBody[Z]; }
    }

    ofstream << wxConvFile.cWX2WC(PovOut3d(wxT("\tscale"), scaleInv)) << ENDL;
    // E炭 1boxel̍ŴAŏIIɂǂc
    ofstream << wxConvFile.cWX2WC(GetPovStrPeriodicOffset());
			    
    ofstream << L"\ttranslate <-0.5, -0.5, -0.5>" << ENDL;

    ofstream << L"}" << ENDL;
}

// ----

void PovVelField_cls::
DrawArrows(){
    int slotIndex[3];
    slotIndex[X] = GetPCData()->mCCompX.mBody;
    slotIndex[Y] = GetPCData()->mCCompY.mBody;
    slotIndex[Z] = GetPCData()->mCCompZ.mBody;

    int tmpsizeX[3], tmpsizeY[3], tmpsizeZ[3];
    LOOP_XYZ {
	tmpsizeX[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(X, slotIndex[xyz]);
	tmpsizeY[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Y, slotIndex[xyz]);
	tmpsizeZ[xyz] = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Z, slotIndex[xyz]);
    }

    // TCYႤ̂͌
    if (! (tmpsizeX[X] == tmpsizeX[Y] && tmpsizeX[X] == tmpsizeX[Z]))
	return;
    if (! (tmpsizeY[X] == tmpsizeY[Y] && tmpsizeY[X] == tmpsizeY[Z]))
	return;
    if (! (tmpsizeZ[X] == tmpsizeZ[Y] && tmpsizeZ[X] == tmpsizeZ[Z]))
	return;
    
    int clrType = GetPCData()->mCTexProperty.mCColorType.mBody;
    mSize[X] = tmpsizeX[X]; mSize[Y] = tmpsizeY[X]; mSize[Z] = tmpsizeZ[X];
    mOffsetZ = mSize[X] * mSize[Y];
    mOffsetY = mSize[X];
    
    unsigned char *pX = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[X])->GetPDF3()->mpMainXYZ[Z];
    unsigned char *pY = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[Y])->GetPDF3()->mpMainXYZ[Z];
    unsigned char *pZ = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(slotIndex[Z])->GetPDF3()->mpMainXYZ[Z];

    if (pX == NULL || pY == NULL || pZ == NULL) return;

    double minRoot[3] = {0, 0, 0};
    double maxRoot[3] = {1, 1, 1};
    GetClippingInformation(minRoot, maxRoot,
			   GetPCData()->mCstrCustomName.msBody);
    double size[3] = {maxRoot[X] - minRoot[X],
		      maxRoot[Y] - minRoot[Y],
		      maxRoot[Z] - minRoot[Z]};
    PovVelFieldData_cls *pCData = GetPCData();

    float init[3]; // o_̈ʒu
    LOOP_XYZ
	init[xyz] = size[xyz] * 0.5f/pCData->mCArrowNumber.mpBody[xyz] + minRoot[xyz];
    
    int arrowNum = 
	pCData->mCArrowNumber.mpBody[X] *
	pCData->mCArrowNumber.mpBody[Y] *
	pCData->mCArrowNumber.mpBody[Z];

    float arrowLength = pCData->mCArrowLength.mBody;

    glBegin(GL_LINES);
    for (unsigned int i = 0; i < arrowNum; i++){
	// o_̈ʒuł̑x𓾂(ܕ)
	float pos[3], f[3], v[3], end[3];
	int iPos[3];
	LOOP_XYZ pos[xyz] = init[xyz] * mSize[xyz];
	LOOP_XYZ iPos[xyz] = (int)pos[xyz];
	LOOP_XYZ f[xyz] = pos[xyz] - iPos[xyz];
	
	LOOP_XYZ if (iPos[xyz] < 0) iPos[xyz] = 0;
	LOOP_XYZ if (iPos[X] >= mSize[X] - 1) iPos[X] = mSize[X]-2;
	    
	int baseOffset = iPos[Z] * mOffsetZ + iPos[Y] * mOffsetY + iPos[X];

	v[X] = (GetInterpolatedValue(pX, baseOffset, f[X], f[Y], f[Z])-128) / 255;
	v[Y] = (GetInterpolatedValue(pY, baseOffset, f[X], f[Y], f[Z])-128) / 255;
	v[Z] = (GetInterpolatedValue(pZ, baseOffset, f[X], f[Y], f[Z])-128) / 255;
	if (clrType != 0){
	    double normv = MyVector::GetNorm(v);
	    glColor4fv(&GCColoring.mpCCustomColor->GetV4CustomColor
		       (normv, TRUE, clrType - 1)[0]);
	}
	
	LOOP_XYZ pos[xyz] = init[xyz];
	LOOP_XYZ end[xyz] = pos[xyz] + v[xyz] * arrowLength / msABoxScale[xyz];

	// K
	double normalVec[3], vunit[3];
	normalVec[X] = v[Y] * mspCCamPos->mTgt[Z] - v[Z] * mspCCamPos->mTgt[Y];
	normalVec[Y] = v[Z] * mspCCamPos->mTgt[X] - v[X] * mspCCamPos->mTgt[Z];
	normalVec[Z] = v[X] * mspCCamPos->mTgt[Y] - v[Y] * mspCCamPos->mTgt[X];
	LOOP_XYZ vunit[xyz] = v[xyz];

	if (MyVector::GetNorm(vunit) == 0) continue;
	if (MyVector::GetNorm(normalVec) == 0) continue;

	MyVector::Normalize(vunit, 1);
	MyVector::Normalize(normalVec, 0.02);
	
	glVertex3fv(pos);
	glVertex3fv(end);

	float arrowPosA[3], arrowPosB[3];
	LOOP_XYZ arrowPosA[xyz] = end[xyz] + ( normalVec[xyz] - vunit[xyz] * 0.04)/msABoxScale[xyz] * pCData->mCArrowHeadSize.mBody;
	LOOP_XYZ arrowPosB[xyz] = end[xyz] + (-normalVec[xyz] - vunit[xyz] * 0.04)/msABoxScale[xyz] * pCData->mCArrowHeadSize.mBody;
	
	glVertex3fv(end);
	glVertex3fv(arrowPosA);
	glVertex3fv(end);
	glVertex3fv(arrowPosB);
	
	init[X] += size[X] * 1.0f/pCData->mCArrowNumber.mpBody[X];
	if (init[X] > maxRoot[X]){
	    init[X] = size[X] * 0.5f/pCData->mCArrowNumber.mpBody[X] + minRoot[X];
	    init[Y] += size[Y] * 1.0f/pCData->mCArrowNumber.mpBody[Y];}
	if (init[Y] > maxRoot[Y]){
	    init[Y] = size[Y] * 0.5f/pCData->mCArrowNumber.mpBody[Y] + minRoot[Y];
	    init[Z] += size[Z] * 1.0f/pCData->mCArrowNumber.mpBody[Z]; }
    }
    glEnd();
}

void PovVelField_cls::
DrawBox(){
    if (GetPCData()->mCbValid.mbBody == FALSE) return;

    double minClip[3] = {0, 0, 0};
    double maxClip[3] = {1, 1, 1};
    GetClippingInformation(minClip, maxClip,
			   GetPCData()->mCstrCustomName.msBody);

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    float alpha;
    
    if (GetPCData()->mCTexProperty.mCTransmitType.mBody != 0)
	alpha = GetPCData()->mCTexProperty.GetLinkedParam();
    else
	alpha = 1;

    unsigned char *pDensity = mpCParentCtrl->mspCDF3SeqCtrlLink->
	GetPDF3Seq(mpCParentCtrl->mSelfDF3Index)->GetPDF3()->mpMainXYZ[Z];

    float clr[3];
    LOOP_RGB clr[rgb] = GetPCData()->mCTexProperty.mCColor.mpBody[rgb];
    glColor4f(clr[0], clr[1], clr[2], alpha);

    glPushMatrix();
    
    glTranslated(0.5 / mSize[X],
		 0.5 / mSize[Y],
		 0.5 / mSize[Z]);
    // E炭 1boxel̍ŴAŏIIɂǂc
    glScaled((double)mSize[X]/(mSize[X]-1),
	     (double)mSize[Y]/(mSize[Y]-1),
	     (double)mSize[Z]/(mSize[Z]-1)); 

    glTranslated(-0.5, -0.5, -0.5);

    if (maxClip[X] > ((double)mSize[X]-1)/mSize[X]) 
	maxClip[X] = ((double)mSize[X]-1)/mSize[X];
    if (maxClip[Y] > ((double)mSize[Y]-1)/mSize[Y]) 
	maxClip[Y] = ((double)mSize[Y]-1)/mSize[Y];
    if (maxClip[Z] > ((double)mSize[Z]-1)/mSize[Z]) 
	maxClip[Z] = ((double)mSize[Z]-1)/mSize[Z];
    
    double PlaneEqXmin[]={+1.0, 0, 0, -minClip[X]};
    double PlaneEqXmax[]={-1.0, 0, 0, maxClip[X]};
    glClipPlane(GL_CLIP_PLANE0, PlaneEqXmin);
    glClipPlane(GL_CLIP_PLANE1, PlaneEqXmax); 
    double PlaneEqYmin[]={0, +1.0, 0, -minClip[Y]};
    double PlaneEqYmax[]={0, -1.0, 0, maxClip[Y]};
    glClipPlane(GL_CLIP_PLANE2, PlaneEqYmin);
    glClipPlane(GL_CLIP_PLANE3, PlaneEqYmax); 
    double PlaneEqZmin[]={0, 0, +1.0, -minClip[Z]};
    double PlaneEqZmax[]={0, 0, -1.0, maxClip[Z]};
    glClipPlane(GL_CLIP_PLANE4, PlaneEqZmin);
    glClipPlane(GL_CLIP_PLANE5, PlaneEqZmax); 
    
    glLineWidth(GetPCData()->mCArrowWidth.mBody);
 
    glDisable(GL_TEXTURE_1D);

    DrawArrows();

    glLineWidth(1);

    glDisable(GL_CLIP_PLANE5);
    glDisable(GL_CLIP_PLANE4);
    glDisable(GL_CLIP_PLANE3);
    glDisable(GL_CLIP_PLANE2);
    glDisable(GL_CLIP_PLANE1);
    glDisable(GL_CLIP_PLANE0);

    glPopMatrix();
    glPopAttrib();
}

// ----

wxString PovVelField_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("PovvelocityField::");
    
    return strStatus;
}

MACRO_SerializeConfP(PovVelField)
MACRO_GetNewPReadDataBlock(PovVelField)
