// -*- c -*- 
// (setq c-basic-offset 4)
// "clsLightSource.h"

#include "clsLightSource.h"

#include "../COMMON/MyVector.h"
#include "../DF3/clsDF3Sequence.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/GL/clsDF3GL.h"
#include "../DF3/GL/clsPreviewProperty.h"
#include "../GLCANVAS/clsStructDrawInfo.h"
#include "../GLCANVAS/clsHitTest.h"
#include "../MYGL/clsGLPerspective.h"
#include "../PARAMETER/clsParametersPanel.h"
#include "../WIDGETCTRL/clsAtoiComboCtrl.h"
#include "../WIDGETCTRL/clsIconSelector.h"
#include "../WIDGETCTRL/EX/clsColourDialogEX.h"

#include "../MYAPI/MsgBox.h"
#include "../POVRAY/clsPovRay.h"

#include "../UNDO/clsUndoManager.h"

#include "../../wxWidgetEnum.h"
#include "../../clsFrameMain.h"
#include "../../clsMain.h"
#include "../../clsTimeCtrl.h"

#include <sstream>

#include <wx/colordlg.h>

extern wxColourDialogEX* GpCClrDialog;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)
#define LOOP_RGBA for (int rgba = 0; rgba < 4; rgba++)
#define ENDL std::endl

#ifdef WIN32
#define ENDLWX wxT("\r\n")
#else
#define ENDLWX wxT("\n")
#endif

extern UndoManager_cls GCUndoList;

BEGIN_EVENT_TABLE(LightSRCGUI_cls, wxDialog)
EVT_BUTTON(wxID_CANCEL,    MultiParametersGUI_cls::OnButtonClose)
EVT_BUTTON(btnLightAdd,    LightSRCGUI_cls::OnButtonAdd)
EVT_BUTTON(btnLightRemove, LightSRCGUI_cls::OnButtonRemove)
EVT_BUTTON(btnLightClr,    LightSRCGUI_cls::OnButtonClr)
END_EVENT_TABLE()

LightSRCGUI_cls::
LightSRCGUI_cls(LightSRC_cls* pCParentLight, LightSRCGUI_cls **ppToThis, 
		wxWindow *pC, wxPoint pos, wxSize size) :
MultiPointInputGUI_cls(pC, pCParentLight, _("Light Source Configure"), pos, size,
		       L"LIGHTSRC_NUM=", L"Light", 4, 
		       pCParentLight->mpBitmapLink)
{
    mspCThis = this;
    mspCParentLight = pCParentLight;
    mVCDisplayList.push_back(_CAMERA_CROSSHAIR);
    //mVIcon.resize(1);
    mVLightTypeIcon.resize(2);
    
    wxIcon iconPointLight, iconSpotLight;
    iconPointLight.CopyFromBitmap(wxBitmap(wxT("ICONS/LightSource.png"), wxBITMAP_TYPE_PNG));
    iconSpotLight.CopyFromBitmap(wxBitmap(wxT("ICONS/SpotLight.png"), wxBITMAP_TYPE_PNG));
    
    //mVIcon[0] = iconPointLight;//.CopyFromBitmap(wxBitmap(wxT("ICONS/LightSource.png"), wxBITMAP_TYPE_PNG));
    mVLightTypeIcon[0] = iconPointLight;
    mVLightTypeIcon[1] = iconSpotLight;

       
    double pInit[4] = {1.2, 1.2, 1.2, 1.2};
    SetInitParam(pInit);

    //SetCloseButtonID(buttonCtrlLightClose);
    SetScreenRectPos(&pCParentLight->mCScreenRect);
   
    // TCU[
    wxBoxSizer       *pBoxSizerSrc   = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerBTM   = new wxBoxSizer(wxHORIZONTAL);

    SetMNumMain(pBoxSizerSrc, btnLightAdd, btnLightRemove);
    SetMNumParam(pBoxSizerSrc);

    pBoxSizerSrc->Add( pBoxSizerBTM,  0, wxALIGN_RIGHT);

    SetSizerAndFit(pBoxSizerSrc);
    
    int minimumWidth = GetClientSize().GetWidth() - 8;
    minimumWidth = mpCMNumMain->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumParam->GetPaddingWidth(minimumWidth);
    mpCMNumMain->SetPaddingWidth(minimumWidth);
    mpCMNumParam->SetPaddingWidth(minimumWidth);

    FuncHookItemNumberChanged(NULL, (MultiParameters_cls*)this);

    RemakeIconList();
}

LightSRCGUI_cls::
~LightSRCGUI_cls(){
}

void LightSRCGUI_cls::
SetMNumParam(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr =
	InitializeMNumPanel(&mpCMNumParam, this);
#ifdef WIN32
    pCurr->mTextWidth  = 90;
    pCurr->mComboWidth = 60;
    pCurr->mButtonWidth = 60;
#else
    pCurr->mTextWidth  = 110;
    pCurr->mComboWidth = 80;
    pCurr->mButtonWidth = 80;
#endif
    int boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    mIndexPos = pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
    
    boxWidth = 2;
    static float clrdummy[3] = {1, 1, 1};
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("emission"));
    mIndexEmit = pCurr->CombinePointer(_GP_DOUBLE, NULL);
    mIndexClr  = pCurr->CombinePointerButton(btnLightClr, _("Color"));
    pCurr->mVCPointers[mIndexClr].mpConnectColorFloat = clrdummy;
    
    mIndexOption = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    mIndexShadow = pCurr->CombinePointer(_GP_BOOL, NULL, _("cast shadow"));
    mIndexChild = pCurr->CombinePointer(_GP_BOOL, NULL, _("Child of cam"));

    wxString listFadeTypeStr[3] = {
	_("No Fade"), _("power 1"), _("power 2")};
    mIndexFadeType = pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, NULL, _("Fade Type"), FALSE, 3, listFadeTypeStr);

    static float fdummy2 = 0.1;
    mIndexFadeDist = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Fade Distance"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &fdummy2);
    pCurr->SetRange(1, 50, _MN_LASTITEM, 50);
    pCurr->CombinePointer(_GP_FLOAT, &fdummy2);
    pCurr->SetRange(0.02, 1);
    
    wxString listTypeStr[4] = {
	_("Point Light"), _("Spot Light"), _("Cylinder Light"), _("Parallel Light") };
    mIndexType = pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, NULL, _("Light Type"), FALSE, 4, listTypeStr);

    wxString listSpotTarget[4] = {
	_("Origin"), _("Anchor Pos"), _("Numeric Abs."), _("Numeric Rel.")};
    mIndexSpotTarget = pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, NULL, _("Spot Target"), FALSE, 4, listSpotTarget);

    boxWidth = 3;
    mIndexTargetPos = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->CombinePointer(_GP_FLOAT, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_FLOAT, NULL, wxT("-"));
    pCurr->CombinePointer(_GP_FLOAT, NULL, wxT("-"));
    
    boxWidth = 2;
    static float fdummy0;
    mIndexSpotAngle = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Spot Angle"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &fdummy0);
    pCurr->SetRange(10, 70, _MN_LASTITEM, 1);
    pCurr->CombinePointer(_GP_FLOAT, &fdummy0);
    pCurr->SetRange(10, 70, _MN_LASTITEM, 1);

    static float fdummy1;
    mIndexBeamRadius = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Beam Radius"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &fdummy1);
    pCurr->SetRange(1, 50, _MN_LASTITEM, 10);
    pCurr->CombinePointer(_GP_FLOAT, &fdummy1);
    pCurr->SetRange(0.1, 5, _MN_LASTITEM);
    
    mIndexArea = mIndexAreaParam = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Area Light"));
    pCurr->CombinePointer(_GP_BOOL, NULL);
    pCurr->CombinePointer(_GP_BOOL, NULL, _("Jitter"));
    
    mIndexAreaParam = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Size/Quality"));
    mIndexAreaSize = pCurr->CombinePointer(_GP_FLOAT, NULL);
    static unsigned int dummy;
    mIndexAreaQuality = pCurr->CombinePointer(_GP_UINT_SPIN, &dummy);
    pCurr->SetRange(2, 12);
    
    mIndexObject = pCurr->CombinePointer(_GP_BOOL, NULL, _("Object Light"));
    mIndexObjectParam = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Size/Emission"));
    mIndexObjectSize = pCurr->CombinePointer(_GP_FLOAT, NULL);
    mIndexObjectEmission = pCurr->CombinePointer(_GP_FLOAT, NULL);
    
#ifdef WIN32
    pCurr->mMinimumWidth = 200;
#else
    pCurr->mMinimumWidth = 230;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(&FuncHookApply, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookApply, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&FuncHookType, _MN_HOOK_COMBOBOX);
    pCurr->SetFuncHook(&FuncHookType, _MN_HOOK_CHECKBOX);

    pCurr->SetColorForXYZ(mIndexPos);
    pCurr->SetColorForXYZ(mIndexTargetPos+1);
}

void LightSRCGUI_cls::
RemakeIconList(){
    mVIcon.clear();
    for (int i = 0; i< mpCParent->mCItemNum.mBody; i++){
	if (mspCParentLight->mVCLightParam[i].mCType.mBody == _LIGHT_SPOT)
	    mVIcon.push_back(mVLightTypeIcon[1]);
	else
	    mVIcon.push_back(mVLightTypeIcon[0]);
    }

    ((IconSelector_cls*)mpCMNumMain->mVCPointers[mIndexIcon].mpPointerCtrl)
	->RemakeImageList(&mVIcon);
}

void LightSRCGUI_cls::
ReconnectMNumParam(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumParam;

    if (mCItemNum.mBody == 0) {
	pCurr->mVCPointers[mIndexPos+0].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[mIndexPos+1].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[mIndexPos+2].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[mIndexEmit].CombinePointer(_GP_DOUBLE, NULL, wxT("-"));
	pCurr->mVCPointers[mIndexShadow].CombinePointer(_GP_BOOL, NULL, _("cast shadow"));
	pCurr->mVCPointers[mIndexChild].CombinePointer(_GP_BOOL, NULL, _("Child of cam"));
	pCurr->mVCPointers[mIndexArea+1].CombinePointer(_GP_BOOL, NULL);
	pCurr->mVCPointers[mIndexArea+2].CombinePointer(_GP_BOOL, NULL, _("Jitter"));
	pCurr->mVCPointers[mIndexFadeType].CombinePointer(_GP_UNSIGNED_INT_LIST, NULL, wxT("-"));
	
	pCurr->mVCPointers[mIndexFadeDist+1].CombinePointer(_GP_FLOAT_SLIDER, NULL);
	pCurr->mVCPointers[mIndexFadeDist+2].CombinePointer(_GP_FLOAT, NULL);
	
	pCurr->mVCPointers[mIndexType].CombinePointer(_GP_UNSIGNED_INT_LIST, NULL, wxT("-"));
	pCurr->mVCPointers[mIndexSpotTarget].CombinePointer(_GP_UNSIGNED_INT_LIST, NULL, wxT("-"));

	pCurr->mVCPointers[mIndexTargetPos+1].CombinePointer(_GP_FLOAT, NULL, wxT("-"));
	pCurr->mVCPointers[mIndexTargetPos+2].CombinePointer(_GP_FLOAT, NULL, wxT("-"));
	pCurr->mVCPointers[mIndexTargetPos+3].CombinePointer(_GP_FLOAT, NULL, wxT("-"));
	
	pCurr->mVCPointers[mIndexSpotAngle+1].CombinePointer(_GP_FLOAT_SLIDER, NULL);
	pCurr->mVCPointers[mIndexSpotAngle+2].CombinePointer(_GP_FLOAT, NULL);
	pCurr->mVCPointers[mIndexBeamRadius+1].CombinePointer(_GP_FLOAT_SLIDER, NULL);
	pCurr->mVCPointers[mIndexBeamRadius+2].CombinePointer(_GP_FLOAT, NULL);

	pCurr->mVCPointers[mIndexObject].CombinePointer(_GP_BOOL, NULL, _("Object Light"));
	pCurr->mVCPointers[mIndexObjectSize].CombinePointer(_GP_BOOL, NULL, _("Object Size"));
	pCurr->mVCPointers[mIndexObjectEmission].CombinePointer(_GP_BOOL, NULL, _("Object Emission"));

	pCurr->UpdateValues();
	HideAndShowAll();
	return;
    }

    double *ptr = mVCParam[mActiveItem].GetPInterpolatedValue(mspCParentLight->mpCParentFrame->mpCParent->mpCTime->GetDoubleIndex());
    pCurr->mVCPointers[mIndexPos + 0].
	CombinePointer(_GP_DOUBLE, &ptr[0], wxT("x"));
    pCurr->mVCPointers[mIndexPos + 1].
	CombinePointer(_GP_DOUBLE, &ptr[1], wxT("y"));
    pCurr->mVCPointers[mIndexPos + 2].
	CombinePointer(_GP_DOUBLE, &ptr[2], wxT("z"));
    pCurr->mVCPointers[mIndexEmit].
	CombinePointer(_GP_DOUBLE, &ptr[3], _("emission"));

    LightParam_cls *pCLightParam = &mspCParentLight->mVCLightParam[mActiveItem];

    pCurr->mVCPointers[mIndexClr].
	mpConnectColorFloat = pCLightParam->mCColor.mpBody;
    
    pCurr->mVCPointers[mIndexShadow].
	CombinePointer(_GP_BOOL, &pCLightParam->mCbShadow.mbBody, 
		       _("cast shadow"));
    pCurr->mVCPointers[mIndexChild].
	CombinePointer(_GP_BOOL, &pCLightParam->mCbChildOfCamera.mbBody, 
		       _("Child of cam"));
    pCurr->mVCPointers[mIndexArea+1].
	CombinePointer(_GP_BOOL, &pCLightParam->mCbAreaLight.mbBody);
    pCurr->mVCPointers[mIndexArea+2].
	CombinePointer(_GP_BOOL, &pCLightParam->mCbAreaJitter.mbBody, 
		       _("Jitter"));

    pCurr->mVCPointers[mIndexObject].
	CombinePointer(_GP_BOOL, &pCLightParam->mCbObject.mbBody, 
		       _("Object Light"));
    pCurr->mVCPointers[mIndexObjectSize].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCObjectSize.mBody, 
		       _("Object Size"));
    pCurr->mVCPointers[mIndexObjectEmission].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCObjectEmission.mBody, 
		       _("Object Emission"));
    
    pCurr->mVCPointers[mIndexFadeType].GetPCtrlWindow()->Freeze();
    pCurr->mVCPointers[mIndexType].GetPCtrlWindow()->Freeze();
    pCurr->mVCPointers[mIndexSpotTarget].GetPCtrlWindow()->Freeze();

    wxString listFadeTypeStr[3] = {
	_("No Fade"), _("power 1"), _("power 2")};
    pCurr->mVCPointers[mIndexFadeType].
	CombinePointer(_GP_UNSIGNED_INT_LIST, &pCLightParam->mCFadeType.mBody,
		       _("Fade Type"), TRUE, 3, listFadeTypeStr, mspCParentLight->mVCLightParam[mActiveItem].mCFadeType.mBody);
    ((wxWindow*)pCurr->mVCPointers[mIndexFadeType].mpPointerCtrl)->Enable(TRUE);

    wxString listTypeStr[4] = {
	_("Point Light"), _("Spot Light"), _("Cylinder Light"), _("Parallel Light") };
    pCurr->mVCPointers[mIndexType].
	CombinePointer(_GP_UNSIGNED_INT_LIST, &pCLightParam->mCType.mBody, 
		       _("Light Type"), TRUE, 4, listTypeStr, 
		       pCLightParam->mCType.mBody);
    ((wxWindow*)pCurr->mVCPointers[mIndexType].mpPointerCtrl)->Enable(TRUE);

    wxString listSpotTarget[4] = {
	_("Origin"), _("Anchor Pos"), _("Numeric Abs."), _("Numeric Rel.")};
    pCurr->mVCPointers[mIndexSpotTarget].
	CombinePointer(_GP_UNSIGNED_INT_LIST,&pCLightParam->mCSpotTarget.mBody, 
		       _("Spot Target"), TRUE, 4, listSpotTarget,
		       pCLightParam->mCSpotTarget.mBody);
    ((wxWindow*)pCurr->mVCPointers[mIndexSpotTarget].mpPointerCtrl)->Enable(TRUE);

    pCurr->mVCPointers[mIndexFadeType].GetPCtrlWindow()->Thaw();
    pCurr->mVCPointers[mIndexType].GetPCtrlWindow()->Thaw();
    pCurr->mVCPointers[mIndexSpotTarget].GetPCtrlWindow()->Thaw();

    pCurr->mVCPointers[mIndexTargetPos+1].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCTargetPos.mpBody[X]);
    pCurr->mVCPointers[mIndexTargetPos+2].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCTargetPos.mpBody[Y]);
    pCurr->mVCPointers[mIndexTargetPos+3].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCTargetPos.mpBody[Z]);
    
    pCurr->mVCPointers[mIndexAreaSize].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCAreaSize.mBody, 
		       _("Area Size"));
    pCurr->mVCPointers[mIndexAreaQuality].
	CombinePointer(_GP_UINT_SPIN, &pCLightParam->mCAreaQuality.mBody, 
		       _("Area Quality"));

    pCurr->mVCPointers[mIndexFadeDist+1].
	CombinePointer(_GP_FLOAT_SLIDER, &pCLightParam->mCFadeDistance.mBody);
    pCurr->mVCPointers[mIndexFadeDist+2].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCFadeDistance.mBody);

    pCurr->mVCPointers[mIndexSpotAngle+1].
	CombinePointer(_GP_FLOAT_SLIDER, &pCLightParam->mCSpotAngle.mBody);
    pCurr->mVCPointers[mIndexSpotAngle+2].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCSpotAngle.mBody);
    
    pCurr->mVCPointers[mIndexBeamRadius+1].
	CombinePointer(_GP_FLOAT_SLIDER, &pCLightParam->mCBeamRadius.mBody);
    pCurr->mVCPointers[mIndexBeamRadius+2].
	CombinePointer(_GP_FLOAT, &pCLightParam->mCBeamRadius.mBody);
    
    HideAndShowAll();

    pCurr->UpdateValues();
}

void LightSRCGUI_cls::
HideAndShowAll(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumParam;
    
    bool bArea = FALSE, bSpot = FALSE, bCylinder = FALSE, bFade = FALSE;
    bool bObject = FALSE, bNumeric = FALSE;
    if (mCItemNum.mBody != 0){
	LightParam_cls *pCLightParam = &mspCParentLight->mVCLightParam[mActiveItem];
	bArea = (pCLightParam->mCbAreaLight.mbBody == TRUE);
	bSpot = (pCLightParam->mCType.mBody == _LIGHT_SPOT);
	bCylinder = (pCLightParam->mCType.mBody == _LIGHT_CYLINDRICAL);
	bFade = (pCLightParam->mCFadeType.mBody != 0);
	bObject = (pCLightParam->mCbObject.mbBody == TRUE);
	bNumeric = (pCLightParam->mCSpotTarget.mBody >= _LSPOT_NUMERIC_ABS);
    }

    pCurr->ShowItems(mIndexFadeDist, 3, bFade);

    pCurr->ShowItems(mIndexArea+2, 1, bArea);
    pCurr->ShowItems(mIndexAreaParam, 3, bArea);

    pCurr->ShowItems(mIndexTargetPos, 4, (bSpot || bCylinder) && bNumeric);
    pCurr->ShowItems(mIndexSpotTarget, 1, bSpot || bCylinder);
    pCurr->ShowItems(mIndexSpotAngle, 3, bSpot);
    pCurr->ShowItems(mIndexBeamRadius, 3, bCylinder);

    pCurr->ShowItems(mIndexObjectParam, 3, bObject);

    RemakeIconList();    
    pCurr->GetSizer()->Layout();
    ResizeDialog();
}

void LightSRCGUI_cls::
FuncHookType(void *pArg, void *pEvent){
    mspCThis->HideAndShowAll();
    mspCThis->RemakeIconList();
}

void LightSRCGUI_cls::
UpdateActiveOnTL(){
    mspCParentLight->UpdateActiveOnTL();
}

// ----

void LightSRCGUI_cls::
OnButtonAdd(wxCommandEvent &event){
    unsigned int color[3] = {255, 255, 255};
    GCUndoList.OpenBlock();
    LightParam_cls CDefaultParam;
    mspCParentLight->AddLight(mCItemNum.mBody, CDefaultParam);
    MultiParametersGUI_cls::OnButtonAdd(event);
    GCUndoList.CloseBlock();
    RemakeIconList();
}

void LightSRCGUI_cls::
OnButtonRemove(wxCommandEvent &event){
    if (mCItemNum.mBody == 0) return;
    GCUndoList.OpenBlock();
    unsigned int current = mActiveItem;
    if (MultiParametersGUI_cls::RemoveItem() == TRUE)
	mspCParentLight->RemoveLight(current);
    GCUndoList.CloseBlock();
    RemakeIconList();
}

void LightSRCGUI_cls::
OnButtonClr(wxCommandEvent &event){
    GpCClrDialog->SetColour(mspCParentLight->GetColour(mActiveItem));
    if (GpCClrDialog->ShowModal() == wxID_OK)
	mspCParentLight->SetColour(mActiveItem, GpCClrDialog->GetPColorFloat());
    /*
    wxColourDialogEX CDialog
	(this, mspCParentLight->GetColour(mActiveItem), TRUE);
    if (CDialog.ShowModal() == wxID_OK)
	mspCParentLight->SetColour(mActiveItem, CDialog.GetPColorFloat());
    ;//mpCParent->mpCBG->mpCMNumBG->UpdateValues();
    */
}

void LightSRCGUI_cls::
FuncHookApply(void *pArg, void *pEvent){
    mspCThis->OnApplyBase();
}

LightSRC_cls* LightSRCGUI_cls::mspCParentLight;
LightSRCGUI_cls* LightSRCGUI_cls::mspCThis;

// --------------------

LightParam_cls::
LightParam_cls() 
{
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
}

void LightParam_cls::
ReDefineList(){
    mCConfList.Clear();
    mCConfList.Add(&mCColor, wxT("LightColor="), wxT("1 1 1"));
    mCConfList.Add(&mCbShadow,   wxT("LightShadow="), wxT("1"));
    mCConfList.Add(&mCFadeType,  wxT("LightFadeType="), wxT("0"));
    mCConfList.Add(&mCSpotTarget, wxT("LightSpotTarget="), wxT("0"));
    mCConfList.Add(&mCTargetPos, wxT("LightSpotPos="), wxT("0 0 0"));
    mCConfList.Add(&mCFadeDistance,  wxT("LightFadeDistance="), wxT("0.1"));
    mCConfList.Add(&mCType,      wxT("LightType="), wxT("0"));
    mCConfList.Add(&mCbChildOfCamera, wxT("LightChildOfCamera="), wxT("0"));
    mCConfList.Add(&mCbAreaLight, wxT("LightArea="), wxT("0"));
    mCConfList.Add(&mCbObject,    wxT("LightObject="), wxT("0"));
    mCConfList.Add(&mCAreaSize,   wxT("LightAreaSize="), wxT("0.2"));
    mCConfList.Add(&mCAreaQuality,wxT("LightAreaQuality="), wxT("5"));
    mCConfList.Add(&mCbAreaJitter, wxT("LightAreaJitetr="), wxT("0"));
    mCConfList.Add(&mCSpotAngle,  wxT("LightSpotAngle="), wxT("35"));
    mCConfList.Add(&mCBeamRadius, wxT("LightBeamRadius="), wxT("1"));
    mCConfList.Add(&mCObjectSize, wxT("LightObjectSize="), wxT("0.01"));
    mCConfList.Add(&mCObjectEmission, wxT("LightObjectEmission="), wxT("1"));
}

wxString LightParam_cls::
Serialize(){
    return mCConfList.SerializeAll();
}

bool LightParam_cls::
ReadStrLine(std::wstring strLine){
    return mCConfList.ReadStrLineAll(strLine);//retVal;
}

// --------------------

LightSRC_cls::
LightSRC_cls(FrameMain_cls *pC, wxBitmap *pBitmap)  
{
    SetTag(wxT("[Light]"));
    mpCParentFrame = pC;
    mpBitmapLink = pBitmap;
    mpCGUI = NULL;
    InitGUI(pC);
    mVCLightParam.reserve(_MAX_ITEM_NUMBER);
}

LightSRC_cls::
~LightSRC_cls(){
}

void LightSRC_cls::
InitGUI(wxWindow* pC){
    mpCGUI = new LightSRCGUI_cls(this, &mpCGUI, pC,
				 mCScreenRect.GetPosition(),
				 mCScreenRect.GetSize()
				 );
    SetPGUI(mpCGUI);
    mpCGUI->UpdateComboBox();
}

void LightSRC_cls::DrawGLPath(unsigned int dispList, double scale,
			      PositionAndDirection_cls mCCameraPos){

    for (unsigned int i = 0; i < mpCGUI->mCItemNum.mBody; i++){
	if (mpCGUI->mVCParam[i].mCbVisible.mbBody == FALSE) continue;

	LightParam_cls *pCL = &mVCLightParam[i];

	glPushMatrix();

	PositionAndDirection_cls CCamInThisCoord = mCCameraPos;
	if (pCL->mCbChildOfCamera.mbBody == TRUE){
	    msCParentObjPos.GLSetMatrixToThisCoordinate();
	    CCamInThisCoord = 
		msCParentObjPos.GetPosAndDirectionInThisCoordinate(mCCameraPos);
	}

	MultiPointInput_cls::DrawGLPath(dispList, scale, i);
	
	double pos[3];
	LOOP_XYZ pos[xyz] = mpCGUI->mVCParam[i].mpInterpolatedValue[xyz];

	glLineStipple(1,0xC0C0);
	glColor4f(0.5f, 0.5f, 0.1f, 1.0f);
	glEnable(GL_LINE_STIPPLE);
	glBegin(GL_LINES);
	glVertex3dv(pos); glVertex3d(pos[X], pos[Y], 0);
	glEnd();
	if (pCL->mCbChildOfCamera.mbBody == TRUE){
	    glBegin(GL_LINE_STRIP);
	    glVertex3d(pos[X], pos[Y], 0); glVertex3d(pos[X], 0, 0);
	    glVertex3d(0, 0, 0);
	    glEnd();
	}
	glDisable(GL_LINE_STIPPLE);

	wxString strTag = mTagStr; strTag << i;
	bool bSelected = mpCParentFrame->IsPanelActive(strTag);
	if (bSelected)
	    glColor4f(1.0f, 1.0f, 0.2f, 1.0f);
	else
	    glColor4f(0.7f, 0.7f, 0.1f, 1.0f);

	glEnable(GL_LINE_STIPPLE);
	if (pCL->mCType.mBody == _LIGHT_SPOT || 
	    pCL->mCType.mBody == _LIGHT_CYLINDRICAL){
	    double targetPos[3] = {0,0,0};
	    if (pCL->mCSpotTarget.mBody == _LSPOT_NUMERIC_ABS)
		LOOP_XYZ targetPos[xyz] = pCL->mCTargetPos.mpBody[xyz];
	    glBegin(GL_LINES);
	    glVertex3dv(pos); glVertex3dv(targetPos);
	    glEnd();

	    PositionAndDirection_cls CLightPps;
	    LOOP_XYZ CLightPps.mPos[xyz] = pos[xyz];
	    LOOP_XYZ CLightPps.mTgt[xyz] = targetPos[xyz] - pos[xyz];
	    CLightPps.RejectInvalidDirection();
	    CLightPps.NormalizeAndOrthogonalize();
	    
	    glPushMatrix();
	    GLOBJ::SetMatrix3dv3(CLightPps.mPos, CLightPps.mTgt, CLightPps.mTop);
	    if (pCL->mCType.mBody == _LIGHT_SPOT){
		glScaled(scale*10, scale*10, scale*10);
		double theta = pCL->mCSpotAngle.mBody / 180 * 6.283 / 2;
		double sintheta = sin(theta), costheta = cos(theta);
		for (float phi = 0; phi < 6.283; phi +=0.31415){
		    glBegin(GL_LINE_STRIP);
		    glVertex3d(sintheta*cos(phi)*0.1, sintheta*sin(phi)*0.1, -costheta*0.1);
		    glVertex3d(sintheta*cos(phi), sintheta*sin(phi), -costheta);
		    
		    glVertex3d(sintheta*cos(phi+0.31415), sintheta*sin(phi+0.31415), -costheta);
		    glEnd();
		}
	    }

	    if (pCL->mCType.mBody == _LIGHT_CYLINDRICAL){
		float radius = pCL->mCBeamRadius.mBody;
		for (float phi = 0; phi < 6.283; phi +=0.31415){
		    glBegin(GL_LINES);
		    glVertex3d(radius*cos(phi), radius*sin(phi), 0);
		    glVertex3d(radius*cos(phi), radius*sin(phi), -1);
		    glEnd();
		}
	    }

	    glPopMatrix();
	}
	    
	if (pCL->mCType.mBody == _LIGHT_PARALLEL) {
	    glBegin(GL_LINES);
	    glVertex3dv(pos); glVertex3d(0, 0, 0);
	    glEnd();
	}
	glDisable(GL_LINE_STIPPLE);

	GLOBJ::SetMatrix3dv3(mpCGUI->mVCParam[i].mpInterpolatedValue,
			     CCamInThisCoord.mTgt, CCamInThisCoord.mTop);
	
	glPushName(_HITTEST_LIGHT);
	glPushName(i);
	
	glScaled(scale, scale, scale);
	GLOBJ::Object::Draw(_GLOBJ_LIGHTSRC);

	glPopName();
	glPopName();

	glPopMatrix();
    }
}

void LightSRC_cls::
DrawSelectedPath(GLGuides_cls *pCGLGuide, double dpos[3], 
		 unsigned int listGL, double scale, unsigned int pathIndex){
    glPushMatrix();

    PositionAndDirection_cls CParentObjPos;
    if (mVCLightParam[pathIndex].mCbChildOfCamera.mbBody == TRUE){
	msCParentObjPos.GLSetMatrixToThisCoordinate();
	msCParentObjPos.RotTransformToThisCoordinate(dpos);
	CParentObjPos = msCParentObjPos;
    }
    
    mpCGUI->DrawSelectedPath(pCGLGuide, dpos, listGL, scale, pathIndex,
			     CParentObjPos);
    glPopMatrix();
}

void LightSRC_cls::
TransformCoordinate(unsigned int pathIndex, double vec[3]){
    if (mVCLightParam[pathIndex].mCbChildOfCamera.mbBody == TRUE)
	msCParentObjPos.RotTransformToThisCoordinate(vec);
}

// ----ʒuȊÕȍ---

void LightSRC_cls::
AddLight(unsigned int index, LightParam_cls CParam){
    GCUndoList.SetNewCommand
	(new ComLightParamAdd_cls(this, index, CParam),
	 new ComLightParamRemove_cls(this, index));
    GCUndoList.ExecCurrent();
}

void LightSRC_cls::
AddLightProc(unsigned int index, LightParam_cls CParam){
    mVCLightParam.push_back(LightParam_cls());

    for (int i = (int)mVCLightParam.size() - 1; i > (int)index; i--)
	mVCLightParam[i] = mVCLightParam[i-1];
    
    mVCLightParam[index] = CParam;
    mVCLightParam[index].ReDefineList();
}

void LightSRC_cls::
RemoveLight(unsigned int index){
    GCUndoList.SetNewCommand
	(new ComLightParamRemove_cls(this, index),
	 new ComLightParamAdd_cls(this, index, mVCLightParam[index]));
    GCUndoList.ExecCurrent();
}

void LightSRC_cls::
RemoveLightProc(unsigned int index){
    for (unsigned int i = index + 1; i < mVCLightParam.size(); i++)
	mVCLightParam[i - 1] = mVCLightParam[i];
    mVCLightParam.resize(mVCLightParam.size() - 1);
}

// ----

void LightSRC_cls::
SetParentObjPos(PositionAndDirection_cls *pCParentObjPos){
    msCParentObjPos = *pCParentObjPos;
}

void LightSRC_cls::
GetShadeOffset(double *pShadeOffset, double doubleIndex){
    if (mpCGUI->mCItemNum.mBody == 0) return;

    mpCGUI->mVCParam[0].GetPInterpolatedValue(doubleIndex);

    double lightPos[3];
    LOOP_XYZ lightPos[xyz] = mpCGUI->mVCParam[0].mpInterpolatedValue[xyz];
    
    if (mVCLightParam[0].mCbChildOfCamera.mbBody == TRUE){
	LOOP_XYZ lightPos[xyz] = 0;
	
	LOOP_XYZ lightPos[xyz] += mpCGUI->mVCParam[0].mpInterpolatedValue[X] * msCParentObjPos.mRgt[xyz];
	LOOP_XYZ lightPos[xyz] += mpCGUI->mVCParam[0].mpInterpolatedValue[Y] * msCParentObjPos.mTop[xyz];
	LOOP_XYZ lightPos[xyz] -= mpCGUI->mVCParam[0].mpInterpolatedValue[Z] * msCParentObjPos.mTgt[xyz];
	
	if (mVCLightParam[0].mCType.mBody != _LIGHT_PARALLEL)
	    LOOP_XYZ lightPos[xyz] += msCParentObjPos.mPos[xyz];
    }
    
    LOOP_XYZ pShadeOffset[xyz] = -lightPos[xyz];
    
    double norm = MyVector::GetNorm(pShadeOffset);
    if (norm != 0)
	LOOP_XYZ pShadeOffset[xyz] /= (norm / DF3GL_cls::mspCProperty->mCShadeOffset.mBody);
}

wxColour LightSRC_cls::
GetColour(int index){
    return wxColour(mVCLightParam[index].mCColor.mpBody[0] * 255,
		    mVCLightParam[index].mCColor.mpBody[1] * 255,
		    mVCLightParam[index].mCColor.mpBody[2] * 255);
}

void LightSRC_cls::
SetColour(int index, float clr[3]){
    LOOP_RGB mVCLightParam[index].mCColor.mpBody[rgb] = clr[rgb];
    mpCGUI->mpCMNumParam->mVCPointers
	[mpCGUI->mIndexClr].UpdateValue();
}

int LightSRC_cls::
GetLightInfo(float lightPos[][4], float lightColor[][4],
	     float lightAttenuateL[], float lightAttenuateQ[]){
    int lightNum = mpCGUI->mCItemNum.mBody;
    if (lightNum == 0) {
	LOOP_XYZ lightPos[0][xyz] = 1; lightPos[0][3] = 0;
	LOOP_RGBA lightColor[0][rgba] = 1;
	return 0; }

    if (lightNum > _MAX_ITEM_NUMBER)
	lightNum = _MAX_ITEM_NUMBER;

    for (int index = 0; index < lightNum; index++){
	if (mVCLightParam[index].mCbChildOfCamera.mbBody == TRUE){
	    LOOP_XYZ lightPos[index][xyz] = 0;
	    
	    LOOP_XYZ lightPos[index][xyz] += mpCGUI->mVCParam[index].mpInterpolatedValue[X] * msCParentObjPos.mRgt[xyz];
	    LOOP_XYZ lightPos[index][xyz] += mpCGUI->mVCParam[index].mpInterpolatedValue[Y] * msCParentObjPos.mTop[xyz];
	    LOOP_XYZ lightPos[index][xyz] -= mpCGUI->mVCParam[index].mpInterpolatedValue[Z] * msCParentObjPos.mTgt[xyz];

	    if (mVCLightParam[index].mCType.mBody != _LIGHT_PARALLEL)
		LOOP_XYZ lightPos[index][xyz] += msCParentObjPos.mPos[xyz];
	}
	else
	    LOOP_XYZ lightPos[index][xyz] = mpCGUI->mVCParam[index].mpInterpolatedValue[xyz];

	if (mVCLightParam[index].mCType.mBody == _LIGHT_CYLINDRICAL){
	    if (lightPos[index][X] == 0 && lightPos[index][Y] == 0)
		lightPos[index][Z] = 25;
	    else {
		double norm = sqrt(lightPos[index][X] * lightPos[index][X]+
				   lightPos[index][Y] * lightPos[index][Y]+
				   lightPos[index][Z] * lightPos[index][Z]);
		LOOP_XYZ lightPos[index][xyz] *= 25 / norm;
	    }
	}

	if (mVCLightParam[index].mCType.mBody == _LIGHT_PARALLEL)
	    lightPos[index][3] = 0;
	else
	    lightPos[index][3] = 1;
	LOOP_RGB lightColor[index][rgb] = 
	    mVCLightParam[index].mCColor.mpBody[rgb] * 
	    mpCGUI->mVCParam[index].mpInterpolatedValue[3];
	LOOP_RGB // ߋƂ̌݊Bǂ̃^C~OŏĂ 
	    if (mVCLightParam[index].mCColor.mpBody[rgb] > 1)
		lightColor[index][rgb] /= 255;
	lightColor[index][3] = 1;

	switch(mVCLightParam[index].mCFadeType.mBody){
	case 0:
	    lightAttenuateL[index] = 0;
	    lightAttenuateQ[index] = 0;
	    break;
	case 1:
	    lightAttenuateL[index] = 
		1.0/mVCLightParam[index].mCFadeDistance.mBody;
	    lightAttenuateQ[index] = 0;
	    break;
	case 2:
	    lightAttenuateL[index] = 0;
	    lightAttenuateQ[index] = 
		1.0/mVCLightParam[index].mCFadeDistance.mBody;
	    break;
	}
    }
    return lightNum;
}

void LightSRC_cls::
SetGLLightOptional(int GLindex, int index, 
		   float lightPos[4], float lightColor[4]){
    float lightDir[4] = {0, 0, -1, 0};
    float lightZero[3] = {0, 0, 0};
    if (index < mpCGUI->mCItemNum.mBody &&
	mVCLightParam[index].mCType.mBody == _LIGHT_SPOT){
	LightParam_cls *pCLightParam = &mVCLightParam[index];
	switch(pCLightParam->mCSpotTarget.mBody){
	case _LSPOT_ORIGIN:
	    for (int i = 0; i < 4; i++) lightDir[i] = -lightPos[i];
	    break;
	case _LSPOT_NUMERIC_ABS:
	    LOOP_XYZ lightDir[xyz] = 
		pCLightParam->mCTargetPos.mpBody[xyz] - lightPos[xyz];
	    break;
	case _LSPOT_NUMERIC_REL:
	    LOOP_XYZ lightDir[xyz] = pCLightParam->mCTargetPos.mpBody[xyz];
	    break;
	}
	
	glLightf(GLindex, GL_SPOT_CUTOFF, pCLightParam->mCSpotAngle.mBody);
	glLightfv(GLindex, GL_SPOT_DIRECTION, lightDir);
	glLightfv(GLindex, GL_AMBIENT, lightZero);
	glLightf(GLindex, GL_SPOT_EXPONENT, 35*300.0f/pCLightParam->mCSpotAngle.mBody/pCLightParam->mCSpotAngle.mBody);

	// FIX_ME!! 8߂̃CgAAmbient݂̂̃CgƂĎgB
	for (int i = 0; i < 4; i++) lightColor[i] = 1; 
	glEnable(GL_LIGHT7);
	glLightfv(GL_LIGHT7, GL_POSITION, lightPos);
	glLightfv(GL_LIGHT7, GL_DIFFUSE, lightZero);
	glLightfv(GL_LIGHT7, GL_AMBIENT, lightColor);
	glLightfv(GL_LIGHT7, GL_SPECULAR, lightZero);
    }
    else if (index < mpCGUI->mCItemNum.mBody &&
	     mVCLightParam[index].mCType.mBody == _LIGHT_CYLINDRICAL){
	LightParam_cls *pCLightParam = &mVCLightParam[index];
	//for (int i = 0; i < 4; i++) lightDir[i] = -lightPos[i];
	switch(pCLightParam->mCSpotTarget.mBody){
	case _LSPOT_ORIGIN:
	    for (int i = 0; i < 4; i++) lightDir[i] = -lightPos[i];
	    break;
	case _LSPOT_NUMERIC_ABS:
	    LOOP_XYZ lightDir[xyz] = 
		pCLightParam->mCTargetPos.mpBody[xyz] - lightPos[xyz];
	    break;
	case _LSPOT_NUMERIC_REL:
	    LOOP_XYZ lightDir[xyz] = pCLightParam->mCTargetPos.mpBody[xyz];
	    break;
	}
	
	glLightf(GLindex, GL_SPOT_CUTOFF, pCLightParam->mCBeamRadius.mBody * 57.3 / 25);
	glLightfv(GLindex, GL_SPOT_DIRECTION, lightDir);
	glLightfv(GLindex, GL_AMBIENT, lightZero);
	glLightf(GLindex, GL_SPOT_EXPONENT, 3*25*25 / pCLightParam->mCBeamRadius.mBody / pCLightParam->mCBeamRadius.mBody);
	
	// FIX_ME!! 8߂̃CgAAmbient݂̂̃CgƂĎgB
	for (int i = 0; i < 4; i++) lightColor[i] = 1; 
	glEnable(GL_LIGHT7);
	glLightfv(GL_LIGHT7, GL_POSITION, lightPos);
	glLightfv(GL_LIGHT7, GL_DIFFUSE, lightZero);
	glLightfv(GL_LIGHT7, GL_AMBIENT, lightColor);
	glLightfv(GL_LIGHT7, GL_SPECULAR, lightZero);
    }
    else{
	glLightf(GLindex, GL_SPOT_CUTOFF, 180.0);
	glLightfv(GLindex, GL_SPOT_DIRECTION, lightDir);
	glLightf(GLindex, GL_SPOT_EXPONENT, 0.0f );
    }
}

wxString LightSRC_cls::
GetPovStrLightSourceForFog(double doubleIndex){
    wxString strLightSphere;
    float lightPos[_MAX_ITEM_NUMBER][4], lightColor[_MAX_ITEM_NUMBER][4];
    float lightAttenuateL[_MAX_ITEM_NUMBER], lightAttenuateQ[_MAX_ITEM_NUMBER];
    GetLightInfo(lightPos, lightColor, lightAttenuateL, lightAttenuateQ);

    for (unsigned int i = 0; i < mpCGUI->mCItemNum.mBody; i++){
	mpCGUI->mVCParam[i].GetPInterpolatedValue(doubleIndex);
	if (mpCGUI->mVCParam[i].mpInterpolatedValue[3] <= 0) continue;
	if (mVCLightParam[i].mCType.mBody == _LIGHT_PARALLEL) continue;

	strLightSphere << PovOut3f(L"\tsphere { ", lightPos[i]) << wxT(",") <<  GetShortString(mVCLightParam[i].mCObjectSize.mBody * 0.75) << wxT("}") << ENDLWX;
    }

    return strLightSphere;
}

wxString LightSRC_cls::
GetPovStrLightSource(double doubleIndex){
    wxString strLight;

    float lightPos[_MAX_ITEM_NUMBER][4], lightColor[_MAX_ITEM_NUMBER][4];
    float lightAttenuateL[_MAX_ITEM_NUMBER], lightAttenuateQ[_MAX_ITEM_NUMBER];
    GetLightInfo(lightPos, lightColor, lightAttenuateL, lightAttenuateQ);

    msstrLightSRCSphere = GetPovStrLightSourceForFog(doubleIndex);
    if (msstrLightSRCSphere != wxT("")){
	strLight << wxT("#macro LIGHTSRC_SPHERE()") << ENDLWX
		 << wxT("union{") << ENDLWX;
	strLight << GetPovStrLightSourceForFog(doubleIndex);
	strLight << wxT("}") << ENDLWX << wxT("#end") << ENDLWX;
    }
    
    for (unsigned int i = 0; i < mpCGUI->mCItemNum.mBody; i++){
	mpCGUI->mVCParam[i].GetPInterpolatedValue(doubleIndex);
	if (mpCGUI->mVCParam[i].mpInterpolatedValue[3] <= 0) continue;
	strLight << ENDLWX;
	strLight << L"light_source{" << ENDLWX;

	if (mVCLightParam[i].mCType.mBody == _LIGHT_PARALLEL){
	    double norm2 = 0; 
	    LOOP_XYZ norm2 += lightPos[i][xyz] * lightPos[i][xyz];
	    double norm = sqrt(norm2);
	    if (norm != 0)
		LOOP_XYZ lightPos[i][xyz] *= 500000 / norm;
	}
	
	strLight << PovOut3f(L"\t", lightPos[i]) << ENDLWX;

	strLight << PovOut3f(L"\tcolor rgb ", lightColor[i], 3) << ENDLWX;

	float pointAt[3] = {0, 0, 0};
	switch (mVCLightParam[i].mCSpotTarget.mBody){
	case _LSPOT_NUMERIC_ABS:
	    LOOP_XYZ pointAt[xyz] = mVCLightParam[i].mCTargetPos.mpBody[xyz];
	    break;
	case _LSPOT_NUMERIC_REL:
	    LOOP_XYZ pointAt[xyz] = lightPos[i][xyz] + mVCLightParam[i].mCTargetPos.mpBody[xyz];
	    break;
	}
	

	if (mVCLightParam[i].mCType.mBody == _LIGHT_SPOT){
	    strLight << wxT("\tspotlight") << ENDLWX 
		     << PovOut3f(wxT("\tpoint_at "), pointAt) << ENDLWX
		     << wxT("\tfalloff ") << mVCLightParam[i].mCSpotAngle.mBody << ENDLWX
		     << wxT("\tradius ") << mVCLightParam[i].mCSpotAngle.mBody * 2 / 3 << ENDLWX;
	}

	if (mVCLightParam[i].mCType.mBody == _LIGHT_CYLINDRICAL){
	    strLight << wxT("\tcylinder") << ENDLWX
		     << PovOut3f(wxT("\tpoint_at "), pointAt) << ENDLWX
		     << wxT("\tfalloff ") << mVCLightParam[i].mCBeamRadius.mBody << ENDLWX
		     << wxT("\tradius ") << mVCLightParam[i].mCBeamRadius.mBody * 0.75 << ENDLWX;
	}

	if (mVCLightParam[i].mCType.mBody == _LIGHT_PARALLEL)
	    strLight << wxT("\tparallel") << ENDLWX 
		     << wxT("\tpoint_at <0, 0, 0>") << ENDLWX;

	if (mVCLightParam[i].mCbAreaLight.mbBody == TRUE){
	    float size = mVCLightParam[i].mCAreaSize.mBody;
	    if (mVCLightParam[i].mCType.mBody == _LIGHT_PARALLEL)
		size *= 500000; // ŝ͉ŁI
	    float quality = mVCLightParam[i].mCAreaQuality.mBody;
	    strLight << wxT("\tarea_light ") << wxT("\tx * ") << GetShortString(size) << wxT(", y * ") << GetShortString(size) << wxT(",") << GetShortString(quality) << wxT(",") << GetShortString(quality) << ENDLWX;
	    // ŕȂ
	    if (quality < 6)
		strLight << wxT("\tcircular orient\n\tadaptive 1") << ENDLWX ;
	    else
		strLight << wxT("\tcircular orient\n\tadaptive 2") << ENDLWX ;
	    if (mVCLightParam[i].mCbAreaJitter.mbBody == TRUE)
		strLight << wxT("\tjitter") << ENDLWX;
	}

	if (mVCLightParam[i].mCFadeType.mBody == 1){
	    strLight << wxT("\tfade_power 1") << ENDLWX;
	    strLight << wxT("\tfade_distance ") << mVCLightParam[i].mCFadeDistance.mBody << ENDLWX;
	}
	if (mVCLightParam[i].mCFadeType.mBody == 2){
	    strLight << wxT("\tfade_power 2") << ENDLWX;
	    strLight << wxT("\tfade_distance ") << mVCLightParam[i].mCFadeDistance.mBody << ENDLWX;
	}

	wxString strPig;
	strPig << wxT("\t\tpigment { aoi pigment_map {") << ENDLWX
	       << wxT("\t\t\t[") 
	       << PovOut3f(wxT("0.5 rgb 0.5 * "), lightColor[i], 3) 
	       << PovOut3f(wxT("] [0.6 rgb 1 * "), lightColor[i], 3) 
	       << PovOut3f(wxT("] [1.0 rgb 2 * "), lightColor[i], 3)
	       << wxT("] } }") << ENDLWX;

	if (mVCLightParam[i].mCbObject.mbBody == TRUE){
	    strLight << wxT("\tlooks_like {") << ENDLWX
		     << wxT("\t\tsphere { <0, 0, 0>,") << GetShortString(mVCLightParam[i].mCObjectSize.mBody) <<  wxT("}") << ENDLWX
		     << wxT("\t\tfinish { ambient 0 diffuse 0 emission ") << GetShortString(mVCLightParam[i].mCObjectEmission.mBody) <<wxT("}") << ENDLWX
		     << strPig << ENDLWX
		     << wxT("\t}") << ENDLWX;
	}
	
	if (mVCLightParam[i].mCbShadow.mbBody == FALSE)
	    strLight << L"\tshadowless" << ENDLWX;
	strLight << L"}" << ENDLWX;
    }
    
    /*
    double shadeOffset[3] = {0, 0, 0};
    GetShadeOffset(shadeOffset, doubleIndex);
    strLight << PovOut3d(L"#declare ShadeOffset=", shadeOffset) << L";" << ENDLWX;
    */
    
    return strLight;
}

wxString LightSRC_cls::
Serialize(){
    wxString sOut;
    sOut << wxT("--LightSource") << ENDLWX;
    sOut << mpCGUI->mCItemNum.Serialize();
    for (unsigned int i = 0; i < mpCGUI->mCItemNum.mBody; i++){
	sOut << mpCGUI->mVCParam[i].Serialize();
	sOut << mVCLightParam[i].Serialize();
    }
    sOut << wxT("--FinishLightSource") << ENDLWX; // [_[1s]vɐił܂̂
    return sOut;
}

// ̃ASY1s]vɐił܂
void LightSRC_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    std::wstring::size_type loc = strLine.find(L"--LightSource");
    if (loc != std::wstring::npos){
	std::wstring strLine;
	std::getline(istream, strLine);
	// UNAĂ炶ȂƂȂ̂mCItemNum𒼐ژMĂ̂
	// ʓ|Ƃ
	int itemNumberOrg = mpCGUI->mCItemNum.mBody;
	if (mpCGUI->mCItemNum.ReadStrLine(strLine)){
	    int itemNumberNew = mpCGUI->mCItemNum.mBody;
	    mpCGUI->mCItemNum.mBody = itemNumberOrg;
	    for (int i = itemNumberOrg - 1; i >= 0; i--) {
		mpCGUI->RemoveItemProc(i);
		RemoveLightProc(i);
	    }
	    
	    std::wstring strLine;
	    for (int i = 0; i < itemNumberNew; i++){
		mpCGUI->AddItemProc(NULL, i);
		LightParam_cls CDefaultParam;
		AddLightProc(i, CDefaultParam);

		// i > 0 ̎ɂ͂łstrLineParameter_Light͂
		if (i == 0) std::getline(istream, strLine);
		mpCGUI->mVCParam[i].ReadStrLine(istream, strLine);

		bool bParamLine = TRUE;
		while (bParamLine){
		    std::getline(istream, strLine);
		    if (mVCLightParam[i].ReadStrLine(strLine) == FALSE)
			bParamLine = FALSE;
		}
		if (mpCGUI->mVCParam[i].mCbShow.mbBody == FALSE)
		    mpCGUI->HideTLPanel(i);
	    }
	    mpCGUI->mspCDF3SeqGUI->mpCPanelList->DeactivateAllPanel(TRUE);
	}
	    
	mpCGUI->FuncHookItemNumberChanged(NULL, (MultiParameters_cls*)mpCGUI);
	mpCGUI->UpdateValues();
	//UpdateActiveOnTL();
	mpCGUI->UpdateComboBox();
    }
}

wxString LightSRC_cls::msstrLightSRCSphere;
PositionAndDirection_cls LightSRC_cls::msCParentObjPos;

// ---------------------------

ComLightParamBase_cls::
ComLightParamBase_cls(LightSRC_cls *pCTgt){
    mpCTgt = pCTgt;
}

ComLightParamAdd_cls::
ComLightParamAdd_cls(LightSRC_cls *pCTgt, unsigned int index,
		     LightParam_cls CParam) :
ComLightParamBase_cls(pCTgt), mCParam(CParam)
{
    mIndex = index;
}

bool ComLightParamAdd_cls::
Execute(bool bFlag){
    mpCTgt->AddLightProc(mIndex, mCParam);
    return _COMMAND_NODEPENDENT;
}

ComLightParamRemove_cls::
ComLightParamRemove_cls(LightSRC_cls *pCTgt, unsigned int index) :
    ComLightParamBase_cls(pCTgt)
{
    mIndex = index;
}

bool ComLightParamRemove_cls::
Execute(bool bFlag){
    mpCTgt->RemoveLightProc(mIndex);
    return _COMMAND_NODEPENDENT;
}
 
