
#ifndef _LIGHT_SOURCE_H
#define _LIGHT_SOURCE_H

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include <GL/glut.h>

#include "../POVRAY/clsPovSerialize.h"
#include "../WIDGETCTRL/clsMultiPointInput.h"

#define _LIGHT_POINT 0
#define _LIGHT_SPOT 1
#define _LIGHT_CYLINDRICAL 2
#define _LIGHT_PARALLEL 3

#define _LSPOT_ORIGIN 0
#define _LSPOT_ANCHOR 1
#define _LSPOT_NUMERIC_ABS 2
#define _LSPOT_NUMERIC_REL 3

class LightSRC_cls; 

class LightSRCGUI_cls : public MultiPointInputGUI_cls {
 public:
    LightSRCGUI_cls(LightSRC_cls* pCParentLight,
		    LightSRCGUI_cls **ppToThis, wxWindow *pC,
		    wxPoint pos, wxSize size);
    ~LightSRCGUI_cls();
    static LightSRCGUI_cls* mspCThis;
    static LightSRC_cls    *mspCParentLight;
    
    void OnButtonAdd    (wxCommandEvent &event);
    void OnButtonRemove (wxCommandEvent &event);
    void OnButtonClr    (wxCommandEvent &event);

    static void FuncHookApply(void *pArg, void *pEvent);
    
    void SetMNumParam(wxBoxSizer *pBoxSizer);
    void RemakeIconList();
    void ReconnectMNumParam();
    void HideAndShowAll();
    void UpdateActiveOnTL();

    static void FuncHookType(void *pArg, void *pEvent);

    int mIndexPos, mIndexEmit, mIndexClr;
    int mIndexTargetPos;
    int mIndexFadeType, mIndexFadeDist, mIndexType, mIndexSpotTarget;
    int mIndexOption, mIndexShadow, mIndexChild;
    int mIndexArea, mIndexObject;
    int mIndexAreaParam, mIndexAreaSize, mIndexAreaQuality;
    int mIndexSpotAngle, mIndexBeamRadius;
    int mIndexObjectParam, mIndexObjectSize, mIndexObjectEmission; 

    std::vector < wxIcon > mVLightTypeIcon;
    
    //void DrawLight();

 private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

class LightParam_cls { 
 public:
    LightParam_cls();
    /*
    LightParam_cls(unsigned int color[3], bool bShadow, unsigned int type,
		   bool bChildOfCamera, bool bArea,
		   float areaSize, int areaQuality, float spotAngle);
    */
    void ReDefineList();

    ConfItemList_cls mCConfList;
    ConfItemFLOATARRAY_cls mCColor;
    ConfItemFLOATARRAY_cls mCTargetPos;
    ConfItemBOOL_cls mCbShadow;
    ConfItemBOOL_cls mCbChildOfCamera;
    ConfItemBOOL_cls mCbAreaLight, mCbObject;
    ConfItemUINT_cls mCFadeType;
    ConfItemUINT_cls mCSpotTarget;
    ConfItemFLOAT_cls mCFadeDistance;
    ConfItemUINT_cls mCType;
    ConfItemFLOAT_cls mCAreaSize;
    ConfItemUINT_cls mCAreaQuality;
    ConfItemBOOL_cls mCbAreaJitter;
    ConfItemFLOAT_cls mCSpotAngle, mCBeamRadius;
    ConfItemFLOAT_cls mCObjectSize, mCObjectEmission;
    wxString Serialize();
    bool ReadStrLine(std::wstring strLine);
};

class LightSRC_cls : 
public MultiPointInput_cls,
public PovSerialize_cls
{ 
    friend class LightSRCGUI_cls; 
    
 public:
    LightSRC_cls(FrameMain_cls *pC, wxBitmap *pBitmap);
    ~LightSRC_cls();

    LightSRCGUI_cls* mpCGUI;
    wxBitmap *mpBitmapLink;
    static wxString msstrLightSRCSphere;
    static PositionAndDirection_cls msCParentObjPos;

    void InitGUI(wxWindow *pC);

    void DrawGLPath(unsigned int dispList, double scale,
		    PositionAndDirection_cls mCCameraPos);
    void DrawSelectedPath(GLGuides_cls *pCGLGuide, double dpos[3], 
			  unsigned int listGL, double scale, 
			  unsigned int pathIndex);
    void TransformCoordinate(unsigned int pathIndex, double vec[3]);

    // -----

    std::vector < LightParam_cls > mVCLightParam;
    void AddLight(unsigned int index, LightParam_cls CParam);
    void AddLightProc(unsigned int index, LightParam_cls CParam);
    void RemoveLight(unsigned int index);
    void RemoveLightProc(unsigned int index);

    static void SetParentObjPos(PositionAndDirection_cls *pCParentObjPos);
    void GetShadeOffset(double *pShadeOffset, double doubleIndex);
    wxColour GetColour(int index);
    void SetColour(int index, float clr[3]);
    int  GetLightInfo(float lightPos[][4], float lightColor[][4],
		      float lightAttenuateL[], float lightAttenuateQ[]);
    void SetGLLightOptional(int GLindex, int index, 
			    float lightPos[4], float lightColor[4]);
    wxString GetPovStrLightSourceForFog(double doubleIndex);
    wxString GetPovStrLightSource(double doubleIndex);

    wxString Serialize();
    void ReadStrLine(std::wistream &istream, std::wstring strLine);
};

// ------------------------

#include "../UNDO/clsUndoManager.h"

class ComLightParamBase_cls{ 
 public:
    ComLightParamBase_cls(LightSRC_cls *pCTgt);

    LightSRC_cls *mpCTgt;
};

class ComLightParamAdd_cls : public CommandItem_cls, ComLightParamBase_cls {
 public:
    ComLightParamAdd_cls(LightSRC_cls *pCTgt, unsigned int index,
			 LightParam_cls CParam);
    LightParam_cls mCParam;
    unsigned int mIndex;

    bool Execute(bool bFlag = _COMMAND_REDO);
};

class ComLightParamRemove_cls : public CommandItem_cls, ComLightParamBase_cls {
 public:
    ComLightParamRemove_cls(LightSRC_cls *pCTgt, unsigned int index);

    unsigned int mIndex;
    bool Execute(bool bFlag = _COMMAND_REDO);
};

#endif
