
// "clsPovDiffObject.h"
// (setq c-basic-offset 4)

#include "clsPovDiffObj.h"
#include "clsPovRay.h"

#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/clsDF3SeqCtrl.h"
#include "../DF3/clsDF3Sequence.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../GLOBJ/GLOBJECTS.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"

extern int displayListSphere;
extern int displayListCylinder;
extern int displayListCylinderOpen;
extern int displayListCone;
extern bool GbGLLockForMultithread;

#include "../../wxWidgetEnum.h"

#include <wx/wx.h>

#ifdef WIN32
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for(int xyz = 0; xyz < 3; xyz++)
#define _MANIP_TRANSLATE 0

#define ENDLWX wxT("\n")

MACRO_FilterConstructor(PovDiffObject,,)

void PovDiffObjectGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    PovDiffObjectData_cls *pCData = mpCParent->GetPCData();
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

    pCurr->mTextWidth = 65;
    int boxWidth = 2;

    pCurr->mItemWidth = -1;
    pCurr->mComboWidth = 108;
    
    wxString listType[4] = { wxT("Plane"), wxT("Box"), wxT("Sphere"), wxT("Cylinder/Cone")};
    mIndexObjType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &pCData->mCObjType.mBody, wxT("Object Type"), TRUE, 4, listType);
    
    mIndexShapeParam = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Shape Param"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &pCData->mCShapeParam.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &pCData->mCShapeParam.mBody);

    pCurr->CombinePointer(_GP_BOOL, &pCData->mCbHideWireframe.mbBody, 
			  _("Hide Wire"));
    
    pCurr->FinishSetting(pBoxSizer, FALSE, index);

    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookHideItem, _MN_HOOK_COMBOBOX_CHANGE);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);

    HideAndShowItems();
}

MACRO_SetCustomFunctions(PovDiffObjectGUI_cls,,,)

void PovDiffObjectGUI_cls::
HideAndShowItems(){
    int type = mpCParent->GetPCData()->mCObjType.mBody;
    mpCMNumPanel->ShowItems
	(mIndexShapeParam, 3, 
	 (type == _OBJECT_TYPE_CONE) || (type == _OBJECT_TYPE_ARROW));

    mpCParent->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
}

MACRO_FuncHookHideItem(PovDiffObject);

// -----

PovDiffObjectData_cls::
PovDiffObjectData_cls(FilterCtrl_cls* pCCtrl, 
		      bool bConnectToStaticParam,
		      int initType) : 
FilterWithParameterData_cls(pCCtrl)
{
    mInitType = initType;
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
    
    /*
      wxString strType; strType << initType;
    mCConfList.Add(&mCObjType,   wxT("PovDiffObjectType="), strType);
    mCConfList.Add(&mCShapeParam, wxT("PovDiffObjectShapeParam="), L"1");
    */
    
    Parameters_cls CParam(10, L"PovDiffObject");
    double initParam[10] = {0, 0, 0,
			    1, 1, 1,
			    0, 0, 0, 0};
    CParam.ExecCommandInitSet(initParam);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void PovDiffObjectData_cls::
ReDefineList(){
    wxString strType; strType << mInitType;
    
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCObjType,   wxT("PovDiffObjectType="), strType);
    mCConfList.Add(&mCShapeParam, wxT("PovDiffObjectShapeParam="), L"1");
    mCConfList.Add(&mCbHideWireframe, wxT("PovDiffObjectHideWire="), L"0");
}

MACRO_MakeInstanceAndAddToList(PovDiffObject)

wxString PovDiffObjectData_cls::
SerializeConfPolymorph(){
    return msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();
}

void PovDiffObjectData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void PovDiffObjectData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* PovDiffObjectData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > PovDiffObjectData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > PovDiffObjectData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// ----

PovDiffObject_cls::
PovDiffObject_cls(FilterCtrl_cls *pCParentCtrl, PovDiffObjectData_cls *pCInitData) :
    FilterWithParameter_cls(pCParentCtrl)
{
    mpCData = new PovDiffObjectData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) *GetPCData() = *pCInitData;
    
    ConstructorCommon(pCParentCtrl, _("DiffObject"), _FILTER_POV_DIFFOBJECT);
    mpCGUIPanel = new PovDiffObjectGUI_cls(this);
}

MACRO_Filter_Destructor_Basic(PovDiffObject);

MACRO_AddFilterItemPolymorphP(PovDiffObject,RefreshTop());
MACRO_InitializeStaticMemberP(PovDiffObject);
MACRO_GetPCData(PovDiffObject);

//--

void PovDiffObject_cls::
SetMatrix(Parameters_cls* pParam, int i, double dpos[3], int flag){
    double pos[3] = {
	pParam->GetPKeyFrameValue(i)[0] + dpos[0],
	pParam->GetPKeyFrameValue(i)[1] + dpos[1],
	pParam->GetPKeyFrameValue(i)[2] + dpos[2]};
    Direction_cls CDir;
    GLOBJ::SetMatrix3dv3(pos, CDir.mTgt, CDir.mTop);
    //glScalef(mScale, mScale, mScale);
}

void PovDiffObject_cls::
DrawGlobal(double viewscale){
    if (GetPCData()->mCbValid.mbBody == FALSE) return;
    if (GetPCData()->GetPParamActive()->mCbVisible.mbBody == FALSE) return;

    glPushMatrix();

    SetGLMatrix(mpCData->GetPInterpolated(), NULL);

    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glColor4f(0.0f, 0.0f, 0.0f, 0.01f);

    glEnable(GL_NORMALIZE);
    switch(GetPCData()->mCObjType.mBody){
    case _OBJECT_TYPE_PLANE:
	glScaled(viewscale*0.1, viewscale*0.1, viewscale*0.1);
	glBegin(GL_QUADS);
	glVertex3d(-20, -20, 0); glVertex3d(+20, -20, 0);
	glVertex3d(+20, +20, 0); glVertex3d(-20, +20, 0);
	glEnd();
	break;
    case _OBJECT_TYPE_BOX:
	DrawBox(); break;
    case _OBJECT_TYPE_SPHERE:
	glScaled(0.5, 0.5, 0.5);
	glCallList(displayListSphere); break;
    case _OBJECT_TYPE_CONE:
	glScaled(0.5, 0.5, 0.5);
	if (GetPCData()->mCShapeParam.mBody == 1)
	    glCallList(displayListCylinder);
	else if (GetPCData()->mCShapeParam.mBody == 0)
	    glCallList(displayListCone);
	else 
	    GLOBJ::Cone::DListDraw(90, 4, GetPCData()->mCShapeParam.mBody,
				   FALSE);
	break; 
    }
    glDisable(GL_NORMALIZE);
    
    glPopMatrix();

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
    glBindTexture(GL_TEXTURE_1D, 0);
    glDisable(GL_TEXTURE_1D);
    glDisable(GL_LIGHTING);
    glDepthMask(GL_FALSE);
}

void PovDiffObject_cls::
DrawGLGuide(GLGuides_cls *pCGLGuides, double dpos[3], double scale,
	    int manipMode, double dScale[3], double dRot[4]){
    
    glPushMatrix();
    if (DrawSelectedKey(pCGLGuides, GetPCData()->GetPParamActive(), NULL, NULL, -1,
			dpos, scale, 0, _MANIP_TRANSLATE, NULL) == FALSE)
	DrawGLGuideCommon(pCGLGuides, dpos, scale, manipMode, dScale, dRot,
			  GetPCData()->GetPInterpolated(),
			  GetPCData()->mCObjType.mBody,
			  GetPCData()->mCShapeParam.mBody,
			  GetPCData()->mCbHideWireframe.mbBody);
			  
    glPopMatrix();
}

wxString PovDiffObject_cls::
GetPovStrDiffObj(wxString strCoordinate){
    if (mpCData->mCbValid.mbBody == FALSE) return wxT("");

    wxString strObjGeom;

    switch (GetPCData()->mCObjType.mBody){
    case _OBJECT_TYPE_PLANE:
	// FIX_ME!! ǂ؂H
	strObjGeom 
	    << wxT("\t\tplane{ z, 0") << ENDLWX;
	break;
    case _OBJECT_TYPE_BOX:
	strObjGeom 
	    << wxT("\t\tbox{ <-0.5, -0.5, -0.5> <0.5, 0.5, 0.5>") << ENDLWX;
	break;
    case _OBJECT_TYPE_SPHERE:
	strObjGeom 
	    << wxT("\t\tsphere{ <0, 0, 0>, 0.5") << ENDLWX;
	break;
    case _OBJECT_TYPE_CONE:
	// FIX_ME!! `̕ωI
	strObjGeom
	    << wxT("\t\tcone{ <0, 0, -0.5>, 0.5, <0, 0, 0.5> 0.5") << ENDLWX;
	break;
    }

    strObjGeom << PovOut3d(wxT("\t\t\tscale"), &mpCData->GetPInterpolated()[_OBJECT_SCALE_X], 6) << ENDLWX;
    
    if (mpCData->GetPInterpolated()[_OBJECT_ROT_Y2] != 0)
	strObjGeom << PovOut3(L"\t\t\trotate", 0, 0, mpCData->GetPInterpolated()[_OBJECT_ROT_Y2]) << ENDLWX; 
    if (mpCData->GetPInterpolated()[_OBJECT_ROT_R] != 0)
	strObjGeom << PovOut3(L"\t\t\trotate", 0, mpCData->GetPInterpolated()[_OBJECT_ROT_R], 0) << ENDLWX; 
    if (mpCData->GetPInterpolated()[_OBJECT_ROT_P] != 0)
	strObjGeom << PovOut3(L"\t\t\trotate", mpCData->GetPInterpolated()[_OBJECT_ROT_P], 0, 0) << ENDLWX; 
    if (mpCData->GetPInterpolated()[_OBJECT_ROT_Y] != 0)
	strObjGeom << PovOut3(L"\t\t\trotate", 0, 0, mpCData->GetPInterpolated()[_OBJECT_ROT_Y]) << ENDLWX; 

    strObjGeom << PovOut3d(wxT("\t\t\ttranslate"), &mpCData->GetPInterpolated()[_OBJECT_TRANS_X], 6) << ENDLWX;

    strObjGeom << strCoordinate; 

    strObjGeom << wxT("\t\t\ttranslate < 0.5, 0.5, 0.5>");
    
    strObjGeom << wxT("\t\t}") << ENDLWX;

    return strObjGeom;
}

wxString PovDiffObject_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("DiffObject::");
    
    return strStatus;
}

MACRO_SerializeConfP(PovDiffObject)
MACRO_GetNewPReadDataBlock(PovDiffObject)
