// -*- c -*- 
// (setq c-basic-offset 4)
// "clsPovRay.h"

#include "clsPovFog.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/clsDF3Sequence.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/clsAtoiComboCtrl.h"
#include "../WIDGETCTRL/EX/clsColourDialogEX.h"

#include "clsLightSource.h"

#include "../../wxWidgetEnum.h"

extern wxColourDialogEX* GpCClrDialog;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)
#define ENDLWX wxT("\n")

MACRO_FilterConstructor(PovFog, WithTarget,);

void PovFogGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){

    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

    pCurr->mTextWidth = 65;

    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    
    wxString listTarget[2] = { _("Volume"), _("Global")};
    pCurr->mComboWidth = 70;
    mIndexTarget = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mIndexOnList,
	 _("Fog target"), TRUE, 2, listTarget, 0);

    int indexCustomA = pCurr->CombinePointerButton
	(buttonCtrlPanelItemsCustomA, _("Color"));
    pCurr->mVCPointers[indexCustomA].mpConnectColorFloat 
	= mpCParent->mCData.mCColor.mpBody;
    
    mIndexEmission = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Emission Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCEmission.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 200);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCEmission.mBody);

    mIndexAbsorption = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Absorption Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCAbsorption.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 200);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCAbsorption.mBody);
    
    mIndexScattering = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Scattering Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCScatter.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 200);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCScatter.mBody);

    wxString listFogType[2] = { wxT("Constant"), wxT("Layered")};
    mIndexGlobalType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCData.mCFogGlobalType.mBody,
	 _("Fog type"), TRUE, 2, listFogType);

    mIndexDistance = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Distance Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCDistance.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 10);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCDistance.mBody);

    mIndexLayerOffset = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Fog Offset"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER,
			  &mpCParent->mCData.mCFogLayerOffset.mBody);
    pCurr->SetRange(-200, 200, _MN_LASTITEM, 20);
    pCurr->CombinePointer(_GP_FLOAT,
			  &mpCParent->mCData.mCFogLayerOffset.mBody);

    mIndexLayerScaleHeight = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Scale Height"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER,
			  &mpCParent->mCData.mCFogLayerScaleHeight.mBody);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 50);
    pCurr->CombinePointer(_GP_FLOAT,
			  &mpCParent->mCData.mCFogLayerScaleHeight.mBody);
    
    pCurr->AlignRight(indexCustomA);
    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    pCurr->UpdateValues();

    mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;

    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookHideItem, _MN_HOOK_COMBOBOX);
    SetHookCommonSettings(pCurr, FuncHookHideItem);

    //mpCParent->UpdateTarget(); ??
    
    wxString targetStr = mpCParent->GetLocaleStrLabel(mpCParent->GetPCData()->mCConfTargetStr.msBody);
    UpdateListMain();
    SetTargetLabelStr(targetStr);

    HideAndShowItems();
}

MACRO_SetCustomFunctions(PovFogGUI_cls,ChooseColor(),,)

void PovFogGUI_cls::
ChooseColor(){
    mpCParent->ChooseColor();
    FilterCtrl_cls::FuncHookRefreshTop(mpCMNumPanel, NULL);
}

void PovFogGUI_cls::
HideAndShowItems(){
    if (mpCMNumPanel->mbPanelOpen == FALSE) return;
    
    bool bGlobal = 
	(mpCParent->GetTargetLabelStr() == _("Global")) ? TRUE : FALSE;
    bool bLayered = bGlobal &
	((mpCParent->mCData.mCFogGlobalType.mBody == 1) ? TRUE : FALSE);
    
    mpCMNumPanel->ShowItems(mIndexGlobalType, 1, bGlobal);
    //mpCMNumPanel->ShowItems(mIndexEmission, 3, !bGlobal);
    //mpCMNumPanel->ShowItems(mIndexAbsorption, 3, !bGlobal);
    mpCMNumPanel->ShowItems(mIndexDistance, 3, /*bGlobal*/FALSE);
    mpCMNumPanel->ShowItems(mIndexLayerOffset, 3, bLayered);
    mpCMNumPanel->ShowItems(mIndexLayerScaleHeight, 3, bLayered);
    
    mpCMNumPanel->GetSizer()->Layout();
    mpCParent->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
}

MACRO_FuncHookHideItem(PovFog);

// -------

PovFogData_cls::
PovFogData_cls(){
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
    /*
    // ʂ̃Xgɒǉ
    mCConfList.Add(&mCEmission,   wxT("FogEmission="), L"0.1");
    mCConfList.Add(&mCAbsorption, wxT("FogAbsorption="), L"0");
    mCConfList.Add(&mCScatter, wxT("FogScattering="), L"0");
    mCConfList.Add(&mCFogGlobalType, wxT("FogGlobalType="), L"0");
    mCConfList.Add(&mCDistance,   wxT("FogDistance="), L"10");
    mCConfList.Add(&mCGlobalFogContainer, wxT("FogGlobalContainer="), L"0");
    mCConfList.Add(&mCFogLayerOffset, wxT("FogLayerOffset="), L"-1");
    mCConfList.Add(&mCFogLayerScaleHeight, wxT("FogLayerScaleHeight="), L"0.1");
    mCConfList.Add(&mCColor,      wxT("FogColor="), L"1 1 1");
    mCConfList.Add(&mCConfTargetStr, wxT("FogTargetStr="), L"");
    */
}

void PovFogData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCEmission,   wxT("FogEmission="), L"0.1");
    mCConfList.Add(&mCAbsorption, wxT("FogAbsorption="), L"0");
    mCConfList.Add(&mCScatter, wxT("FogScattering="), L"0");
    mCConfList.Add(&mCScatterType, wxT("FogScatterType="), L"0");
    mCConfList.Add(&mCScatterColor, wxT("FogScatterColor="), wxT("1 1 1"));
    mCConfList.Add(&mCFogGlobalType, wxT("FogGlobalType="), L"0");
    mCConfList.Add(&mCDistance,   wxT("FogDistance="), L"10");
    mCConfList.Add(&mCGlobalFogContainer, wxT("FogGlobalContainer="), L"0");
    mCConfList.Add(&mCFogLayerOffset, wxT("FogLayerOffset="), L"-1");
    mCConfList.Add(&mCFogLayerScaleHeight, wxT("FogLayerScaleHeight="), L"0.1");
    mCConfList.Add(&mCColor,      wxT("FogColor="), L"1 1 1");
    mCConfList.Add(&mCConfTargetStr, wxT("FogTargetStr="), L"");
    mCConfList.Add(&mCContainerHalfSize, wxT("FogContainerSize="), L"100 100 100");
    
    /*
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCEmission,   wxT("FogEmission="));
    mCConfList.Add(&mCAbsorption, wxT("FogAbsorption="));
    mCConfList.Add(&mCScatter, wxT("FogScattering="));
    mCConfList.Add(&mCFogGlobalType, wxT("FogGlobalType="));
    mCConfList.Add(&mCDistance,   wxT("FogDistance="));
    mCConfList.Add(&mCGlobalFogContainer, wxT("FogGlobalContainer="))
    mCConfList.Add(&mCFogLayerOffset, wxT("FogLayerOffset="));
    mCConfList.Add(&mCFogLayerScaleHeight, wxT("FogLayerScaleHeight="));
    mCConfList.Add(&mCColor,      wxT("FogColor="));
    mCConfList.Add(&mCConfTargetStr, wxT("FogTargetStr="));
    */
}

MACRO_MakeInstanceAndAddToList(PovFog)

// ----------

#include "../MYAPI/MsgBox.h"

PovFog_cls::
PovFog_cls(FilterCtrl_cls *pCParentCtrl, PovFogData_cls *pCInitData)
: FilterBase_cls(pCParentCtrl)
{
    if (pCInitData != NULL) mCData = *pCInitData;
    
    ConstructorCommon(pCParentCtrl, _("Fog"), _FILTER_FOG);
    mpCGUIPanel = new PovFogGUI_cls(this);
    mbWithTexture = FALSE;    
    mbRefreshWhenComming = TRUE;
}

MACRO_Filter_Destructor_Basic(PovFog);
MACRO_GetPData(PovFog);
MACRO_GetPCData(PovFog);

MACRO_AddFilterItemPolymorph(PovFog);

// ---

void PovFog_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls* pCurr = 
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));
    
    pCurr->mTextWidth = 100;
    pCurr->mComboWidth = 70;
    
    int boxWidth = 2;

    wxString listFogContainer[2] = { wxT("BOX"), wxT("SPHERE") };
    
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GetPCData()->mCGlobalFogContainer.mBody,
	 _("Fog Container"), TRUE, 2, listFogContainer);

    pCurr->CombinePointer
	(_GP_FLOAT, &GetPCData()->mCContainerHalfSize.mpBody[0], 
	 _("Half Size"));
    
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Scattering Type"));
    wxString listTarget[5] = { _("Isotoropic"), _("Mie Haze"), _("Mie Murky"), _("Rayleigh"), wxT("Henyey-Greenstein") };
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GetPCData()->mCScatterType.mBody,
	 wxT(""), TRUE, 5, listTarget, 0);

    mIndexScatterColor = pCurr->CombinePointerButton
	(wxID_ANY, _("Color"));
    pCurr->mVCPointers[mIndexScatterColor].mpConnectColorFloat 
	= GetPCData()->mCScatterColor.mpBody;
    
    pCurr->mMinimumWidth = 300;
    pCurr->FinishSettingDLG(FALSE);

    pCurr->SetFuncHook(&FuncHookScatterColor, _MN_HOOK_BUTTON);
}

MACRO_TargetUpdateFunctions(PovFog_cls);

void PovFog_cls::
ChooseColor(){
    wxColour CColor(mCData.mCColor.mpBody[0]*255,
		    mCData.mCColor.mpBody[1]*255, 
		    mCData.mCColor.mpBody[2]*255);
    GpCClrDialog->SetColour(CColor);
    if (GpCClrDialog->ShowModal() == wxID_OK){
	mCData.mCColor.mpBody[0] = (float)GpCClrDialog->GetR()/255;
	mCData.mCColor.mpBody[1] = (float)GpCClrDialog->GetG()/255;
	mCData.mCColor.mpBody[2] = (float)GpCClrDialog->GetB()/255;
    }

    mpCGUIPanel->mpCMNumPanel->UpdateValues();
}

// --

void PovFog_cls::
GetColorAtInf(float pColor[3]){
    
    float max[3];
    LOOP_RGB max[rgb] = mCData.mCEmission.mBody * mCData.mCColor.mpBody[rgb]
	* mCData.mCContainerHalfSize.mpBody[0];
    
    if (mCData.mCAbsorption.mBody == 0){
	if (mCData.mCEmission.mBody != 0)
	    LOOP_RGB {
		pColor[rgb] = mCData.mCColor.mpBody[rgb];
		if (pColor[rgb] > max[rgb]) pColor[rgb] = max[rgb];
	    }
	else
	    LOOP_RGB pColor[rgb] = 0;
	return;
    }
    
    LOOP_RGB{
	pColor[rgb] = mCData.mCEmission.mBody / mCData.mCAbsorption.mBody;
	if (mCData.mCColor.mpBody[rgb] == 0)
	    pColor[rgb] = 0;

	if (pColor[rgb] > max[rgb]) pColor[rgb] = max[rgb];
    }
}

double PovFog_cls::
GetFogDistance(){
    //return mCDistance.mBody
    double distance;

    if (mCData.mCEmission.mBody == 0 && mCData.mCAbsorption.mBody == 0)
	return 9999;
    distance = (mCData.mCEmission.mBody < mCData.mCAbsorption.mBody) ? 
	2.0 / mCData.mCAbsorption.mBody : 2.0 / mCData.mCEmission.mBody; 
    return distance;
}

// CHECK_ME!! It looks that PovRay has some bug with using fog and media.
wxString PovFog_cls::
GetPovStrFogGlobal(){
    wxString strFog;

    strFog << wxT("fog {") << ENDLWX
	   << wxT("\tfog_type ") << mCData.mCFogGlobalType.mBody + 1 << ENDLWX
	   << wxT("\tdistance ") << mCData.mCDistance.mBody << ENDLWX
	   << PovOut3(L"\trgb", 
		      (float)mCData.mCColor.mpBody[0],
		      (float)mCData.mCColor.mpBody[1],
		      (float)mCData.mCColor.mpBody[2]) << ENDLWX;
    if (mCData.mCFogGlobalType.mBody == 1){
	strFog << wxT("\tfog_offset ") << mCData.mCFogLayerOffset.mBody << ENDLWX
	       << wxT("\tfog_alt ") << mCData.mCFogLayerScaleHeight.mBody << ENDLWX
	       << wxT("\tup <0, 0, 1>") << ENDLWX;
    }
    strFog << wxT("}") << ENDLWX;

    return strFog;
}

wxString PovFog_cls::
GetPovStrFogGlobalFunc(wxString strFogTag){
    wxString strFogFunc, strOffset;
    if (mCData.mCFogLayerOffset.mBody > 0) 
	strOffset << wxT("-") << mCData.mCFogLayerOffset.mBody;
    else
	strOffset << wxT("+") << -mCData.mCFogLayerOffset.mBody;
    strFogFunc << wxT("#declare FogFunction") << strFogTag << wxT("=function{") << ENDLWX
	       << wxT("max (0, min (1, exp(-(z") << strOffset << wxT("))))") << ENDLWX
	       << wxT("}") << ENDLWX << ENDLWX;
    return strFogFunc;
}

wxString PovFog_cls::
GetPovStrFogGlobal2(wxString strFogTag){
    wxString strFog;
    strFog << wxT("object {") << ENDLWX
	   << wxT("\tdifference {") << ENDLWX;
    
    float boxSizeM[3], boxSizeP[3];
    LOOP_XYZ{
	boxSizeM[xyz] = -GetPCData()->mCContainerHalfSize.mpBody[0];
	boxSizeP[xyz] = GetPCData()->mCContainerHalfSize.mpBody[0];
    }

    if (mCData.mCGlobalFogContainer.mBody == 0){
	if (mCData.mCFogGlobalType.mBody == 1){
	    // ground fog Ȃ獂͂񂾂낤Ƃ\
	    boxSizeM[Z] /= 10; boxSizeP[Z] /= 10;
	}
	strFog << PovOut3f(wxT("\t\tbox{ "), boxSizeM) << PovOut3f(wxT(","), boxSizeP) << wxT("}") << ENDLWX;
    }
    else
	strFog << wxT("\t\tsphere { <0, 0, 0>,") << GetPCData()->mCContainerHalfSize.mpBody[0] << wxT("}") << ENDLWX;
    
    if (LightSRC_cls::msstrLightSRCSphere != wxT(""))
	strFog << wxT("\t\tobject { LIGHTSRC_SPHERE() }") << ENDLWX;
    
    strFog << wxT("\t}") << ENDLWX
	   << wxT("\tpigment { color rgbt < 0, 0, 0, 1> }") << ENDLWX
	   << wxT("\thollow") 
	   << wxT("\tphotons{ pass_through on reflection off refraction on collect off}") << ENDLWX
	   << wxT("\tinterior{") << ENDLWX
	   << wxT("\tmedia { MediaFog") << strFogTag << wxT(" } }") << ENDLWX
	   << wxT("}") << ENDLWX << ENDLWX;
    return strFog;
}

wxString PovFog_cls::
GetPovStrFog(wxString strVolumeRenderParam){
    wxString strFog;
    if (mCData.mCbValid.mbBody == FALSE) return strFog;

    wxString strFogTag;
    strFogTag << mpCParentCtrl->mpCDF3SeqParent-> mSelfIndex << mpCParentCtrl->GetSelfIndex(this);

    //if (GetTargetLabelStr() == _("Global")) return GetPovStrFogGlobal();
    
    bool bGroundFog =  (GetTargetLabelStr() == _("Global") &&
			mCData.mCFogGlobalType.mBody == 1);

    if (bGroundFog)
	strFog << GetPovStrFogGlobalFunc(strFogTag);
    
    strFog << ENDLWX;
    strFog << wxT("#declare MediaFog") << strFogTag << wxT(" = ") << ENDLWX
	   << wxT("media {") << ENDLWX
	   << wxT("\tdensity { ");
    if (bGroundFog)
	strFog << wxT("function { FogFunction") << strFogTag << wxT("(x, y, z) }") << ENDLWX
	       << wxT("\t\tcolor_map { [0.0 rgb <0,0,0> ] [1.0 ") 
	       << PovOut3(L"rgb", 
			  (float)mCData.mCColor.mpBody[0],
			  (float)mCData.mCColor.mpBody[1],
			  (float)mCData.mCColor.mpBody[2])
	       << wxT("] } }") << ENDLWX;
    else
	strFog << PovOut3(L"color rgb", 
			  (float)mCData.mCColor.mpBody[0],
			  (float)mCData.mCColor.mpBody[1],
			  (float)mCData.mCColor.mpBody[2])
	       << wxT("}") << ENDLWX;

    strFog << strVolumeRenderParam;
    /*
    float emission3f[3], absorb3f[3];
    LOOP_XYZ{
	emission3f[xyz] = mCData.mCEmission.mBody;
	absorb3f[xyz]   = mCData.mCAbsorption.mBody; 
    }
    strFog << PovOut3f(L"\temission rgb", emission3f) << ENDLWX
	   << PovOut3f(L"\tabsorption rgb", absorb3f) << ENDLWX
	   << wxT("}") << ENDLWX << ENDLWX;
    */
    strFog << wxT("\temission ") << mCData.mCEmission.mBody << ENDLWX;
    strFog << wxT("\tabsorption ") << mCData.mCAbsorption.mBody << ENDLWX;
    if (mCData.mCScatter.mBody != 0){
	float scatter3f[3];
	LOOP_XYZ scatter3f[xyz] = mCData.mCScatterColor.mpBody[xyz] * mCData.mCScatter.mBody;
	strFog << wxT("\tscattering {") << GetPCData()->mCScatterType.mBody + 1 << PovOut3f(L",rgb", scatter3f) << wxT("}") << ENDLWX;
    }
    strFog << wxT("}") << ENDLWX << ENDLWX;
    
    if (GetTargetLabelStr() == _("Global"))
	strFog << GetPovStrFogGlobal2(strFogTag);
    
    return strFog;
}

wxString PovFog_cls::
GetPovStrFogForUse(){
    wxString strFog;
    strFog << wxT("\tmedia { MediaFog") << mpCParentCtrl->mpCDF3SeqParent->mSelfIndex << mpCParentCtrl->GetSelfIndex(this) << wxT(" }\n");
    return strFog;
}

void PovFog_cls::
FuncHookScatterColor(void *pArg, void *pEvent){
    PovFog_cls *pCThis = mspCConnectedToDLG;

    wxColour CColor(pCThis->GetPCData()->mCScatterColor.mpBody[0]*255,
		    pCThis->GetPCData()->mCScatterColor.mpBody[1]*255,
		    pCThis->GetPCData()->mCScatterColor.mpBody[2]*255);
    wxColourDialogEX CDialog(pCThis->mpCParentCtrl->mpCFilterPanelsLink,
			     CColor, TRUE);
    if (CDialog.ShowModal() == wxID_OK){
	pCThis->GetPCData()->mCScatterColor.mpBody[0] = (float)CDialog.GetR()/255;
	pCThis->GetPCData()->mCScatterColor.mpBody[1] = (float)CDialog.GetG()/255;
	pCThis->GetPCData()->mCScatterColor.mpBody[2] = (float)CDialog.GetB()/255;
	pCThis->mpCMNumAdvanced->mVCPointers[pCThis->mIndexScatterColor].UpdateValue();
    }
}

// ------

wxString PovFog_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("PovFog::");
    
    return strStatus;
}

wxString PovFog_cls::
SerializeConf(){
    UpdateCommonData();
    mCData.mCConfTargetStr.msBody = 
	GetLocaleIndependentStrLabel(GetTargetLabelStr());
    return mCData.SerializeConf(mFilterType);
}

MACRO_ReadDataBlock(PovFog)

PovFog_cls *PovFog_cls::mspCConnectedToDLG;
