// "clsPovIsosurface.h"
// (setq c-basic-offset 4)

#include "clsPovIsosurface.h"
#include "clsPovRay.h"
#include "clsPovFog.h"

#include "../COMMON/clsFilename.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/clsDF3SeqCtrl.h"
#include "../DF3/clsDF3Sequence.h"
#include "../DF3/GL/clsPreviewProperty.h"
#include "../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../FILTER/clsFilterCtrl.h"
#include "../FILTER/clsClipping.h"
#include "../GLOBJ/GLOBJTransCube.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../MYAPI/MsgBox.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/EX/clsCheckBoxEX.h"
#include "../WIDGETCTRL/EX/clsStaticTextEX.h"
#include "../../wxWidgetEnum.h"

#include <GL/glu.h>
#include <wx/colordlg.h>
#include <wx/file.h>
#include <wx/filename.h>
#include <wx/wx.h>

#include <sstream>

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGBA for (int rgba = 0; rgba < 4; rgba++)
#define ENDL std::endl
#define ENDLWX wxT("\n")

#ifdef WIN32
#define DELIMWX wxT("\\")
#else
#define DELIMWX wxT("/")
#endif

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

MACRO_FilterConstructor(PovIsosurface, WithTarget,);

void PovIsosurfaceGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    PovIsosurfaceData_cls *pCData = mpCParent->GetPCData();
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);
    
    pCurr->mTextWidth = /*-1*/65;
    int boxWidth = 2;

    pCurr->mItemWidth = -1;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    wxString listColor[1] = { wxT("Mono Color") };
    mIndexColorType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &pCData->mCTexProperty.mCColorType.mBody,
	 wxT(""), TRUE, 1, listColor);

    InitializeColorPanel(pCurr, this, &GetColorR, &GetColorG, &GetColorB,
			 _("Texture"));
    mIndexColorBar = pCurr->CombinePointer(_GP_WXPANEL, mpCPanelColorBar);
    
    // lݒ肵Ă
    pCData->GetPInterpolated()[_ISO_CRIT] = 0.5;
    pCData->GetPInterpolated()[_ISO_ALPHA] = 1;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Critical Value"));
    pCurr->CombinePointer(_GP_DOUBLE_SLIDER, 
			  &pCData->GetPInterpolated()[_ISO_CRIT]);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_DOUBLE, 
			  &pCData->GetPInterpolated()[_ISO_CRIT]);
    
    wxString listTarget[2] = { _("Solid"), _("Empty") };
    pCurr->mComboWidth = 108;
    mIndexTarget = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mIndexOnList,
	 _("Interior"), TRUE, 2, listTarget, 1);
    
    pCurr->mItemWidth = -1;

    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    //mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;
    
    UpdateColorList();

    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FilterWithTextureGUI_cls::FuncHookUpdateTextureDLG,
		       _MN_HOOK_COMBOBOX_CHANGE);
    pCurr->SetFuncHook(&FuncHookOpenTextureDLG, _MN_HOOK_PANELCLICK);
    SetHookCommonSettings(pCurr, &FuncHookHideItem);

    wxString targetStr = mpCParent->GetLocaleStrLabel(mpCParent->GetPCData()->mCConfTargetStr.msBody);
    UpdateListMain();
    SetTargetLabelStr(targetStr);
}

MACRO_SetCustomFunctions(PovIsosurfaceGUI_cls,,,);

void PovIsosurfaceGUI_cls::
HideAndShowItems(){
    if (mpCMNumPanel->mbPanelOpen == FALSE) return;
    
    RefreshColorBar();

    mpCMNumPanel->GetSizer()->Layout();
    mpCParent->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
}

MACRO_FuncHookHideItem(PovIsosurface);

void PovIsosurfaceGUI_cls::
HookButton(void *pArg, void *pEvent){
    PovIsosurface_cls *pCParent = PovIsosurface_cls::mspCConnectedToDLG;
    int vertNum = pCParent->mCMarchingCube.GetVertexNumber();
    int vertNumOpt = pCParent->mCMarchingCube.GetVertexNumberOpt();
    int vertNumOptFull = pCParent->mCMarchingCube.GetVertexNumberOptFull();
    float density = pCParent->mCMarchingCube.GetVertexDensity();
    float maxDesity = 
	pCParent->mCMarchingCube.GetMaxVertexDensity
	(PovIsosurface_cls::mspCPreferenceLink->mCThreadIsosurfaceLayer.mBody);
    MsgBoxf(pCParent->mpCMNumAdvanced, "Polygon Information", "%d triangles\n%d vertices (%d MB)\n%d vertices (%d MB) light optimize\n%d vertices (%d MB) middle optimize\nmean density=%f\nmax density=%f\n", 
	    vertNum/3, vertNum, vertNum*3*4/1024/1024, 
	    vertNumOpt, vertNumOpt*3*4/1024/1024, 
	    vertNumOptFull, vertNumOptFull*3*4/1024/1024,
	    density, maxDesity);
}

void PovIsosurfaceGUI_cls::
HookButton2(void *pArg, void *pEvent){
    PovIsosurface_cls *pCParent = PovIsosurface_cls::mspCConnectedToDLG;
    pCParent->mCMarchingCube.OptimizeVertices();
    pCParent->mCMarchingCube.OptimizeVerticesFull();
}

// ----------

PovIsosurfaceData_cls::
PovIsosurfaceData_cls(FilterCtrl_cls* pCCtrl,
		      bool bConnectToStaticParam) :
    FilterWithParameterData_cls(pCCtrl)
{
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;

    Parameters_cls CParam(2, L"Isosurface");
    double initParam[2] = {0.5, 1};
    CParam.ExecCommandInitSet(initParam);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void PovIsosurfaceData_cls::
ReDefineList(){
    // ʂ̃Xgɒǉ
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCbInverse,wxT("IsosurfaceInverse="), L"0");
    mCConfList.Add(&mCConfTargetStr, wxT("IsosurfInteriorStr="), L"");

    mCTexProperty.mCConfList.mbSetWithInitValue = mCConfList.mbSetWithInitValue;
    mCTexProperty.ReDefineList();
}

MACRO_MakeInstanceAndAddToList(PovIsosurface)

wxString PovIsosurfaceData_cls::
SerializeConfPolymorph(){
    return mCTexProperty.Serialize() + msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();
}

void PovIsosurfaceData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    mCTexProperty.ReadStrLine(strLine);
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void PovIsosurfaceData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* PovIsosurfaceData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > PovIsosurfaceData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > PovIsosurfaceData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// ----------

#include "../MYAPI/MsgBox.h"

PovIsosurface_cls::
PovIsosurface_cls(FilterCtrl_cls *pCParentCtrl, PovIsosurfaceData_cls *pCInitData) :
    FilterWithParameter_cls(pCParentCtrl)
{
    mpCData = new PovIsosurfaceData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) *GetPCData() = *pCInitData;
    GetPCData()->mCTexProperty.LinkFilterParameter(this, _ISO_ALPHA);

    ConstructorCommon(pCParentCtrl, _("Isosurface"), _FILTER_ISOSURFACE);
    mpCGUIPanel = new PovIsosurfaceGUI_cls(this);
    ((PovIsosurfaceGUI_cls*)mpCGUIPanel)->mIndexOnList = 1;
}

PovIsosurface_cls::
~PovIsosurface_cls(){
    if (FilterWithTextureGUI_cls::mspCLinkTexpropertyCtrl->mpCConnected == &GetPCData()->mCTexProperty)
	FilterWithTextureGUI_cls::mspCLinkTexpropertyCtrl->ConnectProperty(NULL, wxT("No Item is Linked"));
    if (mspCDF3SeqGUI != NULL)
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
}

MACRO_TargetUpdateFunctions(PovIsosurface_cls);

MACRO_AddFilterItemPolymorphP(PovIsosurface,RefreshTop());
MACRO_InitializeStaticMemberP(PovIsosurface);
MACRO_GetPCData(PovIsosurface);

// --

void PovIsosurface_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls* pCurr =
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));
	
    pCurr->mTextWidth = 100;

    wxString showModeList[2] =
	{wxT("GL_TRIANGLES"), wxT("GL_LINES")};
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &msCShowMode.mBody,
			  _("polygon type"), TRUE, 2, showModeList);
    
    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    pCurr->CombinePointer(_GP_BOOL, &msCbCullFace.mbBody, _("Cull Backside Polygon"));
    pCurr->CombinePointer(_GP_BOOL, &msCbTrimEdge.mbBody, _("Trim Edge"));
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    pCurr->CombinePointer(_GP_BOOL, &msCbReTexOptimize.mbBody, _("Optimize at ReTex"));
    pCurr->CombinePointer(_GP_BOOL, &msCbOffset05.mbBody, _("Offset at Export"));
    
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    pCurr->CombinePointer(_GP_BOOL, &msCbExportNormal.mbBody, _("Smooth/Flat"));
    pCurr->CombinePointer(_GP_BOOL, &msCbHighPrecNormal.mbBody, _("HQ Normal"));
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    pCurr->CombinePointer(_GP_BOOL, &GetPCData()->mCbInverse.mbBody, _("Inverse"));
    pCurr->CombinePointer(_GP_BOOL, &msCbClipForPeriodic.mbBody, _("Clip for Periodic"));
    
    int indexButton = pCurr->CombinePointerButton
	(wxID_ANY, wxT("PolygonNumber"));
    pCurr->CombinePointerButton(wxID_ANY, wxT("Do optimize now"));
    
    pCurr->FinishSettingDLG(FALSE);
    pCurr->SetScreenRectPos(&msCRectMNumStaticMembers);
    
    pCurr->mVCPointers[2].GetPCtrlWindow()->
	SetToolTip(_("Hide polygons facing inside for better performance."));
    pCurr->mVCPointers[3].GetPCtrlWindow()->
	SetToolTip(_("Force isosurface to be closed, setting edge vlume data to 0. "));
    pCurr->mVCPointers[5].GetPCtrlWindow()->
	SetToolTip(_("Do Optimize at Retexuring.\nSaving memory and fast at drawing but slow at calculation."));
    pCurr->mVCPointers[6].GetPCtrlWindow()->
	SetToolTip(_("Offset vertices in exporting .obj file.\nSetting the origin to the center of the box."));
    
    pCurr->SetFuncHook(&PovIsosurfaceGUI_cls::HookButton, _MN_HOOK_BUTTON, 
		       pCurr->mVCPointers[indexButton].mID);
    pCurr->SetFuncHook(&PovIsosurfaceGUI_cls::HookButton2, _MN_HOOK_BUTTON, 
		       pCurr->mVCPointers[indexButton+1].mID);
}

void PovIsosurface_cls::
AdvancedSetting(){
    AdvancedSettingCommon();
    FilterCtrl_cls::FuncHookRefreshTop(mpCGUIPanel->mpCMNumPanel, NULL);
}

void PovIsosurface_cls::
UpdateColors(){
    ((PovIsosurfaceGUI_cls*)mpCGUIPanel)->UpdateColorList();
}

void PovIsosurface_cls::
ExportWavefrontOBJ(wxString strFilePath){

    mCMarchingCube.OptimizeVertices();
    mCMarchingCube.OptimizeVerticesFull();
    
    wxFile CFile(strFilePath, wxFile::write);
    
    wxString strObj;
    mCMarchingCube.GetStrWaveFrontObj(&strObj, msCbOffset05.mbBody,
				      msCbExportNormal.mbBody);
    CFile.Write(strObj);
}

void PovIsosurface_cls::
ExportMesh2Pov(wxString strFilePath){
    mCMarchingCube.OptimizeVertices();
    mCMarchingCube.OptimizeVerticesFull();
    
    wxString strTag = wxT("0");
    wxString strTime = wxT("");
    if (msCbMesh2Sequential.mbBody == TRUE)
	strTime = wxT("_0000"); // FIX_ME!! Ԃ
    
    SplitPathWX_cls CPath(strFilePath);
    wxString strFilePathWithTag;
    strFilePathWithTag << CPath.GetPathWXStr(_FN_DRIVE_DIR_FNAME) << mstrId << CPath.GetPathWXStr(_FN_EXT);
    
    wxFile CFile(strFilePathWithTag, wxFile::write);
    
    wxString strObj;
    mCMarchingCube.GetStrPovMesh2(&strObj, mstrId, strTag, 
				  msCbExportNormal.mbBody);
    
    CFile.Write(strObj);
}

bool PovIsosurface_cls::
IsReprocessRequired(){
    if (mpCData->mCbValid.mbBody == FALSE) return FALSE;
    if (mTargetValue != mpCData->GetPInterpolated()[_ISO_CRIT])
	return TRUE;
    return FALSE;
}

void PovIsosurface_cls::
ClearIsosurface(){
    mCMarchingCube.Clear(0);
}

void PovIsosurface_cls::
RefreshFilter(unsigned char *pData){
    if (mpCParentCtrl->mpCDF3SeqParent->GetPDF3()->IsAllocated() == FALSE)
	return;

    bool bChanged = mpCParentCtrl->IsReprocessRequired
	(GetRemappedIndex(), GetRemappedIndexDouble(), this);
    if (msbRetex == FALSE)
	if ((bChanged | IsReprocessRequired()) == FALSE) return;

    if (GbGLLockForMultithread == TRUE) return; // ɂȂ͂H
    GbGLLockForMultithread = TRUE;
    
    int dataSlotIndex = mpCParentCtrl->mpCDF3SeqParent->mSelfIndex;
    if (pData == NULL)
	pData = mpCParentCtrl->mpCDF3SeqParent->GetPDF3()->mpMainXYZ[Z];
    
    int sizeX = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(X, dataSlotIndex);
    int sizeY = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Y, dataSlotIndex);
    int sizeZ = mspCDF3SeqCtrlLink->GetCurrentSizeOf1Data(Z, dataSlotIndex);
    mCMarchingCube.mTargetValue = mpCData->GetPInterpolated()[_ISO_CRIT];

    double min[3] = {0, 0 ,0}, max[3] = {1, 1, 1};
    //GetClippingInformation(min, max); ČvZ̕Kv̂肪ƂȂ܂œ
    
    mCMarchingCube.ConstructVertices
	(pData, sizeX, sizeY, sizeZ, min[Z], max[Z],
	 mspCPreferenceLink->mCbSSEFilter.mbBody,
	 mspCPreferenceLink->mCThreadIsosurfaceLayer.mBody,
	 bChanged,
	 msCbReTexOptimize.mbBody, msCbTrimEdge.mbBody,
	 msCbHighPrecNormal.mbBody);
    
    /*
    if (msCbReTexOptimize.mbBody == TRUE)
	mCMarchingCube.OptimizeVertices
	    (mspCPreferenceLink->GetThreadNumFilter());
    */

    mTargetValue = mpCData->GetPInterpolated()[_ISO_CRIT];

    GbGLLockForMultithread = FALSE;
}

void PovIsosurface_cls::
DrawBox(double center[3], double scale[3]){
    
    if (GetPCData()->mCbValid.mbBody == FALSE) return;
    if (GetPCData()->GetPParamActive()->mCbVisible.mbBody == FALSE) return;

    if (scale[X] == 0 || scale[Y] == 0 || scale[Z] == 0) return;

    double min[3] = {0, 0, 0};
    double max[3] = {1, 1, 1};
    GetClippingInformation(min, max);
    
    glPushMatrix();
    glTranslated(-0.5, -0.5, -0.5);

    bool bClipping[3] = {TRUE, TRUE, TRUE};
    LOOP_XYZ if (min[xyz] == 0 && max[xyz] == 1) bClipping[xyz] = FALSE;

    if (bClipping[X] == TRUE){
	double PlaneEqXmin[]={+1.0, 0, 0, -min[X]};
	double PlaneEqXmax[]={-1.0, 0, 0, max[X]};
	glClipPlane(GL_CLIP_PLANE0, PlaneEqXmin);
	glClipPlane(GL_CLIP_PLANE1, PlaneEqXmax); }
    if (bClipping[Y] == TRUE){
	double PlaneEqYmin[]={0, +1.0, 0, -min[Y]};
	double PlaneEqYmax[]={0, -1.0, 0, max[Y]};
	glClipPlane(GL_CLIP_PLANE2, PlaneEqYmin);
	glClipPlane(GL_CLIP_PLANE3, PlaneEqYmax); }
    if (bClipping[Z] == TRUE){
	double PlaneEqZmin[]={0, 0, +1.0, -min[Z]};
	double PlaneEqZmax[]={0, 0, -1.0, max[Z]};
	glClipPlane(GL_CLIP_PLANE4, PlaneEqZmin);
	glClipPlane(GL_CLIP_PLANE5, PlaneEqZmax); }

    //GLOBJ::DrawWireBoxMinMax(min, max);

    glEnable(GL_LIGHTING);
    glDepthMask(GL_TRUE);
    
    if (GetPCData()->mCTexProperty.mCRenderType.mBody == _TEX_RENDER_ADD){
	glDisable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    }
    else {
	glEnable(GL_DEPTH_TEST);	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }

    double critValue = mpCData->GetPInterpolated()[_ISO_CRIT];
    
    if (bClipping[X] == TRUE){
	glEnable(GL_CLIP_PLANE0);
	glEnable(GL_CLIP_PLANE1); }
    if (bClipping[Y] == TRUE){
	glEnable(GL_CLIP_PLANE2);
	glEnable(GL_CLIP_PLANE3); }
    if (bClipping[Z] == TRUE){
	glEnable(GL_CLIP_PLANE4);
	glEnable(GL_CLIP_PLANE5); }

    GLuint showType = (msCShowMode.mBody == 0) ? GL_TRIANGLES : GL_LINES;
    mCMarchingCube.GLDraw(center, scale, msCbCullFace.mbBody, showType,
			  &(GetPCData()->mCTexProperty),
			  critValue);
    
    glDisable(GL_CLIP_PLANE5);
    glDisable(GL_CLIP_PLANE4);
    glDisable(GL_CLIP_PLANE3);
    glDisable(GL_CLIP_PLANE2);
    glDisable(GL_CLIP_PLANE1);
    glDisable(GL_CLIP_PLANE0);

    glPopMatrix();
    
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
    glBindTexture(GL_TEXTURE_1D, 0);
    glDisable(GL_TEXTURE_1D);
    glDisable(GL_LIGHTING);
    glDepthMask(GL_FALSE);
    glEnable(GL_DEPTH_TEST);
}

wxString PovIsosurface_cls::
GetMesh2PovFilePath(wxString strTime){
    wxString strBase = msCMesh2Path.msBody;
    
    if (msCbMesh2Sequential.mbBody == TRUE){
	FileSequenceWX_cls CSeq(msCMesh2Path.msBody);
	if (CSeq.mDigitLength == -1)
	    CSeq.AppendDefaultSequenceTag();
	strBase = CSeq.GetModifiedWXStr(0);
    }
    else
	strBase = msCMesh2Path.msBody;
    
    SplitPathWX_cls CPath(strBase);
    /*
    if (CPath.GetPathWXStr(_FN_DRIVE_DIR) == wxT(""))
	strBaseWithTag << CPath.GetPathWXStr(_FN_FNAME);
    else
	strBaseWithTag << CPath.GetPathWXStr(_FN_DRIVE_DIR) << DELIMWX << CPath.GetPathWXStr(_FN_FNAME);
    */
        
    // ΃pXƂĂ΁AԂ
    wxFileName CTmpFileAbs(strBase);
    if (CTmpFileAbs.IsFileReadable() == TRUE)
	return strBase;
    
    wxString strBaseWithTag;
    strBaseWithTag << CPath.GetPathWXStr(_FN_FNAME) 
		   << mstrId << strTime << CPath.GetPathWXStr(_FN_EXT);
    
    // WorkingDirɃt@C΁At@CԂB
    wxString strRelative = msstrWorkingDir;
    strRelative << strBaseWithTag;
    wxFileName CTmpFileRel(strRelative);
    if (CTmpFileRel.IsFileReadable() == TRUE)
	return strBaseWithTag;
    
    return wxT("Isosurface.pov");
}
    
void PovIsosurface_cls::
GetPovStrIsosurfBox(wxString *pstrTarget,
		    wxString strTag, wxString strFilePath,
		    wxString strGeometry, float treeDivideParam,
		    unsigned int depth, double baseAlpha,
		    double clipForPeriodicBoundary[3],
		    DrawingBlock_cls* pCDrawingBlock){
    
    if (mpCData->mCbValid.mbBody == FALSE) return;
    if (IsClippingEmpty() == TRUE) return;

    if (msCbClipForPeriodic.mbBody == FALSE)
	LOOP_XYZ clipForPeriodicBoundary[xyz] = 1;

    for (unsigned int i = 0; i < mVSAlreadyOutputedTag.size(); i++){
	if (strTag == mVSAlreadyOutputedTag[i])
	    return;
    }
    mVSAlreadyOutputedTag.push_back(strTag);
    
    switch (msCOutputType.mBody){
    case _ISO_LINK_TO_DF3:
	*pstrTarget << GetPovStrIsoFunction
	    (mstrId + strTag, L"\""+ strFilePath+ L"\"", strGeometry);
	
	*pstrTarget << GetPovStrIsoSurf
	    (L"ISOSURF_" + mstrId + strTag, L"ISOFUNC_" + mstrId + strTag, clipForPeriodicBoundary);
	break;
    case _ISO_LINK_TO_POV: {
	wxString strTime = wxT("");
	if (msCbMesh2Sequential.mbBody == TRUE)
	    strTime = wxT("_0000"); // FIX_ME!! Ԃ
	wxString strMesh2PovFile = GetMesh2PovFilePath(strTime); 
	wxString strIsoLink;
	wxFileName CTmpFileAbs(strMesh2PovFile);
	if (CTmpFileAbs.IsFileReadable() == TRUE)
	    strIsoLink << wxT("#declare ISOLIN=\"") << strMesh2PovFile << wxT("\"");
	else
	    strIsoLink << wxT("#declare ISOLINK=concat(WORKINGDIR, \"") << strMesh2PovFile << wxT("\")");
	*pstrTarget << strIsoLink << ENDLWX;
	*pstrTarget << wxT("#include ISOLINK") << ENDLWX;
	break; }
    case _ISO_PACK_IN_POV:
	mCMarchingCube.OptimizeVertices();
	mCMarchingCube.OptimizeVerticesFull();
	/*
	wxString strObj;
	mCMarchingCube.GetStrPovMesh2(&strObj, mstrId, strTag, msCbExportNormal.mbBody);
	*pstrTarget << strObj;
	*/
	mCMarchingCube.GetStrPovMesh2(pstrTarget, mstrId, strTag, msCbExportNormal.mbBody);
	break;
    }
    
    *pstrTarget << wxT("#declare ISOFINISH_") << mstrId + strTag << wxT("=finish{") << ENDLWX;
    *pstrTarget << GetPCData()->mCTexProperty.GetPovStrFinish(TRUE)
	      << wxT("}") << ENDLWX;

    *pstrTarget << GetPovStrIsoTex(mstrId + strTag, strFilePath);
    
    if (strFilePath != wxT(""))
	*pstrTarget << GetPovStrIsoBox(mstrId + strTag, strTag, clipForPeriodicBoundary, pCDrawingBlock);
    else
	*pstrTarget << GetPovStrIsoBox(mstrId + strTag, wxT("Empty"), clipForPeriodicBoundary, pCDrawingBlock);
    
    //return *pstrTarget;
}

wxString PovIsosurface_cls::
GetPovStrIsoFunction(wxString strTag, wxString strDensityFile, 
		     wxString strGeometry){
    wxString strFunction;

    strFunction << ENDLWX;
    strFunction << L"#declare " << wxT("ISOFUNC_") << strTag << L"=function{" << ENDLWX;
    if (strDensityFile != wxT("\"\"")){
	strFunction << L"\tpattern{" << ENDLWX
		    << L"\t\tdensity_file df3 " << strDensityFile << ENDLWX
		    << L"\t\tinterpolate 1" << ENDLWX
		    << L"\t\tscale " << strGeometry << ENDLWX
		    << L"\t}" << ENDLWX;
    }
    else
	strFunction << wxT("0");
    strFunction	    << L"}" << ENDLWX;
    
    return strFunction;
}

wxString PovIsosurface_cls::
GetPovStrIsoSurf(wxString strTag, wxString strFunc,
		 double clipForPeriodicBoundary[3]){
    wxString strIsoSurf;

    double min[3] = {0, 0 ,0};
    double max[3] = {1, 1, 1};
    GetClippingInformation(min, max);
    LOOP_XYZ if (max[xyz] > clipForPeriodicBoundary[xyz])
	max[xyz] = clipForPeriodicBoundary[xyz];
    LOOP_XYZ if (min[xyz] < 0.0005) min[xyz] = 0.0005; // Volumei
    LOOP_XYZ if (max[xyz] > 0.9995) max[xyz] = 0.9995;
    
    double critValue = mpCData->GetPInterpolated()[_ISO_CRIT];
    // t֐̂ ad-hoc ɌvZ
    double critLeveled = 0;
    if (mpCParentCtrl->GetTotalLevelMapping(0, this) < critValue)
	for (double level = 0; level < 1; level +=0.0025){
	    if (mpCParentCtrl->GetTotalLevelMapping(level, this) >= critValue){
		critLeveled = level; level = 1; continue;
	    }
	}
    else
	for (double level = 0; level < 1; level +=0.0025){
	    if (mpCParentCtrl->GetTotalLevelMapping(level, this) <= critValue){
		critLeveled = level; level = 1; continue;
	    }
	}
    
    strIsoSurf << ENDLWX;
    strIsoSurf << L"#declare " << strTag << L"=isosurface{" << ENDLWX;
    if (GetPCData()->mCbInverse.mbBody == TRUE)
	strIsoSurf << wxT("\tfunction { + ") << strFunc << L"(x, y, z) }" << ENDLWX;
    else
	strIsoSurf << wxT("\tfunction { - ") << strFunc << L"(x, y, z) }" << ENDLWX;
    strIsoSurf << L"\tcontained_by { box {" << PovOut3d(wxT(""), min) << wxT(",") << PovOut3d(wxT(""), max) << wxT(" } }") << ENDLWX;
    
    if (GetPCData()->mCbInverse.mbBody == TRUE)
	strIsoSurf << L"\tthreshold " << +critLeveled << ENDLWX;
    else
	strIsoSurf << L"\tthreshold " << -critLeveled << ENDLWX;
    strIsoSurf << L"\taccuracy " << msCAccuracy.mBody << ENDLWX;
    strIsoSurf << L"\tmax_gradient " << msCMaxGradient.mBody << ENDLWX;
    if (msCTypeGradient.mBody == 1)
	strIsoSurf << L"\tevaluate " << msCEvaluate.mpBody[0] << L", " << msCEvaluate.mpBody[1] << wxT(", ") << msCEvaluate.mpBody[2] << ENDLWX;
    
    if (msCMaxTrace.mBody == 0)
	strIsoSurf << L"\tall_intersections" << ENDLWX;
    else
	strIsoSurf << L"\tmax_trace " << msCMaxTrace.mBody << ENDLWX;
    //strIsoSurf << L"\ttranslate < -0.5, -0.5, -0.5 >" << ENDLWX;
    strIsoSurf << L"}" << ENDLWX;

    return strIsoSurf;
}

wxString PovIsosurface_cls::
GetPovStrIsoTex(wxString strTag, wxString strDensityFile){
    wxString strIsoTex, strFinishTag;
    strFinishTag << wxT("ISOFINISH_") << strTag;
    
    strIsoTex << ENDLWX;

    strIsoTex << L"#declare " 
	      << wxT("ISOTEX_") << strTag << L"=texture{" << ENDLWX;

    if (GetPCData()->mCTexProperty.mCColorType.mBody == 0)
	strIsoTex << GetPCData()->mCTexProperty.GetPovStrTex(mpCData->GetPInterpolated()[_ISO_ALPHA], strFinishTag);
    else {
	wxString strClrMap, strClr1, strClr2, strClr3;
	float trans = 1.0f - mpCData->GetPInterpolated()[_ISO_ALPHA];
	int transmitType = GetPCData()->mCTexProperty.mCTransmitType.mBody;
	switch (GetPCData()->mCTexProperty.mCMappingType.mBody){
	case _TEX_MAPPING_VALUE: 
	    strClrMap = wxT("MONOCOLOR");
	    strClr1 = GCColoring.GetPovStrClr
		(mpCParentCtrl->mSelfDF3Index,
		 GetPCData()->mCTexProperty.mCColorType.mBody-1 + _CLR_CUSTOM,
		 mpCData->GetPInterpolated()[_ISO_CRIT],
		 trans, transmitType);
	    break;
	case _TEX_MAPPING_PATTERN: {
	    std::vector < wxString > Vstr = 
		GetPCData()->mCTexProperty.GetVPovStrColorMap(_DF3_SLOT_0_FIX_ME, trans);
	    strClrMap = Vstr[0];
	    strClr1 = Vstr[1];
	    strClr2 = Vstr[2];
	    strClr3 = Vstr[3];
	    /*
	    strClrMap = GCColoring.GetPovStrColorMap
		(mpCParentCtrl->mSelfDF3Index,
		 GetPCData()->mCTexProperty.mCColorType.mBody-1 + _CLR_CUSTOM, 
		 FALSE, 0, NULL, FALSE, trans, transmitType,
		 GetPCData()->mCTexProperty.IsHalfMapping(), FALSE);
	    strClr1 = GCColoring.GetPovStrClr
		(mpCParentCtrl->mSelfDF3Index,
		 GetPCData()->mCTexProperty.mCColorType.mBody-1 + _CLR_CUSTOM,
		 0, trans, transmitType);
	    strClr2 = GCColoring.GetPovStrClr
		(mpCParentCtrl->mSelfDF3Index,
		 GetPCData()->mCTexProperty.mCColorType.mBody-1 + _CLR_CUSTOM,
		 1, trans, transmitType);
	    strClr3 = GCColoring.GetPovStrClr
		(mpCParentCtrl->mSelfDF3Index,
		 GetPCData()->mCTexProperty.mCColorType.mBody-1 + _CLR_CUSTOM,
		 0.5, trans, transmitType);
	    */
	    break; }
	}
	
	strIsoTex << GetPCData()->mCTexProperty.GetPovStrTex(mpCData->GetPInterpolated()[_ISO_ALPHA], strFinishTag, strClrMap, 1, 1, 1, TRUE, FALSE/*!!*/, 0, strClr1, strClr2, strClr3);
    }

    strIsoTex << L"}" << ENDLWX;

    return strIsoTex;
}

wxString PovIsosurface_cls::
GetPovStrIsoBox(wxString strBoxLabel, wxString strTag,
		double clipForPeriodicBoundary[3],
		DrawingBlock_cls* pCDrawingBlock){
		
    // GetPovStrIsoSurfƓ߂Ȃ̂œł
    double min[3] = {0, 0 ,0};
    double max[3] = {1, 1, 1};
    GetClippingInformation(min, max);
    LOOP_XYZ if (max[xyz] > clipForPeriodicBoundary[xyz])
	max[xyz] = clipForPeriodicBoundary[xyz];
    LOOP_XYZ if (min[xyz] < 0.0005) min[xyz] = 0.0005; // Volumei
    // ܂

    wxString strIsoBox, strDifference;

    strDifference << pCDrawingBlock->GetPovStrDifference();

    strIsoBox << ENDLWX;
    strIsoBox << L"#declare " << L"ISOBOX_" << strBoxLabel << L"=object{" << ENDLWX;
    strIsoBox << wxT("\tdifference { ") << ENDLWX;
    if (strTag == wxT("Empty"))
	strIsoBox << wxT("\t\tbox{ <0, 0, 0> <0, 0, 0> }") << ENDLWX;
    else if (msCOutputType.mBody == _ISO_LINK_TO_DF3)
	strIsoBox << L"\t" << wxT("object { ISOSURF_") << strBoxLabel << wxT(" }") << ENDLWX;
    else
	strIsoBox << L"\t\tintersection{" << ENDLWX
		  << L"\t\t\tbox {" << PovOut3d(wxT(""), min) << wxT(",") << PovOut3d(wxT(""), max) << L" }" << ENDLWX
		  << L"\t\t\t" << wxT("ISOSURF_") << strBoxLabel << ENDLWX
		  << L"\t\t}" << ENDLWX;
    strIsoBox << strDifference << ENDLWX;
    strIsoBox << wxT("\t}") << ENDLWX;

    if (GetPCData()->mCTexProperty.mCMappingCoord.mBody == _MAPPING_LOCAL){
	double scale[3];
	LOOP_XYZ scale[xyz] = pCDrawingBlock->GetBoxScaleWorldCoord(xyz);
	strIsoBox << L"\ttexture { " << wxT("ISOTEX_") << strBoxLabel << ENDLWX
		  << L"\t\t" << PovOut3(wxT("scale"), 1.0/scale[X], 1.0/scale[Y], 1.0/scale[Z]) << ENDLWX
		  << L"\t\t" << PovOut3(wxT("translate"), 0.5, 0.5, 0.5) << ENDLWX
		  << L" }" << ENDLWX;
    }
    
    wxString targetStr = GetTargetLabelStr();
    
    int selfIndex = mpCParentCtrl->GetSelfIndex(this);
    FilterBase_cls* pFog = mpCParentCtrl->GetPFogTargettingMe(mstrLabel);
    bool bFog = (pFog != NULL);
    if (bFog == TRUE || targetStr != _("Solid"))
	strIsoBox << L"\thollow" << ENDLWX;
    
    if (targetStr != _("Empty") && targetStr != _("Solid")){
	wxString strMediaID = 
	    mpCParentCtrl->GetIdStringFromTag(targetStr);
	strIsoBox << L"\tINTERIOR_" << strMediaID << strTag << L"()" << ENDLWX;
    }

    while (pFog != NULL){
	strIsoBox << L"\tinterior{" << ENDLWX;
	strIsoBox << L"\t" << ((PovFog_cls*)pFog)->GetPovStrFogForUse();
	pFog = mpCParentCtrl->GetPFogTargettingMe(mstrLabel, FALSE);
	strIsoBox << L"\t}" << ENDLWX;
    }

    wxString strIOR = GetPCData()->mCTexProperty.GetPovStrInterior();
    if (strIOR != wxT(""))
	strIsoBox << wxT("\tinterior { ") << strIOR << wxT("}") << ENDLWX;

    double alpha = mpCData->GetPInterpolated()[_ISO_ALPHA];
    double reflectionMin = GetPCData()->mCTexProperty.mCReflectionMin.mBody;
    double reflectionMax = GetPCData()->mCTexProperty.mCReflectionMin.mBody;
    strIsoBox << PovRay_cls::GetPovStrPhotonSetting(alpha, reflectionMin, reflectionMax);
    
    strIsoBox << L"\ttranslate < -0.5, -0.5, -0.5 >" << ENDLWX;
    
    strIsoBox << L"}" << ENDLWX << ENDLWX ;

    if (GetPCData()->mCTexProperty.mCMappingCoord.mBody == _MAPPING_GLOBAL)
	strIsoBox << wxT("#declare ISOTEXGLOBAL_") << strBoxLabel << wxT("=1;") << ENDLWX << ENDLWX;
    
    return strIsoBox;
}

// ------------------

wxString PovIsosurface_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("PovIsosurface::");
    
    return strStatus;
}

wxString PovIsosurface_cls::
SerializeConf(){
    UpdateCommonData();
    GetPCData()->mCConfTargetStr.msBody =
	GetLocaleIndependentStrLabel(GetTargetLabelStr());
    return mpCData->SerializeConf(mFilterType);
}

MACRO_GetNewPReadDataBlock(PovIsosurface)

wxString PovIsosurface_cls::
SerializeStaticConf(){
    return 
	msCMaxTrace.Serialize() + 
	msCTypeGradient.Serialize() +
	msCMaxGradient.Serialize() + 
	msCAccuracy.Serialize() + 
	msCEvaluate.Serialize() + 
	msCbCullFace.Serialize() + 
	msCbTrimEdge.Serialize() + 
	msCbReTexOptimize.Serialize() + 
	msCbOffset05.Serialize() + 
	msCbExportNormal.Serialize() + 
	msCbHighPrecNormal.Serialize() + 
	msCbClipForPeriodic.Serialize() + 
	msCOutputType.Serialize() + 
	msCShowMode.Serialize();
}

void PovIsosurface_cls::
ReadStrLineStaticConf(std::wstring strLine){
    msCMaxTrace.ReadStrLine(strLine);
    msCTypeGradient.ReadStrLine(strLine);
    msCMaxGradient.ReadStrLine(strLine);
    msCAccuracy.ReadStrLine(strLine);
    msCEvaluate.ReadStrLine(strLine);
    msCbCullFace.ReadStrLine(strLine);
    msCbTrimEdge.ReadStrLine(strLine);
    msCbReTexOptimize.ReadStrLine(strLine);
    msCbOffset05.ReadStrLine(strLine);
    msCbExportNormal.ReadStrLine(strLine);
    msCbHighPrecNormal.ReadStrLine(strLine);
    msCbClipForPeriodic.ReadStrLine(strLine);
    msCOutputType.ReadStrLine(strLine);
    msCShowMode.ReadStrLine(strLine);
}

wxRect PovIsosurface_cls::msCRectMNumStaticMembers(wxDefaultPosition, wxDefaultSize);
PovIsosurface_cls *PovIsosurface_cls::mspCConnectedToDLG;

ConfItemINT_cls PovIsosurface_cls::msCMaxTrace(L"PovIsosurfaceMaxTrace=", L"1");
ConfItemUINT_cls PovIsosurface_cls::msCTypeGradient(L"PovIsosurfaceTypeGradient=", L"1");
ConfItemFLOAT_cls PovIsosurface_cls::msCMaxGradient(L"PovIsosurfaceMaxGradient=", L"200");
ConfItemFLOAT_cls PovIsosurface_cls::msCAccuracy(L"PovIsosurfaceAccuracy=", L"0.001");
ConfItemFLOATARRAY_cls PovIsosurface_cls::msCEvaluate(L"PovIsosurfaceEvaluateParm=", L"100 1.2 0.8");
ConfItemBOOL_cls PovIsosurface_cls::msCbCullFace(L"PovIsosurfaceCullFace=", L"0");
ConfItemBOOL_cls PovIsosurface_cls::msCbTrimEdge(L"PovIsosurfaceTrimEdge=", L"1");
ConfItemBOOL_cls PovIsosurface_cls::msCbReTexOptimize(L"PovIsosurfaceReTexOptimize=", L"1");
ConfItemBOOL_cls PovIsosurface_cls::msCbOffset05(L"PovIsosurfaceOffsetAtExport=", L"1");
ConfItemBOOL_cls PovIsosurface_cls::msCbExportNormal(L"PovIsosurfaceExportNormal=", L"1");
ConfItemBOOL_cls PovIsosurface_cls::msCbHighPrecNormal(L"PovIsosurfaceExportNormal=", L"1");
ConfItemBOOL_cls PovIsosurface_cls::msCbClipForPeriodic(L"PovIsosurfaceClipForPeriodic=", L"1");
ConfItemUINT_cls PovIsosurface_cls::msCOutputType(L"PovIsosurfaceOutputType=", L"0");
ConfItemBOOL_cls PovIsosurface_cls::msCbMesh2Sequential(L"PovIsosurfaceMesh2Sequential=", L"0");
ConfItemSTR_cls PovIsosurface_cls::msCMesh2Path(L"PovIsosurfaceMesh2Path=", L"Isosurface.pov");
ConfItemUINT_cls PovIsosurface_cls::msCShowMode(L"PovIsosurfaceShowMode=", L"0");
wxString PovIsosurface_cls::msstrWorkingDir;
