// "clsPovObject.cpp"
// (c-basic-offset 4)

#ifndef _POV_OBJECT_H_
#define _POV_OBJECT_H_

#include "clsPovObjectCommon.h"
#include "PLUGINOBJ/clsPovObjectPlugin.h"

#include "../CONFITEM/clsConfItem.h"
#include "../DF3/clsDF3SeqCtrl.h"
#include "../FILTER/clsFilterBase.h"
#include "../FILTER/clsFilterWithParameter.h"
#include "../FILTER/clsFilterWithTexture.h"
#include "../MYGL/clsTextureMapping.h"
#include "../MYGL/clsPositionAndDirection.h"

class PovObject_cls;
class GLGuides_cls;

#define _POV_OBJECT_GLOBAL_CALL 0
#define _POV_OBJECT_LOCAL_CALL 1

#include <wx/thread.h>

class ThreadReadMesh2Sub : public wxThread {
 public:
    ThreadReadMesh2Sub(unsigned char *pBuffer, int fileSize, int posStart,
		       int itemNum, int skipLine,
		       std::vector < float > *pVNormals);

    virtual ExitCode Entry();

    unsigned char *mpBuffer;
    int mFileSize;
    int mPosStart, mItemNum, mSkipLine;

    std::vector < float > *mpVData;
};

//------------------------------

class PovObjectGUI_cls : public FilterWithTextureGUI_cls {
    
    MACRO_DefGUIMainSettings(PovObject);
    MACRO_DefVirtualCustmFunctions;

    int mIndexColorTypeAndButton;
    int mIndexObjType;//, mIndexColorBar;
    int mIndexText, mIndexFontType, mIndexPath;
    int mIndexMeshOffset, mIndexOpen, mIndexShapeParam;
    int mIndexCustomA, mIndexCustomB;
    int mIndexPluginType, mIndexPluginTree;

    void CustomButtonA();
    void CustomButtonB();

    static void  FuncHookHideItem(void *pArg, void *pEvent);
    static void  FuncHookAssign(void *pArg, void *pEvent);
};

class PovObjectData_cls : public FilterWithParameterData_cls, public FilterWithTextureData_cls {
 public:
    PovObjectData_cls(FilterCtrl_cls* pCCtrl, 
		      bool bConnectToStaticParam = TRUE,
		      int initType = 1);

    void ReDefineList();

    ConfItemUINT_cls     mCObjType;
    ConfItemUINT_cls     mCFontType;
    ConfItemSTR_cls      mCTextStr;
    ConfItemSTR_cls      mCPathStr;
    ConfItemBOOL_cls     mCbUseClipping;
    ConfItemBOOL_cls     mCbUseMeshOffset;
    ConfItemBOOL_cls     mCbOpen;
    ConfItemFLOAT_cls    mCShapeParam;
    ConfItemUINT_cls     mCPluginType;
    ConfItemUINT_cls     mCPluginTree;
    ConfItemUINT_cls     mCPluginQuality;

    // |C^ĊmۂňړȂ悤ɂB(UNDO݂ł܂)
    // *todo _MAX_ITEM_NUMBERɂ鏈
    static std::vector < Parameters_cls > msVCParam[_DF3_MAX_SLOT];
    static std::vector < bool > msVbParamUsed[_DF3_MAX_SLOT];
    static void InitializeStaticMember(int df3index);
    Parameters_cls* GetPParamActive();
    
    // UndoŁẢ\Parameter̊mۂSɕKvȂȂJB
    // !! Complete Me !!
    static void GarbageCollect();
    
    void MakeFilterInstanceAndAddToList
	(int index, FilterCtrl_cls *pCParentCtrl);
    wxString SerializeConfPolymorph();
    void ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine);
};

class PovObject_cls : public PovSerialize_cls, public FilterWithParameter_cls, public FilterObject_cls, public TextureMapping_cls, virtual public PovObjectCommon_cls, public PovObjectPlugin_cls {
 public: 
    PovObject_cls(FilterCtrl_cls *pCParentCtrl,
		  PovObjectData_cls *pCInitData = NULL);
    ~PovObject_cls();

    //--
    virtual void AddFilterItemPolymorph(wxBoxSizer *pBoxSizer, FilterPanels_cls *pCFilterPanel, unsigned int index);
    static void InitializeStaticMember(int df3index);
    PovObjectData_cls* GetPCData();
    // --

    void SetMNumAdvanced(wxWindow *pCParentWindow);
    void UpdateColors();
    void ShowPluginHelp();
    void AutoResize();

    bool IsReprocessRequired();

    int mPrevObjectType;
    float mPrevShapeParam;
    
    // -- Mesh
    std::vector < unsigned int > mVIndex;
    std::vector < unsigned int > mVIndexNormal;
    std::vector < float > mVVertices;
    std::vector < float > mVUV1D;
    std::vector < float > mVNormals;
    wxString mstrLoaded;
    wxString mstrPovLabel;
    unsigned char *mpFieldDataLink;
    unsigned int  mSize[3];

    void ClearMesh();
    void MakeMeshPlaneTile(double offsetX, double offsetY, 
			   double scaleX, double scaleY, int resolution);
    void MakeMeshPlane();
    void MakeMeshBoxVertexZ(double z, double normalz, bool bFlip);
    void MakeMeshBoxVertexY(double y, double normaly, bool bFlip);
    void MakeMeshBoxVertexX(double x, double normalx, bool bFlip);
    void MakeMeshBoxIndex(int offse, int resolution);
    void MakeMeshBox();
    void MakeMeshSphere();
    void MakeMeshCone();
    void ConstructNormalVectors();
    void ReadObjFile(wxString strFile);
    void ReadMesh2File(wxString strFile);
    void SetPFieldData(unsigned char *pData, unsigned int size[3]);
    bool DoesUseMapping();
    /*
    void GetGlobalPos(float globPos[3], float pVertex[3], 
		      double offset[3], double scale[3]);
    */

    void MakeUV1D(TexProperty_cls *pCTexProp, 
		  /*int mappingType, int patternType, int mappingCoord,
		  float pos[3], float mapscale, float mapturb, int axis,
		  */
		  double offset[3], double scale[3]);
    void RefreshFilter(unsigned char *pSource, unsigned int size[3]);
    // --

    void SetMatrix(Parameters_cls *pParam, int keyIndex,
		   double dpos[3], int flag);
    
    void DrawPlane(double delta);
    void DrawGlobal(double scale);
    void DrawGLGuide(GLGuides_cls *pCGLGuides, double dpos[3], double scale,
		     int manupMode, double dScale[3], double dRot[4]);
    
    bool IsTetureAvailable();
    wxString GetPovStrObject(int coordinateType, wxString strBoxLabel,
			     PositionAndDirection_cls CCamPos,
			     wxString strFilePath);

    wxString GetFilterStatusStr();
    wxString SerializeConf();
    static PovObjectData_cls*
	GetNewPReadDataBlock(FilterCtrl_cls* pCCtrl,
			     std::wistream &istream);

};

#endif

