// "clsPovObjectCommon.h"
// (setq c-basic-offset 4)

#include "clsPovObjectCommon.h"

#ifdef WIN32
#include <windows.h>
#endif

#ifdef WIN32
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "../MYGL/clsGLGuides.h"

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for(int xyz = 0; xyz < 3; xyz++)
#define LOOP_YPRY for(int ypry = 0; ypry < 4; ypry++)

void PovObjectCommon_cls::
SetGLMatrix(double *pInterpolated, double dRot[4]){
    glTranslated(pInterpolated[_OBJECT_TRANS_X],
		 pInterpolated[_OBJECT_TRANS_Y],
		 pInterpolated[_OBJECT_TRANS_Z]);
    double YPRY[4];
    if (dRot != NULL)
	LOOP_YPRY YPRY[ypry] = 
	    pInterpolated[_OBJECT_ROT_Y + ypry] + dRot[ypry];
    else
	LOOP_YPRY YPRY[ypry] = pInterpolated[_OBJECT_ROT_Y + ypry];
    glRotated(YPRY[0], 0, 0 ,1);
    glRotated(YPRY[1], 1, 0, 0);
    glRotated(YPRY[2], 0, 1, 0);
    glRotated(YPRY[3], 0, 0, 1);
    glScaled(pInterpolated[_OBJECT_SCALE_X],
	     pInterpolated[_OBJECT_SCALE_Y],
	     pInterpolated[_OBJECT_SCALE_Z]);
}

void PovObjectCommon_cls::
DrawBox(){
    glEnable( GL_NORMALIZE );
    glBegin(GL_QUADS);
    glNormal3d(0, -1, 0); 
    glVertex3d(-0.5, -0.5, -0.5); glVertex3d(+0.5, -0.5, -0.5); 
    glVertex3d(+0.5, -0.5, +0.5); glVertex3d(-0.5, -0.5, +0.5);
		
    glNormal3d(+1,  0, 0);
    glVertex3d(+0.5, -0.5, -0.5); glVertex3d(+0.5, +0.5, -0.5);
    glVertex3d(+0.5, +0.5, +0.5); glVertex3d(+0.5, -0.5, +0.5);

    glNormal3d(0, +1, 0);
    glVertex3d(-0.5, +0.5, -0.5); glVertex3d(-0.5, +0.5, +0.5);
    glVertex3d(+0.5, +0.5, +0.5); glVertex3d(+0.5, +0.5, -0.5);

    glNormal3d(-1, 0, 0);
    glVertex3d(-0.5, +0.5, -0.5); glVertex3d(-0.5, -0.5, -0.5); 
    glVertex3d(-0.5, -0.5, +0.5); glVertex3d(-0.5, +0.5, +0.5); 

    glNormal3d(0, 0, +1);
    glVertex3d(-0.5, -0.5, +0.5); glVertex3d(+0.5, -0.5, +0.5);
    glVertex3d(+0.5, +0.5, +0.5); glVertex3d(-0.5, +0.5, +0.5);

    glNormal3d(0, 0, -1);
    glVertex3d(-0.5, -0.5, -0.5); glVertex3d(-0.5, +0.5, -0.5);
    glVertex3d(+0.5, +0.5, -0.5); glVertex3d(+0.5, -0.5, -0.5);

    glEnd();
    glDisable( GL_NORMALIZE );
}

void PovObjectCommon_cls::
DrawWireframePlane(){
    glBegin(GL_LINES);
    for (int i = -13; i <= 14; i+=3){
	glVertex3d(-13.5, (float)i-0.5, 0); glVertex3d(13.5, (float)i-0.5, 0);
	glVertex3d((float)i-0.5, -13.5, 0); glVertex3d((float)i-0.5, 13.5, 0);
    }
    glEnd();
}

void PovObjectCommon_cls::
DrawWireframeSphere(){
    for (int j = -160; j <= 160; j+=20){
	float theta = 3.14159 * j / 180;
	float costheta = 0.5 * cos(theta);
	float sintheta = 0.5 * sin(theta);
	glBegin(GL_LINE_LOOP);
	for (int i = 0; i < 360; i+=20){
	    float phi = 3.14159 * i / 180;
	    float cosphi = cos(phi);
	    float sinphi = sin(phi);
	    glVertex3d(costheta * cosphi, costheta * sinphi, sintheta);
	}
	glEnd();
    }

    for (int i = 0; i < 360; i+=20){
	float phi = 3.14159 * i / 180;
	float cosphi = cos(phi);
	float sinphi = sin(phi);
	glBegin(GL_LINE_STRIP);
	for (int j = -180; j <= 180; j+=20){
	    float theta = 3.14159 * j / 180;
	    float costheta = 0.5 * cos(theta);
	    float sintheta = 0.5 * sin(theta);
	    glVertex3d(costheta * cosphi, costheta * sinphi, sintheta);
	}
	glEnd();
    }
}

void PovObjectCommon_cls::
DrawWireframeCone(float shapeParam){
    glPushMatrix();
    glScaled(0.5, 0.5, 0.5);
    glBegin(GL_LINE_LOOP);
    for (int i = 0; i < 360; i+=15){
	float phi = 3.14159 * i / 180;
	float cosphi = cos(phi);
	float sinphi = sin(phi);
	glVertex3d(cosphi, sinphi, -1);
    }
    glEnd();
    if (shapeParam != 0){
	glBegin(GL_LINE_LOOP);
	for (int i = 0; i < 360; i+=15){
	    float phi = 3.14159 * i / 180;
	    float cosphi = shapeParam * cos(phi);
	    float sinphi = shapeParam * sin(phi);
	    glVertex3d(cosphi, sinphi, 1);
	}
	glEnd();
    }

    glBegin(GL_LINES);
    for (int i = 0; i < 360; i+=30){
	float phi = 3.14159 * i / 180;
	float cosphi = cos(phi);
	float sinphi = sin(phi);
	glVertex3d(cosphi, sinphi, -1);
	glVertex3d(shapeParam * cosphi, shapeParam * sinphi, 1);
    }
    glEnd();
    
    glPopMatrix();
}

void PovObjectCommon_cls::
DrawWireframeArrow(float shapeParam){
    glPushMatrix();
    glTranslated(0, 0, -0.6 + 0.3 * (2.0 - shapeParam));
    glScaled(0.1, 0.1, 0.3 * (2.0 - shapeParam));
    glScaled(2,2,2);
    DrawWireframeCone(1);
    glPopMatrix();
    glPushMatrix();
    glTranslated(0, 0, 0.6 - 0.3 * shapeParam);
    glScaled(0.25, 0.25, 0.3 * shapeParam);
    glScaled(2,2,2);
    DrawWireframeCone(0);
    glPopMatrix();
}

void PovObjectCommon_cls::
DrawWireframeBox(){
    glBegin(GL_LINE_STRIP);
    glVertex3d(-0.5,-0.5,-0.5); glVertex3d(+0.5,-0.5,-0.5);
    glVertex3d(+0.5,+0.5,-0.5); glVertex3d(-0.5,+0.5,-0.5); 
    glVertex3d(-0.5,-0.5,-0.5); 
    glEnd();

    glBegin(GL_LINE_STRIP);
    glVertex3d(-0.5,-0.5,+0.5); glVertex3d(+0.5,-0.5,+0.5);
    glVertex3d(+0.5,+0.5,+0.5); glVertex3d(-0.5,+0.5,+0.5); 
    glVertex3d(-0.5,-0.5,+0.5); 
    glEnd();

    glBegin(GL_LINES);
    glVertex3d(-0.5,-0.5,-0.5); glVertex3d(-0.5,-0.5,+0.5);
    glVertex3d(+0.5,-0.5,-0.5); glVertex3d(+0.5,-0.5,+0.5);
    glVertex3d(-0.5,+0.5,-0.5); glVertex3d(-0.5,+0.5,+0.5);
    glVertex3d(+0.5,+0.5,-0.5); glVertex3d(+0.5,+0.5,+0.5);
    glEnd();
}

void PovObjectCommon_cls::
DrawWireframe(double dScale[3], double dRot[4], double *pInterpolated, 
	      int objType, float shapeParam){
    glPushMatrix();
    SetGLMatrix(pInterpolated, dRot);
    
    glColor4f(0.5, 1, 0.5, 1);
    switch(objType){
    case _OBJECT_TYPE_PLANE:
	DrawWireframePlane(); break;
    case _OBJECT_TYPE_SPHERE:
	DrawWireframeSphere(); break;
    case _OBJECT_TYPE_CONE:
	DrawWireframeCone(shapeParam); break;
    case _OBJECT_TYPE_ARROW:
	DrawWireframeArrow(shapeParam); break;
    case _OBJECT_TYPE_BOX:
    default:
	DrawWireframeBox(); break;
    }

    glColor4f(1, 1, 0.5, 1);
    glScaled(dScale[X], dScale[Y], dScale[Z]);
    switch(objType){
    case _OBJECT_TYPE_PLANE:
	DrawWireframePlane(); break;
    case _OBJECT_TYPE_SPHERE:
	DrawWireframeSphere(); break;
    case _OBJECT_TYPE_CONE:
	DrawWireframeCone(shapeParam); break;
    case _OBJECT_TYPE_ARROW:
	DrawWireframeArrow(shapeParam); break;
    case _OBJECT_TYPE_BOX:
    default:
	DrawWireframeBox(); break;
    }

    glPopMatrix();
}

void PovObjectCommon_cls::
DrawGLGuideCommon(GLGuides_cls *pCGLGuide, double dpos[3], double scale,
		  int manipMode, double dScale[3], double dRot[4],
		  double *pInterpolated, int objType, float shapeParam,
		  bool bHideWire){

    PositionAndDirection_cls CTmp;
    PositionAndDirection_cls CParentObjPos;
    PositionAndDirection_cls CPosCurrent;
    LOOP_XYZ CPosCurrent.mPos[xyz] = pInterpolated[xyz];
    CPosCurrent.SetDirection(_DIRECTION_X_MINUS);

    CPosCurrent.RotateYPR4d(&pInterpolated[_OBJECT_ROT_Y]);
    //std::vector <int> VDisplayList;
    if (bHideWire == FALSE)
	DrawWireframe(dScale, dRot, pInterpolated, objType, shapeParam);

    pCGLGuide->SetArrowDir(1, 1, 1);
    LOOP_XYZ CParentObjPos.mPos[xyz] = 0;
    
    LOOP_XYZ CPosCurrent.mPos[xyz] += dpos[xyz];
    
    double YPRY[4];
    LOOP_YPRY YPRY[ypry] = 
	pInterpolated[_OBJECT_ROT_Y + ypry] + dRot[ypry];
	
    pCGLGuide->DrawGuide(CPosCurrent.mPos, 2 * scale, 1, NULL, CParentObjPos, manipMode,
			 YPRY);
    
    /*
    DrawCurrent(pCGLGuides, CPosCurrent, VDisplayList, 
		dpos, scale, NULL, CTmp, manupMode);
    */
}
