
#ifndef _POV_VOLUME_H_
#define _POV_VOLUME_H_

#include <vector>

#include "clsPovSerialize.h"
#include "../CONFITEM/clsConfItem.h"
#include "../DF3/GL/clsColoring.h"
#include "../FILTER/clsFilterBase.h"
#include "../MYGL/clsPositionAndDirection.h"

class DrawingBlock_cls;
class LightSRC_cls;
class PanelColorBar_cls; 
class PovRay_cls;
class PovVolume_cls;

#define _POV_VOLUME_MONOCOLOR 0

#define _VOLUME_EMISSION 0
#define _VOLUME_ABSORPTION 1
#define _VOLUME_SCATTER 2

class PovVolumeGUI_cls : public FilterGUI_cls {

    MACRO_DefGUIMainSettings(PovVolume)
    MACRO_DefVirtualCustmFunctions;

    int mIndexColorType, mIndexDataType, mIndexColorBar;
    wxString mListType[_CLR_CUSTOM];

    static bool bWarningForRGBA;

    void ChooseColor();
    void UpdateColorList();

    static void FuncHookComboBox(void *pArg, void *pEvent);
    static void FuncHookPanelClick(void *pArg, void *pEvent);
    
    void RefreshColorBar();
    PanelColorBar_cls *mpCPanelColorBar;
    double GetColor(double value, int rgbIndex);
    static double GetColorR(double value, void* pClue);
    static double GetColorG(double value, void* pClue);
    static double GetColorB(double value, void* pClue);
};

class PovVolumeData_cls : public FilterData_cls {
 public:
    PovVolumeData_cls();

    void ReDefineList();
    ConfItemFLOAT_cls mCEmission, mCAbsorption, mCScatter;
    ConfItemUINT_cls  mCScatterType;
    ConfItemUINT_cls  mCNoiseType;
    ConfItemFLOAT_cls mCNoiseScale;
    ConfItemFLOAT_cls mCCompAbsorption;
    ConfItemINT_cls   mCColoring;
    ConfItemFLOATARRAY_cls mCMonoColor, mCScatterColor;
    ConfItemBOOL_cls  mCbUseLocalSetting;
    ConfItemINT_cls   mCSampleMin, mCSampleMax;
    ConfItemINT_cls  mCMultiValueType;
    ConfItemINT_cls   mCDataForColoring;

    void MakeFilterInstanceAndAddToList
	(int index, FilterCtrl_cls *pCParentCtrl);

    void SetDefaultColour(int df3index);
};

class PovVolume_cls : public PovSerialize_cls, public FilterBase_cls {
 public:
    PovVolume_cls(FilterCtrl_cls *pCParentCtrl,
		  PovVolumeData_cls* pCInitData = NULL);
    static PovRay_cls *mspCPov;
    static LightSRC_cls *mspCLightSRC;

    PovVolumeData_cls mCData;
    FilterData_cls*  GetPData();

    // ---
    virtual void AddFilterItemPolymorph(wxBoxSizer *pBoxSizer, FilterPanels_cls *pCFilterPanels, unsigned int index);
    PovVolumeData_cls* GetPCData();
    // ----
    
    int mIndexScatterColor;

    static PovVolume_cls *mspCConnectedToDLG;
    void SetMNumAdvanced(wxWindow *pCParentWindow);
    void ChooseColor(int x = -1, int y = -1);
    void UpdateColors();

    wxString GetPovStrDensityFile
	(wxString strTag, std::vector < wxString > VSFilePath, 
	 bool bTimeRemap, float remapFactor);
    wxString GetPovStrDProfile
	(wxString strTag, std::vector < wxString > VSFilePath,
	 DrawingBlock_cls *pCDrawingBlock, bool bTimeRemap, float remapFactor);
    wxString GetPovStrDProfileArgument
	(wxString strTranslate, wxString strDensityGeometry,
	 bool bShade, bool bForColor = FALSE);
    wxString GetPovStrVolumeBox
	(wxString strTag, wxString strGeometry,
	 float treeDivideParam, unsigned int depth,
	 double baseAlpha, double doubleIndex, 
	 DrawingBlock_cls* pCDrawingBlock);

    wxString GetPovStrDensity
	(wxString strTag, wxString strTagForTarget,
	 wxString strDensityFile,
	 DrawingBlock_cls *pCDrawingBlock, bool bTimeRemap, float remapFactor);
    
    static wxString GetPovStrCommonDensityModifier(wxString strTab,
						   float treeDivideParam);
    wxString GetPovStrDensityModifier
	(wxString strTab, float treeDivideParam, unsigned int depth);
    wxString GetPovStrInterior(wxString strTag,
			       wxString strDensityGeometry,
			       float treeDivideParam, unsigned int depth,
			       double baseAlpha);
    wxString GetPovStrMedia(float emission3f[3],
			    float absorb3f[3],
			    float scatter3f[3],
			    wxString strDensityProfile, 
			    wxString strDensityGeometry,
			    wxString strTranslate,
			    float treeDivideParam, unsigned int depth,
			    bool bShade = FALSE);
    static wxString GetPovStrBox
	(wxString strBox, 
	 double pMin[3], double pMax[3], bool bClipping,
	 std::vector < FilterBase_cls* > VCPFog,
	 std::vector < wxString > *pVstrInteriorTag,
	 wxString strDifference, DrawingBlock_cls* pCDrawingBlock);
    
    static ConfItemINT_cls   msCInterval;
    static ConfItemINT_cls   msCSampleMin, msCSampleMax;
    static ConfItemINT_cls   msCAALevel;
    static ConfItemFLOAT_cls msCAAThreshould;
    static ConfItemUINT_cls  msCMethod;
    static ConfItemUINT_cls  msCInterpolate;
    static ConfItemFLOAT_cls msCSplitBox;

    static void FuncHookScatterColor(void *pArg, void *pEvent);

    // ----

    wxString GetFilterStatusStr();
    wxString SerializeConf();
    static PovVolumeData_cls ReadDataBlock(std::wistream &istream);

    static wxString SerializeStaticConf();
    static void ReadStrLineStaticConf(std::wstring strLine);

    static void InitializeStaticMember(PovRay_cls* pCPov,
				       LightSRC_cls* pCLightSRC);
};

#endif
