// -*- c -*- 
// "clsDualSlider.h"

#include "clsDualSlider.h"

#include <wx/dcbuffer.h>

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#define _BORDERWIDTH 5 

BEGIN_EVENT_TABLE(DualSlider_cls, wxPanel)
     EVT_PAINT       (DualSlider_cls::OnPaint)
     EVT_MOUSE_EVENTS(DualSlider_cls::OnMouse)
END_EVENT_TABLE()

DualSlider_cls::
DualSlider_cls(wxWindow *parent, wxWindowID id, 
	       int initvalue0, int initvalue1,
	       int min, int max,
	       wxPoint pos, wxSize size) :
wxPanel(parent, id, pos, size, wxNO_BORDER)
{
    SetBackgroundColour(wxColour(wxT("WHEAT")));
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);

    mValue[0] = initvalue0; mValue[1] = initvalue1;
    mMin = min; mMax = max;

    mSelected = -1;

    Connect(GetId(), wxEVT_KEY_DOWN,
	    wxKeyEventHandler(DualSlider_cls::OnKeyDown));
}

void DualSlider_cls::
SetMemberSize(wxSize size){
    mWidth = size.GetWidth();
    mHeight = size.GetHeight();
}

void DualSlider_cls::
SetValueDouble(double value0, double value1){
    mValue[0] = value0 * mScale;
    mValue[1] = value1 * mScale;
    Refresh(FALSE);
}

void DualSlider_cls::
UpdateValueFromSlider(){
    // do nothing
    // mValue͕\ƒĂ
}

void DualSlider_cls::
SetRangeSliderWrapperDouble(int min, int max, int scale){
    mMin = min; mMax = max; mScale = scale;
    //if (mScale > 10) SetPageSize(10);
    //SetRangeSliderWrapperInt(min, max);
    //SetRange(min, max);
}

double DualSlider_cls::
GetValue(int index){
    return (double)mValue[index] / mScale;
}

int DualSlider_cls::
GetValueFromPos(int posX){
    double normalValue = (double)(posX - _BORDERWIDTH) / (mWidth - _BORDERWIDTH*2);
    int value = normalValue * mScale;
    if (value < mMin) value = mMin;
    if (value > mMax) value = mMax;
    return value;
}

int DualSlider_cls::
GetSliderPosition(int index){
    return GetSliderPositionFromValue(mValue[index]);
}

int DualSlider_cls::
GetSliderPositionFromValue(float value){
    float normalValue = (float)(value - mMin) / (mMax - mMin);
    return _BORDERWIDTH + (mWidth - _BORDERWIDTH*2) * normalValue;
}

void DualSlider_cls::
UpdateDisplay(){
    wxAutoBufferedPaintDC dc(this);

    dc.SetBackground(wxBrush(wxColour(wxT("WHEAT")), wxSOLID));
    dc.Clear();

    SetMemberSize(GetClientSize());

    dc.SetPen(*wxGREY_PEN);
    dc.SetBrush(*wxGREY_BRUSH);
    dc.DrawRectangle(_BORDERWIDTH, 7, mWidth - _BORDERWIDTH*2, 3);
    dc.SetPen(*wxLIGHT_GREY_PEN);
    dc.DrawRectangle(_BORDERWIDTH+1, 8, mWidth - _BORDERWIDTH*2 - 1, 2);

    dc.SetPen(*wxGREY_PEN);
    dc.SetBrush(*wxWHITE_BRUSH);
    for (int i = 0; i <= 2; i++){
	dc.DrawRectangle
	    (GetSliderPositionFromValue((float)i/2*mScale)-1, 7, 3, 3);
    }

    wxPoint listPoint[3];
    listPoint[0] = wxPoint(0, 0);
    listPoint[1] = wxPoint(-5, 10);
    listPoint[2] = wxPoint(5, 10);

    if (mSelected == 0)
	dc.SetPen(*wxRED_PEN);
    else
	dc.SetPen(*wxBLACK_PEN);
    dc.SetBrush(*wxBLACK_BRUSH);
    dc.DrawPolygon(3, listPoint, GetSliderPosition(0), 8);

    if (mSelected == 1)
	dc.SetPen(*wxRED_PEN);
    else
	dc.SetPen(*wxBLACK_PEN);
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.DrawPolygon(3, listPoint, GetSliderPosition(1), 8);

    if (mSelected == -1) return;

    wxString Mes;
    Mes << (double)GetValueFromPos(mDraggingPos) / mScale;
    int pos = GetSliderPosition(mSelected);
    if (pos > mWidth / 2) pos -= 24;
    else pos+=8;
    dc.SetFont(*wxSMALL_FONT);
    dc.SetTextForeground(*wxRED);
    dc.DrawText(Mes, pos, 8);
}

void DualSlider_cls::
OnPaint(wxPaintEvent& event){
    event.Skip();
    UpdateDisplay();
}

#include "../MYAPI/MsgBox.h"

void DualSlider_cls::
CalcSelected(wxMouseEvent& event){
    mSelected = -1;
    int posX = event.GetX();
    if (posX > GetSliderPosition(0) - 4 && posX < GetSliderPosition(0) + 4)
	mSelected = 0;
    if (posX > GetSliderPosition(1) - 4 && posX < GetSliderPosition(1) + 4)
	mSelected = 1;
    return;
}

void DualSlider_cls::
OnMouse(wxMouseEvent& event){
    if (event.ButtonDown(wxMOUSE_BTN_LEFT)){
	CalcSelected(event);
	if (mSelected != -1){
	    CaptureMouse();
	    mbCaptureMouse = TRUE;
	}
    }

    if (event.Dragging() == TRUE){
	mDraggingPos = event.GetX();
	if (mSelected != -1)
	    if (mValue[mSelected] != GetValueFromPos(mDraggingPos)){
		mValue[mSelected] = GetValueFromPos(mDraggingPos);
		if (event.m_shiftDown == TRUE){
		    wxCommandEvent com(wxEVT_SCROLL_THUMBTRACK);
		    com.SetId(GetId());
		    GetParent()->AddPendingEvent(com);
		}
	    }
    }

    if (event.ButtonUp(wxMOUSE_BTN_LEFT))
	if (mbCaptureMouse == TRUE){
	    ReleaseMouse();
	    mbCaptureMouse = FALSE;
	    mSelected = -1;
	    wxCommandEvent com(wxEVT_SCROLL_CHANGED);
	    com.SetId(GetId());
	    GetParent()->AddPendingEvent(com);
	}


    event.Skip();
    Refresh(FALSE);
}

wxWindow* DualSlider_cls::
GetPWindow(){
    return this;
}

void DualSlider_cls::
OnKeyDown(wxKeyEvent& event){
    ProcessKeyEvent(event);
}
