// -*- c -*- 
// (setq c-basic-offset 4)
// "clsGeneralPointers.h"

#include "clsGeneralPointers.h"

#include "clsDualSlider.h"
#include "clsAtofTextCtrl.h"
#include "clsAtoiTextCtrl.h"
#include "clsAtoSTextCtrl.h"
#include "clsAtoWXSTextCtrl.h"
#include "clsAtoiSpinCtrl.h"
#include "clsAtoiComboCtrl.h"
#include "clsIconSelector.h"
#include "clsSliderInt.h"
#include "clsSliderFloat.h"
#include "clsSliderDouble.h"
#include "EX/clsButtonEX.h"
#include "EX/clsCheckBoxEX.h"
#include "EX/clsStaticTextEX.h"
#include <wx/valgen.h>

#include "../MYAPI/MsgBox.h"

GeneralPointer_cls::
GeneralPointer_cls() : 
mpPointerVal(NULL), mpPointerCtrl(NULL), mpCTextLabel(NULL),
    mpConnectColorFloat(NULL), mpChoiceList(NULL)
{
    mTitle = wxT("");
    mType = _GP_NOT_DEFINED; mID = 0;
    mProportion = 0;
    mTextWidth = -1; mItemWidth = -1;
 
    mbEnabled  = TRUE;
    mbHookSpase = TRUE;
    mMin = 0; mMax = 0; mScale = 1;

    mAlign  = _GP_ALIGN_LEFT;
    mExpand = _GP_EXPAND;
    mbBorder = TRUE;

    mbConnectSizer = FALSE;
    mSizerIndex = 0;
}

void GeneralPointer_cls::
CombinePointerButton(int item, wxString title, bool enabled, bool bHideButton){
    if (bHideButton == TRUE)
	mType = _GP_BUTTON_HIDE_PANEL;
    else
	mType = _GP_BUTTON;
    mpPointerVal = NULL;
    mTitle       = title;
    mTextWidth   = -1;
    mItemWidth   = -1;
    mbEnabled    = enabled;
    mListMax     = 0;
    mpChoiceList = NULL;
    mInit        = 0;
    mOffset      = item;
}

void GeneralPointer_cls::
CombinePointer(int type, void* pPointerVal, wxString title, bool enabled, unsigned int listMax, wxString *pChoiceList, unsigned int init, unsigned int offset){
    mType        = type;
    if (pPointerVal != NULL || mType != _GP_UNSIGNED_INT_LIST)
	mpPointerVal = pPointerVal; // NULL $B$+$D(B INT_LIST $B$N;~$O0];}(B
    mTitle       = title;
    mbEnabled    = enabled;
    mListMax     = listMax;
    //if (mpChoiceList != NULL) //
    mpChoiceList = pChoiceList;
    mInit        = init;
    mOffset      = offset;
    if (mpPointerCtrl != NULL) ReconnectValidator();
    if (mpPointerCtrl != NULL &&
	type == _GP_UNSIGNED_INT_LIST &&
	listMax != 0) RemakeList();
}

void GeneralPointer_cls::
ReconnectValidator(){
    if (mpPointerVal == NULL) return;
    switch(mType){
    case _GP_BOOL:
	((wxCheckBoxEX*)mpPointerCtrl)->SetValidator(wxGenericValidator((bool*)mpPointerVal));
	break;
    }
}

void GeneralPointer_cls::
RemakeList(){
    AtoiComboCtrl_cls* pCCombo = (AtoiComboCtrl_cls*)mpPointerCtrl;
    pCCombo->Clear();
    
    for (int i = 0; i < mListMax; i++)
	pCCombo->Append(mpChoiceList[i + mOffset]);
    
    if (pCCombo->CheckRange() == TRUE) Assign();
}

void GeneralPointer_cls::
SetRange(double min, double max, int scale){
    mMin = min; mMax = max; mScale = scale;
}

void GeneralPointer_cls::
ApplyRange(){
    switch (mType){
    case _GP_INTEGER:      case _GP_UNSIGNED_INT:
	((AtoiTextCtrl_cls*)mpPointerCtrl)->SetRange((int)mMin, (int)mMax);
	break;
    case _GP_INTEGER_SPIN: case _GP_UINT_SPIN:
	((AtoiSpinCtrl_cls*)mpPointerCtrl)->SetRange((int)mMin, (int)mMax);
	break;
    case _GP_INTEGER_SLIDER:
	((SliderInt_cls*)mpPointerCtrl)->
	    SetRangeSliderWrapperInt((int)mMin, (int)mMax);
	break;
    case _GP_FLOAT_SLIDER:
	((SliderFloat_cls*)mpPointerCtrl)->
	    SetRangeSliderWrapperFloat((int)mMin, (int)mMax, mScale);
	break;
    case _GP_DOUBLE_SLIDER:
	((SliderDouble_cls*)mpPointerCtrl)->
	    SetRangeSliderWrapperDouble((int)mMin, (int)mMax, mScale);
	break;
    case _GP_FLOAT:       case _GP_DOUBLE:
	((AtofTextCtrl_cls*)mpPointerCtrl)->SetRange(mMin, mMax);
	break;
    case _GP_WXSTRING:
	((TextCtrlEX_cls*)mpPointerCtrl)->EnableHookSpace(mbHookSpase);
	break;
    case _GP_WXSTRING_PATH:
	break;
    case _GP_DUALSLIDER:
	((DualSlider_cls*)mpPointerCtrl)->
	    SetRangeSliderWrapperDouble((int)mMin, (int)mMax, mScale);
	break;
    case _GP_UNSIGNED_INT_ICON:
	break;
    }
}

void GeneralPointer_cls::
DisableHookSpace(){
    mbHookSpase = FALSE;
}

// -----

int GeneralPointer_cls::
GetValueInt(){
    if (mpPointerVal == NULL) return 0;
    if (mType == _GP_INTEGER || mType == _GP_INTEGER_SLIDER || mType == _GP_INTEGER_SPIN)
	return *(int*)mpPointerVal;
    else if (mType == _GP_UNSIGNED_INT || mType == _GP_UNSIGNED_INT_LIST ||
	     mType == _GP_UINT_SPIN || mType == _GP_UNSIGNED_INT_ICON)
	return *(unsigned int*)mpPointerVal;
    else 
	MsgBoxExit(NULL, "Error", "Called GetValueInt() for Not-Int member");
    return 0;
}

void GeneralPointer_cls::
SetValueFloat(float value, float value1){
    if (mpPointerVal == NULL) { 
	MsgBox(NULL, "Called SetValueFloat() for NULL member", "Error");
	return; }
    if (mType == _GP_FLOAT || mType == _GP_FLOAT_SLIDER)
	*(float*)mpPointerVal = value;
    else if (mType == _GP_DUALSLIDER){
	*(double*)mpPointerVal = value;
	*((double*)mpPointerVal+1) = value1;
    }
    else
	//printf("error\n");
	MsgBoxExit(NULL, "SetValueFloat() for Not-Float member", "Error");
}

void GeneralPointer_cls::
SetValueInt(int value){
    if (mpPointerVal == NULL) { 
	MsgBox(NULL, "SetValueFloat() for NULL member", "Error");
	return; }
    
    if (mType == _GP_INTEGER || mType == _GP_INTEGER_SPIN || 
	mType == _GP_INTEGER_SLIDER)
	*(int*)mpPointerVal = value;
    else if (mType == _GP_UNSIGNED_INT_LIST || mType == _GP_UNSIGNED_INT 
	     || mType == _GP_UNSIGNED_INT_ICON)
	*(unsigned int*)mpPointerVal = value;
    else
	MsgBoxExit(NULL, "SetValueInt() for Not-Int member", "Error");
}

void GeneralPointer_cls::
SetValueBool(bool value){
    if (mpPointerVal == NULL) { 
	MsgBox(NULL, "SetValueBool() for NULL member", "Error");
	return; }
    
    if (mType == _GP_BOOL)
	*(bool*)mpPointerVal = value;
    else
	MsgBoxExit(NULL, "SetValueBool() for Not-Bool member", "Error");
}

float GeneralPointer_cls::
GetValueFloat(){
    if (mpPointerVal == NULL) return 0;
    if (mType == _GP_FLOAT || mType == _GP_FLOAT_SLIDER)
	return *(float*)mpPointerVal;
    else 
	//printf("error\n");
	MsgBoxExit(NULL, "Called GetValueFloat() for Not-Float member", "Error");
    return 0;
}

double GeneralPointer_cls::
GetValueDouble(int index){
    if (mpPointerVal == NULL) return 0;
    if (mType == _GP_DOUBLE || mType == _GP_DOUBLE_SLIDER)
	return *(double*)mpPointerVal;
    else if (mType == _GP_DUALSLIDER)
	return *((double*)mpPointerVal + index);
    else 
	MsgBoxExit(NULL, "Error", "Called GetValueDouble() for Not-Double member");
    return 0;
}

bool GeneralPointer_cls::
GetValueBool(){
    if (mpPointerVal == NULL) return FALSE;
    if (mType == _GP_BOOL)
	return *(bool*)mpPointerVal;
    else
	MsgBoxExit(NULL, "Called GetValueBool() for Not-Bool member",
		   "Error");
    return FALSE;
}

wxString GeneralPointer_cls::
GetValueWXStr(){
    if (mpPointerVal == NULL) return wxT("");
    if (mType == _GP_WXSTRING || mType == _GP_WXSTRING_PATH)
	return *(wxString*)mpPointerVal;
    else
	MsgBoxExit(NULL, "Called GetValeuWXStr() for Not-WXString member",
		   "Error");
    return wxT("");
}

void GeneralPointer_cls::
UpdateValue(bool bForce){
    switch(mType){
    case _GP_INTEGER:      case _GP_UNSIGNED_INT:
	if (bForce||
	    ((AtoiTextCtrl_cls*)mpPointerCtrl)->GetValue() != GetValueInt())
	    ((AtoiTextCtrl_cls*)mpPointerCtrl)->SetValueInt(GetValueInt());
	break;
    case _GP_INTEGER_SPIN: case _GP_UINT_SPIN: {
	if (bForce||
	    ((AtoiSpinCtrl_cls*)mpPointerCtrl)->GetValue() != GetValueInt())
	    ((AtoiSpinCtrl_cls*)mpPointerCtrl)->SetValueInt(GetValueInt());
	break; }
    case _GP_UNSIGNED_INT_LIST:
	if (bForce||
	    GetValueInt() != ((AtoiComboCtrl_cls*)mpPointerCtrl)->mValue)
	    ((AtoiComboCtrl_cls*)mpPointerCtrl)->SetValueInt(GetValueInt());
	break;
    case _GP_UNSIGNED_INT_ICON:
	if (bForce||GetValueInt() != ((IconSelector_cls*)mpPointerCtrl)->mValue)
	    ((IconSelector_cls*)mpPointerCtrl)->SetValueInt(GetValueInt());
	break;
    case _GP_INTEGER_SLIDER:
	if (bForce||
	    ((SliderInt_cls*)mpPointerCtrl)->GetValue() != GetValueInt())
	    ((SliderInt_cls*)mpPointerCtrl)->SetValueInt(GetValueInt());
	break;
    case _GP_FLOAT_SLIDER:
	if (bForce||
	    ((SliderFloat_cls*)mpPointerCtrl)->GetValue() != GetValueFloat())
	    ((SliderFloat_cls*)mpPointerCtrl)->SetValueFloat(GetValueFloat());
	break;
    case _GP_DUALSLIDER:
	if (GetValueDouble(0) != ((DualSlider_cls*)mpPointerCtrl)->mValue[0] ||
	    GetValueDouble(1) != ((DualSlider_cls*)mpPointerCtrl)->mValue[1])
	    ((DualSlider_cls*)mpPointerCtrl)->SetValueDouble(GetValueDouble(0), GetValueDouble(1));
	break;
    case _GP_FLOAT:
	if (bForce||
	    ((AtofTextCtrl_cls*)mpPointerCtrl)->GetValue() != GetValueFloat())
	    ((AtofTextCtrl_cls*)mpPointerCtrl)->SetValueDouble(GetValueFloat() );
	break;
    case _GP_DOUBLE_SLIDER:
	if (bForce||
	    ((SliderDouble_cls*)mpPointerCtrl)->GetValue() != GetValueDouble())
	    ((SliderDouble_cls*)mpPointerCtrl)->SetValueDouble(GetValueDouble());
	break;
    case _GP_DOUBLE:
	if (bForce||
	    ((AtofTextCtrl_cls*)mpPointerCtrl)->GetValue() != GetValueDouble())
	    ((AtofTextCtrl_cls*)mpPointerCtrl)->SetValueDouble(GetValueDouble());
	break;
    case _GP_BOOL:
	if (GetValueBool() != ((wxCheckBoxEX*)mpPointerCtrl)->GetValue())
	    ((wxCheckBoxEX*)mpPointerCtrl)->SetValue(GetValueBool());
	break;
    case _GP_WXSTRING:
	((AtoWXSTextCtrl_cls*)mpPointerCtrl)->SetValueWXStr(GetValueWXStr());
	break;
    case _GP_WXSTRING_PATH:
	((AtoWXSTextPathCtrl_cls*)mpPointerCtrl)->SetValueWXStr(GetValueWXStr());
	break;
    case _GP_STATIC_TEXT:
	if (mpPointerVal == NULL) break;
	if (((wxStaticTextEX*)mpPointerCtrl)->GetLabel() != *(wxString*)mpPointerVal)
	    ((wxStaticTextEX*)mpPointerCtrl)->SetLabel(*(wxString*)mpPointerVal);
	break;
    case _GP_BUTTON:
	if (((wxButton*)mpPointerCtrl)->GetLabel() !=  mTitle)
	    ((wxButton*)mpPointerCtrl)->SetLabel(mTitle);
	if (mpConnectColorFloat != NULL){
	    if (((wxButton*)mpPointerCtrl)->IsEnabled() == FALSE){
		((wxButton*)mpPointerCtrl)->
		    SetBackgroundColour(wxColour(196, 196, 196));
		((wxWindow*)mpPointerCtrl)->SetForegroundColour(*wxWHITE);
	    }
	    else{
		((wxButton*)mpPointerCtrl)->
		    SetBackgroundColour(wxColour(mpConnectColorFloat[0]*255,
						 mpConnectColorFloat[1]*255,
						 mpConnectColorFloat[2]*255));
		if (mpConnectColorFloat[0] + mpConnectColorFloat[1] + mpConnectColorFloat[2] < 1)
		    ((wxWindow*)mpPointerCtrl)->SetForegroundColour(*wxWHITE);
		else
		    ((wxWindow*)mpPointerCtrl)->SetForegroundColour(*wxBLACK);
	    }
	}
	break;
    }
}

#define MACRO_ASSIGN()				\
if (*porgValue != newValue){\
 *porgValue = newValue;\
 bChanged = TRUE;\
 }

bool GeneralPointer_cls::Assign(){
    bool bChanged = FALSE;
    switch(mType){
    case _GP_INTEGER: {
	int *porgValue = (int*)mpPointerVal;
	int newValue = ((AtoiTextCtrl_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN(); break; }
    case _GP_INTEGER_SPIN: {
	int *porgValue = (int*)mpPointerVal;
	int newValue = ((AtoiSpinCtrl_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN();	break; }
    case _GP_UNSIGNED_INT: {
	unsigned int *porgValue = (unsigned int*)mpPointerVal;
	unsigned int newValue = ((AtoiTextCtrl_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN(); break; }
    case _GP_UINT_SPIN: {
	unsigned int *porgValue = (unsigned int*)mpPointerVal;
	unsigned int newValue = ((AtoiSpinCtrl_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN(); break; }
    case _GP_UNSIGNED_INT_LIST: {
	unsigned int *porgValue = (unsigned int*)mpPointerVal;
	unsigned int newValue = ((AtoiComboCtrl_cls*)mpPointerCtrl)->mValue;
	MACRO_ASSIGN();	break; }
    case _GP_UNSIGNED_INT_ICON: {
	unsigned int *porgValue = (unsigned int*)mpPointerVal;
	unsigned int newValue = ((IconSelector_cls*)mpPointerCtrl)->mValue;
	MACRO_ASSIGN();	break; }
    case _GP_INTEGER_SLIDER: {
	int *porgValue = (int *)mpPointerVal;
	int newValue = ((SliderInt_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN(); break; }
    case _GP_FLOAT_SLIDER: {
	float *porgValue = (float*)mpPointerVal;
	float newValue = ((SliderFloat_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN();	break; }
    case _GP_DUALSLIDER: {
	double *porgValue0 = (double*)mpPointerVal;
	double newValue0 = ((DualSlider_cls*)mpPointerCtrl)->GetValue(0);
	if (*porgValue0 != newValue0){
	    *porgValue0 = newValue0;
	    bChanged = TRUE; }
	double *porgValue1 = (double*)mpPointerVal + 1;
	double newValue1 = ((DualSlider_cls*)mpPointerCtrl)->GetValue(1);
	if (*porgValue1 != newValue1){
	    *porgValue1 = newValue1;
	    bChanged = TRUE; }
	break; }
    case _GP_FLOAT:{
	float *porgValue = (float*)mpPointerVal;
	float newValue = ((AtofTextCtrl_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN();	break; }
    case _GP_DOUBLE_SLIDER:
	*(double*)mpPointerVal = ((SliderDouble_cls*)mpPointerCtrl)->GetValue();
	break;
    case _GP_DOUBLE: {
	double *porgValue = (double*)mpPointerVal;
	double newValue = ((AtofTextCtrl_cls*)mpPointerCtrl)->GetValue();
	MACRO_ASSIGN(); break; }
    case _GP_BOOL: {
	bool *porgValue = (bool*)mpPointerVal;
	bool newValue = (((wxCheckBoxEX*)mpPointerCtrl)->GetValue());
	MACRO_ASSIGN();
	//*(bool*)mpPointerVal = (((wxCheckBoxEX*)mpPointerCtrl)->GetValue());
	break; }
    case _GP_WXSTRING:
	*(wxString*)mpPointerVal = 
	    ((AtoWXSTextCtrl_cls*)mpPointerCtrl)->GetWXStr();
	break;
    case _GP_WXSTRING_PATH:
	*(wxString*)mpPointerVal = 
	    ((AtoWXSTextPathCtrl_cls*)mpPointerCtrl)->GetWXStr();
	break;
    }
    return bChanged;
}

void GeneralPointer_cls::
UpdateFromText(){
    switch(mType){
    case _GP_INTEGER:      case _GP_UNSIGNED_INT:
	((AtoiTextCtrl_cls*)mpPointerCtrl)->UpdateValueFromText();
	break;
    case _GP_INTEGER_SPIN: case _GP_UINT_SPIN:
	((AtoiSpinCtrl_cls*)mpPointerCtrl)->UpdateValueFromText();
	break;
    case _GP_INTEGER_SLIDER:
	((SliderInt_cls*)mpPointerCtrl)->UpdateValueFromSlider();
	break;
    case _GP_FLOAT_SLIDER:
	((SliderFloat_cls*)mpPointerCtrl)->UpdateValueFromSlider();
	break;
    case _GP_DUALSLIDER:
	((DualSlider_cls*)mpPointerCtrl)->UpdateValueFromSlider();
	break;
    case _GP_DOUBLE_SLIDER:
	((SliderDouble_cls*)mpPointerCtrl)->UpdateValueFromSlider();
	break;
    case _GP_FLOAT:
    case _GP_DOUBLE:
	((AtofTextCtrl_cls*)mpPointerCtrl)->UpdateValueFromText();
	break;
    case _GP_BOOL:
	break;
    case _GP_WXSTRING:
	((AtoWXSTextCtrl_cls*)mpPointerCtrl)->UpdateValueFromText();
	break;
    case _GP_WXSTRING_PATH:
	((AtoWXSTextPathCtrl_cls*)mpPointerCtrl)->UpdateValueFromText();
	break;
    }
}

bool GeneralPointer_cls::
IsDirty(){
    if (mpPointerVal == NULL) return FALSE;
    if (mpPointerCtrl == NULL) return FALSE;
    
    bool dirty = FALSE;
    switch(mType){
    case _GP_INTEGER:
	if (*(int*)mpPointerVal != ((AtoiTextCtrl_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_INTEGER_SPIN:
	if (*(int*)mpPointerVal != ((AtoiSpinCtrl_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_UNSIGNED_INT:
	if (*(unsigned int*)mpPointerVal != 
	    (unsigned int)((AtoiTextCtrl_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_UINT_SPIN:
	if (*(unsigned int*)mpPointerVal != 
	    (unsigned int)((AtoiSpinCtrl_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_UNSIGNED_INT_LIST:
	if (*(unsigned int*)mpPointerVal != 
	    (unsigned int)((AtoiComboCtrl_cls*)mpPointerCtrl)->mValue)
	    dirty = TRUE;
	break;
    case _GP_UNSIGNED_INT_ICON:
	if (*(unsigned int*)mpPointerVal != 
	    (unsigned int)((IconSelector_cls*)mpPointerCtrl)->mValue)
	    dirty = TRUE;
	break;
    case _GP_INTEGER_SLIDER:
	if (*(int*)mpPointerVal != ((SliderInt_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_FLOAT_SLIDER:
	if (*(float*)mpPointerVal != 
	    ((SliderFloat_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_DUALSLIDER:
	if (*(double*)mpPointerVal != ((DualSlider_cls*)mpPointerCtrl)->GetValue(0) ||
	    *((double*)mpPointerVal + 1) != ((DualSlider_cls*)mpPointerCtrl)->GetValue(1))
	    dirty = TRUE;
	break;
    case _GP_FLOAT:
	if (*(float*)mpPointerVal != 
	    ((AtofTextCtrl_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_DOUBLE_SLIDER:
	if (*(double*)mpPointerVal !=
	    ((SliderDouble_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_DOUBLE:
	if (*(double*)mpPointerVal != 
	    ((AtofTextCtrl_cls*)mpPointerCtrl)->GetValue())
	    dirty = TRUE;
	break;
    case _GP_BOOL:
	if (*(bool*)mpPointerVal != 
	    (((wxCheckBoxEX*)mpPointerCtrl)->GetValue()))
	    dirty = TRUE;
	break;
    case _GP_WXSTRING:
	if (*(wxString*)mpPointerVal !=
	    ((AtoWXSTextCtrl_cls*)mpPointerCtrl)->GetWXStr()) dirty = TRUE;
	break;
    case _GP_WXSTRING_PATH:
	if (*(wxString*)mpPointerVal !=
	    ((AtoWXSTextPathCtrl_cls*)mpPointerCtrl)->GetWXStr()) dirty = TRUE;
	break;
    }
    return dirty;
}

bool GeneralPointer_cls::
IsTabstop(){
    if (mbEnabled == FALSE) return FALSE;
    switch (mType){
    case _GP_INTEGER:        case _GP_UNSIGNED_INT:
    case _GP_INTEGER_SPIN:   case _GP_UINT_SPIN:
    case _GP_UNSIGNED_INT_LIST: //case _GP_UNSIGNED_INT_ICON:
    case _GP_INTEGER_SLIDER: case _GP_FLOAT_SLIDER: case _GP_DOUBLE_SLIDER:
    case _GP_DUALSLIDER:
    case _GP_FLOAT:          case _GP_DOUBLE:
    case _GP_WXSTRING:       case _GP_WXSTRING_PATH:
	return TRUE;
    }
    return FALSE;
}

bool GeneralPointer_cls::
IsButton(){
    if (mType == _GP_BUTTON_HIDE_PANEL ||
	mType == _GP_BUTTON)
	return TRUE;

    return FALSE;
      
}

wxWindow* GeneralPointer_cls::
GetPCtrlWindow(){
    if (mpPointerCtrl == NULL) return NULL;
    switch(mType){
    case _GP_BOOL:
	return (wxCheckBox*)mpPointerCtrl;
    case _GP_INTEGER:      case _GP_UNSIGNED_INT:
	return (AtoiTextCtrl_cls*)mpPointerCtrl;
    case _GP_INTEGER_SPIN: case _GP_UINT_SPIN:
	return (AtoiSpinCtrl_cls*)mpPointerCtrl;
    case _GP_UNSIGNED_INT_LIST:
	return (AtoiComboCtrl_cls*)mpPointerCtrl;
    case _GP_UNSIGNED_INT_ICON:
	return (IconSelector_cls*)mpPointerCtrl;
    case _GP_INTEGER_SLIDER:
	return (SliderInt_cls*)mpPointerCtrl;
    case _GP_FLOAT_SLIDER:
	return (SliderFloat_cls*)mpPointerCtrl;
    case _GP_DUALSLIDER:
	return (DualSlider_cls*)mpPointerCtrl;
    case _GP_DOUBLE_SLIDER:
	return (SliderDouble_cls*)mpPointerCtrl;
    case _GP_FLOAT: case _GP_DOUBLE:
	return (AtofTextCtrl_cls*)mpPointerCtrl;
    case _GP_WXSTRING:
    case _GP_WXSTRING_PATH:
	return (AtoWXSTextCtrl_cls*)mpPointerCtrl;
    case _GP_WXPANEL:
	return (wxPanel*)mpPointerCtrl;
    case _GP_BUTTON:
    case _GP_BUTTON_HIDE_PANEL:
	return (wxButton*)mpPointerCtrl;
	
    }
    return NULL;
}

void GeneralPointer_cls::
PushBackPrevWindow(wxWindow* pWindow){
    switch(mType){
    case _GP_INTEGER:      case _GP_UNSIGNED_INT:
	((AtoiTextCtrl_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_INTEGER_SPIN: case _GP_UINT_SPIN:
	((AtoiSpinCtrl_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_UNSIGNED_INT_LIST:
	((AtoiComboCtrl_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_UNSIGNED_INT_ICON:
	((IconSelector_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_INTEGER_SLIDER:
	((SliderInt_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_FLOAT_SLIDER:
	((SliderFloat_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_DUALSLIDER:
	((DualSlider_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_DOUBLE_SLIDER:
	((SliderDouble_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_FLOAT: case _GP_DOUBLE:
	((AtofTextCtrl_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    case _GP_WXSTRING:
    case _GP_WXSTRING_PATH:
	((AtoWXSTextCtrl_cls*)mpPointerCtrl)->PushBackPrevWindow(pWindow);
	break;
    }
}

void GeneralPointer_cls::
PushBackNextWindow(wxWindow* pWindow){
    switch(mType){
    case _GP_INTEGER:      case _GP_UNSIGNED_INT:
	((AtoiTextCtrl_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_INTEGER_SPIN: case _GP_UINT_SPIN:
	((AtoiSpinCtrl_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_UNSIGNED_INT_LIST:
	((AtoiComboCtrl_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_UNSIGNED_INT_ICON:
	((IconSelector_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_INTEGER_SLIDER:
	((SliderInt_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_FLOAT_SLIDER:
	((SliderFloat_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_DUALSLIDER:
	((DualSlider_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_DOUBLE_SLIDER:
	((SliderDouble_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_FLOAT: case _GP_DOUBLE:
	((AtofTextCtrl_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    case _GP_WXSTRING:
    case _GP_WXSTRING_PATH:
	((AtoWXSTextCtrl_cls*)mpPointerCtrl)->PushBackNextWindow(pWindow);
	break;
    }
}

void GeneralPointer_cls::
Enable(bool bEnable){
    mbEnabled = bEnable;
    
    if (mType == _GP_BOXSIZERHORIZONTAL)
	mpCTextLabel->Enable(bEnable);
    else
	((wxWindow*)mpPointerCtrl)->Enable(bEnable);
	;
}

void GeneralPointer_cls::
Disable(){
    mbEnabled = FALSE;
    if (mType != _GP_BOXSIZERHORIZONTAL)
	((wxWindow*)mpPointerCtrl)->Disable();
    else
	//((wxBoxSizer*)mpPointerCtrl)->Disable();
	;
}

void GeneralPointer_cls::
Show(bool bShow){
    if (bShow == FALSE) {
	Hide(); return;
    }
    if (mType != _GP_BOXSIZERHORIZONTAL){
	((wxWindow*)mpPointerCtrl)->Show();
    }
    if (mpCTextLabel != NULL) mpCTextLabel->Show();
}

void GeneralPointer_cls::
Hide(){
    if (mType != _GP_BOXSIZERHORIZONTAL){
	((wxWindow*)mpPointerCtrl)->Hide();
    }
    if (mpCTextLabel != NULL) mpCTextLabel->Hide();
}

void GeneralPointer_cls::
UpdateLabel(){
    if (mpCTextLabel != NULL){
	if (mpCTextLabel->GetLabel() != mTitle)
	    mpCTextLabel->SetLabel(mTitle);
    }
}

wxString GeneralPointer_cls::
GetInitStr(){
    unsigned int offseted = mInit - mOffset;
    if (offseted < mMax)
	return mpChoiceList[mInit - mOffset];
    return wxT("Error in GetInitStr()");
}

wxSize GeneralPointer_cls::
GetSizeLabel(){
    return wxSize(mTextWidth, -1);
}

wxSizerFlags GeneralPointer_cls::
GetSizerFlagLabel(){
    if (mTextWidth == -1){
	wxSizerFlags sflag1(1);
	sflag1.Border(wxLEFT, 2).FixedMinSize().Align(wxALIGN_CENTER_VERTICAL);
	return sflag1;
    }
    
    wxSizerFlags sflag0(0);
    sflag0 .Border(wxLEFT, 2).FixedMinSize().Align(wxALIGN_CENTER_VERTICAL);
    return sflag0;
}

wxSizerFlags GeneralPointer_cls::
GetSizerFlagH(int borderWidth){
    if (mItemWidth != -1) mExpand = _GP_NO_EXPAND;
    switch (mExpand){
    case _GP_EXPAND: {
	wxSizerFlags sflag(1);
	sflag.Expand().Border(wxLEFT, borderWidth).FixedMinSize();
	if (mAlign == _GP_ALIGN_RIGHT) sflag.Align(wxRIGHT);
	else
	return sflag;
	break; }
    case _GP_NO_EXPAND: {
	wxSizerFlags sflag(0);
	sflag.Border(wxLEFT, borderWidth).FixedMinSize().Align(wxALIGN_CENTER_VERTICAL);
	if (mAlign == _GP_ALIGN_RIGHT) sflag.Align(wxRIGHT);
	return sflag;
	break; }
    }
    
    wxSizerFlags sflag(0);
    return sflag;
}

wxSizerFlags GeneralPointer_cls::
GetSizerFlagV(){
    wxSizerFlags sflag;
    sflag .Border(wxLEFT, 2).FixedMinSize();

    if (mAlign == _GP_ALIGN_RIGHT) sflag.Right();
    else sflag.Expand();
    
    return sflag;
}

void GeneralPointer_cls::
AlignRight(){
    mAlign  = _GP_ALIGN_RIGHT;
}

int GeneralPointer_cls::
GPConnect_BoxSizer(wxBoxSizer* pBoxSizer, wxWindow* pWindow, int proportion){
    
    wxSizerFlags sflagE(0);
    sflagE.Border(wxLEFT, 2).FixedMinSize().Expand().Align(wxALIGN_CENTER_VERTICAL);
    
    mpPointerCtrl = new wxBoxSizer(wxHORIZONTAL);
    pBoxSizer->Add((wxBoxSizer*)mpPointerCtrl, sflagE);
    
    int id = -1;
    if (mTitle != wxT("")){
	mpCTextLabel = 
	    new wxStaticTextEX(pWindow, wxID_ANY, mTitle,
			       wxPoint(0, -1), GetSizeLabel());
	id = mpCTextLabel->GetId();
	((wxBoxSizer*)mpPointerCtrl)->
	    Add(mpCTextLabel, GetSizerFlagLabel());
    }
    return id;
}

void GeneralPointer_cls::
GPConnect_SText(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    if (mpPointerVal != NULL)
	mpPointerCtrl = new wxStaticTextEX(pWindow, wxID_ANY, *(wxString*)mpPointerVal);
    else
	mpPointerCtrl = new wxStaticTextEX(pWindow, wxID_ANY, mTitle);
    mExpand = _GP_NO_EXPAND;
    pBoxSizer->Add((wxStaticTextEX*)mpPointerCtrl, GetSizerFlagH());
    mID  = ((wxStaticTextEX*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_SLine(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = new wxStaticLine(pWindow, wxID_ANY);
    pBoxSizer->Add((wxStaticLine*)mpPointerCtrl, GetSizerFlagH());
    mID = ((wxStaticLine*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_Button(wxBoxSizer* pBoxSizer, wxWindow* pWindow, int borderWidth){
    mpPointerCtrl = new	wxButtonEX
	(pWindow, wxID_ANY, mTitle, wxDefaultPosition, wxSize(mItemWidth, -1));
    pBoxSizer->Add((wxButtonEX*)mpPointerCtrl, GetSizerFlagH(borderWidth));
    mID = ((wxButtonEX*)mpPointerCtrl)->GetId();
    if (mOffset == wxID_ANY) 
	mOffset = ((wxButtonEX*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_Integer(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = new AtoiTextCtrl_cls
	(pWindow, wxID_ANY, GetValueInt()/*,
					   wxDefaultPosition, wxSize(mItemWidth, -1)*/);
    pBoxSizer
	->Add((AtoiTextCtrl_cls*)mpPointerCtrl, GetSizerFlagH());
    mID  = ((AtoiTextCtrl_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_IntegerSpin(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    if (mItemWidth != -1)
	mpPointerCtrl = new AtoiSpinCtrl_cls
	    (pWindow, wxID_ANY, GetValueInt(),
	     wxDefaultPosition, wxSize(mItemWidth, -1));
    else
	mpPointerCtrl = new AtoiSpinCtrl_cls
	    (pWindow, wxID_ANY, GetValueInt());
    pBoxSizer
	->Add((AtoiSpinCtrl_cls*)mpPointerCtrl, GetSizerFlagH());
    mID  = ((AtoiSpinCtrl_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_List(wxBoxSizer* pBoxSizer, wxWindow* pWindow, int width){
    mMax = mListMax;
    mpPointerCtrl = new AtoiComboCtrl_cls
	(pWindow, wxID_ANY, GetInitStr(), mListMax, mpChoiceList, mOffset,
	 wxSize(width, -1));
    pBoxSizer
	->Add((AtoiComboCtrl_cls*)mpPointerCtrl, GetSizerFlagH());
    mID  = ((AtoiComboCtrl_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_IconSelect(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = new IconSelector_cls
	(pWindow, wxID_ANY, GetValueInt(), mMax, mpVIcon);
    pBoxSizer
	->Add((IconSelector_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((IconSelector_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_SliderInt(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    if (mItemWidth == -1)
	mpPointerCtrl = new SliderInt_cls
	    (pWindow, wxID_ANY, GetValueInt(), (int)mMin, (int)mMax);
    else
	mpPointerCtrl = new SliderInt_cls
	    (pWindow, wxID_ANY, GetValueInt(), (int)mMin, (int)mMax,
	     wxDefaultPosition, wxSize(mItemWidth, -1));
    pBoxSizer
	->Add((SliderInt_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((SliderInt_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_SliderFloat(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    if (mItemWidth == -1)
	mpPointerCtrl = new SliderFloat_cls
	    (pWindow, wxID_ANY, (int)GetValueFloat() * mScale, 
	     (int)mMin, (int)mMax);
    else
	mpPointerCtrl = new SliderFloat_cls
	    (pWindow, wxID_ANY, (int)GetValueFloat() * mScale, 
	     (int)mMin, (int)mMax,
	     wxDefaultPosition, wxSize(mItemWidth, -1));
    
    pBoxSizer
	->Add((SliderFloat_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((SliderFloat_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_DualSlider(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    if (mItemWidth == -1)
	mpPointerCtrl = new DualSlider_cls
	    (pWindow, wxID_ANY, 
	     (int)GetValueDouble(0) * mScale, 
	     (int)GetValueDouble(1) * mScale, 
	     (int)mMin, (int)mMax);
    else
	mpPointerCtrl = new DualSlider_cls
	    (pWindow, wxID_ANY, 
	     (int)GetValueDouble(0) * mScale, 
	     (int)GetValueDouble(1) * mScale, 
	     (int)mMin, (int)mMax,
	     wxDefaultPosition, wxSize(mItemWidth, -1));
    
    pBoxSizer
	->Add((SliderFloat_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((SliderFloat_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_Float(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    if (mItemWidth == -1)
	mpPointerCtrl = new AtofTextCtrl_cls
	    (pWindow, wxID_ANY, GetValueFloat());
    else
	mpPointerCtrl = new AtofTextCtrl_cls
	    (pWindow, wxID_ANY, GetValueFloat(),
	     wxDefaultPosition, wxSize(mItemWidth, -1));
    
    pBoxSizer
	->Add((AtofTextCtrl_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((AtofTextCtrl_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_SliderDouble(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = new SliderDouble_cls
	(pWindow, wxID_ANY, (int)GetValueDouble() * mScale, (int)mMin, (int)mMax);
    pBoxSizer
	->Add((SliderDouble_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((SliderDouble_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_Double(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = new AtofTextCtrl_cls
	(pWindow, wxID_ANY, GetValueDouble());
    pBoxSizer
	->Add((AtofTextCtrl_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((AtofTextCtrl_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_Bool(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    if (mpPointerVal != NULL)
	mpPointerCtrl = new wxCheckBoxEX
	    (pWindow, wxID_ANY, 
	     (mbConnectSizer == TRUE) ? mTitle : wxT(""),
	     wxGenericValidator((bool*)mpPointerVal),
	     wxDefaultPosition, wxSize(mItemWidth, -1));
    else // NULL $B$H7k$S$D$/$H$^$:$$(B
	mpPointerCtrl = new wxCheckBoxEX
	    (pWindow, wxID_ANY, 
	     (mbConnectSizer == TRUE) ? mTitle : wxT(""),
	     wxDefaultValidator,
	     wxDefaultPosition, wxSize(mItemWidth, -1));
    if (mItemWidth != -1) mExpand = _GP_NO_EXPAND;
    pBoxSizer
	->Add((wxCheckBoxEX*)mpPointerCtrl, GetSizerFlagH());
    mID = ((wxCheckBoxEX*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_WXString(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = new AtoWXSTextCtrl_cls
	(pWindow, wxID_ANY, GetValueWXStr());
    pBoxSizer
	->Add((AtoWXSTextCtrl_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((AtoWXSTextCtrl_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_WXString_Path(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = new AtoWXSTextPathCtrl_cls
	(pWindow, wxID_ANY, GetValueWXStr());
    pBoxSizer
	->Add((AtoWXSTextPathCtrl_cls*)mpPointerCtrl, GetSizerFlagH());
    mID = ((AtoWXSTextPathCtrl_cls*)mpPointerCtrl)->GetId();
}

void GeneralPointer_cls::
GPConnect_WXPanel(wxBoxSizer* pBoxSizer, wxWindow* pWindow){
    mpPointerCtrl = mpPointerVal;
    pBoxSizer
	->Add((wxPanel*)mpPointerCtrl, GetSizerFlagH());
    mID = ((wxPanel*)mpPointerCtrl)->GetId();
}

wxBoxSizer* GeneralPointer_cls::
GetPSizer(wxWindow* pWindow){
    // $B%?%$%H%kIU$-?75,%5%$%6!<$rJV$9(B
    wxBoxSizer* pBoxSizerIndividual = new wxBoxSizer(wxHORIZONTAL);
    if (IsButton() == FALSE) 
	if (mTitle != wxT("")){
	    mpCTextLabel = 
		new wxStaticTextEX(pWindow, wxID_ANY, mTitle, wxDefaultPosition, GetSizeLabel());
	    pBoxSizerIndividual->Add(mpCTextLabel, GetSizerFlagLabel());
	}
    return pBoxSizerIndividual;
}
