// -*- c -*- 
// "clsGeneralPointers.cpp"

#ifndef _GENERAL_POINTERS_H_
#define _GENERAL_POINTERS_H_

#include <wx/wx.h>
#include <wx/statline.h>
#include <vector>

#define _GP_NOT_DEFINED  0
#define _GP_INTEGER      1
#define _GP_FLOAT        2
#define _GP_DOUBLE       3
#define _GP_BOOL         4
#define _GP_UNSIGNED_INT 5
#define _GP_UNSIGNED_INT_LIST 6
#define _GP_INTEGER_SLIDER 7
#define _GP_FLOAT_SLIDER 8
#define _GP_DOUBLE_SLIDER 9
#define _GP_STATIC_TEXT  10
#define _GP_INTEGER_SPIN 11
#define _GP_UINT_SPIN    12
#define _GP_WXSTRING     13
#define _GP_WXSTRING_PATH 14
#define _GP_WXPANEL       15
#define _GP_DUALSLIDER  16
#define _GP_UNSIGNED_INT_ICON 17
#define _GP_STATIC_LINE   19
#define _GP_BUTTON 20

#define _GP_BUTTON_HIDE_PANEL 50

#define _GP_BOXSIZERHORIZONTAL 60

#define _GP_ALIGN_LEFT  0
#define _GP_ALIGN_RIGHT 1
#define _GP_NO_EXPAND   0
#define _GP_EXPAND      1

#define _GP_ENABLE      TRUE
#define _GP_DISABLE     FALSE

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

class wxStaticTextEX;

class GeneralPointer_cls {
 public:
    GeneralPointer_cls();
    
    void   CombinePointerButton
	(int item, wxString title,  
	 bool enabled = TRUE, bool bHideButton = FALSE);

    void   CombinePointer
	(int type, void* pPointer, wxString title = wxT(""),
	 bool enabled = TRUE,
	 unsigned int listMax = 0, wxString *pChoiceList = NULL, 
	 unsigned int init = 0, unsigned int offset = 0);
    void   ReconnectValidator();
    void   RemakeList();

    void   SetRange(double mMin, double mMax, int scale = 1);
    void   ApplyRange();
    void   DisableHookSpace();

    void  SetValueFloat(float value, float value1 = 0);
    void  SetValueInt(int value);
    void  SetValueBool(bool value);

    int    GetValueInt();
    float  GetValueFloat();
    double GetValueDouble(int index = 0);
    bool   GetValueBool();
    wxString GetValueWXStr();
    void   UpdateValue(bool bForce = FALSE);
    void   UpdateFromText();
    bool   Assign();
    bool   IsDirty();

    bool   IsTabstop();
    bool   IsButton();

    void*    mpPointerVal;
    void*    mpPointerCtrl;
    wxWindow* GetPCtrlWindow();
    wxStaticTextEX* mpCTextLabel;
    float*   mpConnectColorFloat;

    void   PushBackNextWindow(wxWindow* pWindow);
    void   PushBackPrevWindow(wxWindow* pWindow);
    //std::vector < unsigned int > mVConnectedItem;

    void   Enable(bool bEnable = TRUE);
    void   Disable();
    void   Show(bool bShow = TRUE);
    void   Hide();
    void   UpdateLabel();
	
    wxString mTitle;
    int      mType, mID, mProportion, mTextWidth, mItemWidth;
    bool     mbEnabled;
    bool     mbHookSpase;
    double   mMin, mMax;
    int      mScale;

    // mOffset $B$O(B GP_BUTTON $B$G$b(B item $B$H$7$FMxMQ(B
    unsigned int mListMax, mInit, mOffset;
    wxString *mpChoiceList;
    std::vector < wxIcon > *mpVIcon; // $B2>(B
    wxString GetInitStr();

    wxSize GetSizeLabel();
    wxSizerFlags GetSizerFlagLabel();
    wxSizerFlags GetSizerFlagH(int borderWidth = 2);
    wxSizerFlags GetSizerFlagV();
    void     AlignRight();
    int      mAlign;
    int      mExpand;
    bool     mbBorder;

    bool     mbConnectSizer;
    unsigned int mSizerIndex;

    int  GPConnect_BoxSizer(wxBoxSizer* pBoxSizer, wxWindow* pWindow, int proportion);
    void GPConnect_SText(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_SLine(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_Button(wxBoxSizer* pBoxSizer, wxWindow* pWindow, int borderWidth = 2);
    void GPConnect_Integer(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_IntegerSpin(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_List(wxBoxSizer* pBoxSizer, wxWindow* pWindow, int width);
    void GPConnect_SliderInt(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_SliderFloat(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_SliderDouble(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_Float(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_Double(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_Bool(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_String(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_WXString(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_WXString_Path(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_WXPanel(wxBoxSizer* pBoxSizer, wxWindow* pWindow);
    void GPConnect_DualSlider(wxBoxSizer* pBoxSizer, wxWindow *pWindow);
    void GPConnect_IconSelect(wxBoxSizer* pBoxSizer, wxWindow *pWindow);
    
    wxBoxSizer* GetPSizer(wxWindow* pWindow);
};

#endif
