// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsMultiNumericalInputs.h"

#include <wx/valgen.h>
#include "clsMultiNumericalInputs.h"
#include "clsAtofTextCtrl.h"
#include "clsAtoiTextCtrl.h"
#include "clsAtoiComboCtrl.h"
#include "clsAtoSTextCtrl.h"
#include "clsSliderInt.h"
#include "clsSliderFloat.h"
#include "EX/clsCheckBoxEX.h"
#include "EX/clsStaticTextEX.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"
#include "../MYAPI/MsgBox.h"

extern int GClientDisplayBottomPos;

#define wxComEH wxCommandEventHandler
#define wxMouseEH wxMouseEventHandler

MultiNumericalInputsBase_cls::
MultiNumericalInputsBase_cls(void **ppToSub) :
mpBTNAssign (NULL), mpBTNUpDate(NULL), mpBTNHide(NULL), mpBTNOK(NULL), mpBTNHelp(NULL)
{
    mppToSub = ppToSub;
    mbDirty  = FALSE;
    mbHideButton = TRUE;
    mSizerIndex   = 0;
    mSizerCounter = 0;
    mpBoxSizerSrc = NULL;
    mpDialogAdvanced = NULL;
    
    mDefaultButton = _MN_ASSIGN_UPDATE_HIDE;
    mTextWidth = 40;
    mButtonWidth = -1;//210;
    mComboWidth = -1;
    mCheckBoxWidth = -1;
    mItemWidth = -1;
#ifdef WIN32
    mMinimumWidth = 220;
#else
    mMinimumWidth = 250;
#endif
    mProportion = 0;
    mbAssignHook = TRUE;
    mbPanelOpen = TRUE;
}

int MultiNumericalInputsBase_cls::
CombinePointerButton(int item, wxString title, bool enabled, bool bHideButton){
    unsigned int currentIndex = (unsigned int)mVCPointers.size();

    mVCPointers.push_back(GeneralPointer_cls());
    mVCPointers[currentIndex]
	.CombinePointerButton(item, title, enabled, bHideButton);

    mVCPointers[currentIndex].mItemWidth = mItemWidth;
    if (mButtonWidth != -1)
	mVCPointers[currentIndex].mItemWidth = mButtonWidth;
    
    if (mSizerCounter > 0){
	mVCPointers[currentIndex].mbConnectSizer = TRUE;
	mVCPointers[currentIndex].mSizerIndex = mSizerIndex;
	mSizerCounter--;
    }

    return (int)mVCPointers.size() - 1;
}

int MultiNumericalInputsBase_cls::
CombinePointer(int type, void* pPointerVal, wxString title, bool enabled,
	       unsigned int listMax, wxString *pChoiceList, unsigned int init,
	       unsigned int offset){
    unsigned int currentIndex = (unsigned int)mVCPointers.size();

    mVCPointers.push_back(GeneralPointer_cls());
    mVCPointers[currentIndex]
	.CombinePointer(type, pPointerVal, title, enabled,
			listMax, pChoiceList, init, offset);
    mVCPointers[currentIndex].mTextWidth = mTextWidth;

    // $B$3$NJU$O(BGeneralPointer$BB&$K$G$-$k(B
    mVCPointers[currentIndex].mItemWidth = mItemWidth;
    if (type == _GP_BOOL && mCheckBoxWidth != -1)
	mVCPointers[currentIndex].mItemWidth = mCheckBoxWidth;
    
    if (mSizerCounter > 0){
	mVCPointers[currentIndex].mbConnectSizer = TRUE;
	mVCPointers[currentIndex].mSizerIndex = mSizerIndex;
	mSizerCounter--;
    }

    if (type == _GP_BOXSIZERHORIZONTAL){
	mSizerIndex = currentIndex;
	mSizerCounter = *(int*)pPointerVal;
    }

    return (int)mVCPointers.size() - 1;
}

void MultiNumericalInputsBase_cls::
SetRange(double min, double max, int index, int scale){
    if (index == _MN_LASTITEM)
	mVCPointers[mVCPointers.size() - 1].SetRange(min, max, scale);
    else
	mVCPointers[index].SetRange(min, max, scale);
}

void MultiNumericalInputsBase_cls::
DisableHookSpace(int index){
    if (index == _MN_LASTITEM)
	mVCPointers[mVCPointers.size() - 1].DisableHookSpace();
    else
	mVCPointers[index].DisableHookSpace();
}

void MultiNumericalInputsBase_cls::
AlignRight(int index){
    if (index == _MN_LASTITEM)
	mVCPointers[mVCPointers.size() - 1].AlignRight();
    else
	mVCPointers[index].AlignRight();
}

void MultiNumericalInputsBase_cls::
MakeGeometryBase(wxBoxSizer* pBoxSizerSrc, wxWindow* pWindow){
    
    for (unsigned int i = 0; i < mVCPointers.size(); i++){
	
	// $B;R6!$N%5%$%6!<$rMQ0U$9$k!#(B
	if (mVCPointers[i].mType == _GP_BOXSIZERHORIZONTAL){
	    int id = mVCPointers[i].GPConnect_BoxSizer(pBoxSizerSrc, pWindow, mProportion);
	    if (mType == _MN_PANEL && id != -1)
		pWindow->Connect(id, wxEVT_LEFT_DOWN,
				 wxMouseEventHandler(MultiNumericalInputsPanel_cls::OnMouse));
	    continue;
	}

	wxBoxSizer *pBSizerCurrent;
	if (mVCPointers[i].mbConnectSizer == TRUE)
	    // $B;R6!$N%5%$%6!<$K7k$S$D$1$k(B
	    pBSizerCurrent = (wxBoxSizer*)
		mVCPointers[mVCPointers[i].mSizerIndex].mpPointerCtrl;
	else{
	    // $B?75,%5%$%6!<$K7k$S$D$1$k(B($B%?%$%H%kIU$-(B)
	    pBSizerCurrent = mVCPointers[i].GetPSizer(pWindow);
	    if (mType == _MN_PANEL &&
		mVCPointers[i].mpCTextLabel != NULL)
		pWindow->Connect(mVCPointers[i].mpCTextLabel->GetId(),
				 wxEVT_LEFT_DOWN,
				 wxMouseEH(MultiNumericalInputsPanel_cls::OnMouse));
	}

	switch(mVCPointers[i].mType){
	case _GP_STATIC_TEXT:
	    mVCPointers[i].GPConnect_SText(pBSizerCurrent, pWindow);
	    break;
	case _GP_STATIC_LINE:
	    mVCPointers[i].GPConnect_SLine(pBSizerCurrent, pWindow);
	    break;
	case _GP_BUTTON :
	case _GP_BUTTON_HIDE_PANEL : 
	    // $B2#JB$S%\%?%s$J$i%\!<%@!<$r$O$:$9(B
	    mVCPointers[i].GPConnect_Button(pBSizerCurrent, pWindow, 
					    2*mVCPointers[i].mbBorder);
	    ConnectButton(pWindow, mVCPointers[i].mID); 
	    break;
	case _GP_INTEGER:      case _GP_UNSIGNED_INT:
	    mVCPointers[i].GPConnect_Integer(pBSizerCurrent, pWindow);
	    break;
	case _GP_INTEGER_SPIN: case _GP_UINT_SPIN:
	    mVCPointers[i].GPConnect_IntegerSpin(pBSizerCurrent, pWindow);
	    break;
	case _GP_UNSIGNED_INT_LIST: 
	    mVCPointers[i].GPConnect_List(pBSizerCurrent, pWindow, mComboWidth);
	    ConnectComboBox(pWindow, mVCPointers[i].mID);
	    break; 
	case _GP_UNSIGNED_INT_ICON: {
	    mVCPointers[i].GPConnect_IconSelect(pBSizerCurrent, pWindow);
	    ConnectComboBox(pWindow, mVCPointers[i].mID);
	    break; }
	case _GP_INTEGER_SLIDER:
	    mVCPointers[i].GPConnect_SliderInt(pBSizerCurrent, pWindow);
	    ConnectScroll(pWindow, mVCPointers[i].mID);
	    break; 
	case _GP_FLOAT_SLIDER:
	    mVCPointers[i].GPConnect_SliderFloat(pBSizerCurrent, pWindow);
	    ConnectScroll(pWindow, mVCPointers[i].mID);
	    break;
	case _GP_DUALSLIDER:
	    mVCPointers[i].GPConnect_DualSlider(pBSizerCurrent, pWindow);
	    ConnectScroll(pWindow, mVCPointers[i].mID);
	    break;
	case _GP_FLOAT:
	    mVCPointers[i].GPConnect_Float(pBSizerCurrent, pWindow);
	    break;
	case _GP_DOUBLE_SLIDER:
	    mVCPointers[i].GPConnect_SliderDouble(pBSizerCurrent, pWindow);
	    ConnectScroll(pWindow, mVCPointers[i].mID);
	    break;
	case _GP_DOUBLE:
	    mVCPointers[i].GPConnect_Double(pBSizerCurrent, pWindow);
	    break;
	case _GP_BOOL: 
	    mVCPointers[i].GPConnect_Bool(pBSizerCurrent, pWindow);
	    ConnectCheckBox(pWindow, mVCPointers[i].mID);
	    break; 
	case _GP_WXSTRING:
	    mVCPointers[i].GPConnect_WXString(pBSizerCurrent, pWindow);
	    break;
	case _GP_WXSTRING_PATH:
	    mVCPointers[i].GPConnect_WXString_Path(pBSizerCurrent, pWindow);
	    ConnectButton(pWindow, mVCPointers[i].mID); 
	    break;
	case _GP_WXPANEL:
	    mVCPointers[i].GPConnect_WXPanel(pBSizerCurrent, pWindow);
	    ConnectClick(pWindow, mVCPointers[i].mID);
	    break;
	}
	
	mVCPointers[i].ApplyRange();
	
	if (mVCPointers[i].mbEnabled == FALSE)
	    mVCPointers[i].Disable();
	
	if (mVCPointers[i].mbConnectSizer == FALSE)
	    pBoxSizerSrc->Add(pBSizerCurrent, 
			      mVCPointers[i].GetSizerFlagV()/*.Align(wxALIGN_CENTER_VERTICAL)*/);
    }

    if ((mDefaultButton & 3) != 0){
	wxBoxSizer *pBoxSizerButtons 
	    = new wxBoxSizer(wxHORIZONTAL);
	SetDefaultButtons(pWindow, pBoxSizerButtons);
	pBoxSizerSrc->Add(pBoxSizerButtons, 0, wxEXPAND);
    }

    if (mType == _MN_DIALOG){
	int initIdx = SearchNextTabstop(-1, _MN_FORWARD);
	if (initIdx == -1) initIdx = 0;
	wxWindow *pCWindow = 
	    mVCPointers[initIdx].GetPCtrlWindow();
	if (pCWindow != NULL) pCWindow->SetFocus();
    }
    
    mpPadding = pBoxSizerSrc->Add(mMinimumWidth, -1, 0, 0);

    AdjustTabNavigationOrder();
}

void MultiNumericalInputsBase_cls::
AdjustTabNavigationOrder(){
    std::vector < GeneralPointer_cls > VCPointersText;
    std::vector < GeneralPointer_cls > VCPointersSlider;
    for (unsigned int i = 0; i < mVCPointers.size(); i++){
	switch (mVCPointers[i].mType){
	case _GP_INTEGER: case _GP_UNSIGNED_INT:
	case _GP_FLOAT: case _GP_DOUBLE:
	case _GP_INTEGER_SPIN:
	case _GP_UNSIGNED_INT_LIST:
	case _GP_BUTTON:
	    VCPointersText.push_back(mVCPointers[i]);
	    break;
	case _GP_DOUBLE_SLIDER: 
	case _GP_FLOAT_SLIDER:
	case _GP_INTEGER_SLIDER:
	case _GP_DUALSLIDER:
	    VCPointersSlider.push_back(mVCPointers[i]);
	}
    }
    
    if (VCPointersText.size() > 1) 
	for (unsigned int i = 0; i < VCPointersText.size() - 1; i++){
	    ((wxWindow*)VCPointersText[i+1].mpPointerCtrl)->
		MoveAfterInTabOrder((wxWindow*)VCPointersText[i].mpPointerCtrl);
	}

    if (VCPointersSlider.size() > 1) 
	for (unsigned int i = 0; i < VCPointersSlider.size() - 1; i++){
	    ((wxWindow*)VCPointersSlider[i+1].mpPointerCtrl)->
		MoveAfterInTabOrder((wxWindow*)VCPointersSlider[i].mpPointerCtrl);
	}
}

GeneralPointer_cls* MultiNumericalInputsBase_cls::
GetPLastVCPointer(){
    return &(mVCPointers[GetLastVCIndex()]);
}

unsigned int MultiNumericalInputsBase_cls::
GetLastVCIndex(){
    return (unsigned int)mVCPointers.size() - 1;
}

void MultiNumericalInputsBase_cls::
SetItemTitle(unsigned int index, wxString title){
    if (index > GetLastVCIndex()) return;
    mVCPointers[index].mTitle = title;
}

void MultiNumericalInputsBase_cls::
SetColorForXYZ(unsigned int index){
    mVCPointers[index+0].GetPCtrlWindow()->
	SetBackgroundColour(wxColour(255, 244, 244));
    mVCPointers[index+1].GetPCtrlWindow()->
	SetBackgroundColour(wxColour(244, 255, 244));
    mVCPointers[index+2].GetPCtrlWindow()->
	SetBackgroundColour(wxColour(244, 244, 255));
}

void MultiNumericalInputsBase_cls::
SetDefaultButtons(wxWindow* pC, wxBoxSizer *pSizer){
    if ((mDefaultButton & _MN_HELP) != 0){
	mpBTNHelp = new wxButtonEX(pC, wxID_ANY, wxT("Help"));
	pSizer->Add(mpBTNHelp, 0, wxEXPAND);

	if (mType == _MN_DIALOG)
	    pC->Connect(mpBTNHelp->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputs_cls::OnHelp));
	else
	    pC->Connect(mpBTNHelp->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputsPanel_cls::OnHelp));
    }

    if ((mDefaultButton & _MN_ADVANCED) != 0){
	mpBTNAdvanced = new wxButtonEX(pC, wxID_ANY, wxT("Advanced"));
	pSizer->Add(mpBTNAdvanced, 0, wxEXPAND);
	
	if (mType == _MN_DIALOG)
	    pC->Connect(mpBTNAdvanced->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputs_cls::OnAdvanced));
	else
	    pC->Connect(mpBTNAdvanced->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputsPanel_cls::OnAdvanced));
    }
    
    switch(mDefaultButton & 3){
    case _MN_ASSIGN_UPDATE_HIDE:
	mpBTNAssign = new wxButtonEX(pC, wxID_ANY, wxT(" Assign"));
	mpBTNUpDate = new wxButtonEX(pC, wxID_ANY, wxT("Update"));
	pSizer->Add(mpBTNAssign, 1, wxEXPAND);
	pSizer->Add(mpBTNUpDate, 1, wxEXPAND);
	
	if (mbHideButton == TRUE){
	    mpBTNHide   = new wxButtonEX(pC, wxID_CANCEL, wxT("Hide"), 
					 wxDefaultPosition, wxSize(40, -1));
	    pSizer->Add(mpBTNHide,   0, wxEXPAND);
	}
	if (mType == _MN_DIALOG){
	    pC->Connect(mpBTNAssign->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputs_cls::OnAssign));
	    pC->Connect(mpBTNUpDate->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputs_cls::OnUpdate));
	}
	else{
	    pC->Connect(mpBTNAssign->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
		       wxCommandEventHandler(MultiNumericalInputsPanel_cls::OnAssign));
	    pC->Connect(mpBTNUpDate->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
		       wxCommandEventHandler(MultiNumericalInputsPanel_cls::OnUpdate));
	}
	break;
    case _MN_CANCEL_OK:
	pSizer->Add(-1, -1, 1, wxEXPAND);
	mpBTNOK   = new wxButtonEX(pC, wxID_OK, wxT("OK"));
	if (mType == _MN_DIALOG)
	    pC->Connect(mpBTNOK->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputs_cls::OnOK));
	else
	    pC->Connect(mpBTNOK->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputsPanel_cls::OnOK));
	mpBTNHide = new wxButtonEX(pC, wxID_CANCEL, wxT("Cancel"));
	
	pSizer->Add(mpBTNHide, 0, wxEXPAND);
	pSizer->Add(mpBTNOK,   0, wxEXPAND);
	break;
    case _MN_OK_CANCEL:
	pSizer->Add(-1, -1, 1, wxEXPAND);
	mpBTNOK   = new wxButtonEX(pC, wxID_OK, wxT("OK"));
	if (mType == _MN_DIALOG)
	    pC->Connect(mpBTNOK->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputs_cls::OnOK));
	else
	    pC->Connect(mpBTNOK->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
			wxCommandEventHandler(MultiNumericalInputsPanel_cls::OnOK));
	mpBTNHide = new wxButtonEX(pC, wxID_CANCEL, wxT("Cancel"));
	pSizer->Add(mpBTNOK,   0, wxEXPAND);
	pSizer->Add(mpBTNHide, 0, wxEXPAND);
    }
}

int  MultiNumericalInputsBase_cls::
SearchNextTabstop(int index, int direction){
    if (mVCPointers.size() == 1) return 0;
    int nextIndex = index;
    do {
	nextIndex += direction;
	if (nextIndex == (int)mVCPointers.size()){
	    if (index == -1) 
		return -1;
	    else            
		nextIndex = 0;
	}
	if (nextIndex < 0){
	    if (index == (int)mVCPointers.size()) 
		return -1;
	    else
		nextIndex = (int)mVCPointers.size() -1;
	}
	if (nextIndex == index) break;
    } while (mVCPointers[nextIndex].IsTabstop() == FALSE);
    
    return nextIndex;
}

void MultiNumericalInputsBase_cls::
SetTabRing(std::vector < MultiNumericalInputsBase_cls* > VpList){
    std::vector < int > VInitTabIndex, VLastTabIndex;
    std::vector < wxWindow* > VpInitTab, VpLastTab;
    for (unsigned int i = 0; i < VpList.size(); i++){
	unsigned int initIdx = 
	    VpList[i]->SearchNextTabstop(-1, _MN_FORWARD);
	unsigned int lastIdx = 
	    VpList[i]->SearchNextTabstop((int)VpList[i]->mVCPointers.size(), _MN_BACKWARD);
	VInitTabIndex.push_back(initIdx);
	VLastTabIndex.push_back(lastIdx);
	VpInitTab.push_back(VpList[i]->mVCPointers[initIdx].GetPCtrlWindow());
	VpLastTab.push_back(VpList[i]->mVCPointers[lastIdx].GetPCtrlWindow());
    }

    for (unsigned int i = 0; i < VpList.size(); i++){
	for (unsigned int j = i + 1; j < VpList.size(); j++)
	    VpList[i]->mVCPointers[VLastTabIndex[i]].PushBackNextWindow(VpInitTab[j]);
	for (unsigned int j = 0; j <= i; j++)
	    VpList[i]->mVCPointers[VLastTabIndex[i]].PushBackNextWindow(VpInitTab[j]);
	for (int j = i - 1; j >= 0; j--)
	    VpList[i]->mVCPointers[VInitTabIndex[i]].PushBackPrevWindow(VpLastTab[j]);
	for (int j = (int)VpList.size() - 1 ; j >= (int)i; j--)
	    VpList[i]->mVCPointers[VInitTabIndex[i]].PushBackPrevWindow(VpLastTab[j]);
    }
}

void MultiNumericalInputsBase_cls::
UpdateValues(int index, int exceptionID, bool bForce){
    for (int i = 0; i < (int)mVCPointers.size(); i++){
	if (index != _MN_ALLITEM) i = index;
	if (mVCPointers[i].mID == exceptionID) continue;
	mVCPointers[i].UpdateLabel();
	mVCPointers[i].UpdateValue(bForce);
	if (index != _MN_ALLITEM) return;
    }
    mbDirty = FALSE;
}

void MultiNumericalInputsBase_cls::
UpdateValuesAll(bool bForce){
    UpdateValues(_MN_ALLITEM, _MN_NOITEM, bForce);
}

bool MultiNumericalInputsBase_cls::
CheckDirty(){
    bool dirty = FALSE;
    for (unsigned int i = 0; i < mVCPointers.size(); i++)
	if (mVCPointers[i].IsDirty() == TRUE) dirty = TRUE;
    
    return dirty;
}

void MultiNumericalInputsBase_cls::
UpdateFromText(unsigned int index){
    for (int i = 0; i < (int)mVCPointers.size(); i++){
	if (index != _MN_ASSIGN_ALL) i =index;
	mVCPointers[i].UpdateFromText();
	if (index != _MN_ASSIGN_ALL) return;
    }
    mbDirty = FALSE;
}

void MultiNumericalInputsBase_cls::
AssignBase(unsigned int index){
    
    for (int i = 0; i < (int)mVCPointers.size(); i++){
	if (index != _MN_ASSIGN_ALL) i = index;
	if (mVCPointers[i].mpPointerVal == NULL) continue;

	CacheOriginalVaues(i);
	bool bChange = mVCPointers[i].Assign();
	if (bChange)
	    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
		if (mVpStructFuncHook[i].mType == _MN_HOOK_CHANGE)
		    (*mVpStructFuncHook[i].mpFunc)((*mppToSub), &index);
	
	if (index != _MN_ASSIGN_ALL) break;
    }
    
    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
	if (mVpStructFuncHook[i].mType == _MN_HOOK_ASSIGN){
	    (*mVpStructFuncHook[i].mpFunc)((*mppToSub), &mVCPointers[index]);
	}
    
    if (index == _MN_ASSIGN_ALL) mbDirty = FALSE;
}

void MultiNumericalInputsBase_cls::
AssignBaseById(int id){
    for (unsigned int index = 0; index < mVCPointers.size(); index++)
	if (id == mVCPointers[index].mID)
	    AssignBase(index);
}

void MultiNumericalInputsBase_cls::
ShowItem(unsigned int index, bool bShow){
    mVCPointers[index].Show(bShow);
}

void MultiNumericalInputsBase_cls::
HideItem(unsigned int index){
    mVCPointers[index].Show(FALSE);
}

void MultiNumericalInputsBase_cls::
TogglePanelOpenStatus(unsigned int index){
    mbPanelOpen = !mbPanelOpen;
    for (unsigned int i = index; i < mVCPointers.size(); i++)
	ShowItem(i, mbPanelOpen);
}

void MultiNumericalInputsBase_cls::
ColoringItem(int index, wxColour color){
    if (index != -1){
	mVCPointers[index].GetPCtrlWindow()->SetBackgroundColour(color);
	return;
    }
    
    for (int i = 0; i < mVCPointers.size(); i++){
	if (mVCPointers[i].GetPCtrlWindow() != NULL)
	    mVCPointers[i].GetPCtrlWindow()->SetBackgroundColour(color);
    }
}

void MultiNumericalInputsBase_cls::
OnScrollBase(wxCommandEvent& event){
    int id = event.GetId();
    for (unsigned int i = 0; i < mVCPointers.size(); i++)
	if ((mVCPointers[i].mType == _GP_INTEGER_SLIDER || 
	     mVCPointers[i].mType == _GP_FLOAT_SLIDER ||
	     mVCPointers[i].mType == _GP_DUALSLIDER ||
	     mVCPointers[i].mType == _GP_DOUBLE_SLIDER)
	    && mVCPointers[i].mID   == id)
	    AssignBase(i);
    UpdateValues(_MN_ALLITEM, id);
    event.Skip();
}

void MultiNumericalInputsBase_cls::
OnScrollThumbBase(wxCommandEvent& event){
    int id = event.GetId();
    int index = _MN_NOITEM;
    for (unsigned int i = 0; i < mVCPointers.size(); i++)
	if ((mVCPointers[i].mType == _GP_INTEGER_SLIDER || 
	     mVCPointers[i].mType == _GP_FLOAT_SLIDER ||
	     mVCPointers[i].mType == _GP_DUALSLIDER ||
	     mVCPointers[i].mType == _GP_DOUBLE_SLIDER )
	    && mVCPointers[i].mID == id){
	    AssignBase(i);
	    index = i; 
	}
    
    for (unsigned int i = 0; i < mVCPointers.size(); i++)
	if ((mVCPointers[i].mpPointerVal == mVCPointers[index].mpPointerVal) &&
	    (i != index))
	    UpdateValues(i);
    
    event.Skip();
}

void MultiNumericalInputsBase_cls::
OnComboBoxBase(wxCommandEvent& event, void *pThis){
    bool bChange = FALSE;
    int id = event.GetId();
    for (unsigned int i = 0; i < mVCPointers.size(); i++)
	if ((mVCPointers[i].mType == _GP_UNSIGNED_INT_LIST ||
	     mVCPointers[i].mType == _GP_UNSIGNED_INT_ICON) &&
	    mVCPointers[i].mID   == id){
	    int orgValue = mVCPointers[i].GetValueInt();
	    AssignBase(i);
	    int newValue = mVCPointers[i].GetValueInt();
	    if (orgValue != newValue) bChange = TRUE;
	}

    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++){
	if (mVpStructFuncHook[i].mType == _MN_HOOK_COMBOBOX)
	    (*mVpStructFuncHook[i].mpFunc)(pThis, &event);
	if (mVpStructFuncHook[i].mType == _MN_HOOK_COMBOBOX_CHANGE &&
	    bChange == TRUE)
	    (*mVpStructFuncHook[i].mpFunc)(pThis, &event);
    }
    
    event.Skip();
}

void MultiNumericalInputsBase_cls::
SetDefaultButtonType(int type){
    mDefaultButton = type;
}

void MultiNumericalInputsBase_cls::
CacheOriginalVaues(int index){
    switch (mVCPointers[index].mType){
    case _GP_FLOAT: case _GP_FLOAT_SLIDER:
	mCachedFloat = mVCPointers[index].GetValueFloat();
	break;
    case _GP_INTEGER: case _GP_INTEGER_SLIDER:
    case _GP_INTEGER_SPIN: 
    case _GP_UNSIGNED_INT: case _GP_UNSIGNED_INT_LIST:
    case _GP_UINT_SPIN:
	mCachedInt = mVCPointers[index].GetValueInt();
	break;
    case _GP_BOOL:
	// BOOL$B$N>l9g$@$1$3$N;~E@$G=hM}:Q$_$J$N$GH?E>(B
	mCachedBool = !mVCPointers[index].GetValueBool();
	break;
    }
}

void MultiNumericalInputsBase_cls::
SetFuncHook(void (*pFunc)(void*, void*), int type, int ID, void *pEventOvwr){
    structFuncPointerVoidVoid structFunc;
    structFunc.mpFunc = pFunc;
    structFunc.mType = type;
    structFunc.mId = ID;
    structFunc.mpEventOvwt = pEventOvwr;
    mVpStructFuncHook.push_back(structFunc);
}

// ----------------

BEGIN_EVENT_TABLE(MultiNumericalInputsPanel_cls, wxPanel)
//EVT_BUTTON     (buttonCtrlAssign, MultiNumericalInputsPanel_cls::OnAssign)
     EVT_BUTTON     (wxID_CANCEL, MultiNumericalInputsPanel_cls::OnHide)
     //EVT_BUTTON     (buttonCtrl02, MultiNumericalInputsPanel_cls::OnUpdate)
     // wxID_OK $B$,$&$^$/4XO"IU$1$5$l$J$+$C$?$N$G!"(BConnect$B$G<BAu(B
     EVT_SIZE       (MultiNumericalInputsPanel_cls::OnSize)
     EVT_CHILD_FOCUS(MultiNumericalInputsPanel_cls::OnChildFocusEventCustum)
     EVT_CHAR       (MultiNumericalInputsPanel_cls::OnChar)
     EVT_KEY_DOWN   (MultiNumericalInputsPanel_cls::OnKeyDown)
     EVT_MOUSE_EVENTS(MultiNumericalInputsPanel_cls::OnMouse)
END_EVENT_TABLE()

MultiNumericalInputsPanel_cls::
MultiNumericalInputsPanel_cls(MultiNumericalInputsPanel_cls **ppToThis,
			      wxWindow *parent, wxWindowID id,
			      const wxPoint& pos,
			      const wxSize&  size,
			      long style,
			      const wxString& name) :
MultiNumericalInputsBase_cls((void**)ppToThis),
    wxPanel(parent, id, pos, size, style, name)
{
    mppToThis = ppToThis;
    mType = _MN_PANEL;
    SetBackgroundColour(wxColour(wxT("WHEAT")));
    SetWindowStyle(GetWindowStyle() | wxTAB_TRAVERSAL);
    //mpBTNMinimize = NULL;
    //mpCheckBoxValid = NULL;
    
    InheritAttributes();
}

MultiNumericalInputsPanel_cls::
~MultiNumericalInputsPanel_cls(){
    *mppToThis = NULL;
}

void MultiNumericalInputsPanel_cls::
SetPaddingWidth(int paddingWidth){
    mpPadding->SetInitSize(paddingWidth, -1);
    Fit();
}

int MultiNumericalInputsPanel_cls::
GetPaddingWidth(int minimumWidth){
    int width = GetClientSize().GetWidth();
    if (width > minimumWidth) minimumWidth = width;
    return minimumWidth;
}

void MultiNumericalInputsPanel_cls::
MakeGeometry(){
    if (mpBoxSizerSrc == NULL)
	mpBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    MakeGeometryBase(mpBoxSizerSrc, this);
    this->SetSizer/*AndFit*/(mpBoxSizerSrc);
}

void MultiNumericalInputsPanel_cls::
ConnectButton(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_COMMAND_BUTTON_CLICKED,
		     wxComEH(MultiNumericalInputsPanel_cls::OnButton));
}

void MultiNumericalInputsPanel_cls::
ConnectComboBox(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_COMMAND_COMBOBOX_SELECTED,
		     wxComEH(MultiNumericalInputsPanel_cls::OnComboBox));
}

void MultiNumericalInputsPanel_cls::
ConnectCheckBox(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_COMMAND_CHECKBOX_CLICKED,
		     wxComEH(MultiNumericalInputsPanel_cls::OnCheckBox));
}

void MultiNumericalInputsPanel_cls::
ConnectScroll(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_SCROLL_CHANGED,
		     wxComEH(MultiNumericalInputsPanel_cls::OnScroll));
    pWindow->Connect(id, wxEVT_SCROLL_THUMBTRACK,
		     wxComEH(MultiNumericalInputsPanel_cls::OnScrollThumb));
}

void MultiNumericalInputsPanel_cls::
ConnectClick(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_LEFT_DOWN,
		     wxMouseEH(MultiNumericalInputsPanel_cls::OnMouseOnItem));
    pWindow->Connect(id, wxEVT_RIGHT_DOWN,
		     wxMouseEH(MultiNumericalInputsPanel_cls::OnMouseOnItem));
}

void MultiNumericalInputsPanel_cls::
OnAssign(wxCommandEvent& WXUNUSED(event)){
    AssignBase(_MN_ASSIGN_ALL);
    if (mpBTNAssign != NULL)
	mpBTNAssign->SetLabel(wxT(" Assign"));

    wxPaintEvent dummyEvent;
    GetParent()->AddPendingEvent(dummyEvent);
}

void MultiNumericalInputsPanel_cls::
OnHide(wxCommandEvent& event){
    ShowPanel(FALSE);
    wxSizeEvent dummyEvent;
    GetParent()->AddPendingEvent(dummyEvent);
}

void MultiNumericalInputsPanel_cls::
OnShow(wxCommandEvent& event){
    ShowPanel(TRUE);
}

void MultiNumericalInputsPanel_cls::
OnButton(wxCommandEvent& event){
    int id = event.GetId();
    for (unsigned int i = 0; i < mVCPointers.size(); i++){
	if (mVCPointers[i].mType == _GP_BUTTON && mVCPointers[i].mID   == id){
	    if (mVCPointers[i].mpPointerVal != NULL)
		((void(*)())mVCPointers[i].mpPointerVal)();
	    else{
		event.SetId(mVCPointers[i].mOffset);
		GetParent()->AddPendingEvent(event);
	    }
	}
	if (mVCPointers[i].mType == _GP_WXSTRING_PATH && mVCPointers[i].mID == id){
	    //mVCPointers[i].UpdateFromText();
	    //VCPointers[i].Assign();
	    AssignBase(i);
	}
	
    }

    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
	if (mVpStructFuncHook[i].mType == _MN_HOOK_BUTTON &&
	    (mVpStructFuncHook[i].mId == id || 
	     mVpStructFuncHook[i].mId == _MN_HOOK_ID_ANY)){
	    (*mVpStructFuncHook[i].mpFunc)(this, &event);
	}

}

void MultiNumericalInputsPanel_cls::
OnCheckBox(wxCommandEvent& event){
    wxPaintEvent dummyEvent;
    wxGetTopLevelParent(this)->AddPendingEvent(dummyEvent);
    
    int id = event.GetId();
    unsigned int index;
    for (unsigned int i = 0; i < mVCPointers.size(); i++){
	if (id == mVCPointers[i].mID) index = i;
    }
    
    CacheOriginalVaues(index);

    event.Skip();
    
    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++){
	if (mVpStructFuncHook[i].mType == _MN_HOOK_CHECKBOX)
	    (*mVpStructFuncHook[i].mpFunc)(this, &event);
	if (mVpStructFuncHook[i].mType == _MN_HOOK_CHANGE)
	    (*mVpStructFuncHook[i].mpFunc)(this, &index);
    }
}

void MultiNumericalInputsPanel_cls::
OnComboBox(wxCommandEvent& event){
    OnComboBoxBase(event, this);
}

void MultiNumericalInputsPanel_cls::
OnScroll(wxCommandEvent& event){
    OnScrollBase(event);
    GetParent()->AddPendingEvent(event);
}

void MultiNumericalInputsPanel_cls::
OnScrollThumb(wxCommandEvent& event){
    OnScrollThumbBase(event);
}

void MultiNumericalInputsPanel_cls::
OnUpdate(wxCommandEvent& WXUNUSED(event)){
    UpdateValues();
    mpBTNAssign->SetLabel(wxT("Assign"));
}

void MultiNumericalInputsPanel_cls::
OnOK(wxCommandEvent& WXUNUSED(event)){
    wxCommandEvent dummyEvent;
    OnAssign(dummyEvent);
    OnHide(dummyEvent);
}

void MultiNumericalInputsPanel_cls::
OnChar(wxKeyEvent& event){
    if (event.GetKeyCode() == WXK_RETURN || event.GetKeyCode() == WXK_SPACE){
	AssignBaseById(event.GetId()); 
	wxGetTopLevelParent(this)->AddPendingEvent(event);
	UpdateValues();
	if (mpBTNAssign != NULL)
	    mpBTNAssign->SetLabel(wxT("Assign"));
	/*
	else
	    GetParent()->Refresh(FALSE);
	*/
    }
}

void MultiNumericalInputsPanel_cls::
OnKeyDown(wxKeyEvent& event){
    //wxPostEvent(GetParent()->GetEventHandler(), event);
    if (event.GetKeyCode() == WXK_DELETE){
	GetParent()->AddPendingEvent(event);
    }

    wxGetTopLevelParent(this)->AddPendingEvent(event);
}

void MultiNumericalInputsPanel_cls::
OnMouse(wxMouseEvent& event){
    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
	if (mVpStructFuncHook[i].mType == _MN_HOOK_MOUSE)
	    (*mVpStructFuncHook[i].mpFunc)(this, &event);
    event.Skip();
}

/*
void MultiNumericalInputsPanel_cls::
OnMouseOnHide(wxMouseEvent& event){
    if (mpBTNMinimize == NULL) return;
    event.SetId(mpBTNMinimize->GetId());
    mpBTNMinimize->AddPendingEvent(event);
}
*/

void MultiNumericalInputsPanel_cls::
OnMouseOnItem(wxMouseEvent& event){
    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
	if (mVpStructFuncHook[i].mType == _MN_HOOK_PANELCLICK)
	    (*mVpStructFuncHook[i].mpFunc)(this, &event);
}

void MultiNumericalInputsPanel_cls::
OnChildFocusEventCustum(wxChildFocusEvent& event){
    if (mbDirty == FALSE){
	mbDirty = CheckDirty();
	if (mbDirty == TRUE){
	    if (mpBTNAssign != NULL){
		mpBTNAssign->SetLabel(wxT("*Assign"));
		mpBTNAssign->Refresh(TRUE);
	    }
	    Refresh(TRUE);
	}
    }
    event.Skip();
}

void MultiNumericalInputsPanel_cls::
OnSize(wxSizeEvent& event){
    UpdateValues();
    event.Skip();
}

void MultiNumericalInputsPanel_cls::
OnHelp(wxCommandEvent& event){
    wxMessageBox(mstrHelp, wxT("Help Message"), wxOK | wxICON_INFORMATION, this);
}

void MultiNumericalInputsPanel_cls::
OnAdvanced(wxCommandEvent& event){
    //wxMessageBox(mstrHelp, wxT("Help Message"), wxOK | wxICON_INFORMATION, this);
}

/*
void MultiNumericalInputsPanel_cls::
SetMinimizeButtonLabel(wxString strLabel){
    // FIX_ME $B0BA4AuCV$N(Bif$B$,$J$$(B
    mpBTNMinimize->SetLabel(strLabel);
    mVCPointers[mIndexMinimize].GetPCtrlWindow()->SetLabel(strLabel);
}
*/

void MultiNumericalInputsPanel_cls::
RefreshBoth(bool mode){
    Refresh(mode);
}

void MultiNumericalInputsPanel_cls::
ShowPanel(bool mode){
    bool bShownPrev = IsShownOnScreen();

    Show(mode);
    
    if (bShownPrev == mode) return;

    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
	if (mVpStructFuncHook[i].mType == _MN_HOOK_SHOW){
	    (*mVpStructFuncHook[i].mpFunc)(this, &mode);
	}
}

void MultiNumericalInputsPanel_cls::
ShowItems(int index, int num, bool bShow){
    for (int i = index; i < index+num; i++) mVCPointers[i].Show(bShow);
}

void MultiNumericalInputsPanel_cls::
EnableItems(int index, int num, bool bShow){
    for (int i = index; i < index+num; i++) mVCPointers[i].Enable(bShow);
}

void MultiNumericalInputsPanel_cls::
SetHilight(bool bHilight){
    
    for (size_t i = 0; i < mVCPointers.size(); i++)
	if (mVCPointers[i].mType == _GP_BUTTON_HIDE_PANEL)
	    if (bHilight == TRUE){
		((wxButton*)mVCPointers[i].mpPointerCtrl)->SetBackgroundColour(wxColour(wxT("LIGHT BLUE")));
	    }
	    else{
		((wxButton*)mVCPointers[i].mpPointerCtrl)->SetBackgroundColour(wxColour(wxNullColour));
	    }

    if (bHilight == TRUE) SetFocus();
}

void MultiNumericalInputsPanel_cls::
AddToSizer(wxBoxSizer *pBoxSizer, int index){
    wxSizerFlags sflag0(0), sflag0E(0), sflag1E(1);
    sflag0          .Border(wxALL, 2);
    sflag0E.Expand().Border(wxALL, 2);
    sflag1E.Expand();
    if (index == -1)
	pBoxSizer->Add(this, sflag0);
    else
	pBoxSizer->Insert(index, this, sflag0);
}

int MultiNumericalInputsPanel_cls::
FinishSetting(wxBoxSizer *pBoxSizer, bool vis, int index){
    MakeGeometry();
    ShowPanel(vis);
    if (pBoxSizer != NULL) AddToSizer(pBoxSizer, index);
    UpdateValuesAll(TRUE);
    GetParent()->Layout();
    return 0;
}

// ------------------------------------------------------------------

BEGIN_EVENT_TABLE(MultiNumericalInputs_cls, wxDialog)
//EVT_BUTTON     (buttonCtrlAssign, MultiNumericalInputs_cls::OnAssign)
     EVT_BUTTON     (wxID_CANCEL, MultiNumericalInputs_cls::OnHide)
     //EVT_BUTTON     (buttonCtrl02, MultiNumericalInputs_cls::OnUpdate)
     // wxID_OK $B$,$&$^$/4XO"IU$1$5$l$J$+$C$?$N$G!"(BConnect$B$G<BAu(B
     EVT_SIZE       (MultiNumericalInputs_cls::OnSize)
     EVT_CHILD_FOCUS(MultiNumericalInputs_cls::OnChildFocusEventCustum)
     EVT_CHAR       (MultiNumericalInputs_cls::OnChar)
END_EVENT_TABLE()

MultiNumericalInputs_cls::
MultiNumericalInputs_cls(MultiNumericalInputs_cls **ppToThis,
			 wxWindow *parent, wxWindowID id, wxString title,
			 const wxPoint& pos,
			 const wxSize&  size,
			 long style,
			 const wxString& name) :
MultiNumericalInputsBase_cls((void**)ppToThis),
    wxDialogEXKey(parent, id, title, pos, size, NULL, style, name)
{
    mppToThis = ppToThis;
    mType     = _MN_DIALOG;
    
    SetBackgroundColour(wxColour(wxT("WHEAT")));
    SetWindowStyle(GetWindowStyle() | (wxTAB_TRAVERSAL
				       & ~wxMINIMIZE_BOX
				       & ~wxMAXIMIZE_BOX));
    mTitle = title;
}

MultiNumericalInputs_cls::
~MultiNumericalInputs_cls(){
    //MakeModal(FALSE);
    *mppToThis = NULL;
}

void MultiNumericalInputs_cls::
MakeGeometry(){
    if (mpBoxSizerSrc == NULL)
	mpBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    
    MakeGeometryBase(mpBoxSizerSrc, this);
    SetSizerAndFit(mpBoxSizerSrc);
}

void MultiNumericalInputs_cls::
OpenDialogEX(){
    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
	if (mVpStructFuncHook[i].mType == _MN_HOOK_OPEN)
	    (*mVpStructFuncHook[i].mpFunc)(this, NULL);
    ShowModal();
}

void MultiNumericalInputs_cls::
ConnectButton(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_COMMAND_BUTTON_CLICKED,
		     wxComEH(MultiNumericalInputs_cls::OnButton));
}

void MultiNumericalInputs_cls::
ConnectComboBox(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_COMMAND_COMBOBOX_SELECTED,
		     wxComEH(MultiNumericalInputs_cls::OnComboBox));
}

void MultiNumericalInputs_cls::
OnScroll(wxCommandEvent& event){
    OnScrollBase(event);
}

void MultiNumericalInputs_cls::
OnScrollThumb(wxCommandEvent& event){
    OnScrollThumbBase(event);
}

void MultiNumericalInputs_cls::
ConnectCheckBox(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_COMMAND_CHECKBOX_CLICKED,
		     wxComEH(MultiNumericalInputs_cls::OnCheckBox));
}

void MultiNumericalInputs_cls::
ConnectScroll(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_SCROLL_CHANGED,
		     wxComEH(MultiNumericalInputs_cls::OnScroll));
    pWindow->Connect(id, wxEVT_SCROLL_THUMBTRACK,
		     wxComEH(MultiNumericalInputs_cls::OnScrollThumb));
}

void MultiNumericalInputs_cls::
ConnectClick(wxWindow *pWindow, int id){
    pWindow->Connect(id, wxEVT_LEFT_DOWN,
		     wxMouseEH(MultiNumericalInputs_cls::OnMouseOnItem));
    pWindow->Connect(id, wxEVT_RIGHT_DOWN,
		     wxMouseEH(MultiNumericalInputs_cls::OnMouseOnItem));
}

void MultiNumericalInputs_cls::
OnAssign(wxCommandEvent& WXUNUSED(event)){
    AssignBase(_MN_ASSIGN_ALL);
    if (mpBTNAssign != NULL)
	mpBTNAssign->SetLabel(wxT(" Assign"));

    SetLabel(mTitle);
    wxPaintEvent dummyEvent;
    GetParent()->AddPendingEvent(dummyEvent);
}

void MultiNumericalInputs_cls::
OnHide(wxCommandEvent& WXUNUSED(event)){
    if ((mDefaultButton & 3) == _MN_CANCEL_OK ||
	(mDefaultButton & 3) == _MN_OK_CANCEL)
	UpdateValues();
    CloseEXKey(wxID_CANCEL);
    wxPaintEvent dummyEvent;
    GetParent()->AddPendingEvent(dummyEvent);
}

void MultiNumericalInputs_cls::
OnButton(wxCommandEvent& event){
    int id = event.GetId();
    for (unsigned int i = 0; i < mVCPointers.size(); i++)
	if (mVCPointers[i].mType == _GP_BUTTON && mVCPointers[i].mID == id){
	    if (mVCPointers[i].mpPointerVal != NULL)
		((void(*)())mVCPointers[i].mpPointerVal)();
	    else{
		event.SetId(mVCPointers[i].mOffset);
		GetParent()->AddPendingEvent(event);
	    }
	}

    Freeze();
    for (unsigned int i = 0; i < mVpStructFuncHook.size(); i++)
	if (mVpStructFuncHook[i].mType == _MN_HOOK_BUTTON &&
	    (mVpStructFuncHook[i].mId == id ||
	     mVpStructFuncHook[i].mId == _MN_HOOK_ID_ANY)){
	    (*mVpStructFuncHook[i].mpFunc)(this, &event);
	}
    Thaw();
}

void MultiNumericalInputs_cls::
OnCheckBox(wxCommandEvent& event){
    wxPaintEvent dummyEvent;
    GetParent()->AddPendingEvent(dummyEvent);
}

void MultiNumericalInputs_cls::
OnComboBox(wxCommandEvent& event){
    OnComboBoxBase(event, this);
}

void MultiNumericalInputs_cls::
OnUpdate(wxCommandEvent& WXUNUSED(event)){
    UpdateValues();
    mpBTNAssign->SetLabel(wxT(" Assign"));
    SetLabel(mTitle);
}

void MultiNumericalInputs_cls::
OnOK(wxCommandEvent& WXUNUSED(event)){
    UpdateFromText(_MN_ASSIGN_ALL);
    wxCommandEvent dummyEvent;
    OnAssign(dummyEvent);
    CloseEXKey(wxID_OK);
    wxSizeEvent dummyEvent2;
    GetParent()->AddPendingEvent(dummyEvent2);
}

void MultiNumericalInputs_cls::
OnChildFocusEventCustum(wxChildFocusEvent& event){
    if (mbDirty == FALSE){
	mbDirty = CheckDirty();
	if (mbDirty == TRUE && mpBTNAssign != NULL){
	    mpBTNAssign->SetLabel(wxT("*Assign"));
	    mpBTNAssign->Refresh(TRUE);
	    SetLabel(wxT("*") + mTitle);
	}
    }
    event.Skip();
}

void MultiNumericalInputs_cls::
OnMouseOnItem(wxMouseEvent& event){
}

void MultiNumericalInputs_cls::
OnSize(wxSizeEvent& event){
    UpdateValues();
    event.Skip();
}

void MultiNumericalInputs_cls::
OnHelp(wxCommandEvent& event){
    wxMessageBox(mstrHelp, wxT("Help Message"), wxOK | wxICON_INFORMATION, this);
}

void MultiNumericalInputs_cls::
OnAdvanced(wxCommandEvent& event){
    if (mpDialogAdvanced != NULL){
	mpDialogAdvanced->UpdateValues();
	mpDialogAdvanced->OpenDialogEX();/*ShowModal();*/
    }
}

void MultiNumericalInputs_cls::
OnChar(wxKeyEvent& event){
    if (event.GetKeyCode() == WXK_RETURN || event.GetKeyCode() == WXK_SPACE){
	AssignBaseById(event.GetId()); 
	UpdateValues();
    }
}

int MultiNumericalInputs_cls::
FinishSettingDLG(bool vis, bool modal, bool bMakeGeom){
    if (bMakeGeom) MakeGeometry();

    wxRect rect = GetScreenRect();
    //MsgBoxf(this, "!!", "%d %d %d", rect.y, rect.height, GDisplayHeight);

    if (rect.y + rect.height > GClientDisplayBottomPos)
	Move(rect.x, GClientDisplayBottomPos - rect.height);
    
    if (modal == TRUE) return ShowModal();
    
    Show(vis);
    UpdateValuesAll(TRUE);
    return 0;
}
