// -*- c -*-
// "clsMultiPointInput.h"

#include "clsMultiPointInput.h"
#include "clsMultiNumericalInputs.h"
#include "clsAtoiComboCtrl.h"

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"
#include "../../clsFrameMain.h"
#include "../../clsMain.h"
#include "../../clsTimeCtrl.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/clsDF3SeqCtrl.h"
#include "../DF3/clsDF3Sequence.h"
#include "../MYAPI/MsgBox.h"
#include "../PARAMETER/clsParametersPanel.h"
#include "../PARAMETER/clsTimeLinePanel.h"
#include "../UNDO/clsUndoManager.h"

extern UndoManager_cls GCUndoList;

MultiPointInputGUI_cls::
MultiPointInputGUI_cls(wxWindow *pC, MultiPointInput_cls* pCParent,
		       wxString strTitle,
		       wxPoint pos, wxSize size, 
		       std::wstring strItemNum, std::wstring strTag, 
		       unsigned int dimention, wxBitmap *pBitmap) :
    MultiParameters_cls(strItemNum, strTag, dimention),
	MultiParametersGUI_cls(this, pC, strTitle, pos, size, pBitmap)
    {
    mpCParentPInputBase = pCParent;
}

void MultiPointInputGUI_cls::
UpdateValues(){
    mpCMNumMain->UpdateValues();
    mpCMNumParam->UpdateValues();
}

void MultiPointInputGUI_cls::
UpdateInterpolatedValues(double doubleIndex){
    if (mCItemNum.mBody == 0) return;
    for (unsigned int i = 0; i < mCItemNum.mBody; i++)
	mVCParam[i].GetPInterpolatedValue(doubleIndex);
    if (IsShownOnScreen() == TRUE) UpdateValues();
}

void MultiPointInputGUI_cls::
OnApplyBase(){
    if (mCItemNum.mBody == 0) return;
    mVCParam[mActiveItem].SetKeyIfChanged
	(mpCParentPInputBase->mpCParentFrame->mpCParent->mpCTime->GetDoubleIndex());
    mpCParentPInputBase->mpCParentFrame->mbPathChanged = TRUE;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    mpCParentPInputBase->mpCParentFrame->AddPendingEvent(com);
    mpCParentPInputBase->UpdateItems(TRUE);
}

void MultiPointInputGUI_cls::
UpdateComboBoxPropagate(){
    if (mCItemNum.mBody > 0)
	mpCMNumParam->Enable(TRUE);
    else
	mpCMNumParam->Disable();
}

void MultiPointInputGUI_cls::
AddItemPropagate(){
    UpdateComboBox(); 
    mpCMNumMain->UpdateValues();
    //mpCMNumParam->UpdateValues();
}

void MultiPointInputGUI_cls::
RemoveItemPropagate(){
    UpdateComboBox();
    UpdateActiveOnTL();
    mpCMNumMain->UpdateValues();
    //mpCMNumParam->UpdateValues();
}

// -----------------------------------------

MultiPointInput_cls::
MultiPointInput_cls()  
{
}

void MultiPointInput_cls::
SetPGUI(MultiPointInputGUI_cls* pCGUI){
    mpCGUIBase = pCGUI;
    SetPDialog(pCGUI);
}

void MultiPointInput_cls::
SetTag(wxString strTag){
    mTagStr = strTag;
}

void MultiPointInput_cls::
SetParent(FrameMain_cls *pC){
    if (mpCParentFrame == NULL){
	mCScreenRect = wxRect(wxDefaultPosition, wxDefaultSize);
	mpCParentFrame = pC;
	InitGUI(pC);
    }
}

void MultiPointInput_cls::
UpdateItems(bool bForce){
    mpCGUIBase->mpCMNumMain->UpdateValuesAll(bForce);
    mpCGUIBase->mpCMNumParam->UpdateValuesAll(bForce);
}

int MultiPointInput_cls::
GetSelectedIndex(){
    for (unsigned int i = 0; i < mpCGUIBase->mCItemNum.mBody; i++){
	wxString strTag = mTagStr; strTag << i;
	if (mpCParentFrame->IsPanelActive(strTag) == TRUE) return i;
    }
    return -1; //!!
}

void MultiPointInput_cls::
DrawGLPath(unsigned int dispList, double scale, int index){

    if (index == -1)
	for (unsigned int i = 0; i < mpCGUIBase->mCItemNum.mBody; i++){
	    wxString strTag = mTagStr; strTag << i;
	    bool bSelected = mpCParentFrame->IsPanelActive(strTag);
	    mpCGUIBase->DrawPath(i, bSelected, dispList, scale);
	}
    else {
	wxString strTag = mTagStr; strTag << index;
	bool bSelected = mpCParentFrame->IsPanelActive(strTag);
	mpCGUIBase->DrawPath(index, bSelected, dispList, scale);
    }

    UpdateInterpolatedValues(); //pX`p̌vZCurrentlς̂߂
}

void MultiPointInput_cls::
UpdateActiveOnTL(){
    wxString strLabel;
    strLabel << mTagStr << mpCGUIBase->mActiveItem;
    mpCParentFrame->mpCDF3SeqCtrl->mpCGUI->ActivatePanel(strLabel, TRUE);
    
    mpCGUIBase->ReconnectMNumParam();
    wxCommandEvent dummy(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    mpCGUIBase->GetParent()->AddPendingEvent(dummy);
}

void MultiPointInput_cls::
UpdateActiveFromTag(wxString strTag){
    for (int index = 0; index < mpCGUIBase->mCItemNum.mBody; index++){
	wxString strLabel;
	strLabel << mTagStr << index;
	if (strLabel == strTag){
	    mpCGUIBase->mActiveItem = index;
	    mpCGUIBase->ReconnectMNumParam();
	    UpdateItems();
	}
    }
}

bool MultiPointInput_cls::
UpdateInterpolatedValues(){
    mpCGUIBase->UpdateInterpolatedValues
	(mpCParentFrame->mpCParent->mpCTime->GetDoubleIndex());
    return TRUE;
}
