// -*- c -*-
// "clsDrawingBox.h"
// (setq c-basic-offset 4)

#include "clsDrawingBox.h"

#include <GL/glu.h>

extern int displayListTransCube[4];
extern int displayListWireBox[5];

#include "SRC/COMMON/MyVector.h"
#include "SRC/DF3/clsDF3SeqCtrl.h"
#include "SRC/DF3/clsDF3Sequence.h"
#include "SRC/DRAWBOX/clsBoxCopy.h"
#include "SRC/FILTER/clsFilterCtrl.h"
#include "SRC/GLOBJ/GLOBJTransCube.h"
#include "SRC/MYGL/clsTextureProperty.h"
#include "SRC/POVRAY/clsPovRay.h"
#include "SRC/POVRAY/clsPovVolume.h"
#include "SRC/POVRAY/clsTreeBoxRule.h"

#include <fstream>
#include <sstream>

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

#define ENDL std::endl
#define ENDLWX wxT("\n")

#ifdef WIN32
#define DELIMWX wxT("\\")
#else
#define DELIMWX wxT("/")
#endif

// $B:G=i$N(Bindex$B$O(B8$BJ,LZ0J30$NDI2C$rA[Dj(B
double DrawingBlock_cls::smGeomMin[2][8][3] = {
    {{-0.5, -0.5, -0.5}, {0, -0.5, -0.5}, {-0.5, 0, -0.5}, {0, 0, -0.5},
     {-0.5, -0.5, 0},    {0, -0.5, 0},    {-0.5, 0, 0},    {0, 0, 0} } ,
    {{-0.5, -0.5, -0.5}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
     {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0} } };
double DrawingBlock_cls::smGeomMax[2][8][3] = {
    {{0, 0, 0},   {0.5, 0, 0},   {0, 0.5, 0},   {0.5, 0.5, 0},
     {0, 0, 0.5}, {0.5, 0, 0.5}, {0, 0.5, 0.5}, {0.5, 0.5, 0.5}} ,
    {{0.5, 0.5, 0.5}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
     {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0} } };
double DrawingBlock_cls::smGeomCenter[2][8][3] = {
    {{-0.5,-0.5,-0.5},{+0.5,-0.5,-0.5},{-0.5,+0.5,-0.5},{+0.5,+0.5,-0.5},
     {-0.5,-0.5,+0.5},{+0.5,-0.5,+0.5},{-0.5,+0.5,+0.5},{+0.5,+0.5,+0.5} },
    {{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
     {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0} } };

std::vector < wxString > DrawingBlock_cls::msVSAlreadyCountedTag;

DrawingBlock_cls::
DrawingBlock_cls(){
    mDepth = 0;
    mDepthOffset = 0;
    mbFileExist = FALSE;
    mCb8Tree.SetItem(L"DrawingBoxUse8TRee=", L"0");
    mCb8TreeForCopy.SetItem(L"DrawingBoxUse8TReeForCopy=", L"0");
    mBoxType = _DBOX_8BOX;
    
    mCBoxMin[X].SetItem(L"DrawingBoxMinX=", L"-1");
    mCBoxMin[Y].SetItem(L"DrawingBoxMinY=", L"-1");
    mCBoxMin[Z].SetItem(L"DrawingBoxMinZ=", L"-1");
    mCBoxMax[X].SetItem(L"DrawingBoxMaxX=", L"1");
    mCBoxMax[Y].SetItem(L"DrawingBoxMaxY=", L"1");
    mCBoxMax[Z].SetItem(L"DrawingBoxMaxZ=", L"1");
    LOOP_XYZ {
	mbMirroring[xyz] = FALSE;
	mCopyOffset[xyz] = 0;
    }

    mbOriginalOne = FALSE;
    mbVisibility = FALSE;
    mDistance = 0;
    mTreeDivideParam = 0;
    mpCParentBlock = NULL;
}

DrawingBlock_cls* DrawingBlock_cls::
GetRootBlock(){
    if (mDepth == 0) return this;
    return mpCParentBlock->GetRootBlock();
}

void DrawingBlock_cls::
SetRoot(){
    LOOP_XYZ mBoxMinWorldCoord[xyz] = 
	mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz];
    LOOP_XYZ mBoxMaxWorldCoord[xyz] = 
	mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz+3];
    mbOriginalOne = TRUE;
    TexProperty_cls::SetBoxGeom(mBoxMinWorldCoord, mBoxMaxWorldCoord,
				mspCRotate->GetPInterpolatedValue(msCurrentDoubleIndex));
    
}

void DrawingBlock_cls::
SetShowParam(bool bShowWire, bool bShowCube, bool bSelected){
    mbShowWire = bShowWire;
    mbShowCube = bShowCube;
    mbSelected = bSelected;
}

// ----

#include "SRC/MYAPI/MsgBox.h"

ThreadMakeChilds_cls::
ThreadMakeChilds_cls() : wxThread (wxTHREAD_JOINABLE)
{
}
 
void ThreadMakeChilds_cls::
SetParameters(DrawingBlock_cls *pC,
	      double pos[3], int treePath, CheckTreeFile_cls* pCheckFile, 
	      bool bForceMake){
    mpC = pC;
    LOOP_XYZ mPos[xyz] = pos[xyz];
    mTreePath = treePath;
    mpCheckFile = pCheckFile;
    mbForceMake = bForceMake;

    if ( Create() != wxTHREAD_NO_ERROR ) 
	MsgBoxExit(NULL, "Error!", "Can't create thread!");
}

wxThread::ExitCode ThreadMakeChilds_cls::
Entry(){
    // $B8=:_5!G=$7$F$$$J$$!*(B
    mpC->mVCChildBox.clear();
    mpC->mVCTreePath.push_back(mTreePath);

    if (mpC->mCb8Tree.mbBody == FALSE) 
	return (ExitCode)_DBOX_VALIDBOX;
    if (mpC->mCb8TreeForCopy.mbBody == FALSE && mpC->mbOriginalOne == FALSE) 
	return (ExitCode)_DBOX_VALIDBOX;
    
    /* msFilePath$B$,JQ99$K$J$C$?$N$G2>$K%3%a%s%H%"%&%H(B
    wchar_t buf[16];
    unsigned int depthTMP = mpC->mDepth + mpC->mDepthOffset;
    swprintf(buf, 16, L"%d.df3", depthTMP-1);
    std::wstring::size_type strpos = mpC->msFilePath.find(buf);
    if (strpos != std::wstring::npos){
#ifdef WIN32
	swprintf(buf, 16, L"%d_%02d\\%d.df3", depthTMP, mTreePath, depthTMP);
#else
	swprintf(buf, 16, L"%d_%02d/%d.df3", depthTMP, mTreePath, depthTMP);
#endif
	mpC->msFilePath.replace(strpos, 6, buf);
    }
    
    mpC->mbFileExist = TRUE;
    if (mpCheckFile != NULL){
	mpC->mbFileExist = mpCheckFile->IsFileExist(mpC->msFilePath, mpC->GetTreePathStr());
	if (mpC->mbFileExist == FALSE && 
	    mbForceMake == _CHECK_TREE_STOP_IF_FILE_NOT_EXIST)
	    return _DBOX_INVALIDBOX;
    }
    */
    
    mpC->CalcDistance(mPos);

    float size = (mpC->GetBoxScaleWorldCoord(X) + mpC->GetBoxScaleWorldCoord(Y) + mpC->GetBoxScaleWorldCoord(Z)) / 3.0;
    mpC->mTreeDivideParam = ( mpC->msCritScale * mpC->mHotSpotMod * size) / mpC->mDistance;
    if (mpC->mTreeDivideParam < 1) return (ExitCode)_DBOX_VALIDBOX;
    
    if (mpC->mDepth >= mpC->msDepthMax + (mpC->mHotSpotMod != 1))
	return (ExitCode)_DBOX_VALIDBOX;
    
    DrawingBlock_cls CChildTemplate = /**this*/*mpC;
    ThreadMakeChilds_cls AThread[8];
    for (unsigned int index = 0; index < 8; index++){
	mpC->mVCChildBox.push_back(CChildTemplate);
	mpC->mVCChildBox[index].SetGeometryChild(_TREE_OCTOTREE, index, mpC);
    }

    // $B%"%I%l%9$N0LCV$,(Bvector$B$N:F3NJ]$GJQ$o$k$N$G!"%"%I%l%9$O@_Dj8e$KEO$9(B
    bool childValidation = TRUE;
    for (unsigned int index = 0; index < 4; index++){
	AThread[index].SetParameters
	    (&(mpC->mVCChildBox[index]),
	     mPos, index, mpCheckFile, mbForceMake);
    }
    
    for (unsigned int index = 0; index < 4; index++)
	AThread[index].Run();

    for (unsigned int index = 4; index < 8; index++){
	if (mpC->mVCChildBox[index].MakeChilds
	    (mPos, index, mpCheckFile, mbForceMake) 
	    == _DBOX_INVALIDBOX){
	    mpC->mVCChildBox.clear(); index = 8; continue;
	}
    }
    
    for (unsigned int index = 0; index < 4; index++)
	if (AThread[index].Wait() == _DBOX_INVALIDBOX){
	    childValidation = FALSE;
	}
    
    if (childValidation == FALSE)
	mpC->mVCChildBox.clear();
    
    return (ExitCode)_DBOX_VALIDBOX;
}

// ----

bool DrawingBlock_cls::
MakeChilds(double pos[3], int treePath, 
	   CheckTreeFile_cls* pCheckFile, bool bForceMake){
    mVCChildBox.clear();
    mVCTreePath.push_back(treePath);

    if (mCb8Tree.mbBody == FALSE) 
	return _DBOX_VALIDBOX;
    if (mCb8TreeForCopy.mbBody == FALSE && mbOriginalOne == FALSE) 
	return _DBOX_VALIDBOX;

    wchar_t buf[16];
    unsigned int depthTMP = mDepth + mDepthOffset;
    
    swprintf(buf, 16, L"%d.df3", depthTMP-1);
    wxString strFrom(buf);
    
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_OCTOTREE)
	swprintf(buf, 16, L"%d_%02d%s%d.df3", depthTMP, treePath, DELIMWX, depthTMP);
    else
	swprintf(buf, 16, L"%d.df3", depthTMP);
    
    wxString strTo(buf);

    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mVsFilePath[i].Replace(strFrom, strTo);
    
    mbFileExist = TRUE;
    if (pCheckFile != NULL){
	mbFileExist = pCheckFile->IsFileExist(mVsFilePath/*[_DF3_SLOT_0_FIX_ME].c_str()*/, GetTreePathStr().c_str(), mspCTreeBoxRuleLink->msCStrDataSlot.msBody);
	if (mbFileExist == FALSE && 
	    bForceMake == _CHECK_TREE_STOP_IF_FILE_NOT_EXIST)
	    return _DBOX_INVALIDBOX;
    }

    CalcDistance(pos);
    
    // OCTOTREE$B$N;~$O(BTREE$B$NBG$A@Z$j$r7W;;!#$=$&$G$J$1$l$P:G8e$^$G(B
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_OCTOTREE){
	float size = (GetBoxScaleWorldCoord(X) + GetBoxScaleWorldCoord(Y) + GetBoxScaleWorldCoord(Z)) / 3.0;
	mTreeDivideParam = ( msCritScale * mHotSpotMod * size) / mDistance;
	if (mTreeDivideParam < 1 && mDepth >= msDepthMin) 
	return _DBOX_VALIDBOX;
    }
    
    if (mDepth >= msDepthMax + (mHotSpotMod != 1))
	return _DBOX_VALIDBOX;
    
    DrawingBlock_cls CChildTemplate = *this;
    LOOP_XYZ CChildTemplate.mbMirroring[xyz] = FALSE; //$B;R6!$O%_%i!<$7$J$$(B

    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_OCTOTREE){
	for (unsigned int index = 0; index < 8; index++){
	    mVCChildBox.push_back(CChildTemplate);
	    mVCChildBox[index].SetGeometryChild(_TREE_OCTOTREE, index, this);
	    if (mVCChildBox[index].MakeChilds
		(pos, index, pCheckFile, bForceMake) 
		== _DBOX_INVALIDBOX){
		mVCChildBox.clear(); index = 8; continue;
	    }
	}
    }
    else{
	mVCChildBox.push_back(CChildTemplate);
	mVCChildBox[0].SetGeometryChild(_TREE_NESTED, 0, this);
	if (mVCChildBox[0].MakeChilds
	    (pos, 0, pCheckFile, bForceMake) 
	    == _DBOX_INVALIDBOX){
	    mVCChildBox.clear();
	}
    }
    
    return _DBOX_VALIDBOX;
}

wxString DrawingBlock_cls::
GetTreePathStr(){
    wxString strPath;
    for (unsigned int i = 0; i < mVCTreePath.size(); i++)
	strPath << mVCTreePath[i];
    return strPath;
}

int DrawingBlock_cls::
GetBoxNumber(bool bCheckFileExist){
    int initNum = 0;
    if (mVCChildBox.size() == 0 && mbVisibility == TRUE){
	bool checked = FALSE;
	for (unsigned int i = 0; i < msVSAlreadyCountedTag.size(); i++){
	    if (GetTreePathStr() == msVSAlreadyCountedTag[i]){
		checked = TRUE;
		i = (unsigned int)msVSAlreadyCountedTag.size();
	    }
	}
	if (checked == FALSE){
	    if (bCheckFileExist == FALSE || mbFileExist == TRUE)
		initNum++;
	    msVSAlreadyCountedTag.push_back(GetTreePathStr());
	}
    }
    int childNum = 0;
    for  (unsigned int i = 0; i < mVCChildBox.size(); i++)
	childNum += mVCChildBox[i].GetBoxNumber(bCheckFileExist);
    initNum += childNum;
    if (mVCChildBox.size() != 0 && childNum == 0)
	initNum ++; // $B;R6!$,(BInvalid$B$J$i<+J,$r;H$&(B
    return initNum;
}

void DrawingBlock_cls::
SetFilePath(std::vector < wxString> VstrPath){
    mVsFilePath = VstrPath;
}

void DrawingBlock_cls::
CheckMinAndMax(){
    LOOP_XYZ{
	if (mCBoxMin[xyz].mBody > mCBoxMax[xyz].mBody){
	    double tmp = mCBoxMin[xyz].mBody;
	    mCBoxMin[xyz].mBody = mCBoxMax[xyz].mBody;
	    mCBoxMax[xyz].mBody = tmp;
	}
    }
}

void DrawingBlock_cls::
SetOffset(unsigned int axis, int offset){
    mCopyOffset[axis] = offset;
    double boxScale = GetBoxScaleWorldCoord(axis);
    mBoxMinWorldCoord[axis] += offset * boxScale;
    mBoxMaxWorldCoord[axis] += offset * boxScale;
}

double DrawingBlock_cls::
GetRootBoxScale(unsigned int axis){
    return  
	mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis + 3] -
	mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis];
}

double DrawingBlock_cls::
GetBoxScale(unsigned int axis, bool bAbsolute){
    double scale;
    if (mDepth > 0)
      scale = mCBoxMax[axis].mBody - mCBoxMin[axis].mBody;
    else
	scale = 
	    mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis + 3] -
	    mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis];
    
    if (bAbsolute == _DBOX_ABSOLUTE && scale < 0) scale *= -1;
    return scale;
}

double DrawingBlock_cls::
GetBoxCenter(unsigned int axis){
    if (mDepth > 0)
	return (mCBoxMax[axis].mBody + mCBoxMin[axis].mBody) / 2;
    double baseOffset = 
	(mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis + 3] + 
	 mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis]) / 2;
    double scale = 
	(mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis + 3] - 
	 mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[axis]);
    return baseOffset + msTranslate[axis]*scale;
}

double DrawingBlock_cls::
GetBoxScaleWorldCoord(unsigned int axis, bool bAbsolute){
    double scale = mBoxMaxWorldCoord[axis] - mBoxMinWorldCoord[axis];
    if (bAbsolute == _DBOX_ABSOLUTE && scale < 0) scale *= -1;
    return scale;
}

double DrawingBlock_cls::
GetBoxCenterWorldCoord(unsigned int axis){
    return (mBoxMinWorldCoord[axis] + mBoxMaxWorldCoord[axis]) / 2;
}

void DrawingBlock_cls::
CheckVis(PositionAndDirection_cls CCam, 
	 unsigned int renderType, float radian){

    mbVisibility = TRUE;

    // $B8=>u!"%5%$%:$O(BX$BJ}8~$N$_$G9M$($F$$$k!#(B
    // $B$$$:$l(Bxyz$B$=$l$>$l$GH=CG$9$k(B

    // $BGX8e$+$I$&$+(B
    double rel[3];
    LOOP_XYZ rel[xyz] = GetBoxCenterWorldCoord(xyz) - CCam.mPos[xyz];
    double projectedDist = MyVector::GetScalarProduct(rel, CCam.mTgt);
    if (projectedDist < - GetBoxScaleWorldCoord(X, _DBOX_ABSOLUTE) * 0.867)
	mbVisibility = FALSE;

    // $B;kLn3Q(B 
    if (renderType == _POV_PERSPECTIVE){
	double projectedPos[3];
	LOOP_XYZ projectedPos[xyz] = 
	    CCam.mPos[xyz] + projectedDist * CCam.mTgt[xyz];
	LOOP_XYZ rel[xyz] = GetBoxCenterWorldCoord(xyz) - projectedPos[xyz];
	double sideDist = MyVector::GetNorm(rel);
	double critDist = projectedDist * tan(radian);
	if (sideDist > critDist + GetBoxScaleWorldCoord(X, _DBOX_ABSOLUTE) * 0.867)
	    mbVisibility = FALSE;
    }

    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED)
	if (mDepth < msDepthMin)
	    mbVisibility = FALSE;

    if (radian > 1.57) mbVisibility = TRUE;
    
    for (unsigned i = 0; i < mVCChildBox.size(); i++)
	mVCChildBox[i].CheckVis(CCam, /*anc,*/ renderType, radian);
}

void DrawingBlock_cls::
CalcDistance(double pos[3]){
    double centerWC[3];
    LOOP_XYZ 
	if (mbMirroring[xyz] == TRUE) 
	    centerWC[xyz] = -GetBoxCenterWorldCoord(xyz);
	else
	    centerWC[xyz] = GetBoxCenterWorldCoord(xyz);

    MyVector::Rotate(centerWC, mspCRotate->mpInterpolatedValue[3], Z);
    MyVector::Rotate(centerWC, mspCRotate->mpInterpolatedValue[2], Y);
    MyVector::Rotate(centerWC, mspCRotate->mpInterpolatedValue[1], X);
    MyVector::Rotate(centerWC, mspCRotate->mpInterpolatedValue[0], Z);
    
    mDistance = MyVector::GetNormRelative(pos, centerWC);
    mHotSpotMod = msCHotSpot.CalcModifier(pos, centerWC, GetBoxScaleWorldCoord(X));
}

void DrawingBlock_cls::
ConvertMinAndMaxToChild(double min[3], double max[3], 
			double minRoot[3], double maxRoot[3]){
    if (mDepth == 0){
	LOOP_XYZ min[xyz] = minRoot[xyz];
	LOOP_XYZ max[xyz] = maxRoot[xyz];
	return; 
    }

    double factorA[3], factorB[3];
    LOOP_XYZ factorA[xyz] = 1.0 / (mBoxMaxWorldCoord[xyz] - mBoxMinWorldCoord[xyz]);
    LOOP_XYZ factorB[xyz] = - mBoxMinWorldCoord[xyz] / (mBoxMaxWorldCoord[xyz] - mBoxMinWorldCoord[xyz]);
    
    double posMin[3], posMax[3];
    LOOP_XYZ posMin[xyz] = minRoot[xyz] * (mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz+3] - mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz]) + mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz];
    LOOP_XYZ posMax[xyz] = maxRoot[xyz] * (mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz+3] - mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz]) + mspCBoxSize->GetPInterpolatedValue(msCurrentDoubleIndex)[xyz];

    LOOP_XYZ min[xyz] = factorA[xyz] * posMin[xyz] + factorB[xyz];
    LOOP_XYZ max[xyz] = factorA[xyz] * posMax[xyz] + factorB[xyz];

    LOOP_XYZ if (min[xyz] < 0) min[xyz] = 0;
    LOOP_XYZ if (max[xyz] > 1) max[xyz] = 1;

    //LOOP_XYZ min[xyz] = mCBoxMin[xyz].mBody;
    //LOOP_XYZ max[xyz] = mCBoxMax[xyz].mBody;
}

bool DrawingBlock_cls::
IsNestOnDensity(){
    bool bNestOnDensity = FALSE;
    if (mVCChildBox.size() != 0 &&
	mDepth >= mspCTreeBoxRuleLink->mCDepthMin.mBody)
	if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED &&
	    mDepth == mspCTreeBoxRuleLink->mCDepthMax.mBody - 1)
	    bNestOnDensity = TRUE;
    return bNestOnDensity;
}

void DrawingBlock_cls::
DrawCube(){
    if (mbShowCube == FALSE) return;
    if (mbVisibility == FALSE) return;
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_OCTOTREE &&
	mVCChildBox.size() != 0) return;
    glTranslated(-0.5, -0.5, -0.5);
    if (mCb8Tree.mbBody == FALSE)
	glCallList(displayListTransCube[1]);
    else
	glCallList(displayListTransCube[(mDepth < 4) ? mDepth : 3]);
    glTranslated(+0.5, +0.5, +0.5);
}

void DrawingBlock_cls::
DrawRoudedCubeA(double rad){
    glBegin(GL_LINES);
    glVertex3d(-0.5+rad,-0.5,-0.5+rad); glVertex3d(-0.5+rad,-0.5,+0.5-rad);
    glVertex3d(-0.5+rad,-0.5,-0.5+rad); glVertex3d(+0.5-rad,-0.5,-0.5+rad);
    glVertex3d(-0.5,-0.5+rad,-0.5+rad); glVertex3d(-0.5,-0.5+rad,+0.5-rad);
    glVertex3d(-0.5,-0.5+rad,-0.5+rad); glVertex3d(-0.5,+0.5-rad,-0.5+rad);
    glVertex3d(-0.5+rad,-0.5+rad,-0.5); glVertex3d(-0.5+rad,+0.5-rad,-0.5);
    glVertex3d(-0.5+rad,-0.5+rad,-0.5); glVertex3d(+0.5-rad,-0.5+rad,-0.5);

    glEnd();
}

void DrawingBlock_cls::
DrawWire(){
    if (mbShowWire == FALSE) return;
    if (mbVisibility == FALSE) return;
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_OCTOTREE){
	if (mVCChildBox.size() != 0) return;
    }
        
    //glEnable(GL_COLOR_MATERIAL);
    glDisable(GL_TEXTURE_2D);
    int type = (mbFileExist == TRUE) ? 0:1 + mbSelected * 2;
    switch(type){
    case 0:
	glColor4f(1.f, 1.f, 1.f, 0.4f); break;
    case 1:
	glColor4f(1.f, 0.5f, 0.5f, 0.6f); break;
    case 2:
	glColor4f(1.f, 1.f, 1.f, 1.0f); break;
    case 3:
	glColor4f(1.f, 0.5f, 0.5f, 1.0f); break;
    case 4:
	glColor4f(1.f, 1.f, 0.2, 1.f); break;
    }

    bool bRound = FALSE; // $B0l1~4]$_$r$D$1$k$H$-$O6hJL$7$FI=<((B
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED &&
	mspCTreeBoxRuleLink->mCRoundedCurve.mBody > 0.02 && mDepth >= 1)
	bRound = TRUE;

    if (bRound == FALSE){
	glTranslated(-0.5, -0.5, -0.5);
	glCallList
	    (displayListWireBox[(mbFileExist == TRUE) ? 0:1 + mbSelected * 2]);
	glTranslated(+0.5, +0.5, +0.5);
    }
    else{
	float rad = 0.02;//mspCTreeBoxRuleLink->mCRoundedCurve.mBody;
	glPushMatrix();
	DrawRoudedCubeA(rad);
	glRotatef(180, 0, 0, 1);
	DrawRoudedCubeA(rad);
	glRotatef(180, 1, 0, 0);
	DrawRoudedCubeA(rad);
	glRotatef(180, 0, 0, 1);
	DrawRoudedCubeA(rad);
	glPopMatrix();
    }

}

void DrawingBlock_cls::
DrawClip(bool bRoot){
    if (mbVisibility == FALSE) return;
    
    if (mDepth != 0) return;
    
    glPushMatrix();
    glTranslated(-0.5, -0.5, -0.5);

    if (bRoot == TRUE){
	for (unsigned int i = 0; i < msVClipMin.size(); i++){
	    if (msVActiveClip[i] == TRUE)
		glColor4f(0.4f, 1.0f, 0.4f, 1.0f);
	    else
		glColor4f(0.2f, 0.8f, 0.2f, 0.6f);
	    if (msVVisibleClip[i] == TRUE)
		GLOBJ::DrawWireBoxMinMax(&msVClipMin[i][0], 
					 &msVClipMax[i][0]);
	}
    }
    else {
	double min[3], max[3];
	double minRoot[3], maxRoot[3];
	LOOP_XYZ minRoot[xyz] = msClipMin[xyz];
	LOOP_XYZ maxRoot[xyz] = msClipMax[xyz];
	ConvertMinAndMaxToChild(min, max, minRoot, maxRoot);
	glColor4f(0.2f, 0.8f, 0.2f, 0.6f);
	if (mbVisibility == TRUE)
	    GLOBJ::DrawWireBoxMinMax(min, max);
    }
    
    glPopMatrix();
}

PositionAndDirection_cls DrawingBlock_cls::
GetInvertedPosDir(PositionAndDirection_cls CPos, bool bMirror[3],
		  bool bNormal, bool bWorldCord){
    CPos.RotateAroundOriginYaw(-mspCRotate->mpInterpolatedValue[0]);
    CPos.RotateAroundOriginRoll(-mspCRotate->mpInterpolatedValue[1]);
    CPos.RotateAroundOriginPitch(-mspCRotate->mpInterpolatedValue[2]);
    CPos.RotateAroundOriginYaw(-mspCRotate->mpInterpolatedValue[3]);
    
    double scale[3];
    if (bWorldCord == TRUE)
	LOOP_XYZ scale[xyz] = GetBoxScaleWorldCoord(xyz);
    else
	LOOP_XYZ scale[xyz] = GetBoxScale(xyz);

    if (bNormal == TRUE) // $BLL$NK!@~$K4X$9$k=hM}$O5U$K$J$k(B
	LOOP_XYZ {
	    CPos.mTgt[xyz] *= scale[xyz];
	    CPos.mTop[xyz] *= scale[xyz];
	    CPos.mRgt[xyz] *= scale[xyz];
	    CPos.mPos[xyz] *= scale[xyz];
	}
    else
	LOOP_XYZ {
	    if (scale[xyz] != 0){
		CPos.mTgt[xyz] /= scale[xyz];
		CPos.mTop[xyz] /= scale[xyz];
		CPos.mRgt[xyz] /= scale[xyz];
		CPos.mPos[xyz] /= scale[xyz];
	    }
	}
    LOOP_XYZ
	if (bMirror[xyz] == TRUE){
	    CPos.mTgt[xyz] *= -1;
	    CPos.mTop[xyz] *= -1;
	    CPos.mRgt[xyz] *= -1;
	    CPos.mPos[xyz] *= -1;
	}
    return CPos;
}

void DrawingBlock_cls::
DrawDataBlock(PositionAndDirection_cls CCam, bool bShowData, bool bCubeAndWire,
	      double *pShadeOffset){

    glPushMatrix();

    if (mbMirroring[Z] == TRUE) glScalef( 1,  1, -1);
    if (mbMirroring[Y] == TRUE) glScalef( 1, -1,  1);
    if (mbMirroring[X] == TRUE) glScalef(-1,  1,  1);

    if (mDepth == 0)
	glTranslatef(mCopyOffset[X] * GetBoxScale(X),
		     mCopyOffset[Y] * GetBoxScale(Y),
		     mCopyOffset[Z] * GetBoxScale(Z));
    
    double boxCenter[3]; LOOP_XYZ boxCenter[xyz] = GetBoxCenter(xyz);
    double boxScale[3];  LOOP_XYZ boxScale[xyz] = GetBoxScale(xyz);

    glTranslated(boxCenter[X], boxCenter[Y], boxCenter[Z]);
    glScaled    (boxScale[X],  boxScale[Y],  boxScale[Z]); 

    if (bCubeAndWire){
	glEnable(GL_LINE_SMOOTH);
	DrawCube();
	DrawWire();
	DrawClip(mDepth == 0 ? TRUE : FALSE);
	if (bShowData == TRUE && mDepth == 0) // Isosurf$B$H$+?F0J30$GIA$$$A$cBLL\(B
	    mspCDF3SeqCtrlLink->DrawItems(boxCenter, boxScale);
	glDisable(GL_LINE_SMOOTH);
    }

    PositionAndDirection_cls CShadeOffset;
    LOOP_XYZ CShadeOffset.mPos[xyz] = pShadeOffset[xyz];

    // $B%+%a%i0LCV!"8w8;J}8~$K2sE>$N5U1i;;(B
    CCam = GetInvertedPosDir(CCam, mbMirroring);
    CShadeOffset = GetInvertedPosDir(CShadeOffset, mbMirroring, FALSE);
    
    if (bShowData == TRUE && mDepth == 0)
	LOOP_XYZ
	    mspCDF3SeqCtrlLink
	    ->DrawDataSlot(CCam, GetBoxScale(X), GetBoxScale(Y), GetBoxScale(Z),
			   xyz, mbMirroring, msClipMin, msClipMax,
			   CShadeOffset.mPos);
        
    for (unsigned i = 0; i < mVCChildBox.size(); i++)
	mVCChildBox[i].DrawDataBlock(CCam, FALSE, bCubeAndWire, pShadeOffset);
    
    glPopMatrix();
}

void DrawingBlock_cls::
SetGeometryChild(int treetype, unsigned int index, 
		 DrawingBlock_cls *pCParentBlock){
    mpCParentBlock = pCParentBlock;
    mDepth++ ;
    double minWC[3], maxWC[3];
    LOOP_XYZ minWC[xyz] = mBoxMinWorldCoord[xyz];
    LOOP_XYZ maxWC[xyz] = mBoxMaxWorldCoord[xyz];

    double geomMin[3], geomMax[3];
    LOOP_XYZ{
	geomMin[xyz] = smGeomMin[treetype][index][xyz];
	geomMax[xyz] = smGeomMax[treetype][index][xyz];
    }
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED)
	LOOP_XYZ {
	    geomMin[xyz] *= mspCTreeBoxRuleLink->mCNestScale.mBody;
	    geomMax[xyz] *= mspCTreeBoxRuleLink->mCNestScale.mBody;
	    geomMin[xyz] += mspCTreeBoxRuleLink->mCNestOffset.mpBody[xyz];
	    geomMax[xyz] += mspCTreeBoxRuleLink->mCNestOffset.mpBody[xyz];
	}
    
    LOOP_XYZ{
	mCBoxMin[xyz].mBody = geomMin[xyz];
	mCBoxMax[xyz].mBody = geomMax[xyz];
	mBoxMinWorldCoord[xyz] = 
	    minWC[xyz] + (maxWC[xyz] - minWC[xyz]) * (geomMin[xyz] + 0.5);
	mBoxMaxWorldCoord[xyz] = 
	    maxWC[xyz] - (maxWC[xyz] - minWC[xyz]) * (0.5 - geomMax[xyz]);
    }
}

wxString DrawingBlock_cls::
GetPovStrDataBlock(wxString strTab, unsigned int type, wxString strFilterIndex,
		   bool bAThreeViewShow0[3], bool bAThreeViewShow1[3]){
    wxString strInfo;
    wxString strHeader;
    strHeader << strTab << L"union {" << ENDLWX;

    bool bOmitInfo = FALSE;

    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED ||
	mVCChildBox.size() == 0){
	
	if (mbVisibility == TRUE)
	    switch (type){
	    case _DBOX_VOLUME:
		strInfo << strTab << L"\tobject { BOX_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX;
		break;
	    case _DBOX_ISOSURF:
		strInfo << strTab << L"\tobject { ISOBOX_" << strFilterIndex + GetTreePathStr() << wxT(" }") << ENDLWX;
		break;
	    case _DBOX_CROSS_SECTION:
		strInfo << strTab << L"\tobject { OBJECT_C_SECTION_" << strFilterIndex + GetTreePathStr() << wxT(" }") << ENDLWX;
		break;
	    case _DBOX_THREE_VIEW:
		if (bAThreeViewShow0[X] == FALSE &&
		    bAThreeViewShow0[Y] == FALSE &&
		    bAThreeViewShow0[Z] == FALSE &&
		    bAThreeViewShow1[X] == FALSE &&
		    bAThreeViewShow1[Y] == FALSE &&
		    bAThreeViewShow1[Z] == FALSE) break;
		if (bAThreeViewShow0[X] == TRUE)
		    strInfo << strTab << L"\tobject { OBJECT_THREE_VIEW_X_MINUS_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX;
		if (bAThreeViewShow0[Y] == TRUE)
		    strInfo << strTab << L"\tobject { OBJECT_THREE_VIEW_Y_MINUS_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX;
		if (bAThreeViewShow0[Z] == TRUE)
		    strInfo << strTab << L"\tobject { OBJECT_THREE_VIEW_Z_MINUS_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX;
		if (bAThreeViewShow1[X] == TRUE)
		    strInfo << strTab << L"\tobject { OBJECT_THREE_VIEW_X_PLUS_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX;
		if (bAThreeViewShow1[Y] == TRUE)
		    strInfo << strTab << L"\tobject { OBJECT_THREE_VIEW_Y_PLUS_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX;
		if (bAThreeViewShow1[Z] == TRUE)
		    strInfo << strTab << L"\tobject { OBJECT_THREE_VIEW_Z_PLUS_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX;
		break;
	    case _DBOX_STREAMLINE:
		if (mDepth == 0){
		    strInfo << strTab << L"\tobject { OBJECT_STREAM_LINES_" << strFilterIndex + GetTreePathStr() << ENDLWX
			    << wxT("#ifdef (STREAMLINETEXGLOBAL_") << strFilterIndex + GetTreePathStr() << wxT(")") << ENDLWX
			    << wxT("\t\ttexture { STREAMLINE_TEX_") << strFilterIndex + GetTreePathStr() << wxT(" }") << ENDLWX
			    << wxT("#end") << ENDLWX
			    << wxT(" \t} ") << ENDLWX;
		    strInfo << wxT("#ifdef (OBJECT_STREAM_POINTS_") << strFilterIndex + GetTreePathStr() << wxT(")");
		    strInfo << strTab << L"\tobject { OBJECT_STREAM_POINTS_" << strFilterIndex + GetTreePathStr() << wxT("}") << ENDLWX 
			    << wxT("#end") << ENDLWX;
		}
		break;
	    case _DBOX_VELOCITYFIELD:
		if (mDepth == 0){
		    strInfo << strTab << L"\tobject { OBJECT_VELOCITYFIELD_" << strFilterIndex + GetTreePathStr() << ENDLWX
			    << wxT("\t\ttexture { VELOCITYFIELD_TEX_") << strFilterIndex + GetTreePathStr() << wxT(" }") << ENDLWX
			    << wxT("\t}") << ENDLWX;
		}
		break;
	    }
	// $BH"$,;kLn30$G$bIA$/!"$J$*$+$D?F0J30$@$H!JB?=EIA2h$J$N$G!KIA$+$J$$(B
	if (type == _DBOX_POVOBJ)
	    if (GetTreePathStr() == wxT("0")){
		strInfo << strTab << L"\tPOVOBJ_" << strFilterIndex << ENDLWX;
		bOmitInfo = TRUE;
	}
    }

    if (mVCChildBox.size() != 0 && IsNestOnDensity() == FALSE)
	for (unsigned int i = 0; i < mVCChildBox.size(); i++)
	    strInfo << mVCChildBox[i].GetPovStrDataBlock(strTab + L"\t", type, strFilterIndex);
    
    if (strInfo == wxT("")) return wxT("");
    
    if (bOmitInfo == TRUE){
	strInfo << wxT("}") << ENDLWX;
	return strHeader + strInfo;
    }

    if (mbMirroring[X] == TRUE) 
	strInfo << strTab << L"\tscale <-1, 1, 1>" << ENDLWX; 
    if (mbMirroring[Y] == TRUE) 
	strInfo << strTab << L"\tscale <1, -1, 1>" << ENDLWX;
    if (mbMirroring[Z] == TRUE)
	strInfo << strTab << L"\tscale <1, 1, -1>" << ENDLWX;
    
    strInfo << strTab << L"\tscale <" <<
	GetBoxScale(X) << L"," <<
	GetBoxScale(Y) << L"," <<
	GetBoxScale(Z) << L">" << ENDLWX;
    
    double center[3];
    LOOP_XYZ
	center[xyz] = GetBoxCenter(xyz) * ((mbMirroring[xyz] == TRUE) ? -1 : 1);
    if (center[X] != 0 || center[Y] != 0 || center[Z] != 0)
	strInfo << strTab << L"\ttranslate <" <<
	    center[X] << L"," <<
	    center[Y] << L"," <<
	    center[Z] << L">" << ENDLWX;
    
    if (mDepth == 0){
	float trans[3];
	LOOP_XYZ
	    trans[xyz] = mCopyOffset[xyz] * GetBoxScale(xyz);
	if (trans[X] != 0 || trans[Y] != 0 || trans[Z] != 0)
	    strInfo << strTab << PovOut3fS(L"\ttranslate", trans) << ENDLWX;
	
	strInfo << PovOutRotation(wxT("\t"), mspCRotate->mpInterpolatedValue);
    }

    if (type == _DBOX_ISOSURF)
	strInfo << wxT("#ifdef (ISOTEXGLOBAL_") << strFilterIndex + GetTreePathStr() << wxT(")") << ENDLWX
		<< wxT("\ttexture { ISOTEX_") << strFilterIndex + GetTreePathStr() << wxT(" }") << ENDLWX
		<< wxT("#end") << ENDLWX;
    if (type == _DBOX_CROSS_SECTION)
	strInfo << wxT("#ifdef (CROSSTEXGLOBAL_") << strFilterIndex + GetTreePathStr() << wxT(")") << ENDLWX
		<< wxT("\ttexture { C_SECTION_TEX_") << strFilterIndex + GetTreePathStr() << wxT(" }") << ENDLWX
		<< wxT("#end") << ENDLWX;

    if (type == _DBOX_THREE_VIEW)
	strInfo << wxT("#ifdef (THREEVIEWGLOBAL_") << strFilterIndex + GetTreePathStr() << wxT(")") << ENDLWX
		<< wxT("\ttexture { THREE_VIEW_TEX_") << strFilterIndex + GetTreePathStr() << wxT(" }") << ENDLWX
		<< wxT("#end") << ENDLWX;

    if (type == _DBOX_VOLUME && mDepth == 0)
	strInfo << wxT("\ttexture { BOX_TEX }") << ENDLWX;
    
    strInfo << strTab << L"}" << ENDLWX;

    return strHeader + strInfo;
}

wxString DrawingBlock_cls::
GetPovStrInverseCoordinate(){
    wxString strInfo;
    wxString strTab = wxT("\t\t\t");

    if (mspCRotate->mpInterpolatedValue[0] != 0)
	strInfo << strTab << PovOut3(L"rotate", 0, 0, -mspCRotate->mpInterpolatedValue[0]) << ENDLWX; 
    if (mspCRotate->mpInterpolatedValue[1] != 0)
	strInfo << strTab << PovOut3(L"rotate", -mspCRotate->mpInterpolatedValue[1], 0, 0) << ENDLWX; 
    if (mspCRotate->mpInterpolatedValue[2] != 0)
	strInfo << strTab << PovOut3(L"rotate", 0, -mspCRotate->mpInterpolatedValue[2], 0) << ENDLWX; 
    if (mspCRotate->mpInterpolatedValue[3] != 0)
	strInfo << strTab << PovOut3(L"rotate", 0, 0, -mspCRotate->mpInterpolatedValue[3]) << ENDLWX;

    double center[3]; 
    LOOP_XYZ center[xyz] =
	GetRootBlock()->GetBoxCenter(xyz) *
	((mbMirroring[xyz] == TRUE) ? -1 : 1);
    if (center[X] != 0 || center[Y] != 0 || center[Z] != 0)
	strInfo << strTab << L"translate <" <<
	    -center[X] << L"," <<
	    -center[Y] << L"," <<
	    -center[Z] << L">" << ENDLWX;
    
    strInfo << strTab << L"scale <" <<
	1.0/GetRootBoxScale(X) << L"," <<
	1.0/GetRootBoxScale(Y) << L"," <<
	1.0/GetRootBoxScale(Z) << L">" << ENDLWX;

    // FIX_ME!! OCTOTREE$B$N>l9gDI2C(B
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED)
	for (int depth = 0 ; depth < mDepth; depth++){
	    wxString strTranslate;
	    if (mspCTreeBoxRuleLink->mCNestOffset.mpBody[X] != 0 ||
		mspCTreeBoxRuleLink->mCNestOffset.mpBody[Y] != 0 ||
		mspCTreeBoxRuleLink->mCNestOffset.mpBody[Z] != 0)
		strTranslate << PovOut3f(L"translate -1 *", mspCTreeBoxRuleLink->mCNestOffset.mpBody);
	    
	    strInfo << strTab << strTranslate << wxT("scale ") << GetShortString(1.0/mspCTreeBoxRuleLink->mCNestScale.mBody) << ENDLWX;
	}
    
    if (mbMirroring[Z] == TRUE)
	strInfo << strTab << L"\tscale <1, 1, -1>" << ENDLWX;
    if (mbMirroring[Y] == TRUE) 
	strInfo << strTab << L"\tscale <1, -1, 1>" << ENDLWX;
    if (mbMirroring[X] == TRUE) 
	strInfo << strTab << L"\tscale <-1, 1, 1>" << ENDLWX; 

    // NestedTree$B$N>l9g!"F1$8>l=j$K=E$J$k$H$^$:$$(B
    /*
    strInfo << strTab << L"\t translate <" << 0.000001 * mDepth << L"," << 0.000001 * mDepth << L"," << 0.000001 * mDepth << L">" << ENDLWX;
    */
    
    return strInfo;
}

wxString DrawingBlock_cls::
GetPovStrDifference(){
    bool bDifferenceCond = FALSE;
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED){
	bDifferenceCond = TRUE;
	if (mDepth == mspCTreeBoxRuleLink->mCDepthMax.mBody - 1)
	    bDifferenceCond = FALSE;
    }
    
    if (bDifferenceCond == FALSE) return wxT("");
    if (mVCChildBox.size() == 0) return wxT("");

    float minPos[3], maxPos[3];
    mspCTreeBoxRuleLink->GetChildNestPos(minPos, maxPos);

    LOOP_XYZ { // CHECK_ME!!
	minPos[xyz] -= 0.0002f;
	maxPos[xyz] += 0.0002f;
    }
    
    wxString strDifference;
    if (mspCTreeBoxRuleLink->mCRoundedCurve.mBody != 0)
	strDifference << wxT("\t\tRound_Box_Merge ( ") 
		      << PovOut3f(wxT(""), minPos) << PovOut3f(wxT(","), maxPos)
		      << wxT(", ") << mspCTreeBoxRuleLink->mCRoundedCurve.mBody * mspCTreeBoxRuleLink->mCNestScale.mBody << wxT(" )") << ENDLWX;
    else
	strDifference << wxT("\t\tbox { ") 
		      << PovOut3f(wxT(""), minPos) << PovOut3f(wxT(","), maxPos)
		      << wxT(" }") << ENDLWX;
    return strDifference;
}

wxString DrawingBlock_cls::
GetPovBoxTex(){
    wxString strTex;
    strTex << wxT("#declare BOX_TEX_FINISH=finish{") << ENDLWX;
    strTex << mspCTexture->GetPovStrFinish(TRUE)
	   << wxT("}") << ENDLWX;
    
    strTex << wxT("#declare BOX_TEX=texture{") << ENDLWX;
    float trans = 1.0f - msTexAlpha;
    int transmitType = 1;
    wxString strClrMap, strClr1, strClr2, strClr3;
    if (mspCTexture->mCColorType.mBody == 0)
	strTex << mspCTexture->GetPovStrTex(msTexAlpha, wxT("BOX_TEX_FINISH"), wxT(""), 1, 1, 1, TRUE, TRUE);
    else{
	std::vector < wxString > Vstr = 
		    mspCTexture->GetVPovStrColors(_DF3_SLOT_0_FIX_ME, trans);
	strClrMap = Vstr[0];
	strClr1 = Vstr[1];
	strClr2 = Vstr[2];
	strClr3 = Vstr[3];
	
	double globalBoxMin[3], globalBoxMax[3];
	LOOP_XYZ globalBoxMin[xyz] = TexProperty_cls::msBoxMinWorld[xyz];
	LOOP_XYZ globalBoxMax[xyz] = TexProperty_cls::msBoxMaxWorld[xyz];
	double globalBoxCenter[3], globalBoxSize[3];
	LOOP_XYZ globalBoxCenter[xyz] = (globalBoxMax[xyz] + globalBoxMin[xyz]) / 2;
	LOOP_XYZ globalBoxSize[xyz] = (globalBoxMax[xyz] - globalBoxMin[xyz]);
	
	wxString strTransDF3;
	strTransDF3 << wxT("\ttranslate <-0.5, -0.5, -0.5>") << ENDLWX
		    << PovOut3d(wxT("\tscale"), globalBoxSize) << ENDLWX
		    << PovOut3d(wxT("\ttranslate"), globalBoxCenter) << ENDLWX
		    << PovOutRotation(wxT("\t"), TexProperty_cls::msBoxRotParam);
	mspCTexture->SetDensityFile(mspCDF3SeqCtrlLink->GetPDF3Active()->GetFilePathWXStr()/*wxT("strFilePath")*/);
	mspCTexture->SetTranslationForDensityFile(strTransDF3);
	strTex << mspCTexture->GetPovStrTex
	    (trans, 
	     wxT("BOX_TEX_FINISH"), strClrMap, 1, 1, 1, TRUE, TRUE, 0, 
	     strClr1, strClr2, strClr3);
    }
    strTex << wxT("}") << ENDLWX;
    
    return strTex;
}

void DrawingBlock_cls::
OutputPovDataBlockDensity(std::wfstream &ofstream,
			  PovRay_cls *pCPov, 
			  PositionAndDirection_cls CCamGlobalCoord,
			  int frame){
   
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED ||
	mVCChildBox.size() != 0){
	for (unsigned int i = 0; i < mVCChildBox.size(); i++)
	    mVCChildBox[i].OutputPovDataBlockDensity(ofstream, pCPov, msCCamGlobalCoord/*CCamGlobalCoord*/, frame);
	if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_OCTOTREE)
	    return ;
    }

    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED ||
	mVCChildBox.size() == 0){
	PositionAndDirection_cls CCamLocalCoord; 
	CCamLocalCoord = GetInvertedPosDir(msCCamGlobalCoord/*CCamGlobalCoord*/, mbMirroring);
	if (mbVisibility == TRUE){
	    // msFilePath $B$O$3$N;~E@$G(B TreeRule$B$K$h$k(BConverted$B$5$l$?L>A0(B
	    //ofstream << wxConvFile.cWX2WC
	    mspCDF3SeqCtrlLink->GetPovStrDensityBox
		(ofstream, GetTreePathStr(), mVsFilePath,
		 mspCDF3SeqCtrlLink->GetPDF3Seq(_DF3_SLOT_0_FIX_ME)->SerializeLayerSize(PovVolume_cls::msCInterpolate.mBody),
		 this, CCamLocalCoord, frame);
	}
	
	// $B?F$N;~$O(BPovObject$B$J$I$NCf$G(BGlobal$B$K$"$k$b$N$rI,$:IA$/(B
	if (GetTreePathStr() == wxT("0")){
	    /*ssInfo <<*/
	    ofstream << wxConvFile.cWX2WC
		(mspCDF3SeqCtrlLink->
		 GetPovStrGlobalObject(GetTreePathStr(), msCCamGlobalCoord/*CCamGlobalCoord*/,
				       mVsFilePath));
	}
    }

    //ssInfo << ENDLWX;
    ofstream << L"\n";
    //return ssInfo;
}

void DrawingBlock_cls::
ReadStrLine(std::wstring strLine){
    mCb8Tree.ReadStrLine(strLine);
    mCb8TreeForCopy.ReadStrLine(strLine);
    LOOP_XYZ{
	mCBoxMin[xyz].ReadStrLine(strLine);
	mCBoxMax[xyz].ReadStrLine(strLine);
    }
}

void DrawingBlock_cls::
ReadStrLineTex(std::wistream &istream, std::wstring strLine){
    std::wstring::size_type loc = strLine.find(L"--TextureForVolumeBox");
    if (loc != std::wstring::npos){	
	std::wstring wstrLine;
	while (TRUE){							
	    std::getline(istream, wstrLine);				
	    mspCTexture->ReadStrLine(wstrLine);
	    if (wstrLine.find(L"--END") != std::wstring::npos) break;	
	}
    }
}

wxString DrawingBlock_cls::
SerializeConf(){
    wxString str = mCb8Tree.Serialize() + mCb8TreeForCopy.Serialize();
    LOOP_XYZ {
	//str.append(mCBoxMin[xyz].Serialize());
	str << mCBoxMin[xyz].Serialize();
	//str.append(mCBoxMax[xyz].Serialize());
	str << mCBoxMax[xyz].Serialize();
    }

    str << L"--TextureForVolumeBox" << ENDLWX;
    str << mspCTexture->Serialize();
    str << L"--END" << ENDLWX;

    return str;
}

// ---------

int DrawingBlock_cls::
GetDepthFromName(wxString strFileName){
    int depth = 0;
    for (int i = 0; i < 16; i++){
	wchar_t buf[256];
	swprintf(buf, 256, L"_%d.df3", i);
	int find = strFileName.Find(buf);
	if (find != wxNOT_FOUND) {
	    depth = i; break;
	}
    }
    return depth;
}

void DrawingBlock_cls::
GainDrawingBlockCopies(std::vector < DrawingBlock_cls > &VCDrawingCls,
		       DrawingBlock_cls* pCThis, 
		       std::vector < wxString > VstrFilePath,
		       CheckTreeFile_cls* pCheckFile,
		       bool bForceMake){
    pCThis->mVCTreePath.clear();
    pCThis->SetFilePath(VstrFilePath);
    VCDrawingCls.push_back(*pCThis);

    int rootDepthOffset = GetDepthFromName(VstrFilePath[_DF3_SLOT_0_FIX_ME]);
    //mspCTreeBoxRuleLink->mCDepthOffset.mBody = rootDepthOffset;
    pCThis->mDepthOffset = rootDepthOffset;
    
    bool bMirr[3] = {FALSE, FALSE, FALSE};
    bool bCopy[3] = {FALSE, FALSE, FALSE};
    int copyFrom[3] = {0, 0, 0};
    int copyTo[3]   = {0, 0, 0};
    if (mspCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_BOXCOPY) == TRUE){
	BoxCopyData_cls* pCData =
	    (BoxCopyData_cls*)mspCDF3SeqCtrlLink->GetFirstPFilterDataInAllDF3(_FILTER_BOXCOPY);
	if (pCData != NULL && pCData->mCbValid.mbBody == TRUE){
	    LOOP_XYZ{
		bMirr[xyz] = pCData->mCbMirroring.mpbBody[xyz];
		bCopy[xyz] = pCData->mCbCopying.mpbBody[xyz];
		copyFrom[xyz] = pCData->mCCopyFrom.mpBody[xyz];
		copyTo  [xyz] = pCData->mCCopyTo.mpBody[xyz];
	    }
	}
    }

    LOOP_XYZ
	if (bMirr[xyz] == TRUE){
	    // $B%_%i!<(Bc$B$rG[CV(B($B:GBg(B8$B%V%m%C%/(B)$B$K$9$k(B
	    unsigned int size = (unsigned int)VCDrawingCls.size();
	    for (unsigned int i = 0; i < size; i++){
		VCDrawingCls.push_back(VCDrawingCls[i]);
		int current = (int)VCDrawingCls.size() - 1;
		VCDrawingCls[current].mbOriginalOne = FALSE;
		VCDrawingCls[current].mbMirroring[xyz] = TRUE;
	    }
	}
    
    LOOP_XYZ
	if (bCopy[xyz] == TRUE){
	    // $B%3%T!<$rG[CV(B
	    unsigned int size = (unsigned int)VCDrawingCls.size();
	    for (unsigned int i = 0; i < size; i++){
		for (int j = copyFrom[xyz]; j <= copyTo[xyz]; j++){
		    if (j == 0) continue;
		    VCDrawingCls.push_back(VCDrawingCls[i]);
		    int current = (int)VCDrawingCls.size() - 1;
		    //VCDrawingCls[current].mCopyOffset[xyz] = j;
		    VCDrawingCls[current].mbOriginalOne = FALSE;
		    VCDrawingCls[current].SetOffset(xyz, j);
		}
	    }
	}
    
    // $B;R6!G[CV(B
    unsigned int size = (unsigned int)VCDrawingCls.size();
    for (unsigned int i = 0; i < size; i++){
	if (pCheckFile != NULL)
	    VCDrawingCls[i].mbFileExist = 
		pCheckFile->IsFileExist(VCDrawingCls[i].mVsFilePath/*[_DF3_SLOT_0_FIX_ME].c_str()*/, L"0", mspCTreeBoxRuleLink->msCStrDataSlot.msBody);
	
	if (msThread == TRUE){
	    ThreadMakeChilds_cls thread;
	    thread.SetParameters
		(&(VCDrawingCls[i]),
		 msCurrentFocus, 0, pCheckFile, bForceMake);
	    thread.Run();
	    thread.Wait();
	}
	else{
	    VCDrawingCls[i].MakeChilds
		(msCurrentFocus, 0, pCheckFile, bForceMake);
	}
    }
    return;
}

// ---------

bool operator>(const DrawingBlock_cls& CLeft, 
	       const DrawingBlock_cls& CRight){
    return CLeft.mDistance > CRight.mDistance;
}

bool DrawingBlock_cls::msParamForDebugging = FALSE;

// ---------

void DrawingBlock_cls::
InitializeStaticMembers(FrameMain_cls *pCFrame, DF3SeqCtrl_cls *pCDF3SeqCtrl,
			FilterCtrl_cls* pCFilterCtrl,
			TreeBoxRule_cls* pCTreeBoxRule){
    mspCDF3SeqCtrlLink = pCDF3SeqCtrl;
    mspCFilterCtrlLink = pCFilterCtrl;
    mspCTreeBoxRuleLink = pCTreeBoxRule;
    msCHotSpot.SetParent(pCFrame);
    msTexAlpha = 0;
    mspCTexture = new TexProperty_cls;
    mspCTexture->LinkAlphaParameter(&msTexAlpha);
    mspCTexture->mCTransmitType.mBody = _TRANSMIT_TRANSMIT;
    mspCTexture->mCSpecular.mBody = 0;
    
    double boxsize[6] = {-1, -1, -1, 1, 1, 1};
    mspCBoxSize->ExecCommandInitSet(boxsize);
    double rotate[4] = {0, 0 ,0, 0};
    mspCRotate->ExecCommandInitSet(rotate);
}

void DrawingBlock_cls::
SetDefaultClipping(){
    LOOP_XYZ msClipMin[xyz] = 0;
    LOOP_XYZ msClipMax[xyz] = 1;
    msVClipMin.clear(); msVClipMax.clear(); 
    msVActiveClip.clear(); msVVisibleClip.clear();
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mspCDF3SeqCtrlLink->GetPDF3Seq(i)->mpCFilterCtrl->GetClippingInformation(_("Volume"), msClipMin, msClipMax, &msVClipMin, &msVClipMax, &msVActiveClip, &msVVisibleClip, (i==0));
}

Parameters_cls* DrawingBlock_cls::mspCBoxSize;
Parameters_cls* DrawingBlock_cls::mspCRotate;
double DrawingBlock_cls::msTranslate[3] = {0, 0, 0};
double DrawingBlock_cls::msClipMin[3];
double DrawingBlock_cls::msClipMax[3];
double DrawingBlock_cls::msCurrentDoubleIndex = 0;
PositionAndDirection_cls DrawingBlock_cls::msCCamGlobalCoord;
std::vector < std::vector < double > > DrawingBlock_cls::msVClipMin;
std::vector < std::vector < double > > DrawingBlock_cls::msVClipMax;
std::vector < bool > DrawingBlock_cls::msVActiveClip;
std::vector < bool > DrawingBlock_cls::msVVisibleClip;

void DrawingBlock_cls::
SetTreeParam(double currentFocus[3], 
	     unsigned int depthMax, unsigned int depthMin,
	     double critScale){
    LOOP_XYZ msCurrentFocus[xyz] = currentFocus[xyz];
    msDepthMax = depthMax;
    msDepthMin = depthMin;
    msCritScale = critScale;
}

double       DrawingBlock_cls::msCurrentFocus[3];
unsigned int DrawingBlock_cls::msDepthMax;
unsigned int DrawingBlock_cls::msDepthMin;
double       DrawingBlock_cls::msCritScale;
bool         DrawingBlock_cls::msThread = FALSE;
DF3SeqCtrl_cls* DrawingBlock_cls::mspCDF3SeqCtrlLink;
FilterCtrl_cls* DrawingBlock_cls::mspCFilterCtrlLink;
TreeBoxRule_cls* DrawingBlock_cls::mspCTreeBoxRuleLink;

HotSpots_cls DrawingBlock_cls::msCHotSpot;
TexProperty_cls* DrawingBlock_cls::mspCTexture;
double DrawingBlock_cls::msTexAlpha;
