// -*- c -*-
// "clsDrawingBox.cpp"

#ifndef _DRAWING_BOX_H_
#define _DRAWING_BOX_H_

#include "SRC/CONFITEM/clsConfItem.h"
#include "SRC/DRAWBOX/clsHotSpot.h"
#include "SRC/MYGL/clsPositionAndDirection.h"
#include "SRC/POVRAY/clsCheckTreeFile.h"
#include "SRC/POVRAY/clsPovSerialize.h"

class DF3SeqCtrl_cls; 
class PovRay_cls; 

#include "wx/thread.h"

#define _DBOX_NOBOX 0
#define _DBOX_4BOX 1
#define _DBOX_8BOX 2

#define _DBOX_INVALIDBOX 0
#define _DBOX_VALIDBOX 1

#define _DBOX_RELATIVE 0
#define _DBOX_ABSOLUTE 1

#define _DBOX_VOLUME 0
#define _DBOX_ISOSURF 1
#define _DBOX_CROSS_SECTION 2
#define _DBOX_POVOBJ 3
#define _DBOX_THREE_VIEW 4
#define _DBOX_STREAMLINE 5
#define _DBOX_VELOCITYFIELD 6

class DrawingBlock_cls;
class DF3SeqCtrl_cls;
class FilterCtrl_cls;
class HotSpots_cls; 
class TexProperty_cls;
class TreeBoxRule_cls;

class ThreadMakeChilds_cls : public wxThread {
 public:
    ThreadMakeChilds_cls();
    void SetParameters(DrawingBlock_cls *pC,
		       double pos[3], int treePath, 
		       CheckTreeFile_cls* pCheckFile, 
		       bool bForceMake);

    virtual ExitCode Entry();

    DrawingBlock_cls *mpC;
    double mPos[3];
    int    mTreePath;
    CheckTreeFile_cls *mpCheckFile;
    bool   mbForceMake;

};

class DrawingBlock_cls : public PovSerialize_cls {
 public:
    DrawingBlock_cls();
    DrawingBlock_cls *mpCParentBlock;
    DrawingBlock_cls* GetRootBlock();

    unsigned int mDepth, mDepthOffset;
    void     SetRoot();
    void     SetShowParam
	(bool bShowWire, bool bShowCube, bool bSelected);
    bool     MakeChilds
	(double pos[3], int treePath, 
	 CheckTreeFile_cls *pCheckFile, bool bForceMake);
    std::vector < DrawingBlock_cls > mVCChildBox;
    std::vector < unsigned int >     mVCTreePath;
    wxString GetTreePathStr();
    static std::vector < wxString > msVSAlreadyCountedTag;
    int GetBoxNumber(bool bCheckFileExist);
    
    bool        mbFileExist;
    std::vector < wxString > mVsFilePath;
    void SetFilePath(std::vector < wxString> VstrPath);

    bool mbShowWire, mbShowCube, mbSelected;
    ConfItemBOOL_cls mCb8Tree;
    ConfItemBOOL_cls mCb8TreeForCopy;

    void CheckMinAndMax();
    ConfItemDOUBLE_cls mCBoxMin[3]; // c[Ȃǎg
    ConfItemDOUBLE_cls mCBoxMax[3];
    
    static DF3SeqCtrl_cls* mspCDF3SeqCtrlLink;
    static FilterCtrl_cls* mspCFilterCtrlLink;
    static TreeBoxRule_cls* mspCTreeBoxRuleLink;

    static PositionAndDirection_cls msCCamGlobalCoord;
    static double msCurrentDoubleIndex;
    static Parameters_cls* mspCBoxSize; // e̕TCY
    static Parameters_cls* mspCRotate;
    static double msTranslate[3];
    static double msClipMin[3]; // |킹̍ŏIIȃNbv
    static double msClipMax[3];
    static std::vector < std::vector < double > > msVClipMin;
    static std::vector < std::vector < double > > msVClipMax;
    static std::vector < bool > msVActiveClip;
    static std::vector < bool > msVVisibleClip;
    static TexProperty_cls* mspCTexture;
    static double msTexAlpha;

    //----
    static void InitializeStaticMembers
	(FrameMain_cls *pCFrame, DF3SeqCtrl_cls *pCDF3SeqCtrl,
	 FilterCtrl_cls *pCFilterCtrl, TreeBoxRule_cls *pCTreeBoxRule);
    static void SetDefaultClipping();
    
    static void SetTreeParam(double currentFocus[3], 
			     unsigned int depthMax, unsigned int depthMin, 
			     double critScale);
    static double msCurrentFocus[3];
    static unsigned int msDepthMax;
    static unsigned int msDepthMin;
    static double msCritScale;
    static bool   msThread;

    // FrameMainł̕ωɕω̂ŁAConf͕KvȂ
    bool mbMirroring[3]; // 
    int  mCopyOffset[3]; // +- index ɕsړ
    void SetOffset(unsigned int axis, int offset);

    // ----
    
    static double GetRootBoxScale(unsigned int axis);
    double GetBoxScale(unsigned int axis, bool bAbsolute = _DBOX_RELATIVE);
    double GetBoxCenter(unsigned int axis);

    bool   mbOriginalOne;
    bool   mbVisibility;
    double mHotSpotMod;
    double mDistance;
    double GetBoxScaleWorldCoord(unsigned int axis, bool bAbsolute = _DBOX_RELATIVE);
    double GetBoxCenterWorldCoord(unsigned int axis);
    double mBoxMinWorldCoord[3];
    double mBoxMaxWorldCoord[3];

    void   CheckVis(PositionAndDirection_cls CCamPos,
		    unsigned int renderType, float radian);
    void   CalcDistance(double pos[3]);
    void   ConvertMinAndMaxToChild(double min[3], double max[3], 
				   double minRoot[3], double maxRoot[3]);
    bool   IsNestOnDensity();

    static HotSpots_cls msCHotSpot;

    unsigned     int mBoxType;
    float  mTreeDivideParam;
    void DrawCube();
    void DrawRoudedCubeA(double rad);
    void DrawWire();
    void DrawClip(bool bRoot);
    PositionAndDirection_cls 
	GetInvertedPosDir(PositionAndDirection_cls CPos, bool bMirror[3], 
			  bool bNormal = TRUE, bool bWorldCord = FALSE);
    void DrawDataBlock(PositionAndDirection_cls CCam,
		       bool bShowData, bool bCubeAndWire, 
		       double *pShadeOffset);
    
    void SetGeometryChild(int treetype, unsigned int index, 
			  DrawingBlock_cls *pCParentBlock);
    static double smGeomMin[2][8][3];
    static double smGeomMax[2][8][3];
    static double smGeomCenter[2][8][3];

    void ReadStrLine(std::wstring strLine);
    static void ReadStrLineTex(std::wistream &istream, std::wstring strLine);
    wxString SerializeConf();
    wxString GetPovStrDataBlock
	(wxString strTab, unsigned int type, wxString strFilterIndex = wxT(""),
	 bool bAThreeViewShow0[3] = NULL, bool bAThreeViewShow1[3] = NULL);
    wxString GetPovStrDifference();
    wxString GetPovBoxTex();
    void OutputPovDataBlockDensity
	(std::wfstream &ofstream,
	 PovRay_cls *pCPov, PositionAndDirection_cls CCamGlobalCoord,
	 int frame);
    wxString GetPovStrInverseCoordinate();
    
    static bool msParamForDebugging;

    static int  GetDepthFromName(wxString strFileName);
    static void GainDrawingBlockCopies
	(std::vector < DrawingBlock_cls > &VCDrawingCls,
	 DrawingBlock_cls* pCThis, std::vector < wxString > VstrPath,
	 CheckTreeFile_cls* pCheckFile=NULL,
	 bool bForceMake = _CHECK_TREE_MAKE_EVEN_IF_FILE_NOT_EXIST);
};

// rpIy[^ ()
bool operator>(const DrawingBlock_cls& CLeft, 
	       const DrawingBlock_cls& CRight);

#endif
