// -*- c++ -*- 
// "clsFrameMain.cpp"

#ifndef _FRAME_MAIN_H_
#define _FRAME_MAIN_H_

#include <wx/wx.h>
#include <wx/splitter.h>
#include <wx/stopwatch.h>

class CameraPathCtrl_cls;
class CameraPathGLGUI_cls;
class ControlPanel_cls;
class DF3SeqCtrl_cls;
class DF3Sequence_cls;
class FrameMain_cls; 
class FilterCtrl_cls;
class FilterPanelPain_cls;
class GLCanvasCtrl_cls;
class GLExtraFrame_cls;
class GLInfo_cls;
class Main_cls;
class MainGLCanvas_cls;
class PovRay_cls;
class PreviewPanel_cls;
class StereoGram_cls;
class Scene_cls;

class wxButtonEX;
class wxCheckBoxEX; 

#include "clsDrawingBox.h"
#include "SRC/COMMON/clsMultiThreadCtrl.h"
#include "SRC/WIDGETCTRL/clsMultiNumericalInputs.h"
#include "SRC/WIDGETCTRL/clsEditString.h"

#include "SRC/MAINCANVAS/clsFilePathEditor.h"

#include "SRC/COMMON/CommonDefinitions.h"

#define _FM_DEFAULT_CONF false
#define _FM_PROJECT_CONF true

class FrameMain_cls : public wxFrame, public MultiThreadCtrl_cls {
 public:
    FrameMain_cls(const wxString& title, Main_cls *pMain, int posx, int posy,
		  int width, int height);
    ~FrameMain_cls();
    
    static FrameMain_cls *mspCThis;
    Main_cls *mpCParent;
    bool mbForceFocus;
    bool mbOpenProject;
    wxString mstrErr;

    // event handlers (these functions should _not_ be virtual)
    void NormalizeActiveCameraPath();
    void UpdateCameraPos();
    void ForceNewDrawGLPanel(int index = -1);
    void PaintGLPanel(int index = -1, bool bForce = TRUE);
        
    int  mUpdateLevel;
    void UpdateDrawingData();
    bool IsIntersectCanvas(wxRect rect);
    void RefreshPanels();
    void RefreshCameraNumerical();
    void RefreshCameraCombobox();
    void FlexGridSizerAdjust(unsigned int type);
    void SplitterWindowResize();
    
    // ----------

    bool mbViaIdle;
    void ProcPlayTex();
    void OnIdle(wxIdleEvent& event);
    
    // -----------
    
    wxString    GetCurrentActiveFileWXStr();
    wxString    GetFilePathWXStr(int index);
    wxString    GetFilePathNextWXStr(int index);
    
    // -----------
    
    bool   IsPanelActive(wxString strTag);
    
    static GLCanvasCtrl_cls*  mpCCanvas[4];
    DF3SeqCtrl_cls*           mpCDF3SeqCtrl;
    GLInfo_cls*               mpCGLInfo;

    CameraPathCtrl_cls*       mpCPathCtrl;
    PovRay_cls*               mpCPovRay;
    StereoGram_cls*           mpCStereo;
    
    wxSplitterWindow*        mpCSplitterMain;
    PreviewPanel_cls*        mpCPanelTop;
    FilterPanelPain_cls*     mpCFilterPain;
    ControlPanel_cls*        mpCControlPanel;
    GLExtraFrame_cls*        mpCGLExtFrame;
    
    wxStopWatch              mCSWMain;
    void   FitSplitterWindow(bool bForce = FALSE);

    void   AddBitmapMenu(wxMenu *pMenu, int ID, wxString strMain, wxString strSub, 
			 wxBitmap *pBitmap = &wxNullBitmap);
    void   AddBitmapMenu(wxMenu *pMenu, int ID, wxString strMain, wxString strSub, 
			 wxString strFile);

    wxMenu *mpCMenuPovRay, *mpCMenuDialog;
    wxMenu *mpCMenuFilter, *mpCMenuPanel;
    wxMenu *mpCMenuEdit,   *mpCMenuBoxConf;
    wxMenu *mpCMenuHelp,   *mpCMenuFilterPrimitive;

    wxBitmap *mpBitmapPreviewConfig, *mpBitmapCameraConfig;
    wxBitmap *mpBitmapStereo,        *mpBitmapPovray;
    wxBitmap *mpBitmapPreference,    *mpBitmapScene;
    wxBitmap *mpBitmapTexture,       *mpBitmapTree;
    wxBitmap *mpBitmapQuit;
    
    wxMenuBar *mpCMenuBar;

    wxString mstrStatus[3];
    int    mStatusBarStack[3];
    void   PushStatusTextMain(wxString strStatus, int index=0);
    void   PopStatusTextMain(int index = 0);
    
    void   UpdateMenu();
    void   UpdateStatusBar();

    // ------

    GLCanvasCtrl_cls* GetPCGLCanvasCtrl(int index);
    MainGLCanvas_cls* GetPCGLCanvas(int index);
    void ReconnectCameraPathGUI();
    void SetMNumBoxInformation     (wxBoxSizer *pBoxSizer, wxWindow *pC);
    void SetMNumPlayOption         (wxBoxSizer *pBoxSizer, wxWindow *pC);
    void SetMNumDownSizing         (wxBoxSizer *pBoxSizer, wxWindow *pC);
    
    void SetMNumScene     ();
    MainGLCanvas_cls* mpCConnectedGLPanel;
    static MultiNumericalInputsPanel_cls* mpCMNumBoxInformation;
    static MultiNumericalInputsPanel_cls* mpCMNumPlayOption;
    static MultiNumericalInputsPanel_cls* mpCMNumDownSizing;
    //static MultiNumericalInputs_cls* mpCMNumRenderAnim;
    //void OnButtonMultipanels(wxCommandEvent &event);
    wxRect mCRectMNumScene;
    wxRect mCRectMNumDownSizing;
    wxRect mCRectMNumCopy;

    //FilterCtrl_cls* GetPFilterCtrlActive();

    // -----

    void OnPaint(wxPaintEvent& event);
    void OnCloseData(wxCommandEvent& event);
    void OnQuit (wxCommandEvent& event);
    void OnDebug();
    void OnAbout(wxCommandEvent& event);
    void OnOpen (wxCommandEvent& event);
    void OnAdd  (wxCommandEvent& event);
    void OnOpenEdit (wxCommandEvent& event);
    void OnAddEdit  (wxCommandEvent& event);
    void OnMenuPopStatusBar(wxCommandEvent& event);
    
    bool mbHasOpenedProject;
    void OnProjectOpen  (wxCommandEvent& event);
    void UpdateAfterProjectOpen();
    void OpenProjectMain(wxString strFilename);
    void OnProjectSave  (wxCommandEvent& event);
    void OnProjectSaveAs(wxCommandEvent& event);

    void OnImportPath   (wxCommandEvent& event);
    void OnImportRenderParam(wxCommandEvent& event);
    void OnExportDF3    (wxCommandEvent& event);
    void OnExportDF3Seq (wxCommandEvent& event);
    void OnExportImgSeq (wxCommandEvent& event);
    void OnExportIsosurfaceOBJ(wxCommandEvent& event);
    void OnExportIsosurfacePOV(wxCommandEvent& event);
    
    void OnSaveImage      (wxCommandEvent& event);
    void OnSaveImgSeq     (wxCommandEvent& event);
    void OnSaveImgSeqBegin(wxCommandEvent& event);
    void OnPlayTex        (wxCommandEvent& event);
    void OnPlayTexInv     (wxCommandEvent& event);
    void OnPlayTexStop    (wxCommandEvent& event);

    void OnRedo        (wxCommandEvent& event);
    void OnUndo        (wxCommandEvent& event);

    void OnMenuLightSource     (wxCommandEvent& event);
    void OnMenuHotSpot         (wxCommandEvent& event);
    void OnMenuPovRay          (wxCommandEvent& event);
    void OnMenuPovRayAdvanced  (wxCommandEvent& event);
    void OnMenuPovRayError     (wxCommandEvent& event);
    void OnMenuTree            (wxCommandEvent& event);
    void OnMenuStereogram      (wxCommandEvent& event);
    void OnMenuCameraPath      (wxCommandEvent& event);
    void OnMenuTimeRemap       (wxCommandEvent& event);
    void OnMenuGoFrame         (wxCommandEvent& event);
    void OnCommandPreviewConfig   (wxCommandEvent& event);
    void OnMenuSetAnchorPath   (wxCommandEvent& event);
    void OnMenuSetAnchorZero   (wxCommandEvent& event);

    void OnMenuFilterSwapUp    (wxCommandEvent& event);
    void OnMenuFilterSwapDown  (wxCommandEvent& event);
    void OnMenuIsosurface      (wxCommandEvent& event);
    void OnMenuCrossSection    (wxCommandEvent& event);
    void OnMenuThreeView       (wxCommandEvent& event);
    void OnMenuFog             (wxCommandEvent& event);
    void OnMenuBoxCopy         (wxCommandEvent& event);
    void OnMenuClipping        (wxCommandEvent& event);
    void WarningForMultiVolume ();
    void OnMenuVolume          (wxCommandEvent& event);
    void OnMenuBlurGauss       (wxCommandEvent& event);
    void OnMenuInvDiffuse      (wxCommandEvent& event);
    void OnMenuLevelAndGamma   (wxCommandEvent& event);
    void OnMenuFixPosition     (wxCommandEvent& event);
    void OnMenuDataRotate      (wxCommandEvent& event);
    void OnMenuLayerColorCurve (wxCommandEvent& event);
    void AddPovObject(int type);
    void OnMenuPovObjPlane     (wxCommandEvent& event);
    void OnMenuPovObjBox       (wxCommandEvent& event);
    void OnMenuPovObjWiredBox  (wxCommandEvent& event);
    void OnMenuPovObjSphere    (wxCommandEvent& event);
    void OnMenuPovObjCone      (wxCommandEvent& event);
    void OnMenuPovObjArrow     (wxCommandEvent& event);
    void OnMenuPovObjText      (wxCommandEvent& event);
    void OnMenuPovObjMesh      (wxCommandEvent& event);
    void OnMenuPovObjPlugin    (wxCommandEvent& event);
    void OnMenuCenterOfMass    (wxCommandEvent& event);
    void OnMenuMorphOpenClose  (wxCommandEvent& event);
    void OnMenuPovDiffObject   (wxCommandEvent& event);
    void OnMenuPovStreamLine   (wxCommandEvent& event);
    void OnMenuPovVelField     (wxCommandEvent& event);
    void OnMenuFilterAdvanced  (wxCommandEvent& event);
    
    void OnMenuExportPovRaySnap(wxCommandEvent& event);
    void OnMenuExportPovRayAnim(wxCommandEvent& event);
    void OnMenuRenderPovRaySnap(wxCommandEvent& event);
    void OnMenuRenderPovRayAnim(wxCommandEvent& event);
    void PovRaySnapMain(bool bRender);
    void OnMenuPovWine         (wxCommandEvent& event);
    void OnMenuAutoRescale     (wxCommandEvent& event);
    void OnMenuViewPanelType   (wxCommandEvent& event);
    void OnMenuViewExtra       (wxCommandEvent& event);
    void OnMenuToolTip         (wxCommandEvent& event);
    
    void OnMenuAddLight        (wxCommandEvent& event);
    void OnMenuAddHotSpot      (wxCommandEvent& event);
    void OnMenuTextureDLG      (wxCommandEvent& event);
    void OnMenuCustomColor     (wxCommandEvent& event);
    void OnMenuPreference      (wxCommandEvent& event);
    void OnMenuScene           (wxCommandEvent& event);
    void OnMenuSetCamPos       (wxCommandEvent& event);
    void OnMenuSetAncPos       (wxCommandEvent& event);
    void OnMenuRefresh         (wxCommandEvent& event);
    void OnMenuRefreshTL       (wxCommandEvent& event);
    void OnMenuRefreshViewList (wxCommandEvent& event);
    void OnMenuReTexture       (wxCommandEvent& event);
    void OnMenuReLoad          (wxCommandEvent& event);
    void OnMenuReLoadSnap      (wxCommandEvent& event);
    void OnMenuRefreshPanels   (wxCommandEvent& event);
    void OnMenuRefreshCameraNumerical(wxCommandEvent& event);
    void OnMenuRefreshCameraCombobox(wxCommandEvent& event);
    void OnMenuRefreshCombo    (wxCommandEvent& event);
    void OnMenuRefreshColorBar (wxCommandEvent& event);
    void OnMenuSetFocus        (wxCommandEvent& event);
    void OnMenuSplitterResize  (wxCommandEvent& event);
    void OnMenuDF3GUIChanged   (wxCommandEvent& event);
    void OnSash(wxCommandEvent& event);


    // ---------

    void OnKeyDown(wxKeyEvent& event);
    void OnKeyUp(wxKeyEvent& event);
    void OnFocusEventCustum     (wxFocusEvent & event);
    
    bool mbDragCtrlKey;
    void OnMouseMotion(wxMouseEvent& event);

    // -----

    void DeselectAllIf(wxString labelStr = wxT(""));
    void SetKeyPath(unsigned int type);

    static void FuncHookParamChanged(void *pArg, void *pEvent);
    static void FuncHookBoxInfoChanged(void *pArg, void *pEvent);
    static void FuncHookDownSizingChanged(void *pArg, void *pEvent);
    static void FuncHookMouseOnBoxConf(void *pArg, void *pEvent);
    static void FuncHookPanelLayout(void *pArg, void *pEvent);
    void OnCommandEventRefresh(wxCommandEvent &event);
    //void OnButtonDownSizeConf(wxCommandEvent &event);
    void OnButtonRefEditor(wxCommandEvent &event);
    
    // -----

    void GainDrawingBlockCopies
	(std::vector < DrawingBlock_cls > &DrawingBlockCopies,
	 bool bForceCheck = _CHECK_TREE_OMIT_FILE_CHECK,
	 bool bForceMake  = _CHECK_TREE_MAKE_EVEN_IF_FILE_NOT_EXIST);
    wxString GetPovStrObjects
    (std::vector < DrawingBlock_cls > &VCDrawingCls);
    /*
    void OutputPovStrObjectsDensity
    (std::wfstream &ofstream,
     std::vector < DrawingBlock_cls > &VCDrawingCls, 
     PositionAndDirection_cls CCamGlobalCoord, int frame);
    */

    void SaveConfigure(wxString strFileOut, bool flag);
    void ReadConfigure(std::wstring strFileIn);
    ConfItemSTR_cls      mCLastOpenedFileDF3;
    ConfItemSTR_cls      mCLastOpenedFileOSP;
    ConfItemSTR_cls      mCLastOpenedFileIMG;
    ConfItemSTR_cls      mCLastExportedFileDF3;
    ConfItemSTR_cls      mCLastExportedFileIMG;
    ConfItemUINT_cls     mCPreviewPanelType;
    
    // -----

    void InitializeOnIdle();
    bool mbIdle;
    bool mbPathChanged;
    clock_t mClockLastDraw;

    int  mMilliSleep;
    long int  mWaitDuration;

    static FrameMain_cls*            mpThis;

    // --- ChildDialogGeometry
    bool IsIntersectDialogs(wxRect rect);

    unsigned int mParamForDebugginng;
    
 private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

#endif
