// -*- c++ -*-
// (setq c-basic-offset 4)

#include "clsTimeCtrl.h"
#include "SRC/PARAMETER/clsParameters.h"
#include "SRC/PARAMETER/clsParametersPanel.h"
#include "SRC/PARAMETER/clsTimeLinePanel.h"
#include "SRC/DF3/clsDF3SeqCtrl.h"
#include "SRC/DF3/clsDF3SeqGUI.h"
#include "SRC/DF3/clsDF3Sequence.h"
#include "SRC/UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "SRC/WIDGETCTRL/clsAtoiTextCtrl.h"
#include "SRC/WIDGETCTRL/EX/clsSliderEX.h"


#ifdef WIN32
#define ENDLWX wxT("\r\n");
#else
#define ENDLWX wxT("\n");
#endif

TimeCtrl_cls::
TimeCtrl_cls() :
    mCTimeRemap(1, L"TimeRemap")
{
    mFrame = 0;
    mCustomRegionMin = 0;
    mCustomRegionMax = 99;

    mbPlayTex = FALSE; 
    mPlayMode = _TIME_CTRL_PLAY_FULL;
    mPlayDirection = +1;

    mCConfListTime.Add(&mCMaxFrame, L"DF3MaxFrame=", L"99");
    mCConfListTime.Add(&mCWAMin, L"DF3WorkAreaMin=", L"0");
    mCConfListTime.Add(&mCWAMax, L"DF3WorkAreaMax=", L"99");
    mCConfListTime.Add(&mCAAMin, L"DF3ActiveAreaMin=", L"0");
    mCConfListTime.Add(&mCAAMax, L"DF3ActiveAreaMax=", L"99");

    mCConfListTime.Add(&mCbUseTimeRemap, L"DF3SequenceUseTimeRemap=", L"0");
    mCConfListTime.Add(&mCStep, L"DF3SequenceStep=", L"1");
    mCConfListTime.Add(&mCbRepeat, L"DF3SequenceRepeat=", L"0");

    mCConfListTime.Add(&mCSceneLengthType, L"DF3SequenceSceneLengthType=", L"0");

    double initZero = 0;
    mCTimeRemap.ExecCommandInitSet(&initZero);

    TimeLinePanel_cls::msVBSnippingTag.resize(99);
}

void TimeCtrl_cls::
DispatchTimeData(){
    Parameters_cls::SetFrameNumber(mCMaxFrame.mBody, mCStep.mBody);
    ParametersPanel_cls::SetDoubleIndex
	(mCMaxFrame.mBody, mCAAMin.mBody, mCAAMax.mBody, 
	 GetDoubleIndex(),
	 GetDoubleIndexAAMin(), GetDoubleIndexAAMax(),
	 GetDoubleIndexWAMin(), GetDoubleIndexWAMax());

    //mspCDF3SeqCtrlLink->mpCGUI->mpCParent->UpdatePreFetchList();
}

void TimeCtrl_cls::
UpdateDF3SeqGUI(bool bUpdateParent){
    DispatchTimeData();
    mCTimeRemap.GetPInterpolatedValue(GetDoubleIndex());
    mspCDF3SeqCtrlLink->mpCGUI->UpdateDisplay(bUpdateParent);
}

void TimeCtrl_cls::
UpdateDF3SeqGUISlider(){
    int tickFreq = GetInterval(mCAAMax.mBody - mCAAMin.mBody);
    mspCDF3SeqCtrlLink->mpCGUI->mpCCurrentIndex->SetValueInt(mFrame);
}

void TimeCtrl_cls::
SetWorkingArea(int maxIndex){
    mCWAMin.mBody = mCAAMin.mBody = 0;
    mCWAMax.mBody = mCAAMax.mBody = maxIndex;
    Parameters_cls::SetFrameNumber(mCMaxFrame.mBody, 
				   mCStep.mBody);
}

void TimeCtrl_cls::
CheckWorkingArea(){
    if (mCWAMax.mBody > mCMaxFrame.mBody)
	mCWAMax.mBody = mCMaxFrame.mBody;
    if (mCWAMin.mBody > mCWAMax.mBody){
	unsigned int max = mCWAMax.mBody;
	mCWAMax.mBody = mCWAMin.mBody;
	mCWAMin.mBody = max;
    }

    if (mCAAMax.mBody > mCMaxFrame.mBody)
	mCAAMax.mBody = mCMaxFrame.mBody;
    if (mCAAMin.mBody > mCAAMax.mBody){
	unsigned int max = mCAAMax.mBody;
	mCAAMax.mBody = mCAAMin.mBody;
	mCAAMin.mBody = max;
    }
    
    if (mFrame < mCAAMin.mBody) 
	mFrame = mCAAMin.mBody;
    if (mFrame > mCAAMax.mBody)
	mFrame = mCAAMax.mBody;

    mFrame = CalcNextStep(0);

    UpdateDF3SeqGUISlider();
}

void TimeCtrl_cls::
RemapParameterKeys(double rate){
    Parameters_cls::SetFrameNumber(mCMaxFrame.mBody, mCStep.mBody);
    GCUndoList.OpenBlock();
    mspCDF3SeqCtrlLink->mpCGUI->mpCPanelList->RemapKeys(rate);
    GCUndoList.CloseBlock();
}

void TimeCtrl_cls::
SetBUseTimeRemap(bool bUse){
    mCbUseTimeRemap.mbBody = bUse;
    if (mCbUseTimeRemap.mbBody == TRUE){
	// \[Xłlast = (double)mMaxIndexǖȂH
	double init = 0, last = (double)mCMaxFrame.mBody;//, zero = 0;

	mCTimeRemap.mCbSelectable.mbBody = TRUE;
	mspCDF3SeqCtrlLink->mpCGUI->ShowItem(wxT("[Time Remap]"));
	
	GCUndoList.OpenBlock();
	mCTimeRemap.InitSet(&init);

	KeyFrame_cls CKey(1);//, CKeySlave(0);
	CKey.Set(GetDoubleIndex(last), _PARAM_HERMITE_AUTO, &last);
	mCTimeRemap.SetKey(CKey);
	
	/*
	mCTimeRemap.SetKey(GetDoubleIndex(last), &last, &zero, &zero, _PARAM_HERMITE_AUTO,
			   NULL, NULL, NULL);
	*/
	GCUndoList.CloseBlock();
    }
    else{
	mCTimeRemap.mCbSelectable.mbBody = FALSE;
	mspCDF3SeqCtrlLink->mpCGUI->ShowItem(wxT("[Time Remap]"), FALSE);
    }
    mspCDF3SeqCtrlLink->mpCGUI->UpdateItemsVisivility();
    UpdateDF3SeqGUI(TRUE);
}

unsigned int TimeCtrl_cls::
GetRemappedIndex(int frame){
    if (frame == _TIME_CTRL_CURRENT) frame = mFrame;
    int remappedIndex;
    if (mCbUseTimeRemap.mbBody == FALSE)
	remappedIndex = frame;
    else
	remappedIndex = (int)(*mCTimeRemap.GetPInterpolatedValue
			      (GetDoubleIndex(frame)) + 1e-5);
    if (remappedIndex < 0) remappedIndex = 0;
    return remappedIndex;
}

double TimeCtrl_cls::
GetRemappedIndexDouble(int frame){
    if (frame == _TIME_CTRL_CURRENT) frame = mFrame;
    double remappedIndex;
    if (mCbUseTimeRemap.mbBody == FALSE)
	remappedIndex = frame;
    else
	remappedIndex = (*mCTimeRemap.GetPInterpolatedValue(GetDoubleIndex(frame)));
    if (remappedIndex < 0) remappedIndex = 0;
    return remappedIndex;
}

int TimeCtrl_cls::
GetInterval(int frameNumber){
    int interval = 1;
    if (frameNumber > 1000)     interval = 10;
    else if (frameNumber > 500) interval = 5;
    else if (frameNumber > 200) interval = 2;
    return interval;
}

double TimeCtrl_cls::
GetDoubleIndex(int frame){
    if (mCMaxFrame.mBody == 0) return 0;
    if (frame == -1)
	return (double)mFrame / (mCMaxFrame.mBody);
    return (double)frame / (mCMaxFrame.mBody);
}

double TimeCtrl_cls::
GetDoubleIndexWAMin(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCWAMin.mBody / mCMaxFrame.mBody;
}

double TimeCtrl_cls::
GetDoubleIndexWAMax(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCWAMax.mBody / mCMaxFrame.mBody;
}

double TimeCtrl_cls::
GetDoubleIndexAAMin(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCAAMin.mBody / mCMaxFrame.mBody;
}

double TimeCtrl_cls::
GetDoubleIndexAAMax(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCAAMax.mBody / mCMaxFrame.mBody;
}

int TimeCtrl_cls::
GetOffsetForImgSeq(){
    return (mFrame - mCWAMin.mBody) / mCStep.mBody;
}

unsigned int TimeCtrl_cls::
CalcNextStep(int increment, bool bForceWA, bool bCheckRepeat){
    unsigned int nextStep = mFrame + mCStep.mBody * increment;
    
    if (nextStep > mCMaxFrame.mBody){
	if (increment < 0) 
	    if (bCheckRepeat && mCbRepeat.mbBody)
		nextStep = mCMaxFrame.mBody;
	    else
		nextStep = 0;
	else
	    if (bCheckRepeat && mCbRepeat.mbBody)
		nextStep = 0;
	    else
		nextStep = mCMaxFrame.mBody;
    }
    
    if ((increment != 0) && (bForceWA == TRUE)){
	if (nextStep > mCWAMax.mBody) nextStep = mCWAMax.mBody; 
	if (nextStep < mCWAMin.mBody) nextStep = mCWAMin.mBody;
    }
    
    nextStep /= mCStep.mBody;
    nextStep *= mCStep.mBody;
    return nextStep;
}

unsigned int TimeCtrl_cls::
GetIndexFromDouble(double doubleIndex){
    if (mCMaxFrame.mBody == 0) return 0;
    return (doubleIndex * mCMaxFrame.mBody) + 0.1;
}

bool TimeCtrl_cls::
IsWorkingArea(unsigned int frame){
    if (frame < mCWAMin.mBody) return FALSE;
    if (frame > mCWAMax.mBody) return FALSE;
    if (frame > mCMaxFrame.mBody) return FALSE;

    return TRUE;
}

void TimeCtrl_cls::
SetIndexWithinWA(){
    if (mFrame > mCWAMax.mBody) mFrame = mCWAMax.mBody;
    if (mFrame < mCWAMin.mBody) mFrame = mCWAMin.mBody;
}

void TimeCtrl_cls::
SetMaxFrame(unsigned int maxFrame){
    mCMaxFrame.mBody = maxFrame;
    if (mFrame > maxFrame) mFrame = 0;

    CheckAfterMaxFrameChange();
}

void TimeCtrl_cls::
CheckAfterMaxFrameChange(){
    static unsigned int prevMaxFrame = 99;

    if (mCMaxFrame.mBody != prevMaxFrame){
	if (prevMaxFrame == mCAAMax.mBody)
	    mCAAMax.mBody = mCMaxFrame.mBody;
	if (prevMaxFrame == mCWAMax.mBody)
	    mCWAMax.mBody = mCMaxFrame.mBody;
	if (mCAAMax.mBody > mCMaxFrame.mBody)
	    mCAAMax.mBody = mCMaxFrame.mBody;
	prevMaxFrame = mCMaxFrame.mBody;
    }

    mCustomRegionMin = 0; 
    mCustomRegionMax = mCMaxFrame.mBody;

    TimeLinePanel_cls::msVBSnippingTag.resize(mCMaxFrame.mBody);
}

bool TimeCtrl_cls::
OpenCustomRegionDialog(wxWindow *pC){
    MultiNumericalInputs_cls *pCurr;

    pCurr = new MultiNumericalInputs_cls (&pCurr, pC, wxID_ANY, wxT("Render Area"));
    pCurr->SetDefaultButtonType(_MN_OK_CANCEL_DEFAULT);
    pCurr->mTextWidth = 80;
    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("init/fin"));
    pCurr->CombinePointer(_GP_UINT_SPIN, &mCustomRegionMin);
    pCurr->SetRange(0, mCMaxFrame.mBody);
    pCurr->CombinePointer(_GP_UINT_SPIN, &mCustomRegionMax);
    pCurr->SetRange(0, mCMaxFrame.mBody);

    bool bRet = (pCurr->FinishSettingDLG(TRUE, _MN_DLG_MODAL, _MN_DLG_MAKEGEOM)
		 == wxID_OK);
    
    /*
    pCurr->FinishSettingDLG(FALSE);
    bool bRet = (pCurr->ShowModal() != wxID_CANCEL);
    */
    delete pCurr;
    
    return bRet;
}

void TimeCtrl_cls::
ZoomIn(){
    int widthR = mCAAMax.mBody - mFrame;
    int widthL = mFrame - mCAAMin.mBody;
    int center = mFrame;
    mCAAMax.mBody = center + widthR / 2;
    mCAAMin.mBody = center - widthL / 2;
    CheckWorkingArea();
    UpdateDF3SeqGUI();
}

void TimeCtrl_cls::
ZoomOut(){
    int widthR = mCAAMax.mBody - mFrame; if (widthR == 0) widthR =1;
    int widthL = mFrame - mCAAMin.mBody; if (widthL == 0) widthL =1;
    int center = mFrame;
    mCAAMax.mBody = center + widthR * 2;
    if ((center - widthL * 2) < 0)
	mCAAMin.mBody = 0;
    else
	mCAAMin.mBody = center - widthL * 2;
    CheckWorkingArea();
    UpdateDF3SeqGUI();
}

void TimeCtrl_cls::
EmptySnippingTag(){
    if (TimeLinePanel_cls::msVBSnippingTag.size() != mCMaxFrame.mBody + 1)
	TimeLinePanel_cls::msVBSnippingTag.resize(mCMaxFrame.mBody + 1);
    for (unsigned int i = 0; 
	 i < TimeLinePanel_cls::msVBSnippingTag.size(); i++)
	TimeLinePanel_cls::msVBSnippingTag[i] = FALSE;
}

void TimeCtrl_cls::
SnipNext(){
    unsigned int i = mFrame + mCStep.mBody;
    mFrame = mCMaxFrame.mBody;
    for (; i < mCMaxFrame.mBody; i++)
	if (TimeLinePanel_cls::msVBSnippingTag[i] == TRUE) {
	    mFrame = i;  
	    break;
	}
    
    CheckWorkingArea();
    UpdateDF3SeqGUI();
}

void TimeCtrl_cls::
SnipPrev(){
    int i = mFrame - mCStep.mBody;
    mFrame = 0;
    for (; i > 0; i--)
	if (TimeLinePanel_cls::msVBSnippingTag[i] == TRUE) {
	    mFrame = i; 
	    break;
	}
    
    CheckWorkingArea();
    UpdateDF3SeqGUI();
}

void TimeCtrl_cls::
SetWAInHere(){
    mCWAMin.mBody = mFrame;
    CheckWorkingArea();
    DispatchTimeData();
    mspCDF3SeqCtrlLink->mpCGUI->UpdateDisplay();
    mspCDF3SeqCtrlLink->mpCGUI->UpdatePanelsBitmapState();
}

void TimeCtrl_cls::
SetWAOutHere(){
    mCWAMax.mBody = mFrame;
    CheckWorkingArea();
    DispatchTimeData();
    mspCDF3SeqCtrlLink->mpCGUI->UpdateDisplay();
    mspCDF3SeqCtrlLink->mpCGUI->UpdatePanelsBitmapState();
}

void TimeCtrl_cls::
SetIndexFromDouble(double doubleIndex){
    double index = GetIndexFromDouble(doubleIndex);
    mFrame = (unsigned int)index;
}

void TimeCtrl_cls::
SetIndex(unsigned int index){
    mFrame = index;
    mFrame = CalcNextStep(0);
    CheckWorkingArea();
}

void TimeCtrl_cls::
FrameStep(int step, bool bForceWA){
    mFrame = CalcNextStep(step, bForceWA);
    UpdateDF3SeqGUI(TRUE);
}

void TimeCtrl_cls::
FrameMoveL(int step){
    unsigned int nextStep = CalcNextStep(step, TRUE, FALSE);
    bool condition = mspCDF3SeqCtrlLink->IsPlayTexConditionSatisfied(nextStep);
    while (mspCDF3SeqCtrlLink->IsPlayTexConditionSatisfied(nextStep) ==
	   condition){
	mFrame = nextStep;
	nextStep = CalcNextStep(step, TRUE, FALSE);
	if (nextStep == mFrame) break;
	if (condition == FALSE) mFrame = nextStep;
    }
    SetIndexWithinWA();
}

// -----

// _łSubset_FramegĂȂ̂Ŗgp
wxString TimeCtrl_cls::
GetPovStrLength(){
    wxString strPov;

    strPov << wxT("Initial_Frame=0") << ENDLWX;
    strPov << wxT("Final_Frame=") << mCMaxFrame.mBody << ENDLWX;
    strPov << wxT("Subset_Start_Frame=") << mCWAMin.mBody << ENDLWX;
    strPov << wxT("Subset_End_Frame=") << mCWAMax.mBody << ENDLWX;

    return strPov;
}

void TimeCtrl_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    mCConfListTime.ReadStrLineAll(strLine);
    std::wstring::size_type loc = strLine.find(L"--TimeRemap");
    if (loc != std::wstring::npos){
	std::wstring strLine;
	std::getline(istream, strLine);
	mCTimeRemap.ReadStrLine(istream, strLine);
    }
    
    Parameters_cls::SetFrameNumber(mCMaxFrame.mBody, mCStep.mBody);
}

wxString TimeCtrl_cls::
SerializeParam(){
    wxString sOut;
    sOut << mCConfListTime.SerializeAll();
    if (mCbUseTimeRemap.mbBody == TRUE){
	sOut << wxT("--TimeRemap\n");
	sOut << mCTimeRemap.Serialize();
    }
    return sOut;
}

//DF3SeqGUI_cls* TimeCtrl_cls::mspCDF3SeqGUI = NULL;
DF3SeqCtrl_cls* TimeCtrl_cls::mspCDF3SeqCtrlLink = NULL;
